reference = {
    "journal": {
        "values": {
            "settings": {
                "appGeocoders": [
                    {
                        "singleLineFieldName": "SingleLine",
                        "name": "Esri World Geocoder",
                        "url": "https://geocode.arcgis.com/arcgis/rest/services/World/GeocodeServer",
                    },
                    {
                        "singleLineFieldName": "SingleLine",
                        "name": "san",
                        "placeholder": "",
                        "url": "https://sampleserver5.arcgisonline.com/arcgis/rest/services/Locators/SanDiego/GeocodeServer",
                    },
                ],
                "layout": {"id": "side"},
            },
            "title": "title",
            "story": {"storage": "WEBAPP", "sections": []},
            "template": {
                "name": "Map Journal",
                "createdWith": "1.14.1",
                "editedWith": "1.14.1",
            },
        }
    }
}

storymap_2 = {
    "root": "n-4xkUEe",
    "nodes": {
        "n-4xkUEe": {
            "type": "story",
            "data": {"storyTheme": "r-vlc4Kp"},
            "config": {"coverDate": "first-published"},
            "children": ["n-aTn8ak", "n-1AItUD", "n-cOeTah"],
        },
        "n-aTn8ak": {
            "type": "storycover",
            "data": {
                "type": "minimal",
                "title": "",
                "summary": "",
                "byline": "",
                "titlePanelPosition": "start",
            },
        },
        "n-1AItUD": {
            "type": "navigation",
            "data": {"links": []},
            "config": {"isHidden": True},
        },
        "n-cOeTah": {"type": "credits"},
    },
    "resources": {
        "r-vlc4Kp": {
            "type": "story-theme",
            "data": {"themeId": "summit", "themeBaseVariableOverrides": {}},
        }
    },
}
