templates = {
    "blank fullscreen": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "FIT_WINDOW",
                "type": "NORMAL",
                "layout": {"LARGE": "layout_0"},
                "isDefault": True,
            }
        },
        "layouts": {"layout_0": {"type": "FIXED", "content": {}, "order": []}},
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": False},
        "template": "blankfullscreen",
        "attributes": {"portalUrl": ""},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "widgets": {},
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dialogs": {},
        "dataSources": {},
        "messageConfigs": {},
    },
    "blank scrolling": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {"LARGE": "layout_0"},
                "isDefault": True,
            }
        },
        "layouts": {"layout_0": {"type": "FLOW", "content": {}, "order": []}},
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": False},
        "template": "blankscrollable",
        "attributes": {"portalUrl": ""},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "widgets": {},
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dialogs": {},
        "dataSources": {},
        "messageConfigs": {},
    },
    "foldable": {
        "pages": {
            "home": {
                "type": "NORMAL",
                "mode": "FIT_WINDOW",
                "layout": {
                    "LARGE": "home-layout-large",
                    "SMALL": "layout_5",
                    "MEDIUM": "layout_10",
                },
                "isVisible": True,
                "id": "home",
                "isDefault": True,
                "header": False,
            }
        },
        "layouts": {
            "home-layout-large": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "71.3715%",
                            "height": "60px",
                            "right": "0px",
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "STRETCH"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "60px",
                            "width": "31.2500%",
                            "height": "52%",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_6",
                        "setting": {
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"}
                        },
                    },
                },
            },
            "layout_1": {
                "type": "FIXED",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "14px",
                            "top": "50%",
                            "width": "36px",
                            "height": "36px",
                        },
                        "widgetId": "widget_2",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "90px",
                            "top": "50%",
                            "width": "279px",
                            "height": "42px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "vCenter": True,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "392px",
                            "top": "50%",
                            "width": "210px",
                            "height": "32px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_4",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "889px",
                            "top": "0px",
                            "width": "300px",
                            "height": "50px",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_5",
                        "setting": {"autoProps": {"left": True, "height": "STRETCH"}},
                    },
                },
            },
            "layout_2": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_8"},
                    "1": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_9"},
                },
            },
            "layout_3": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_8",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_4": {"type": "FIXED"},
            "layout_5": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "71.3715%",
                            "height": "60px",
                            "right": "0px",
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "STRETCH"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "60px",
                            "width": "31.2500%",
                            "height": "52%",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_6",
                        "setting": {
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"}
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_6": {"type": "FIXED"},
            "layout_7": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "14px",
                            "top": "50%",
                            "width": "36px",
                            "height": "36px",
                        },
                        "widgetId": "widget_2",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "90px",
                            "top": "3px",
                            "width": "279px",
                            "height": "48px",
                            "bottom": "1px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "autoProps": {"height": "STRETCH"},
                            "hCenter": False,
                        },
                        "isPending": True,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "392px",
                            "top": "0px",
                            "width": "210px",
                            "height": "48px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_4",
                        "isPending": True,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "STRETCH"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "282px",
                            "top": "0px",
                            "width": "200px",
                            "height": "46px",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {"left": True, "height": "STRETCH"},
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_8": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_8"},
                    "1": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_9"},
                },
            },
            "layout_9": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_8",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_10": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "71.3715%",
                            "height": "60px",
                            "right": "0px",
                        },
                        "widgetId": "widget_10",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "STRETCH"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "60px",
                            "width": "31.2500%",
                            "height": "52%",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_6",
                        "setting": {
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"}
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_11": {"type": "FIXED"},
            "layout_12": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "14px",
                            "top": "50%",
                            "width": "36px",
                            "height": "36px",
                        },
                        "widgetId": "widget_2",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "90px",
                            "top": "50%",
                            "width": "279px",
                            "height": "42px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "vCenter": True,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "392px",
                            "top": "50%",
                            "width": "210px",
                            "height": "32px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_4",
                        "isPending": True,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "889px",
                            "top": "0px",
                            "width": "300px",
                            "height": "50px",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_5",
                        "setting": {"autoProps": {"left": True, "height": "STRETCH"}},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_13": {
                "type": "FIXED",
                "content": {
                    "0": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_8"},
                    "1": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_9"},
                },
                "order": ["0", "1"],
            },
            "layout_14": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_8",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_15": {"type": "FIXED"},
        },
        "widgets": {
            "widget_1": {
                "id": "widget_1",
                "uri": "widgets/layout/fixed/",
                "config": {},
                "layouts": {"DEFAULT": {"LARGE": "layout_1"}},
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--primary-700)"},
                    "boxShadow": {
                        "offsetX": {"distance": 0, "unit": "px"},
                        "offsetY": {"distance": 2, "unit": "px"},
                        "blur": {"distance": 4, "unit": "px"},
                        "spread": {"distance": 0, "unit": "px"},
                        "color": "rgba(0,0,0,0.5)",
                    },
                },
                "version": "1.5.0",
            },
            "widget_2": {
                "id": "widget_2",
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "cropShape": "circle",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "svgViewBox": "0 0 28 28",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "version": "1.5.0",
            },
            "widget_3": {
                "id": "widget_3",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px; color: rgb(255, 255, 255);">Here is the title</span></p>',
                    "style": {
                        "verticalAlign": "center",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "version": "1.5.0",
            },
            "widget_4": {
                "id": "widget_4",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: rgb(255, 255, 255);">Here is the subtitle</span></p>',
                    "style": {
                        "verticalAlign": "center",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "version": "1.5.0",
            },
            "widget_5": {
                "id": "widget_5",
                "uri": "widgets/common/controller/",
                "config": {
                    "behavior": {
                        "onlyOpenOne": True,
                        "displayType": "STACK",
                        "vertical": False,
                        "size": {},
                    },
                    "appearance": {
                        "space": 0,
                        "advanced": False,
                        "card": {
                            "showLabel": False,
                            "labelGrowth": 10,
                            "avatar": {
                                "type": "primary",
                                "size": "default",
                                "shape": "rectangle",
                            },
                        },
                    },
                },
                "layouts": {
                    "controller": {
                        "LARGE": "layout_2",
                        "SMALL": "layout_8",
                        "MEDIUM": "layout_13",
                    },
                    "_openwidget": {
                        "LARGE": "layout_3",
                        "SMALL": "layout_9",
                        "MEDIUM": "layout_14",
                    },
                },
                "widgets": ["widget_8", "widget_9"],
                "version": "1.5.0",
            },
            "widget_6": {
                "id": "widget_6",
                "uri": "widgets/arcgis/arcgis-map/",
                "config": {
                    "toolConifg": {"canZoom": True, "canHome": True, "canSearch": True},
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 36978595.474472,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -10491677.231363693,
                                "y": 4628100.902328416,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -15584217.803833922,
                            "ymin": 1683135.0765579282,
                            "xmax": -5399136.658893464,
                            "ymax": 7573066.728098903,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_4",
                        "SMALL": "layout_6",
                        "MEDIUM": "layout_11",
                    }
                },
                "version": "1.5.0",
            },
            "widget_7": {
                "id": "widget_7",
                "uri": "widgets/layout/fixed/",
                "config": {},
                "layouts": {"DEFAULT": {"SMALL": "layout_7"}},
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--primary-700)"},
                    "boxShadow": {
                        "offsetX": {"distance": 0, "unit": "px"},
                        "offsetY": {"distance": 2, "unit": "px"},
                        "blur": {"distance": 4, "unit": "px"},
                        "spread": {"distance": 0, "unit": "px"},
                        "color": "rgba(0,0,0,0.5)",
                    },
                },
                "version": "1.5.0",
            },
            "widget_8": {
                "id": "widget_8",
                "uri": "widgets/arcgis/map-layers/",
                "config": {"useMapWidget": True},
                "useMapWidgetIds": ["widget_6"],
                "version": "1.5.0",
            },
            "widget_9": {
                "id": "widget_9",
                "uri": "widgets/arcgis/legend/",
                "config": {},
                "useMapWidgetIds": ["widget_6"],
                "version": "1.5.0",
            },
            "widget_10": {
                "id": "widget_10",
                "uri": "widgets/layout/fixed/",
                "config": {},
                "layouts": {"DEFAULT": {"MEDIUM": "layout_12"}},
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--primary-700)"},
                    "boxShadow": {
                        "offsetX": {"distance": 0, "unit": "px"},
                        "offsetY": {"distance": 2, "unit": "px"},
                        "blur": {"distance": 4, "unit": "px"},
                        "spread": {"distance": 0, "unit": "px"},
                        "color": "rgba(0,0,0,0.5)",
                    },
                },
                "version": "1.5.0",
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "foldable",
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dialogs": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"home": []}],
        "timestamp": 1620599410273,
        "header": {"layout": {"LARGE": "layout_15"}, "height": {"LARGE": 75}},
        "attributes": {"portalUrl": ""},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "launchpad": {
        "pages": {
            "home": {
                "type": "NORMAL",
                "mode": "FIT_WINDOW",
                "layout": {"LARGE": "home-layout-large", "SMALL": "layout_54"},
                "isVisible": True,
                "id": "home",
                "isDefault": True,
                "header": False,
            }
        },
        "layouts": {
            "home-layout-large": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "39.0720%",
                            "height": "52.1173%",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_23",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                            }
                        },
                    }
                },
            },
            "layout_51": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "top": "706px",
                            "width": "400px",
                            "height": "60px",
                            "bottom": "25px",
                        },
                        "widgetId": "widget_24",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": True, "left": False, "right": True},
                            "hCenter": True,
                            "vCenter": False,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "top": "0px",
                            "width": "360px",
                            "height": "46px",
                        },
                        "widgetId": "widget_25",
                        "setting": {
                            "lockParent": True,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                    },
                },
            },
            "layout_52": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_8"},
                    "1": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_9"},
                },
            },
            "layout_53": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_9",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_54": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "39.0720%",
                            "height": "52.1173%",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_23",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                            }
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_55": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "top": "706px",
                            "width": "400px",
                            "height": "60px",
                            "bottom": "1px",
                        },
                        "widgetId": "widget_24",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": True, "left": False, "right": True},
                            "hCenter": True,
                            "vCenter": False,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "top": "0px",
                            "width": "360px",
                            "height": "46px",
                        },
                        "widgetId": "widget_25",
                        "isPending": True,
                        "setting": {
                            "lockParent": True,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_56": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_8"},
                    "1": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_9"},
                },
            },
            "layout_57": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_9",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_58": {"type": "FIXED"},
        },
        "widgets": {
            "widget_23": {
                "id": "widget_23",
                "uri": "widgets/arcgis/arcgis-map/",
                "config": {
                    "toolConifg": {"canZoom": True, "canHome": True, "canSearch": True},
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 18489297.737236,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -12286049.3093898,
                                "y": 9276508.600679886,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -14641532.773025164,
                            "ymin": 7804025.687794642,
                            "xmax": -9930565.845754435,
                            "ymax": 10748991.513565129,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "layouts": {
                    "MapFixedLayout": {"LARGE": "layout_51", "SMALL": "layout_55"}
                },
                "version": "1.5.0",
            },
            "widget_24": {
                "id": "widget_24",
                "uri": "widgets/common/controller/",
                "config": {
                    "behavior": {
                        "onlyOpenOne": True,
                        "displayType": "STACK",
                        "vertical": False,
                        "size": {},
                    },
                    "appearance": {
                        "space": 0,
                        "advanced": False,
                        "card": {
                            "showLabel": False,
                            "labelGrowth": 10,
                            "avatar": {
                                "type": "primary",
                                "size": "default",
                                "shape": "circle",
                            },
                        },
                    },
                },
                "layouts": {
                    "controller": {"LARGE": "layout_52", "SMALL": "layout_56"},
                    "_openwidget": {"LARGE": "layout_53", "SMALL": "layout_57"},
                },
                "widgets": ["widget_8", "widget_9"],
                "version": "1.5.0",
            },
            "widget_25": {
                "id": "widget_25",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": "<p style='text-align: center;'><span style='font-size: 24px; color: rgb(255, 255, 255);'>Here is the title</span></p>",
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--primary-600)"}
                },
                "version": "1.5.0",
            },
            "widget_8": {
                "id": "widget_8",
                "uri": "widgets/arcgis/map-layers/",
                "config": {"useMapWidget": True},
                "useMapWidgetIds": ["widget_23"],
                "version": "1.5.0",
            },
            "widget_9": {
                "id": "widget_9",
                "uri": "widgets/arcgis/legend/",
                "config": {},
                "useMapWidgetIds": ["widget_23"],
                "version": "1.5.0",
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "launchpad",
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dialogs": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"home": []}],
        "timestamp": 1620599623072,
        "header": {"layout": {"LARGE": "layout_58"}, "height": {"LARGE": 75}},
        "attributes": {"portalUrl": ""},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "jewelrybox": {
        "pages": {
            "home": {
                "type": "NORMAL",
                "mode": "FIT_WINDOW",
                "layout": {
                    "LARGE": "layout_1",
                    "SMALL": "layout_17",
                    "MEDIUM": "layout_21",
                },
                "isVisible": True,
                "isDefault": True,
                "id": "home",
                "header": False,
            }
        },
        "layouts": {
            "layout_1": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "39.072%",
                            "height": "60px",
                            "right": "0px",
                        },
                        "widgetId": "widget_1",
                        "setting": {"autoProps": {"width": "STRETCH"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "60px",
                            "width": "78.1441%",
                            "height": "52.1173%",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_6",
                        "setting": {
                            "autoProps": {"height": "STRETCH", "width": "STRETCH"}
                        },
                    },
                },
            },
            "layout_2": {
                "type": "FIXED",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "14px",
                            "top": "50%",
                            "width": "36px",
                            "height": "36px",
                        },
                        "setting": {
                            "vCenter": True,
                            "autoProps": {"top": False, "bottom": True},
                        },
                        "widgetId": "widget_2",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "82px",
                            "top": "50%",
                            "width": "291px",
                            "height": "42px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "380px",
                            "top": "50%",
                            "width": "207px",
                            "height": "32px",
                            "bottom": "0px",
                            "right": "NaNpx",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "hCenter": False,
                            "vCenter": True,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "728px",
                            "top": "0px",
                            "width": "300px",
                            "height": "50px",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_5",
                        "setting": {"autoProps": {"left": True, "height": "STRETCH"}},
                    },
                },
            },
            "layout_3": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_14"},
                    "1": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_15"},
                },
            },
            "layout_4": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_14"},
                    "1": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_15"},
                },
            },
            "layout_5": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_14",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_6": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_7": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "298px",
                            "height": "350px",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                            }
                        },
                    }
                },
            },
            "layout_8": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "110px",
                            "height": "133px",
                            "right": "10px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_8",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "10px",
                            "width": "465px",
                            "height": "29px",
                            "right": "20px",
                        },
                        "widgetId": "widget_9",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": "STRETCH"},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "34px",
                            "width": "253px",
                            "height": "46px",
                            "right": "20px",
                            "bottom": "20px",
                        },
                        "widgetId": "widget_10",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_9": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "110px",
                            "height": "133px",
                            "right": "10px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_8",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "10px",
                            "width": "465px",
                            "height": "29px",
                            "right": "20px",
                        },
                        "widgetId": "widget_9",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": "STRETCH"},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "34px",
                            "width": "253px",
                            "height": "46px",
                            "right": "20px",
                            "bottom": "20px",
                        },
                        "widgetId": "widget_10",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_10": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "110px",
                            "height": "133px",
                            "right": "10px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_8",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "10px",
                            "width": "465px",
                            "height": "29px",
                            "right": "20px",
                        },
                        "widgetId": "widget_9",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": "STRETCH"},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "34px",
                            "width": "253px",
                            "height": "46px",
                            "right": "20px",
                            "bottom": "20px",
                        },
                        "widgetId": "widget_10",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_11": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "110px",
                            "height": "133px",
                            "right": "10px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_8",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "10px",
                            "width": "465px",
                            "height": "29px",
                            "right": "20px",
                        },
                        "widgetId": "widget_9",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": "STRETCH"},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "34px",
                            "width": "253px",
                            "height": "46px",
                            "right": "20px",
                            "bottom": "20px",
                        },
                        "widgetId": "widget_10",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_12": {"type": "FIXED"},
            "layout_13": {"type": "FIXED"},
            "layout_14": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "356px",
                            "height": "298px",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_11",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
            },
            "layout_15": {"type": "FIXED"},
            "layout_16": {"type": "FIXED"},
            "layout_17": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "39.072%",
                            "height": "60px",
                            "right": "0px",
                        },
                        "widgetId": "widget_12",
                        "setting": {"autoProps": {"width": "STRETCH"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "60px",
                            "width": "78.1441%",
                            "height": "52.1173%",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_13",
                        "setting": {
                            "autoProps": {"height": "STRETCH", "width": "STRETCH"}
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_18": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "14px",
                            "top": "50%",
                            "width": "36px",
                            "height": "36px",
                        },
                        "setting": {
                            "vCenter": True,
                            "autoProps": {"top": False, "bottom": True},
                        },
                        "widgetId": "widget_2",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "82px",
                            "top": "2px",
                            "width": "291px",
                            "height": "50px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "364px",
                            "top": "0px",
                            "width": "207px",
                            "height": "50px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "243px",
                            "top": "0px",
                            "width": "200px",
                            "height": "50px",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {"left": True, "height": "STRETCH"},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_19": {
                "type": "FIXED",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                            },
                        },
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "356px",
                            "height": "298px",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_10",
                        "isPending": True,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": "0", "top": "0", "right": "0", "bottom": "0"},
                        "widgetId": "widget_11",
                        "setting": {
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"}
                        },
                    },
                },
                "order": ["1", "2"],
            },
            "layout_20": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                            },
                        },
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "298px",
                            "height": "350px",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_7",
                    }
                },
                "order": ["0"],
            },
            "layout_21": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "39.072%",
                            "height": "60px",
                            "right": "0px",
                        },
                        "widgetId": "widget_16",
                        "setting": {"autoProps": {"width": "STRETCH"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "60px",
                            "width": "78.1441%",
                            "height": "52.1173%",
                            "right": "0px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_17",
                        "setting": {
                            "autoProps": {"height": "STRETCH", "width": "STRETCH"}
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_22": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "14px",
                            "top": "50%",
                            "width": "36px",
                            "height": "36px",
                        },
                        "setting": {
                            "vCenter": True,
                            "autoProps": {"top": False, "bottom": True},
                        },
                        "widgetId": "widget_2",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "82px",
                            "top": "50%",
                            "width": "291px",
                            "height": "42px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "380px",
                            "top": "50%",
                            "width": "207px",
                            "height": "32px",
                            "bottom": "0px",
                            "right": "NaNpx",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "hCenter": False,
                            "vCenter": True,
                            "autoProps": {"top": False, "bottom": True},
                        },
                        "isPending": True,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "370px",
                            "top": 0,
                            "width": "230px",
                            "height": "60px",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_5",
                        "setting": {"autoProps": {"left": True, "height": "STRETCH"}},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_23": {
                "type": "FIXED",
                "content": {
                    "0": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_14"},
                    "1": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_15"},
                },
                "order": ["0", "1"],
            },
            "layout_24": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_14",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_25": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "298px",
                            "height": "350px",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                            }
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_26": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "110px",
                            "height": "133px",
                            "right": "10px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_8",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "10px",
                            "width": "465px",
                            "height": "29px",
                            "right": "20px",
                        },
                        "widgetId": "widget_9",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": "STRETCH"},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "34px",
                            "width": "253px",
                            "height": "46px",
                            "right": "20px",
                            "bottom": "20px",
                        },
                        "widgetId": "widget_10",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_27": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "0px",
                            "width": "110px",
                            "height": "133px",
                            "right": "10px",
                            "bottom": "0px",
                        },
                        "widgetId": "widget_8",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "10px",
                            "width": "465px",
                            "height": "29px",
                            "right": "20px",
                        },
                        "widgetId": "widget_9",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": "STRETCH"},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "125px",
                            "top": "34px",
                            "width": "253px",
                            "height": "46px",
                            "right": "20px",
                            "bottom": "20px",
                        },
                        "widgetId": "widget_10",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"},
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_28": {"type": "FIXED"},
            "layout_29": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "356px",
                            "height": "298px",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_11",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                            }
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_30": {"type": "FIXED"},
        },
        "widgets": {
            "widget_1": {
                "id": "widget_1",
                "uri": "widgets/layout/fixed/",
                "config": {},
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--primary-600)"},
                    "boxShadow": {
                        "offsetX": {"distance": 0, "unit": "px"},
                        "offsetY": {"distance": 2, "unit": "px"},
                        "blur": {"distance": 4, "unit": "px"},
                        "spread": {"distance": 0, "unit": "px"},
                        "color": "rgba(0,0,0,0.5)",
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_2"}},
                "version": "1.5.0",
            },
            "widget_2": {
                "id": "widget_2",
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "cropShape": "circle",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "svgViewBox": "0 0 28 28",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "version": "1.5.0",
            },
            "widget_3": {
                "id": "widget_3",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 24px; color: var(--light);">Here is the title</span></p>',
                    "style": {
                        "verticalAlign": "center",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "version": "1.5.0",
            },
            "widget_4": {
                "id": "widget_4",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--light); font-size: 16px;">Here is the subtitle</span></p>',
                    "style": {
                        "verticalAlign": "center",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "version": "1.5.0",
            },
            "widget_5": {
                "id": "widget_5",
                "uri": "widgets/common/controller/",
                "config": {
                    "behavior": {
                        "onlyOpenOne": True,
                        "displayType": "STACK",
                        "vertical": False,
                        "size": {},
                    },
                    "appearance": {
                        "space": 0,
                        "advanced": False,
                        "card": {
                            "showLabel": False,
                            "labelGrowth": 10,
                            "avatar": {
                                "type": "primary",
                                "size": "default",
                                "shape": "rectangle",
                            },
                        },
                    },
                },
                "layouts": {
                    "controller": {
                        "LARGE": "layout_3",
                        "SMALL": "layout_4",
                        "MEDIUM": "layout_23",
                    },
                    "_openwidget": {
                        "LARGE": "layout_5",
                        "SMALL": "layout_6",
                        "MEDIUM": "layout_24",
                    },
                },
                "widgets": ["widget_14", "widget_15"],
                "version": "1.5.0",
            },
            "widget_6": {
                "id": "widget_6",
                "uri": "widgets/layout/sidebar/",
                "config": {
                    "direction": "HORIZONTAL",
                    "collapseSide": "FIRST",
                    "overlay": False,
                    "size": "398px",
                    "divider": {"visible": True},
                    "resizable": True,
                    "toggleBtn": {
                        "visible": True,
                        "icon": "LEFT",
                        "offsetX": 15,
                        "offsetY": 0,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 15,
                        "height": 40,
                        "color": {
                            "normal": {
                                "icon": {
                                    "useTheme": True,
                                    "color": "var(--primary-700)",
                                },
                                "bg": {
                                    "useTheme": True,
                                    "color": "var(--light-400)",
                                    "opacity": 0.6,
                                },
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--light-400)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "0 4px 4px 0"}},
                        "collapseStyle": {"style": {"borderRadius": "0 4px 4px 0"}},
                        "template": "rect",
                    },
                    "defaultState": 1,
                },
                "layouts": {
                    "FIRST": {"LARGE": "layout_7"},
                    "SECOND": {"LARGE": "layout_14"},
                },
                "version": "1.5.0",
            },
            "widget_7": {
                "id": "widget_7",
                "uri": "widgets/common/list/",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE4",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "isInitialed": True,
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {
                                "LARGE": {"width": 395.984375, "height": 110},
                                "MEDIUM": {"width": 395.984375, "height": 110},
                                "SMALL": {"width": 356.015625, "height": 110},
                            },
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0, 0, 0, 1)",
                                    "width": {"distance": 1, "unit": "px"},
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {
                                "LARGE": {"width": 395.984375, "height": 110},
                                "MEDIUM": {"width": 395.984375, "height": 110},
                                "SMALL": {"width": 356.015625, "height": 110},
                            },
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0, 0, 0, 1)",
                                    "width": {"distance": 1, "unit": "px"},
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {
                                "LARGE": {"width": 395.984375, "height": 110},
                                "MEDIUM": {"width": 395.984375, "height": 110},
                                "SMALL": {"width": 356.015625, "height": 110},
                            },
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary)",
                                    "width": {"distance": 1, "unit": "px"},
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                    "isItemStyleConfirm": True,
                },
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_8",
                        "SMALL": "layout_9",
                        "MEDIUM": "layout_26",
                    },
                    "SELECTED": {
                        "LARGE": "layout_10",
                        "SMALL": "layout_11",
                        "MEDIUM": "layout_27",
                    },
                    "HOVER": {
                        "LARGE": "layout_12",
                        "SMALL": "layout_13",
                        "MEDIUM": "layout_28",
                    },
                },
                "version": "1.5.0",
            },
            "widget_8": {
                "id": "widget_8",
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "version": "1.5.0",
            },
            "widget_9": {
                "id": "widget_9",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "",
                    "placeholder": '<h5><span style="font-size: 16px;">Double click to edit text</span></h5>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "version": "1.5.0",
            },
            "widget_10": {
                "id": "widget_10",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "",
                    "placeholder": '<p style=""><span style="font-size: 13px;">Double click to edit text</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "version": "1.5.0",
            },
            "widget_11": {
                "id": "widget_11",
                "uri": "widgets/arcgis/arcgis-map/",
                "config": {
                    "toolConifg": {"canZoom": True, "canHome": True, "canSearch": True},
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 18489297.737236,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -10970311.977193125,
                                "y": 4800743.33460336,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -12795016.716416366,
                            "ymin": 3269556.783995116,
                            "xmax": -9145607.237969883,
                            "ymax": 6331929.885211604,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_15",
                        "SMALL": "layout_16",
                        "MEDIUM": "layout_30",
                    }
                },
                "version": "1.5.0",
            },
            "widget_12": {
                "id": "widget_12",
                "uri": "widgets/layout/fixed/",
                "config": {},
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--primary)"},
                    "boxShadow": {
                        "offsetX": {"distance": 0, "unit": "px"},
                        "offsetY": {"distance": 2, "unit": "px"},
                        "blur": {"distance": 4, "unit": "px"},
                        "spread": {"distance": 0, "unit": "px"},
                        "color": "rgba(0,0,0,0.5)",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_18"}},
                "version": "1.5.0",
            },
            "widget_13": {
                "id": "widget_13",
                "uri": "widgets/layout/sidebar/",
                "config": {
                    "direction": "VERTICAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "140px",
                    "divider": {"visible": True},
                    "resizable": True,
                    "toggleBtn": {
                        "visible": True,
                        "icon": "DOWN",
                        "offsetX": 0,
                        "offsetY": -15,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 40,
                        "height": 15,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": True, "color": "colors.primary"},
                                "bg": {
                                    "useTheme": True,
                                    "color": "var(--light-400)",
                                    "opacity": 0.6,
                                },
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--light-400)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "4px 4px 0 0"}},
                        "collapseStyle": {"style": {"borderRadius": "4px 4px 0 0"}},
                        "template": "rect",
                    },
                    "defaultState": 1,
                },
                "layouts": {
                    "FIRST": {"SMALL": "layout_19"},
                    "SECOND": {"SMALL": "layout_20"},
                },
                "version": "1.5.0",
            },
            "widget_14": {
                "id": "widget_14",
                "uri": "widgets/arcgis/map-layers/",
                "config": {"useMapWidget": True},
                "useMapWidgetIds": ["widget_11"],
                "version": "1.5.0",
            },
            "widget_15": {
                "id": "widget_15",
                "uri": "widgets/arcgis/legend/",
                "config": {},
                "useMapWidgetIds": ["widget_11"],
                "version": "1.5.0",
            },
            "widget_16": {
                "id": "widget_16",
                "uri": "widgets/layout/fixed/",
                "config": {},
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--primary)"},
                    "boxShadow": {
                        "offsetX": {"distance": 0, "unit": "px"},
                        "offsetY": {"distance": 2, "unit": "px"},
                        "blur": {"distance": 4, "unit": "px"},
                        "spread": {"distance": 0, "unit": "px"},
                        "color": "rgba(0,0,0,0.5)",
                    },
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_22"}},
                "version": "1.5.0",
            },
            "widget_17": {
                "id": "widget_17",
                "uri": "widgets/layout/sidebar/",
                "config": {
                    "direction": "HORIZONTAL",
                    "collapseSide": "FIRST",
                    "overlay": False,
                    "size": "398px",
                    "divider": {"visible": True},
                    "resizable": True,
                    "toggleBtn": {
                        "visible": True,
                        "icon": "LEFT",
                        "offsetX": 15,
                        "offsetY": 0,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 15,
                        "height": 40,
                        "color": {
                            "normal": {
                                "icon": {
                                    "useTheme": True,
                                    "color": "var(--primary-700)",
                                },
                                "bg": {
                                    "useTheme": True,
                                    "color": "var(--light-400)",
                                    "opacity": 0.6,
                                },
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--light-400)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "0 4px 4px 0"}},
                        "collapseStyle": {"style": {"borderRadius": "0 4px 4px 0"}},
                        "template": "rect",
                    },
                    "defaultState": 1,
                },
                "layouts": {
                    "FIRST": {"MEDIUM": "layout_25"},
                    "SECOND": {"MEDIUM": "layout_29"},
                },
                "version": "1.5.0",
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/morandi/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "jewelrybox",
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"home": []}],
        "timestamp": 1622674876738,
        "attributes": {"portalUrl": ""},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "billboard": {
        "pages": {
            "home": {
                "type": "NORMAL",
                "mode": "FIT_WINDOW",
                "layout": {"LARGE": "layout_0", "SMALL": "layout_5"},
                "isVisible": True,
                "id": "home",
                "isDefault": True,
                "header": False,
            }
        },
        "layouts": {
            "layout_0": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "31.2500%",
                            "height": "50.0000%",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                            }
                        },
                    }
                },
            },
            "layout_1": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "47px",
                            "top": "0px",
                            "width": "200px",
                            "height": "62px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                    }
                },
            },
            "layout_2": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_8"},
                    "1": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_9"},
                },
            },
            "layout_3": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_9",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_5": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "31.2500%",
                            "height": "50.0000%",
                            "right": 0,
                            "bottom": 0,
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                            }
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_6": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "14px",
                            "top": "648px",
                            "width": "303px",
                            "height": "62px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_7": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_8"},
                    "1": {"type": "WIDGET", "bbox": {}, "widgetId": "widget_9"},
                },
            },
            "layout_9": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_8",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
        },
        "widgets": {
            "widget_1": {
                "id": "widget_1",
                "uri": "widgets/arcgis/arcgis-map/",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                        "canBaseMap": True,
                        "canMeasure": True,
                        "canFullScreen": True,
                        "canScaleBar": False,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 18489297.737236,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -11041175.861734187,
                                "y": 4754179.003170503,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -13492052.736669427,
                            "ymin": 3222992.4525622595,
                            "xmax": -8590298.986798948,
                            "ymax": 6285365.553778747,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "layouts": {
                    "MapFixedLayout": {"LARGE": "layout_1", "SMALL": "layout_6"}
                },
                "useDataSources": [],
                "version": "1.5.0",
            },
            "widget_2": {
                "id": "widget_2",
                "uri": "widgets/common/controller/",
                "config": {
                    "behavior": {
                        "onlyOpenOne": True,
                        "displayType": "STACK",
                        "vertical": False,
                        "size": {},
                    },
                    "appearance": {
                        "space": 0,
                        "advanced": False,
                        "card": {
                            "showLabel": False,
                            "labelGrowth": 10,
                            "avatar": {
                                "type": "primary",
                                "size": "default",
                                "shape": "rectangle",
                            },
                        },
                    },
                },
                "layouts": {
                    "controller": {"LARGE": "layout_2", "SMALL": "layout_7"},
                    "_openwidget": {"LARGE": "layout_3", "SMALL": "layout_9"},
                },
                "widgets": ["widget_8", "widget_9"],
                "useDataSources": [],
                "useDataSourcesEnabled": True,
                "version": "1.5.0",
            },
            "widget_8": {
                "id": "widget_8",
                "uri": "widgets/arcgis/map-layers/",
                "config": {"useMapWidget": True},
                "useMapWidgetIds": ["widget_1"],
                "version": "1.5.0",
            },
            "widget_9": {
                "id": "widget_9",
                "uri": "widgets/arcgis/legend/",
                "config": {},
                "useMapWidgetIds": ["widget_1"],
                "version": "1.5.0",
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/vivid/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "billboard",
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"home": []}],
        "timestamp": 1622673850666,
        "attributes": {"portalUrl": ""},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "journey": {
        "pages": {
            "page_1": {
                "id": "page_1",
                "mode": "FIT_WINDOW",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_98",
                    "MEDIUM": "layout_378",
                    "SMALL": "layout_385",
                },
                "isVisible": True,
                "isDefault": True,
                "oneByOneEffect": {"type": "FADE_IN", "option": {"direction": None}},
                "header": False,
            }
        },
        "layouts": {
            "layout_98": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "62.5000%",
                            "height": "50.0000%",
                        },
                        "widgetId": "widget_145",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_99": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_146",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
            },
            "layout_100": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "-4.1667%",
                            "top": "50%",
                            "bottom": "28.3208%",
                            "width": "100.0000%",
                            "height": "100.00%",
                        },
                        "setting": {
                            "hCenter": True,
                            "vCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
            },
            "layout_101": {"type": "FIXED", "order": [], "content": {}},
            "layout_181": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_331",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
            },
            "layout_182": {
                "type": "COLUMN",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_332",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_333",
                    },
                },
                "setting": {"space": 0},
            },
            "layout_183": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": "0px", "width": "11px", "height": "70px"},
                        "widgetId": "widget_516",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_378": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "62.5000%",
                            "height": "50.0000%",
                        },
                        "widgetId": "widget_517",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_379": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_146",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_380": {"type": "FIXED"},
            "layout_381": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "-4.1667%",
                            "top": "50%",
                            "bottom": "28.3208%",
                            "width": "100.0000%",
                            "height": "100.00%",
                        },
                        "setting": {
                            "hCenter": True,
                            "vCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_382": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_518",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_383": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_519",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_333",
                    },
                },
                "order": ["0", "1"],
                "setting": {"space": 0},
            },
            "layout_384": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": "0px", "width": "11px", "height": "70px"},
                        "widgetId": "widget_516",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_385": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "62.5000%",
                            "height": "50.0000%",
                        },
                        "widgetId": "widget_520",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_386": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_146",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_387": {"type": "FIXED"},
            "layout_388": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "-4.1667%",
                            "top": "50%",
                            "bottom": "28.3208%",
                            "width": "100.0000%",
                            "height": "100.00%",
                        },
                        "setting": {
                            "hCenter": True,
                            "vCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_389": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_521",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_390": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_522",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_333",
                    },
                },
                "order": ["0", "1"],
                "setting": {"space": 0},
            },
            "layout_391": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_516",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/dark/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "journey",
        "widgets": {
            "widget_145": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "HORIZONTAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "400px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "RIGHT",
                        "offsetX": -15,
                        "offsetY": 0,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 15,
                        "height": 60,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 0 0 92px"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 0 0 92px"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_145",
                "layouts": {
                    "FIRST": {"LARGE": "layout_99"},
                    "SECOND": {"LARGE": "layout_100"},
                },
                "style": {
                    "background": {
                        "color": "var(--primary-500)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_146": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                        "canLocate": True,
                        "canCompass": True,
                        "canLayers": True,
                        "canBaseMap": True,
                        "canMeasure": True,
                        "canFullScreen": True,
                        "canScaleBar": True,
                        "canSelect": False,
                    },
                    "isUseCustomMapState": True,
                    "layoutIndex": 0,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 18489297.737236,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -11094987.529646927,
                                "y": 4587852.029622002,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -12777825.144372921,
                            "ymin": 3056665.479013758,
                            "xmax": -9412149.914920934,
                            "ymax": 6119038.580230245,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_146",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_101",
                        "MEDIUM": "layout_380",
                        "SMALL": "layout_387",
                    }
                },
            },
            "widget_331": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_331",
                "layouts": {"DEFAULT": {"LARGE": "layout_182"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_332": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_332",
                "layouts": {"DEFAULT": {"LARGE": "layout_183"}},
            },
            "widget_333": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "description": "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -11830275.262593094,
                                "ymin": 286735.42243510764,
                                "xmax": -4462968.728356624,
                                "ymax": 6988734.062477579,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -8146621.995474859,
                                    "y": 3637734.7424563435,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_333editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "description": "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -5619597.781539567,
                                "ymin": -380400.1638864768,
                                "xmax": 1747708.752696902,
                                "ymax": 6321598.476155994,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -1935944.5144213326,
                                    "y": 2970599.156134759,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_333editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "description": "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -4203662.698719097,
                                "ymin": -3579108.914078116,
                                "xmax": 3163643.8355173715,
                                "ymax": 3122889.7259643557,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -520009.4316008631,
                                    "y": -228109.59405688022,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_333editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": True,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "PUSH", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_333",
                "style": {
                    "background": {
                        "color": "var(--primary-100)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
                "useMapWidgetIds": ["widget_146"],
            },
            "widget_516": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="color: var(--light); font-size: 18px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_516",
            },
            "widget_517": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "VERTICAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "400px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "DOWN",
                        "offsetX": 0,
                        "offsetY": -15,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 60,
                        "height": 15,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_517",
                "layouts": {
                    "FIRST": {"MEDIUM": "layout_379"},
                    "SECOND": {"MEDIUM": "layout_381"},
                },
                "style": {
                    "background": {
                        "color": "var(--primary-500)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_518": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_518",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_383"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_519": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_519",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_384"}},
            },
            "widget_520": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "VERTICAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "320px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "DOWN",
                        "offsetX": 0,
                        "offsetY": -15,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 60,
                        "height": 15,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_520",
                "layouts": {
                    "FIRST": {"SMALL": "layout_386"},
                    "SECOND": {"SMALL": "layout_388"},
                },
                "style": {
                    "background": {
                        "color": "var(--primary-500)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_521": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_521",
                "layouts": {"DEFAULT": {"SMALL": "layout_390"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_522": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_522",
                "layouts": {"DEFAULT": {"SMALL": "layout_391"}},
            },
        },
        "widgetsManifest": {},
        "views": {
            "view_4": {
                "id": "view_4",
                "layout": {
                    "LARGE": "layout_181",
                    "MEDIUM": "layout_382",
                    "SMALL": "layout_389",
                },
            }
        },
        "sections": {"section_2": {"id": "section_2", "views": ["view_4"]}},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_1": []}],
        "timestamp": 1622674575065,
        "attributes": {"portalUrl": ""},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "ribbon": {
        "pages": {
            "page_1": {
                "id": "page_1",
                "mode": "FIT_WINDOW",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_98",
                    "MEDIUM": "layout_147",
                    "SMALL": "layout_151",
                },
                "isVisible": True,
                "isDefault": True,
                "oneByOneEffect": {"type": "FADE_IN", "option": {"direction": None}},
                "header": False,
            }
        },
        "layouts": {
            "layout_98": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "62.5000%",
                            "height": "50.0000%",
                        },
                        "widgetId": "widget_145",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_99": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_146",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
            },
            "layout_100": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "240px",
                            "top": 0,
                            "bottom": "-19px",
                            "width": "100.00%",
                            "height": "230px",
                        },
                        "widgetId": "widget_147",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "CUSTOM",
                                "width": "CUSTOM",
                            },
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    }
                },
            },
            "layout_101": {"type": "FIXED", "order": [], "content": {}},
            "layout_102": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": True,
                        },
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "100.0000%",
                            "height": "38.5645%",
                            "right": "0.0000%",
                            "bottom": "61.4355%",
                        },
                        "widgetId": "widget_149",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "26.8500%",
                            "top": "40.6216%",
                            "width": 463,
                            "height": 69,
                            "right": 269,
                            "bottom": "50.7534%",
                        },
                        "widgetId": "widget_150",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "31px",
                            "top": "50.2930%",
                            "width": "93.8000%",
                            "height": "45.7402%",
                            "right": "31px",
                            "bottom": "32px",
                        },
                        "widgetId": "widget_151",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_144": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "41.67%",
                            "height": "100.00%",
                            "right": 278,
                            "bottom": "1px",
                        },
                        "widgetId": "widget_220",
                        "isPending": True,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 28,
                            "width": "63.0252%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_221",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_145": {
                "type": "FIXED",
                "content": {
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                                "height": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": True,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "50%",
                            "top": "486px",
                            "width": "840.00px",
                            "height": "35.3801%",
                            "right": "79px",
                            "bottom": "133px",
                        },
                        "widgetId": "widget_226",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "AUTO",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "44.8777%",
                            "top": "60.00px",
                            "width": 100,
                            "height": 53,
                            "right": 427,
                            "bottom": "89.7751%",
                        },
                        "widgetId": "widget_227",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.5656%",
                            "top": "120.00px",
                            "bottom": "87.8191%",
                            "width": 100,
                            "height": 16,
                        },
                        "widgetId": "widget_228",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "29px",
                            "top": "180.00px",
                            "bottom": "58.5395%",
                            "width": "840.00px",
                            "height": "260.00px",
                        },
                        "widgetId": "widget_230",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "39.5299%",
                            "top": "89.4319%",
                            "bottom": "40.00px",
                            "width": "23.0126%",
                            "height": "5.2301%",
                        },
                        "widgetId": "widget_234",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["2", "3", "4", "6", "7"],
            },
            "layout_146": {
                "type": "ROW",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": 300},
                        "widgetId": "widget_231",
                        "setting": {"heightMode": "fit"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": 300},
                        "widgetId": "widget_232",
                        "setting": {"heightMode": "fit"},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": 300},
                        "widgetId": "widget_233",
                        "setting": {"heightMode": "fit"},
                    },
                },
            },
            "layout_147": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "62.5000%",
                            "height": "50.0000%",
                        },
                        "widgetId": "widget_235",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_148": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_146",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_149": {"type": "FIXED", "content": {}, "order": []},
            "layout_150": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "83.3333%",
                            "height": "46.9925%",
                        },
                        "widgetId": "widget_147",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_151": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "62.5000%",
                            "height": "50.0000%",
                        },
                        "widgetId": "widget_236",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_152": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_146",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_153": {"type": "FIXED", "content": {}, "order": []},
            "layout_154": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "83.3333%",
                            "height": "46.9925%",
                        },
                        "widgetId": "widget_147",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    }
                },
                "order": ["0"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "widgets": {
            "widget_145": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "VERTICAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "211px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "DOWN",
                        "offsetX": 0,
                        "offsetY": -15,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 60,
                        "height": 15,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_145",
                "layouts": {
                    "FIRST": {"LARGE": "layout_99"},
                    "SECOND": {"LARGE": "layout_100"},
                },
            },
            "widget_146": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                        "canLocate": True,
                        "canCompass": True,
                        "canLayers": True,
                        "canBaseMap": True,
                        "canMeasure": True,
                        "canFullScreen": True,
                        "canScaleBar": True,
                        "canSelect": False,
                    },
                    "isUseCustomMapState": True,
                    "layoutIndex": 0,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 18489297.737236,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -11114555.408887926,
                                "y": 4812882.640893511,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -14245416.087447913,
                            "ymin": 3377089.5015851418,
                            "xmax": -7983694.73032794,
                            "ymax": 6248675.780201879,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_146",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_101",
                        "MEDIUM": "layout_149",
                        "SMALL": "layout_153",
                    }
                },
            },
            "widget_147": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -15699632.289930115,
                                "ymin": 2454795.292715268,
                                "xmax": -5172113.258272158,
                                "ymax": 7258709.646380749,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -10435872.774101136,
                                    "y": 4856752.469548008,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_147editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -14753341.879759885,
                                "ymin": 4403977.732500888,
                                "xmax": -4225822.848101929,
                                "ymax": 9207892.086166369,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -9489582.363930907,
                                    "y": 6805934.909333629,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_147editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -12365296.242075045,
                                "ymin": 6021882.09169864,
                                "xmax": -1837777.2104170881,
                                "ymax": 10825796.445364121,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7101536.726246066,
                                    "y": 8423839.26853138,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_147editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 4,
                            "name": "Bookmark-4",
                            "title": "Bookmark-4",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -10571089.877137579,
                                "ymin": 4544430.77197486,
                                "xmax": -43570.84547962248,
                                "ymax": 9348345.12564034,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -5307330.361308601,
                                    "y": 6946387.9488076,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_147editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 5,
                            "name": "Bookmark-5",
                            "title": "Bookmark-5",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -12365296.242075045,
                                "ymin": 6021882.09169864,
                                "xmax": -1837777.2104170881,
                                "ymax": 10825796.445364121,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7101536.726246066,
                                    "y": 8423839.26853138,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_147editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 6,
                            "name": "Bookmark-6",
                            "title": "Bookmark-6",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -16093665.170740096,
                                "ymin": 7368396.781969945,
                                "xmax": -5566146.139082139,
                                "ymax": 12172311.135635424,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -10829905.654911118,
                                    "y": 9770353.958802685,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_147editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 7,
                            "name": "Bookmark-7",
                            "title": "Bookmark-7",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -12365296.242075045,
                                "ymin": 6021882.09169864,
                                "xmax": -1837777.2104170881,
                                "ymax": 10825796.445364121,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7101536.726246066,
                                    "y": 8423839.26853138,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_147editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 8,
                            "name": "Bookmark-8",
                            "title": "Bookmark-8",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -17731919.293022163,
                                "ymin": 10583882.679512825,
                                "xmax": -7204400.261364207,
                                "ymax": 15387797.033178303,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -12468159.777193185,
                                    "y": 12985839.856345564,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_147editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "runtimeAddAllow": True,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder/",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                    "displayFromWeb": False,
                },
                "id": "widget_147",
                "useMapWidgetIds": ["widget_146"],
            },
            "widget_149": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_149",
            },
            "widget_150": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="text-align: center;"><strong style="color: var(--dark-800); font-size: 32px;">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_150",
            },
            "widget_151": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="text-align: center;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_151",
            },
            "widget_220": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_220",
            },
            "widget_221": {
                "uri": "widgets/survey123/",
                "version": "1.5.0",
                "config": {
                    "surveyItemId": "",
                    "portalUrl": "",
                    "embeds": ["fullScreen", "onSubmitted"],
                    "hides": ["navbar", "header", "description", "footer", "theme"],
                },
                "id": "widget_221",
            },
            "widget_226": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="text-align: center;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_226",
            },
            "widget_227": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 32px; color: var(--dark-800);">About</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 14px; color: var(--dark-800);">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_227",
            },
            "widget_228": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {"type": "Style0", "color": "#000", "size": "2px"},
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "Default"},
                },
                "id": "widget_228",
            },
            "widget_230": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_230",
                "layouts": {"DEFAULT": {"LARGE": "layout_146"}},
            },
            "widget_231": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_231",
            },
            "widget_232": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_232",
            },
            "widget_233": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_233",
            },
            "widget_234": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "transparent",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "#076fe5",
                                    "width": "1px",
                                },
                                "text": {"color": "#076fe5", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#076fe5", "size": 14},
                            },
                            "hover": {
                                "background": {"color": "#076fe5", "fillType": "fill"},
                                "border": {
                                    "type": "solid",
                                    "color": "#076fe5",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "#fff", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#fff", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_234",
            },
            "widget_235": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "VERTICAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "215px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "DOWN",
                        "offsetX": 0,
                        "offsetY": -15,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 60,
                        "height": 15,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_235",
                "layouts": {
                    "FIRST": {"MEDIUM": "layout_148"},
                    "SECOND": {"MEDIUM": "layout_150"},
                },
            },
            "widget_236": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "VERTICAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "211px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "DOWN",
                        "offsetX": 0,
                        "offsetY": -15,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 60,
                        "height": 15,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_236",
                "layouts": {
                    "FIRST": {"SMALL": "layout_152"},
                    "SECOND": {"SMALL": "layout_154"},
                },
            },
        },
        "template": "ribbon",
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_1": []}],
        "timestamp": 00000000000000,
        "attributes": {"portalUrl": ""},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "general": {
        "pages": {
            "home": {
                "type": "NORMAL",
                "mode": "AUTO_SCROLL",
                "layout": {"LARGE": "home-layout-large", "MEDIUM": "layout_357"},
                "isVisible": True,
                "id": "home",
                "icon": {
                    "svg": '<svg viewBox="0 0 10 12" xmlns="http://www.w3.org/2000/svg"><path d="M1 0h8a1 1 0 011 1v10a1 1 0 01-1 1H1a1 1 0 01-1-1V1a1 1 0 011-1zm0 1v10h8V1H1zm2 2h4a.5.5 0 010 1H3a.5.5 0 010-1zm0 2.5h4a.5.5 0 010 1H3a.5.5 0 010-1z" fill="#000" fill-rule="nonzero"></path></svg>',
                    "properties": {
                        "filename": None,
                        "originalName": None,
                        "path": None,
                        "color": "#fff",
                        "size": 16,
                        "inlineSvg": True,
                    },
                },
                "heightInBuilder": {"LARGE": 3313, "SMALL": 4899, "MEDIUM": 2937},
                "isDefault": True,
            }
        },
        "layouts": {
            "home-layout-large": {
                "type": "FLOW",
                "order": ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "355px",
                            "top": "-300px",
                            "width": "300px",
                            "height": "600px",
                        },
                        "widgetId": "widget_228",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "801px",
                            "top": "548px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_8",
                        "setting": {"heightMode": "auto"},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "4036px",
                            "width": "1280px",
                            "height": "328px",
                            "right": 1280,
                            "bottom": 4364.12109375,
                        },
                        "widgetId": "widget_538",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "713px",
                            "top": "976px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_41",
                    },
                    "5": {
                        "type": "WIDGET",
                        "widgetId": "widget_43",
                        "bbox": {
                            "left": "0px",
                            "top": "1139px",
                            "width": "1920px",
                            "height": 400,
                        },
                        "isPending": False,
                        "setting": {"heightMode": "auto"},
                    },
                    "6": {
                        "type": "WIDGET",
                        "widgetId": "widget_48",
                        "setting": {"heightMode": "auto"},
                        "bbox": {"height": 400},
                    },
                    "7": {
                        "type": "WIDGET",
                        "widgetId": "widget_53",
                        "setting": {"heightMode": "auto"},
                        "bbox": {"height": 400},
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "841px",
                            "top": "2618px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_58",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "164px",
                            "top": "3929px",
                            "width": "800px",
                            "height": 380,
                        },
                        "widgetId": "widget_549",
                        "setting": {"heightMode": "auto"},
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "3278px",
                            "width": "1920px",
                            "height": "620px",
                        },
                        "widgetId": "widget_213",
                        "isPending": False,
                        "setting": {"heightMode": "auto"},
                    },
                },
            },
            "layout_6": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": 250},
                        "widgetId": "widget_62",
                        "setting": {
                            "maintainedByLayout": True,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_20": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": 150},
                        "widgetId": "widget_42",
                        "setting": {
                            "autoHeight": True,
                            "style": {"alignSelf": "center"},
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_21": {
                "content": {
                    "0": {
                        "bbox": {"left": 2, "width": 3, "height": 400},
                        "type": "WIDGET",
                        "widgetId": "widget_44",
                        "setting": {"heightMode": "ratio", "aspectRatio": 1},
                    },
                    "1": {
                        "bbox": {"left": 5, "width": 5, "height": 135},
                        "type": "WIDGET",
                        "widgetId": "widget_45",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
                "type": "ROW",
            },
            "layout_22": {
                "content": {
                    "0": {
                        "bbox": {"width": 1102.15625, "height": 61.1015625},
                        "type": "WIDGET",
                        "widgetId": "widget_46",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_47",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"}
                        },
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_23": {
                "content": {
                    "0": {
                        "bbox": {"left": 2, "width": 3, "height": 400},
                        "type": "WIDGET",
                        "widgetId": "widget_49",
                        "setting": {"heightMode": "ratio", "aspectRatio": 1},
                    },
                    "1": {
                        "bbox": {"left": 5, "width": 5, "height": 135},
                        "type": "WIDGET",
                        "widgetId": "widget_50",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
                "type": "ROW",
            },
            "layout_24": {
                "content": {
                    "0": {
                        "bbox": {"width": 465.5, "height": 69.609375},
                        "type": "WIDGET",
                        "widgetId": "widget_51",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_52",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_25": {
                "content": {
                    "0": {
                        "bbox": {"left": 2, "width": 3, "height": 400},
                        "type": "WIDGET",
                        "widgetId": "widget_54",
                        "setting": {"heightMode": "ratio", "aspectRatio": 1},
                    },
                    "1": {
                        "bbox": {"left": 5, "width": 5, "height": 135},
                        "type": "WIDGET",
                        "widgetId": "widget_55",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
                "type": "ROW",
            },
            "layout_26": {
                "content": {
                    "0": {
                        "bbox": {"width": 943, "height": 64.7890625},
                        "type": "WIDGET",
                        "widgetId": "widget_56",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_57",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_27": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": 201},
                        "widgetId": "widget_212",
                        "setting": {
                            "maintainedByLayout": True,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_29": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 943, "height": 105.13671875},
                        "widgetId": "widget_9",
                        "setting": {
                            "autoProps": {"width": "STRETCH", "height": "AUTO"}
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "110px",
                            "width": "943px",
                            "height": "70px",
                        },
                        "widgetId": "widget_63",
                        "isPending": False,
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "aspectRatio": 0.22,
                        },
                    },
                },
                "type": "COLUMN",
            },
            "layout_100": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 943, "height": 95.21484375},
                        "widgetId": "widget_59",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "AUTO"}},
                        "bbox": {"width": 943, "height": "86.06px"},
                        "widgetId": "widget_61",
                    },
                },
                "type": "COLUMN",
            },
            "layout_101": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": 600},
                        "widgetId": "widget_214",
                        "setting": {
                            "autoHeight": True,
                            "style": {"alignSelf": "flex-start"},
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_102": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "194px",
                            "top": "120px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_215",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "4px",
                            "top": "289px",
                            "width": "800px",
                            "height": "400px",
                        },
                        "widgetId": "widget_216",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
            },
            "layout_103": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 2, "height": 50},
                        "widgetId": "widget_217",
                    }
                },
            },
            "layout_107": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 600},
                        "widgetId": "widget_229",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_108": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 1898, "height": 126.0859375},
                        "widgetId": "widget_230",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "aspectRatio": 0.2,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "535px",
                            "top": "284px",
                            "width": "800px",
                            "height": "400px",
                        },
                        "widgetId": "widget_231",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
            },
            "layout_109": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 2, "height": "50px"},
                        "widgetId": "widget_232",
                        "setting": {
                            "autoProps": {"height": "CUSTOM"},
                            "heightMode": "fixed",
                            "aspectRatio": "100:30",
                        },
                    }
                },
            },
            "layout_347": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": 400},
                        "widgetId": "widget_539",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_348": {
                "type": "ROW",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": 600},
                        "widgetId": "widget_540",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": 600},
                        "widgetId": "widget_541",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "8px",
                            "top": "NaNpx",
                            "width": "4px",
                            "height": "600px",
                        },
                        "widgetId": "widget_542",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
            },
            "layout_349": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 120, "height": 121.05859375},
                        "widgetId": "widget_543",
                        "isPending": False,
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "heightMode": "ratio",
                            "aspectRatio": "100:100",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "width": "224px",
                            "height": "165px",
                            "left": "26px",
                            "right": "250px",
                            "top": "188px",
                            "bottom": "302px",
                        },
                        "widgetId": "widget_544",
                        "isPending": False,
                        "setting": {
                            "heightMode": "fixed",
                            "aspectRatio": 0.74,
                            "autoProps": {"height": "AUTO", "width": "STRETCH"},
                        },
                    },
                },
                "type": "COLUMN",
            },
            "layout_350": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "120px", "height": "120px"},
                        "widgetId": "widget_545",
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-33px",
                            "top": "189px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_546",
                        "setting": {
                            "autoProps": {"height": "AUTO", "width": "STRETCH"}
                        },
                    },
                },
                "type": "COLUMN",
            },
            "layout_351": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "15px",
                            "top": "-107px",
                            "width": "120px",
                            "height": "120px",
                        },
                        "widgetId": "widget_547",
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-68px",
                            "top": "217px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_548",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
            },
            "layout_352": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": 400},
                        "widgetId": "widget_550",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_353": {
                "type": "ROW",
                "order": ["0", "3", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": 600},
                        "widgetId": "widget_551",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 8,
                            "top": "NaNpx",
                            "width": "4px",
                            "height": "600px",
                        },
                        "widgetId": "widget_553",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": 600},
                        "widgetId": "widget_631",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
            },
            "layout_354": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 263.546875, "height": 307.4609375},
                        "widgetId": "widget_554",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:100"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-14px",
                            "top": "198px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_555",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
            },
            "layout_356": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "1px",
                            "top": "2px",
                            "width": "300px",
                            "height": "300px",
                        },
                        "widgetId": "widget_558",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:100"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-66px",
                            "top": "256px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_559",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
            },
            "layout_357": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "355px",
                            "top": "-300px",
                            "width": "300px",
                            "height": 266,
                        },
                        "widgetId": "widget_560",
                        "setting": {"heightMode": "auto"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "801px",
                            "top": "548px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_563",
                        "setting": {"heightMode": "auto"},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "4036px",
                            "width": "1280px",
                            "height": 356,
                            "right": 1280,
                            "bottom": 4364.12109375,
                        },
                        "widgetId": "widget_577",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "713px",
                            "top": "976px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_565",
                    },
                    "4": {
                        "type": "WIDGET",
                        "widgetId": "widget_566",
                        "bbox": {
                            "left": "0px",
                            "top": "1139px",
                            "width": "1920px",
                            "height": 300,
                        },
                        "isPending": False,
                        "setting": {"heightMode": "auto"},
                    },
                    "5": {
                        "type": "WIDGET",
                        "widgetId": "widget_568",
                        "setting": {"heightMode": "auto"},
                        "bbox": {"height": 300},
                    },
                    "6": {
                        "type": "WIDGET",
                        "widgetId": "widget_570",
                        "setting": {"heightMode": "auto"},
                        "bbox": {"height": 300},
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "1444px",
                            "width": "600px",
                            "height": 179,
                            "right": 600,
                            "bottom": 1696.22265625,
                        },
                        "widgetId": "widget_572",
                        "isPending": False,
                        "setting": {"heightMode": "auto"},
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "164px",
                            "top": "3929px",
                            "width": "800px",
                            "height": 316,
                        },
                        "widgetId": "widget_582",
                        "setting": {"heightMode": "auto"},
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "1658px",
                            "width": "600px",
                            "height": 246,
                            "right": 600,
                            "bottom": 2056.60546875,
                        },
                        "widgetId": "widget_574",
                        "isPending": False,
                        "setting": {"heightMode": "auto"},
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
            },
            "layout_358": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 600},
                        "widgetId": "widget_561",
                        "setting": {
                            "heightMode": "fit",
                            "aspectRatio": 1.02,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_359": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 578, "height": 86.171875},
                        "widgetId": "widget_230",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "535px",
                            "top": "284px",
                            "width": "800px",
                            "height": "400px",
                        },
                        "widgetId": "widget_562",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
            "layout_360": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "50.00px"},
                        "widgetId": "widget_232",
                        "setting": {
                            "autoProps": {"height": "CUSTOM"},
                            "heightMode": "fixed",
                            "aspectRatio": "100:30",
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_361": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 159},
                        "widgetId": "widget_564",
                        "setting": {
                            "maintainedByLayout": True,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_362": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 943, "height": 105.13671875},
                        "widgetId": "widget_9",
                        "setting": {
                            "autoProps": {"width": "STRETCH", "height": "AUTO"},
                            "heightMode": "fit",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "110px",
                            "width": "943px",
                            "height": "70px",
                        },
                        "widgetId": "widget_63",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
            "layout_363": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 150},
                        "widgetId": "widget_42",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_364": {
                "type": "ROW",
                "content": {
                    "0": {
                        "bbox": {"left": 0, "width": 5, "height": 400},
                        "type": "WIDGET",
                        "widgetId": "widget_44",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:100"},
                    },
                    "1": {
                        "bbox": {"left": 5, "width": 7, "height": 135},
                        "type": "WIDGET",
                        "widgetId": "widget_567",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_365": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"width": 1102.15625, "height": 61.1015625},
                        "type": "WIDGET",
                        "widgetId": "widget_46",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_47",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_366": {
                "type": "ROW",
                "content": {
                    "0": {
                        "bbox": {"left": 0, "width": 5, "height": 400},
                        "type": "WIDGET",
                        "widgetId": "widget_49",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:100"},
                    },
                    "1": {
                        "bbox": {"left": 5, "width": 7, "height": 135},
                        "type": "WIDGET",
                        "widgetId": "widget_569",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_367": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"width": 465.5, "height": 69.609375},
                        "type": "WIDGET",
                        "widgetId": "widget_51",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_52",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_368": {
                "type": "ROW",
                "content": {
                    "0": {
                        "bbox": {"left": 0, "width": 5, "height": 400},
                        "type": "WIDGET",
                        "widgetId": "widget_54",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:100"},
                    },
                    "1": {
                        "bbox": {"left": 5, "width": 7, "height": 135},
                        "type": "WIDGET",
                        "widgetId": "widget_571",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_369": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"width": 943, "height": 64.7890625},
                        "type": "WIDGET",
                        "widgetId": "widget_56",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_57",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_370": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 201},
                        "widgetId": "widget_573",
                        "setting": {
                            "maintainedByLayout": True,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_371": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 578, "height": 61.83203125},
                        "widgetId": "widget_59",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "aspectRatio": 0.22,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {
                            "width": "578px",
                            "height": "133px",
                            "left": -1.23046875,
                            "right": 576.76953125,
                            "top": 66.40625,
                            "bottom": 199.40625,
                        },
                        "widgetId": "widget_61",
                        "isPending": False,
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
            "layout_372": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 378},
                        "widgetId": "widget_575",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_373": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-2px",
                            "top": "60px",
                            "width": "578px",
                            "height": "42px",
                            "right": 575.94921875,
                            "bottom": 102.40625,
                        },
                        "widgetId": "widget_215",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "4px",
                            "top": "289px",
                            "width": "800px",
                            "height": "400px",
                        },
                        "widgetId": "widget_576",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
            "layout_374": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": 50},
                        "widgetId": "widget_217",
                        "setting": {"heightMode": "fixed"},
                    }
                },
                "order": ["0"],
            },
            "layout_375": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 400},
                        "widgetId": "widget_578",
                        "setting": {"heightMode": "fit", "autoHeight": True},
                    }
                },
                "order": ["0"],
            },
            "layout_376": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": 600},
                        "widgetId": "widget_579",
                        "setting": {"heightMode": "fit", "autoHeight": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": 600},
                        "widgetId": "widget_580",
                        "setting": {"heightMode": "fit", "autoHeight": True},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "8px",
                            "top": "NaNpx",
                            "width": "4px",
                            "height": "600px",
                        },
                        "widgetId": "widget_581",
                        "setting": {"heightMode": "fit", "autoHeight": True},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_377": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 120, "height": 121.05859375},
                        "widgetId": "widget_543",
                        "isPending": False,
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "heightMode": "ratio",
                            "aspectRatio": "100:100",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "width": "132px",
                            "height": "102px",
                            "left": "26px",
                            "right": "250px",
                            "top": "188px",
                            "bottom": "302px",
                        },
                        "widgetId": "widget_544",
                        "isPending": False,
                        "setting": {
                            "heightMode": "fixed",
                            "aspectRatio": 0.77,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
            "layout_378": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "120px", "height": "120px"},
                        "widgetId": "widget_545",
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-33px",
                            "top": "189px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_546",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
            "layout_379": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "15px",
                            "top": "-107px",
                            "width": "120px",
                            "height": "120px",
                        },
                        "widgetId": "widget_547",
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-68px",
                            "top": "217px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_548",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
            "layout_380": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 400},
                        "widgetId": "widget_583",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_381": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": 600},
                        "widgetId": "widget_584",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 8,
                            "top": "NaNpx",
                            "width": "4px",
                            "height": "600px",
                        },
                        "widgetId": "widget_586",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": 600},
                        "widgetId": "widget_634",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "3", "2"],
            },
            "layout_382": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 263.546875, "height": 307.4609375},
                        "widgetId": "widget_554",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:100"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-14px",
                            "top": "198px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_555",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
            "layout_384": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "1px",
                            "top": "2px",
                            "width": "300px",
                            "height": "300px",
                        },
                        "widgetId": "widget_558",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:100"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-66px",
                            "top": "256px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_559",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
            "layout_430": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 263.546875, "height": 307.4609375},
                        "widgetId": "widget_632",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:100"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-14px",
                            "top": "198px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_633",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
            "layout_431": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 263.546875, "height": 307.4609375},
                        "widgetId": "widget_635",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:100"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_636",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                    },
                },
                "type": "COLUMN",
                "order": ["0", "1"],
            },
        },
        "widgets": {
            "widget_8": {
                "id": "widget_8",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [60, 10, 60, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_6"}},
                "version": "1.5.0",
            },
            "widget_9": {
                "id": "widget_9",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center; line-height: 2;"><span style="font-family: &quot;Avenir Next&quot;; font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_41": {
                "id": "widget_41",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [30, 10, 30, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_20"}},
                "version": "1.5.0",
            },
            "widget_42": {
                "id": "widget_42",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center; line-height: 4;"><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_43": {
                "id": "widget_43",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 40, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_21"}},
                "version": "1.5.0",
            },
            "widget_44": {
                "id": "widget_44",
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "version": "1.5.0",
            },
            "widget_45": {
                "id": "widget_45",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_22"}},
                "version": "1.5.0",
            },
            "widget_46": {
                "id": "widget_46",
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_47": {
                "id": "widget_47",
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_48": {
                "id": "widget_48",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 40, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_23"}},
                "version": "1.5.0",
            },
            "widget_49": {
                "id": "widget_49",
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "version": "1.5.0",
            },
            "widget_50": {
                "id": "widget_50",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_24"}},
                "version": "1.5.0",
            },
            "widget_51": {
                "id": "widget_51",
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_52": {
                "id": "widget_52",
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_53": {
                "id": "widget_53",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 40, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_25"}},
                "version": "1.5.0",
            },
            "widget_54": {
                "id": "widget_54",
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "version": "1.5.0",
            },
            "widget_55": {
                "id": "widget_55",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_26"}},
                "version": "1.5.0",
            },
            "widget_56": {
                "id": "widget_56",
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_57": {
                "id": "widget_57",
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_58": {
                "id": "widget_58",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [60, 10, 60, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_27"}},
                "version": "1.5.0",
            },
            "widget_59": {
                "id": "widget_59",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center; line-height: 2;"><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "useDataSourcesEnabled": False,
                "version": "1.5.0",
            },
            "widget_61": {
                "id": "widget_61",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;">Double click to edit text</p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_62": {
                "id": "widget_62",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_29"}},
                "version": "1.5.0",
            },
            "widget_63": {
                "id": "widget_63",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;">Double click to edit text</p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_212": {
                "id": "widget_212",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_100"}},
                "version": "1.5.0",
            },
            "widget_213": {
                "id": "widget_213",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [100, 10, 100, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_101"}},
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--light-300)"}
                },
                "version": "1.5.0",
            },
            "widget_214": {
                "id": "widget_214",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [0, 0, 0, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_102"}},
                "version": "1.5.0",
            },
            "widget_215": {
                "id": "widget_215",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_216": {
                "id": "widget_216",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 0,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_103"}},
                "version": "1.5.0",
            },
            "widget_217": {
                "id": "widget_217",
                "uri": "widgets/common/button/",
                "config": {
                    "functionConfig": {"toolTip": "", "linkParam": {}},
                    "styleConfig": {
                        "name": "Default",
                        "themeStyle": {"className": "d-block", "color": "primary"},
                        "customStyle": {"regular": {}, "hover": {}, "clicked": {}},
                    },
                },
                "version": "1.5.0",
            },
            "widget_228": {
                "id": "widget_228",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [100, 20, 100, 20], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_107"}},
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--light-300)"}
                },
                "version": "1.5.0",
            },
            "widget_229": {
                "id": "widget_229",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [20, 5, 20, 5], "unit": "%"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_108"}},
                "version": "1.5.0",
            },
            "widget_230": {
                "id": "widget_230",
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 2rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "version": "1.5.0",
            },
            "widget_231": {
                "id": "widget_231",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_109"}},
                "version": "1.5.0",
            },
            "widget_232": {
                "id": "widget_232",
                "uri": "widgets/common/button/",
                "config": {
                    "functionConfig": {"toolTip": "", "linkParam": {}},
                    "styleConfig": {
                        "name": "Default",
                        "themeStyle": {"className": "d-block", "color": "primary"},
                        "customStyle": {"regular": {}, "hover": {}, "clicked": {}},
                    },
                },
                "version": "1.5.0",
            },
            "widget_538": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_538",
                "layouts": {"DEFAULT": {"LARGE": "layout_347"}},
                "version": "1.5.0",
            },
            "widget_539": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_539",
                "layouts": {"DEFAULT": {"LARGE": "layout_348"}},
                "version": "1.5.0",
            },
            "widget_540": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 40,
                    "style": {
                        "padding": {"number": [40, 20, 40, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_540",
                "layouts": {"DEFAULT": {"LARGE": "layout_349"}},
                "version": "1.5.0",
            },
            "widget_541": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 40,
                    "style": {
                        "padding": {"number": [40, 20, 40, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_541",
                "layouts": {"DEFAULT": {"LARGE": "layout_350"}},
                "version": "1.5.0",
            },
            "widget_542": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 40,
                    "style": {
                        "padding": {"number": [40, 20, 40, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_542",
                "layouts": {"DEFAULT": {"LARGE": "layout_351"}},
                "version": "1.5.0",
            },
            "widget_543": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                        "imageFillMode": "FILL",
                    },
                    "styleConfig": {},
                },
                "id": "widget_543",
                "version": "1.5.0",
            },
            "widget_544": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_544",
                "style": {"background": {"fillType": "fit", "image": {"url": ""}}},
                "version": "1.5.0",
            },
            "widget_545": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_545",
                "version": "1.5.0",
            },
            "widget_546": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_546",
                "version": "1.5.0",
            },
            "widget_547": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_547",
                "version": "1.5.0",
            },
            "widget_548": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_548",
                "version": "1.5.0",
            },
            "widget_549": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 40, 10], "unit": "px"}},
                },
                "id": "widget_549",
                "layouts": {"DEFAULT": {"LARGE": "layout_352"}},
                "version": "1.5.0",
            },
            "widget_550": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_550",
                "layouts": {"DEFAULT": {"LARGE": "layout_353"}},
                "version": "1.5.0",
            },
            "widget_551": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_551",
                "layouts": {"DEFAULT": {"LARGE": "layout_354"}},
                "version": "1.5.0",
            },
            "widget_553": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_553",
                "layouts": {"DEFAULT": {"LARGE": "layout_356"}},
                "version": "1.5.0",
            },
            "widget_554": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_554",
                "version": "1.5.0",
            },
            "widget_555": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_555",
                "version": "1.5.0",
            },
            "widget_558": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_558",
                "version": "1.5.0",
            },
            "widget_559": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_559",
                "version": "1.5.0",
            },
            "widget_560": {
                "id": "widget_560",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 5, 20, 5], "unit": "%"}},
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_358"}},
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--light-300)"}
                },
                "version": "1.5.0",
            },
            "widget_561": {
                "id": "widget_561",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [0, 0, 0, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_359"}},
                "version": "1.5.0",
            },
            "widget_562": {
                "id": "widget_562",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_360"}},
                "version": "1.5.0",
            },
            "widget_563": {
                "id": "widget_563",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 40, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_361"}},
                "version": "1.5.0",
            },
            "widget_564": {
                "id": "widget_564",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [20, 0, 20, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_362"}},
                "version": "1.5.0",
            },
            "widget_565": {
                "id": "widget_565",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_363"}},
                "version": "1.5.0",
            },
            "widget_566": {
                "id": "widget_566",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 40, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_364"}},
                "version": "1.5.0",
            },
            "widget_567": {
                "id": "widget_567",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_365"}},
                "version": "1.5.0",
            },
            "widget_568": {
                "id": "widget_568",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 40, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_366"}},
                "version": "1.5.0",
            },
            "widget_569": {
                "id": "widget_569",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_367"}},
                "version": "1.5.0",
            },
            "widget_570": {
                "id": "widget_570",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 40, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_368"}},
                "version": "1.5.0",
            },
            "widget_571": {
                "id": "widget_571",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_369"}},
                "version": "1.5.0",
            },
            "widget_572": {
                "id": "widget_572",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 40, 10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_370"}},
                "version": "1.5.0",
            },
            "widget_573": {
                "id": "widget_573",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [20, 0, 20, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_371"}},
                "version": "1.5.0",
            },
            "widget_574": {
                "id": "widget_574",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_372"}},
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--light-300)"}
                },
                "version": "1.5.0",
            },
            "widget_575": {
                "id": "widget_575",
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [60, 0, 60, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_373"}},
                "version": "1.5.0",
            },
            "widget_576": {
                "id": "widget_576",
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 0,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_374"}},
                "version": "1.5.0",
            },
            "widget_577": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_577",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_375"}},
                "version": "1.5.0",
            },
            "widget_578": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_578",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_376"}},
                "version": "1.5.0",
            },
            "widget_579": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 40,
                    "style": {
                        "padding": {"number": [40, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_579",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_377"}},
                "version": "1.5.0",
            },
            "widget_580": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 40,
                    "style": {
                        "padding": {"number": [40, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_580",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_378"}},
                "version": "1.5.0",
            },
            "widget_581": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 40,
                    "style": {
                        "padding": {"number": [40, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_581",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_379"}},
                "version": "1.5.0",
            },
            "widget_582": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 40, 10], "unit": "px"}},
                },
                "id": "widget_582",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_380"}},
                "version": "1.5.0",
            },
            "widget_583": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_583",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_381"}},
                "version": "1.5.0",
            },
            "widget_584": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_584",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_382"}},
                "version": "1.5.0",
            },
            "widget_586": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_586",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_384"}},
                "version": "1.5.0",
            },
            "widget_631": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_631",
                "layouts": {"DEFAULT": {"LARGE": "layout_430"}},
                "version": "1.5.0",
            },
            "widget_632": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_632",
                "version": "1.5.0",
            },
            "widget_633": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_633",
                "version": "1.5.0",
            },
            "widget_634": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_634",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_431"}},
                "version": "1.5.0",
            },
            "widget_635": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_635",
                "version": "1.5.0",
            },
            "widget_636": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_636",
                "version": "1.5.0",
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "general",
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"home": []}],
        "timestamp": 0000000000000,
        "attributes": {"portalUrl": ""},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "introduction": {
        "pages": {
            "page_1": {
                "type": "NORMAL",
                "mode": "AUTO_SCROLL",
                "layout": {"LARGE": "layout_110", "MEDIUM": "layout_450"},
                "isVisible": True,
                "id": "page_1",
                "icon": {
                    "svg": '<svg viewBox="0 0 10 12" xmlns="http://www.w3.org/2000/svg"><path d="M1 0h8a1 1 0 011 1v10a1 1 0 01-1 1H1a1 1 0 01-1-1V1a1 1 0 011-1zm0 1v10h8V1H1zm2 2h4a.5.5 0 010 1H3a.5.5 0 010-1zm0 2.5h4a.5.5 0 010 1H3a.5.5 0 010-1z" fill="#000" fill-rule="nonzero"></path></svg>',
                    "properties": {
                        "filename": None,
                        "originalName": None,
                        "path": None,
                        "color": "#fff",
                        "size": 16,
                        "inlineSvg": True,
                    },
                },
                "heightInBuilder": {"LARGE": 2250, "SMALL": 3771, "MEDIUM": 1797},
                "isDefault": True,
            }
        },
        "layouts": {
            "layout_110": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "101px",
                            "width": "1920px",
                            "height": "320px",
                        },
                        "widgetId": "widget_233",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "801px",
                            "top": "548px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_235",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-110px",
                            "top": "-191px",
                            "width": "800px",
                            "height": 420,
                        },
                        "widgetId": "widget_643",
                        "setting": {"heightMode": "auto"},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "1095px",
                            "width": "1920px",
                            "height": "62px",
                        },
                        "widgetId": "widget_245",
                        "isPending": False,
                        "setting": {"heightMode": "auto"},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-63px",
                            "top": "-198px",
                            "width": "800px",
                            "height": 420,
                        },
                        "widgetId": "widget_677",
                        "setting": {"heightMode": "auto"},
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "15px",
                            "top": "-200px",
                            "width": "800px",
                            "height": 420,
                        },
                        "widgetId": "widget_679",
                        "setting": {"heightMode": "auto"},
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
            },
            "layout_111": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 535},
                        "widgetId": "widget_234",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:50"},
                    }
                },
                "order": ["0"],
            },
            "layout_112": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": 191},
                        "widgetId": "widget_236",
                        "setting": {
                            "maintainedByLayout": True,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_113": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 943, "height": 105.13671875},
                        "widgetId": "widget_237",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "110px",
                            "width": "943px",
                            "height": "70px",
                        },
                        "widgetId": "widget_238",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_118": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": 125},
                        "widgetId": "widget_246",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "auto",
                            "offsetX": 1.2,
                            "autoHeight": True,
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
            },
            "layout_431": {
                "content": {
                    "0": {
                        "bbox": {"height": "auto", "width": "160px"},
                        "type": "WIDGET",
                        "widgetId": "widget_633",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "100:100",
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_634",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_432": {
                "content": {
                    "0": {
                        "bbox": {"height": "auto", "width": "160px"},
                        "type": "WIDGET",
                        "widgetId": "widget_637",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "100:100",
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_638",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_433": {
                "content": {
                    "0": {
                        "bbox": {"height": "auto", "width": "160px"},
                        "type": "WIDGET",
                        "widgetId": "widget_641",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "100:100",
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_642",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_434": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": "400px"},
                        "widgetId": "widget_644",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_435": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 4, "height": 486},
                        "widgetId": "widget_632",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 4, "width": 4, "height": 486},
                        "widgetId": "widget_636",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {
                            "left": "8px",
                            "top": "NaNpx",
                            "width": "4px",
                            "height": "486px",
                        },
                        "widgetId": "widget_640",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_437": {
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_647",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"},
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_648",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_649",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-19px",
                            "top": "501px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_658",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "type": "COLUMN",
            },
            "layout_441": {
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_663",
                        "setting": {
                            "fitContainer": True,
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_664",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_665",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "93px",
                            "top": "306px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_674",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "type": "COLUMN",
            },
            "layout_442": {
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_667",
                        "setting": {
                            "fitContainer": True,
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_668",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_669",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "69px",
                            "top": "321px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_675",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "type": "COLUMN",
            },
            "layout_443": {
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_671",
                        "setting": {
                            "fitContainer": True,
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_672",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_673",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-51px",
                            "top": "321px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_676",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "type": "COLUMN",
            },
            "layout_444": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": "400px"},
                        "widgetId": "widget_678",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_445": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 4, "height": "410px"},
                        "widgetId": "widget_646",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 4, "width": 4, "height": "410px"},
                        "widgetId": "widget_681",
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 8, "width": 4, "height": "400px"},
                        "widgetId": "widget_686",
                    },
                },
                "order": ["0", "3", "4"],
            },
            "layout_446": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": 400},
                        "widgetId": "widget_680",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_447": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 4, "height": 410},
                        "widgetId": "widget_662",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 4, "width": 4, "height": 410},
                        "widgetId": "widget_666",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {
                            "left": "8px",
                            "top": "NaNpx",
                            "width": "4px",
                            "height": "410px",
                        },
                        "widgetId": "widget_670",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_448": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_682",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"},
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_683",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_684",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-19px",
                            "top": "501px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_685",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_449": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_687",
                        "setting": {
                            "fitContainer": True,
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                        "isPending": False,
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_688",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_689",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-33px",
                            "top": "480px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_690",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_450": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "101px",
                            "width": "1920px",
                            "height": "320px",
                        },
                        "widgetId": "widget_691",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "801px",
                            "top": "548px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_692",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-110px",
                            "top": "-191px",
                            "width": "800px",
                            "height": 420,
                        },
                        "widgetId": "widget_694",
                        "setting": {"heightMode": "auto"},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "1095px",
                            "width": "1920px",
                            "height": "62px",
                        },
                        "widgetId": "widget_699",
                        "isPending": False,
                        "setting": {"heightMode": "auto"},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-63px",
                            "top": "-198px",
                            "width": "800px",
                            "height": 420,
                        },
                        "widgetId": "widget_700",
                        "setting": {"heightMode": "auto"},
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "15px",
                            "top": "-200px",
                            "width": "800px",
                            "height": 420,
                        },
                        "widgetId": "widget_705",
                        "setting": {"heightMode": "auto"},
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
            },
            "layout_451": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 535},
                        "widgetId": "widget_234",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:50"},
                    }
                },
                "order": ["0"],
            },
            "layout_452": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "191px"},
                        "widgetId": "widget_693",
                        "setting": {
                            "maintainedByLayout": True,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_453": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 943, "height": 105.13671875},
                        "widgetId": "widget_237",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "top": "110px",
                            "width": "943px",
                            "height": "70px",
                        },
                        "widgetId": "widget_238",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_454": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_695",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_455": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 4, "height": 486},
                        "widgetId": "widget_696",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 4, "width": 4, "height": 486},
                        "widgetId": "widget_697",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {
                            "left": "8px",
                            "top": "NaNpx",
                            "width": "4px",
                            "height": "486px",
                        },
                        "widgetId": "widget_698",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_456": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto", "width": "160px"},
                        "type": "WIDGET",
                        "widgetId": "widget_633",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "100:100",
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_634",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_457": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto", "width": "160px"},
                        "type": "WIDGET",
                        "widgetId": "widget_637",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "100:100",
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_638",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_458": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto", "width": "160px"},
                        "type": "WIDGET",
                        "widgetId": "widget_641",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "100:100",
                            "autoProps": {"width": "CUSTOM", "height": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_642",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_459": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "125px"},
                        "widgetId": "widget_246",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "auto",
                            "offsetX": 1.2,
                            "autoHeight": True,
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_460": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_701",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_461": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 4, "height": "410px"},
                        "widgetId": "widget_702",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 4, "width": 4, "height": "410px"},
                        "widgetId": "widget_703",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 8, "width": 4, "height": "400px"},
                        "widgetId": "widget_704",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_462": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_647",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"},
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_648",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_649",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-19px",
                            "top": "501px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_658",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_463": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_682",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"},
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_683",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_684",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-19px",
                            "top": "501px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_685",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_464": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_687",
                        "setting": {
                            "fitContainer": True,
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                        "isPending": False,
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_688",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_689",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-33px",
                            "top": "480px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_690",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_465": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_706",
                        "setting": {
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_466": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 4, "height": 410},
                        "widgetId": "widget_707",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 4, "width": 4, "height": 410},
                        "widgetId": "widget_708",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {
                            "left": "8px",
                            "top": "NaNpx",
                            "width": "4px",
                            "height": "410px",
                        },
                        "widgetId": "widget_709",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_467": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_663",
                        "setting": {
                            "fitContainer": True,
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_664",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_665",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "93px",
                            "top": "306px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_674",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_468": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_667",
                        "setting": {
                            "fitContainer": True,
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_668",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_669",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "69px",
                            "top": "321px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_675",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_469": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "auto"},
                        "type": "WIDGET",
                        "widgetId": "widget_671",
                        "setting": {
                            "fitContainer": True,
                            "heightMode": "ratio",
                            "aspectRatio": "100:60",
                        },
                    },
                    "1": {
                        "bbox": {"height": "63px"},
                        "type": "WIDGET",
                        "widgetId": "widget_672",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_673",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-51px",
                            "top": "321px",
                            "width": "220px",
                            "height": "32px",
                        },
                        "widgetId": "widget_676",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
        },
        "widgets": {
            "widget_233": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_233",
                "layouts": {"DEFAULT": {"LARGE": "layout_111"}},
                "version": "1.5.0",
            },
            "widget_234": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M0 0v28h28V0H0z",
                                "cropShape": "rectangle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_234",
                "version": "1.5.0",
            },
            "widget_235": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [80, 10, 80, 10], "unit": "px"}},
                },
                "id": "widget_235",
                "layouts": {"DEFAULT": {"LARGE": "layout_112"}},
                "version": "1.5.0",
            },
            "widget_236": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_236",
                "layouts": {"DEFAULT": {"LARGE": "layout_113"}},
                "version": "1.5.0",
            },
            "widget_237": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center; line-height: 1.5;"><span style="font-family: &quot;Avenir Next&quot;; font-size: 36px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_237",
                "version": "1.5.0",
            },
            "widget_238": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_238",
                "version": "1.5.0",
            },
            "widget_245": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [60, 10, 60, 10], "unit": "px"}},
                },
                "id": "widget_245",
                "layouts": {"DEFAULT": {"LARGE": "layout_118"}},
                "version": "1.5.0",
            },
            "widget_246": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 36px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_246",
                "version": "1.5.0",
            },
            "widget_632": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [40, 0, 40, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_632",
                "layouts": {"DEFAULT": {"LARGE": "layout_431"}},
                "style": {"background": {"fillType": "center", "image": {"url": ""}}},
                "version": "1.5.0",
            },
            "widget_633": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M14.124 0l12.124 7v14l-12.124 7L2 21V7z",
                                "cropShape": "hexagon",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_633",
                "style": {"background": {"fillType": "center", "image": {"url": ""}}},
                "version": "1.5.0",
            },
            "widget_634": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_634",
                "version": "1.5.0",
            },
            "widget_636": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [40, 0, 40, 0], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_636",
                "layouts": {"DEFAULT": {"LARGE": "layout_432"}},
                "version": "1.5.0",
            },
            "widget_637": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M14.124 0l12.124 7v14l-12.124 7L2 21V7z",
                                "cropShape": "hexagon",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_637",
                "version": "1.5.0",
            },
            "widget_638": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_638",
                "version": "1.5.0",
            },
            "widget_640": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [40, 0, 40, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_640",
                "layouts": {"DEFAULT": {"LARGE": "layout_433"}},
                "version": "1.5.0",
            },
            "widget_641": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M14.124 0l12.124 7v14l-12.124 7L2 21V7z",
                                "cropShape": "hexagon",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_641",
                "version": "1.5.0",
            },
            "widget_642": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_642",
                "version": "1.5.0",
            },
            "widget_643": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_643",
                "layouts": {"DEFAULT": {"LARGE": "layout_434"}},
                "version": "1.5.0",
            },
            "widget_644": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_644",
                "layouts": {"DEFAULT": {"LARGE": "layout_435"}},
                "version": "1.5.0",
            },
            "widget_646": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_646",
                "layouts": {"DEFAULT": {"LARGE": "layout_437"}},
                "version": "1.5.0",
            },
            "widget_647": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_647",
                "version": "1.5.0",
            },
            "widget_648": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_648",
                "version": "1.5.0",
            },
            "widget_649": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 13px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_649",
                "version": "1.5.0",
            },
            "widget_658": {
                "uri": "widgets/common/button/",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {"color": "#f0f0f0", "fillType": "fill"},
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {"color": "#fff", "fillType": "fill"},
                                "border": {
                                    "type": "solid",
                                    "color": "#c5c5c5",
                                    "width": "1px",
                                },
                                "text": {"color": "#181818", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_658",
                "version": "1.5.0",
            },
            "widget_662": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_662",
                "layouts": {"DEFAULT": {"LARGE": "layout_441"}},
                "version": "1.5.0",
            },
            "widget_663": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_663",
                "version": "1.5.0",
            },
            "widget_664": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_664",
                "version": "1.5.0",
            },
            "widget_665": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 13px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_665",
                "version": "1.5.0",
            },
            "widget_666": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_666",
                "layouts": {"DEFAULT": {"LARGE": "layout_442"}},
                "version": "1.5.0",
            },
            "widget_667": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_667",
                "version": "1.5.0",
            },
            "widget_668": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_668",
                "version": "1.5.0",
            },
            "widget_669": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 13px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_669",
                "version": "1.5.0",
            },
            "widget_670": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_670",
                "layouts": {"DEFAULT": {"LARGE": "layout_443"}},
                "version": "1.5.0",
            },
            "widget_671": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_671",
                "version": "1.5.0",
            },
            "widget_672": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_672",
                "version": "1.5.0",
            },
            "widget_673": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 13px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_673",
                "version": "1.5.0",
            },
            "widget_674": {
                "uri": "widgets/common/button/",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {"themeStyle": {"quickStyleType": "default"}},
                },
                "id": "widget_674",
                "version": "1.5.0",
            },
            "widget_675": {
                "uri": "widgets/common/button/",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {"themeStyle": {"quickStyleType": "default"}},
                },
                "id": "widget_675",
                "version": "1.5.0",
            },
            "widget_676": {
                "uri": "widgets/common/button/",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {"themeStyle": {"quickStyleType": "default"}},
                },
                "id": "widget_676",
                "version": "1.5.0",
            },
            "widget_677": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_677",
                "layouts": {"DEFAULT": {"LARGE": "layout_444"}},
                "version": "1.5.0",
            },
            "widget_678": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_678",
                "layouts": {"DEFAULT": {"LARGE": "layout_445"}},
                "version": "1.5.0",
            },
            "widget_679": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_679",
                "layouts": {"DEFAULT": {"LARGE": "layout_446"}},
                "version": "1.5.0",
            },
            "widget_680": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_680",
                "layouts": {"DEFAULT": {"LARGE": "layout_447"}},
                "version": "1.5.0",
            },
            "widget_681": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_681",
                "layouts": {"DEFAULT": {"LARGE": "layout_448"}},
                "version": "1.5.0",
            },
            "widget_682": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_682",
                "version": "1.5.0",
            },
            "widget_683": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_683",
                "version": "1.5.0",
            },
            "widget_684": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 13px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_684",
                "version": "1.5.0",
            },
            "widget_685": {
                "uri": "widgets/common/button/",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {"color": "#f0f0f0", "fillType": "fill"},
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {"color": "#fff", "fillType": "fill"},
                                "border": {
                                    "type": "solid",
                                    "color": "#c5c5c5",
                                    "width": "1px",
                                },
                                "text": {"color": "#181818", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_685",
                "version": "1.5.0",
            },
            "widget_686": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_686",
                "layouts": {"DEFAULT": {"LARGE": "layout_449"}},
                "version": "1.5.0",
            },
            "widget_687": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_687",
                "version": "1.5.0",
            },
            "widget_688": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_688",
                "version": "1.5.0",
            },
            "widget_689": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style=""><span style="font-size: 13px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_689",
                "version": "1.5.0",
            },
            "widget_690": {
                "uri": "widgets/common/button/",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {"themeStyle": {"quickStyleType": "default"}},
                },
                "id": "widget_690",
                "version": "1.5.0",
            },
            "widget_691": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_691",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_451"}},
                "version": "1.5.0",
            },
            "widget_692": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [80, 10, 80, 10], "unit": "px"}},
                },
                "id": "widget_692",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_452"}},
                "version": "1.5.0",
            },
            "widget_693": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_693",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_453"}},
                "version": "1.5.0",
            },
            "widget_694": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_694",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_454"}},
                "version": "1.5.0",
            },
            "widget_695": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_695",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_455"}},
                "version": "1.5.0",
            },
            "widget_696": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [40, 0, 40, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_696",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_456"}},
                "style": {"background": {"fillType": "center", "image": {"url": ""}}},
                "version": "1.5.0",
            },
            "widget_697": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [40, 0, 40, 0], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_697",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_457"}},
                "version": "1.5.0",
            },
            "widget_698": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [40, 0, 40, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_698",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_458"}},
                "version": "1.5.0",
            },
            "widget_699": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [60, 10, 60, 10], "unit": "px"}},
                },
                "id": "widget_699",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_459"}},
                "version": "1.5.0",
            },
            "widget_700": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_700",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_460"}},
                "version": "1.5.0",
            },
            "widget_701": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_701",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_461"}},
                "version": "1.5.0",
            },
            "widget_702": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_702",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_462"}},
                "version": "1.5.0",
            },
            "widget_703": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_703",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_463"}},
                "version": "1.5.0",
            },
            "widget_704": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_704",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_464"}},
                "version": "1.5.0",
            },
            "widget_705": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_705",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_465"}},
                "version": "1.5.0",
            },
            "widget_706": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_706",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_466"}},
                "version": "1.5.0",
            },
            "widget_707": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_707",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_467"}},
                "version": "1.5.0",
            },
            "widget_708": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_708",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_468"}},
                "version": "1.5.0",
            },
            "widget_709": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_709",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_469"}},
                "version": "1.5.0",
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "introduction",
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_1": []}],
        "timestamp": 1624304713265,
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "gallery": {
        "pages": {
            "page_2": {
                "id": "page_2",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {"LARGE": "layout_127"},
                "isVisible": True,
                "icon": {
                    "svg": '<svg viewBox="0 0 10 12" xmlns="http://www.w3.org/2000/svg"><path d="M1 0h8a1 1 0 011 1v10a1 1 0 01-1 1H1a1 1 0 01-1-1V1a1 1 0 011-1zm0 1v10h8V1H1zm2 2h4a.5.5 0 010 1H3a.5.5 0 010-1zm0 2.5h4a.5.5 0 010 1H3a.5.5 0 010-1z" fill="#000" fill-rule="nonzero"></path></svg>',
                    "properties": {
                        "filename": None,
                        "originalName": None,
                        "path": None,
                        "color": "#fff",
                        "size": 16,
                        "inlineSvg": True,
                    },
                },
                "heightInBuilder": {"LARGE": 2836, "MEDIUM": 2139, "SMALL": 5056},
                "isDefault": True,
            }
        },
        "layouts": {
            "layout_127": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "772px",
                            "top": "-44px",
                            "width": "300px",
                            "height": "300px",
                        },
                        "widgetId": "widget_259",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "717px",
                            "top": "580px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_261",
                    },
                    "2": {"type": "WIDGET", "widgetId": "widget_265"},
                    "3": {"type": "WIDGET", "widgetId": "widget_278"},
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "563px",
                            "top": "1811px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_291",
                    },
                    "5": {"type": "WIDGET", "widgetId": "widget_295"},
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "811px",
                            "top": "2312px",
                            "width": "300px",
                            "height": "600px",
                        },
                        "widgetId": "widget_308",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6"],
            },
            "layout_128": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 496.203125},
                        "widgetId": "widget_260",
                        "setting": {"heightMode": "ratio", "aspectRatio": "100:50"},
                    }
                },
            },
            "layout_129": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "246px"},
                        "widgetId": "widget_262",
                        "setting": {
                            "heightMode": "fit",
                            "maintainedByLayout": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_130": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "top": "NaNpx",
                            "width": "NaNpx",
                            "height": "156px",
                        },
                        "widgetId": "widget_263",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "419px",
                            "top": "-30px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_264",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "type": "COLUMN",
            },
            "layout_131": {
                "content": {
                    "0": {
                        "bbox": {"left": 0, "width": 4, "height": "NaNpx"},
                        "type": "WIDGET",
                        "widgetId": "widget_266",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "NaN:null",
                            "autoHeight": True,
                        },
                    },
                    "1": {
                        "bbox": {"left": 4, "width": 4, "height": "NaNpx"},
                        "type": "WIDGET",
                        "widgetId": "widget_270",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "NaN:null",
                            "autoHeight": True,
                        },
                    },
                    "3": {
                        "bbox": {"left": 8, "width": 4, "height": "NaNpx"},
                        "type": "WIDGET",
                        "widgetId": "widget_493",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "NaN:null",
                            "autoHeight": True,
                        },
                    },
                },
                "order": ["0", "1", "3"],
                "type": "ROW",
            },
            "layout_132": {
                "content": {
                    "0": {
                        "bbox": {"width": 371.328125, "height": "260px"},
                        "type": "WIDGET",
                        "widgetId": "widget_267",
                        "setting": {
                            "fitContainer": False,
                            "heightMode": "ratio",
                            "aspectRatio": "100:80",
                        },
                    },
                    "1": {
                        "bbox": {"height": "80.00px"},
                        "type": "WIDGET",
                        "widgetId": "widget_268",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_269",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1", "2"],
                "type": "COLUMN",
            },
            "layout_133": {
                "content": {
                    "0": {
                        "bbox": {"width": 365.328125, "height": "260px"},
                        "type": "WIDGET",
                        "widgetId": "widget_271",
                        "setting": {
                            "fitContainer": False,
                            "heightMode": "ratio",
                            "aspectRatio": "100:80",
                        },
                    },
                    "1": {
                        "bbox": {"height": "80.00px"},
                        "type": "WIDGET",
                        "widgetId": "widget_272",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_273",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1", "2"],
                "type": "COLUMN",
            },
            "layout_135": {
                "content": {
                    "0": {
                        "bbox": {"left": 0, "width": 4, "height": "NaNpx"},
                        "type": "WIDGET",
                        "widgetId": "widget_279",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "NaN:null",
                            "autoHeight": True,
                        },
                    },
                    "1": {
                        "bbox": {"left": 4, "width": 4, "height": "NaNpx"},
                        "type": "WIDGET",
                        "widgetId": "widget_283",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "NaN:null",
                            "autoHeight": True,
                        },
                    },
                    "3": {
                        "bbox": {"left": 8, "width": 4, "height": "NaNpx"},
                        "type": "WIDGET",
                        "widgetId": "widget_497",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "NaN:null",
                            "autoHeight": True,
                        },
                    },
                },
                "order": ["0", "1", "3"],
                "type": "ROW",
            },
            "layout_136": {
                "content": {
                    "0": {
                        "bbox": {"height": "260px"},
                        "type": "WIDGET",
                        "widgetId": "widget_280",
                        "setting": {
                            "fitContainer": False,
                            "heightMode": "ratio",
                            "aspectRatio": "100:80",
                        },
                    },
                    "1": {
                        "bbox": {"height": "80.00px"},
                        "type": "WIDGET",
                        "widgetId": "widget_281",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_282",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1", "2"],
                "type": "COLUMN",
            },
            "layout_137": {
                "content": {
                    "0": {
                        "bbox": {"height": "260px"},
                        "type": "WIDGET",
                        "widgetId": "widget_284",
                        "setting": {
                            "fitContainer": False,
                            "heightMode": "ratio",
                            "aspectRatio": "100:80",
                        },
                    },
                    "1": {
                        "bbox": {"height": "80.00px"},
                        "type": "WIDGET",
                        "widgetId": "widget_285",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_286",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1", "2"],
                "type": "COLUMN",
            },
            "layout_139": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "160px"},
                        "widgetId": "widget_292",
                        "setting": {
                            "maintainedByLayout": True,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_140": {
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "top": "NaNpx",
                            "width": "NaNpx",
                            "height": "156.00px",
                        },
                        "widgetId": "widget_293",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "455px",
                            "top": "-26px",
                            "width": "360px",
                            "height": "70px",
                        },
                        "widgetId": "widget_294",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "type": "COLUMN",
            },
            "layout_141": {
                "content": {
                    "0": {
                        "bbox": {"left": 0, "width": 3},
                        "type": "WIDGET",
                        "widgetId": "widget_296",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "NaN:null",
                            "autoHeight": True,
                        },
                    },
                    "1": {
                        "bbox": {"left": 3, "width": 3},
                        "type": "WIDGET",
                        "widgetId": "widget_299",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "NaN:12",
                            "autoHeight": True,
                        },
                    },
                    "2": {
                        "bbox": {"left": 6, "width": 3},
                        "type": "WIDGET",
                        "widgetId": "widget_302",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "NaN:null",
                            "autoHeight": True,
                        },
                    },
                    "3": {
                        "bbox": {"left": 9, "width": 3},
                        "type": "WIDGET",
                        "widgetId": "widget_305",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": "NaN:null",
                            "autoHeight": True,
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
                "type": "ROW",
            },
            "layout_142": {
                "content": {
                    "0": {
                        "bbox": {"height": "200px"},
                        "type": "WIDGET",
                        "widgetId": "widget_297",
                        "setting": {
                            "fitContainer": False,
                            "heightMode": "ratio",
                            "aspectRatio": "100:80",
                        },
                    },
                    "1": {
                        "bbox": {"height": "80.00px"},
                        "type": "WIDGET",
                        "widgetId": "widget_298",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_143": {
                "content": {
                    "0": {
                        "bbox": {"height": "200px"},
                        "type": "WIDGET",
                        "widgetId": "widget_300",
                        "setting": {
                            "fitContainer": False,
                            "heightMode": "ratio",
                            "aspectRatio": "100:80",
                        },
                    },
                    "1": {
                        "bbox": {
                            "height": "80.00px",
                            "left": -0.8203125,
                            "right": 262.1796875,
                            "top": 209.0546875,
                            "bottom": 239.0546875,
                            "width": 263,
                        },
                        "type": "WIDGET",
                        "widgetId": "widget_301",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_144": {
                "content": {
                    "0": {
                        "bbox": {"height": "200px"},
                        "type": "WIDGET",
                        "widgetId": "widget_303",
                        "setting": {
                            "fitContainer": False,
                            "heightMode": "ratio",
                            "aspectRatio": "100:80",
                        },
                    },
                    "1": {
                        "bbox": {"height": "80.00px"},
                        "type": "WIDGET",
                        "widgetId": "widget_304",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_145": {
                "content": {
                    "0": {
                        "bbox": {"height": "200px"},
                        "type": "WIDGET",
                        "widgetId": "widget_306",
                        "setting": {
                            "autoProps": {"height": "CUSTOM"},
                            "heightMode": "ratio",
                            "aspectRatio": "100:80",
                        },
                        "isPending": False,
                    },
                    "1": {
                        "bbox": {"height": "80.00px"},
                        "type": "WIDGET",
                        "widgetId": "widget_307",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1"],
                "type": "COLUMN",
            },
            "layout_146": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 428.74609375},
                        "widgetId": "widget_309",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "setting": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 40, 10], "unit": "px"}},
                },
            },
            "layout_147": {
                "order": ["0", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": 1278, "height": 132.4375},
                        "widgetId": "widget_310",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "autoProps": {"width": "CUSTOM"},
                        },
                        "bbox": {"width": "200.00px", "height": "50px"},
                        "widgetId": "widget_312",
                    },
                },
                "type": "COLUMN",
            },
            "layout_304": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"width": 365.328125, "height": "260px"},
                        "type": "WIDGET",
                        "widgetId": "widget_494",
                        "setting": {
                            "fitContainer": False,
                            "heightMode": "ratio",
                            "aspectRatio": "100:80",
                        },
                    },
                    "1": {
                        "bbox": {"height": "80.00px"},
                        "type": "WIDGET",
                        "widgetId": "widget_495",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_496",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_305": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "bbox": {"height": "260px"},
                        "type": "WIDGET",
                        "widgetId": "widget_498",
                        "setting": {
                            "fitContainer": False,
                            "heightMode": "ratio",
                            "aspectRatio": "100:80",
                        },
                    },
                    "1": {
                        "bbox": {"height": "80.00px"},
                        "type": "WIDGET",
                        "widgetId": "widget_499",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                    "2": {
                        "bbox": {"height": "53px"},
                        "type": "WIDGET",
                        "widgetId": "widget_500",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "order": ["0", "1", "2"],
            },
        },
        "widgets": {
            "widget_259": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_259",
                "layouts": {"DEFAULT": {"LARGE": "layout_128"}},
                "version": "1.5.0",
            },
            "widget_260": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_260",
                "version": "1.5.0",
            },
            "widget_261": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [60, 10, 60, 10], "unit": "px"}},
                },
                "id": "widget_261",
                "layouts": {"DEFAULT": {"LARGE": "layout_129"}},
                "version": "1.5.0",
            },
            "widget_262": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_262",
                "layouts": {"DEFAULT": {"LARGE": "layout_130"}},
                "version": "1.5.0",
            },
            "widget_263": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 36px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_263",
                "version": "1.5.0",
            },
            "widget_264": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;">Double click to edit text</p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_264",
                "version": "1.5.0",
            },
            "widget_265": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 80, 20, 80], "unit": "px"}},
                },
                "id": "widget_265",
                "layouts": {"DEFAULT": {"LARGE": "layout_131"}},
                "version": "1.5.0",
            },
            "widget_266": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 40, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_266",
                "style": {
                    "background": {"image": {"url": ""}},
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": {"distance": 1, "unit": "px"},
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_132"}},
                "version": "1.5.0",
            },
            "widget_267": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_267",
                "version": "1.5.0",
            },
            "widget_268": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_268",
                "version": "1.5.0",
            },
            "widget_269": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_269",
                "version": "1.5.0",
            },
            "widget_270": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 40, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_270",
                "style": {
                    "boxShadow": {
                        "offsetX": {"distance": 4, "unit": "px"},
                        "offsetY": {"distance": 4, "unit": "px"},
                        "blur": {"distance": 10, "unit": "px"},
                        "spread": {"distance": 0, "unit": "px"},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": {"distance": 1, "unit": "px"},
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_133"}},
                "version": "1.5.0",
            },
            "widget_271": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_271",
                "version": "1.5.0",
            },
            "widget_272": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_272",
                "version": "1.5.0",
            },
            "widget_273": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_273",
                "version": "1.5.0",
            },
            "widget_278": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 80, 20, 80], "unit": "px"}},
                },
                "id": "widget_278",
                "layouts": {"DEFAULT": {"LARGE": "layout_135"}},
                "version": "1.5.0",
            },
            "widget_279": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 40, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_279",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": {"distance": 1, "unit": "px"},
                    }
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_136"}},
                "version": "1.5.0",
            },
            "widget_280": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_280",
                "version": "1.5.0",
            },
            "widget_281": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_281",
                "version": "1.5.0",
            },
            "widget_282": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_282",
                "version": "1.5.0",
            },
            "widget_283": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 40, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_283",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": {"distance": 1, "unit": "px"},
                    }
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_137"}},
                "version": "1.5.0",
            },
            "widget_284": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_284",
                "version": "1.5.0",
            },
            "widget_285": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_285",
                "version": "1.5.0",
            },
            "widget_286": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_286",
                "version": "1.5.0",
            },
            "widget_291": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [60, 20, 60, 20], "unit": "px"}},
                },
                "id": "widget_291",
                "layouts": {"DEFAULT": {"LARGE": "layout_139"}},
                "version": "1.5.0",
            },
            "widget_292": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_292",
                "layouts": {"DEFAULT": {"LARGE": "layout_140"}},
                "version": "1.5.0",
            },
            "widget_293": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 36px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_293",
                "version": "1.5.0",
            },
            "widget_294": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_294",
                "version": "1.5.0",
            },
            "widget_295": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 80, 20, 80], "unit": "px"}},
                },
                "id": "widget_295",
                "layouts": {"DEFAULT": {"LARGE": "layout_141"}},
                "version": "1.5.0",
            },
            "widget_296": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [0, 0, 40, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_296",
                "layouts": {"DEFAULT": {"LARGE": "layout_142"}},
                "version": "1.5.0",
            },
            "widget_297": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_297",
                "version": "1.5.0",
            },
            "widget_298": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_298",
                "version": "1.5.0",
            },
            "widget_299": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [0, 0, 40, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_299",
                "layouts": {"DEFAULT": {"LARGE": "layout_143"}},
                "version": "1.5.0",
            },
            "widget_300": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_300",
                "version": "1.5.0",
            },
            "widget_301": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_301",
                "version": "1.5.0",
            },
            "widget_302": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [0, 0, 40, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_302",
                "layouts": {"DEFAULT": {"LARGE": "layout_144"}},
                "version": "1.5.0",
            },
            "widget_303": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_303",
                "version": "1.5.0",
            },
            "widget_304": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_304",
                "version": "1.5.0",
            },
            "widget_305": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [0, 0, 40, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_305",
                "layouts": {"DEFAULT": {"LARGE": "layout_145"}},
                "version": "1.5.0",
            },
            "widget_306": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_306",
                "version": "1.5.0",
            },
            "widget_307": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_307",
                "version": "1.5.0",
            },
            "widget_308": {
                "uri": "widgets/layout/row/",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [100, 10, 100, 10], "unit": "px"}},
                },
                "id": "widget_308",
                "style": {
                    "background": {"image": {"url": ""}, "color": "var(--light-300)"}
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_146"}},
                "version": "1.5.0",
            },
            "widget_309": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_309",
                "layouts": {"DEFAULT": {"LARGE": "layout_147"}},
                "version": "1.5.0",
            },
            "widget_310": {
                "uri": "widgets/common/text/",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 36px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_310",
                "style": {"background": {"image": {"url": ""}, "fillType": "fit"}},
                "version": "1.5.0",
            },
            "widget_312": {
                "uri": "widgets/common/button/",
                "config": {
                    "functionConfig": {"toolTip": "", "linkParam": {}},
                    "styleConfig": {
                        "name": "Default",
                        "themeStyle": {"className": "d-block", "color": "primary"},
                        "customStyle": {"regular": {}, "hover": {}, "clicked": {}},
                    },
                },
                "id": "widget_312",
                "version": "1.5.0",
            },
            "widget_493": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 40, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_493",
                "style": {
                    "boxShadow": {
                        "offsetX": {"distance": 4, "unit": "px"},
                        "offsetY": {"distance": 4, "unit": "px"},
                        "blur": {"distance": 10, "unit": "px"},
                        "spread": {"distance": 0, "unit": "px"},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": {"distance": 1, "unit": "px"},
                    },
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_304"}},
                "version": "1.5.0",
            },
            "widget_494": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_494",
                "version": "1.5.0",
            },
            "widget_495": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_495",
                "version": "1.5.0",
            },
            "widget_496": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_496",
                "version": "1.5.0",
            },
            "widget_497": {
                "uri": "widgets/layout/column/",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 40, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_497",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": {"distance": 1, "unit": "px"},
                    }
                },
                "layouts": {"DEFAULT": {"LARGE": "layout_305"}},
                "version": "1.5.0",
            },
            "widget_498": {
                "uri": "widgets/common/image/",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_498",
                "version": "1.5.0",
            },
            "widget_499": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_499",
                "version": "1.5.0",
            },
            "widget_500": {
                "uri": "widgets/common/text/",
                "config": {
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 1rem;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_500",
                "version": "1.5.0",
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "gallery",
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_2": []}],
        "timestamp": 1624314119393,
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "epic": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_0",
                    "MEDIUM": "layout_482",
                    "SMALL": "layout_527",
                },
                "isDefault": True,
                "isVisible": True,
                "bodyBackgroundColor": "",
                "header": False,
                "footer": False,
                "oneByOneEffect": {"type": "FLOAT_IN", "option": {"direction": None}},
            }
        },
        "layouts": {
            "layout_0": {
                "type": "FLOW",
                "content": {
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-263px",
                            "right": "1017px",
                            "top": "80px",
                            "bottom": "180px",
                            "width": "1280px",
                            "height": "160px",
                        },
                        "widgetId": "widget_16",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "220px",
                            "right": "580px",
                            "top": "316px",
                            "bottom": "386px",
                            "width": "360px",
                            "height": "80px",
                        },
                        "widgetId": "widget_18",
                        "setting": {
                            "heightMode": "auto",
                            "offsetX": 0,
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "314px",
                            "right": "614px",
                            "top": "476px",
                            "bottom": "776px",
                            "width": "300px",
                            "height": "40%",
                        },
                        "widgetId": "widget_20",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "offsetX": 0,
                            "width": "100%",
                        },
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35px",
                            "right": "535px",
                            "top": "1229px",
                            "bottom": "1729px",
                            "width": "500px",
                            "height": "430px",
                        },
                        "widgetId": "widget_34",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "1833px",
                            "bottom": "1888px",
                            "width": "1200px",
                            "height": "60px",
                        },
                        "widgetId": "widget_54",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "151px",
                            "right": "551px",
                            "top": "1960px",
                            "bottom": "2260px",
                            "width": "400px",
                            "height": "450px",
                        },
                        "widgetId": "widget_69",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "941px",
                            "bottom": "1041px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_73",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "17": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "1337px",
                            "bottom": "1437px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_75",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "18": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "42px",
                            "right": "1242px",
                            "top": "1940px",
                            "bottom": "2040px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_77",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "19": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-91px",
                            "right": "269px",
                            "top": "2369px",
                            "bottom": "2439px",
                            "width": "360px",
                            "height": 89,
                        },
                        "widgetId": "widget_82",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "20": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "2561px",
                            "bottom": "2651px",
                            "width": "1200px",
                            "height": 118,
                        },
                        "widgetId": "widget_85",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "23": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "3400px",
                            "bottom": "3640px",
                            "width": "1200px",
                            "height": "240px",
                        },
                        "widgetId": "widget_119",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "24": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "86px",
                            "right": "486px",
                            "top": "3260px",
                            "bottom": "3660px",
                            "width": "400px",
                            "height": "500px",
                        },
                        "widgetId": "widget_123",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "27": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_140",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "width": "100%",
                        },
                    },
                    "28": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_145",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "width": "100%",
                        },
                    },
                    "29": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "4347px",
                            "bottom": "4447px",
                            "width": "1200px",
                            "height": "320px",
                        },
                        "widgetId": "widget_150",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "30": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-18px",
                            "right": "1182px",
                            "top": "4200px",
                            "bottom": "4440px",
                            "width": "1200px",
                            "height": "240px",
                        },
                        "widgetId": "widget_152",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "33": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-177px",
                            "right": "223px",
                            "top": "-199px",
                            "bottom": "201px",
                            "width": "400px",
                            "height": "480px",
                        },
                        "widgetId": "widget_439",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                    },
                    "34": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-121px",
                            "right": "179px",
                            "top": "-191px",
                            "bottom": "200px",
                            "width": "300px",
                            "height": 343,
                        },
                        "widgetId": "widget_543",
                        "setting": {
                            "heightMode": "auto",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "width": "100%",
                        },
                    },
                },
                "order": [
                    "5",
                    "6",
                    "7",
                    "16",
                    "34",
                    "17",
                    "11",
                    "12",
                    "18",
                    "14",
                    "19",
                    "20",
                    "33",
                    "23",
                    "24",
                    "30",
                    "27",
                    "28",
                    "29",
                ],
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
            },
            "layout_10": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_17",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
            },
            "layout_11": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": 70},
                        "widgetId": "widget_19",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_12": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 300},
                        "widgetId": "widget_21",
                        "setting": {
                            "heightMode": "ratio",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:40",
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
            },
            "layout_19": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": 414},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    }
                },
            },
            "layout_20": {
                "type": "FIXED",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "52.2007%",
                            "top": "50%",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_35",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "height": "CUSTOM",
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1506%",
                            "right": "17.3743%",
                            "top": 21,
                            "bottom": "86.2333%",
                            "width": "100.00%",
                            "height": "8.7379%",
                        },
                        "widgetId": "widget_36",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0295%",
                            "right": 0,
                            "top": "21.0255%",
                            "bottom": "25.3337%",
                            "width": "48.1422%",
                            "height": "53.6408%",
                        },
                        "widgetId": "widget_37",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": "20.00px",
                            "width": "18.6757%",
                            "height": "42px",
                        },
                        "widgetId": "widget_38",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"bottom": False, "top": True},
                        },
                    },
                },
            },
            "layout_26": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "offsetX": -10,
                            "autoProps": {"height": False},
                        },
                        "bbox": {"left": 2, "width": 2, "height": 30},
                        "widgetId": "widget_53",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 11, "width": 1, "height": 60},
                        "widgetId": "widget_128",
                        "setting": {"heightMode": "fit"},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_31": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 300},
                        "widgetId": "widget_70",
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_33": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_74",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_34": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_76",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_35": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "50px"},
                        "widgetId": "widget_78",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_37": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "50px"},
                        "widgetId": "widget_83",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "flex-end"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 4, "height": "50px"},
                        "widgetId": "widget_86",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "flex-end"},
                        },
                    },
                },
            },
            "layout_38": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 4, "height": 70},
                        "widgetId": "widget_84",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 5, "width": 4, "height": 70},
                        "widgetId": "widget_87",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                        "bbox": {"left": 10, "width": 2, "height": "42px"},
                        "widgetId": "widget_88",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_47": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 400},
                        "widgetId": "widget_120",
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_48": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "vCenter": False, "hCenter": False},
                        "bbox": {
                            "left": 0,
                            "right": "65.7059%",
                            "top": 73,
                            "bottom": "30.1771%",
                            "width": "34.2954%",
                            "height": "35.9447%",
                        },
                        "widgetId": "widget_121",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "49.3209%",
                            "right": 0,
                            "top": "34.1014%",
                            "bottom": 10,
                            "width": "50.8489%",
                            "height": "61.2903%",
                        },
                        "widgetId": "widget_122",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_49": {
                "type": "ROW",
                "order": ["1", "0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": 400},
                        "widgetId": "widget_124",
                        "setting": {"heightMode": "fit"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "490px"},
                        "widgetId": "widget_129",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
            },
            "layout_50": {"type": "FIXED"},
            "layout_54": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.58%",
                            "top": "50%",
                            "width": "13.39%",
                            "height": "55.81%",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "21.20%",
                            "width": "226px",
                            "height": "35%",
                            "right": "39.05%",
                            "top": "50%",
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "230px",
                            "top": "50%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.25%",
                            "bottom": "20.93%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_55": {"type": "FIXED"},
            "layout_57": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "9px",
                            "top": "10px",
                            "width": "29.8968%",
                            "height": "86.5369%",
                            "right": "67.07189999999997%",
                            "bottom": "8px",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35.0191%",
                            "width": 238,
                            "height": "43.6104%",
                            "right": "3.1237%",
                            "top": 10,
                            "bottom": 74,
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "274px",
                            "top": "69.6731999999997%",
                            "width": "27.8170%",
                            "height": "23.8722%",
                            "right": "2.2273999999999994%",
                            "bottom": "6.163100000000027%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_60": {
                "type": "ROW",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 89},
                        "widgetId": "widget_141",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": 89},
                        "widgetId": "widget_142",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": 89},
                        "widgetId": "widget_143",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": 89},
                        "widgetId": "widget_144",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
            },
            "layout_61": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 89},
                        "widgetId": "widget_146",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": 89},
                        "widgetId": "widget_147",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": 89},
                        "widgetId": "widget_148",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": 89},
                        "widgetId": "widget_149",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_62": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": 50},
                        "widgetId": "widget_151",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_63": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 400},
                        "widgetId": "widget_153",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_64": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO", "width": "AUTO"},
                        },
                        "bbox": {
                            "left": 0,
                            "right": "65.7059%",
                            "top": 73,
                            "bottom": "30.1771%",
                            "width": "34.2954%",
                            "height": "35.9447%",
                        },
                        "widgetId": "widget_154",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "49.3209%",
                            "right": 0,
                            "top": "34.1014%",
                            "bottom": 10,
                            "width": "50.8489%",
                            "height": "61.2903%",
                        },
                        "widgetId": "widget_155",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO", "width": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_342": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_440",
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_343": {
                "type": "FIXED",
                "order": ["1", "0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "51.5763%",
                            "top": "87.9151%",
                            "bottom": "10.00px",
                            "width": "15.1778%",
                            "height": "9.3280%",
                        },
                        "widgetId": "widget_441",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "1": {
                        "sectionId": "section_7",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "3.1819%",
                            "top": "50%",
                            "bottom": "0.2506%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                },
            },
            "layout_344": {
                "type": "FIXED",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_442",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "STRETCH",
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_443",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_444",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
            },
            "layout_379": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_469",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_470",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_471",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_381": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_475",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_476",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_477",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_466": {
                "type": "ROW",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "391px",
                        },
                        "widgetId": "widget_544",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 3,
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "391px",
                        },
                        "widgetId": "widget_571",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 6,
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "391px",
                        },
                        "widgetId": "widget_577",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 9,
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "391px",
                        },
                        "widgetId": "widget_583",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                        },
                    },
                },
            },
            "layout_474": {
                "type": "FIXED",
                "content": {
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_570",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_862",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_863",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["6", "7", "8"],
            },
            "layout_475": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_568",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_569",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_476": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_572",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_573",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_574",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_477": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_575",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_576",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_478": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_578",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_579",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_580",
                        "setting": {
                            "lockParent": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "top": False,
                                "bottom": True,
                                "left": False,
                                "right": True,
                            },
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_479": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_581",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_582",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_480": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_584",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_585",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_586",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_481": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_587",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_588",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_482": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-263px",
                            "right": "1017px",
                            "top": "80px",
                            "bottom": "180px",
                            "width": "1280px",
                            "height": "160px",
                        },
                        "widgetId": "widget_590",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "220px",
                            "right": "580px",
                            "top": "316px",
                            "bottom": "386px",
                            "width": "360px",
                            "height": "80px",
                        },
                        "widgetId": "widget_591",
                        "setting": {
                            "heightMode": "auto",
                            "offsetX": 0,
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "314px",
                            "right": "614px",
                            "top": "476px",
                            "bottom": "776px",
                            "width": "300px",
                            "height": "700px",
                        },
                        "widgetId": "widget_592",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "offsetX": 0,
                            "width": "100%",
                        },
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "941px",
                            "bottom": "1041px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_593",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-121px",
                            "right": "179px",
                            "top": "-191px",
                            "bottom": "200px",
                            "width": "300px",
                            "height": "320px",
                        },
                        "widgetId": "widget_594",
                        "setting": {
                            "heightMode": "auto",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": "BOTTOM"},
                                }
                            },
                            "width": "100%",
                        },
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "1337px",
                            "bottom": "1437px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_595",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35px",
                            "right": "535px",
                            "top": "1229px",
                            "bottom": "1729px",
                            "width": "500px",
                            "height": "430px",
                        },
                        "widgetId": "widget_596",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "1833px",
                            "bottom": "1888px",
                            "width": "1200px",
                            "height": "60px",
                        },
                        "widgetId": "widget_597",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "42px",
                            "right": "1242px",
                            "top": "1940px",
                            "bottom": "2040px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_598",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "151px",
                            "right": "551px",
                            "top": "1960px",
                            "bottom": "2260px",
                            "width": "400px",
                            "height": "450px",
                        },
                        "widgetId": "widget_599",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-91px",
                            "right": "269px",
                            "top": "2369px",
                            "bottom": "2439px",
                            "width": "360px",
                            "height": 140,
                        },
                        "widgetId": "widget_600",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "2561px",
                            "bottom": "2651px",
                            "width": "1200px",
                            "height": 118,
                        },
                        "widgetId": "widget_601",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-177px",
                            "right": "223px",
                            "top": "-199px",
                            "bottom": "201px",
                            "width": "400px",
                            "height": "480px",
                        },
                        "widgetId": "widget_602",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "3400px",
                            "bottom": "3640px",
                            "width": "1200px",
                            "height": "240px",
                        },
                        "widgetId": "widget_604",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "15": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "86px",
                            "right": "486px",
                            "top": "3260px",
                            "bottom": "3660px",
                            "width": "400px",
                            "height": "500px",
                        },
                        "widgetId": "widget_606",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-18px",
                            "right": "1182px",
                            "top": "4200px",
                            "bottom": "4440px",
                            "width": "1200px",
                            "height": 198,
                        },
                        "widgetId": "widget_607",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "17": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_609",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "width": "100%",
                        },
                    },
                    "18": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_610",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "width": "100%",
                        },
                    },
                    "19": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "4347px",
                            "bottom": "4447px",
                            "width": "1200px",
                            "height": "240px",
                        },
                        "widgetId": "widget_611",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "21": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "27px",
                            "right": "116px",
                            "top": "-50px",
                            "bottom": "38px",
                            "width": "88px",
                            "height": "320px",
                        },
                        "widgetId": "widget_613",
                        "setting": {"heightMode": "auto", "width": "100%"},
                    },
                },
                "order": [
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "21",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                    "16",
                    "17",
                    "18",
                    "19",
                ],
            },
            "layout_484": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_17",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_485": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_19",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_486": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 300},
                        "widgetId": "widget_21",
                        "setting": {
                            "heightMode": "ratio",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:50",
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_487": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_74",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_488": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "391px"},
                        "widgetId": "widget_544",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "391px"},
                        "widgetId": "widget_571",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_489": {
                "type": "FIXED",
                "content": {
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": "42.1053%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_570",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_864",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_865",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["2", "3", "4"],
            },
            "layout_490": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_568",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_569",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_491": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_572",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_573",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": "49.9590%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_574",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_492": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_575",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_576",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_493": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_578",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_579",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": "48.0429%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_580",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_494": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_581",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_582",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_495": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_584",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_585",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_586",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_496": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_587",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_588",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_497": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_76",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_498": {
                "type": "ROW",
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": 414},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_499": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "50%",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_35",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "height": "CUSTOM",
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1394%",
                            "right": 0,
                            "top": "22px",
                            "bottom": "76.3285%",
                            "width": "47.8606%",
                            "height": "18.3575%",
                        },
                        "widgetId": "widget_36",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "51.9932%",
                            "right": "1px",
                            "top": "18.4820%",
                            "bottom": "22.3647%",
                            "width": "47.8678%",
                            "height": "59.1533%",
                        },
                        "widgetId": "widget_37",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": "20.00px",
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_38",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"bottom": False, "top": True},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_503": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "offsetX": -10,
                            "autoProps": {"height": False},
                        },
                        "bbox": {"left": 2, "width": 2, "height": 30},
                        "widgetId": "widget_53",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "60px"},
                        "widgetId": "widget_128",
                        "setting": {"heightMode": "fit"},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_504": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_78",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_505": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 300},
                        "widgetId": "widget_70",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_506": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 50},
                        "widgetId": "widget_83",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "flex-end"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 3, "height": 50},
                        "widgetId": "widget_86",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "flex-end"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_507": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 4, "height": 70},
                        "widgetId": "widget_84",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 5, "width": 4, "height": 70},
                        "widgetId": "widget_87",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                        "bbox": {"left": 9, "width": 3, "height": "42px"},
                        "widgetId": "widget_88",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_508": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_603",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_509": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "sectionId": "section_7",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "3.1819%",
                            "top": "50%",
                            "bottom": "0.2506%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "51.5763%",
                            "top": "87.9151%",
                            "bottom": "10.00px",
                            "width": "15.1778%",
                            "height": "9.3280%",
                        },
                        "widgetId": "widget_441",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_510": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_442",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_443",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_444",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_511": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_469",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_470",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_471",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_512": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_475",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_476",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_477",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_513": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 400},
                        "widgetId": "widget_605",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_514": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {
                            "left": "1px",
                            "right": "57.3866%",
                            "top": "74px",
                            "bottom": "5.0300%",
                            "width": "42.4211%",
                            "height": "61.1872%",
                        },
                        "widgetId": "widget_121",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "49.3209%",
                            "right": 0,
                            "top": "34.1014%",
                            "bottom": 10,
                            "width": "50.8489%",
                            "height": "61.2903%",
                        },
                        "widgetId": "widget_122",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_515": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "490px"},
                        "widgetId": "widget_129",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": 400},
                        "widgetId": "widget_124",
                        "setting": {"heightMode": "fit"},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_516": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "9px",
                            "top": "10px",
                            "width": "29.8968%",
                            "height": "86.5369%",
                            "right": "67.07189999999997%",
                            "bottom": "8px",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35.0191%",
                            "width": 238,
                            "height": "43.6104%",
                            "right": "3.1237%",
                            "top": 10,
                            "bottom": 74,
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "274px",
                            "top": "69.6731999999997%",
                            "width": "27.8170%",
                            "height": "23.8722%",
                            "right": "2.2273999999999994%",
                            "bottom": "6.163100000000027%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_517": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.58%",
                            "top": "50%",
                            "width": "13.39%",
                            "height": "55.81%",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "21.20%",
                            "width": "226px",
                            "height": "35%",
                            "right": "39.05%",
                            "top": "50%",
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "230px",
                            "top": "50%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.25%",
                            "bottom": "20.93%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_518": {"type": "FIXED"},
            "layout_519": {"type": "FIXED"},
            "layout_520": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 400},
                        "widgetId": "widget_608",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_521": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "top": False,
                                "bottom": True,
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "3.2039%",
                            "top": 0,
                            "bottom": "71.6895%",
                            "width": "100.00%",
                            "height": "22.8311%",
                        },
                        "widgetId": "widget_154",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": "34.1014%",
                            "bottom": 10,
                            "width": "100.00%",
                            "height": "61.2903%",
                        },
                        "widgetId": "widget_155",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_522": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 89},
                        "widgetId": "widget_141",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": 89},
                        "widgetId": "widget_142",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": 89},
                        "widgetId": "widget_143",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": 89},
                        "widgetId": "widget_144",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_523": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 89},
                        "widgetId": "widget_146",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": 89},
                        "widgetId": "widget_147",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": 89},
                        "widgetId": "widget_148",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": 89},
                        "widgetId": "widget_149",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_524": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_151",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_526": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:120",
                        },
                        "bbox": {"left": 0, "width": 6, "height": "89px"},
                        "widgetId": "widget_577",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:120",
                        },
                        "bbox": {"left": 6, "width": 6, "height": "89px"},
                        "widgetId": "widget_583",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_527": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-263px",
                            "right": "1017px",
                            "top": "80px",
                            "bottom": "180px",
                            "width": "1280px",
                            "height": "160px",
                        },
                        "widgetId": "widget_615",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "220px",
                            "right": "580px",
                            "top": "316px",
                            "bottom": "386px",
                            "width": "360px",
                            "height": "80px",
                        },
                        "widgetId": "widget_616",
                        "setting": {
                            "heightMode": "auto",
                            "offsetX": 0,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "314px",
                            "right": "614px",
                            "top": "476px",
                            "bottom": "776px",
                            "width": "300px",
                            "height": "700px",
                        },
                        "widgetId": "widget_617",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "offsetX": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "941px",
                            "bottom": "1041px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_618",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-121px",
                            "right": "179px",
                            "top": "-191px",
                            "bottom": "200px",
                            "width": "300px",
                            "height": 343,
                        },
                        "widgetId": "widget_619",
                        "setting": {
                            "heightMode": "auto",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": "BOTTOM"},
                                }
                            },
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "1337px",
                            "bottom": "1437px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_620",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35px",
                            "right": "535px",
                            "top": "1229px",
                            "bottom": "1729px",
                            "width": "500px",
                            "height": "430px",
                        },
                        "widgetId": "widget_621",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "1833px",
                            "bottom": "1888px",
                            "width": "1200px",
                            "height": "60px",
                        },
                        "widgetId": "widget_622",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "42px",
                            "right": "1242px",
                            "top": "1940px",
                            "bottom": "2040px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_623",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "151px",
                            "right": "551px",
                            "top": "1960px",
                            "bottom": "2260px",
                            "width": "400px",
                            "height": "450px",
                        },
                        "widgetId": "widget_624",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-91px",
                            "right": "269px",
                            "top": "2369px",
                            "bottom": "2439px",
                            "width": "360px",
                            "height": 89,
                        },
                        "widgetId": "widget_625",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "2561px",
                            "bottom": "2651px",
                            "width": "1200px",
                            "height": 118,
                        },
                        "widgetId": "widget_626",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-177px",
                            "right": "223px",
                            "top": "-199px",
                            "bottom": "201px",
                            "width": "400px",
                            "height": "480px",
                        },
                        "widgetId": "widget_627",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "18px",
                            "right": "342px",
                            "top": "3883px",
                            "bottom": "4199px",
                            "width": "324px",
                            "height": "316px",
                        },
                        "widgetId": "widget_629",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "15": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "86px",
                            "right": "486px",
                            "top": "3260px",
                            "bottom": "3660px",
                            "width": "400px",
                            "height": "500px",
                        },
                        "widgetId": "widget_631",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-18px",
                            "right": "1182px",
                            "top": "4200px",
                            "bottom": "4440px",
                            "width": "1200px",
                            "height": "240px",
                        },
                        "widgetId": "widget_632",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "17": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_634",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "18": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_635",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "19": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "4347px",
                            "bottom": "4447px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_636",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                },
                "order": [
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                    "16",
                    "17",
                    "18",
                    "19",
                ],
            },
            "layout_529": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_639",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_530": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_640",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_531": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_641",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_532": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_642",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_533": {
                "type": "ROW",
                "content": {
                    "4": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_643",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["4"],
            },
            "layout_534": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_568",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_569",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_570",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_535": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_568",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_569",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_536": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_572",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_573",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_574",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_537": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_575",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_576",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_538": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_578",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_579",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_580",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_539": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_581",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_582",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_540": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_584",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_585",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_586",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_541": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_587",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_588",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_542": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_644",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_543": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_645",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_544": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "52.2007%",
                            "top": "50%",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_35",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.8125%",
                            "right": 0,
                            "top": 0,
                            "bottom": "73.3010%",
                            "width": "48.0000%",
                            "height": "26.6990%",
                        },
                        "widgetId": "widget_36",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.7941%",
                            "right": "1px",
                            "top": "22.7695%",
                            "bottom": "24.7405%",
                            "width": "46.9385%",
                            "height": "52.4900%",
                        },
                        "widgetId": "widget_37",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0126%",
                            "right": "7.5654%",
                            "top": "87.2434%",
                            "bottom": "16.00px",
                            "width": "40.4219%",
                            "height": "10.1449%",
                        },
                        "widgetId": "widget_38",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"bottom": False, "top": True},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_548": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_646",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_549": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_647",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_550": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_648",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_551": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_649",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_552": {
                "type": "ROW",
                "content": {
                    "3": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_650",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["3"],
            },
            "layout_553": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_651",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_554": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "sectionId": "section_7",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "3.1819%",
                            "top": "50%",
                            "bottom": "0.2506%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "51.5763%",
                            "top": "87.9151%",
                            "bottom": "10.00px",
                            "width": "15.1778%",
                            "height": "9.3280%",
                        },
                        "widgetId": "widget_441",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_555": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_442",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_443",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_444",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "AUTO",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_556": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_469",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_470",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_471",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_557": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_475",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_476",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_477",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_558": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_652",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_559": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                        "bbox": {
                            "left": "0px",
                            "right": 0,
                            "top": "1px",
                            "bottom": "35.6364%",
                            "width": "100.0000%",
                            "height": "64.0000%",
                        },
                        "widgetId": "widget_121",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "38.2026%",
                            "bottom": 36,
                            "width": "100.0000%",
                            "height": "48.7179%",
                        },
                        "widgetId": "widget_122",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO", "width": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_560": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_653",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_561": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "9px",
                            "top": "10px",
                            "width": "39.0599%",
                            "height": "86.5343%",
                            "right": "57.9401%",
                            "bottom": "10px",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "45.8711%",
                            "width": "153px",
                            "height": "43.6103%",
                            "right": "3.1198%",
                            "top": "10px",
                            "bottom": "75px",
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "274px",
                            "top": "69.6731999999997%",
                            "width": "27.8170%",
                            "height": "23.8722%",
                            "right": "2.2273999999999994%",
                            "bottom": "6.163100000000027%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_562": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.58%",
                            "top": "50%",
                            "width": "13.39%",
                            "height": "55.81%",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "21.20%",
                            "width": "226px",
                            "height": "35%",
                            "right": "39.05%",
                            "top": "50%",
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "230px",
                            "top": "50%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.25%",
                            "bottom": "20.93%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_563": {"type": "FIXED"},
            "layout_564": {"type": "FIXED"},
            "layout_565": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_654",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_566": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO", "width": "CUSTOM"},
                        },
                        "bbox": {
                            "left": 0,
                            "right": "0.0000%",
                            "top": 0,
                            "bottom": "76.8844%",
                            "width": "100.0000%",
                            "height": "23.1156%",
                        },
                        "widgetId": "widget_154",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": "34.1014%",
                            "bottom": 10,
                            "width": "100.00%",
                            "height": "61.2903%",
                        },
                        "widgetId": "widget_155",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_567": {
                "type": "ROW",
                "content": {
                    "4": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_655",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["4"],
            },
            "layout_568": {
                "type": "ROW",
                "content": {
                    "4": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_656",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["4"],
            },
            "layout_569": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_657",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_572": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "50px"},
                        "widgetId": "widget_17",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_573": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_19",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_574": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "296px", "height": 237},
                        "widgetId": "widget_21",
                        "setting": {
                            "heightMode": "fixed",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_575": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_74",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_576": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_544",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": 1.13,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_571",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": 1.13,
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_577",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": 1.13,
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_583",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": 1.13,
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_577": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "50px"},
                        "widgetId": "widget_76",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_578": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"width": "auto", "height": 414},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_579": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "offsetX": -10,
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"width": "auto", "height": 30},
                        "widgetId": "widget_53",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 60},
                        "widgetId": "widget_128",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "isPending": True,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_580": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "50px"},
                        "widgetId": "widget_78",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_581": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 300},
                        "widgetId": "widget_70",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_582": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "100.00%", "height": 50},
                        "widgetId": "widget_83",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM", "width": "CUSTOM"},
                            "style": {"alignSelf": "flex-end"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "302px", "height": "58px"},
                        "widgetId": "widget_84",
                    },
                },
                "order": ["0", "2"],
            },
            "layout_583": {
                "type": "COLUMN",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"width": "auto", "height": "58px"},
                        "widgetId": "widget_87",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"width": "auto", "height": 42},
                        "widgetId": "widget_88",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": "CUSTOM"}},
                        "bbox": {"width": "100.00%", "height": "50px"},
                        "widgetId": "widget_86",
                    },
                },
                "order": ["3", "1", "2"],
            },
            "layout_584": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "400px"},
                        "widgetId": "widget_628",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_585": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "296px", "height": 275},
                        "widgetId": "widget_630",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_586": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "490px"},
                        "widgetId": "widget_129",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 400},
                        "widgetId": "widget_124",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_587": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "296px", "height": 255},
                        "widgetId": "widget_633",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_588": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_141",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_142",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_143",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_144",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_589": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_146",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_147",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_148",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_149",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_590": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_151",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_670": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_838",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1506%",
                            "right": "17.3743%",
                            "top": 21,
                            "bottom": "86.2333%",
                            "width": "30.4523%",
                            "height": "8.7379%",
                        },
                        "widgetId": "widget_839",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0295%",
                            "right": 0,
                            "top": "21.0255%",
                            "bottom": "25.3337%",
                            "width": "48.1422%",
                            "height": "53.6408%",
                        },
                        "widgetId": "widget_840",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": 9,
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_841",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_671": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_842",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1506%",
                            "right": "17.3743%",
                            "top": 21,
                            "bottom": "86.2333%",
                            "width": "30.4523%",
                            "height": "8.7379%",
                        },
                        "widgetId": "widget_843",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0102%",
                            "right": "1px",
                            "top": "25.9473%",
                            "bottom": "17.6876%",
                            "width": "47.7975%",
                            "height": "56.3651%",
                        },
                        "widgetId": "widget_844",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": 9,
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_845",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_672": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_846",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.8125%",
                            "right": 0,
                            "top": 0,
                            "bottom": "73.3010%",
                            "width": "48.0000%",
                            "height": "26.6990%",
                        },
                        "widgetId": "widget_847",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.6458%",
                            "right": 0,
                            "top": "29.7747%",
                            "bottom": "19.9939%",
                            "width": "48.0000%",
                            "height": "50.2427%",
                        },
                        "widgetId": "widget_848",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0126%",
                            "right": "7.5654%",
                            "top": "87.2434%",
                            "bottom": "11px",
                            "width": "40.4219%",
                            "height": "10.1449%",
                        },
                        "widgetId": "widget_849",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_673": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_850",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1506%",
                            "right": "17.3743%",
                            "top": 21,
                            "bottom": "86.2333%",
                            "width": "30.4523%",
                            "height": "8.7379%",
                        },
                        "widgetId": "widget_851",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0295%",
                            "right": 0,
                            "top": "21.0255%",
                            "bottom": "25.3337%",
                            "width": "48.1422%",
                            "height": "53.6408%",
                        },
                        "widgetId": "widget_852",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": 9,
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_853",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_674": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_854",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1506%",
                            "right": "17.3743%",
                            "top": 21,
                            "bottom": "86.2333%",
                            "width": "30.4523%",
                            "height": "8.7379%",
                        },
                        "widgetId": "widget_855",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0102%",
                            "right": "1px",
                            "top": "25.9473%",
                            "bottom": "17.6876%",
                            "width": "47.7975%",
                            "height": "56.3651%",
                        },
                        "widgetId": "widget_856",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": 9,
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_857",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_675": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_858",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.8125%",
                            "right": 0,
                            "top": 0,
                            "bottom": "73.3010%",
                            "width": "48.0000%",
                            "height": "26.6990%",
                        },
                        "widgetId": "widget_859",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.6458%",
                            "right": 0,
                            "top": "29.7747%",
                            "bottom": "19.9939%",
                            "width": "48.0000%",
                            "height": "50.2427%",
                        },
                        "widgetId": "widget_860",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0126%",
                            "right": "7.5654%",
                            "top": "87.2434%",
                            "bottom": "11px",
                            "width": "40.4219%",
                            "height": "10.1449%",
                        },
                        "widgetId": "widget_861",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/ink/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "epic",
        "widgets": {
            "widget_16": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_16",
                "layouts": {"DEFAULT": {"LARGE": "layout_10"}},
            },
            "widget_17": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_17",
            },
            "widget_18": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_18",
                "layouts": {"DEFAULT": {"LARGE": "layout_11"}},
            },
            "widget_19": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. </span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_19",
            },
            "widget_20": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_20",
                "layouts": {"DEFAULT": {"LARGE": "layout_12"}},
            },
            "widget_21": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_21",
            },
            "widget_34": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_34",
                "layouts": {"DEFAULT": {"LARGE": "layout_19"}},
            },
            "widget_35": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_35",
            },
            "widget_36": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_36",
            },
            "widget_37": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_37",
            },
            "widget_38": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_38",
            },
            "widget_53": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_1",
                        "views": ["view_1", "view_31", "view_32"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "default",
                        "standard": {
                            "scrollable": False,
                            "gap": "10px",
                            "showIcon": True,
                            "alternateIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><path d="M6 0a6 6 0 110 12A6 6 0 016 0zm0 2a4 4 0 100 8 4 4 0 000-8z" fill="#000" fill-rule="nonzero"></path></svg>',
                                "properties": {
                                    "filename": "circle",
                                    "originalName": None,
                                    "path": None,
                                    "color": "#fff",
                                    "size": 8,
                                    "inlineSvg": False,
                                },
                            },
                            "activedIcon": {
                                "svg": '<svg viewBox="0 0 6 6" xmlns="http://www.w3.org/2000/svg"><circle cx="1104" cy="1049" r="3" transform="translate(-1101 -1046)" fill="#000" fill-rule="nonzero"></circle></svg>',
                                "properties": {"filename": "dot-10", "size": 8},
                            },
                            "showText": False,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_53",
            },
            "widget_54": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_54",
                "layouts": {"DEFAULT": {"LARGE": "layout_26"}},
            },
            "widget_69": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_69",
                "layouts": {"DEFAULT": {"LARGE": "layout_31"}},
            },
            "widget_70": {
                "uri": "widgets/common/embed/",
                "version": "1.5.0",
                "config": {"embedType": "url"},
                "id": "widget_70",
            },
            "widget_73": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_73",
                "layouts": {"DEFAULT": {"LARGE": "layout_33"}},
            },
            "widget_74": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_74",
            },
            "widget_75": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_75",
                "layouts": {"DEFAULT": {"LARGE": "layout_34"}},
            },
            "widget_76": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_76",
            },
            "widget_77": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_77",
                "layouts": {"DEFAULT": {"LARGE": "layout_35"}},
            },
            "widget_78": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_78",
            },
            "widget_82": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_82",
                "layouts": {"DEFAULT": {"LARGE": "layout_37"}},
            },
            "widget_83": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_83",
            },
            "widget_84": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_84",
            },
            "widget_85": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_85",
                "layouts": {"DEFAULT": {"LARGE": "layout_38"}},
            },
            "widget_86": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p><br></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_86",
            },
            "widget_87": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit.\ufeff</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_87",
            },
            "widget_88": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M2 8c0-.291.231-.527.516-.527h9.508L8.527 3.9a.535.535 0 010-.746.508.508 0 01.73 0L14 8l-4.743 4.846a.508.508 0 01-.73 0 .535.535 0 010-.746l3.497-3.573H2.516A.522.522 0 012 8z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "RIGHT",
                        },
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_3",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {
                                "iconProps": {"color": "var(--primary-600)", "size": 14}
                            },
                            "regular": {
                                "iconProps": {"color": "var(--dark-600)", "size": 14}
                            },
                        },
                        "themeStyle": {"quickStyleType": "tertiary"},
                    },
                },
                "id": "widget_88",
                "useDataSources": [],
            },
            "widget_119": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_119",
                "layouts": {"DEFAULT": {"LARGE": "layout_47"}},
            },
            "widget_120": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_120",
                "layouts": {"DEFAULT": {"LARGE": "layout_48"}},
            },
            "widget_121": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_121",
            },
            "widget_122": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_122",
            },
            "widget_123": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_123",
                "layouts": {"DEFAULT": {"LARGE": "layout_49"}},
            },
            "widget_124": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "id": "widget_124",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_50",
                        "MEDIUM": "layout_519",
                        "SMALL": "layout_564",
                    }
                },
                "config": {
                    "toolConifg": {"canZoom": True, "canHome": True, "canSearch": True},
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 18489297.737236,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -10764505.062957548,
                                "y": 4586292.545509916,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -13479548.307646286,
                            "ymin": 3055105.994901672,
                            "xmax": -8049461.8182688095,
                            "ymax": 6117479.09611816,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
            },
            "widget_128": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_1",
                        "views": ["view_1", "view_31", "view_32"],
                    },
                    "display": {
                        "advanced": True,
                        "vertical": False,
                        "type": "navButtonGroup",
                        "navStyle": "tertiary",
                        "standard": {
                            "showPageNumber": True,
                            "previousText": "",
                            "previousIcon": None,
                            "nextText": "",
                            "nextIcon": None,
                            "current": 1,
                            "totalPage": 4,
                            "disablePrevious": True,
                            "step": 1,
                        },
                        "variant": {
                            "root": {},
                            "item": {
                                "default": {
                                    "color": "#181818",
                                    "bg": "transparent",
                                    "border": {"width": "0px", "color": "transparent"},
                                    "size": "18px",
                                    "icon": {"size": "18px"},
                                },
                                "hover": {
                                    "color": "var(--dark-100)",
                                    "bg": "transparent",
                                    "border": {"width": "0px", "color": "transparent"},
                                    "size": "18px",
                                    "icon": {"size": "18px"},
                                },
                                "active": {"color": "#fff", "bg": "#076fe5"},
                                "disabled": {
                                    "color": "#8b8b8b",
                                    "bg": "transparent",
                                    "border": {"width": "0px", "color": "transparent"},
                                    "size": "18px",
                                    "icon": {"size": "18px"},
                                },
                                "focus": {
                                    "border": {"width": "0px", "color": "transparent"}
                                },
                            },
                        },
                    },
                },
                "id": "widget_128",
            },
            "widget_129": {
                "uri": "widgets/common/list/",
                "version": "1.5.0",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE5",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "scrollBarOpen": True,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {"LARGE": {"width": 386.66, "height": 152.28}},
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0, 0, 0, 1)",
                                    "width": "1px",
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {"LARGE": {"width": 386.66, "height": 152.28}},
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0, 0, 0, 1)",
                                    "width": "1px",
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {"LARGE": {"width": 386.66, "height": 152.28}},
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(133, 55, 241, 1)",
                                    "width": "1px",
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                },
                "id": "widget_129",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_57",
                        "MEDIUM": "layout_516",
                        "SMALL": "layout_561",
                    },
                    "SELECTED": {
                        "LARGE": "layout_54",
                        "MEDIUM": "layout_517",
                        "SMALL": "layout_562",
                    },
                    "HOVER": {
                        "LARGE": "layout_55",
                        "MEDIUM": "layout_518",
                        "SMALL": "layout_563",
                    },
                },
                "useDataSourcesEnabled": True,
            },
            "widget_133": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "cropShape": "rectangle",
                                "svgPath": "M0 0v28h28V0H0z",
                                "svgViewBox": "0 0 28 28",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_133",
            },
            "widget_134": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 18px;">Double click to edit text</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_134",
            },
            "widget_135": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {"toolTip": "", "linkParam": {}},
                    "styleConfig": {
                        "name": "Default",
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                            "clicked": {},
                        },
                        "useCustom": False,
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_135",
            },
            "widget_140": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_140",
                "layouts": {"DEFAULT": {"LARGE": "layout_60"}},
            },
            "widget_141": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_141",
            },
            "widget_142": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_142",
            },
            "widget_143": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_143",
            },
            "widget_144": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_144",
            },
            "widget_145": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_145",
                "layouts": {"DEFAULT": {"LARGE": "layout_61"}},
            },
            "widget_146": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_146",
            },
            "widget_147": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_147",
            },
            "widget_148": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_148",
            },
            "widget_149": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_149",
            },
            "widget_150": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_150",
                "layouts": {"DEFAULT": {"LARGE": "layout_62"}},
            },
            "widget_151": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_151",
            },
            "widget_152": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_152",
                "layouts": {"DEFAULT": {"LARGE": "layout_63"}},
            },
            "widget_153": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_153",
                "layouts": {"DEFAULT": {"LARGE": "layout_64"}},
            },
            "widget_154": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_154",
            },
            "widget_155": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_155",
            },
            "widget_439": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_439",
                "layouts": {"DEFAULT": {"LARGE": "layout_342"}},
            },
            "widget_440": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_440",
                "layouts": {"DEFAULT": {"LARGE": "layout_343"}},
            },
            "widget_441": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_7",
                        "views": ["view_28", "view_29", "view_30"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "default",
                        "standard": {
                            "scrollable": False,
                            "gap": "10px",
                            "showIcon": True,
                            "alternateIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><path d="M6 0a6 6 0 110 12A6 6 0 016 0zm0 2a4 4 0 100 8 4 4 0 000-8z" fill="#000" fill-rule="nonzero"></path></svg>',
                                "properties": {
                                    "filename": "circle",
                                    "originalName": None,
                                    "path": None,
                                    "color": "#fff",
                                    "size": 8,
                                    "inlineSvg": False,
                                },
                            },
                            "activedIcon": {
                                "svg": '<svg viewBox="0 0 6 6" xmlns="http://www.w3.org/2000/svg"><circle cx="1104" cy="1049" r="3" transform="translate(-1101 -1046)" fill="#000" fill-rule="nonzero"></circle></svg>',
                                "properties": {"filename": "dot-10", "size": 8},
                            },
                            "showText": False,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_441",
            },
            "widget_442": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_442",
            },
            "widget_443": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_443",
                "style": {
                    "borderRadius": {"number": [100], "unit": "px"},
                    "border": {"type": "solid", "color": "#000000", "width": "2px"},
                },
            },
            "widget_444": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 24px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_444",
            },
            "widget_469": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_469",
            },
            "widget_470": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_470",
                "style": {
                    "borderRadius": {"number": [100], "unit": "px"},
                    "border": {"type": "solid", "color": "#000000", "width": "2px"},
                },
            },
            "widget_471": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 24px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_471",
            },
            "widget_475": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_475",
            },
            "widget_476": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_476",
                "style": {
                    "borderRadius": {"number": [100], "unit": "px"},
                    "border": {"type": "solid", "color": "#000000", "width": "2px"},
                },
            },
            "widget_477": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 24px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_477",
            },
            "widget_543": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_543",
                "layouts": {"DEFAULT": {"LARGE": "layout_466"}},
            },
            "widget_544": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 350.625,
                    "widgetHeight": 396.203125,
                },
                "id": "widget_544",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_474",
                        "MEDIUM": "layout_489",
                        "SMALL": "layout_534",
                    },
                    "HOVER": {
                        "LARGE": "layout_475",
                        "MEDIUM": "layout_490",
                        "SMALL": "layout_535",
                    },
                },
            },
            "widget_568": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_568",
            },
            "widget_569": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_569",
            },
            "widget_570": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_570",
            },
            "widget_571": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 350.625,
                    "widgetHeight": 396.203125,
                },
                "id": "widget_571",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_476",
                        "MEDIUM": "layout_491",
                        "SMALL": "layout_536",
                    },
                    "HOVER": {
                        "LARGE": "layout_477",
                        "MEDIUM": "layout_492",
                        "SMALL": "layout_537",
                    },
                },
            },
            "widget_572": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_572",
            },
            "widget_573": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_573",
            },
            "widget_574": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_574",
            },
            "widget_575": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_575",
            },
            "widget_576": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_576",
            },
            "widget_577": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 350.625,
                    "widgetHeight": 396.203125,
                },
                "id": "widget_577",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_478",
                        "MEDIUM": "layout_493",
                        "SMALL": "layout_538",
                    },
                    "HOVER": {
                        "LARGE": "layout_479",
                        "MEDIUM": "layout_494",
                        "SMALL": "layout_539",
                    },
                },
            },
            "widget_578": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_578",
            },
            "widget_579": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_579",
            },
            "widget_580": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_580",
            },
            "widget_581": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_581",
            },
            "widget_582": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_582",
            },
            "widget_583": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 350.625,
                    "widgetHeight": 396.203125,
                },
                "id": "widget_583",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_480",
                        "MEDIUM": "layout_495",
                        "SMALL": "layout_540",
                    },
                    "HOVER": {
                        "LARGE": "layout_481",
                        "MEDIUM": "layout_496",
                        "SMALL": "layout_541",
                    },
                },
            },
            "widget_584": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_584",
            },
            "widget_585": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_585",
            },
            "widget_586": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_586",
            },
            "widget_587": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_587",
            },
            "widget_588": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_588",
            },
            "widget_590": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_590",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_484"}},
            },
            "widget_591": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_591",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_485"}},
            },
            "widget_592": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_592",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_486"}},
            },
            "widget_593": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_593",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_487"}},
            },
            "widget_594": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_594",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_488"}},
            },
            "widget_595": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_595",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_497"}},
            },
            "widget_596": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_596",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_498"}},
            },
            "widget_597": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_597",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_503"}},
            },
            "widget_598": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_598",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_504"}},
            },
            "widget_599": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_599",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_505"}},
            },
            "widget_600": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_600",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_506"}},
            },
            "widget_601": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_601",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_507"}},
            },
            "widget_602": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_602",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_508"}},
            },
            "widget_603": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_603",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_509"}},
            },
            "widget_604": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_604",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_513"}},
            },
            "widget_605": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_605",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_514"}},
            },
            "widget_606": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_606",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_515"}},
            },
            "widget_607": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_607",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_520"}},
            },
            "widget_608": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_608",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_521"}},
            },
            "widget_609": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_609",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_522"}},
            },
            "widget_610": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_610",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_523"}},
            },
            "widget_611": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_611",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_524"}},
            },
            "widget_613": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_613",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_526"}},
            },
            "widget_615": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_615",
                "layouts": {"DEFAULT": {"SMALL": "layout_529"}},
            },
            "widget_616": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_616",
                "layouts": {"DEFAULT": {"SMALL": "layout_530"}},
            },
            "widget_617": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_617",
                "layouts": {"DEFAULT": {"SMALL": "layout_531"}},
            },
            "widget_618": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_618",
                "layouts": {"DEFAULT": {"SMALL": "layout_532"}},
            },
            "widget_619": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_619",
                "layouts": {"DEFAULT": {"SMALL": "layout_533"}},
            },
            "widget_620": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_620",
                "layouts": {"DEFAULT": {"SMALL": "layout_542"}},
            },
            "widget_621": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_621",
                "layouts": {"DEFAULT": {"SMALL": "layout_543"}},
            },
            "widget_622": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_622",
                "layouts": {"DEFAULT": {"SMALL": "layout_548"}},
            },
            "widget_623": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_623",
                "layouts": {"DEFAULT": {"SMALL": "layout_549"}},
            },
            "widget_624": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_624",
                "layouts": {"DEFAULT": {"SMALL": "layout_550"}},
            },
            "widget_625": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_625",
                "layouts": {"DEFAULT": {"SMALL": "layout_551"}},
            },
            "widget_626": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_626",
                "layouts": {"DEFAULT": {"SMALL": "layout_552"}},
            },
            "widget_627": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_627",
                "layouts": {"DEFAULT": {"SMALL": "layout_553"}},
            },
            "widget_628": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_628",
                "layouts": {"DEFAULT": {"SMALL": "layout_554"}},
            },
            "widget_629": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_629",
                "layouts": {"DEFAULT": {"SMALL": "layout_558"}},
            },
            "widget_630": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_630",
                "layouts": {"DEFAULT": {"SMALL": "layout_559"}},
            },
            "widget_631": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_631",
                "layouts": {"DEFAULT": {"SMALL": "layout_560"}},
            },
            "widget_632": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_632",
                "layouts": {"DEFAULT": {"SMALL": "layout_565"}},
            },
            "widget_633": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_633",
                "layouts": {"DEFAULT": {"SMALL": "layout_566"}},
            },
            "widget_634": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_634",
                "layouts": {"DEFAULT": {"SMALL": "layout_567"}},
            },
            "widget_635": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_635",
                "layouts": {"DEFAULT": {"SMALL": "layout_568"}},
            },
            "widget_636": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 40, 10], "unit": "px"}},
                },
                "id": "widget_636",
                "layouts": {"DEFAULT": {"SMALL": "layout_569"}},
            },
            "widget_639": {
                "uri": "widgets/layout/column/",
                "id": "widget_639",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_572"}},
            },
            "widget_640": {
                "uri": "widgets/layout/column/",
                "id": "widget_640",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_573"}},
            },
            "widget_641": {
                "uri": "widgets/layout/column/",
                "id": "widget_641",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_574"}},
            },
            "widget_642": {
                "uri": "widgets/layout/column/",
                "id": "widget_642",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_575"}},
            },
            "widget_643": {
                "uri": "widgets/layout/column/",
                "id": "widget_643",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_576"}},
            },
            "widget_644": {
                "uri": "widgets/layout/column/",
                "id": "widget_644",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_577"}},
            },
            "widget_645": {
                "uri": "widgets/layout/column/",
                "id": "widget_645",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_578"}},
            },
            "widget_646": {
                "uri": "widgets/layout/column/",
                "id": "widget_646",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_579"}},
            },
            "widget_647": {
                "uri": "widgets/layout/column/",
                "id": "widget_647",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_580"}},
            },
            "widget_648": {
                "uri": "widgets/layout/column/",
                "id": "widget_648",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_581"}},
            },
            "widget_649": {
                "uri": "widgets/layout/column/",
                "id": "widget_649",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_582"}},
            },
            "widget_650": {
                "uri": "widgets/layout/column/",
                "id": "widget_650",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_583"}},
            },
            "widget_651": {
                "uri": "widgets/layout/column/",
                "id": "widget_651",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_584"}},
            },
            "widget_652": {
                "uri": "widgets/layout/column/",
                "id": "widget_652",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_585"}},
            },
            "widget_653": {
                "uri": "widgets/layout/column/",
                "id": "widget_653",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_586"}},
            },
            "widget_654": {
                "uri": "widgets/layout/column/",
                "id": "widget_654",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_587"}},
            },
            "widget_655": {
                "uri": "widgets/layout/column/",
                "id": "widget_655",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_588"}},
            },
            "widget_656": {
                "uri": "widgets/layout/column/",
                "id": "widget_656",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_589"}},
            },
            "widget_657": {
                "uri": "widgets/layout/column/",
                "id": "widget_657",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_590"}},
            },
            "widget_838": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_838",
            },
            "widget_839": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_839",
            },
            "widget_840": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_840",
            },
            "widget_841": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_841",
            },
            "widget_842": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_842",
            },
            "widget_843": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_843",
            },
            "widget_844": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_844",
            },
            "widget_845": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_845",
            },
            "widget_846": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_846",
            },
            "widget_847": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_847",
            },
            "widget_848": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_848",
            },
            "widget_849": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_849",
            },
            "widget_850": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_850",
            },
            "widget_851": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_851",
            },
            "widget_852": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_852",
            },
            "widget_853": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_853",
            },
            "widget_854": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_854",
            },
            "widget_855": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_855",
            },
            "widget_856": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_856",
            },
            "widget_857": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_857",
            },
            "widget_858": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_858",
            },
            "widget_859": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_859",
            },
            "widget_860": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_860",
            },
            "widget_861": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_861",
            },
            "widget_862": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_862",
            },
            "widget_863": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_863",
            },
            "widget_864": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_864",
            },
            "widget_865": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_865",
            },
        },
        "widgetsManifest": {},
        "views": {
            "view_1": {
                "id": "view_1",
                "layout": {
                    "LARGE": "layout_20",
                    "MEDIUM": "layout_499",
                    "SMALL": "layout_544",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "",
                "backgroundIMage": {},
            },
            "view_28": {
                "id": "view_28",
                "layout": {
                    "LARGE": "layout_344",
                    "MEDIUM": "layout_510",
                    "SMALL": "layout_555",
                },
            },
            "view_29": {
                "id": "view_29",
                "layout": {
                    "LARGE": "layout_379",
                    "MEDIUM": "layout_511",
                    "SMALL": "layout_556",
                },
            },
            "view_30": {
                "id": "view_30",
                "layout": {
                    "LARGE": "layout_381",
                    "MEDIUM": "layout_512",
                    "SMALL": "layout_557",
                },
            },
            "view_31": {
                "id": "view_31",
                "layout": {
                    "LARGE": "layout_670",
                    "MEDIUM": "layout_671",
                    "SMALL": "layout_672",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "",
                "backgroundIMage": {},
            },
            "view_32": {
                "id": "view_32",
                "layout": {
                    "LARGE": "layout_673",
                    "MEDIUM": "layout_674",
                    "SMALL": "layout_675",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "",
                "backgroundIMage": {},
            },
        },
        "sections": {
            "section_1": {
                "id": "section_1",
                "views": ["view_1", "view_31", "view_32"],
                "transition": {"type": "PUSH"},
            },
            "section_7": {
                "id": "section_7",
                "views": ["view_28", "view_29", "view_30"],
                "transition": {"type": "PUSH"},
            },
        },
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_0": []}],
        "timestamp": 1622673507976,
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "snapshot": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_0",
                    "MEDIUM": "layout_821",
                    "SMALL": "layout_853",
                },
                "isDefault": True,
                "isVisible": True,
                "header": False,
                "bodyBackgroundColor": "",
                "footer": False,
            }
        },
        "layouts": {
            "layout_0": {
                "type": "FLOW",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1280px",
                            "top": "512px",
                            "bottom": "2782px",
                            "width": "1280px",
                            "height": "1800px",
                        },
                        "widgetId": "widget_6",
                        "setting": {"heightMode": "auto", "width": "100%"},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "525px",
                            "right": "905px",
                            "top": "-30px",
                            "bottom": "30px",
                            "width": "380px",
                            "height": 93,
                        },
                        "widgetId": "widget_7",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6px",
                            "right": "1286px",
                            "top": "1px",
                            "bottom": "501px",
                            "width": "1280px",
                            "height": "480px",
                        },
                        "widgetId": "widget_152",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "isFloating": True,
                            "lockParent": True,
                            "floatingArea": 9,
                            "offsetY": 40,
                            "offsetX": 40,
                            "heightMode": "fixed",
                            "width": "64px",
                        },
                        "bbox": {
                            "top": 0,
                            "left": 61.6282958984375,
                            "right": 0,
                            "width": "128px",
                            "height": "48px",
                            "bottom": 0,
                        },
                        "widgetId": "widget_216",
                    },
                },
                "order": ["3", "2", "1", "6"],
                "setting": {
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}}
                },
            },
            "layout_3": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "1460.00px"},
                        "setting": {
                            "heightMode": "fixed",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {
                                        "direction": None,
                                        "configType": "DEFAULT",
                                    },
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "flex-start"},
                            "aspectRatio": 1.01,
                        },
                    }
                },
            },
            "layout_5": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": 60},
                        "widgetId": "widget_8",
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_103": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "480.00px"},
                        "widgetId": "widget_153",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:40",
                        },
                    }
                },
            },
            "layout_104": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_237",
                        "setting": {
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "44.3056%",
                            "bottom": "43.9382%",
                            "width": 960,
                            "height": "11.7563%",
                        },
                        "widgetId": "widget_238",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                            },
                            "vCenter": False,
                            "hCenter": True,
                        },
                        "isPending": False,
                    },
                },
            },
            "layout_161": {
                "type": "FIXED",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "960.00px",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_265",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_784",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_803",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_822",
                        "isPending": False,
                    },
                },
                "order": ["1", "2", "3", "4"],
            },
            "layout_163": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_266",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_267",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_268",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_164": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_269",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_270",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_272",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "3"],
            },
            "layout_759": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_785",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_786",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_787",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_762": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_794",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_795",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_796",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_765": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_804",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_805",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_806",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_768": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_813",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_814",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_815",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_771": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_823",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_824",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_825",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_774": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_832",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_833",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_834",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_777": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "960.00px",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_841",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_854",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_867",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_880",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_778": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_842",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_843",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_844",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_780": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_848",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_849",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_850",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_782": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_855",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_856",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_857",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_784": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_861",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_862",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_863",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_786": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_868",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_869",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_870",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_788": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_874",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_875",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_876",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_790": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_881",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_882",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_883",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_792": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_887",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_888",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_889",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_799": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "960.00px",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_906",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_919",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_932",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_945",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_800": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_907",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_908",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_909",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_802": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_913",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_914",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_915",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_804": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_920",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_921",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_922",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_806": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_926",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_927",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_928",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_808": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_933",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_934",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_935",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_810": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_939",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_940",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_941",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_812": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_946",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_947",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_948",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_814": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_952",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_953",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_954",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_821": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6px",
                            "right": "1286px",
                            "top": "1px",
                            "bottom": "501px",
                            "width": "1280px",
                            "height": "480px",
                        },
                        "widgetId": "widget_971",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "525px",
                            "right": "905px",
                            "top": "-30px",
                            "bottom": "30px",
                            "width": "380px",
                            "height": 93,
                        },
                        "widgetId": "widget_973",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1280px",
                            "top": "512px",
                            "bottom": "2782px",
                            "width": "1280px",
                            "height": "1800px",
                        },
                        "widgetId": "widget_974",
                        "setting": {"heightMode": "auto", "width": "100%"},
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "isFloating": True,
                            "lockParent": True,
                            "floatingArea": 9,
                            "offsetY": 40,
                            "offsetX": 40,
                            "heightMode": "fixed",
                            "width": "64px",
                        },
                        "bbox": {
                            "top": 0,
                            "left": 61.6282958984375,
                            "right": 0,
                            "width": "128px",
                            "height": "48px",
                            "bottom": 0,
                        },
                        "widgetId": "widget_216",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_822": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "480.00px"},
                        "widgetId": "widget_972",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:40",
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_823": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_237",
                        "setting": {
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "44.3056%",
                            "bottom": "43.9382%",
                            "width": "90.00%",
                            "height": "11.7563%",
                        },
                        "widgetId": "widget_238",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                            },
                            "vCenter": False,
                            "hCenter": True,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_824": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": "60px"},
                        "widgetId": "widget_8",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_825": {
                "type": "ROW",
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "1460.00px"},
                        "setting": {
                            "heightMode": "fixed",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {
                                        "direction": None,
                                        "configType": "DEFAULT",
                                    },
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "flex-start"},
                            "aspectRatio": 1.01,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_826": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_265",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_784",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_803",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_822",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_827": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_266",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_267",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_268",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_828": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_269",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_270",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_272",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_829": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_785",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_786",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_787",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_830": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_794",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_795",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_796",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_831": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_804",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_805",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_806",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_832": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_813",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_814",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_815",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_833": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_823",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_824",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_825",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_834": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_832",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_833",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_834",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_835": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_841",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_854",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_867",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_880",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_836": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_842",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_843",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_844",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_837": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_848",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_849",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_850",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_838": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_855",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_856",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_857",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_839": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_861",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_862",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_863",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_840": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_868",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_869",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_870",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_841": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_874",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_875",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_876",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_842": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_881",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_882",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_883",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_843": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_887",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_888",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_889",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_844": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_906",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_919",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_932",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_945",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_845": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_907",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_908",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_909",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_846": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_913",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_914",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_915",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_847": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_920",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_921",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_922",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_848": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_926",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_927",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_928",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_849": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_933",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_934",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_935",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_850": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_939",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_940",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_941",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_851": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_946",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_947",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_948",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_852": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_952",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_953",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_954",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_853": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6px",
                            "right": "1286px",
                            "top": "1px",
                            "bottom": "501px",
                            "width": "1280px",
                            "height": "480px",
                        },
                        "widgetId": "widget_975",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "525px",
                            "right": "905px",
                            "top": "-30px",
                            "bottom": "30px",
                            "width": "380px",
                            "height": 93,
                        },
                        "widgetId": "widget_977",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1280px",
                            "top": "512px",
                            "bottom": "2782px",
                            "width": "1280px",
                            "height": "1800px",
                        },
                        "widgetId": "widget_978",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "isFloating": True,
                            "lockParent": True,
                            "floatingArea": 9,
                            "offsetY": 40,
                            "offsetX": 40,
                            "heightMode": "auto",
                            "width": "64px",
                        },
                        "bbox": {
                            "top": 0,
                            "left": 61.6282958984375,
                            "right": 0,
                            "width": "128px",
                            "height": "48px",
                            "bottom": 0,
                        },
                        "widgetId": "widget_216",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_854": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_979",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_855": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_237",
                        "setting": {
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "50%",
                            "bottom": "43.9382%",
                            "width": "90.00%",
                            "height": "11.7563%",
                        },
                        "widgetId": "widget_238",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_856": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_980",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_857": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_981",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_858": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_265",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_784",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_803",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_822",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_859": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_266",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_267",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_268",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_860": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_269",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_270",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_272",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_861": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_785",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_786",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_787",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_862": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_794",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_795",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_796",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_863": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_804",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_805",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_806",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_864": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_813",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_814",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_815",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_865": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_823",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_824",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_825",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_866": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_832",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_833",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_834",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_867": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_841",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_854",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_867",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_880",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_868": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_842",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_843",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_844",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_869": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_848",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_849",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_850",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_870": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_855",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_856",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_857",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_871": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_861",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_862",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_863",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_872": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_868",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_869",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_870",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_873": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_874",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_875",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_876",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_874": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_881",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_882",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_883",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_875": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_887",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_888",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_889",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_876": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_906",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_919",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_932",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_945",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_877": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_907",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_908",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_909",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_878": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_913",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_914",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_915",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_879": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_920",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_921",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_922",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_880": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_926",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_927",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_928",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_881": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_933",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_934",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_935",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_882": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_939",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_940",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_941",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_883": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_946",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_947",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_948",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_884": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_952",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_953",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_954",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_885": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "320.00px"},
                        "widgetId": "widget_976",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:40",
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_886": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 60},
                        "widgetId": "widget_8",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_887": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"width": "auto", "height": "1460.00px"},
                        "setting": {
                            "heightMode": "fixed",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {
                                        "direction": None,
                                        "configType": "DEFAULT",
                                    },
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "flex-start"},
                            "aspectRatio": 1.01,
                        },
                    }
                },
                "order": ["0"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/morandi/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "snapshot",
        "widgets": {
            "widget_6": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_6",
                "layouts": {"DEFAULT": {"LARGE": "layout_3"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_7": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_7",
                "layouts": {"DEFAULT": {"LARGE": "layout_5"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_8": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_1",
                        "views": ["view_2", "view_16", "view_17"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "underline",
                        "standard": {
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_8",
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_149": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_149",
            },
            "widget_150": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="text-align: center;"><strong style="color: var(--dark-800); font-size: 32px;">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_150",
            },
            "widget_151": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="text-align: center;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_151",
            },
            "widget_152": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_152",
                "layouts": {"DEFAULT": {"LARGE": "layout_103"}},
            },
            "widget_153": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_153",
                "layouts": {"DEFAULT": {"LARGE": "layout_104"}},
            },
            "widget_216": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "textExpression": None,
                        "linkParam": {"linkType": "NONE", "value": ""},
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_216",
                "useDataSources": [],
            },
            "widget_237": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_237",
            },
            "widget_238": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px; text-shadow: rgba(0, 0, 0, 0.2) 0px 2px 0px, rgba(0, 0, 0, 0.3) -2px 4px 2px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_238",
            },
            "widget_265": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_4",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_265",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_163",
                        "MEDIUM": "layout_827",
                        "SMALL": "layout_859",
                    },
                    "HOVER": {
                        "LARGE": "layout_164",
                        "MEDIUM": "layout_828",
                        "SMALL": "layout_860",
                    },
                },
            },
            "widget_266": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_266",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_267": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_267",
            },
            "widget_268": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title\ufeff</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_268",
            },
            "widget_269": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_269",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_270": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_270",
            },
            "widget_272": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_272",
            },
            "widget_784": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_5",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_784",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_759",
                        "MEDIUM": "layout_829",
                        "SMALL": "layout_861",
                    },
                    "HOVER": {
                        "LARGE": "layout_762",
                        "MEDIUM": "layout_830",
                        "SMALL": "layout_862",
                    },
                },
            },
            "widget_785": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_785",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_786": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_786",
            },
            "widget_787": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_787",
            },
            "widget_794": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_794",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_795": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_795",
            },
            "widget_796": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_796",
            },
            "widget_803": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_6",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_803",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_765",
                        "MEDIUM": "layout_831",
                        "SMALL": "layout_863",
                    },
                    "HOVER": {
                        "LARGE": "layout_768",
                        "MEDIUM": "layout_832",
                        "SMALL": "layout_864",
                    },
                },
            },
            "widget_804": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_804",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_805": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_805",
            },
            "widget_806": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_806",
            },
            "widget_813": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_813",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_814": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_814",
            },
            "widget_815": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_815",
            },
            "widget_822": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_7",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_822",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_771",
                        "MEDIUM": "layout_833",
                        "SMALL": "layout_865",
                    },
                    "HOVER": {
                        "LARGE": "layout_774",
                        "MEDIUM": "layout_834",
                        "SMALL": "layout_866",
                    },
                },
            },
            "widget_823": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_823",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_824": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_824",
            },
            "widget_825": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_825",
            },
            "widget_832": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_832",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_833": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_833",
            },
            "widget_834": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_834",
            },
            "widget_841": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_8",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_841",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_778",
                        "MEDIUM": "layout_836",
                        "SMALL": "layout_868",
                    },
                    "HOVER": {
                        "LARGE": "layout_780",
                        "MEDIUM": "layout_837",
                        "SMALL": "layout_869",
                    },
                },
            },
            "widget_842": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_842",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_843": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_843",
            },
            "widget_844": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_844",
            },
            "widget_848": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_848",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_849": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_849",
            },
            "widget_850": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_850",
            },
            "widget_854": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_9",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_854",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_782",
                        "MEDIUM": "layout_838",
                        "SMALL": "layout_870",
                    },
                    "HOVER": {
                        "LARGE": "layout_784",
                        "MEDIUM": "layout_839",
                        "SMALL": "layout_871",
                    },
                },
            },
            "widget_855": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_855",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_856": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_856",
            },
            "widget_857": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_857",
            },
            "widget_861": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_861",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_862": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_862",
            },
            "widget_863": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_863",
            },
            "widget_867": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_10",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_867",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_786",
                        "MEDIUM": "layout_840",
                        "SMALL": "layout_872",
                    },
                    "HOVER": {
                        "LARGE": "layout_788",
                        "MEDIUM": "layout_841",
                        "SMALL": "layout_873",
                    },
                },
            },
            "widget_868": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_868",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_869": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_869",
            },
            "widget_870": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_870",
            },
            "widget_874": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_874",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_875": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_875",
            },
            "widget_876": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_876",
            },
            "widget_880": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "NONE",
                        "openType": "_self",
                        "value": "page_1,view_11",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_880",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_790",
                        "MEDIUM": "layout_842",
                        "SMALL": "layout_874",
                    },
                    "HOVER": {
                        "LARGE": "layout_792",
                        "MEDIUM": "layout_843",
                        "SMALL": "layout_875",
                    },
                },
            },
            "widget_881": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_881",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_882": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_882",
            },
            "widget_883": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_883",
            },
            "widget_887": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_887",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_888": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_888",
            },
            "widget_889": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_889",
            },
            "widget_906": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "NONE",
                        "openType": "_self",
                        "value": "page_1,view_12",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_906",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_800",
                        "MEDIUM": "layout_845",
                        "SMALL": "layout_877",
                    },
                    "HOVER": {
                        "LARGE": "layout_802",
                        "MEDIUM": "layout_846",
                        "SMALL": "layout_878",
                    },
                },
            },
            "widget_907": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_907",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_908": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_908",
            },
            "widget_909": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeff\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_909",
            },
            "widget_913": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_913",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_914": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_914",
            },
            "widget_915": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_915",
            },
            "widget_919": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_13",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_919",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_804",
                        "MEDIUM": "layout_847",
                        "SMALL": "layout_879",
                    },
                    "HOVER": {
                        "LARGE": "layout_806",
                        "MEDIUM": "layout_848",
                        "SMALL": "layout_880",
                    },
                },
            },
            "widget_920": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_920",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_921": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_921",
            },
            "widget_922": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_922",
            },
            "widget_926": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_926",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_927": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_927",
            },
            "widget_928": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_928",
            },
            "widget_932": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "NONE",
                        "openType": "_self",
                        "value": "page_1,view_14",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_932",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_808",
                        "MEDIUM": "layout_849",
                        "SMALL": "layout_881",
                    },
                    "HOVER": {
                        "LARGE": "layout_810",
                        "MEDIUM": "layout_850",
                        "SMALL": "layout_882",
                    },
                },
            },
            "widget_933": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_933",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_934": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_934",
            },
            "widget_935": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_935",
            },
            "widget_939": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_939",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_940": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_940",
            },
            "widget_941": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_941",
            },
            "widget_945": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "NONE",
                        "openType": "_self",
                        "value": "page_1,view_15",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_945",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_812",
                        "MEDIUM": "layout_851",
                        "SMALL": "layout_883",
                    },
                    "HOVER": {
                        "LARGE": "layout_814",
                        "MEDIUM": "layout_852",
                        "SMALL": "layout_884",
                    },
                },
            },
            "widget_946": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_946",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_947": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_947",
            },
            "widget_948": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_948",
            },
            "widget_952": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_952",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_953": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_953",
            },
            "widget_954": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_954",
            },
            "widget_971": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_971",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_822"}},
            },
            "widget_972": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_972",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_823"}},
            },
            "widget_973": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_973",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_824"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_974": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_974",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_825"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_975": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_975",
                "layouts": {"DEFAULT": {"SMALL": "layout_854"}},
            },
            "widget_976": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_976",
                "layouts": {"DEFAULT": {"SMALL": "layout_855"}},
            },
            "widget_977": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_977",
                "layouts": {"DEFAULT": {"SMALL": "layout_856"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_978": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_978",
                "layouts": {"DEFAULT": {"SMALL": "layout_857"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_979": {
                "uri": "widgets/layout/column/",
                "id": "widget_979",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_885"}},
            },
            "widget_980": {
                "uri": "widgets/layout/column/",
                "id": "widget_980",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_886"}},
            },
            "widget_981": {
                "uri": "widgets/layout/column/",
                "id": "widget_981",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_887"}},
            },
        },
        "widgetsManifest": {},
        "views": {
            "view_2": {
                "id": "view_2",
                "layout": {
                    "LARGE": "layout_161",
                    "MEDIUM": "layout_826",
                    "SMALL": "layout_858",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "",
                "backgroundIMage": {},
            },
            "view_16": {
                "id": "view_16",
                "layout": {
                    "LARGE": "layout_777",
                    "MEDIUM": "layout_835",
                    "SMALL": "layout_867",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "",
                "backgroundIMage": {},
            },
            "view_17": {
                "id": "view_17",
                "layout": {
                    "LARGE": "layout_799",
                    "MEDIUM": "layout_844",
                    "SMALL": "layout_876",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "",
                "backgroundIMage": {},
            },
        },
        "sections": {
            "section_1": {
                "id": "section_1",
                "views": ["view_2", "view_16", "view_17"],
                "transition": {"type": "PUSH"},
            }
        },
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_0": []}],
        "timestamp": 1619662282719,
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "summary": {
        "pages": {
            "page_1": {
                "id": "page_1",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_2",
                    "MEDIUM": "layout_632",
                    "SMALL": "layout_647",
                },
                "isVisible": True,
                "isDefault": True,
                "header": False,
                "bodyBackgroundColor": "",
                "footer": False,
                "oneByOneEffect": {"type": "FLOAT_IN", "option": {"direction": None}},
            }
        },
        "layouts": {
            "layout_2": {
                "type": "FLOW",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "294px",
                            "right": "654px",
                            "top": "-32px",
                            "bottom": "38px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_181",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "102px",
                            "right": "462px",
                            "top": "-34px",
                            "bottom": "36px",
                            "width": "360px",
                            "height": "420px",
                        },
                        "widgetId": "widget_184",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "576px",
                            "right": "876px",
                            "top": "-293px",
                            "bottom": "307px",
                            "width": "300px",
                            "height": "340px",
                        },
                        "widgetId": "widget_201",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 60,
                            "width": "100%",
                        },
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "64px",
                            "right": "1216px",
                            "top": "921px",
                            "bottom": "1081px",
                            "width": "1152px",
                            "height": "160px",
                        },
                        "widgetId": "widget_212",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-56px",
                            "right": "344px",
                            "top": "904px",
                            "bottom": "1304px",
                            "width": "400px",
                            "height": 437,
                        },
                        "widgetId": "widget_659",
                        "setting": {"heightMode": "auto", "width": "100%"},
                    },
                },
                "order": ["1", "2", "3", "4", "5"],
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
            },
            "layout_68": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": 50},
                        "widgetId": "widget_182",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 4, "height": 50},
                        "widgetId": "widget_183",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    },
                },
            },
            "layout_69": {
                "type": "ROW",
                "order": ["1", "2"],
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": 600},
                        "widgetId": "widget_186",
                        "setting": {"heightMode": "fit"},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "400px"},
                        "widgetId": "widget_190",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
            },
            "layout_70": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "595px", "height": "100px"},
                        "widgetId": "widget_185",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "flex-start"},
                            "autoProps": {"width": False},
                        },
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_187",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": False}},
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_188",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": False}},
                        "bbox": {"width": 583, "height": "80px"},
                        "widgetId": "widget_189",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_71": {
                "type": "COLUMN",
                "order": ["0", "1", "2", "3", "4", "6", "7", "5", "8", "9"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": 40},
                        "widgetId": "widget_191",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "10px"},
                        "widgetId": "widget_192",
                        "setting": {"autoProps": {"width": False}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_193",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_194",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_195",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_196",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_197",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_198",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_199",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_200",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                },
            },
            "layout_72": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": 600},
                        "widgetId": "widget_202",
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_73": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_203",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_204",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_205",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_206",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_207",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_208",
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_209",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_210",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7"],
            },
            "layout_74": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "offsetY": 20,
                        },
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_211",
                    }
                },
                "order": ["0"],
            },
            "layout_592": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 8, "height": "100.00%"},
                        "widgetId": "widget_660",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "391px"},
                        "widgetId": "widget_661",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:115",
                        },
                    },
                },
            },
            "layout_593": {"type": "FIXED"},
            "layout_629": {"type": "FIXED", "content": {}, "order": []},
            "layout_631": {
                "type": "FIXED",
                "content": {
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "14px",
                            "right": "28px",
                            "top": "65.2241%",
                            "bottom": "123px",
                            "width": "89.9235%",
                            "height": "9.1771%",
                        },
                        "widgetId": "widget_803",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.2701%",
                            "right": "6.7127%",
                            "top": "76.2881%",
                            "bottom": "3.9791%",
                            "width": "90.0173%",
                            "height": "19.7328%",
                        },
                        "widgetId": "widget_804",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM"},
                        },
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": "47.1321%",
                            "width": "100.0000%",
                            "height": "55.00%",
                        },
                        "widgetId": "widget_805",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_806",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.3418%",
                            "right": 73,
                            "top": "52.5180%",
                            "bottom": "42.5211%",
                            "width": 76,
                            "height": 32,
                        },
                        "widgetId": "widget_807",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["4", "5", "6", "9", "10"],
            },
            "layout_632": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "294px",
                            "right": "654px",
                            "top": "-32px",
                            "bottom": "38px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_808",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "102px",
                            "right": "462px",
                            "top": "-34px",
                            "bottom": "36px",
                            "width": "360px",
                            "height": "420px",
                        },
                        "widgetId": "widget_809",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "64px",
                            "right": "1216px",
                            "top": "921px",
                            "bottom": "1081px",
                            "width": "1152px",
                            "height": "160px",
                        },
                        "widgetId": "widget_814",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-56px",
                            "right": "344px",
                            "top": "904px",
                            "bottom": "1304px",
                            "width": "400px",
                            "height": 316,
                        },
                        "widgetId": "widget_815",
                        "setting": {"heightMode": "auto", "width": "100%"},
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "86px",
                            "right": "263px",
                            "top": "544px",
                            "bottom": "678px",
                            "width": "177px",
                            "height": 154,
                        },
                        "widgetId": "widget_816",
                        "setting": {"heightMode": "auto", "width": "100%"},
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "42px",
                            "right": "262px",
                            "top": "-60px",
                            "bottom": "259px",
                            "width": "221px",
                            "height": 339,
                        },
                        "widgetId": "widget_818",
                        "setting": {"heightMode": "auto", "width": "100%"},
                    },
                },
                "order": ["0", "1", "5", "6", "3", "4"],
            },
            "layout_633": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_182",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_634": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600px"},
                        "widgetId": "widget_810",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_635": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "595px", "height": "100px"},
                        "widgetId": "widget_185",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "flex-start"},
                            "autoProps": {"width": False},
                        },
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_187",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": False}},
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_188",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {"width": False, "height": "AUTO"},
                        },
                        "bbox": {"width": 583, "height": "80.00px"},
                        "widgetId": "widget_189",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_636": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": 40},
                        "widgetId": "widget_191",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "10px"},
                        "widgetId": "widget_192",
                        "setting": {"autoProps": {"width": False}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_193",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_194",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_195",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_197",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_198",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_196",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_199",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_200",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
            },
            "layout_638": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_203",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_204",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_205",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_206",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_207",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_208",
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_209",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_210",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7"],
            },
            "layout_639": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "offsetY": 20,
                        },
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_211",
                    }
                },
                "order": ["0"],
            },
            "layout_640": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "100.00%"},
                        "widgetId": "widget_660",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "391px"},
                        "widgetId": "widget_661",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": 1.15,
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_641": {"type": "FIXED"},
            "layout_642": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "62.1755%",
                            "bottom": 111,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_803",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "72.2257%",
                            "bottom": "3.9076%",
                            "width": "90.1186%",
                            "height": "24.0343%",
                        },
                        "widgetId": "widget_804",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "55.00%",
                        },
                        "widgetId": "widget_805",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_806",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "65.1124%",
                            "right": 12,
                            "top": "49.8659%",
                            "bottom": "39.1470%",
                            "width": 76,
                            "height": 32,
                        },
                        "widgetId": "widget_807",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
            },
            "layout_643": {"type": "FIXED"},
            "layout_644": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "maintainedByLayout": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 12, "height": "554px"},
                        "widgetId": "widget_817",
                    }
                },
                "order": ["0"],
            },
            "layout_645": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {
                            "left": "-51px",
                            "right": "169px",
                            "top": "-44px",
                            "bottom": "356px",
                            "width": "221px",
                            "height": "400px",
                        },
                        "widgetId": "widget_811",
                    },
                    "1": {
                        "widgetId": "widget_183",
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-51px",
                            "right": "169px",
                            "top": "-44px",
                            "bottom": "356px",
                            "width": "221px",
                            "height": "400px",
                        },
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["1", "0"],
            },
            "layout_646": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 12, "height": "319px"},
                        "widgetId": "widget_813",
                    }
                },
                "order": ["0"],
            },
            "layout_647": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "102px",
                            "right": "462px",
                            "top": "-34px",
                            "bottom": "36px",
                            "width": "360px",
                            "height": "420px",
                        },
                        "widgetId": "widget_820",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "18px",
                            "right": "342px",
                            "top": "1236px",
                            "bottom": "1351px",
                            "width": "324px",
                            "height": "115px",
                        },
                        "widgetId": "widget_825",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "width": "100%",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "14px",
                            "right": "316px",
                            "top": "-19px",
                            "bottom": "277px",
                            "width": "302px",
                            "height": 316,
                        },
                        "widgetId": "widget_832",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "15px",
                            "right": "317px",
                            "top": "6px",
                            "bottom": "406px",
                            "width": "302px",
                            "height": 420,
                        },
                        "widgetId": "widget_833",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "9px",
                            "right": "311px",
                            "top": "-13px",
                            "bottom": "334px",
                            "width": "302px",
                            "height": "360px",
                        },
                        "widgetId": "widget_834",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-3px",
                            "right": "311px",
                            "top": "-23px",
                            "bottom": "68px",
                            "width": "314px",
                            "height": 111,
                        },
                        "widgetId": "widget_837",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                    },
                },
                "order": ["10", "1", "5", "3", "6", "7"],
            },
            "layout_649": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_828",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_650": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "595px", "height": "100px"},
                        "widgetId": "widget_185",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "flex-start"},
                            "autoProps": {"width": False, "height": "AUTO"},
                        },
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_187",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {"width": False, "height": "AUTO"},
                        },
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_188",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": False}},
                        "bbox": {"width": 583, "height": "80px"},
                        "widgetId": "widget_189",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_651": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "294px", "height": "40px"},
                        "widgetId": "widget_191",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "10px"},
                        "widgetId": "widget_192",
                        "setting": {"autoProps": {"width": False}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_193",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_194",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_195",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_197",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_198",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_196",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_199",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_200",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
            },
            "layout_653": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_203",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_204",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_205",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_206",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_207",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_208",
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_209",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_210",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7"],
            },
            "layout_654": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_830",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_656": {"type": "FIXED"},
            "layout_657": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "62.1755%",
                            "bottom": 111,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_803",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "72.2257%",
                            "bottom": "3.9076%",
                            "width": "89.9281%",
                            "height": "23.8693%",
                        },
                        "widgetId": "widget_804",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "55.00%",
                        },
                        "widgetId": "widget_805",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_806",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "67.6259%",
                            "right": 14,
                            "top": "51.3400%",
                            "bottom": "40.6198%",
                            "width": 76,
                            "height": 32,
                        },
                        "widgetId": "widget_807",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
            },
            "layout_658": {"type": "FIXED"},
            "layout_660": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 600},
                        "widgetId": "widget_821",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "400px"},
                        "widgetId": "widget_822",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "300px", "height": "92px"},
                        "widgetId": "widget_183",
                    },
                },
                "order": ["0", "2", "1"],
            },
            "layout_662": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "offsetY": 20,
                        },
                        "bbox": {"width": "auto", "height": "92px"},
                        "widgetId": "widget_211",
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_664": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 12, "height": "296px"},
                        "widgetId": "widget_824",
                    }
                },
                "order": ["0"],
            },
            "layout_665": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 12, "height": "100.00%"},
                        "widgetId": "widget_660",
                    }
                },
                "order": ["0"],
            },
            "layout_666": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 0, "width": 12, "height": "347px"},
                        "widgetId": "widget_661",
                    }
                },
                "order": ["0"],
            },
            "layout_669": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 12, "height": "91px"},
                        "widgetId": "widget_182",
                    }
                },
                "order": ["0"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/ink/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "summary",
        "widgets": {
            "widget_181": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_181",
                "layouts": {"DEFAULT": {"LARGE": "layout_68"}},
            },
            "widget_182": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_182",
            },
            "widget_183": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_183",
            },
            "widget_184": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_184",
                "layouts": {"DEFAULT": {"LARGE": "layout_69"}},
            },
            "widget_185": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-800); font-size: 20px;">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. </span></p>',
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_185",
            },
            "widget_186": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_186",
                "layouts": {"DEFAULT": {"LARGE": "layout_70"}},
            },
            "widget_187": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p><br></p>",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_187",
            },
            "widget_188": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p><br></p>",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_188",
            },
            "widget_189": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-100); font-size: 16px;">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. \ufeff</span></p>',
                    "placeholder": '<p><span style="color: var(--dark-100); font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_189",
            },
            "widget_190": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_190",
                "layouts": {"DEFAULT": {"LARGE": "layout_71"}},
            },
            "widget_191": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="color: var(--dark-800); font-size: 20px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_191",
            },
            "widget_192": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_192",
            },
            "widget_193": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_193",
            },
            "widget_194": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_194",
            },
            "widget_195": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_195",
            },
            "widget_196": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_196",
            },
            "widget_197": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_197",
            },
            "widget_198": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_198",
            },
            "widget_199": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_199",
            },
            "widget_200": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_200",
            },
            "widget_201": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_201",
                "layouts": {"DEFAULT": {"LARGE": "layout_72"}},
            },
            "widget_202": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_202",
                "layouts": {"DEFAULT": {"LARGE": "layout_73"}},
            },
            "widget_203": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="color: var(--dark-800); font-size: 20px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_203",
            },
            "widget_204": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_204",
            },
            "widget_205": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_205",
            },
            "widget_206": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_206",
            },
            "widget_207": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_207",
            },
            "widget_208": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_208",
            },
            "widget_209": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_209",
            },
            "widget_210": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_210",
            },
            "widget_211": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_211",
            },
            "widget_212": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_212",
                "layouts": {"DEFAULT": {"LARGE": "layout_74"}},
            },
            "widget_659": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_659",
                "layouts": {"DEFAULT": {"LARGE": "layout_592"}},
            },
            "widget_660": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {"canZoom": True, "canHome": True, "canSearch": True},
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 18489297.737236,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -11153691.167369908,
                                "y": 4754179.003170505,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -13709745.393225523,
                            "ymin": 3301263.969526261,
                            "xmax": -8597636.941514293,
                            "ymax": 6207094.036814749,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_660",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_593",
                        "MEDIUM": "layout_641",
                        "SMALL": "layout_656",
                    }
                },
            },
            "widget_661": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE3",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 447.453125,
                    "widgetHeight": 514.5625,
                },
                "id": "widget_661",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_631",
                        "MEDIUM": "layout_642",
                        "SMALL": "layout_657",
                    },
                    "HOVER": {
                        "LARGE": "layout_629",
                        "MEDIUM": "layout_643",
                        "SMALL": "layout_658",
                    },
                },
            },
            "widget_803": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style=""><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_803",
            },
            "widget_804": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_804",
            },
            "widget_805": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_805",
            },
            "widget_806": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "success-100",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_806",
            },
            "widget_807": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {"text": "Share", "textExpression": None},
                    "styleConfig": {
                        "themeStyle": {"quickStyleType": "primary"},
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--white)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                    },
                },
                "id": "widget_807",
            },
            "widget_808": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_808",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_633"}},
            },
            "widget_809": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_809",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_634"}},
            },
            "widget_810": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_810",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_635"}},
            },
            "widget_811": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_811",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_636"}},
            },
            "widget_813": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_813",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_638"}},
            },
            "widget_814": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_814",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_639"}},
            },
            "widget_815": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_815",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_640"}},
            },
            "widget_816": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_816",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_644"}},
            },
            "widget_817": {
                "id": "widget_817",
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_645"}},
            },
            "widget_818": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_818",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_646"}},
            },
            "widget_820": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_820",
                "layouts": {"DEFAULT": {"SMALL": "layout_649"}},
            },
            "widget_821": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_821",
                "layouts": {"DEFAULT": {"SMALL": "layout_650"}},
            },
            "widget_822": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_822",
                "layouts": {"DEFAULT": {"SMALL": "layout_651"}},
            },
            "widget_824": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_824",
                "layouts": {"DEFAULT": {"SMALL": "layout_653"}},
            },
            "widget_825": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_825",
                "layouts": {"DEFAULT": {"SMALL": "layout_654"}},
            },
            "widget_828": {
                "uri": "widgets/layout/column/",
                "id": "widget_828",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_660"}},
            },
            "widget_830": {
                "uri": "widgets/layout/column/",
                "id": "widget_830",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_662"}},
            },
            "widget_832": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 40, 10], "unit": "px"}},
                },
                "id": "widget_832",
                "layouts": {"DEFAULT": {"SMALL": "layout_664"}},
            },
            "widget_833": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_833",
                "layouts": {"DEFAULT": {"SMALL": "layout_665"}},
            },
            "widget_834": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_834",
                "layouts": {"DEFAULT": {"SMALL": "layout_666"}},
            },
            "widget_837": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_837",
                "layouts": {"DEFAULT": {"SMALL": "layout_669"}},
            },
        },
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_1": []}],
        "timestamp": 1622673365846,
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "timeline": {
        "pages": {
            "page_4": {
                "id": "page_4",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_5",
                    "SMALL": "layout_308",
                    "MEDIUM": "layout_326",
                },
                "isVisible": True,
                "isDefault": True,
                "header": False,
                "bodyBackgroundColor": "",
                "footer": False,
                "oneByOneEffect": {"type": "FLOAT_IN", "option": {"direction": None}},
            }
        },
        "layouts": {
            "layout_5": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "229px",
                            "right": "589px",
                            "top": "62px",
                            "bottom": "132px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_244",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetY": 20,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-35px",
                            "right": "265px",
                            "top": "164px",
                            "bottom": "214px",
                            "width": "300px",
                            "height": "2100px",
                        },
                        "widgetId": "widget_246",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 20,
                            "width": "100%",
                        },
                    },
                },
                "order": ["0", "1"],
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
            },
            "layout_100": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_245",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
            },
            "layout_101": {
                "type": "ROW",
                "order": ["0", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 1, "height": "1940px"},
                        "widgetId": "widget_247",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "STRETCH"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 1, "width": 11, "height": 600},
                        "widgetId": "widget_249",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
            },
            "layout_102": {
                "type": "COLUMN",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "280px"},
                        "widgetId": "widget_250",
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_248",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_254",
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_255",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_259",
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_260",
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_264",
                        "isPending": False,
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_265",
                        "isPending": False,
                    },
                    "11": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_269",
                        "isPending": False,
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_270",
                        "isPending": False,
                    },
                    "13": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_273",
                        "isPending": False,
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_274",
                        "isPending": False,
                    },
                },
                "order": [
                    "4",
                    "1",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                ],
            },
            "layout_103": {
                "type": "ROW",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "180px"},
                        "widgetId": "widget_251",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "180px"},
                        "widgetId": "widget_252",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 5, "height": "180px"},
                        "widgetId": "widget_253",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
            },
            "layout_104": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 180},
                        "widgetId": "widget_256",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 7, "height": 180},
                        "widgetId": "widget_258",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "2"],
            },
            "layout_105": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 2, "height": 180},
                        "widgetId": "widget_261",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 5, "height": 180},
                        "widgetId": "widget_262",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 2, "height": 180},
                        "widgetId": "widget_263",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "2", "1"],
            },
            "layout_106": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "180px"},
                        "widgetId": "widget_266",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "180px"},
                        "widgetId": "widget_267",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 5, "height": "180px"},
                        "widgetId": "widget_268",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_107": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 5, "height": "180px"},
                        "widgetId": "widget_271",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 6, "height": 180},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_108": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "180px"},
                        "widgetId": "widget_275",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 5, "height": 180},
                        "widgetId": "widget_276",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_308": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "229px",
                            "right": "589px",
                            "top": "62px",
                            "bottom": "132px",
                            "width": "360px",
                            "height": "100px",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetY": 20,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-35px",
                            "right": "265px",
                            "top": "164px",
                            "bottom": "214px",
                            "width": "300px",
                            "height": "3400px",
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 20,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                            "width": "100%",
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_309": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_418",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_310": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 1, "height": "3300px"},
                        "type": "WIDGET",
                        "widgetId": "widget_419",
                        "setting": {
                            "autoProps": {"height": "STRETCH"},
                            "autoHeight": True,
                            "heightMode": "fixed",
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 1, "width": 11, "height": 600},
                        "widgetId": "widget_411",
                    },
                },
                "order": ["2", "3"],
            },
            "layout_311": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_248",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "280px"},
                        "widgetId": "widget_412",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_254",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_413",
                        "isPending": False,
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_259",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_414",
                        "isPending": False,
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_264",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_415",
                        "isPending": False,
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "8": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_269",
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_416",
                        "isPending": False,
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "10": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_273",
                        "isPending": False,
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_417",
                        "isPending": False,
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11"],
            },
            "layout_312": {
                "type": "ROW",
                "content": {
                    "3": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_420",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["3"],
            },
            "layout_313": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_421",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_314": {
                "type": "ROW",
                "content": {
                    "3": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_422",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["3"],
            },
            "layout_315": {
                "type": "ROW",
                "content": {
                    "3": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_423",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["3"],
            },
            "layout_316": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_424",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_317": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_425",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_318": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "50px"},
                        "widgetId": "widget_245",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_319": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "20px", "height": "3260px"},
                        "widgetId": "widget_247",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False, "width": False},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_320": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_251",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_252",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_253",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_321": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_256",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_258",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_322": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_261",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_263",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_262",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_323": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_266",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_267",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_268",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_324": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_271",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_325": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_275",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_276",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_326": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "229px",
                            "right": "589px",
                            "top": "62px",
                            "bottom": "132px",
                            "width": "360px",
                            "height": "80px",
                        },
                        "widgetId": "widget_426",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetY": 20,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-35px",
                            "right": "265px",
                            "top": "164px",
                            "bottom": "214px",
                            "width": "300px",
                            "height": "2100px",
                        },
                        "widgetId": "widget_427",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 20,
                            "width": "100%",
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_327": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 50},
                        "widgetId": "widget_245",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_328": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 1, "height": "1940px"},
                        "widgetId": "widget_247",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "STRETCH"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 1, "width": 11, "height": 600},
                        "widgetId": "widget_428",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_329": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_248",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "280px"},
                        "widgetId": "widget_429",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_254",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_430",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_259",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_431",
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_264",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_432",
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_269",
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_433",
                        "isPending": False,
                    },
                    "10": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_273",
                        "isPending": False,
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_434",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11"],
            },
            "layout_330": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "180px"},
                        "widgetId": "widget_251",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "180px"},
                        "widgetId": "widget_252",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 5, "height": "180px"},
                        "widgetId": "widget_253",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_331": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 180},
                        "widgetId": "widget_256",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 7, "height": 180},
                        "widgetId": "widget_258",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_332": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 2, "height": 180},
                        "widgetId": "widget_261",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 2, "height": 180},
                        "widgetId": "widget_263",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 5, "height": 180},
                        "widgetId": "widget_262",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_333": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "180px"},
                        "widgetId": "widget_266",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "180px"},
                        "widgetId": "widget_267",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 5, "height": "180px"},
                        "widgetId": "widget_268",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_334": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 5, "height": "180px"},
                        "widgetId": "widget_271",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 6, "height": 180},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_335": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "180px"},
                        "widgetId": "widget_275",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 5, "height": 180},
                        "widgetId": "widget_276",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/ink/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "timeline",
        "widgets": {
            "widget_244": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_244",
                "layouts": {"DEFAULT": {"LARGE": "layout_100"}},
            },
            "widget_245": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_245",
            },
            "widget_246": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_246",
                "layouts": {"DEFAULT": {"LARGE": "layout_101"}},
            },
            "widget_247": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Vertical",
                    "strokeStyle": {
                        "type": "Style1",
                        "color": "var(--light-700)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "Point0", "pointSize": 5},
                    "pointEnd": {"pointStyle": "Point0", "pointSize": 5},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_247",
            },
            "widget_248": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 24px; color: var(--dark-800);">1920\ufeff</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_248",
            },
            "widget_249": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_249",
                "layouts": {"DEFAULT": {"LARGE": "layout_102"}},
            },
            "widget_250": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_250",
                "layouts": {"DEFAULT": {"LARGE": "layout_103"}},
            },
            "widget_251": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_251",
            },
            "widget_252": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_252",
            },
            "widget_253": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_253",
            },
            "widget_254": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--dark-800); font-size: 24px;">1965</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_254",
            },
            "widget_255": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_255",
                "layouts": {"DEFAULT": {"LARGE": "layout_104"}},
            },
            "widget_256": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_256",
            },
            "widget_258": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_258",
            },
            "widget_259": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 24px; color: var(--dark-800);">1993</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_259",
            },
            "widget_260": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_260",
                "layouts": {"DEFAULT": {"LARGE": "layout_105"}},
            },
            "widget_261": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_261",
            },
            "widget_262": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_262",
            },
            "widget_263": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_263",
            },
            "widget_264": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 24px; color: var(--dark-800);">2008</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_264",
            },
            "widget_265": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_265",
                "layouts": {"DEFAULT": {"LARGE": "layout_106"}},
            },
            "widget_266": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_266",
            },
            "widget_267": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_267",
            },
            "widget_268": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_268",
            },
            "widget_269": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--dark-800); font-size: 24px;">2015</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_269",
            },
            "widget_270": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_270",
                "layouts": {"DEFAULT": {"LARGE": "layout_107"}},
            },
            "widget_271": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_271",
            },
            "widget_272": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_272",
            },
            "widget_273": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 24px; color: var(--dark-800);">2020</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_273",
            },
            "widget_274": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_274",
                "layouts": {"DEFAULT": {"LARGE": "layout_108"}},
            },
            "widget_275": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_275",
            },
            "widget_276": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_276",
            },
            "widget_409": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_409",
                "layouts": {"DEFAULT": {"SMALL": "layout_309"}},
            },
            "widget_410": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_410",
                "layouts": {"DEFAULT": {"SMALL": "layout_310"}},
            },
            "widget_411": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_411",
                "layouts": {"DEFAULT": {"SMALL": "layout_311"}},
            },
            "widget_412": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_412",
                "layouts": {"DEFAULT": {"SMALL": "layout_312"}},
            },
            "widget_413": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_413",
                "layouts": {"DEFAULT": {"SMALL": "layout_313"}},
            },
            "widget_414": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_414",
                "layouts": {"DEFAULT": {"SMALL": "layout_314"}},
            },
            "widget_415": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_415",
                "layouts": {"DEFAULT": {"SMALL": "layout_315"}},
            },
            "widget_416": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_416",
                "layouts": {"DEFAULT": {"SMALL": "layout_316"}},
            },
            "widget_417": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_417",
                "layouts": {"DEFAULT": {"SMALL": "layout_317"}},
            },
            "widget_418": {
                "uri": "widgets/layout/column/",
                "id": "widget_418",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_318"}},
            },
            "widget_419": {
                "uri": "widgets/layout/column/",
                "id": "widget_419",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_319"}},
            },
            "widget_420": {
                "uri": "widgets/layout/column/",
                "id": "widget_420",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_320"}},
            },
            "widget_421": {
                "uri": "widgets/layout/column/",
                "id": "widget_421",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_321"}},
            },
            "widget_422": {
                "uri": "widgets/layout/column/",
                "id": "widget_422",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_322"}},
            },
            "widget_423": {
                "uri": "widgets/layout/column/",
                "id": "widget_423",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_323"}},
            },
            "widget_424": {
                "uri": "widgets/layout/column/",
                "id": "widget_424",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_324"}},
            },
            "widget_425": {
                "uri": "widgets/layout/column/",
                "id": "widget_425",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_325"}},
            },
            "widget_426": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_426",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_327"}},
            },
            "widget_427": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_427",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_328"}},
            },
            "widget_428": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_428",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_329"}},
            },
            "widget_429": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_429",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_330"}},
            },
            "widget_430": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_430",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_331"}},
            },
            "widget_431": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_431",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_332"}},
            },
            "widget_432": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_432",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_333"}},
            },
            "widget_433": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_433",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_334"}},
            },
            "widget_434": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_434",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_335"}},
            },
        },
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_4": []}],
        "timestamp": 1623030418665,
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
        "dialogs": {},
    },
    "scenic": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_0",
                    "MEDIUM": "layout_821",
                    "SMALL": "layout_853",
                },
                "isDefault": True,
                "isVisible": True,
                "header": True,
                "bodyBackgroundColor": "",
                "footer": True,
            },
            "page_1": {
                "id": "page_1",
                "mode": "FIT_WINDOW",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_98",
                    "SMALL": "layout_371",
                    "MEDIUM": "layout_488",
                },
                "isVisible": True,
                "isDefault": False,
                "oneByOneEffect": {"type": "FADE_IN", "option": {"direction": None}},
                "bodyBackgroundColor": "",
                "header": False,
            },
        },
        "layouts": {
            "layout_0": {
                "type": "FLOW",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1280px",
                            "top": "512px",
                            "bottom": "2782px",
                            "width": "1280px",
                            "height": "1800px",
                        },
                        "widgetId": "widget_6",
                        "setting": {"heightMode": "auto", "width": "100%"},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "525px",
                            "right": "905px",
                            "top": "-30px",
                            "bottom": "30px",
                            "width": "380px",
                            "height": 93,
                        },
                        "widgetId": "widget_7",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6px",
                            "right": "1286px",
                            "top": "1px",
                            "bottom": "501px",
                            "width": "1280px",
                            "height": "480px",
                        },
                        "widgetId": "widget_152",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "isFloating": True,
                            "lockParent": True,
                            "floatingArea": 9,
                            "offsetY": 40,
                            "offsetX": 40,
                            "heightMode": "fixed",
                            "width": "64px",
                        },
                        "bbox": {
                            "top": 0,
                            "left": 61.6282958984375,
                            "right": 0,
                            "width": "128px",
                            "height": "48px",
                            "bottom": 0,
                        },
                        "widgetId": "widget_216",
                    },
                },
                "order": ["3", "2", "1", "6"],
                "setting": {
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}}
                },
            },
            "layout_1": {
                "type": "FIXED",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20.00px",
                            "right": "95.7060%",
                            "top": 19,
                            "bottom": 20,
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.00px",
                            "right": "66.4266%",
                            "top": 19,
                            "bottom": 14,
                            "width": "28.1250%",
                            "height": "53.5211%",
                        },
                        "widgetId": "widget_2",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "93.7500%",
                            "right": "20.00px",
                            "top": "50%",
                            "bottom": "28.1690%",
                            "width": 80,
                            "height": 40,
                        },
                        "widgetId": "widget_144",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                },
            },
            "layout_3": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "1460.00px"},
                        "setting": {
                            "heightMode": "fixed",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {
                                        "direction": None,
                                        "configType": "DEFAULT",
                                    },
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "flex-start"},
                            "aspectRatio": 1.01,
                        },
                    }
                },
            },
            "layout_5": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": 60},
                        "widgetId": "widget_8",
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_98": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "62.5000%",
                            "height": "50.0000%",
                        },
                        "widgetId": "widget_145",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_99": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_146",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
            },
            "layout_100": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "-4.1667%",
                            "top": "50%",
                            "bottom": "28.3208%",
                            "width": "100.0000%",
                            "height": "100.00%",
                        },
                        "setting": {
                            "hCenter": True,
                            "vCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
            },
            "layout_101": {"type": "FIXED", "order": [], "content": {}},
            "layout_102": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": True,
                        },
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "100.0000%",
                            "height": "38.5645%",
                            "right": "0.0000%",
                            "bottom": "61.4355%",
                        },
                        "widgetId": "widget_149",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "26.8500%",
                            "top": "40.6216%",
                            "width": 463,
                            "height": 69,
                            "right": 269,
                            "bottom": "50.7534%",
                        },
                        "widgetId": "widget_150",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "31px",
                            "top": "50.2930%",
                            "width": "93.8000%",
                            "height": "45.7402%",
                            "right": "31px",
                            "bottom": "32px",
                        },
                        "widgetId": "widget_151",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_103": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "480.00px"},
                        "widgetId": "widget_153",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:40",
                        },
                    }
                },
            },
            "layout_104": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_237",
                        "setting": {
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "44.3056%",
                            "bottom": "43.9382%",
                            "width": 960,
                            "height": "11.7563%",
                        },
                        "widgetId": "widget_238",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                            },
                            "vCenter": False,
                            "hCenter": True,
                        },
                        "isPending": False,
                    },
                },
            },
            "layout_143": {
                "type": "FIXED",
                "order": ["1", "2"],
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "46.2028%",
                            "top": "69.9531%",
                            "bottom": 33,
                            "width": "40.00%",
                            "height": "13.5000%",
                        },
                        "widgetId": "widget_218",
                        "setting": {
                            "hCenter": True,
                            "vCenter": False,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "45.0000%",
                            "top": "35.9063%",
                            "bottom": "40.0938%",
                            "width": "10.0000%",
                            "height": "24.0000%",
                        },
                        "widgetId": "widget_219",
                        "setting": {
                            "autoProps": {
                                "height": "AUTO",
                                "width": "AUTO",
                                "left": False,
                                "right": True,
                            },
                            "hCenter": True,
                            "vCenter": False,
                        },
                        "isPending": False,
                    },
                },
            },
            "layout_161": {
                "type": "FIXED",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "960.00px",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_265",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_784",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_803",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_822",
                        "isPending": False,
                    },
                },
                "order": ["1", "2", "3", "4"],
            },
            "layout_163": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_266",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_267",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_268",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_164": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_269",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_270",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_272",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "3"],
            },
            "layout_181": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_331",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
            },
            "layout_182": {
                "type": "COLUMN",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_332",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_333",
                    },
                },
            },
            "layout_183": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_148",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_330",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_184": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_334",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_185": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_335",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_338",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_186": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_336",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_337",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_187": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_339",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_188": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_340",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_343",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_189": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_341",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_342",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_190": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_344",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_191": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_345",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_348",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_192": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_346",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_347",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_193": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_349",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_194": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_350",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_353",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_195": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_351",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_352",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_196": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_354",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_197": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_355",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_358",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_198": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_356",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_357",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_199": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_359",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_200": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_360",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_363",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_201": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_361",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_362",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_202": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_364",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_203": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_365",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_368",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_204": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_366",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_367",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_205": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_369",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_206": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_370",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_373",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_207": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_371",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_372",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_208": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_374",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_209": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_375",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_378",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_210": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_376",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_377",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_211": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_379",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_212": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_380",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_383",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_213": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_381",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_382",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_214": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_384",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_215": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_385",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_388",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_216": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_386",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_387",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_297": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20.00px",
                            "right": "95.7060%",
                            "top": 19,
                            "bottom": 20,
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.00px",
                            "right": "66.4266%",
                            "top": 19,
                            "bottom": 14,
                            "width": "28.1250%",
                            "height": "53.5211%",
                        },
                        "widgetId": "widget_2",
                        "isPending": True,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "93.7500%",
                            "right": "20.00px",
                            "top": "50%",
                            "bottom": "28.1690%",
                            "width": 80,
                            "height": 40,
                        },
                        "widgetId": "widget_144",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_371": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "62.5000%",
                            "height": "50.0000%",
                        },
                        "widgetId": "widget_456",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_372": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_146",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_373": {"type": "FIXED"},
            "layout_374": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "-4.1667%",
                            "top": "50%",
                            "bottom": "28.3208%",
                            "width": "100.0000%",
                            "height": "100.00%",
                        },
                        "setting": {
                            "hCenter": True,
                            "vCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_375": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_457",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_376": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_458",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_333",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_377": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_148",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_330",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_378": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_459",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_379": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_460",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_338",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_380": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_336",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_337",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_381": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_461",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_382": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "354px", "height": "54px"},
                        "widgetId": "widget_462",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_343",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_383": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_341",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_342",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_384": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_463",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_385": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_464",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_348",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_386": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_346",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_347",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_387": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_465",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_388": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_466",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_353",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_389": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_351",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_352",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_390": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_467",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_391": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_468",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_358",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_392": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_356",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_357",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_393": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_469",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_394": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_470",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_363",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_395": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_361",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_362",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_396": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_471",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_397": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_472",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_368",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_398": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_366",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_367",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_399": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_473",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_400": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_474",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_373",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_401": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_371",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_372",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_402": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_475",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_403": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_476",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_378",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_404": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_376",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_377",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_405": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_477",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_406": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_478",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_383",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_407": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_381",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_382",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_408": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_479",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_409": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_480",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_388",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_410": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32px"},
                        "widgetId": "widget_386",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_387",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_488": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "62.5000%",
                            "height": "50.0000%",
                        },
                        "widgetId": "widget_513",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_489": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_146",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": True,
                                "width": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_490": {"type": "FIXED"},
            "layout_491": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "-4.1667%",
                            "top": "50%",
                            "bottom": "28.3208%",
                            "width": "100.0000%",
                            "height": "100.00%",
                        },
                        "setting": {
                            "hCenter": True,
                            "vCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_492": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_514",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_493": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_515",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_333",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_494": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_148",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_330",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_495": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_516",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_496": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_517",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_338",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_497": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_336",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_337",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_498": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_518",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_499": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_519",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_343",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_500": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_341",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_342",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_501": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_520",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_502": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_521",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_348",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_503": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_346",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_347",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_504": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_522",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_505": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_523",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_353",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_506": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_351",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_352",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_507": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_524",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_508": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_525",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_358",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_509": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_356",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_357",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_510": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_526",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_511": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_527",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_363",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_512": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_361",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_362",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_513": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_528",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_514": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_529",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_368",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_515": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_366",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_367",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_516": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_530",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_517": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_531",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_373",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_518": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_371",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_372",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_519": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_532",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_520": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_533",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_378",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_521": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_376",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_377",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_522": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_534",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_523": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_535",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_383",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_524": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_381",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_382",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_525": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "24.6231%",
                            "top": "50%",
                            "bottom": "24.6231%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_536",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_526": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "396px", "height": "54.00px"},
                        "widgetId": "widget_537",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"height": "STRETCH"}},
                        "bbox": {"width": "396px", "height": 687},
                        "widgetId": "widget_388",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_527": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"left": 0, "width": 2, "height": "32.00px"},
                        "widgetId": "widget_386",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 2, "width": 10, "height": "34px"},
                        "widgetId": "widget_387",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_528": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20.00px",
                            "right": "95.7060%",
                            "top": 19,
                            "bottom": 20,
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.00px",
                            "right": "66.4266%",
                            "top": 19,
                            "bottom": 14,
                            "width": "28.1250%",
                            "height": "53.5211%",
                        },
                        "widgetId": "widget_2",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "93.7500%",
                            "right": "20.00px",
                            "top": "50%",
                            "bottom": "28.1690%",
                            "width": 80,
                            "height": 40,
                        },
                        "widgetId": "widget_144",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_529": {
                "type": "FIXED",
                "content": {
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "378px",
                            "top": "12.00px",
                            "width": "40.00px",
                            "height": "40px",
                            "right": "20px",
                            "bottom": "381px",
                        },
                        "widgetId": "widget_538",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": True,
                                "top": False,
                            },
                            "order": 0,
                            "lockParent": False,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "width": "CUSTOM",
                            },
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "50%",
                            "right": 19,
                            "top": 167,
                            "bottom": "25.7421%",
                            "width": "93.3110%",
                            "height": "48.1191%",
                        },
                        "widgetId": "widget_539",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "top": False,
                                "bottom": True,
                                "left": False,
                                "right": True,
                            },
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "43.5355%",
                            "top": 56,
                            "bottom": "82.9692%",
                            "width": 136,
                            "height": 53,
                        },
                        "widgetId": "widget_540",
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "width": "CUSTOM",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": True,
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "50%",
                            "right": 19,
                            "top": "77.8648%",
                            "bottom": "14.9252%",
                            "width": "93.3110%",
                            "height": "7.2100%",
                        },
                        "widgetId": "widget_541",
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "right": False,
                                "left": True,
                            },
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "59.9912%",
                            "right": "110.00px",
                            "top": "84.6739%",
                            "bottom": "20.00px",
                            "width": "80px",
                            "height": "32.00px",
                        },
                        "widgetId": "widget_542",
                        "isPending": True,
                    },
                    "10": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "right": False,
                                "left": True,
                            },
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "79.2920%",
                            "right": "20.00px",
                            "top": "84.4355%",
                            "bottom": "20.00px",
                            "width": "80px",
                            "height": "32.00px",
                        },
                        "widgetId": "widget_543",
                        "isPending": True,
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "42.6427%",
                            "top": "17.2414%",
                            "bottom": "80.2508%",
                            "width": "14.3573%",
                            "height": "2.5078%",
                        },
                        "widgetId": "widget_544",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["5", "6", "7", "8", "9", "10", "11"],
                "setting": {"lockDescendants": True},
            },
            "layout_530": {
                "type": "FIXED",
                "content": {
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "378px",
                            "top": "12.00px",
                            "width": "40.00px",
                            "height": "40px",
                            "right": "20px",
                            "bottom": "381px",
                        },
                        "widgetId": "widget_545",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": True,
                                "top": False,
                            },
                            "order": 0,
                            "lockParent": False,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "right": False,
                                "left": True,
                            },
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "59.9912%",
                            "right": "110.00px",
                            "top": "84.6739%",
                            "bottom": "20.00px",
                            "width": "80px",
                            "height": "32.00px",
                        },
                        "widgetId": "widget_549",
                        "isPending": True,
                    },
                    "10": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "right": False,
                                "left": True,
                            },
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "79.2920%",
                            "right": "20.00px",
                            "top": "84.4355%",
                            "bottom": "20.00px",
                            "width": "80px",
                            "height": "32.00px",
                        },
                        "widgetId": "widget_550",
                        "isPending": True,
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "3.2813%",
                            "top": "9.9561%",
                            "bottom": "3.5474%",
                            "width": "92.8271%",
                            "height": "86.4966%",
                        },
                        "widgetId": "widget_551",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["5", "9", "10", "11"],
                "setting": {"lockDescendants": True},
            },
            "layout_678": {
                "type": "FIXED",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "46.2028%",
                            "top": "69.9531%",
                            "bottom": 33,
                            "width": "40.00%",
                            "height": "13.5000%",
                        },
                        "widgetId": "widget_218",
                        "setting": {
                            "hCenter": True,
                            "vCenter": False,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "241px",
                            "top": "35.8984%",
                            "bottom": "37.1016%",
                            "width": "24.4722%",
                            "height": "27.0000%",
                        },
                        "widgetId": "widget_219",
                        "setting": {
                            "autoProps": {
                                "height": "AUTO",
                                "width": "AUTO",
                                "left": False,
                                "right": True,
                            },
                            "hCenter": True,
                            "vCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["1", "2"],
            },
            "layout_679": {
                "type": "FIXED",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "46.2028%",
                            "top": "69.9531%",
                            "bottom": 33,
                            "width": "40.00%",
                            "height": "13.5000%",
                        },
                        "widgetId": "widget_218",
                        "setting": {
                            "hCenter": True,
                            "vCenter": False,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "45.0000%",
                            "top": "35.9063%",
                            "bottom": "40.0938%",
                            "width": "10.0000%",
                            "height": "24.0000%",
                        },
                        "widgetId": "widget_219",
                        "setting": {
                            "autoProps": {
                                "height": "AUTO",
                                "width": "AUTO",
                                "left": False,
                                "right": True,
                            },
                            "hCenter": True,
                            "vCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["1", "2"],
            },
            "layout_759": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_785",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_786",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_787",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_762": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_794",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_795",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_796",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_765": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_804",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_805",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_806",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_768": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_813",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_814",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_815",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_771": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_823",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_824",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_825",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_774": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_832",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_833",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_834",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_777": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "960.00px",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_841",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_854",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_867",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_880",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_778": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_842",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_843",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_844",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_780": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_848",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_849",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_850",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_782": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_855",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_856",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_857",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_784": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_861",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_862",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_863",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_786": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_868",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_869",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_870",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_788": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_874",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_875",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_876",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_790": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_881",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_882",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_883",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_792": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_887",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_888",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_889",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_799": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "960.00px",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_906",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_919",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_932",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": 960,
                            "height": 320,
                        },
                        "widgetId": "widget_945",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_800": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_907",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_908",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_909",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_802": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_913",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_914",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_915",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_804": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_920",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_921",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_922",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_806": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_926",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_927",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_928",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_808": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_933",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_934",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_935",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_810": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_939",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_940",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_941",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_812": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_946",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_947",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_948",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_814": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_952",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_953",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_954",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_821": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6px",
                            "right": "1286px",
                            "top": "1px",
                            "bottom": "501px",
                            "width": "1280px",
                            "height": "480px",
                        },
                        "widgetId": "widget_971",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "525px",
                            "right": "905px",
                            "top": "-30px",
                            "bottom": "30px",
                            "width": "380px",
                            "height": 93,
                        },
                        "widgetId": "widget_973",
                        "setting": {"heightMode": "fixed", "width": "100%"},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1280px",
                            "top": "512px",
                            "bottom": "2782px",
                            "width": "1280px",
                            "height": "1800px",
                        },
                        "widgetId": "widget_974",
                        "setting": {"heightMode": "auto", "width": "100%"},
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "isFloating": True,
                            "lockParent": True,
                            "floatingArea": 9,
                            "offsetY": 40,
                            "offsetX": 40,
                            "heightMode": "fixed",
                            "width": "64px",
                        },
                        "bbox": {
                            "top": 0,
                            "left": 61.6282958984375,
                            "right": 0,
                            "width": "128px",
                            "height": "48px",
                            "bottom": 0,
                        },
                        "widgetId": "widget_216",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_822": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "480.00px"},
                        "widgetId": "widget_972",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:40",
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_823": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_237",
                        "setting": {
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "44.3056%",
                            "bottom": "43.9382%",
                            "width": "90.00%",
                            "height": "11.7563%",
                        },
                        "widgetId": "widget_238",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                            },
                            "vCenter": False,
                            "hCenter": True,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_824": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": "60px"},
                        "widgetId": "widget_8",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_825": {
                "type": "ROW",
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "1460.00px"},
                        "setting": {
                            "heightMode": "fixed",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {
                                        "direction": None,
                                        "configType": "DEFAULT",
                                    },
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "flex-start"},
                            "aspectRatio": 1.01,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_826": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_265",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_784",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_803",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_822",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_827": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_266",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_267",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_268",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_828": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_269",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_270",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_272",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_829": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_785",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_786",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_787",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_830": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_794",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_795",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_796",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_831": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_804",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_805",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_806",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_832": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_813",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_814",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_815",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_833": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_823",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_824",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_825",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_834": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_832",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_833",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_834",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_835": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_841",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_854",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_867",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_880",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_836": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_842",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_843",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_844",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_837": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_848",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_849",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_850",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_838": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_855",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_856",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_857",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_839": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_861",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_862",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_863",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_840": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_868",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_869",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_870",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_841": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_874",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_875",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_876",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_842": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_881",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_882",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_883",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_843": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_887",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_888",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_889",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_844": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_906",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_919",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_932",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_945",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_845": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_907",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_908",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_909",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_846": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_913",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_914",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_915",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_847": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_920",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_921",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_922",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_848": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_926",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_927",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_928",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_849": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_933",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_934",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_935",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_850": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_939",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_940",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_941",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_851": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_946",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_947",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_948",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_852": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_952",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_953",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_954",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_853": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6px",
                            "right": "1286px",
                            "top": "1px",
                            "bottom": "501px",
                            "width": "1280px",
                            "height": "480px",
                        },
                        "widgetId": "widget_975",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "525px",
                            "right": "905px",
                            "top": "-30px",
                            "bottom": "30px",
                            "width": "380px",
                            "height": 93,
                        },
                        "widgetId": "widget_977",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1280px",
                            "top": "512px",
                            "bottom": "2782px",
                            "width": "1280px",
                            "height": "1800px",
                        },
                        "widgetId": "widget_978",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "isFloating": True,
                            "lockParent": True,
                            "floatingArea": 9,
                            "offsetY": 40,
                            "offsetX": 40,
                            "heightMode": "auto",
                            "width": "64px",
                        },
                        "bbox": {
                            "top": 0,
                            "left": 61.6282958984375,
                            "right": 0,
                            "width": "128px",
                            "height": "48px",
                            "bottom": 0,
                        },
                        "widgetId": "widget_216",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_854": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_979",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_855": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_237",
                        "setting": {
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "50%",
                            "bottom": "43.9382%",
                            "width": "90.00%",
                            "height": "11.7563%",
                        },
                        "widgetId": "widget_238",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_856": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_980",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_857": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_981",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_858": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_265",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_784",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_803",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_822",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_859": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_266",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_267",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_268",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_860": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_269",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_270",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_272",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_861": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_785",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_786",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_787",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_862": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_794",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_795",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_796",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_863": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_804",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_805",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_806",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_864": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_813",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_814",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_815",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_865": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_823",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_824",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_825",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_866": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_832",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_833",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_834",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_867": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_841",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_854",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_867",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_880",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_868": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_842",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_843",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_844",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_869": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_848",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_849",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_850",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_870": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_855",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_856",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_857",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_871": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_861",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_862",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_863",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_872": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_868",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_869",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_870",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_873": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_874",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_875",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_876",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_874": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_881",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_882",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_883",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_875": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_887",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_888",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_889",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_876": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "16.4883%",
                            "top": "2.00%",
                            "bottom": "79.4129%",
                            "width": "100.00%",
                            "height": "320.00px",
                        },
                        "widgetId": "widget_906",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "25.2205%",
                            "bottom": "52.8316%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_919",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "48.5082%",
                            "bottom": "29.5439%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_932",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "14.7361%",
                            "top": "71.7957%",
                            "bottom": "6.2564%",
                            "width": "100.00%",
                            "height": 320,
                        },
                        "widgetId": "widget_945",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_877": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0055%",
                        },
                        "widgetId": "widget_907",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_908",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_909",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_878": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_913",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_914",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_915",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_879": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_920",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_921",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_922",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_880": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_926",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_927",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_928",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_881": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_933",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_934",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_935",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_882": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_939",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_940",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_941",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_883": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_946",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_947",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 56,
                            "top": "50%",
                            "bottom": "41.0373%",
                            "width": "89.9461%",
                            "height": "17.9293%",
                        },
                        "widgetId": "widget_948",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_884": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": 0,
                            "width": "100.0000%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_952",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_953",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.8703%",
                            "top": "50%",
                            "bottom": "16.1530%",
                            "width": "80.00%",
                            "height": "17.5879%",
                        },
                        "widgetId": "widget_954",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_885": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "320.00px"},
                        "widgetId": "widget_976",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:40",
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_886": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 60},
                        "widgetId": "widget_8",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_887": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"width": "auto", "height": "1460.00px"},
                        "setting": {
                            "heightMode": "fixed",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {
                                        "direction": None,
                                        "configType": "DEFAULT",
                                    },
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "flex-start"},
                            "aspectRatio": 1.01,
                        },
                    }
                },
                "order": ["0"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/morandi/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "scenic",
        "widgets": {
            "widget_1": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_1",
            },
            "widget_2": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 18px; color: var(--light);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_2",
            },
            "widget_6": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_6",
                "layouts": {"DEFAULT": {"LARGE": "layout_3"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_7": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_7",
                "layouts": {"DEFAULT": {"LARGE": "layout_5"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_8": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_1",
                        "views": ["view_2", "view_16", "view_17"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "underline",
                        "standard": {
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_8",
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_144": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "About",
                        "textExpression": None,
                        "linkParam": {"linkType": "DIALOG", "value": "dialog_1"},
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "white", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_144",
                "useDataSources": [],
            },
            "widget_145": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "HORIZONTAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "400px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "RIGHT",
                        "offsetX": -15,
                        "offsetY": 0,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 15,
                        "height": 60,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 0 0 92px"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 0 0 92px"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_145",
                "layouts": {
                    "FIRST": {"LARGE": "layout_99"},
                    "SECOND": {"LARGE": "layout_100"},
                },
                "style": {
                    "background": {
                        "color": "var(--primary-500)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_146": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                        "canLocate": True,
                        "canCompass": True,
                        "canLayers": True,
                        "canBaseMap": True,
                        "canMeasure": True,
                        "canFullScreen": True,
                        "canScaleBar": True,
                        "canSelect": False,
                    },
                    "isUseCustomMapState": False,
                    "layoutIndex": 0,
                },
                "id": "widget_146",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_101",
                        "SMALL": "layout_373",
                        "MEDIUM": "layout_490",
                    }
                },
            },
            "widget_148": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_148",
                "useDataSources": [],
            },
            "widget_149": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_149",
            },
            "widget_150": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="text-align: center;"><strong style="color: var(--dark-800); font-size: 32px;">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_150",
            },
            "widget_151": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="text-align: center;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_151",
            },
            "widget_152": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_152",
                "layouts": {"DEFAULT": {"LARGE": "layout_103"}},
            },
            "widget_153": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_153",
                "layouts": {"DEFAULT": {"LARGE": "layout_104"}},
            },
            "widget_216": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "textExpression": None,
                        "linkParam": {"linkType": "DIALOG", "value": "dialog_2"},
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_216",
                "useDataSources": [],
            },
            "widget_218": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;">Powered by esri\ufeff</p>',
                    "placeholder": '<p style="text-align: center;">Double click to edit text</p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_218",
            },
            "widget_219": {
                "uri": "widgets/common/share/",
                "version": "1.5.0",
                "config": {
                    "uiMode": "INLINE",
                    "popup": {
                        "icon": "",
                        "items": ["embed", "qrcode", "email", "facebook", "twitter"],
                        "tooltip": "",
                    },
                    "inline": {
                        "items": ["facebook", "twitter", "email"],
                        "design": {
                            "direction": "HORIZONTAL",
                            "hideLabel": True,
                            "btnRad": 0,
                            "btnColor": "",
                            "iconColor": "",
                            "size": "lg",
                        },
                    },
                },
                "id": "widget_219",
            },
            "widget_237": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_237",
            },
            "widget_238": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px; text-shadow: rgba(0, 0, 0, 0.2) 0px 2px 0px, rgba(0, 0, 0, 0.3) -2px 4px 2px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_238",
            },
            "widget_265": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_4",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_265",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_163",
                        "MEDIUM": "layout_827",
                        "SMALL": "layout_859",
                    },
                    "HOVER": {
                        "LARGE": "layout_164",
                        "MEDIUM": "layout_828",
                        "SMALL": "layout_860",
                    },
                },
            },
            "widget_266": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_266",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_267": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_267",
            },
            "widget_268": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title\ufeff</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_268",
            },
            "widget_269": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_269",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_270": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_270",
            },
            "widget_272": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_272",
            },
            "widget_330": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View1-card01</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_330",
            },
            "widget_331": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_331",
                "layouts": {"DEFAULT": {"LARGE": "layout_182"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_332": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_332",
                "layouts": {"DEFAULT": {"LARGE": "layout_183"}},
            },
            "widget_333": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_333",
                "style": {
                    "background": {
                        "color": "var(--primary-100)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_334": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_334",
                "layouts": {"DEFAULT": {"LARGE": "layout_185"}},
            },
            "widget_335": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_335",
                "layouts": {"DEFAULT": {"LARGE": "layout_186"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_336": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_336",
                "useDataSources": [],
            },
            "widget_337": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View1-card02</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_337",
            },
            "widget_338": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_338",
            },
            "widget_339": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_339",
                "layouts": {"DEFAULT": {"LARGE": "layout_188"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_340": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_340",
                "layouts": {"DEFAULT": {"LARGE": "layout_189"}},
            },
            "widget_341": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_341",
                "useDataSources": [],
            },
            "widget_342": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View1-card03</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_342",
            },
            "widget_343": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_343",
            },
            "widget_344": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_344",
                "layouts": {"DEFAULT": {"LARGE": "layout_191"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_345": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_345",
                "layouts": {"DEFAULT": {"LARGE": "layout_192"}},
            },
            "widget_346": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_346",
                "useDataSources": [],
            },
            "widget_347": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View1-card04</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_347",
            },
            "widget_348": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_348",
            },
            "widget_349": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_349",
                "layouts": {"DEFAULT": {"LARGE": "layout_194"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_350": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_350",
                "layouts": {"DEFAULT": {"LARGE": "layout_195"}},
            },
            "widget_351": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_351",
                "useDataSources": [],
            },
            "widget_352": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View2-card01</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_352",
            },
            "widget_353": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_353",
            },
            "widget_354": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_354",
                "layouts": {"DEFAULT": {"LARGE": "layout_197"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_355": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_355",
                "layouts": {"DEFAULT": {"LARGE": "layout_198"}},
            },
            "widget_356": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_356",
                "useDataSources": [],
            },
            "widget_357": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View2-card02</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_357",
            },
            "widget_358": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_358",
            },
            "widget_359": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_359",
                "layouts": {"DEFAULT": {"LARGE": "layout_200"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_360": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_360",
                "layouts": {"DEFAULT": {"LARGE": "layout_201"}},
            },
            "widget_361": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_361",
                "useDataSources": [],
            },
            "widget_362": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View2-card03</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_362",
            },
            "widget_363": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_363",
            },
            "widget_364": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_364",
                "layouts": {"DEFAULT": {"LARGE": "layout_203"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_365": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_365",
                "layouts": {"DEFAULT": {"LARGE": "layout_204"}},
            },
            "widget_366": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_366",
                "useDataSources": [],
            },
            "widget_367": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View2-card04</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_367",
            },
            "widget_368": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_368",
            },
            "widget_369": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_369",
                "layouts": {"DEFAULT": {"LARGE": "layout_206"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_370": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_370",
                "layouts": {"DEFAULT": {"LARGE": "layout_207"}},
            },
            "widget_371": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_371",
                "useDataSources": [],
            },
            "widget_372": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View3-card01\ufeff</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_372",
            },
            "widget_373": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_373",
            },
            "widget_374": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_374",
                "layouts": {"DEFAULT": {"LARGE": "layout_209"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_375": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_375",
                "layouts": {"DEFAULT": {"LARGE": "layout_210"}},
            },
            "widget_376": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_376",
                "useDataSources": [],
            },
            "widget_377": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View3-card02\ufeff</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_377",
            },
            "widget_378": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_378",
            },
            "widget_379": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_379",
                "layouts": {"DEFAULT": {"LARGE": "layout_212"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_380": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_380",
                "layouts": {"DEFAULT": {"LARGE": "layout_213"}},
            },
            "widget_381": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_381",
                "useDataSources": [],
            },
            "widget_382": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View3-card03\ufeff</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_382",
            },
            "widget_383": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_383",
            },
            "widget_384": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_384",
                "layouts": {"DEFAULT": {"LARGE": "layout_215"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_385": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_385",
                "layouts": {"DEFAULT": {"LARGE": "layout_216"}},
            },
            "widget_386": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M14 8a.522.522 0 01-.516.527H3.976L7.473 12.1a.535.535 0 010 .746.508.508 0 01-.73 0L2 8l4.743-4.846a.508.508 0 01.73 0 .535.535 0 010 .746L3.976 7.473h9.508c.285 0 .516.236.516.527z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {
                                "iconProps": {"color": "var(--black)", "size": 14}
                            },
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_386",
                "useDataSources": [],
            },
            "widget_387": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--light); font-size: 18px;">View3-card04</strong></p>',
                    "placeholder": '<p><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_387",
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_388": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": False,
                    "isInitialed": True,
                    "bookmarks": [],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_388",
            },
            "widget_456": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "VERTICAL",
                    "collapseSide": "FIRST",
                    "overlay": False,
                    "size": "389px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "UP",
                        "offsetX": 0,
                        "offsetY": 15,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 60,
                        "height": 15,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-800)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "0 0 92px 92px"}},
                        "collapseStyle": {"style": {"borderRadius": "0 0 92px 92px"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_456",
                "layouts": {
                    "FIRST": {"SMALL": "layout_372"},
                    "SECOND": {"SMALL": "layout_374"},
                },
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_457": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_457",
                "layouts": {"DEFAULT": {"SMALL": "layout_376"}},
                "style": {
                    "background": {
                        "color": "var(--primary-300)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_458": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_458",
                "layouts": {"DEFAULT": {"SMALL": "layout_377"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_459": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_459",
                "layouts": {"DEFAULT": {"SMALL": "layout_379"}},
            },
            "widget_460": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_460",
                "layouts": {"DEFAULT": {"SMALL": "layout_380"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_461": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_461",
                "layouts": {"DEFAULT": {"SMALL": "layout_382"}},
                "style": {
                    "background": {
                        "color": "var(--primary-300)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_462": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_462",
                "layouts": {"DEFAULT": {"SMALL": "layout_383"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_463": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_463",
                "layouts": {"DEFAULT": {"SMALL": "layout_385"}},
                "style": {
                    "background": {
                        "color": "var(--primary-300)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_464": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_464",
                "layouts": {"DEFAULT": {"SMALL": "layout_386"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_465": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_465",
                "layouts": {"DEFAULT": {"SMALL": "layout_388"}},
                "style": {
                    "background": {
                        "color": "var(--primary-300)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_466": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_466",
                "layouts": {"DEFAULT": {"SMALL": "layout_389"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_467": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_467",
                "layouts": {"DEFAULT": {"SMALL": "layout_391"}},
                "style": {
                    "background": {
                        "color": "var(--primary-300)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_468": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_468",
                "layouts": {"DEFAULT": {"SMALL": "layout_392"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_469": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_469",
                "layouts": {"DEFAULT": {"SMALL": "layout_394"}},
                "style": {
                    "background": {
                        "color": "var(--primary-300)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_470": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_470",
                "layouts": {"DEFAULT": {"SMALL": "layout_395"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_471": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_471",
                "layouts": {"DEFAULT": {"SMALL": "layout_397"}},
                "style": {
                    "background": {
                        "color": "var(--primary-300)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_472": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_472",
                "layouts": {"DEFAULT": {"SMALL": "layout_398"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_473": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_473",
                "layouts": {"DEFAULT": {"SMALL": "layout_400"}},
                "style": {
                    "background": {
                        "color": "var(--primary-300)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_474": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_474",
                "layouts": {"DEFAULT": {"SMALL": "layout_401"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_475": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_475",
                "layouts": {"DEFAULT": {"SMALL": "layout_403"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_476": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_476",
                "layouts": {"DEFAULT": {"SMALL": "layout_404"}},
            },
            "widget_477": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_477",
                "layouts": {"DEFAULT": {"SMALL": "layout_406"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_478": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_478",
                "layouts": {"DEFAULT": {"SMALL": "layout_407"}},
            },
            "widget_479": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_479",
                "layouts": {"DEFAULT": {"SMALL": "layout_409"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_480": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_480",
                "layouts": {"DEFAULT": {"SMALL": "layout_410"}},
            },
            "widget_513": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "VERTICAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "400px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "DOWN",
                        "offsetX": 0,
                        "offsetY": -15,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 60,
                        "height": 15,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_513",
                "layouts": {
                    "FIRST": {"MEDIUM": "layout_489"},
                    "SECOND": {"MEDIUM": "layout_491"},
                },
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_514": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_514",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_493"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_515": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_515",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_494"}},
            },
            "widget_516": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_516",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_496"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_517": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_517",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_497"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_518": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_518",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_499"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_519": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_519",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_500"}},
            },
            "widget_520": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_520",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_502"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_521": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_521",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_503"}},
            },
            "widget_522": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_522",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_505"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_523": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_523",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_506"}},
            },
            "widget_524": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_524",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_508"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_525": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_525",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_509"}},
            },
            "widget_526": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_526",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_511"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_527": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_527",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_512"}},
            },
            "widget_528": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_528",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_514"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_529": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_529",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_515"}},
            },
            "widget_530": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_530",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_517"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_531": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_531",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_518"}},
            },
            "widget_532": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_532",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_520"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_533": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_533",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_521"}},
            },
            "widget_534": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_534",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_523"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_535": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_535",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_524"}},
            },
            "widget_536": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_536",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_526"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_537": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_537",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_527"}},
            },
            "widget_538": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox="0 0 14 14" xmlns="http://www.w3.org/2000/svg"><g fill-rule="nonzero" fill="none"><path d="M7.745 7l4.1 4.1a.527.527 0 01-.745.746L7 7.746l-4.1 4.1a.527.527 0 01-.746-.746l4.1-4.1-4.1-4.1a.527.527 0 01.746-.746l4.1 4.1 4.1-4.1a.527.527 0 01.746.746L7.746 7z" fill="#000"></path><path d="M1 1h12v12H1z"></path></g></svg>',
                            "position": "LEFT",
                        },
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "current_page",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "transparent",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "transparent",
                                    "width": "1px",
                                },
                                "text": {"color": "#181818", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-100)", "size": 20},
                            },
                            "hover": {
                                "background": {"fillType": "fill"},
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"color": "#076fe5", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark)", "size": 20},
                            },
                        },
                    },
                },
                "id": "widget_538",
            },
            "widget_539": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_539",
            },
            "widget_540": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 32px; color: var(--dark-800);">About</strong><strong style="font-size: 14px; color: var(--dark-800);">\ufeff</strong></p>',
                    "placeholder": '<p><strong style="font-size: 14px; color: var(--dark-800);">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_540",
            },
            "widget_541": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_541",
            },
            "widget_542": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "OK",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "current_page",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "themeStyle": {"quickStyleType": "primary"},
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#fff", "size": 16}},
                            "hover": {"iconProps": {"size": 16}},
                        },
                    },
                },
                "id": "widget_542",
            },
            "widget_543": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "Cancel",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "current_page",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {"themeStyle": {"quickStyleType": "default"}},
                },
                "id": "widget_543",
            },
            "widget_544": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {"type": "Style0", "color": "#000", "size": "2px"},
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "Default"},
                },
                "id": "widget_544",
            },
            "widget_545": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox="0 0 14 14" xmlns="http://www.w3.org/2000/svg"><g fill-rule="nonzero" fill="none"><path d="M7.745 7l4.1 4.1a.527.527 0 01-.745.746L7 7.746l-4.1 4.1a.527.527 0 01-.746-.746l4.1-4.1-4.1-4.1a.527.527 0 01.746-.746l4.1 4.1 4.1-4.1a.527.527 0 01.746.746L7.746 7z" fill="#000"></path><path d="M1 1h12v12H1z"></path></g></svg>',
                            "position": "LEFT",
                        },
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "current_page",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "transparent",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "transparent",
                                    "width": "1px",
                                },
                                "text": {"color": "#181818", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-100)", "size": 20},
                            },
                            "hover": {
                                "background": {"fillType": "fill"},
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"color": "#076fe5", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark)", "size": 20},
                            },
                        },
                    },
                },
                "id": "widget_545",
            },
            "widget_549": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "OK",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "current_page",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "themeStyle": {"quickStyleType": "primary"},
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#fff", "size": 16}},
                            "hover": {"iconProps": {"size": 16}},
                        },
                    },
                },
                "id": "widget_549",
            },
            "widget_550": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "Cancel",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "current_page",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {"themeStyle": {"quickStyleType": "default"}},
                },
                "id": "widget_550",
            },
            "widget_551": {
                "uri": "widgets/survey123/",
                "version": "1.5.0",
                "config": {
                    "surveyItemId": "",
                    "portalUrl": "",
                    "embeds": ["fullScreen", "onSubmitted"],
                    "hides": ["navbar", "header", "description", "footer", "theme"],
                },
                "id": "widget_551",
            },
            "widget_784": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_5",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_784",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_759",
                        "MEDIUM": "layout_829",
                        "SMALL": "layout_861",
                    },
                    "HOVER": {
                        "LARGE": "layout_762",
                        "MEDIUM": "layout_830",
                        "SMALL": "layout_862",
                    },
                },
            },
            "widget_785": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_785",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_786": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_786",
            },
            "widget_787": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_787",
            },
            "widget_794": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_794",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_795": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_795",
            },
            "widget_796": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_796",
            },
            "widget_803": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_6",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_803",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_765",
                        "MEDIUM": "layout_831",
                        "SMALL": "layout_863",
                    },
                    "HOVER": {
                        "LARGE": "layout_768",
                        "MEDIUM": "layout_832",
                        "SMALL": "layout_864",
                    },
                },
            },
            "widget_804": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_804",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_805": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_805",
            },
            "widget_806": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_806",
            },
            "widget_813": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_813",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_814": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_814",
            },
            "widget_815": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_815",
            },
            "widget_822": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_7",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_822",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_771",
                        "MEDIUM": "layout_833",
                        "SMALL": "layout_865",
                    },
                    "HOVER": {
                        "LARGE": "layout_774",
                        "MEDIUM": "layout_834",
                        "SMALL": "layout_866",
                    },
                },
            },
            "widget_823": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_823",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_824": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_824",
            },
            "widget_825": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_825",
            },
            "widget_832": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_832",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_833": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_833",
            },
            "widget_834": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_834",
            },
            "widget_841": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_8",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_841",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_778",
                        "MEDIUM": "layout_836",
                        "SMALL": "layout_868",
                    },
                    "HOVER": {
                        "LARGE": "layout_780",
                        "MEDIUM": "layout_837",
                        "SMALL": "layout_869",
                    },
                },
            },
            "widget_842": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_842",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_843": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_843",
            },
            "widget_844": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_844",
            },
            "widget_848": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_848",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_849": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_849",
            },
            "widget_850": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_850",
            },
            "widget_854": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_9",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_854",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_782",
                        "MEDIUM": "layout_838",
                        "SMALL": "layout_870",
                    },
                    "HOVER": {
                        "LARGE": "layout_784",
                        "MEDIUM": "layout_839",
                        "SMALL": "layout_871",
                    },
                },
            },
            "widget_855": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_855",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_856": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_856",
            },
            "widget_857": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_857",
            },
            "widget_861": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_861",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_862": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_862",
            },
            "widget_863": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_863",
            },
            "widget_867": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_10",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_867",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_786",
                        "MEDIUM": "layout_840",
                        "SMALL": "layout_872",
                    },
                    "HOVER": {
                        "LARGE": "layout_788",
                        "MEDIUM": "layout_841",
                        "SMALL": "layout_873",
                    },
                },
            },
            "widget_868": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_868",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_869": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_869",
            },
            "widget_870": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_870",
            },
            "widget_874": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_874",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_875": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_875",
            },
            "widget_876": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_876",
            },
            "widget_880": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_11",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_880",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_790",
                        "MEDIUM": "layout_842",
                        "SMALL": "layout_874",
                    },
                    "HOVER": {
                        "LARGE": "layout_792",
                        "MEDIUM": "layout_843",
                        "SMALL": "layout_875",
                    },
                },
            },
            "widget_881": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_881",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_882": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_882",
            },
            "widget_883": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_883",
            },
            "widget_887": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_887",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_888": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_888",
            },
            "widget_889": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_889",
            },
            "widget_906": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_12",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_906",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_800",
                        "MEDIUM": "layout_845",
                        "SMALL": "layout_877",
                    },
                    "HOVER": {
                        "LARGE": "layout_802",
                        "MEDIUM": "layout_846",
                        "SMALL": "layout_878",
                    },
                },
            },
            "widget_907": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_907",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_908": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_908",
            },
            "widget_909": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeff\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_909",
            },
            "widget_913": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_913",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_914": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_914",
            },
            "widget_915": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_915",
            },
            "widget_919": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_13",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_919",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_804",
                        "MEDIUM": "layout_847",
                        "SMALL": "layout_879",
                    },
                    "HOVER": {
                        "LARGE": "layout_806",
                        "MEDIUM": "layout_848",
                        "SMALL": "layout_880",
                    },
                },
            },
            "widget_920": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_920",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_921": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_921",
            },
            "widget_922": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_922",
            },
            "widget_926": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_926",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_927": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_927",
            },
            "widget_928": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_928",
            },
            "widget_932": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_14",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_932",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_808",
                        "MEDIUM": "layout_849",
                        "SMALL": "layout_881",
                    },
                    "HOVER": {
                        "LARGE": "layout_810",
                        "MEDIUM": "layout_850",
                        "SMALL": "layout_882",
                    },
                },
            },
            "widget_933": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_933",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_934": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_934",
            },
            "widget_935": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_935",
            },
            "widget_939": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_939",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_940": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_940",
            },
            "widget_941": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_941",
            },
            "widget_945": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "10px",
                                "spread": "0px",
                                "color": "rgba(0,0,0,0.2)",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE6",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "linkParam": {
                        "linkType": "VIEW",
                        "openType": "_self",
                        "value": "page_1,view_15",
                    },
                    "widgetWidth": 960,
                    "widgetHeight": 320,
                },
                "id": "widget_945",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_812",
                        "MEDIUM": "layout_851",
                        "SMALL": "layout_883",
                    },
                    "HOVER": {
                        "LARGE": "layout_814",
                        "MEDIUM": "layout_852",
                        "SMALL": "layout_884",
                    },
                },
            },
            "widget_946": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_946",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_947": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_947",
            },
            "widget_948": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 40px;">\ufeffCard title</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 40px;">Card title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_948",
            },
            "widget_952": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "NONE",
                            "value": "page_1",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_952",
                "style": {
                    "boxShadow": {
                        "offsetX": "0px",
                        "offsetY": "0px",
                        "blur": "10px",
                        "spread": "0px",
                        "color": "#000000",
                    }
                },
            },
            "widget_953": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_953",
            },
            "widget_954": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_954",
            },
            "widget_971": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_971",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_822"}},
            },
            "widget_972": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_972",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_823"}},
            },
            "widget_973": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_973",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_824"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_974": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_974",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_825"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_975": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_975",
                "layouts": {"DEFAULT": {"SMALL": "layout_854"}},
            },
            "widget_976": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_976",
                "layouts": {"DEFAULT": {"SMALL": "layout_855"}},
            },
            "widget_977": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_977",
                "layouts": {"DEFAULT": {"SMALL": "layout_856"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_978": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_978",
                "layouts": {"DEFAULT": {"SMALL": "layout_857"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_979": {
                "uri": "widgets/layout/column/",
                "id": "widget_979",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_885"}},
            },
            "widget_980": {
                "uri": "widgets/layout/column/",
                "id": "widget_980",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_886"}},
            },
            "widget_981": {
                "uri": "widgets/layout/column/",
                "id": "widget_981",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_887"}},
            },
        },
        "widgetsManifest": {},
        "views": {
            "view_2": {
                "id": "view_2",
                "layout": {
                    "LARGE": "layout_161",
                    "MEDIUM": "layout_826",
                    "SMALL": "layout_858",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "",
                "backgroundIMage": {},
            },
            "view_4": {
                "id": "view_4",
                "layout": {
                    "LARGE": "layout_181",
                    "SMALL": "layout_375",
                    "MEDIUM": "layout_492",
                },
            },
            "view_5": {
                "id": "view_5",
                "layout": {
                    "LARGE": "layout_184",
                    "SMALL": "layout_378",
                    "MEDIUM": "layout_495",
                },
            },
            "view_6": {
                "id": "view_6",
                "layout": {
                    "LARGE": "layout_187",
                    "SMALL": "layout_381",
                    "MEDIUM": "layout_498",
                },
            },
            "view_7": {
                "id": "view_7",
                "layout": {
                    "LARGE": "layout_190",
                    "SMALL": "layout_384",
                    "MEDIUM": "layout_501",
                },
            },
            "view_8": {
                "id": "view_8",
                "layout": {
                    "LARGE": "layout_193",
                    "SMALL": "layout_387",
                    "MEDIUM": "layout_504",
                },
            },
            "view_9": {
                "id": "view_9",
                "layout": {
                    "LARGE": "layout_196",
                    "SMALL": "layout_390",
                    "MEDIUM": "layout_507",
                },
            },
            "view_10": {
                "id": "view_10",
                "layout": {
                    "LARGE": "layout_199",
                    "SMALL": "layout_393",
                    "MEDIUM": "layout_510",
                },
            },
            "view_11": {
                "id": "view_11",
                "layout": {
                    "LARGE": "layout_202",
                    "SMALL": "layout_396",
                    "MEDIUM": "layout_513",
                },
            },
            "view_12": {
                "id": "view_12",
                "layout": {
                    "LARGE": "layout_205",
                    "SMALL": "layout_399",
                    "MEDIUM": "layout_516",
                },
            },
            "view_13": {
                "id": "view_13",
                "layout": {
                    "LARGE": "layout_208",
                    "SMALL": "layout_402",
                    "MEDIUM": "layout_519",
                },
            },
            "view_14": {
                "id": "view_14",
                "layout": {
                    "LARGE": "layout_211",
                    "SMALL": "layout_405",
                    "MEDIUM": "layout_522",
                },
            },
            "view_15": {
                "id": "view_15",
                "layout": {
                    "LARGE": "layout_214",
                    "SMALL": "layout_408",
                    "MEDIUM": "layout_525",
                },
            },
            "view_16": {
                "id": "view_16",
                "layout": {
                    "LARGE": "layout_777",
                    "MEDIUM": "layout_835",
                    "SMALL": "layout_867",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "",
                "backgroundIMage": {},
            },
            "view_17": {
                "id": "view_17",
                "layout": {
                    "LARGE": "layout_799",
                    "MEDIUM": "layout_844",
                    "SMALL": "layout_876",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "",
                "backgroundIMage": {},
            },
        },
        "sections": {
            "section_1": {
                "id": "section_1",
                "views": ["view_2", "view_16", "view_17"],
                "transition": {"type": "PUSH"},
            },
            "section_2": {
                "id": "section_2",
                "views": [
                    "view_4",
                    "view_5",
                    "view_6",
                    "view_7",
                    "view_8",
                    "view_9",
                    "view_10",
                    "view_11",
                    "view_12",
                    "view_13",
                    "view_14",
                    "view_15",
                ],
            },
        },
        "dialogs": {
            "dialog_1": {
                "id": "dialog_1",
                "mode": "FIXED",
                "sizeMode": {
                    "LARGE": {"position": "CENTER", "width": "960px", "height": "640px"}
                },
                "dialogBackground": {
                    "color": "var(--light)",
                    "fillType": "fit",
                    "image": {"url": ""},
                },
                "overlayColor": "var(--light-500)",
                "layout": {"LARGE": "layout_529"},
                "index": 0,
                "isSplash": False,
                "closeWhenClickOverlay": True,
            },
            "dialog_2": {
                "id": "dialog_2",
                "mode": "FIXED",
                "sizeMode": {
                    "LARGE": {"position": "CENTER", "width": "640px", "height": "640px"}
                },
                "dialogBackground": {
                    "color": "var(--light)",
                    "fillType": "fit",
                    "image": {"url": ""},
                },
                "overlayColor": "var(--light-500)",
                "layout": {"LARGE": "layout_530"},
                "index": 1,
                "isSplash": False,
                "closeWhenClickOverlay": True,
            },
        },
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_0": []}, {"page_1": []}],
        "header": {
            "layout": {
                "LARGE": "layout_1",
                "SMALL": "layout_297",
                "MEDIUM": "layout_528",
            },
            "height": {"LARGE": "72px"},
            "sticky": True,
            "backgroundColor": "var(--primary-500)",
        },
        "timestamp": 1619661676048,
        "footer": {
            "layout": {
                "LARGE": "layout_143",
                "MEDIUM": "layout_678",
                "SMALL": "layout_679",
            },
            "height": {"LARGE": "200px"},
            "backgroundColor": "",
        },
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "exhibition": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_0",
                    "MEDIUM": "layout_482",
                    "SMALL": "layout_527",
                },
                "isDefault": True,
                "isVisible": True,
                "bodyBackgroundColor": "",
                "header": True,
                "footer": True,
                "oneByOneEffect": {"type": "FLOAT_IN", "option": {"direction": None}},
            },
            "page_1": {
                "id": "page_1",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_2",
                    "MEDIUM": "layout_696",
                    "SMALL": "layout_708",
                },
                "isVisible": True,
                "isDefault": False,
                "header": True,
                "bodyBackgroundColor": "",
                "footer": True,
                "oneByOneEffect": {"type": "FLOAT_IN", "option": {"direction": None}},
            },
            "page_2": {
                "id": "page_2",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_3",
                    "MEDIUM": "layout_252",
                    "SMALL": "layout_676",
                },
                "isVisible": True,
                "isDefault": False,
                "header": True,
                "bodyBackgroundColor": "",
                "footer": True,
                "oneByOneEffect": {"type": "FLOAT_IN", "option": {"direction": None}},
            },
            "page_3": {
                "id": "page_3",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_4",
                    "MEDIUM": "layout_290",
                    "SMALL": "layout_440",
                },
                "isVisible": True,
                "isDefault": False,
                "header": True,
                "bodyBackgroundColor": "",
                "footer": True,
                "oneByOneEffect": {"type": "FLOAT_IN", "option": {"direction": None}},
            },
            "page_4": {
                "id": "page_4",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_5",
                    "SMALL": "layout_308",
                    "MEDIUM": "layout_326",
                },
                "isVisible": True,
                "isDefault": False,
                "header": True,
                "bodyBackgroundColor": "",
                "footer": True,
                "oneByOneEffect": {"type": "FLOAT_IN", "option": {"direction": None}},
            },
        },
        "layouts": {
            "layout_0": {
                "type": "FLOW",
                "content": {
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-263px",
                            "right": "1017px",
                            "top": "80px",
                            "bottom": "180px",
                            "width": "1280px",
                            "height": "160px",
                        },
                        "widgetId": "widget_16",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "220px",
                            "right": "580px",
                            "top": "316px",
                            "bottom": "386px",
                            "width": "360px",
                            "height": "80px",
                        },
                        "widgetId": "widget_18",
                        "setting": {"heightMode": "auto", "offsetX": 0, "offsetY": 0},
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "314px",
                            "right": "614px",
                            "top": "476px",
                            "bottom": "776px",
                            "width": "300px",
                            "height": "60%",
                        },
                        "widgetId": "widget_20",
                        "setting": {"heightMode": "fixed", "offsetY": 0, "offsetX": 0},
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35px",
                            "right": "535px",
                            "top": "1229px",
                            "bottom": "1729px",
                            "width": "500px",
                            "height": "430px",
                        },
                        "widgetId": "widget_34",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "1833px",
                            "bottom": "1888px",
                            "width": "1200px",
                            "height": "60px",
                        },
                        "widgetId": "widget_54",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                        "isPending": False,
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "151px",
                            "right": "551px",
                            "top": "1960px",
                            "bottom": "2260px",
                            "width": "400px",
                            "height": "450px",
                        },
                        "widgetId": "widget_69",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "941px",
                            "bottom": "1041px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_73",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "17": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "1337px",
                            "bottom": "1437px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_75",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "18": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "42px",
                            "right": "1242px",
                            "top": "1940px",
                            "bottom": "2040px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_77",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "19": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-91px",
                            "right": "269px",
                            "top": "2369px",
                            "bottom": "2439px",
                            "width": "360px",
                            "height": 89,
                        },
                        "widgetId": "widget_82",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                    },
                    "20": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "2561px",
                            "bottom": "2651px",
                            "width": "1200px",
                            "height": 118,
                        },
                        "widgetId": "widget_85",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                        "isPending": False,
                    },
                    "23": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "3400px",
                            "bottom": "3640px",
                            "width": "1200px",
                            "height": "240px",
                        },
                        "widgetId": "widget_119",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                        "isPending": False,
                    },
                    "24": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "86px",
                            "right": "486px",
                            "top": "3260px",
                            "bottom": "3660px",
                            "width": "400px",
                            "height": "500px",
                        },
                        "widgetId": "widget_123",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                    },
                    "27": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_140",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    },
                    "28": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_145",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    },
                    "29": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "4347px",
                            "bottom": "4447px",
                            "width": "1200px",
                            "height": "200px",
                        },
                        "widgetId": "widget_150",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "30": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-18px",
                            "right": "1182px",
                            "top": "4200px",
                            "bottom": "4440px",
                            "width": "1200px",
                            "height": "240px",
                        },
                        "widgetId": "widget_152",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                        "isPending": False,
                    },
                    "33": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-177px",
                            "right": "223px",
                            "top": "-199px",
                            "bottom": "201px",
                            "width": "400px",
                            "height": "480px",
                        },
                        "widgetId": "widget_439",
                        "setting": {"heightMode": "fixed"},
                    },
                    "34": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-121px",
                            "right": "179px",
                            "top": "-191px",
                            "bottom": "200px",
                            "width": "300px",
                            "height": 343,
                        },
                        "widgetId": "widget_543",
                        "setting": {
                            "heightMode": "auto",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": "BOTTOM"},
                                }
                            },
                        },
                    },
                },
                "order": [
                    "5",
                    "6",
                    "7",
                    "16",
                    "34",
                    "17",
                    "11",
                    "12",
                    "18",
                    "14",
                    "19",
                    "20",
                    "33",
                    "23",
                    "24",
                    "30",
                    "27",
                    "28",
                    "29",
                ],
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
            },
            "layout_2": {
                "type": "FLOW",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "294px",
                            "right": "654px",
                            "top": "-32px",
                            "bottom": "38px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_181",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "102px",
                            "right": "462px",
                            "top": "-34px",
                            "bottom": "36px",
                            "width": "360px",
                            "height": "420px",
                        },
                        "widgetId": "widget_184",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "576px",
                            "right": "876px",
                            "top": "-293px",
                            "bottom": "307px",
                            "width": "300px",
                            "height": "340px",
                        },
                        "widgetId": "widget_201",
                        "setting": {"heightMode": "fixed", "offsetY": 60},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "64px",
                            "right": "1216px",
                            "top": "921px",
                            "bottom": "1081px",
                            "width": "1152px",
                            "height": "160px",
                        },
                        "widgetId": "widget_212",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-56px",
                            "right": "344px",
                            "top": "904px",
                            "bottom": "1304px",
                            "width": "400px",
                            "height": 437,
                        },
                        "widgetId": "widget_659",
                        "setting": {"heightMode": "fixed"},
                    },
                },
                "order": ["1", "2", "3", "4", "5"],
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
            },
            "layout_3": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "97px",
                            "right": "457px",
                            "top": "55px",
                            "bottom": "125px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_217",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "width": "100%",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20px",
                            "right": "520px",
                            "top": "-38px",
                            "bottom": "462px",
                            "width": "500px",
                            "height": "360px",
                        },
                        "widgetId": "widget_219",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "643px",
                            "bottom": "803px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_226",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "820px",
                            "bottom": "1180px",
                            "width": "1200px",
                            "height": "500px",
                        },
                        "widgetId": "widget_228",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
            },
            "layout_4": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "138px",
                            "right": "498px",
                            "top": "87px",
                            "bottom": "157px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_236",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetY": 20,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "67px",
                            "right": "567px",
                            "top": "26px",
                            "bottom": "526px",
                            "width": "500px",
                            "height": "630px",
                        },
                        "widgetId": "widget_238",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                    },
                },
                "order": ["0", "1"],
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
            },
            "layout_5": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "229px",
                            "right": "589px",
                            "top": "62px",
                            "bottom": "132px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_244",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetY": 20,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-35px",
                            "right": "265px",
                            "top": "164px",
                            "bottom": "214px",
                            "width": "300px",
                            "height": "2100px",
                        },
                        "widgetId": "widget_246",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                    },
                },
                "order": ["0", "1"],
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
            },
            "layout_6": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "41.67%",
                            "height": "100.00%",
                            "right": 278,
                            "bottom": "1px",
                        },
                        "widgetId": "widget_3",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "47.9023%",
                            "top": "16.8175%",
                            "width": "45.8750%",
                            "height": "48px",
                            "right": "6.2227%",
                            "bottom": "73.6656%",
                        },
                        "widgetId": "widget_4",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": 383,
                            "top": "46.5219%",
                            "width": "45.8750%",
                            "height": "32.0000%",
                            "right": "6.2227%",
                            "bottom": 107,
                        },
                        "widgetId": "widget_5",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "47.9219%",
                            "right": "55px",
                            "top": "32.0818%",
                            "bottom": "59.1406%",
                            "width": "45.1426%",
                            "height": "8.7776%",
                        },
                        "widgetId": "widget_6",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_10": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_17",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
            },
            "layout_11": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": 70},
                        "widgetId": "widget_19",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_12": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600.00px"},
                        "widgetId": "widget_21",
                        "setting": {
                            "heightMode": "fit",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
            },
            "layout_19": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": 414},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    }
                },
            },
            "layout_20": {
                "type": "FIXED",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "52.2007%",
                            "top": "50%",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_35",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1464%",
                            "right": "2.4252%",
                            "top": "22px",
                            "bottom": "84.5411%",
                            "width": "45.00%",
                            "height": "10.1449%",
                        },
                        "widgetId": "widget_36",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0295%",
                            "right": 0,
                            "top": "21.0255%",
                            "bottom": "25.3337%",
                            "width": "48.1422%",
                            "height": "53.6408%",
                        },
                        "widgetId": "widget_37",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": "20.00px",
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_38",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"bottom": False, "top": True},
                        },
                    },
                },
            },
            "layout_26": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "offsetX": -10,
                            "autoProps": {"height": False},
                        },
                        "bbox": {"left": 2, "width": 2, "height": 30},
                        "widgetId": "widget_53",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 11, "width": 1, "height": 60},
                        "widgetId": "widget_128",
                        "setting": {"heightMode": "fit"},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_31": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 300},
                        "widgetId": "widget_70",
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_33": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_74",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_34": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "50px"},
                        "widgetId": "widget_76",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_35": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "50px"},
                        "widgetId": "widget_78",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_36": {
                "type": "FIXED",
                "order": ["0", "1", "2", "3", "4", "5"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "11.5198%",
                            "right": "41.1067%",
                            "top": 0,
                            "bottom": 0,
                            "width": "47.3736%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_79",
                        "isPending": True,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.9026%",
                            "right": "90.9117%",
                            "top": "18px",
                            "bottom": "18px",
                            "width": "5.1857%",
                            "height": "52.0000%",
                        },
                        "widgetId": "widget_80",
                        "isPending": True,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "89.4221%",
                            "right": "1.6717%",
                            "top": 13,
                            "bottom": "16.6667%",
                            "width": "8.9063%",
                            "height": "66.6667%",
                        },
                        "widgetId": "widget_81",
                        "isPending": True,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.142500000000007%",
                            "right": "55.73250000000003%",
                            "top": "5px",
                            "bottom": "5px",
                            "width": "38.1250%",
                            "height": "83.3333%",
                        },
                        "widgetId": "widget_158",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.9207999999999995%",
                            "right": "94.49479999999998%",
                            "top": "10px",
                            "bottom": "13px",
                            "width": "4.5844%",
                            "height": "62.3958%",
                        },
                        "widgetId": "widget_159",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "92.0224%",
                            "right": 22,
                            "top": "13px",
                            "bottom": "23.459400000000002%",
                            "width": "6.2500%",
                            "height": "53.3333%",
                        },
                        "widgetId": "widget_160",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
            },
            "layout_37": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "50px"},
                        "widgetId": "widget_83",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "flex-end"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 4, "height": "50px"},
                        "widgetId": "widget_86",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "flex-end"},
                        },
                    },
                },
            },
            "layout_38": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 4, "height": 70},
                        "widgetId": "widget_84",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 5, "width": 4, "height": 70},
                        "widgetId": "widget_87",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                        "bbox": {"left": 10, "width": 2, "height": "42px"},
                        "widgetId": "widget_88",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_47": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 400},
                        "widgetId": "widget_120",
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_48": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "vCenter": False, "hCenter": False},
                        "bbox": {
                            "left": 0,
                            "right": "65.7059%",
                            "top": 73,
                            "bottom": "30.1771%",
                            "width": "34.2954%",
                            "height": "35.9447%",
                        },
                        "widgetId": "widget_121",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "49.3209%",
                            "right": 0,
                            "top": "34.1014%",
                            "bottom": 10,
                            "width": "50.8489%",
                            "height": "61.2903%",
                        },
                        "widgetId": "widget_122",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_49": {
                "type": "ROW",
                "order": ["1", "0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": 400},
                        "widgetId": "widget_124",
                        "setting": {"heightMode": "fit"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "490px"},
                        "widgetId": "widget_129",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
            },
            "layout_50": {"type": "FIXED"},
            "layout_54": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.58%",
                            "top": "50%",
                            "width": "13.39%",
                            "height": "55.81%",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "21.20%",
                            "width": "226px",
                            "height": "35%",
                            "right": "39.05%",
                            "top": "50%",
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "230px",
                            "top": "50%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.25%",
                            "bottom": "20.93%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_55": {"type": "FIXED"},
            "layout_57": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "9px",
                            "top": "10px",
                            "width": "29.8968%",
                            "height": "86.5369%",
                            "right": "67.07189999999997%",
                            "bottom": "8px",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35.0191%",
                            "width": 238,
                            "height": "43.6104%",
                            "right": "3.1237%",
                            "top": 10,
                            "bottom": 74,
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "274px",
                            "top": "69.6731999999997%",
                            "width": "27.8170%",
                            "height": "23.8722%",
                            "right": "2.2273999999999994%",
                            "bottom": "6.163100000000027%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_60": {
                "type": "ROW",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 89},
                        "widgetId": "widget_141",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": 89},
                        "widgetId": "widget_142",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": 89},
                        "widgetId": "widget_143",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": 89},
                        "widgetId": "widget_144",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
            },
            "layout_61": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 89},
                        "widgetId": "widget_146",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": 89},
                        "widgetId": "widget_147",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": 89},
                        "widgetId": "widget_148",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": 89},
                        "widgetId": "widget_149",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_62": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_151",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_63": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 400},
                        "widgetId": "widget_153",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_64": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO", "width": "AUTO"},
                        },
                        "bbox": {
                            "left": 0,
                            "right": "65.7059%",
                            "top": 73,
                            "bottom": "30.1771%",
                            "width": "34.2954%",
                            "height": "35.9447%",
                        },
                        "widgetId": "widget_154",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "49.3209%",
                            "right": 0,
                            "top": "34.1014%",
                            "bottom": 10,
                            "width": "50.8489%",
                            "height": "61.2903%",
                        },
                        "widgetId": "widget_155",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO", "width": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_66": {
                "type": "FIXED",
                "order": [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                    "16",
                    "17",
                    "18",
                ],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "4.3569%",
                            "right": "71.6587%",
                            "top": "8.0324%",
                            "bottom": "70.9676%",
                            "width": "23.9844%",
                            "height": "21.0000%",
                        },
                        "widgetId": "widget_163",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "36.2927%",
                            "right": "53.7299%",
                            "top": "9.4688%",
                            "bottom": "83.1312%",
                            "width": "9.9774%",
                            "height": "7.4000%",
                        },
                        "widgetId": "widget_164",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "58.0750%",
                            "right": "33.2813%",
                            "top": "9.4688%",
                            "bottom": "83.1312%",
                            "width": "8.6438%",
                            "height": "7.4000%",
                        },
                        "widgetId": "widget_165",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "79.3420%",
                            "right": "11.4225%",
                            "top": "9.4688%",
                            "bottom": "83.1312%",
                            "width": "9.2355%",
                            "height": "7.4000%",
                        },
                        "widgetId": "widget_166",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.8139%",
                            "right": "57.1392%",
                            "top": "21.2180%",
                            "bottom": "68.7820%",
                            "width": "8.0469%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_167",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.8120%",
                            "right": "57.1411%",
                            "top": "32.7693%",
                            "bottom": "57.2307%",
                            "width": "8.0469%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_168",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.8108%",
                            "right": "57.1423%",
                            "top": "45.1555%",
                            "bottom": "44.8445%",
                            "width": "8.0469%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_169",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.8108%",
                            "right": "57.1423%",
                            "top": "57.8375%",
                            "bottom": "32.1625%",
                            "width": "8.0469%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_170",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "56.6255%",
                            "right": "35.3276%",
                            "top": "21.2156%",
                            "bottom": "68.7844%",
                            "width": "8.0469%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_171",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "56.6248%",
                            "right": "35.3284%",
                            "top": "32.7687%",
                            "bottom": "57.2313%",
                            "width": "8.0469%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_172",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "56.6248%",
                            "right": "35.3284%",
                            "top": "45.1531%",
                            "bottom": "44.8469%",
                            "width": "8.0469%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_173",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "77.8635%",
                            "right": "14.0896%",
                            "top": "21.2156%",
                            "bottom": "68.7844%",
                            "width": "8.0469%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_174",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "77.8625%",
                            "right": "14.0906%",
                            "top": "32.7656%",
                            "bottom": "57.2344%",
                            "width": "8.0469%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_175",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "77.8625%",
                            "right": "14.0906%",
                            "top": "45.1531%",
                            "bottom": "44.8469%",
                            "width": "8.0469%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_176",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "36.2903%",
                            "right": "3.9288%",
                            "top": "72.1813%",
                            "bottom": "17.8188%",
                            "width": "59.7809%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_177",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "15": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "36.2903%",
                            "right": "53.7329%",
                            "top": "84.5656%",
                            "bottom": "8.0344%",
                            "width": "10.0000%",
                            "height": "7.4000%",
                        },
                        "widgetId": "widget_178",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "48.898600000000016%",
                            "right": "31.726399999999984%",
                            "top": "84.96559999999998%",
                            "bottom": "7.634400000000001%",
                            "width": "19.3750%",
                            "height": "7.4000%",
                        },
                        "widgetId": "widget_179",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "17": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.57799999999982%",
                            "right": "7.047000000000011%",
                            "top": "84.96869999999998%",
                            "bottom": "7.631300000000001%",
                            "width": "19.3750%",
                            "height": "7.4000%",
                        },
                        "widgetId": "widget_180",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "18": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "4.3567%",
                            "right": "68.2996%",
                            "top": 185,
                            "bottom": "51.4375%",
                            "width": "27.3438%",
                            "height": "11.6000%",
                        },
                        "widgetId": "widget_903",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
            },
            "layout_68": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": 50},
                        "widgetId": "widget_182",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 4, "height": 50},
                        "widgetId": "widget_183",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    },
                },
            },
            "layout_69": {
                "type": "ROW",
                "order": ["1", "2"],
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": 600},
                        "widgetId": "widget_186",
                        "setting": {"heightMode": "fit"},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "400px"},
                        "widgetId": "widget_190",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
            },
            "layout_70": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "595px", "height": "100px"},
                        "widgetId": "widget_185",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "flex-start"},
                            "autoProps": {"width": False},
                        },
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_187",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": False}},
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_188",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": False}},
                        "bbox": {"width": 583, "height": "80px"},
                        "widgetId": "widget_189",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_71": {
                "type": "COLUMN",
                "order": ["0", "1", "2", "3", "4", "6", "7", "5", "8", "9"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": 40},
                        "widgetId": "widget_191",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "10px"},
                        "widgetId": "widget_192",
                        "setting": {"autoProps": {"width": False}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_193",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_194",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_195",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_196",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_197",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_198",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_199",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_200",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                },
            },
            "layout_72": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": 600},
                        "widgetId": "widget_202",
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_73": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_203",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_204",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_205",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_206",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_207",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_208",
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_209",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_210",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7"],
            },
            "layout_74": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "offsetY": 20,
                        },
                        "bbox": {
                            "left": "0px",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "4px",
                            "height": "50px",
                        },
                        "widgetId": "widget_211",
                    }
                },
                "order": ["0"],
            },
            "layout_79": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": 50},
                        "widgetId": "widget_218",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
            },
            "layout_80": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "sectionId": "section_4",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 8, "height": 360},
                        "setting": {"heightMode": "fit"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": 320},
                        "widgetId": "widget_220",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": True},
                            "style": {"alignSelf": "center"},
                            "offsetX": 0,
                            "offsetY": 0,
                        },
                    },
                },
            },
            "layout_81": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "lockParent": True},
                        "bbox": {
                            "left": "0px",
                            "right": 0,
                            "top": "1px",
                            "bottom": "0.2950%",
                            "width": "100.0000%",
                            "height": "99.4100%",
                        },
                        "widgetId": "widget_222",
                    }
                },
                "order": ["0"],
            },
            "layout_82": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.1057%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_223",
                        "setting": {"lockParent": True},
                    }
                },
            },
            "layout_83": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00px",
                            "right": "1px",
                            "top": "0px",
                            "bottom": 0,
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_221",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
            },
            "layout_84": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "34.1793%",
                            "top": "0.00%",
                            "bottom": "-11.2760%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_224",
                        "setting": {"lockParent": True},
                    }
                },
            },
            "layout_85": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.2523%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_225",
                        "setting": {"lockParent": True},
                    }
                },
            },
            "layout_86": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": 50},
                        "widgetId": "widget_227",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_87": {
                "type": "ROW",
                "content": {
                    "0": {
                        "sectionId": "section_5",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 8, "height": 360},
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": 320},
                        "widgetId": "widget_234",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": True},
                            "style": {"alignSelf": "center"},
                            "offsetX": 0,
                            "offsetY": 0,
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_88": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "lockParent": True},
                        "bbox": {
                            "left": "0px",
                            "right": 0,
                            "top": "1px",
                            "bottom": "0.2950%",
                            "width": "100.0000%",
                            "height": "99.4100%",
                        },
                        "widgetId": "widget_229",
                    }
                },
                "order": ["0"],
            },
            "layout_89": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.1057%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_230",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_90": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00px",
                            "right": "1px",
                            "top": "0px",
                            "bottom": 0,
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_231",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_91": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "34.1793%",
                            "top": "0.00%",
                            "bottom": "-11.2760%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_232",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_92": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.2523%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_233",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_93": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.0372%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_235",
                    }
                },
            },
            "layout_94": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": 50},
                        "widgetId": "widget_237",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "offsetY": 20,
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": "8px", "width": "4px", "height": "60px"},
                        "widgetId": "widget_241",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    },
                },
            },
            "layout_95": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "sectionId": "section_6",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": 500},
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_96": {
                "type": "FIXED",
                "order": ["1", "0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 54,
                            "right": "68.5931%",
                            "top": 62,
                            "bottom": "12.0724%",
                            "width": "320px",
                            "height": "380px",
                        },
                        "widgetId": "widget_239",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": True,
                            "autoProps": {"width": False, "height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "27.2496%",
                            "top": "0.00%",
                            "bottom": "29.3763%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_240",
                        "setting": {"lockParent": True},
                    },
                },
            },
            "layout_97": {"type": "FIXED"},
            "layout_98": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "25.9776%",
                            "right": 0,
                            "top": 0,
                            "bottom": "0.0000%",
                            "width": "74.1935%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_242",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": True,
                        },
                        "bbox": {
                            "left": 0,
                            "right": "74.1100%",
                            "top": 0,
                            "bottom": 298,
                            "width": "25.8913%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_243",
                        "isPending": False,
                    },
                },
            },
            "layout_99": {"type": "FIXED", "order": [], "content": {}},
            "layout_100": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 5, "height": "50px"},
                        "widgetId": "widget_245",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
            },
            "layout_101": {
                "type": "ROW",
                "order": ["0", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 1, "height": "1940px"},
                        "widgetId": "widget_247",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 1, "width": 11, "height": 600},
                        "widgetId": "widget_249",
                        "setting": {"heightMode": "fit"},
                    },
                },
            },
            "layout_102": {
                "type": "COLUMN",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "280px"},
                        "widgetId": "widget_250",
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_248",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_254",
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_255",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_259",
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_260",
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_264",
                        "isPending": False,
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_265",
                        "isPending": False,
                    },
                    "11": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_269",
                        "isPending": False,
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_270",
                        "isPending": False,
                    },
                    "13": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_273",
                        "isPending": False,
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_274",
                        "isPending": False,
                    },
                },
                "order": [
                    "4",
                    "1",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                ],
            },
            "layout_103": {
                "type": "ROW",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "180px"},
                        "widgetId": "widget_251",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "180px"},
                        "widgetId": "widget_252",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 5, "height": "180px"},
                        "widgetId": "widget_253",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
            },
            "layout_104": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 180},
                        "widgetId": "widget_256",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 7, "height": 180},
                        "widgetId": "widget_258",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "2"],
            },
            "layout_105": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 2, "height": 180},
                        "widgetId": "widget_261",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 5, "height": 180},
                        "widgetId": "widget_262",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 2, "height": 180},
                        "widgetId": "widget_263",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "2", "1"],
            },
            "layout_106": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "180px"},
                        "widgetId": "widget_266",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "180px"},
                        "widgetId": "widget_267",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 5, "height": "180px"},
                        "widgetId": "widget_268",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_107": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 5, "height": "180px"},
                        "widgetId": "widget_271",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 6, "height": 180},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_108": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "180px"},
                        "widgetId": "widget_275",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 5, "height": 180},
                        "widgetId": "widget_276",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_109": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": 0,
                            "top": 0,
                            "width": "41.67%",
                            "height": "100.00%",
                            "right": 278,
                            "bottom": "1px",
                        },
                        "widgetId": "widget_277",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "0.00%",
                            "top": "10.00%",
                            "width": "45.83%",
                            "height": "45.00px",
                            "right": "6.25%",
                            "bottom": "76.0029%",
                        },
                        "widgetId": "widget_278",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": True,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "383px",
                            "top": "40.00%",
                            "width": "45.4868%",
                            "height": "240.00px",
                            "right": "6.4732%",
                            "bottom": "105px",
                        },
                        "widgetId": "widget_279",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "48.2367%",
                            "right": "6.7956%",
                            "top": "24.1027%",
                            "bottom": "58.2329%",
                            "width": "44.9677%",
                            "height": "80.00px",
                        },
                        "widgetId": "widget_280",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "92.2203%",
                            "right": 9,
                            "top": 0,
                            "bottom": "89.9598%",
                            "width": "6.6416%",
                            "height": "10.0402%",
                        },
                        "widgetId": "widget_902",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
            },
            "layout_110": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "8.337999999999992%",
                            "right": "14.497299999999996%",
                            "top": "4px",
                            "bottom": "6px",
                            "width": "77.1647%",
                            "height": "83.3073%",
                        },
                        "widgetId": "widget_158",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.9193%",
                            "right": "90.7913%",
                            "top": "10px",
                            "bottom": "12px",
                            "width": "8.2894%",
                            "height": "62.3698%",
                        },
                        "widgetId": "widget_159",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "87.2845%",
                            "right": "10px",
                            "top": "13px",
                            "bottom": "25.0260%",
                            "width": "10.9863%",
                            "height": "53.3073%",
                        },
                        "widgetId": "widget_160",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_150": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.4983%",
                            "right": "23.0463%",
                            "top": "4px",
                            "bottom": "6px",
                            "width": "63.4554%",
                            "height": "83.3073%",
                        },
                        "widgetId": "widget_158",
                        "isPending": True,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.9201%",
                            "right": "87.5043%",
                            "top": "10px",
                            "bottom": "12px",
                            "width": "11.5755%",
                            "height": "62.3698%",
                        },
                        "widgetId": "widget_159",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.7274%",
                            "right": "6px",
                            "top": "13px",
                            "bottom": "25.4948%",
                            "width": "16.5408%",
                            "height": "53.3073%",
                        },
                        "widgetId": "widget_160",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "12.4870%",
                            "right": "68.3464%",
                            "top": 0,
                            "bottom": 0,
                            "width": "19.1667%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_435",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_252": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 40, 0, 40], "unit": "px"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "97px",
                            "right": "457px",
                            "top": "55px",
                            "bottom": "125px",
                            "width": "360px",
                            "height": "100px",
                        },
                        "widgetId": "widget_387",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "width": "100%",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20px",
                            "right": "520px",
                            "top": "-38px",
                            "bottom": "462px",
                            "width": "500px",
                            "height": "360px",
                        },
                        "widgetId": "widget_388",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "643px",
                            "bottom": "803px",
                            "width": "1200px",
                            "height": "100px",
                        },
                        "widgetId": "widget_389",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "820px",
                            "bottom": "1180px",
                            "width": "1200px",
                            "height": "360px",
                        },
                        "widgetId": "widget_390",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "329px",
                            "right": "499px",
                            "top": "-110px",
                            "bottom": "210px",
                            "width": "170px",
                            "height": 48,
                        },
                        "widgetId": "widget_391",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "376px",
                            "right": "546px",
                            "top": "1038px",
                            "bottom": "1358px",
                            "width": "170px",
                            "height": 53,
                        },
                        "widgetId": "widget_392",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                    },
                },
                "order": ["0", "1", "4", "2", "3", "5"],
            },
            "layout_253": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 50},
                        "widgetId": "widget_218",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_254": {
                "type": "ROW",
                "content": {
                    "0": {
                        "sectionId": "section_4",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "340px"},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_255": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "lockParent": True},
                        "bbox": {
                            "left": "0px",
                            "right": 0,
                            "top": "2px",
                            "bottom": "0.5900%",
                            "width": "100.0000%",
                            "height": "98.8201%",
                        },
                        "widgetId": "widget_222",
                    }
                },
                "order": ["0"],
            },
            "layout_256": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.1057%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_223",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_257": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00px",
                            "right": "1px",
                            "top": "0px",
                            "bottom": 0,
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_221",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_258": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "34.1793%",
                            "top": "0.00%",
                            "bottom": "-11.2760%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_224",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_259": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.2523%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_225",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_260": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 50},
                        "widgetId": "widget_227",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_261": {
                "type": "ROW",
                "content": {
                    "0": {
                        "sectionId": "section_5",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "340px"},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_262": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "lockParent": True},
                        "bbox": {
                            "left": "0px",
                            "right": 0,
                            "top": "1px",
                            "bottom": "0.2950%",
                            "width": "100.0000%",
                            "height": "99.4100%",
                        },
                        "widgetId": "widget_229",
                    }
                },
                "order": ["0"],
            },
            "layout_263": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.1057%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_230",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_264": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00px",
                            "right": "1px",
                            "top": "0px",
                            "bottom": 0,
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_231",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_265": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "34.1793%",
                            "top": "0.00%",
                            "bottom": "-11.2760%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_232",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_266": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.2523%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_233",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_267": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.0372%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_235",
                    }
                },
                "order": ["0"],
            },
            "layout_268": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 6, "height": "200px"},
                        "widgetId": "widget_220",
                        "isPending": True,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 3, "width": 6, "height": "200px"},
                        "widgetId": "widget_862",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_269": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 6, "height": "240px"},
                        "widgetId": "widget_234",
                        "isPending": True,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 3, "width": 6, "height": "240px"},
                        "widgetId": "widget_863",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_290": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 40, 0, 40], "unit": "px"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "138px",
                            "right": "498px",
                            "top": "87px",
                            "bottom": "157px",
                            "width": "360px",
                            "height": "100px",
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetY": 20,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "67px",
                            "right": "567px",
                            "top": "26px",
                            "bottom": "526px",
                            "width": "500px",
                            "height": "700px",
                        },
                        "widgetId": "widget_402",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "339px",
                            "right": "509px",
                            "top": "-27px",
                            "bottom": "33px",
                            "width": "170px",
                            "height": 80,
                        },
                        "widgetId": "widget_403",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                    },
                },
                "order": ["0", "2", "1"],
            },
            "layout_291": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 50},
                        "widgetId": "widget_237",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "offsetY": 20,
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_292": {
                "type": "ROW",
                "content": {
                    "0": {
                        "sectionId": "section_6",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "600px"},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_293": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "27.2496%",
                            "top": "0.00%",
                            "bottom": "29.3763%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_240",
                        "setting": {"lockParent": True},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 11,
                            "right": "44.9525%",
                            "top": 218,
                            "bottom": "1.5703%",
                            "width": 263,
                            "height": 371,
                        },
                        "widgetId": "widget_239",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": True,
                            "autoProps": {"width": False, "height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_294": {"type": "FIXED"},
            "layout_295": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "29.2%",
                            "right": "-1px",
                            "top": "0px",
                            "bottom": "0.0%",
                            "width": "71.0%",
                            "height": "100.0%",
                        },
                        "widgetId": "widget_242",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "top": True,
                                "bottom": False,
                            },
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": True,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": True,
                                "bottom": False,
                            },
                        },
                        "bbox": {
                            "left": "0px",
                            "right": "70.1%",
                            "top": "0px",
                            "bottom": "0px",
                            "width": "29.9%",
                            "height": "100.0%",
                        },
                        "widgetId": "widget_243",
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_296": {"type": "FIXED"},
            "layout_297": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 0, "width": 12, "height": 60},
                        "widgetId": "widget_241",
                    }
                },
                "order": ["0"],
            },
            "layout_308": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "229px",
                            "right": "589px",
                            "top": "62px",
                            "bottom": "132px",
                            "width": "360px",
                            "height": "100px",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetY": 20,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-35px",
                            "right": "265px",
                            "top": "164px",
                            "bottom": "214px",
                            "width": "300px",
                            "height": "3400px",
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_309": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_418",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_310": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 1, "height": "3300px"},
                        "type": "WIDGET",
                        "widgetId": "widget_419",
                        "setting": {
                            "autoProps": {"height": False},
                            "autoHeight": True,
                            "heightMode": "fixed",
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 1, "width": 11, "height": 600},
                        "widgetId": "widget_411",
                    },
                },
                "order": ["2", "3"],
            },
            "layout_311": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_248",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "280px"},
                        "widgetId": "widget_412",
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_254",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_413",
                        "isPending": False,
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_259",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_414",
                        "isPending": False,
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_264",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_415",
                        "isPending": False,
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "8": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_269",
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_416",
                        "isPending": False,
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "10": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_273",
                        "isPending": False,
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_417",
                        "isPending": False,
                        "setting": {
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11"],
            },
            "layout_312": {
                "type": "ROW",
                "content": {
                    "3": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_420",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["3"],
            },
            "layout_313": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_421",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_314": {
                "type": "ROW",
                "content": {
                    "3": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_422",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["3"],
            },
            "layout_315": {
                "type": "ROW",
                "content": {
                    "3": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_423",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["3"],
            },
            "layout_316": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_424",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_317": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_425",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_318": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "50px"},
                        "widgetId": "widget_245",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_319": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "20px", "height": "3260px"},
                        "widgetId": "widget_247",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False, "width": False},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_320": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_251",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_252",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_253",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_321": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_256",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_258",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_322": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_261",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_263",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_262",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_323": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_266",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_267",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_268",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_324": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_271",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_325": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "180px"},
                        "widgetId": "widget_275",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 180},
                        "widgetId": "widget_276",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_326": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "229px",
                            "right": "589px",
                            "top": "62px",
                            "bottom": "132px",
                            "width": "360px",
                            "height": "80px",
                        },
                        "widgetId": "widget_426",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetY": 20,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-35px",
                            "right": "265px",
                            "top": "164px",
                            "bottom": "214px",
                            "width": "300px",
                            "height": "2100px",
                        },
                        "widgetId": "widget_427",
                        "setting": {"heightMode": "fixed", "offsetY": 20},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_327": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 50},
                        "widgetId": "widget_245",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_328": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 1, "height": "1940px"},
                        "widgetId": "widget_247",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 1, "width": 11, "height": 600},
                        "widgetId": "widget_428",
                        "setting": {"heightMode": "fit"},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_329": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_248",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "280px"},
                        "widgetId": "widget_429",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_254",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_430",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_259",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_431",
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_264",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_432",
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_269",
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_433",
                        "isPending": False,
                    },
                    "10": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "1079px", "height": "40px"},
                        "widgetId": "widget_273",
                        "isPending": False,
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {"width": "1079px", "height": "280px"},
                        "widgetId": "widget_434",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11"],
            },
            "layout_330": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "180px"},
                        "widgetId": "widget_251",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "180px"},
                        "widgetId": "widget_252",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 5, "height": "180px"},
                        "widgetId": "widget_253",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_331": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 180},
                        "widgetId": "widget_256",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 7, "height": 180},
                        "widgetId": "widget_258",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_332": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 2, "height": 180},
                        "widgetId": "widget_261",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 2, "height": 180},
                        "widgetId": "widget_263",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 5, "height": 180},
                        "widgetId": "widget_262",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_333": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "180px"},
                        "widgetId": "widget_266",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "180px"},
                        "widgetId": "widget_267",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 5, "height": "180px"},
                        "widgetId": "widget_268",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_334": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 5, "height": "180px"},
                        "widgetId": "widget_271",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 6, "height": 180},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_335": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "180px"},
                        "widgetId": "widget_275",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 5, "height": 180},
                        "widgetId": "widget_276",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_342": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_440",
                        "setting": {"heightMode": "fit"},
                    }
                },
            },
            "layout_343": {
                "type": "FIXED",
                "order": ["1", "0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "51.5763%",
                            "top": "87.9151%",
                            "bottom": "10.00px",
                            "width": "15.1778%",
                            "height": "9.3280%",
                        },
                        "widgetId": "widget_441",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "1": {
                        "sectionId": "section_7",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "3.1819%",
                            "top": "50%",
                            "bottom": "0.2506%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                },
            },
            "layout_344": {
                "type": "FIXED",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_442",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_443",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_444",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
            },
            "layout_379": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_469",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_470",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_471",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_381": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_475",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_476",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_477",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_440": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "138px",
                            "right": "498px",
                            "top": "87px",
                            "bottom": "157px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_526",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetY": 20,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "67px",
                            "right": "567px",
                            "top": "26px",
                            "bottom": "526px",
                            "width": "500px",
                            "height": "630px",
                        },
                        "widgetId": "widget_527",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 20,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_441": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_528",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_442": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_529",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_443": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": True,
                                "bottom": False,
                            },
                        },
                        "bbox": {
                            "left": "0.0%",
                            "right": "13.9%",
                            "top": "0.0%",
                            "bottom": "0.0%",
                            "width": "100.00%",
                            "height": "100.0%",
                        },
                        "widgetId": "widget_911",
                    }
                },
                "order": ["0"],
            },
            "layout_444": {"type": "FIXED"},
            "layout_445": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "bbox": {
                            "left": "0.0%",
                            "right": "13.9%",
                            "top": "0.0%",
                            "bottom": "0.4%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_909",
                    }
                },
                "order": ["0"],
            },
            "layout_446": {"type": "FIXED"},
            "layout_447": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 50},
                        "widgetId": "widget_237",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "offsetY": 20,
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "60px"},
                        "widgetId": "widget_241",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_448": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "sectionId": "section_6",
                        "type": "SECTION",
                        "bbox": {"width": "auto", "height": 500},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_466": {
                "type": "ROW",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "391px",
                        },
                        "widgetId": "widget_544",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 3,
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "391px",
                        },
                        "widgetId": "widget_571",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 6,
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "391px",
                        },
                        "widgetId": "widget_577",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 9,
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "391px",
                        },
                        "widgetId": "widget_583",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                        },
                    },
                },
            },
            "layout_474": {
                "type": "FIXED",
                "content": {
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": 0,
                            "bottom": "50.0000%",
                            "width": "100.00%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_570",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "top": False,
                                "bottom": True,
                                "left": False,
                                "right": True,
                            },
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_904",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_905",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "lockLayout": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["6", "7", "8"],
            },
            "layout_475": {
                "type": "FIXED",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_569",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "4.18%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_568",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["2", "1"],
            },
            "layout_476": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_572",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_573",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_574",
                        "setting": {
                            "lockParent": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "top": False,
                                "bottom": True,
                                "left": False,
                                "right": True,
                            },
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_477": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.18%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_575",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_576",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_478": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_578",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_579",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_580",
                        "setting": {
                            "lockParent": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "top": False,
                                "bottom": True,
                                "left": False,
                                "right": True,
                            },
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_479": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_581",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_582",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_480": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_584",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_585",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_586",
                        "setting": {
                            "lockParent": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "top": False,
                                "bottom": True,
                                "left": False,
                                "right": True,
                            },
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_481": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_587",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_588",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_482": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-263px",
                            "right": "1017px",
                            "top": "80px",
                            "bottom": "180px",
                            "width": "1280px",
                            "height": "160px",
                        },
                        "widgetId": "widget_590",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "220px",
                            "right": "580px",
                            "top": "316px",
                            "bottom": "386px",
                            "width": "360px",
                            "height": "80px",
                        },
                        "widgetId": "widget_591",
                        "setting": {"heightMode": "auto", "offsetX": 0, "offsetY": 0},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "314px",
                            "right": "614px",
                            "top": "476px",
                            "bottom": "776px",
                            "width": "300px",
                            "height": "700px",
                        },
                        "widgetId": "widget_592",
                        "setting": {"heightMode": "auto", "offsetY": 0, "offsetX": 0},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "941px",
                            "bottom": "1041px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_593",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-121px",
                            "right": "179px",
                            "top": "-191px",
                            "bottom": "200px",
                            "width": "300px",
                            "height": "320px",
                        },
                        "widgetId": "widget_594",
                        "setting": {
                            "heightMode": "auto",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": "BOTTOM"},
                                }
                            },
                        },
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "1337px",
                            "bottom": "1437px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_595",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35px",
                            "right": "535px",
                            "top": "1229px",
                            "bottom": "1729px",
                            "width": "500px",
                            "height": "430px",
                        },
                        "widgetId": "widget_596",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "1833px",
                            "bottom": "1888px",
                            "width": "1200px",
                            "height": "60px",
                        },
                        "widgetId": "widget_597",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "42px",
                            "right": "1242px",
                            "top": "1940px",
                            "bottom": "2040px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_598",
                        "setting": {
                            "heightMode": "fixed",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "151px",
                            "right": "551px",
                            "top": "1960px",
                            "bottom": "2260px",
                            "width": "400px",
                            "height": "450px",
                        },
                        "widgetId": "widget_599",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-91px",
                            "right": "269px",
                            "top": "2369px",
                            "bottom": "2439px",
                            "width": "360px",
                            "height": 140,
                        },
                        "widgetId": "widget_600",
                        "setting": {"heightMode": "auto", "offsetY": 0},
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "2561px",
                            "bottom": "2651px",
                            "width": "1200px",
                            "height": 118,
                        },
                        "widgetId": "widget_601",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                        "isPending": False,
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-177px",
                            "right": "223px",
                            "top": "-199px",
                            "bottom": "201px",
                            "width": "400px",
                            "height": "480px",
                        },
                        "widgetId": "widget_602",
                        "setting": {"heightMode": "fixed"},
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "3400px",
                            "bottom": "3640px",
                            "width": "1200px",
                            "height": "240px",
                        },
                        "widgetId": "widget_604",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                        "isPending": False,
                    },
                    "15": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "86px",
                            "right": "486px",
                            "top": "3260px",
                            "bottom": "3660px",
                            "width": "400px",
                            "height": "500px",
                        },
                        "widgetId": "widget_606",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-18px",
                            "right": "1182px",
                            "top": "4200px",
                            "bottom": "4440px",
                            "width": "1200px",
                            "height": 198,
                        },
                        "widgetId": "widget_607",
                        "setting": {"heightMode": "fixed", "offsetY": 0},
                        "isPending": False,
                    },
                    "17": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_609",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    },
                    "18": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_610",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    },
                    "19": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "4347px",
                            "bottom": "4447px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_611",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                        },
                        "isPending": False,
                    },
                    "21": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "27px",
                            "right": "116px",
                            "top": "-50px",
                            "bottom": "38px",
                            "width": "88px",
                            "height": "320px",
                        },
                        "widgetId": "widget_613",
                        "setting": {"heightMode": "auto"},
                    },
                },
                "order": [
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "21",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                    "16",
                    "17",
                    "18",
                    "19",
                ],
            },
            "layout_484": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_17",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_485": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_19",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_486": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 300},
                        "widgetId": "widget_21",
                        "setting": {
                            "heightMode": "fit",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_487": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_74",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_488": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "391px"},
                        "widgetId": "widget_544",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "391px"},
                        "widgetId": "widget_571",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"width": False, "height": False},
                            "aspectRatio": "100:120",
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_489": {
                "type": "FIXED",
                "content": {
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": "49.9590%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_570",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_906",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_907",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["2", "3", "4"],
            },
            "layout_490": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_568",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_569",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_491": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_572",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_573",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": 0,
                            "bottom": "49.9590%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_574",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "top": False,
                                "bottom": True,
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                            "vCenter": False,
                            "hCenter": True,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_492": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_575",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_576",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_493": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_578",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_579",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": "48.0429%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_580",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_494": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_581",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_582",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_495": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_584",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_585",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_586",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_496": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_587",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_588",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_497": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_76",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_498": {
                "type": "ROW",
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": 414},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_499": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "52.2007%",
                            "top": "50%",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_35",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1394%",
                            "right": 0,
                            "top": "22px",
                            "bottom": "76.3285%",
                            "width": "47.8606%",
                            "height": "18.3575%",
                        },
                        "widgetId": "widget_36",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "51.9911%",
                            "right": "1px",
                            "top": "18.7802%",
                            "bottom": "21.9684%",
                            "width": "47.8638%",
                            "height": "59.2514%",
                        },
                        "widgetId": "widget_37",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": 0,
                            "bottom": "20.00px",
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_38",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"top": True, "bottom": False},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_503": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "offsetX": -10,
                            "autoProps": {"height": False},
                        },
                        "bbox": {"left": 2, "width": 2, "height": 30},
                        "widgetId": "widget_53",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 10, "width": 2, "height": "60px"},
                        "widgetId": "widget_128",
                        "setting": {"heightMode": "fit"},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_504": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_78",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_505": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 300},
                        "widgetId": "widget_70",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_506": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "50px"},
                        "widgetId": "widget_83",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "flex-end"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 4, "height": "50px"},
                        "widgetId": "widget_86",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "flex-end"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_507": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 0, "width": 4, "height": 70},
                        "widgetId": "widget_84",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"left": 5, "width": 4, "height": 70},
                        "widgetId": "widget_87",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                        "bbox": {"left": 9, "width": 3, "height": "42px"},
                        "widgetId": "widget_88",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_508": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_603",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_509": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "sectionId": "section_7",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "3.1819%",
                            "top": "50%",
                            "bottom": "0.2506%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "51.5763%",
                            "top": "87.9151%",
                            "bottom": "10.00px",
                            "width": "15.1778%",
                            "height": "9.3280%",
                        },
                        "widgetId": "widget_441",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_510": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_442",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_443",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_444",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_511": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_469",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_470",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_471",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_512": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_475",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_476",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_477",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_513": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 400},
                        "widgetId": "widget_605",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_514": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {
                            "left": "1px",
                            "right": "57.3866%",
                            "top": "74px",
                            "bottom": "5.0300%",
                            "width": "42.4211%",
                            "height": "61.1872%",
                        },
                        "widgetId": "widget_121",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "49.3209%",
                            "right": 0,
                            "top": "34.1014%",
                            "bottom": 10,
                            "width": "50.8489%",
                            "height": "61.2903%",
                        },
                        "widgetId": "widget_122",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_515": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "490px"},
                        "widgetId": "widget_129",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": 400},
                        "widgetId": "widget_124",
                        "setting": {"heightMode": "fit"},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_516": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "9px",
                            "top": "10px",
                            "width": "29.8968%",
                            "height": "86.5369%",
                            "right": "67.07189999999997%",
                            "bottom": "8px",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35.0191%",
                            "width": 238,
                            "height": "43.6104%",
                            "right": "3.1237%",
                            "top": 10,
                            "bottom": 74,
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "274px",
                            "top": "69.6731999999997%",
                            "width": "27.8170%",
                            "height": "23.8722%",
                            "right": "2.2273999999999994%",
                            "bottom": "6.163100000000027%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_517": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.58%",
                            "top": "50%",
                            "width": "13.39%",
                            "height": "55.81%",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "21.20%",
                            "width": "226px",
                            "height": "35%",
                            "right": "39.05%",
                            "top": "50%",
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "230px",
                            "top": "50%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.25%",
                            "bottom": "20.93%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_518": {"type": "FIXED"},
            "layout_519": {"type": "FIXED"},
            "layout_520": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": 400},
                        "widgetId": "widget_608",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_521": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "top": False,
                                "bottom": True,
                                "left": False,
                                "right": True,
                            },
                        },
                        "bbox": {
                            "left": "50%",
                            "right": "3.2039%",
                            "top": 0,
                            "bottom": "71.6895%",
                            "width": "100.00%",
                            "height": "22.8311%",
                        },
                        "widgetId": "widget_154",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": "34.1014%",
                            "bottom": 10,
                            "width": "100.00%",
                            "height": "61.2903%",
                        },
                        "widgetId": "widget_155",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_522": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 89},
                        "widgetId": "widget_141",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": 89},
                        "widgetId": "widget_142",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": 89},
                        "widgetId": "widget_143",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": 89},
                        "widgetId": "widget_144",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_523": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": 89},
                        "widgetId": "widget_146",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": 89},
                        "widgetId": "widget_147",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": 89},
                        "widgetId": "widget_148",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": 89},
                        "widgetId": "widget_149",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": False},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_524": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_151",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_526": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:120",
                        },
                        "bbox": {"left": 0, "width": 6, "height": "89px"},
                        "widgetId": "widget_577",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:120",
                        },
                        "bbox": {"left": 6, "width": 6, "height": "89px"},
                        "widgetId": "widget_583",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_527": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-263px",
                            "right": "1017px",
                            "top": "80px",
                            "bottom": "180px",
                            "width": "1280px",
                            "height": "160px",
                        },
                        "widgetId": "widget_615",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "220px",
                            "right": "580px",
                            "top": "316px",
                            "bottom": "386px",
                            "width": "360px",
                            "height": "80px",
                        },
                        "widgetId": "widget_616",
                        "setting": {
                            "heightMode": "auto",
                            "offsetX": 0,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "314px",
                            "right": "614px",
                            "top": "476px",
                            "bottom": "776px",
                            "width": "300px",
                            "height": "700px",
                        },
                        "widgetId": "widget_617",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "offsetX": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "941px",
                            "bottom": "1041px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_618",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-121px",
                            "right": "179px",
                            "top": "-191px",
                            "bottom": "200px",
                            "width": "300px",
                            "height": 343,
                        },
                        "widgetId": "widget_619",
                        "setting": {
                            "heightMode": "auto",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "NONE",
                                    "option": {"direction": None},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": "BOTTOM"},
                                }
                            },
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "1337px",
                            "bottom": "1437px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_620",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35px",
                            "right": "535px",
                            "top": "1229px",
                            "bottom": "1729px",
                            "width": "500px",
                            "height": "430px",
                        },
                        "widgetId": "widget_621",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "1833px",
                            "bottom": "1888px",
                            "width": "1200px",
                            "height": "60px",
                        },
                        "widgetId": "widget_622",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "42px",
                            "right": "1242px",
                            "top": "1940px",
                            "bottom": "2040px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_623",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "151px",
                            "right": "551px",
                            "top": "1960px",
                            "bottom": "2260px",
                            "width": "400px",
                            "height": "450px",
                        },
                        "widgetId": "widget_624",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-91px",
                            "right": "269px",
                            "top": "2369px",
                            "bottom": "2439px",
                            "width": "360px",
                            "height": 89,
                        },
                        "widgetId": "widget_625",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "2561px",
                            "bottom": "2651px",
                            "width": "1200px",
                            "height": 118,
                        },
                        "widgetId": "widget_626",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-177px",
                            "right": "223px",
                            "top": "-199px",
                            "bottom": "201px",
                            "width": "400px",
                            "height": "480px",
                        },
                        "widgetId": "widget_627",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "18px",
                            "right": "342px",
                            "top": "3883px",
                            "bottom": "4199px",
                            "width": "324px",
                            "height": "316px",
                        },
                        "widgetId": "widget_629",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "15": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "86px",
                            "right": "486px",
                            "top": "3260px",
                            "bottom": "3660px",
                            "width": "400px",
                            "height": "500px",
                        },
                        "widgetId": "widget_631",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-18px",
                            "right": "1182px",
                            "top": "4200px",
                            "bottom": "4440px",
                            "width": "1200px",
                            "height": "240px",
                        },
                        "widgetId": "widget_632",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "17": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_634",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "18": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10px",
                            "right": "310px",
                            "top": "3878px",
                            "bottom": "4178px",
                            "width": "300px",
                            "height": "160px",
                        },
                        "widgetId": "widget_635",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "19": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "43px",
                            "right": "1243px",
                            "top": "4347px",
                            "bottom": "4447px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_636",
                        "setting": {
                            "heightMode": "auto",
                            "width": "100%",
                            "offsetX": None,
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                },
                "order": [
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                    "16",
                    "17",
                    "18",
                    "19",
                ],
            },
            "layout_529": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_639",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_530": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_640",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_531": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_641",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_532": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_642",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_533": {
                "type": "ROW",
                "content": {
                    "4": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_643",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["4"],
            },
            "layout_534": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_568",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_569",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_570",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_535": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_568",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_569",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_536": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_572",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_573",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_574",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_537": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_575",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_576",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_538": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_578",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_579",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_580",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_539": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_581",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_582",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_540": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 19,
                            "right": 19,
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "87.1622%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_584",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.4189%",
                            "right": "3.7162%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "89.8649%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_585",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_586",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"top": False, "bottom": True},
                            "vCenter": False,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_541": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "89.9281%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_587",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "89.9281%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_588",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_542": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_644",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_543": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_645",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_544": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "52.2007%",
                            "top": "50%",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_35",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "height": "CUSTOM",
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.8125%",
                            "right": 0,
                            "top": 0,
                            "bottom": "73.3010%",
                            "width": "48.0000%",
                            "height": "26.6990%",
                        },
                        "widgetId": "widget_36",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.6458%",
                            "right": 0,
                            "top": "29.7747%",
                            "bottom": "19.9939%",
                            "width": "48.0000%",
                            "height": "50.2427%",
                        },
                        "widgetId": "widget_37",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0126%",
                            "right": "7.5654%",
                            "top": "87.2434%",
                            "bottom": "10.00px",
                            "width": "40.4219%",
                            "height": "10.1449%",
                        },
                        "widgetId": "widget_38",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"bottom": False, "top": True},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_548": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_646",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_549": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_647",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_550": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_648",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_551": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_649",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_552": {
                "type": "ROW",
                "content": {
                    "3": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_650",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["3"],
            },
            "layout_553": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_651",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_554": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "sectionId": "section_7",
                        "type": "SECTION",
                        "bbox": {
                            "left": "50%",
                            "right": "3.1819%",
                            "top": "50%",
                            "bottom": "0.2506%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "51.5763%",
                            "top": "87.9151%",
                            "bottom": "10.00px",
                            "width": "15.1778%",
                            "height": "9.3280%",
                        },
                        "widgetId": "widget_441",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_555": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_442",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_443",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_444",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "AUTO",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_556": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_469",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_470",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_471",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_557": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "56.3587%",
                            "top": "50%",
                            "bottom": "8.7629%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_475",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "CUSTOM",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "44.8762%",
                            "top": "15.00%",
                            "bottom": "46.1689%",
                            "width": 128,
                            "height": 128,
                        },
                        "widgetId": "widget_476",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "33.9539%",
                            "top": "58.9873%",
                            "bottom": "23.2911%",
                            "width": "31.9149%",
                            "height": "17.7215%",
                        },
                        "widgetId": "widget_477",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_558": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_652",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_559": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                        "bbox": {
                            "left": "1px",
                            "right": 0,
                            "top": "1px",
                            "bottom": "35.6364%",
                            "width": "99.6689%",
                            "height": "64.0000%",
                        },
                        "widgetId": "widget_121",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "38.2026%",
                            "bottom": 36,
                            "width": "100.0000%",
                            "height": "48.7179%",
                        },
                        "widgetId": "widget_122",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO", "width": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_560": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_653",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_561": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "9px",
                            "top": "10px",
                            "width": "39.0599%",
                            "height": "86.5343%",
                            "right": "57.9401%",
                            "bottom": "10px",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "45.8711%",
                            "width": "153px",
                            "height": "43.6103%",
                            "right": "3.1198%",
                            "top": "10px",
                            "bottom": "75px",
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "274px",
                            "top": "69.6731999999997%",
                            "width": "27.8170%",
                            "height": "23.8722%",
                            "right": "2.2273999999999994%",
                            "bottom": "6.163100000000027%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_562": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.58%",
                            "top": "50%",
                            "width": "13.39%",
                            "height": "55.81%",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_133",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "21.20%",
                            "width": "226px",
                            "height": "35%",
                            "right": "39.05%",
                            "top": "50%",
                        },
                        "widgetId": "widget_134",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": False,
                        },
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "230px",
                            "top": "50%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.25%",
                            "bottom": "20.93%",
                        },
                        "widgetId": "widget_135",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_563": {"type": "FIXED"},
            "layout_564": {"type": "FIXED"},
            "layout_565": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_654",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_566": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO", "width": "CUSTOM"},
                        },
                        "bbox": {
                            "left": 0,
                            "right": "0.0000%",
                            "top": 0,
                            "bottom": "76.8844%",
                            "width": "100.0000%",
                            "height": "23.1156%",
                        },
                        "widgetId": "widget_154",
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": 0,
                            "top": "34.1014%",
                            "bottom": 10,
                            "width": "100.00%",
                            "height": "61.2903%",
                        },
                        "widgetId": "widget_155",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_567": {
                "type": "ROW",
                "content": {
                    "4": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_655",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["4"],
            },
            "layout_568": {
                "type": "ROW",
                "content": {
                    "4": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_656",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["4"],
            },
            "layout_569": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_657",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_572": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "50px"},
                        "widgetId": "widget_17",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_573": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_19",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_574": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "296px", "height": 237},
                        "widgetId": "widget_21",
                        "setting": {
                            "heightMode": "fixed",
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_575": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_74",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_576": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_544",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:120",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_571",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:120",
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_577",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:120",
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_583",
                        "setting": {
                            "heightMode": "ratio",
                            "autoProps": {"height": "CUSTOM"},
                            "aspectRatio": "100:120",
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_577": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "50px"},
                        "widgetId": "widget_76",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_578": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "sectionId": "section_1",
                        "type": "SECTION",
                        "bbox": {"width": "auto", "height": 414},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_579": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "offsetX": -10,
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"width": "auto", "height": 30},
                        "widgetId": "widget_53",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 60},
                        "widgetId": "widget_128",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "isPending": True,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_580": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "50px"},
                        "widgetId": "widget_78",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_581": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 300},
                        "widgetId": "widget_70",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_582": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 50},
                        "widgetId": "widget_83",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "flex-end"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "302px", "height": "58px"},
                        "widgetId": "widget_84",
                    },
                },
                "order": ["0", "2"],
            },
            "layout_583": {
                "type": "COLUMN",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "bbox": {"width": "auto", "height": "58px"},
                        "widgetId": "widget_87",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "bbox": {"width": "auto", "height": 42},
                        "widgetId": "widget_88",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "302px", "height": "50px"},
                        "widgetId": "widget_86",
                    },
                },
                "order": ["3", "1", "2"],
            },
            "layout_584": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "400px"},
                        "widgetId": "widget_628",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_585": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "296px", "height": 275},
                        "widgetId": "widget_630",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_586": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "490px"},
                        "widgetId": "widget_129",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 400},
                        "widgetId": "widget_124",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_587": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "296px", "height": 255},
                        "widgetId": "widget_633",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_588": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_141",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_142",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_143",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_144",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_589": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_146",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_147",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_148",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 89},
                        "widgetId": "widget_149",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_590": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_151",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 20,
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_592": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 8, "height": "400px"},
                        "widgetId": "widget_660",
                        "setting": {"heightMode": "fit"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": "8px", "width": "4px", "height": "210px"},
                        "widgetId": "widget_874",
                        "setting": {"heightMode": "fit"},
                    },
                },
            },
            "layout_593": {"type": "FIXED"},
            "layout_670": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_838",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1506%",
                            "right": "17.3743%",
                            "top": 21,
                            "bottom": "86.2333%",
                            "width": "30.4523%",
                            "height": "8.7379%",
                        },
                        "widgetId": "widget_839",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0295%",
                            "right": 0,
                            "top": "21.0255%",
                            "bottom": "25.3337%",
                            "width": "48.1422%",
                            "height": "53.6408%",
                        },
                        "widgetId": "widget_840",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": 9,
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_841",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_671": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_842",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1506%",
                            "right": "17.3743%",
                            "top": 21,
                            "bottom": "86.2333%",
                            "width": "30.4523%",
                            "height": "8.7379%",
                        },
                        "widgetId": "widget_843",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0102%",
                            "right": "1px",
                            "top": "25.9473%",
                            "bottom": "17.6876%",
                            "width": "47.7975%",
                            "height": "56.3651%",
                        },
                        "widgetId": "widget_844",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": 9,
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_845",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_672": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_846",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.8125%",
                            "right": 0,
                            "top": 0,
                            "bottom": "73.3010%",
                            "width": "48.0000%",
                            "height": "26.6990%",
                        },
                        "widgetId": "widget_847",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.6458%",
                            "right": 0,
                            "top": "29.7747%",
                            "bottom": "19.9939%",
                            "width": "48.0000%",
                            "height": "50.2427%",
                        },
                        "widgetId": "widget_848",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0126%",
                            "right": "7.5654%",
                            "top": "87.2434%",
                            "bottom": "11px",
                            "width": "40.4219%",
                            "height": "10.1449%",
                        },
                        "widgetId": "widget_849",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_673": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_850",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1506%",
                            "right": "17.3743%",
                            "top": 21,
                            "bottom": "86.2333%",
                            "width": "30.4523%",
                            "height": "8.7379%",
                        },
                        "widgetId": "widget_851",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0295%",
                            "right": 0,
                            "top": "21.0255%",
                            "bottom": "25.3337%",
                            "width": "48.1422%",
                            "height": "53.6408%",
                        },
                        "widgetId": "widget_852",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": 9,
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_853",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_674": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_854",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.1506%",
                            "right": "17.3743%",
                            "top": 21,
                            "bottom": "86.2333%",
                            "width": "30.4523%",
                            "height": "8.7379%",
                        },
                        "widgetId": "widget_855",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0102%",
                            "right": "1px",
                            "top": "25.9473%",
                            "bottom": "17.6876%",
                            "width": "47.7975%",
                            "height": "56.3651%",
                        },
                        "widgetId": "widget_856",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0267%",
                            "right": "29.2989%",
                            "top": "87.6562%",
                            "bottom": 9,
                            "width": "18.6757%",
                            "height": "10.1942%",
                        },
                        "widgetId": "widget_857",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_675": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": "52.2007%",
                            "top": "2px",
                            "bottom": "10px",
                            "width": "47.6298%",
                            "height": "97.1354%",
                        },
                        "widgetId": "widget_858",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.8125%",
                            "right": 0,
                            "top": 0,
                            "bottom": "73.3010%",
                            "width": "48.0000%",
                            "height": "26.6990%",
                        },
                        "widgetId": "widget_859",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.6458%",
                            "right": 0,
                            "top": "29.7747%",
                            "bottom": "19.9939%",
                            "width": "48.0000%",
                            "height": "50.2427%",
                        },
                        "widgetId": "widget_860",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "52.0126%",
                            "right": "7.5654%",
                            "top": "87.2434%",
                            "bottom": "11px",
                            "width": "40.4219%",
                            "height": "10.1449%",
                        },
                        "widgetId": "widget_861",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_676": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "97px",
                            "right": "457px",
                            "top": "55px",
                            "bottom": "125px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_864",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 20,
                            "width": "100%",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20px",
                            "right": "520px",
                            "top": "-38px",
                            "bottom": "462px",
                            "width": "500px",
                            "height": "360px",
                        },
                        "widgetId": "widget_865",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 20,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "41px",
                            "right": "1241px",
                            "top": "643px",
                            "bottom": "803px",
                            "width": "1200px",
                            "height": "160px",
                        },
                        "widgetId": "widget_866",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 20,
                            "width": "100%",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "40px",
                            "right": "1240px",
                            "top": "820px",
                            "bottom": "1180px",
                            "width": "1200px",
                            "height": "500px",
                        },
                        "widgetId": "widget_867",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 20,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_677": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_868",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_678": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_869",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_679": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "lockParent": True},
                        "bbox": {
                            "left": "0px",
                            "right": 0,
                            "top": "1px",
                            "bottom": "0.2950%",
                            "width": "100.0000%",
                            "height": "99.4100%",
                        },
                        "widgetId": "widget_222",
                    }
                },
                "order": ["0"],
            },
            "layout_680": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.1057%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_223",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_681": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00px",
                            "right": "1px",
                            "top": "0px",
                            "bottom": 0,
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_221",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_682": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "34.1793%",
                            "top": "0.00%",
                            "bottom": "-11.2760%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_224",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_683": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.2523%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_225",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_684": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_870",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_685": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_871",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_686": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "lockParent": True},
                        "bbox": {
                            "left": "0px",
                            "right": 0,
                            "top": "1px",
                            "bottom": "0.2950%",
                            "width": "100.0000%",
                            "height": "99.4100%",
                        },
                        "widgetId": "widget_229",
                    }
                },
                "order": ["0"],
            },
            "layout_687": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.1057%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_230",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_688": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00px",
                            "right": "1px",
                            "top": "0px",
                            "bottom": 0,
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_231",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_689": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "34.1793%",
                            "top": "0.00%",
                            "bottom": "-11.2760%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_232",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_690": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.2523%",
                            "top": "0.00%",
                            "bottom": "6.8249%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_233",
                        "setting": {"lockParent": True},
                    }
                },
                "order": ["0"],
            },
            "layout_691": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "48.0372%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_235",
                    }
                },
                "order": ["0"],
            },
            "layout_692": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 50},
                        "widgetId": "widget_218",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO", "width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_693": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "sectionId": "section_4",
                        "type": "SECTION",
                        "bbox": {"width": "auto", "height": 360},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_220",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetX": 0,
                            "offsetY": 0,
                        },
                        "isPending": True,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_872",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetX": 0,
                            "offsetY": 0,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_694": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 50},
                        "widgetId": "widget_227",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_695": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "sectionId": "section_5",
                        "type": "SECTION",
                        "bbox": {"width": "auto", "height": 360},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_234",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetX": 0,
                            "offsetY": 0,
                        },
                        "isPending": True,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_873",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetX": 0,
                            "offsetY": 0,
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_696": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "294px",
                            "right": "654px",
                            "top": "-32px",
                            "bottom": "38px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_875",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "30px",
                            "right": "570px",
                            "top": "121px",
                            "bottom": "474px",
                            "width": "540px",
                            "height": "353px",
                        },
                        "widgetId": "widget_876",
                        "setting": {"heightMode": "auto", "offsetY": 0},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "576px",
                            "right": "876px",
                            "top": "-293px",
                            "bottom": "307px",
                            "width": "300px",
                            "height": "340px",
                        },
                        "widgetId": "widget_879",
                        "setting": {"heightMode": "fixed", "offsetY": 60},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "64px",
                            "right": "1216px",
                            "top": "921px",
                            "bottom": "1081px",
                            "width": "1152px",
                            "height": "160px",
                        },
                        "widgetId": "widget_881",
                        "setting": {
                            "heightMode": "fixed",
                            "offsetY": 0,
                            "width": "100%",
                        },
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-56px",
                            "right": "344px",
                            "top": "904px",
                            "bottom": "1304px",
                            "width": "400px",
                            "height": 437,
                        },
                        "widgetId": "widget_882",
                        "setting": {"heightMode": "fixed"},
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "171px",
                            "right": "347px",
                            "top": "502px",
                            "bottom": "636px",
                            "width": "177px",
                            "height": 154,
                        },
                        "widgetId": "widget_883",
                        "setting": {"heightMode": "auto"},
                    },
                },
                "order": ["0", "1", "5", "2", "3", "4"],
            },
            "layout_697": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_182",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_698": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600px"},
                        "widgetId": "widget_877",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_699": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "595px", "height": "100px"},
                        "widgetId": "widget_185",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "flex-start"},
                            "autoProps": {"width": False},
                        },
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_187",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": False}},
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_188",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": False}},
                        "bbox": {"width": 583, "height": "80px"},
                        "widgetId": "widget_189",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_700": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": 40},
                        "widgetId": "widget_191",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "10px"},
                        "widgetId": "widget_192",
                        "setting": {"autoProps": {"width": False}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_193",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_194",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_195",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_197",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_198",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_196",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_199",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_200",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
            },
            "layout_701": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600px"},
                        "widgetId": "widget_880",
                        "setting": {"heightMode": "fit"},
                    }
                },
                "order": ["0"],
            },
            "layout_702": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_203",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_204",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_205",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_206",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_207",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_208",
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_209",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_210",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7"],
            },
            "layout_703": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "offsetY": 20,
                        },
                        "bbox": {"left": 0, "width": 12, "height": "50px"},
                        "widgetId": "widget_211",
                    }
                },
                "order": ["0"],
            },
            "layout_704": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "400px"},
                        "widgetId": "widget_660",
                        "setting": {"heightMode": "fit"},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "210px"},
                        "widgetId": "widget_874",
                        "setting": {"heightMode": "fit"},
                    },
                },
                "order": ["0", "1"],
            },
            "layout_705": {"type": "FIXED"},
            "layout_706": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                            "maintainedByLayout": True,
                        },
                        "bbox": {"left": 0, "width": 12, "height": "554px"},
                        "widgetId": "widget_884",
                    }
                },
                "order": ["0"],
            },
            "layout_707": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {
                            "left": "83px",
                            "right": "304px",
                            "top": "-68px",
                            "bottom": "332px",
                            "width": "221px",
                            "height": "400px",
                        },
                        "widgetId": "widget_878",
                    },
                    "1": {
                        "widgetId": "widget_183",
                        "type": "WIDGET",
                        "bbox": {
                            "left": "83px",
                            "right": "304px",
                            "top": "-68px",
                            "bottom": "332px",
                            "width": "221px",
                            "height": "400px",
                        },
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["1", "0"],
            },
            "layout_708": {
                "type": "FLOW",
                "setting": {
                    "style": {"padding": {"number": [0, 5, 0, 5], "unit": "%"}}
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "294px",
                            "right": "654px",
                            "top": "-32px",
                            "bottom": "38px",
                            "width": "360px",
                            "height": "160px",
                        },
                        "widgetId": "widget_885",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "width": "100%",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "102px",
                            "right": "462px",
                            "top": "-34px",
                            "bottom": "36px",
                            "width": "360px",
                            "height": "420px",
                        },
                        "widgetId": "widget_886",
                        "setting": {
                            "heightMode": "auto",
                            "offsetY": 0,
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-56px",
                            "right": "344px",
                            "top": "904px",
                            "bottom": "1304px",
                            "width": "400px",
                            "height": 437,
                        },
                        "widgetId": "widget_892",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "18px",
                            "right": "342px",
                            "top": "947px",
                            "bottom": "1262px",
                            "width": "324px",
                            "height": "315px",
                        },
                        "widgetId": "widget_899",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "18px",
                            "right": "342px",
                            "top": "1264px",
                            "bottom": "1332px",
                            "width": "324px",
                            "height": "68px",
                        },
                        "widgetId": "widget_900",
                        "setting": {"heightMode": "auto", "width": "100%"},
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "18px",
                            "right": "342px",
                            "top": "1572px",
                            "bottom": "1992px",
                            "width": "324px",
                            "height": "420px",
                        },
                        "widgetId": "widget_901",
                        "setting": {"heightMode": "fixed"},
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "6", "7", "8", "4"],
            },
            "layout_709": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_893",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_710": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_894",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_711": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "595px", "height": "100px"},
                        "widgetId": "widget_185",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "flex-start"},
                            "autoProps": {"width": False},
                        },
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_187",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": False}},
                        "bbox": {"width": 583, "height": "60px"},
                        "widgetId": "widget_188",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "autoProps": {"width": False}},
                        "bbox": {"width": 583, "height": "80px"},
                        "widgetId": "widget_189",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_712": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": 40},
                        "widgetId": "widget_191",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "10px"},
                        "widgetId": "widget_192",
                        "setting": {"autoProps": {"width": False}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_193",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_194",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_195",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_197",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_198",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_196",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_199",
                        "setting": {
                            "autoProps": {"width": False},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_200",
                        "setting": {"autoProps": {"width": False}},
                        "isPending": False,
                    },
                    "10": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "300px", "height": "92px"},
                        "widgetId": "widget_183",
                    },
                },
                "order": ["10", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
            },
            "layout_714": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "40px"},
                        "widgetId": "widget_203",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "488px", "height": "16px"},
                        "widgetId": "widget_204",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "296px", "height": "40px"},
                        "widgetId": "widget_205",
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_206",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_207",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_208",
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "40px"},
                        "widgetId": "widget_209",
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "484px", "height": "16px"},
                        "widgetId": "widget_210",
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7"],
            },
            "layout_716": {
                "type": "ROW",
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_897",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_717": {"type": "FIXED"},
            "layout_718": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto"},
                        "widgetId": "widget_182",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                            "offsetY": 20,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_719": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 600},
                        "widgetId": "widget_887",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "400.00px"},
                        "widgetId": "widget_888",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_722": {
                "type": "COLUMN",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "296px", "height": 329},
                        "widgetId": "widget_874",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_724": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 0, "width": 12, "height": "295px"},
                        "widgetId": "widget_890",
                    }
                },
                "order": ["0"],
            },
            "layout_725": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {
                            "left": "0px",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "10px",
                            "height": "48px",
                        },
                        "widgetId": "widget_211",
                    }
                },
                "order": ["0"],
            },
            "layout_726": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_660",
                    }
                },
                "order": ["0"],
            },
            "layout_727": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.8333%",
                            "right": "72.1667%",
                            "top": "18.9375%",
                            "bottom": "54.2625%",
                            "width": "24.0000%",
                            "height": "26.8000%",
                        },
                        "widgetId": "widget_163",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.8047%",
                            "right": "53.6901%",
                            "top": "18.9406%",
                            "bottom": "73.6594%",
                            "width": "11.5000%",
                            "height": "7.4000%",
                        },
                        "widgetId": "widget_164",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "56.3320%",
                            "right": "32.1628%",
                            "top": "18.9406%",
                            "bottom": "73.6594%",
                            "width": "11.5000%",
                            "height": "7.4000%",
                        },
                        "widgetId": "widget_165",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "77.8568%",
                            "right": "10.6380%",
                            "top": "18.9375%",
                            "bottom": "73.6625%",
                            "width": "11.5000%",
                            "height": "7.4000%",
                        },
                        "widgetId": "widget_166",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.8073%",
                            "right": "56.0990%",
                            "top": "29.0188%",
                            "bottom": "60.9813%",
                            "width": "9.1667%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_167",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.8073%",
                            "right": "56.0990%",
                            "top": "41.6969%",
                            "bottom": "48.3031%",
                            "width": "9.1667%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_168",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.8047%",
                            "right": "56.1016%",
                            "top": "54.3750%",
                            "bottom": "35.6250%",
                            "width": "9.1667%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_169",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.8047%",
                            "right": "56.1016%",
                            "top": "67.0469%",
                            "bottom": "22.9531%",
                            "width": "9.1667%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_170",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "56.3294%",
                            "right": "34.5768%",
                            "top": "29.0188%",
                            "bottom": "60.9813%",
                            "width": "9.1667%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_171",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "56.3281%",
                            "right": "34.5781%",
                            "top": "41.6969%",
                            "bottom": "48.3031%",
                            "width": "9.1667%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_172",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "56.3307%",
                            "right": "34.5755%",
                            "top": "54.3750%",
                            "bottom": "35.6250%",
                            "width": "9.1667%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_173",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "77.8490%",
                            "right": "13.0573%",
                            "top": "28.7188%",
                            "bottom": "61.2813%",
                            "width": "9.1667%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_174",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "77.8438%",
                            "right": "13.0625%",
                            "top": "41.9250%",
                            "bottom": "48.0750%",
                            "width": "9.1667%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_175",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "77.8516%",
                            "right": "13.0547%",
                            "top": "54.3062%",
                            "bottom": "35.6938%",
                            "width": "9.1667%",
                            "height": "10.0000%",
                        },
                        "widgetId": "widget_176",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "5.6432%",
                            "top": "78.9844%",
                            "bottom": "17.8156%",
                            "width": "90.0000%",
                            "height": "3.2000%",
                        },
                        "widgetId": "widget_177",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "15": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0000%",
                            "right": "69.2259%",
                            "top": "84.9625%",
                            "bottom": "7.8375%",
                            "width": "25.7741%",
                            "height": "7.2000%",
                        },
                        "widgetId": "widget_178",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.8047%",
                            "right": "40.3678%",
                            "top": "84.8656%",
                            "bottom": "3.5344%",
                            "width": "24.8275%",
                            "height": "11.6000%",
                        },
                        "widgetId": "widget_179",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "17": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "65.7188%",
                            "right": "7.0495%",
                            "top": "84.5687%",
                            "bottom": "3.8312%",
                            "width": "27.2318%",
                            "height": "11.6000%",
                        },
                        "widgetId": "widget_180",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "18": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "72.1693%",
                            "top": 20,
                            "bottom": "86.4961%",
                            "width": "24.0000%",
                            "height": "9.6000%",
                        },
                        "widgetId": "widget_903",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                },
                "order": [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                    "16",
                    "17",
                    "18",
                ],
            },
            "layout_728": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "4.1667%",
                            "top": "13.9185%",
                            "bottom": "72.9386%",
                            "width": "91.6667%",
                            "height": "13.1429%",
                        },
                        "widgetId": "widget_163",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "5.0000%",
                            "top": "28.8605%",
                            "bottom": "65.9967%",
                            "width": "90.0000%",
                            "height": "5.1429%",
                        },
                        "widgetId": "widget_164",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "5.0000%",
                            "top": "44.2121%",
                            "bottom": "50.6451%",
                            "width": "90.0000%",
                            "height": "5.1429%",
                        },
                        "widgetId": "widget_165",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "4.7352%",
                            "top": "59.5647%",
                            "bottom": "35.2924%",
                            "width": "90.0000%",
                            "height": "5.1429%",
                        },
                        "widgetId": "widget_166",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.3646%",
                            "right": "79.4792%",
                            "top": "35.8080%",
                            "bottom": "58.4777%",
                            "width": "15.2778%",
                            "height": 40,
                        },
                        "widgetId": "widget_167",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "28.1424%",
                            "right": "56.7014%",
                            "top": "35.8080%",
                            "bottom": "58.4777%",
                            "width": "15.2778%",
                            "height": 40,
                        },
                        "widgetId": "widget_168",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50.9180%",
                            "right": "33.9258%",
                            "top": "35.8013%",
                            "bottom": "58.4844%",
                            "width": "15.2778%",
                            "height": 40,
                        },
                        "widgetId": "widget_169",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.6936%",
                            "right": "11.1502%",
                            "top": "35.8013%",
                            "bottom": "58.4844%",
                            "width": "15.2778%",
                            "height": 40,
                        },
                        "widgetId": "widget_170",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "6.0200%",
                            "right": "78.8238%",
                            "top": "51.1540%",
                            "bottom": "43.1317%",
                            "width": "15.2778%",
                            "height": 40,
                        },
                        "widgetId": "widget_171",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "CUSTOM", "width": "AUTO"},
                        },
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "28.1424%",
                            "right": "56.7014%",
                            "top": "51.1585%",
                            "bottom": "43.1272%",
                            "width": "15.2778%",
                            "height": 40,
                        },
                        "widgetId": "widget_172",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50.2648%",
                            "right": "34.5790%",
                            "top": "51.1585%",
                            "bottom": "43.1272%",
                            "width": "15.2778%",
                            "height": 40,
                        },
                        "widgetId": "widget_173",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.3646%",
                            "right": "79.4792%",
                            "top": "67.4018%",
                            "bottom": "26.8839%",
                            "width": "15.2778%",
                            "height": 40,
                        },
                        "widgetId": "widget_174",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "28.1424%",
                            "right": "56.7014%",
                            "top": "67.3973%",
                            "bottom": "26.8884%",
                            "width": "15.2778%",
                            "height": 40,
                        },
                        "widgetId": "widget_175",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50.9201%",
                            "right": "33.9236%",
                            "top": "67.4040%",
                            "bottom": "26.8817%",
                            "width": "15.2778%",
                            "height": 40,
                        },
                        "widgetId": "widget_176",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO"},
                        },
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "5.0000%",
                            "top": "74.9129%",
                            "bottom": "22.8013%",
                            "width": "90.0000%",
                            "height": "2.2857%",
                        },
                        "widgetId": "widget_177",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "15": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "5.0000%",
                            "top": "78.9955%",
                            "bottom": "15.8616%",
                            "width": "90.0000%",
                            "height": "5.1429%",
                        },
                        "widgetId": "widget_178",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "5.0000%",
                            "top": "85.5446%",
                            "bottom": "9.7411%",
                            "width": "90.0000%",
                            "height": "4.7143%",
                        },
                        "widgetId": "widget_179",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "17": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "5.0000%",
                            "top": "91.6696%",
                            "bottom": 25,
                            "width": "90.0000%",
                            "height": "4.7143%",
                        },
                        "widgetId": "widget_180",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                            },
                        },
                    },
                    "18": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "29.2535%",
                            "right": "28.8976%",
                            "top": 27,
                            "bottom": "84.6228%",
                            "width": "41.9444%",
                            "height": "11.5714%",
                        },
                        "widgetId": "widget_903",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": "AUTO", "height": "AUTO"},
                        },
                    },
                },
                "order": [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                    "16",
                    "17",
                    "18",
                ],
            },
            "layout_730": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "298px", "height": 276},
                        "widgetId": "widget_242",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "252px", "height": "210px"},
                        "widgetId": "widget_243",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_732": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "298px", "height": 230},
                        "widgetId": "widget_240",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "252px", "height": "260px"},
                        "widgetId": "widget_239",
                    },
                },
                "order": ["0", "1"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/ink/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "exhibition",
        "widgets": {
            "widget_3": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_3",
            },
            "widget_4": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style=""><strong style="font-size: 30px; color: var(--dark-800);">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_4",
            },
            "widget_5": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. </span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_5",
            },
            "widget_6": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 20px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_6",
            },
            "widget_16": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_16",
                "layouts": {"DEFAULT": {"LARGE": "layout_10"}},
            },
            "widget_17": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_17",
            },
            "widget_18": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_18",
                "layouts": {"DEFAULT": {"LARGE": "layout_11"}},
            },
            "widget_19": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. </span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_19",
            },
            "widget_20": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_20",
                "layouts": {"DEFAULT": {"LARGE": "layout_12"}},
            },
            "widget_21": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_21",
            },
            "widget_34": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_34",
                "layouts": {"DEFAULT": {"LARGE": "layout_19"}},
            },
            "widget_35": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_35",
            },
            "widget_36": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_36",
            },
            "widget_37": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_37",
            },
            "widget_38": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_38",
            },
            "widget_53": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_1",
                        "views": ["view_1", "view_31", "view_32"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "default",
                        "standard": {
                            "scrollable": False,
                            "gap": "10px",
                            "showIcon": True,
                            "alternateIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><path d="M6 0a6 6 0 110 12A6 6 0 016 0zm0 2a4 4 0 100 8 4 4 0 000-8z" fill="#000" fill-rule="nonzero"></path></svg>',
                                "properties": {
                                    "filename": "circle",
                                    "originalName": None,
                                    "path": None,
                                    "color": "#fff",
                                    "size": 8,
                                    "inlineSvg": False,
                                },
                            },
                            "activedIcon": {
                                "svg": '<svg viewBox="0 0 6 6" xmlns="http://www.w3.org/2000/svg"><circle cx="1104" cy="1049" r="3" transform="translate(-1101 -1046)" fill="#000" fill-rule="nonzero"></circle></svg>',
                                "properties": {"filename": "dot-10", "size": 8},
                            },
                            "showText": False,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_53",
            },
            "widget_54": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_54",
                "layouts": {"DEFAULT": {"LARGE": "layout_26"}},
            },
            "widget_69": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_69",
                "layouts": {"DEFAULT": {"LARGE": "layout_31"}},
            },
            "widget_70": {
                "uri": "widgets/common/embed/",
                "version": "1.5.0",
                "config": {"embedType": "url"},
                "id": "widget_70",
            },
            "widget_73": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_73",
                "layouts": {"DEFAULT": {"LARGE": "layout_33"}},
            },
            "widget_74": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_74",
            },
            "widget_75": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_75",
                "layouts": {"DEFAULT": {"LARGE": "layout_34"}},
            },
            "widget_76": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_76",
            },
            "widget_77": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_77",
                "layouts": {"DEFAULT": {"LARGE": "layout_35"}},
            },
            "widget_78": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_78",
            },
            "widget_79": {
                "uri": "widgets/common/menu/",
                "version": "1.5.0",
                "config": {
                    "type": "nav",
                    "menuStyle": "underline",
                    "vertical": False,
                    "advanced": True,
                    "standard": {
                        "textAlign": "center",
                        "gap": "0px",
                        "showText": True,
                        "showIcon": False,
                        "iconPosition": "start",
                    },
                    "variant": {
                        "size": {
                            "fontSize": "0.875rem",
                            "lineHeight": 1.5,
                            "paddingX": "1rem",
                            "paddingY": "0.25rem",
                        },
                        "gutter": "0.25rem",
                        "icon": {"spacing": "0.5rem"},
                        "root": {"border": {"width": 0}, "borderRadius": "0.125rem"},
                        "item": {
                            "default": {
                                "color": "#181818",
                                "bg": "transparent",
                                "border": {"color": "transparent", "width": "2px"},
                                "borderRadius": "0.125rem",
                                "size": "18px",
                            },
                            "hover": {
                                "color": "#000000",
                                "border": {"color": "#000000", "width": "2px"},
                                "borderRadius": "0.125rem",
                                "size": "18px",
                            },
                            "active": {
                                "color": "#000000",
                                "border": {"color": "#000000", "width": "2px"},
                                "borderRadius": "0.125rem",
                                "size": "18px",
                            },
                            "disabled": {
                                "color": "#989898",
                                "shadow": "none",
                                "borderRadius": "0.125rem",
                            },
                            "focus": {
                                "shadow": "inset 0 0 0 1px #000",
                                "borderRadius": "0.125rem",
                            },
                        },
                    },
                    "paper": {
                        "bg": "#f0f0f0",
                        "color": "#050505",
                        "paddingX": 0,
                        "paddingY": 0,
                        "border": {"width": 0, "color": "transparent"},
                    },
                },
                "id": "widget_79",
            },
            "widget_80": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_80",
            },
            "widget_81": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {"color": "#f0f0f0", "fillType": "fill"},
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {"color": "#fff", "fillType": "fill"},
                                "border": {
                                    "type": "solid",
                                    "color": "#c5c5c5",
                                    "width": "0px",
                                },
                                "text": {"color": "#181818", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_81",
            },
            "widget_82": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_82",
                "layouts": {"DEFAULT": {"LARGE": "layout_37"}},
            },
            "widget_83": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_83",
            },
            "widget_84": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_84",
            },
            "widget_85": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_85",
                "layouts": {"DEFAULT": {"LARGE": "layout_38"}},
            },
            "widget_86": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p><br></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_86",
            },
            "widget_87": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit.\ufeff</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_87",
            },
            "widget_88": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M2 8c0-.291.231-.527.516-.527h9.508L8.527 3.9a.535.535 0 010-.746.508.508 0 01.73 0L14 8l-4.743 4.846a.508.508 0 01-.73 0 .535.535 0 010-.746l3.497-3.573H2.516A.522.522 0 012 8z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "RIGHT",
                        },
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_3",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {
                                "iconProps": {"color": "var(--primary-600)", "size": 14}
                            },
                            "regular": {
                                "iconProps": {"color": "var(--dark-600)", "size": 14}
                            },
                        },
                        "themeStyle": {"quickStyleType": "tertiary"},
                    },
                },
                "id": "widget_88",
                "useDataSources": [],
            },
            "widget_119": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_119",
                "layouts": {"DEFAULT": {"LARGE": "layout_47"}},
            },
            "widget_120": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_120",
                "layouts": {"DEFAULT": {"LARGE": "layout_48"}},
            },
            "widget_121": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_121",
            },
            "widget_122": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_122",
            },
            "widget_123": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_123",
                "layouts": {"DEFAULT": {"LARGE": "layout_49"}},
            },
            "widget_124": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "id": "widget_124",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_50",
                        "MEDIUM": "layout_519",
                        "SMALL": "layout_564",
                    }
                },
                "config": {
                    "toolConifg": {"canZoom": True, "canHome": True, "canSearch": True},
                    "isUseCustomMapState": False,
                },
            },
            "widget_128": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_1",
                        "views": ["view_1", "view_31", "view_32"],
                    },
                    "display": {
                        "advanced": True,
                        "vertical": False,
                        "type": "navButtonGroup",
                        "navStyle": "tertiary",
                        "standard": {
                            "showPageNumber": True,
                            "previousText": "",
                            "previousIcon": None,
                            "nextText": "",
                            "nextIcon": None,
                            "current": 1,
                            "totalPage": 4,
                            "disablePrevious": True,
                            "step": 1,
                        },
                        "variant": {
                            "root": {},
                            "item": {
                                "default": {
                                    "color": "#181818",
                                    "bg": "transparent",
                                    "border": {"width": "0px", "color": "transparent"},
                                    "size": "18px",
                                    "icon": {"size": "18px"},
                                },
                                "hover": {
                                    "color": "var(--dark-100)",
                                    "bg": "transparent",
                                    "border": {"width": "0px", "color": "transparent"},
                                    "size": "18px",
                                    "icon": {"size": "18px"},
                                },
                                "active": {"color": "#fff", "bg": "#076fe5"},
                                "disabled": {
                                    "color": "#8b8b8b",
                                    "bg": "transparent",
                                    "border": {"width": "0px", "color": "transparent"},
                                    "size": "18px",
                                    "icon": {"size": "18px"},
                                },
                                "focus": {
                                    "border": {"width": "0px", "color": "transparent"}
                                },
                            },
                        },
                    },
                },
                "id": "widget_128",
            },
            "widget_129": {
                "uri": "widgets/common/list/",
                "version": "1.5.0",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE5",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "scrollBarOpen": True,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {"LARGE": {"width": 386.66, "height": 152.28}},
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0, 0, 0, 1)",
                                    "width": "1px",
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {"LARGE": {"width": 386.66, "height": 152.28}},
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0, 0, 0, 1)",
                                    "width": "1px",
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {"LARGE": {"width": 386.66, "height": 152.28}},
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(133, 55, 241, 1)",
                                    "width": "1px",
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                },
                "id": "widget_129",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_57",
                        "MEDIUM": "layout_516",
                        "SMALL": "layout_561",
                    },
                    "SELECTED": {
                        "LARGE": "layout_54",
                        "MEDIUM": "layout_517",
                        "SMALL": "layout_562",
                    },
                    "HOVER": {
                        "LARGE": "layout_55",
                        "MEDIUM": "layout_518",
                        "SMALL": "layout_563",
                    },
                },
                "useDataSourcesEnabled": True,
            },
            "widget_133": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "cropShape": "rectangle",
                                "svgPath": "M0 0v28h28V0H0z",
                                "svgViewBox": "0 0 28 28",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_133",
            },
            "widget_134": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 18px;">Double click to edit text</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_134",
            },
            "widget_135": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {"toolTip": "", "linkParam": {}},
                    "styleConfig": {
                        "name": "Default",
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                            "clicked": {},
                        },
                        "useCustom": False,
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_135",
            },
            "widget_140": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_140",
                "layouts": {"DEFAULT": {"LARGE": "layout_60"}},
            },
            "widget_141": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_141",
            },
            "widget_142": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_142",
            },
            "widget_143": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_143",
            },
            "widget_144": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_144",
            },
            "widget_145": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_145",
                "layouts": {"DEFAULT": {"LARGE": "layout_61"}},
            },
            "widget_146": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_146",
            },
            "widget_147": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_147",
            },
            "widget_148": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_148",
            },
            "widget_149": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_149",
            },
            "widget_150": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [100, 10, 100, 10], "unit": "px"}},
                },
                "id": "widget_150",
                "layouts": {"DEFAULT": {"LARGE": "layout_62"}},
            },
            "widget_151": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_151",
            },
            "widget_152": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_152",
                "layouts": {"DEFAULT": {"LARGE": "layout_63"}},
            },
            "widget_153": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_153",
                "layouts": {"DEFAULT": {"LARGE": "layout_64"}},
            },
            "widget_154": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style=""><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_154",
            },
            "widget_155": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_155",
            },
            "widget_158": {
                "uri": "widgets/common/menu/",
                "version": "1.5.0",
                "config": {
                    "type": "nav",
                    "menuStyle": "underline",
                    "vertical": False,
                    "advanced": True,
                    "standard": {
                        "textAlign": "center",
                        "gap": "0px",
                        "showText": True,
                        "showIcon": False,
                        "iconPosition": "start",
                    },
                    "variant": {
                        "size": {
                            "fontSize": "0.875rem",
                            "lineHeight": 1.5,
                            "paddingX": "1rem",
                            "paddingY": "0.25rem",
                        },
                        "gutter": "0.25rem",
                        "icon": {"spacing": "0.5rem"},
                        "root": {"border": {"width": 0}, "borderRadius": "0.125rem"},
                        "item": {
                            "default": {
                                "color": "var(--dark-800)",
                                "bg": "transparent",
                                "border": {"color": "", "width": "2px"},
                                "borderRadius": "0.125rem",
                            },
                            "hover": {
                                "color": "var(--dark-800)",
                                "border": {"color": "var(--dark-800)", "width": "2px"},
                                "borderRadius": "0.125rem",
                            },
                            "active": {
                                "color": "var(--primary-500)",
                                "border": {
                                    "color": "var(--primary-500)",
                                    "width": "2px",
                                },
                                "borderRadius": "0.125rem",
                                "bg": "",
                            },
                            "disabled": {
                                "color": "#989898",
                                "shadow": "none",
                                "borderRadius": "0.125rem",
                            },
                            "focus": {
                                "shadow": "inset 0 0 0 1px #000",
                                "borderRadius": "0.125rem",
                            },
                        },
                    },
                    "paper": {
                        "bg": "#f0f0f0",
                        "color": "#050505",
                        "paddingX": 0,
                        "paddingY": 0,
                        "border": {"width": 0, "color": "transparent"},
                    },
                },
                "id": "widget_158",
            },
            "widget_159": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "page_0",
                            "openType": "_self",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_159",
            },
            "widget_160": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "linkParam": {"linkType": "DIALOG", "value": "dialog_1"}
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_160",
                "useDataSources": [],
            },
            "widget_163": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 30px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_163",
            },
            "widget_164": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 20px; color: var(--dark-800);">Text</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_164",
            },
            "widget_165": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 20px; color: var(--dark-800);">Text</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_165",
            },
            "widget_166": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 20px; color: var(--dark-800);">Text</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_166",
            },
            "widget_167": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_167",
            },
            "widget_168": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_168",
            },
            "widget_169": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_169",
            },
            "widget_170": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_170",
            },
            "widget_171": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_171",
            },
            "widget_172": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_172",
            },
            "widget_173": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_173",
            },
            "widget_174": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_174",
            },
            "widget_175": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_175",
            },
            "widget_176": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "undefinedpx",
                                },
                                "text": {"color": "var(--dark-100)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "rgba(0,0,0,0)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "18px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                        },
                    },
                },
                "id": "widget_176",
            },
            "widget_177": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-800)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_177",
            },
            "widget_178": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 20px;">Text</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_178",
            },
            "widget_179": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p><br></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_179",
            },
            "widget_180": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p><br></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_180",
            },
            "widget_181": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_181",
                "layouts": {"DEFAULT": {"LARGE": "layout_68"}},
            },
            "widget_182": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_182",
            },
            "widget_183": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_183",
            },
            "widget_184": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_184",
                "layouts": {"DEFAULT": {"LARGE": "layout_69"}},
            },
            "widget_185": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-800); font-size: 20px;">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. </span></p>',
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_185",
            },
            "widget_186": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_186",
                "layouts": {"DEFAULT": {"LARGE": "layout_70"}},
            },
            "widget_187": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p><br></p>",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_187",
            },
            "widget_188": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p><br></p>",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_188",
            },
            "widget_189": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-100); font-size: 16px;">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. \ufeff</span></p>',
                    "placeholder": '<p><span style="color: var(--dark-100); font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_189",
            },
            "widget_190": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_190",
                "layouts": {"DEFAULT": {"LARGE": "layout_71"}},
            },
            "widget_191": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="color: var(--dark-800); font-size: 20px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_191",
            },
            "widget_192": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_192",
            },
            "widget_193": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_193",
            },
            "widget_194": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_194",
            },
            "widget_195": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_195",
            },
            "widget_196": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_196",
            },
            "widget_197": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_197",
            },
            "widget_198": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_198",
            },
            "widget_199": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_199",
            },
            "widget_200": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_200",
            },
            "widget_201": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_201",
                "layouts": {"DEFAULT": {"LARGE": "layout_72"}},
            },
            "widget_202": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_202",
                "layouts": {"DEFAULT": {"LARGE": "layout_73"}},
            },
            "widget_203": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="color: var(--dark-800); font-size: 20px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_203",
            },
            "widget_204": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_204",
            },
            "widget_205": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_205",
            },
            "widget_206": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_206",
            },
            "widget_207": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_207",
            },
            "widget_208": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_208",
            },
            "widget_209": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_209",
            },
            "widget_210": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--dark-100)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_210",
            },
            "widget_211": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p><br></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_211",
            },
            "widget_212": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_212",
                "layouts": {"DEFAULT": {"LARGE": "layout_74"}},
            },
            "widget_217": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_217",
                "layouts": {"DEFAULT": {"LARGE": "layout_79"}},
            },
            "widget_218": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_218",
            },
            "widget_219": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_219",
                "layouts": {"DEFAULT": {"LARGE": "layout_80"}},
            },
            "widget_220": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_4",
                        "views": [
                            "view_15",
                            "view_16",
                            "view_17",
                            "view_18",
                            "view_19",
                        ],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "underline",
                        "standard": {
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": True,
                        "advanced": True,
                        "variant": {
                            "size": {
                                "fontSize": "0.875rem",
                                "lineHeight": 1.5,
                                "paddingX": "1rem",
                                "paddingY": "0.5rem",
                            },
                            "gutter": "0.25rem",
                            "icon": {"spacing": "0.5rem"},
                            "root": {
                                "border": {"width": 0},
                                "borderRadius": "0.125rem",
                            },
                            "item": {
                                "default": {
                                    "color": "var(--dark-400)",
                                    "bg": "transparent",
                                    "border": {"color": "transparent", "width": "2px"},
                                    "borderRadius": "0.125rem",
                                    "size": "18px",
                                },
                                "hover": {
                                    "color": "var(--black)",
                                    "border": {"color": "transparent", "width": "2px"},
                                    "borderRadius": "0.125rem",
                                    "size": "18px",
                                },
                                "active": {
                                    "color": "var(--primary-600)",
                                    "border": {
                                        "color": "var(--primary-600)",
                                        "width": "2px",
                                    },
                                    "borderRadius": "0.125rem",
                                    "size": "18px",
                                },
                                "disabled": {
                                    "color": "#5e5e5e",
                                    "shadow": "none",
                                    "borderRadius": "0.125rem",
                                },
                            },
                        },
                    },
                },
                "id": "widget_220",
            },
            "widget_221": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_221",
            },
            "widget_222": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_222",
            },
            "widget_223": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_223",
            },
            "widget_224": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_224",
            },
            "widget_225": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_225",
            },
            "widget_226": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_226",
                "layouts": {"DEFAULT": {"LARGE": "layout_86"}},
            },
            "widget_227": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_227",
            },
            "widget_228": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 100, 10], "unit": "px"}},
                },
                "id": "widget_228",
                "layouts": {"DEFAULT": {"LARGE": "layout_87"}},
            },
            "widget_229": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_229",
            },
            "widget_230": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_230",
            },
            "widget_231": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_231",
            },
            "widget_232": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_232",
            },
            "widget_233": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_233",
            },
            "widget_234": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_5",
                        "views": [
                            "view_20",
                            "view_21",
                            "view_22",
                            "view_23",
                            "view_24",
                            "view_25",
                        ],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "underline",
                        "standard": {
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": True,
                        "advanced": True,
                        "variant": {
                            "size": {
                                "fontSize": "0.875rem",
                                "lineHeight": 1.5,
                                "paddingX": "1rem",
                                "paddingY": "0.5rem",
                            },
                            "gutter": "0.25rem",
                            "icon": {"spacing": "0.5rem"},
                            "root": {
                                "border": {"width": 0},
                                "borderRadius": "0.125rem",
                            },
                            "item": {
                                "default": {
                                    "color": "var(--dark-400)",
                                    "bg": "transparent",
                                    "border": {"color": "transparent", "width": "2px"},
                                    "borderRadius": "0.125rem",
                                    "size": "18px",
                                },
                                "hover": {
                                    "color": "var(--black)",
                                    "border": {"color": "transparent", "width": "2px"},
                                    "borderRadius": "0.125rem",
                                },
                                "active": {
                                    "color": "var(--primary-600)",
                                    "border": {
                                        "color": "var(--primary-600)",
                                        "width": "2px",
                                    },
                                    "borderRadius": "0.125rem",
                                },
                                "disabled": {
                                    "color": "#5e5e5e",
                                    "shadow": "none",
                                    "borderRadius": "0.125rem",
                                },
                            },
                        },
                    },
                },
                "id": "widget_234",
            },
            "widget_235": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_235",
            },
            "widget_236": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_236",
                "layouts": {"DEFAULT": {"LARGE": "layout_94"}},
            },
            "widget_237": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_237",
            },
            "widget_238": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 100, 10], "unit": "px"}},
                },
                "id": "widget_238",
                "layouts": {"DEFAULT": {"LARGE": "layout_95"}},
            },
            "widget_239": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "description": "Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim. ",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -16398178.009336105,
                                "ymin": 3062953.580262496,
                                "xmax": -10156024.531457132,
                                "ymax": 5685049.398556485,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 18489297.737236,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -13277101.270396618,
                                    "y": 4374001.4894094905,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_239editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "description": "Ne malis tollit scriptorem sit, \nmundi verear patrioque et per. Possim temporibus ex vim, \nvim option facilis iracundia te. Quem choro philosophia \nvel ut. Decore fabulas partiendo eam id, sea malis \neuismod percipit in.",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -8623925.38605369,
                                "ymin": 4685505.9247800475,
                                "xmax": -8526391.737961939,
                                "ymax": 4726476.171940847,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 288895.277144,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -8575158.562007815,
                                    "y": 4705991.048360447,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_239editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "description": "Quem choro philosophia \nvel ut. Decore fabulas partiendo eam id, sea malis \neuismod percipit in. Rebum facilisis concludaturque ex \neos wrror noster.",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -9107656.4409315,
                                "ymin": 3296529.4831420346,
                                "xmax": -9010122.792839749,
                                "ymax": 3337499.7303028335,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 288895.277144,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -9058889.616885625,
                                    "y": 3317014.606722434,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_239editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 4,
                            "name": "Bookmark-4",
                            "title": "Bookmark-4",
                            "description": "Audiam salutatus ut ius. Ne malis tollit scriptorem sit, \nmundi verear patrioque et per. Possim temporibus ex vim, \nvim option facilis iracundia te. Quem choro philosophia \nvel ut. Decore fabulas partiendo eam id.",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -8335970.787958424,
                                "ymin": 4929045.093821363,
                                "xmax": -8140903.491774579,
                                "ymax": 5011749.958425955,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 577790.554289,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -8238437.139866501,
                                    "y": 4970397.526123659,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_239editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 5,
                            "name": "Bookmark-5",
                            "title": "Bookmark-5",
                            "description": "Vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. ",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -10811553.699782591,
                                "ymin": 3390984.361120283,
                                "xmax": -10421419.10741524,
                                "ymax": 3554865.349763622,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 1155581.108577,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -10616486.403598916,
                                    "y": 3472924.8554419526,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_239editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": True,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_239",
                "layouts": {},
                "useMapWidgetIds": ["widget_240"],
            },
            "widget_240": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {"canZoom": True, "canHome": True, "canSearch": True},
                    "isUseCustomMapState": False,
                },
                "id": "widget_240",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_97",
                        "MEDIUM": "layout_294",
                        "SMALL": "layout_444",
                    }
                },
            },
            "widget_241": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_6",
                        "views": ["view_26", "view_27"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "underline",
                        "standard": {
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_241",
            },
            "widget_242": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {"canZoom": True, "canHome": True, "canSearch": True},
                    "isUseCustomMapState": False,
                },
                "id": "widget_242",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_99",
                        "MEDIUM": "layout_296",
                        "SMALL": "layout_446",
                    }
                },
            },
            "widget_243": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "LIST",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "California",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -14555128.383324426,
                                "ymin": 3649989.9574924936,
                                "xmax": -11999074.15746881,
                                "ymax": 5098013.021326488,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -13277101.270396618,
                                    "y": 4374001.4894094905,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Washington D.C.",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -8615096.909286765,
                                "ymin": 4683365.687988066,
                                "xmax": -8535220.214728864,
                                "ymax": 4728616.408732829,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 288895.277144,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -8575158.562007815,
                                    "y": 4705991.048360447,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Orlando",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -9098827.964164576,
                                "ymin": 3294389.2463500523,
                                "xmax": -9018951.269606674,
                                "ymax": 3339639.967094816,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 288895.277144,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -9058889.616885625,
                                    "y": 3317014.606722434,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 4,
                            "name": "New York",
                            "title": "Bookmark-4",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -8318313.83442454,
                                "ymin": 4925146.805378817,
                                "xmax": -8158560.445308462,
                                "ymax": 5015648.2468685005,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 577790.554289,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -8238437.139866501,
                                    "y": 4970397.526123659,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 5,
                            "name": "Houston",
                            "title": "Bookmark-5",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -10776239.792714857,
                                "ymin": 3382423.4139523474,
                                "xmax": -10456733.014482975,
                                "ymax": 3563426.296931558,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 1155581.108577,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -10616486.403598916,
                                    "y": 3472924.8554419526,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 6,
                            "name": "Chicago",
                            "title": "Bookmark-6",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -9835076.40552504,
                                "ymin": 5098451.5472655045,
                                "xmax": -9675323.016408961,
                                "ymax": 5188952.988755188,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 577790.554289,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -9755199.710967,
                                    "y": 5143702.268010346,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 7,
                            "name": "Columbus",
                            "title": "Bookmark-7",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -9319700.55899942,
                                "ymin": 4815210.803606071,
                                "xmax": -9159947.16988334,
                                "ymax": 4905712.245095754,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 577790.554289,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -9239823.86444138,
                                    "y": 4860461.524350912,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 8,
                            "name": "Jeckson",
                            "title": "Bookmark-8",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 2354203.151888173,
                                "ymin": 4700929.749766148,
                                "xmax": 2359195.44529818,
                                "ymax": 4703757.919812773,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 18055.954822,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 2356699.2985931765,
                                    "y": 4702343.8347894605,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 9,
                            "name": "Rome",
                            "title": "Bookmark-9",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 1311132.7005729368,
                                "ymin": 5101332.425147486,
                                "xmax": 1470886.0896890163,
                                "ymax": 5191833.866637169,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 577790.554289,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 1391009.3951309766,
                                    "y": 5146583.145892328,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 10,
                            "name": "Dallas",
                            "title": "Bookmark-10",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -10855060.164231844,
                                "ymin": 3820694.6641620174,
                                "xmax": -10695306.775115766,
                                "ymax": 3911196.105651701,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 577790.554289,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -10775183.469673805,
                                    "y": 3865945.384906859,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 11,
                            "name": "Denver",
                            "title": "Bookmark-11",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -11767534.89831523,
                                "ymin": 4783024.994480948,
                                "xmax": -11607781.509199152,
                                "ymax": 4873526.435970631,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 577790.554289,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -11687658.203757191,
                                    "y": 4828275.71522579,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 12,
                            "name": "Toronto",
                            "title": "Bookmark-12",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -8917023.451758007,
                                "ymin": 5366036.130784131,
                                "xmax": -8757270.062641928,
                                "ymax": 5456537.572273814,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 577790.554289,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -8837146.757199967,
                                    "y": 5411286.851528972,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 13,
                            "name": "San francisco",
                            "title": "Bookmark-13",
                            "type": "2d",
                            "imgParam": {},
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -13707568.6824539,
                                "ymin": 4502774.637364612,
                                "xmax": -13547815.293337822,
                                "ymax": 4593276.078854295,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 577790.554289,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -13627691.987895861,
                                    "y": 4548025.358109454,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_243editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_243",
                "useMapWidgetIds": ["widget_242"],
            },
            "widget_244": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_244",
                "layouts": {"DEFAULT": {"LARGE": "layout_100"}},
            },
            "widget_245": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 30px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_245",
            },
            "widget_246": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_246",
                "layouts": {"DEFAULT": {"LARGE": "layout_101"}},
            },
            "widget_247": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Vertical",
                    "strokeStyle": {
                        "type": "Style1",
                        "color": "var(--light-700)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "Point0", "pointSize": 5},
                    "pointEnd": {"pointStyle": "Point0", "pointSize": 5},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_247",
            },
            "widget_248": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 24px; color: var(--dark-800);">1920\ufeff</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_248",
            },
            "widget_249": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_249",
                "layouts": {"DEFAULT": {"LARGE": "layout_102"}},
            },
            "widget_250": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_250",
                "layouts": {"DEFAULT": {"LARGE": "layout_103"}},
            },
            "widget_251": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_251",
            },
            "widget_252": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_252",
            },
            "widget_253": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_253",
            },
            "widget_254": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--dark-800); font-size: 24px;">1965</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_254",
            },
            "widget_255": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_255",
                "layouts": {"DEFAULT": {"LARGE": "layout_104"}},
            },
            "widget_256": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_256",
            },
            "widget_258": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_258",
            },
            "widget_259": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 24px; color: var(--dark-800);">1993</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_259",
            },
            "widget_260": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_260",
                "layouts": {"DEFAULT": {"LARGE": "layout_105"}},
            },
            "widget_261": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_261",
            },
            "widget_262": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_262",
            },
            "widget_263": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_263",
            },
            "widget_264": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 24px; color: var(--dark-800);">2008</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_264",
            },
            "widget_265": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_265",
                "layouts": {"DEFAULT": {"LARGE": "layout_106"}},
            },
            "widget_266": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_266",
            },
            "widget_267": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_267",
            },
            "widget_268": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_268",
            },
            "widget_269": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--dark-800); font-size: 24px;">2015</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_269",
            },
            "widget_270": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_270",
                "layouts": {"DEFAULT": {"LARGE": "layout_107"}},
            },
            "widget_271": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_271",
            },
            "widget_272": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_272",
            },
            "widget_273": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="font-size: 24px; color: var(--dark-800);">2020</strong></p>',
                    "placeholder": '<p><strong style="font-size: 20px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_273",
            },
            "widget_274": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_274",
                "layouts": {"DEFAULT": {"LARGE": "layout_108"}},
            },
            "widget_275": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_275",
            },
            "widget_276": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. Possim temporibus ex vim, vim option.Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque et per. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_276",
            },
            "widget_277": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_277",
            },
            "widget_278": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style=""><strong style="color: var(--dark-800); font-size: 30px;">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_278",
            },
            "widget_279": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 18px;">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. </span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_279",
            },
            "widget_280": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 20px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_280",
            },
            "widget_387": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_387",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_253"}},
            },
            "widget_388": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_388",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_254"}},
            },
            "widget_389": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_389",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_260"}},
            },
            "widget_390": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 100, 10], "unit": "px"}},
                },
                "id": "widget_390",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_261"}},
            },
            "widget_391": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_391",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_268"}},
            },
            "widget_392": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_392",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_269"}},
            },
            "widget_401": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_401",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_291"}},
            },
            "widget_402": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 100, 10], "unit": "px"}},
                },
                "id": "widget_402",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_292"}},
            },
            "widget_403": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_403",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_297"}},
            },
            "widget_409": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_409",
                "layouts": {"DEFAULT": {"SMALL": "layout_309"}},
            },
            "widget_410": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_410",
                "layouts": {"DEFAULT": {"SMALL": "layout_310"}},
            },
            "widget_411": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_411",
                "layouts": {"DEFAULT": {"SMALL": "layout_311"}},
            },
            "widget_412": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_412",
                "layouts": {"DEFAULT": {"SMALL": "layout_312"}},
            },
            "widget_413": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_413",
                "layouts": {"DEFAULT": {"SMALL": "layout_313"}},
            },
            "widget_414": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_414",
                "layouts": {"DEFAULT": {"SMALL": "layout_314"}},
            },
            "widget_415": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_415",
                "layouts": {"DEFAULT": {"SMALL": "layout_315"}},
            },
            "widget_416": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_416",
                "layouts": {"DEFAULT": {"SMALL": "layout_316"}},
            },
            "widget_417": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_417",
                "layouts": {"DEFAULT": {"SMALL": "layout_317"}},
            },
            "widget_418": {
                "uri": "widgets/layout/column/",
                "id": "widget_418",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_318"}},
            },
            "widget_419": {
                "uri": "widgets/layout/column/",
                "id": "widget_419",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_319"}},
            },
            "widget_420": {
                "uri": "widgets/layout/column/",
                "id": "widget_420",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_320"}},
            },
            "widget_421": {
                "uri": "widgets/layout/column/",
                "id": "widget_421",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_321"}},
            },
            "widget_422": {
                "uri": "widgets/layout/column/",
                "id": "widget_422",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_322"}},
            },
            "widget_423": {
                "uri": "widgets/layout/column/",
                "id": "widget_423",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_323"}},
            },
            "widget_424": {
                "uri": "widgets/layout/column/",
                "id": "widget_424",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_324"}},
            },
            "widget_425": {
                "uri": "widgets/layout/column/",
                "id": "widget_425",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_325"}},
            },
            "widget_426": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_426",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_327"}},
            },
            "widget_427": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_427",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_328"}},
            },
            "widget_428": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_428",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_329"}},
            },
            "widget_429": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_429",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_330"}},
            },
            "widget_430": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_430",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_331"}},
            },
            "widget_431": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_431",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_332"}},
            },
            "widget_432": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_432",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_333"}},
            },
            "widget_433": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_433",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_334"}},
            },
            "widget_434": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_434",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_335"}},
            },
            "widget_435": {
                "uri": "widgets/common/menu/",
                "version": "1.5.0",
                "config": {
                    "type": "drawer",
                    "menuStyle": "default",
                    "vertical": True,
                    "advanced": False,
                    "standard": {
                        "icon": {
                            "svg": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg" class="css-1i7frhi jimu-icon"><path d="M2 1a1 1 0 100 2h12a1 1 0 100-2H2zm0-1h12a2 2 0 010 4H2a2 2 0 010-4zm0 7a1 1 0 100 2h12a1 1 0 100-2H2zm0-1h12a2 2 0 010 4H2a2 2 0 010-4zm0 7a1 1 0 100 2h12a1 1 0 100-2H2zm0-1h12a2 2 0 010 4H2a2 2 0 010-4z" fill="currentColor" fill-rule="nonzero"></path></svg>',
                            "properties": {
                                "color": "#121212",
                                "size": 20,
                                "inlineSvg": True,
                            },
                        },
                        "anchor": "left",
                        "submenuMode": "foldable",
                        "textAlign": "center",
                        "gap": "0px",
                    },
                },
                "id": "widget_435",
            },
            "widget_439": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_439",
                "layouts": {"DEFAULT": {"LARGE": "layout_342"}},
            },
            "widget_440": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_440",
                "layouts": {"DEFAULT": {"LARGE": "layout_343"}},
            },
            "widget_441": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_7",
                        "views": ["view_28", "view_29", "view_30"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "default",
                        "standard": {
                            "scrollable": False,
                            "gap": "10px",
                            "showIcon": True,
                            "alternateIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><path d="M6 0a6 6 0 110 12A6 6 0 016 0zm0 2a4 4 0 100 8 4 4 0 000-8z" fill="#000" fill-rule="nonzero"></path></svg>',
                                "properties": {
                                    "filename": "circle",
                                    "originalName": None,
                                    "path": None,
                                    "color": "#fff",
                                    "size": 8,
                                    "inlineSvg": False,
                                },
                            },
                            "activedIcon": {
                                "svg": '<svg viewBox="0 0 6 6" xmlns="http://www.w3.org/2000/svg"><circle cx="1104" cy="1049" r="3" transform="translate(-1101 -1046)" fill="#000" fill-rule="nonzero"></circle></svg>',
                                "properties": {"filename": "dot-10", "size": 8},
                            },
                            "showText": False,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_441",
            },
            "widget_442": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_442",
            },
            "widget_443": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_443",
                "style": {
                    "borderRadius": {"number": [100], "unit": "px"},
                    "border": {"type": "solid", "color": "#000000", "width": "2px"},
                },
            },
            "widget_444": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 24px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_444",
            },
            "widget_469": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_469",
            },
            "widget_470": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_470",
                "style": {
                    "borderRadius": {"number": [100], "unit": "px"},
                    "border": {"type": "solid", "color": "#000000", "width": "2px"},
                },
            },
            "widget_471": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 24px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_471",
            },
            "widget_475": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_475",
            },
            "widget_476": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_476",
                "style": {
                    "borderRadius": {"number": [100], "unit": "px"},
                    "border": {"type": "solid", "color": "#000000", "width": "2px"},
                },
            },
            "widget_477": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 24px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_477",
            },
            "widget_526": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_526",
                "layouts": {"DEFAULT": {"SMALL": "layout_441"}},
            },
            "widget_527": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 100, 10], "unit": "px"}},
                },
                "id": "widget_527",
                "layouts": {"DEFAULT": {"SMALL": "layout_442"}},
            },
            "widget_528": {
                "uri": "widgets/layout/column/",
                "id": "widget_528",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_447"}},
            },
            "widget_529": {
                "uri": "widgets/layout/column/",
                "id": "widget_529",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_448"}},
            },
            "widget_543": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_543",
                "layouts": {"DEFAULT": {"LARGE": "layout_466"}},
            },
            "widget_544": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 275.5,
                    "widgetHeight": 330.59375,
                },
                "id": "widget_544",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_474",
                        "MEDIUM": "layout_489",
                        "SMALL": "layout_534",
                    },
                    "HOVER": {
                        "LARGE": "layout_475",
                        "MEDIUM": "layout_490",
                        "SMALL": "layout_535",
                    },
                },
            },
            "widget_568": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_568",
            },
            "widget_569": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_569",
            },
            "widget_570": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_570",
            },
            "widget_571": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 275.5,
                    "widgetHeight": 330.59375,
                },
                "id": "widget_571",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_476",
                        "MEDIUM": "layout_491",
                        "SMALL": "layout_536",
                    },
                    "HOVER": {
                        "LARGE": "layout_477",
                        "MEDIUM": "layout_492",
                        "SMALL": "layout_537",
                    },
                },
            },
            "widget_572": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_572",
            },
            "widget_573": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_573",
            },
            "widget_574": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_574",
            },
            "widget_575": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_575",
            },
            "widget_576": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_576",
            },
            "widget_577": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 275.5,
                    "widgetHeight": 330.59375,
                },
                "id": "widget_577",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_478",
                        "MEDIUM": "layout_493",
                        "SMALL": "layout_538",
                    },
                    "HOVER": {
                        "LARGE": "layout_479",
                        "MEDIUM": "layout_494",
                        "SMALL": "layout_539",
                    },
                },
            },
            "widget_578": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_578",
            },
            "widget_579": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_579",
            },
            "widget_580": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_580",
            },
            "widget_581": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_581",
            },
            "widget_582": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_582",
            },
            "widget_583": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--secondary-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 275.5,
                    "widgetHeight": 330.59375,
                },
                "id": "widget_583",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_480",
                        "MEDIUM": "layout_495",
                        "SMALL": "layout_540",
                    },
                    "HOVER": {
                        "LARGE": "layout_481",
                        "MEDIUM": "layout_496",
                        "SMALL": "layout_541",
                    },
                },
            },
            "widget_584": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_584",
            },
            "widget_585": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_585",
            },
            "widget_586": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_586",
            },
            "widget_587": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_587",
            },
            "widget_588": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_588",
            },
            "widget_590": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_590",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_484"}},
            },
            "widget_591": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_591",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_485"}},
            },
            "widget_592": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_592",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_486"}},
            },
            "widget_593": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_593",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_487"}},
            },
            "widget_594": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_594",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_488"}},
            },
            "widget_595": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_595",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_497"}},
            },
            "widget_596": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_596",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_498"}},
            },
            "widget_597": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_597",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_503"}},
            },
            "widget_598": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_598",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_504"}},
            },
            "widget_599": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_599",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_505"}},
            },
            "widget_600": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_600",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_506"}},
            },
            "widget_601": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_601",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_507"}},
            },
            "widget_602": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_602",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_508"}},
            },
            "widget_603": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_603",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_509"}},
            },
            "widget_604": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_604",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_513"}},
            },
            "widget_605": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_605",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_514"}},
            },
            "widget_606": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_606",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_515"}},
            },
            "widget_607": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_607",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_520"}},
            },
            "widget_608": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_608",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_521"}},
            },
            "widget_609": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_609",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_522"}},
            },
            "widget_610": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_610",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_523"}},
            },
            "widget_611": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [60, 10, 60, 10], "unit": "px"}},
                },
                "id": "widget_611",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_524"}},
            },
            "widget_613": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_613",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_526"}},
            },
            "widget_615": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_615",
                "layouts": {"DEFAULT": {"SMALL": "layout_529"}},
            },
            "widget_616": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_616",
                "layouts": {"DEFAULT": {"SMALL": "layout_530"}},
            },
            "widget_617": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_617",
                "layouts": {"DEFAULT": {"SMALL": "layout_531"}},
            },
            "widget_618": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_618",
                "layouts": {"DEFAULT": {"SMALL": "layout_532"}},
            },
            "widget_619": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_619",
                "layouts": {"DEFAULT": {"SMALL": "layout_533"}},
            },
            "widget_620": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_620",
                "layouts": {"DEFAULT": {"SMALL": "layout_542"}},
            },
            "widget_621": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_621",
                "layouts": {"DEFAULT": {"SMALL": "layout_543"}},
            },
            "widget_622": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_622",
                "layouts": {"DEFAULT": {"SMALL": "layout_548"}},
            },
            "widget_623": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_623",
                "layouts": {"DEFAULT": {"SMALL": "layout_549"}},
            },
            "widget_624": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_624",
                "layouts": {"DEFAULT": {"SMALL": "layout_550"}},
            },
            "widget_625": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_625",
                "layouts": {"DEFAULT": {"SMALL": "layout_551"}},
            },
            "widget_626": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_626",
                "layouts": {"DEFAULT": {"SMALL": "layout_552"}},
            },
            "widget_627": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_627",
                "layouts": {"DEFAULT": {"SMALL": "layout_553"}},
            },
            "widget_628": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_628",
                "layouts": {"DEFAULT": {"SMALL": "layout_554"}},
            },
            "widget_629": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_629",
                "layouts": {"DEFAULT": {"SMALL": "layout_558"}},
            },
            "widget_630": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_630",
                "layouts": {"DEFAULT": {"SMALL": "layout_559"}},
            },
            "widget_631": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_631",
                "layouts": {"DEFAULT": {"SMALL": "layout_560"}},
            },
            "widget_632": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_632",
                "layouts": {"DEFAULT": {"SMALL": "layout_565"}},
            },
            "widget_633": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_633",
                "layouts": {"DEFAULT": {"SMALL": "layout_566"}},
            },
            "widget_634": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_634",
                "layouts": {"DEFAULT": {"SMALL": "layout_567"}},
            },
            "widget_635": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_635",
                "layouts": {"DEFAULT": {"SMALL": "layout_568"}},
            },
            "widget_636": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [60, 10, 60, 10], "unit": "px"}},
                },
                "id": "widget_636",
                "layouts": {"DEFAULT": {"SMALL": "layout_569"}},
            },
            "widget_639": {
                "uri": "widgets/layout/column/",
                "id": "widget_639",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_572"}},
            },
            "widget_640": {
                "uri": "widgets/layout/column/",
                "id": "widget_640",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_573"}},
            },
            "widget_641": {
                "uri": "widgets/layout/column/",
                "id": "widget_641",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_574"}},
            },
            "widget_642": {
                "uri": "widgets/layout/column/",
                "id": "widget_642",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_575"}},
            },
            "widget_643": {
                "uri": "widgets/layout/column/",
                "id": "widget_643",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_576"}},
            },
            "widget_644": {
                "uri": "widgets/layout/column/",
                "id": "widget_644",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_577"}},
            },
            "widget_645": {
                "uri": "widgets/layout/column/",
                "id": "widget_645",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_578"}},
            },
            "widget_646": {
                "uri": "widgets/layout/column/",
                "id": "widget_646",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_579"}},
            },
            "widget_647": {
                "uri": "widgets/layout/column/",
                "id": "widget_647",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_580"}},
            },
            "widget_648": {
                "uri": "widgets/layout/column/",
                "id": "widget_648",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_581"}},
            },
            "widget_649": {
                "uri": "widgets/layout/column/",
                "id": "widget_649",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_582"}},
            },
            "widget_650": {
                "uri": "widgets/layout/column/",
                "id": "widget_650",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_583"}},
            },
            "widget_651": {
                "uri": "widgets/layout/column/",
                "id": "widget_651",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_584"}},
            },
            "widget_652": {
                "uri": "widgets/layout/column/",
                "id": "widget_652",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_585"}},
            },
            "widget_653": {
                "uri": "widgets/layout/column/",
                "id": "widget_653",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_586"}},
            },
            "widget_654": {
                "uri": "widgets/layout/column/",
                "id": "widget_654",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_587"}},
            },
            "widget_655": {
                "uri": "widgets/layout/column/",
                "id": "widget_655",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_588"}},
            },
            "widget_656": {
                "uri": "widgets/layout/column/",
                "id": "widget_656",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_589"}},
            },
            "widget_657": {
                "uri": "widgets/layout/column/",
                "id": "widget_657",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_590"}},
            },
            "widget_659": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_659",
                "layouts": {"DEFAULT": {"LARGE": "layout_592"}},
            },
            "widget_660": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {"canZoom": True, "canHome": True, "canSearch": True},
                    "isUseCustomMapState": False,
                },
                "id": "widget_660",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_593",
                        "MEDIUM": "layout_705",
                        "SMALL": "layout_717",
                    }
                },
            },
            "widget_838": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_838",
            },
            "widget_839": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_839",
            },
            "widget_840": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_840",
            },
            "widget_841": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_841",
            },
            "widget_842": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_842",
            },
            "widget_843": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_843",
            },
            "widget_844": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_844",
            },
            "widget_845": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_845",
            },
            "widget_846": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_846",
            },
            "widget_847": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_847",
            },
            "widget_848": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_848",
            },
            "widget_849": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_849",
            },
            "widget_850": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_850",
            },
            "widget_851": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_851",
            },
            "widget_852": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_852",
            },
            "widget_853": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_853",
            },
            "widget_854": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_854",
            },
            "widget_855": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_855",
            },
            "widget_856": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_856",
            },
            "widget_857": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_857",
            },
            "widget_858": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {
                            "linkType": "VIEW",
                            "openType": "_self",
                            "value": "page_0,view_1",
                        },
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_858",
            },
            "widget_859": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 24px; color: var(--dark-800);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_859",
            },
            "widget_860": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 18px; color: var(--dark-800);">Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Decore fabulas partiendo eam id, sea malis euismod percipit in. \ufeffAudiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. Possim temporibus ex vim, vim option facilis iracundiate. Quem choro philosophia vel ut. Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper.</span></p>',
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_860",
            },
            "widget_861": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "hover": {"iconProps": {"size": 14}},
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "primary"},
                    },
                },
                "id": "widget_861",
            },
            "widget_862": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_4",
                        "views": [
                            "view_15",
                            "view_16",
                            "view_17",
                            "view_18",
                            "view_19",
                        ],
                    },
                    "display": {
                        "type": "navButtonGroup",
                        "navStyle": "tertiary",
                        "standard": {
                            "showPageNumber": True,
                            "previousText": "",
                            "previousIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="M2.479 5.47L8.019.162a.592.592 0 01.813 0 .534.534 0 010 .779L3.552 6l5.28 5.06a.534.534 0 010 .779.592.592 0 01-.813 0l-5.54-5.31a.726.726 0 010-1.058z" fill="#000"></path><path d="M0 0h12v12H0z"></path></g></svg>',
                                "properties": {"filename": "left-arrow-12", "size": 12},
                            },
                            "nextText": "",
                            "nextIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="M9.521 5.47L3.981.162a.592.592 0 00-.813 0 .534.534 0 000 .779L8.448 6l-5.28 5.06a.534.534 0 000 .779.592.592 0 00.813 0l5.54-5.31a.726.726 0 000-1.058z" fill="#000"></path><path d="M0 0h12v12H0z"></path></g></svg>',
                                "properties": {
                                    "filename": "right-arrow-12",
                                    "size": 12,
                                },
                            },
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_862",
            },
            "widget_863": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_5",
                        "views": [
                            "view_20",
                            "view_21",
                            "view_22",
                            "view_23",
                            "view_24",
                            "view_25",
                        ],
                    },
                    "display": {
                        "type": "navButtonGroup",
                        "navStyle": "tertiary",
                        "standard": {
                            "showPageNumber": True,
                            "previousText": "",
                            "previousIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="M2.479 5.47L8.019.162a.592.592 0 01.813 0 .534.534 0 010 .779L3.552 6l5.28 5.06a.534.534 0 010 .779.592.592 0 01-.813 0l-5.54-5.31a.726.726 0 010-1.058z" fill="#000"></path><path d="M0 0h12v12H0z"></path></g></svg>',
                                "properties": {"filename": "left-arrow-12", "size": 12},
                            },
                            "nextText": "",
                            "nextIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="M9.521 5.47L3.981.162a.592.592 0 00-.813 0 .534.534 0 000 .779L8.448 6l-5.28 5.06a.534.534 0 000 .779.592.592 0 00.813 0l5.54-5.31a.726.726 0 000-1.058z" fill="#000"></path><path d="M0 0h12v12H0z"></path></g></svg>',
                                "properties": {
                                    "filename": "right-arrow-12",
                                    "size": 12,
                                },
                            },
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_863",
            },
            "widget_864": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_864",
                "layouts": {"DEFAULT": {"SMALL": "layout_677"}},
            },
            "widget_865": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_865",
                "layouts": {"DEFAULT": {"SMALL": "layout_678"}},
            },
            "widget_866": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_866",
                "layouts": {"DEFAULT": {"SMALL": "layout_684"}},
            },
            "widget_867": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 100, 10], "unit": "px"}},
                },
                "id": "widget_867",
                "layouts": {"DEFAULT": {"SMALL": "layout_685"}},
            },
            "widget_868": {
                "uri": "widgets/layout/column/",
                "id": "widget_868",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_692"}},
            },
            "widget_869": {
                "uri": "widgets/layout/column/",
                "id": "widget_869",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_693"}},
            },
            "widget_870": {
                "uri": "widgets/layout/column/",
                "id": "widget_870",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_694"}},
            },
            "widget_871": {
                "uri": "widgets/layout/column/",
                "id": "widget_871",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_695"}},
            },
            "widget_872": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_4",
                        "views": [
                            "view_15",
                            "view_16",
                            "view_17",
                            "view_18",
                            "view_19",
                        ],
                    },
                    "display": {
                        "type": "navButtonGroup",
                        "navStyle": "tertiary",
                        "standard": {
                            "showPageNumber": True,
                            "previousText": "",
                            "previousIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="M2.479 5.47L8.019.162a.592.592 0 01.813 0 .534.534 0 010 .779L3.552 6l5.28 5.06a.534.534 0 010 .779.592.592 0 01-.813 0l-5.54-5.31a.726.726 0 010-1.058z" fill="#000"></path><path d="M0 0h12v12H0z"></path></g></svg>',
                                "properties": {"filename": "left-arrow-12", "size": 12},
                            },
                            "nextText": "",
                            "nextIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="M9.521 5.47L3.981.162a.592.592 0 00-.813 0 .534.534 0 000 .779L8.448 6l-5.28 5.06a.534.534 0 000 .779.592.592 0 00.813 0l5.54-5.31a.726.726 0 000-1.058z" fill="#000"></path><path d="M0 0h12v12H0z"></path></g></svg>',
                                "properties": {
                                    "filename": "right-arrow-12",
                                    "size": 12,
                                },
                            },
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_872",
            },
            "widget_873": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_5",
                        "views": [
                            "view_20",
                            "view_21",
                            "view_22",
                            "view_23",
                            "view_24",
                            "view_25",
                        ],
                    },
                    "display": {
                        "type": "navButtonGroup",
                        "navStyle": "tertiary",
                        "standard": {
                            "showPageNumber": True,
                            "previousText": "",
                            "previousIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="M2.479 5.47L8.019.162a.592.592 0 01.813 0 .534.534 0 010 .779L3.552 6l5.28 5.06a.534.534 0 010 .779.592.592 0 01-.813 0l-5.54-5.31a.726.726 0 010-1.058z" fill="#000"></path><path d="M0 0h12v12H0z"></path></g></svg>',
                                "properties": {"filename": "left-arrow-12", "size": 12},
                            },
                            "nextText": "",
                            "nextIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="M9.521 5.47L3.981.162a.592.592 0 00-.813 0 .534.534 0 000 .779L8.448 6l-5.28 5.06a.534.534 0 000 .779.592.592 0 00.813 0l5.54-5.31a.726.726 0 000-1.058z" fill="#000"></path><path d="M0 0h12v12H0z"></path></g></svg>',
                                "properties": {
                                    "filename": "right-arrow-12",
                                    "size": 12,
                                },
                            },
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_873",
            },
            "widget_874": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_874-snap1.jpg",
                                "originalName": "widget_874-snap1",
                                "url": "${appResourceUrl}/images/widget_874/widget_874-snap1.jpg",
                                "type": "IMAGE",
                                "size": 121400,
                                "created": 1593430183683,
                                "resourcesPrefix": "images/widget_874",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -15712742.020351829,
                                "ymin": -820162.0511185443,
                                "xmax": 411190.4742321037,
                                "ymax": 8102790.882777418,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7650775.773059863,
                                    "y": 3641314.415829437,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_874editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "description": "Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. \ufeff",
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "description": "Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. \ufeff",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_874-snap2.jpg",
                                "originalName": "widget_874-snap2",
                                "url": "${appResourceUrl}/images/widget_874/widget_874-snap2.jpg",
                                "type": "IMAGE",
                                "size": 123453,
                                "created": 1593430192008,
                                "resourcesPrefix": "images/widget_874",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -10236793.314003259,
                                "ymin": -6674397.610450352,
                                "xmax": 5887139.180580674,
                                "ymax": 2248555.323445611,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -2174827.0667112926,
                                    "y": -2212921.14350237,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_874editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "description": "Audiam salutatus ut ius. Ne malis tollit scriptorem sit, mundi verear patrioque etper. \ufeff",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_874-snap3.jpg",
                                "originalName": "widget_874-snap3",
                                "url": "${appResourceUrl}/images/widget_874/widget_874-snap3.jpg",
                                "type": "IMAGE",
                                "size": 58492,
                                "created": 1593430360839,
                                "resourcesPrefix": "images/widget_874",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -46849544.581695534,
                                "ymin": 5458670.724887071,
                                "xmax": -42818561.45804956,
                                "ymax": 7689408.958361061,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 18489297.737236,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -44834053.019872546,
                                    "y": 6574039.841624066,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_874editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "runtimeAddAllow": False,
                    "autoPlayAllow": True,
                    "autoInterval": 3,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_874",
                "useMapWidgetIds": ["widget_660"],
            },
            "widget_875": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_875",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_697"}},
            },
            "widget_876": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_876",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_698"}},
            },
            "widget_877": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_877",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_699"}},
            },
            "widget_878": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_878",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_700"}},
            },
            "widget_879": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_879",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_701"}},
            },
            "widget_880": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_880",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_702"}},
            },
            "widget_881": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_881",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_703"}},
            },
            "widget_882": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_882",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_704"}},
            },
            "widget_883": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_883",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_706"}},
            },
            "widget_884": {
                "id": "widget_884",
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"MEDIUM": "layout_707"}},
            },
            "widget_885": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_885",
                "layouts": {"DEFAULT": {"SMALL": "layout_709"}},
            },
            "widget_886": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_886",
                "layouts": {"DEFAULT": {"SMALL": "layout_710"}},
            },
            "widget_887": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_887",
                "layouts": {"DEFAULT": {"SMALL": "layout_711"}},
            },
            "widget_888": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_888",
                "layouts": {"DEFAULT": {"SMALL": "layout_712"}},
            },
            "widget_890": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_890",
                "layouts": {"DEFAULT": {"SMALL": "layout_714"}},
            },
            "widget_892": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_892",
                "layouts": {"DEFAULT": {"SMALL": "layout_716"}},
            },
            "widget_893": {
                "uri": "widgets/layout/column/",
                "id": "widget_893",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_718"}},
            },
            "widget_894": {
                "uri": "widgets/layout/column/",
                "id": "widget_894",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_719"}},
            },
            "widget_897": {
                "uri": "widgets/layout/column/",
                "id": "widget_897",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_722"}},
            },
            "widget_899": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_899",
                "layouts": {"DEFAULT": {"SMALL": "layout_724"}},
            },
            "widget_900": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_900",
                "layouts": {"DEFAULT": {"SMALL": "layout_725"}},
            },
            "widget_901": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_901",
                "layouts": {"DEFAULT": {"SMALL": "layout_726"}},
            },
            "widget_902": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "icon": {
                            "data": '<svg viewBox="0 0 14 14" xmlns="http://www.w3.org/2000/svg"><g fill-rule="nonzero" fill="none"><path d="M7.745 7l4.1 4.1a.527.527 0 01-.745.746L7 7.746l-4.1 4.1a.527.527 0 01-.746-.746l4.1-4.1-4.1-4.1a.527.527 0 01.746-.746l4.1 4.1 4.1-4.1a.527.527 0 01.746.746L7.746 7z" fill="#000"></path><path d="M1 1h12v12H1z"></path></g></svg>',
                            "position": "LEFT",
                        },
                        "text": "",
                        "textExpression": None,
                        "linkParam": {
                            "linkType": "PAGE",
                            "value": "current_page",
                            "openType": "_self",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "transparent",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "transparent",
                                    "width": "1px",
                                },
                                "text": {"color": "#141414", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [0, 0, 0, 0]},
                                "iconProps": {"color": "var(--dark-100)", "size": 20},
                            },
                            "hover": {
                                "background": {"fillType": "fill"},
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"color": "#484848", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [0, 0, 0, 0]},
                                "iconProps": {"color": "var(--dark)", "size": 20},
                            },
                        },
                    },
                },
                "id": "widget_902",
                "useDataSources": [],
            },
            "widget_903": {
                "uri": "widgets/common/share/",
                "version": "1.5.0",
                "config": {
                    "uiMode": "INLINE",
                    "popup": {
                        "icon": "",
                        "items": ["embed", "qrcode", "email", "facebook", "twitter"],
                        "tooltip": "",
                    },
                    "inline": {
                        "items": ["facebook", "twitter", "email"],
                        "design": {
                            "direction": "HORIZONTAL",
                            "hideLabel": True,
                            "btnRad": 0,
                            "btnColor": "",
                            "iconColor": "",
                            "size": "lg",
                        },
                    },
                },
                "id": "widget_903",
            },
            "widget_904": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_904",
            },
            "widget_905": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_905",
            },
            "widget_906": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_906",
            },
            "widget_907": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_907",
            },
            "widget_909": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_909",
                "layouts": {"DEFAULT": {"SMALL": "layout_730"}},
            },
            "widget_911": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_911",
                "layouts": {"DEFAULT": {"SMALL": "layout_732"}},
            },
        },
        "widgetsManifest": {},
        "views": {
            "view_1": {
                "id": "view_1",
                "layout": {
                    "LARGE": "layout_20",
                    "MEDIUM": "layout_499",
                    "SMALL": "layout_544",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "var(--light)",
                "backgroundIMage": {},
            },
            "view_15": {
                "id": "view_15",
                "layout": {
                    "LARGE": "layout_81",
                    "MEDIUM": "layout_255",
                    "SMALL": "layout_679",
                },
            },
            "view_16": {
                "id": "view_16",
                "layout": {
                    "LARGE": "layout_82",
                    "MEDIUM": "layout_256",
                    "SMALL": "layout_680",
                },
            },
            "view_17": {
                "id": "view_17",
                "layout": {
                    "LARGE": "layout_83",
                    "MEDIUM": "layout_257",
                    "SMALL": "layout_681",
                },
            },
            "view_18": {
                "id": "view_18",
                "layout": {
                    "LARGE": "layout_84",
                    "MEDIUM": "layout_258",
                    "SMALL": "layout_682",
                },
            },
            "view_19": {
                "id": "view_19",
                "layout": {
                    "LARGE": "layout_85",
                    "MEDIUM": "layout_259",
                    "SMALL": "layout_683",
                },
            },
            "view_20": {
                "id": "view_20",
                "layout": {
                    "LARGE": "layout_88",
                    "MEDIUM": "layout_262",
                    "SMALL": "layout_686",
                },
            },
            "view_21": {
                "id": "view_21",
                "layout": {
                    "LARGE": "layout_89",
                    "MEDIUM": "layout_263",
                    "SMALL": "layout_687",
                },
            },
            "view_22": {
                "id": "view_22",
                "layout": {
                    "LARGE": "layout_90",
                    "MEDIUM": "layout_264",
                    "SMALL": "layout_688",
                },
            },
            "view_23": {
                "id": "view_23",
                "layout": {
                    "LARGE": "layout_91",
                    "MEDIUM": "layout_265",
                    "SMALL": "layout_689",
                },
            },
            "view_24": {
                "id": "view_24",
                "layout": {
                    "LARGE": "layout_92",
                    "MEDIUM": "layout_266",
                    "SMALL": "layout_690",
                },
            },
            "view_25": {
                "id": "view_25",
                "layout": {
                    "LARGE": "layout_93",
                    "MEDIUM": "layout_267",
                    "SMALL": "layout_691",
                },
            },
            "view_26": {
                "id": "view_26",
                "layout": {
                    "LARGE": "layout_96",
                    "MEDIUM": "layout_293",
                    "SMALL": "layout_443",
                },
            },
            "view_27": {
                "id": "view_27",
                "layout": {
                    "LARGE": "layout_98",
                    "MEDIUM": "layout_295",
                    "SMALL": "layout_445",
                },
            },
            "view_28": {
                "id": "view_28",
                "layout": {
                    "LARGE": "layout_344",
                    "MEDIUM": "layout_510",
                    "SMALL": "layout_555",
                },
            },
            "view_29": {
                "id": "view_29",
                "layout": {
                    "LARGE": "layout_379",
                    "MEDIUM": "layout_511",
                    "SMALL": "layout_556",
                },
            },
            "view_30": {
                "id": "view_30",
                "layout": {
                    "LARGE": "layout_381",
                    "MEDIUM": "layout_512",
                    "SMALL": "layout_557",
                },
            },
            "view_31": {
                "id": "view_31",
                "layout": {
                    "LARGE": "layout_670",
                    "MEDIUM": "layout_671",
                    "SMALL": "layout_672",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "var(--light)",
                "backgroundIMage": {},
            },
            "view_32": {
                "id": "view_32",
                "layout": {
                    "LARGE": "layout_673",
                    "MEDIUM": "layout_674",
                    "SMALL": "layout_675",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "var(--light)",
                "backgroundIMage": {},
            },
        },
        "sections": {
            "section_1": {
                "id": "section_1",
                "views": ["view_1", "view_31", "view_32"],
                "transition": {"type": "PUSH"},
            },
            "section_4": {
                "id": "section_4",
                "views": ["view_15", "view_16", "view_17", "view_18", "view_19"],
            },
            "section_5": {
                "id": "section_5",
                "views": [
                    "view_20",
                    "view_21",
                    "view_22",
                    "view_23",
                    "view_24",
                    "view_25",
                ],
            },
            "section_6": {"id": "section_6", "views": ["view_26", "view_27"]},
            "section_7": {
                "id": "section_7",
                "views": ["view_28", "view_29", "view_30"],
                "transition": {"type": "PUSH"},
            },
        },
        "dialogs": {
            "dialog_1": {
                "id": "dialog_1",
                "mode": "FIXED",
                "sizeMode": {
                    "LARGE": {"position": "CENTER", "width": "800px", "height": "500px"}
                },
                "closeWhenClickOverlay": True,
                "dialogBackground": {
                    "color": "var(--light)",
                    "fillType": "fit",
                    "image": {"url": ""},
                },
                "overlayColor": "var(--dark-900)",
                "layout": {"LARGE": "layout_109"},
                "index": 0,
                "isSplash": False,
                "aspectRatio": False,
                "effect": {
                    "type": "FADE_IN",
                    "option": {"direction": None, "configType": "GENTLE"},
                },
            }
        },
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [
            {"page_0": []},
            {"page_1": []},
            {"page_2": []},
            {"page_3": []},
            {"page_4": []},
        ],
        "timestamp": 1624314774146,
        "header": {
            "layout": {
                "LARGE": "layout_36",
                "MEDIUM": "layout_110",
                "SMALL": "layout_150",
            },
            "height": {"LARGE": "60px"},
            "backgroundColor": "",
            "sticky": False,
        },
        "footer": {
            "layout": {
                "LARGE": "layout_66",
                "MEDIUM": "layout_727",
                "SMALL": "layout_728",
            },
            "height": {"LARGE": "500px", "SMALL": "700px"},
            "backgroundColor": "",
        },
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "dart": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "FIT_WINDOW",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_0",
                    "SMALL": "layout_5",
                    "MEDIUM": "layout_10",
                },
                "isDefault": True,
                "isVisible": True,
                "header": False,
            }
        },
        "layouts": {
            "layout_0": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_1",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "lockLayout": True,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "80.0918%",
                            "bottom": "24.00px",
                            "width": "100%",
                            "height": "120.00px",
                        },
                        "widgetId": "widget_2",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "width": "STRETCH",
                                "bottom": False,
                                "top": True,
                                "left": False,
                                "right": False,
                            },
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "FLY_IN",
                                    "option": {"direction": "BOTTOM"},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_1": {"type": "FIXED", "order": [], "content": {}},
            "layout_2": {
                "type": "FIXED",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "72.4099%",
                            "right": "16.00px",
                            "top": "50%",
                            "bottom": "25.8475%",
                            "width": "30.00%",
                            "height": "44.0678%",
                        },
                        "widgetId": "widget_3",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20.00px",
                            "right": "58.6237%",
                            "top": 19,
                            "bottom": "47.8946%",
                            "width": "40.00%",
                            "height": "27.9661%",
                        },
                        "widgetId": "widget_7",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                                "width": "CUSTOM",
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20.00px",
                            "right": "58.5143%",
                            "top": 62,
                            "bottom": "22.0339%",
                            "width": "40.00%",
                            "height": "27.9661%",
                        },
                        "widgetId": "widget_9",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                                "width": "CUSTOM",
                            },
                        },
                    },
                },
            },
            "layout_3": {
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "NaNpx",
                            "height": "NaNpx",
                        },
                        "widgetId": "widget_4",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "NaNpx",
                            "height": "NaNpx",
                        },
                        "widgetId": "widget_5",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "NaNpx",
                            "height": "NaNpx",
                        },
                        "widgetId": "widget_6",
                    },
                },
            },
            "layout_4": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_6",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_5": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_1",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "lockLayout": True,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 53,
                            "right": "14.8394%",
                            "top": "87.8901%",
                            "bottom": 12,
                            "width": "70.3212%",
                            "height": "75px",
                        },
                        "widgetId": "widget_8",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "width": "CUSTOM",
                                "bottom": False,
                                "top": True,
                            },
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "FLY_IN",
                                    "option": {"direction": "BOTTOM"},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_6": {"type": "FIXED"},
            "layout_7": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.7963%",
                            "right": 0,
                            "top": 15,
                            "bottom": 0,
                            "width": "100.00%",
                            "height": "44.0678%",
                        },
                        "widgetId": "widget_3",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "AUTO",
                                "width": "CUSTOM",
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 16,
                            "right": "70.6395%",
                            "top": "50%",
                            "bottom": "34.7590%",
                            "width": "40.00%",
                            "height": "27.9661%",
                        },
                        "widgetId": "widget_7",
                        "isPending": True,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                                "width": "CUSTOM",
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_8": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "NaNpx",
                            "height": "NaNpx",
                        },
                        "widgetId": "widget_4",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "NaNpx",
                            "height": "NaNpx",
                        },
                        "widgetId": "widget_5",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "NaNpx",
                            "height": "NaNpx",
                        },
                        "widgetId": "widget_6",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_9": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_6",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_10": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_1",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "lockLayout": True,
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "80.0918%",
                            "bottom": "24.00px",
                            "width": "100%",
                            "height": "120.00px",
                        },
                        "widgetId": "widget_10",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "width": "STRETCH",
                                "bottom": False,
                                "top": True,
                                "left": False,
                                "right": False,
                            },
                            "effect": {
                                "SCROLL_IN": {
                                    "type": "FLY_IN",
                                    "option": {"direction": "BOTTOM"},
                                }
                            },
                            "oneByOneEffect": {
                                "SCROLL_IN": {
                                    "type": "FLOAT_IN",
                                    "option": {"direction": None},
                                }
                            },
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_11": {"type": "FIXED"},
            "layout_12": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "54.5339%",
                            "right": "17px",
                            "top": "50%",
                            "bottom": "31.8750%",
                            "width": "42.6328%",
                            "height": "36.4583%",
                        },
                        "widgetId": "widget_3",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "AUTO",
                                "width": "CUSTOM",
                                "top": False,
                                "bottom": True,
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "22px",
                            "right": "45.4622%",
                            "top": "19px",
                            "bottom": "48.7630%",
                            "width": "50.8711%",
                            "height": "35.0000%",
                        },
                        "widgetId": "widget_7",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                                "width": "CUSTOM",
                            },
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20.00px",
                            "right": "58.5143%",
                            "top": 62,
                            "bottom": "22.0339%",
                            "width": "40.00%",
                            "height": "27.9661%",
                        },
                        "widgetId": "widget_9",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "height": "AUTO",
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                                "width": "CUSTOM",
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_13": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "NaNpx",
                            "height": "NaNpx",
                        },
                        "widgetId": "widget_4",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "NaNpx",
                            "height": "NaNpx",
                        },
                        "widgetId": "widget_5",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "NaNpx",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "NaNpx",
                            "height": "NaNpx",
                        },
                        "widgetId": "widget_6",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_14": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "widgetId": "widget_6",
                        "bbox": {"left": 0, "top": 0, "bottom": 0, "right": 0},
                    }
                },
                "order": ["0"],
            },
            "layout_15": {"type": "FIXED"},
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "dart",
        "widgets": {
            "widget_1": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {"canZoom": True, "canHome": True, "canSearch": True},
                    "isUseCustomMapState": False,
                },
                "id": "widget_1",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_1",
                        "SMALL": "layout_6",
                        "MEDIUM": "layout_11",
                    }
                },
            },
            "widget_2": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "id": "widget_2",
                "layouts": {"DEFAULT": {"LARGE": "layout_2"}},
                "style": {
                    "background": {
                        "color": "var(--primary-500)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
            "widget_3": {
                "uri": "widgets/common/controller/",
                "version": "1.5.0",
                "config": {
                    "behavior": {
                        "onlyOpenOne": False,
                        "openStarts": [],
                        "displayType": "STACK",
                        "vertical": False,
                        "size": {},
                    },
                    "appearance": {
                        "space": 0,
                        "advanced": False,
                        "card": {
                            "showLabel": False,
                            "labelGrowth": 10,
                            "avatar": {
                                "type": "primary",
                                "size": "lg",
                                "shape": "circle",
                            },
                        },
                    },
                },
                "id": "widget_3",
                "layouts": {
                    "controller": {
                        "LARGE": "layout_3",
                        "SMALL": "layout_8",
                        "MEDIUM": "layout_13",
                    },
                    "_openwidget": {
                        "LARGE": "layout_4",
                        "SMALL": "layout_9",
                        "MEDIUM": "layout_14",
                    },
                },
            },
            "widget_4": {
                "uri": "widgets/arcgis/legend/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_4",
            },
            "widget_5": {
                "uri": "widgets/arcgis/map-layers/",
                "version": "1.5.0",
                "config": {"useMapWidget": True},
                "id": "widget_5",
            },
            "widget_6": {
                "uri": "widgets/common/share/",
                "version": "1.5.0",
                "config": {
                    "uiMode": "POPUP",
                    "popup": {
                        "icon": "",
                        "items": ["embed", "qrcode", "email", "facebook", "twitter"],
                        "tooltip": "",
                    },
                    "inline": {
                        "items": [
                            "facebook",
                            "twitter",
                            "embed",
                            "qrcode",
                            "email",
                            "sharelink",
                        ],
                        "design": {
                            "direction": "HORIZONTAL",
                            "hideLabel": False,
                            "btnRad": 0,
                            "btnColor": "",
                            "iconColor": "",
                            "size": "default",
                        },
                    },
                },
                "id": "widget_6",
            },
            "widget_7": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="font-size: 1.5rem; color: var(--light);">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_7",
            },
            "widget_8": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "id": "widget_8",
                "layouts": {"DEFAULT": {"SMALL": "layout_7"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
            },
            "widget_9": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><strong style="color: var(--light); font-size: 1rem;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_9",
            },
            "widget_10": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "id": "widget_10",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_12"}},
                "style": {
                    "background": {
                        "color": "var(--primary-500)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
            },
        },
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dialogs": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_0": []}],
        "timestamp": 1619663963914,
        "header": {"layout": {"LARGE": "layout_15"}, "height": {"LARGE": 75}},
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "pocket": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "FIT_WINDOW",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_0",
                    "SMALL": "layout_4",
                    "MEDIUM": "layout_8",
                },
                "isDefault": True,
                "isVisible": True,
                "header": False,
            }
        },
        "layouts": {
            "layout_0": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_1",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_1": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
            },
            "layout_2": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
            },
            "layout_3": {"type": "FIXED"},
            "layout_4": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_3",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_5": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_6": {"type": "FIXED"},
            "layout_7": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_8": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_4",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_9": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_10": {"type": "FIXED"},
            "layout_11": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_12": {"type": "FIXED"},
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "pocket",
        "widgets": {
            "widget_1": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "HORIZONTAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "300px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "RIGHT",
                        "offsetX": -15,
                        "offsetY": 0,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 15,
                        "height": 60,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 0 0 92px"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 0 0 92px"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_1",
                "layouts": {
                    "FIRST": {"LARGE": "layout_1"},
                    "SECOND": {"LARGE": "layout_2"},
                },
            },
            "widget_2": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": False,
                },
                "id": "widget_2",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_3",
                        "SMALL": "layout_6",
                        "MEDIUM": "layout_10",
                    }
                },
            },
            "widget_3": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "VERTICAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "300px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "DOWN",
                        "offsetX": 0,
                        "offsetY": -15,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 60,
                        "height": 15,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_3",
                "layouts": {
                    "FIRST": {"SMALL": "layout_5"},
                    "SECOND": {"SMALL": "layout_7"},
                },
            },
            "widget_4": {
                "uri": "widgets/layout/sidebar/",
                "version": "1.5.0",
                "config": {
                    "direction": "VERTICAL",
                    "collapseSide": "SECOND",
                    "overlay": False,
                    "size": "300px",
                    "divider": {
                        "visible": True,
                        "lineStyle": {
                            "type": "solid",
                            "color": "var(--light-500)",
                            "width": "1px",
                        },
                    },
                    "resizable": True,
                    "toggleBtn": {
                        "visible": False,
                        "icon": "DOWN",
                        "offsetX": 0,
                        "offsetY": -15,
                        "position": "CENTER",
                        "iconSize": 14,
                        "width": 60,
                        "height": 15,
                        "color": {
                            "normal": {
                                "icon": {"useTheme": False, "color": "#FFFFFF"},
                                "bg": {"useTheme": True, "color": "var(--primary)"},
                            },
                            "hover": {
                                "bg": {"useTheme": True, "color": "var(--primary-700)"}
                            },
                        },
                        "expandStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                        "collapseStyle": {"style": {"borderRadius": "92px 92px 0 0"}},
                    },
                    "defaultState": 1,
                },
                "id": "widget_4",
                "layouts": {
                    "FIRST": {"MEDIUM": "layout_9"},
                    "SECOND": {"MEDIUM": "layout_11"},
                },
            },
        },
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dialogs": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_0": []}],
        "timestamp": 1619659761502,
        "header": {"layout": {"LARGE": "layout_12"}, "height": {"LARGE": 75}},
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "quick navigation": {
        "pages": {
            "page_1": {
                "id": "page_1",
                "mode": "FIT_WINDOW",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_42",
                    "MEDIUM": "layout_46",
                    "SMALL": "layout_50",
                },
                "isDefault": True,
                "isVisible": True,
                "header": False,
            }
        },
        "layouts": {
            "layout_42": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "85.0000%",
                            "top": 0,
                            "bottom": 0,
                            "width": "15.00%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_33",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "15.0000%",
                            "right": "21.7031%",
                            "top": 0,
                            "bottom": 0,
                            "width": "65.00%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_34",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "2": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {
                            "left": "80.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "20.00%",
                            "height": "100.0000%",
                        },
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": True, "left": False},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_43": {
                "type": "FIXED",
                "order": ["3"],
                "content": {
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0%",
                            "right": "0.0%",
                            "top": "33.3%",
                            "bottom": "0.0%",
                            "width": "100.0%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_48",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "top": True,
                                "bottom": False,
                            }
                        },
                    }
                },
            },
            "layout_44": {"type": "FIXED"},
            "layout_45": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_37",
                        "setting": {
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                    }
                },
            },
            "layout_46": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "70.4106%",
                            "top": 0,
                            "bottom": 0,
                            "width": "29.5894%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_39",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "29.7265%",
                            "right": "32.9997%",
                            "top": 0,
                            "bottom": 0,
                            "width": "37.2738%",
                            "height": "100.0000%",
                        },
                        "widgetId": "widget_34",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "2": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {
                            "left": "67.0029%",
                            "right": "0px",
                            "top": 0,
                            "bottom": 0,
                            "width": "32.9945%",
                            "height": "100.0000%",
                        },
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": True, "left": False},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_47": {"type": "FIXED"},
            "layout_48": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10.00px",
                            "right": "10.00px",
                            "top": "10.00px",
                            "bottom": "94.5573%",
                            "width": "100%",
                            "height": "3.6667%",
                        },
                        "widgetId": "widget_35",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "height": "AUTO",
                                "right": False,
                                "left": False,
                                "width": "STRETCH",
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "0.0000%",
                            "top": "7.4292%",
                            "bottom": "90.9076%",
                            "width": "100.0000%",
                            "height": "5.5679%",
                        },
                        "widgetId": "widget_36",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 14,
                            "right": 14,
                            "top": "9.0924%",
                            "bottom": "73.5479%",
                            "width": "75.8938%",
                            "height": "7.1726%",
                        },
                        "widgetId": "widget_38",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_49": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "2px",
                            "top": "2px",
                            "bottom": "2px",
                            "width": "100.0000%",
                            "height": "99.5842%",
                        },
                        "widgetId": "widget_37",
                        "setting": {
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_50": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": "582px",
                            "width": "100.0000%",
                            "height": "19.1298%",
                        },
                        "widgetId": "widget_40",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": "86px",
                            "bottom": "270px",
                            "width": "100.0000%",
                            "height": "50.5556%",
                        },
                        "widgetId": "widget_34",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                    "2": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {
                            "left": 0,
                            "right": "3px",
                            "top": "450px",
                            "bottom": "-13px",
                            "width": "99.1059%",
                            "height": "39.2839%",
                        },
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "height": "CUSTOM",
                            },
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_51": {"type": "FIXED"},
            "layout_52": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10.00px",
                            "right": "10.00px",
                            "top": "10.00px",
                            "bottom": "94.5573%",
                            "width": "100%",
                            "height": "3.6667%",
                        },
                        "widgetId": "widget_35",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "height": "AUTO",
                                "right": False,
                                "left": False,
                                "width": "STRETCH",
                            },
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "0.0000%",
                            "top": "26.5224%",
                            "bottom": "61.6899%",
                            "width": "100.0000%",
                            "height": "5.5679%",
                        },
                        "widgetId": "widget_36",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 1,
                            "top": "39.0469%",
                            "bottom": "39.5879%",
                            "width": "99.7222%",
                            "height": "21.0550%",
                        },
                        "widgetId": "widget_38",
                        "isPending": True,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "12px",
                            "right": "16px",
                            "top": "35.25300000000003%",
                            "bottom": "43.38179999999997%",
                            "width": "92.1441%",
                            "height": "21.0550%",
                        },
                        "widgetId": "widget_47",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                        },
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_53": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": 0,
                            "top": "2px",
                            "bottom": "18px",
                            "width": "99.4394%",
                            "height": "92.9787%",
                        },
                        "widgetId": "widget_37",
                        "setting": {
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "CUSTOM",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_54": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_41",
                        "setting": {
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_55": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "2px",
                            "top": "2px",
                            "bottom": "2px",
                            "width": "100.0000%",
                            "height": "99.5842%",
                        },
                        "widgetId": "widget_42",
                        "setting": {
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_56": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": 0,
                            "top": "2px",
                            "bottom": "18px",
                            "width": "99.4394%",
                            "height": "92.9787%",
                        },
                        "widgetId": "widget_43",
                        "setting": {
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "CUSTOM",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_57": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_44",
                        "setting": {
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_58": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "2px",
                            "top": "2px",
                            "bottom": "2px",
                            "width": "100.0000%",
                            "height": "99.5842%",
                        },
                        "widgetId": "widget_45",
                        "setting": {
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "STRETCH",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_59": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2px",
                            "right": 0,
                            "top": "2px",
                            "bottom": "18px",
                            "width": "99.4394%",
                            "height": "92.9787%",
                        },
                        "widgetId": "widget_46",
                        "setting": {
                            "autoProps": {
                                "width": "STRETCH",
                                "height": "CUSTOM",
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_60": {"type": "FIXED"},
            "layout_61": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "199px", "height": "58px"},
                        "widgetId": "widget_35",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "216px", "height": "16px"},
                        "widgetId": "widget_36",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {"style": {}},
                        "bbox": {"width": "174px", "height": "167px"},
                        "widgetId": "widget_38",
                    },
                },
                "order": ["0", "1", "2"],
                "setting": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "quicknavigation",
        "widgets": {
            "widget_33": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_33",
                "layouts": {"DEFAULT": {"LARGE": "layout_43"}},
            },
            "widget_34": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": None,
                },
                "id": "widget_34",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_44",
                        "MEDIUM": "layout_47",
                        "SMALL": "layout_51",
                    }
                },
            },
            "widget_35": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_35",
            },
            "widget_36": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {"type": "Style0", "color": "#000", "size": "1px"},
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_36",
            },
            "widget_37": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059278519,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059284230,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap2.jpg",
                                "originalName": "widget_37-snap2",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap2.jpg",
                                "type": "IMAGE",
                                "size": 356127,
                                "created": 1600059289718,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -4819999.0638817595,
                                "ymin": -9757359.631022941,
                                "xmax": 16274174.757916152,
                                "ymax": 8538607.459311984,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 5727087.847017196,
                                    "y": -609376.0858554794,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": False,
                    "style": {"id": "default"},
                    "direction": "VERTICAL",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_37",
                "useMapWidgetIds": ["widget_34"],
            },
            "widget_38": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_2",
                        "views": ["view_11", "view_12", "view_13"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "underline",
                        "standard": {
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                            "gap": "20px",
                        },
                        "vertical": True,
                        "advanced": True,
                        "variant": {
                            "size": {
                                "fontSize": "0.875rem",
                                "lineHeight": 1.5,
                                "paddingX": "1rem",
                                "paddingY": "0.25rem",
                            },
                            "gutter": "0.25rem",
                            "icon": {"spacing": "0.5rem"},
                            "root": {
                                "border": {"width": 0},
                                "borderRadius": "0.125rem",
                            },
                            "item": {
                                "default": {
                                    "color": "#181818",
                                    "bg": "transparent",
                                    "border": {"color": "transparent", "width": "2px"},
                                    "borderRadius": "0.125rem",
                                },
                                "hover": {
                                    "color": "#076fe5",
                                    "border": {"color": "transparent", "width": "2px"},
                                    "borderRadius": "0.125rem",
                                },
                                "active": {
                                    "color": "#076fe5",
                                    "border": {"color": "#076fe5", "width": "2px"},
                                    "borderRadius": "0.125rem",
                                },
                                "disabled": {
                                    "color": "#989898",
                                    "shadow": "none",
                                    "borderRadius": "0.125rem",
                                },
                            },
                        },
                    },
                },
                "id": "widget_38",
            },
            "widget_39": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_39",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_48"}},
            },
            "widget_40": {
                "uri": "widgets/layout/fixed/",
                "version": "1.5.0",
                "config": None,
                "id": "widget_40",
                "layouts": {"DEFAULT": {"SMALL": "layout_52"}},
            },
            "widget_41": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059278519,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059284230,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap2.jpg",
                                "originalName": "widget_37-snap2",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap2.jpg",
                                "type": "IMAGE",
                                "size": 356127,
                                "created": 1600059289718,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -4819999.0638817595,
                                "ymin": -9757359.631022941,
                                "xmax": 16274174.757916152,
                                "ymax": 8538607.459311984,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 5727087.847017196,
                                    "y": -609376.0858554794,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": False,
                    "style": {"id": "default"},
                    "direction": "VERTICAL",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_41",
                "useMapWidgetIds": ["widget_34"],
            },
            "widget_42": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059278519,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark--2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059284230,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark--3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap2.jpg",
                                "originalName": "widget_37-snap2",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap2.jpg",
                                "type": "IMAGE",
                                "size": 356127,
                                "created": 1600059289718,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -4819999.0638817595,
                                "ymin": -9757359.631022941,
                                "xmax": 16274174.757916152,
                                "ymax": 8538607.459311984,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 5727087.847017196,
                                    "y": -609376.0858554794,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": False,
                    "style": {"id": "default"},
                    "direction": "VERTICAL",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_42",
                "useMapWidgetIds": ["widget_34"],
            },
            "widget_43": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059278519,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark--2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059284230,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark--3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap2.jpg",
                                "originalName": "widget_37-snap2",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap2.jpg",
                                "type": "IMAGE",
                                "size": 356127,
                                "created": 1600059289718,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -4819999.0638817595,
                                "ymin": -9757359.631022941,
                                "xmax": 16274174.757916152,
                                "ymax": 8538607.459311984,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 5727087.847017196,
                                    "y": -609376.0858554794,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": False,
                    "style": {"id": "default"},
                    "direction": "VERTICAL",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_43",
                "useMapWidgetIds": ["widget_34"],
            },
            "widget_44": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059278519,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059284230,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap2.jpg",
                                "originalName": "widget_37-snap2",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap2.jpg",
                                "type": "IMAGE",
                                "size": 356127,
                                "created": 1600059289718,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -4819999.0638817595,
                                "ymin": -9757359.631022941,
                                "xmax": 16274174.757916152,
                                "ymax": 8538607.459311984,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 5727087.847017196,
                                    "y": -609376.0858554794,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": False,
                    "style": {"id": "default"},
                    "direction": "VERTICAL",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_44",
                "useMapWidgetIds": ["widget_34"],
            },
            "widget_45": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059278519,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark--2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059284230,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark--3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap2.jpg",
                                "originalName": "widget_37-snap2",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap2.jpg",
                                "type": "IMAGE",
                                "size": 356127,
                                "created": 1600059289718,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -4819999.0638817595,
                                "ymin": -9757359.631022941,
                                "xmax": 16274174.757916152,
                                "ymax": 8538607.459311984,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 5727087.847017196,
                                    "y": -609376.0858554794,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": False,
                    "style": {"id": "default"},
                    "direction": "VERTICAL",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_45",
                "useMapWidgetIds": ["widget_34"],
            },
            "widget_46": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059278519,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark--2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap1.jpg",
                                "originalName": "widget_37-snap1",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap1.jpg",
                                "type": "IMAGE",
                                "size": 289775,
                                "created": 1600059284230,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -18144272.523465287,
                                "ymin": -8144844.082319291,
                                "xmax": 2949901.298332623,
                                "ymax": 10151123.008015634,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7597185.612566332,
                                    "y": 1003139.4628481712,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark--3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_37-snap2.jpg",
                                "originalName": "widget_37-snap2",
                                "url": "${appResourceUrl}/images/widget_37/widget_37-snap2.jpg",
                                "type": "IMAGE",
                                "size": 356127,
                                "created": 1600059289718,
                                "resourcesPrefix": "images/widget_37",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -4819999.0638817595,
                                "ymin": -9757359.631022941,
                                "xmax": 16274174.757916152,
                                "ymax": 8538607.459311984,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 5727087.847017196,
                                    "y": -609376.0858554794,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_37editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": False,
                    "style": {"id": "default"},
                    "direction": "VERTICAL",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_46",
                "useMapWidgetIds": ["widget_34"],
            },
            "widget_47": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_2",
                        "views": ["view_11", "view_12", "view_13"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "underline",
                        "standard": {
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_47",
            },
            "widget_48": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_48",
                "layouts": {"DEFAULT": {"LARGE": "layout_61"}},
            },
        },
        "widgetsManifest": {},
        "views": {
            "view_11": {
                "id": "view_11",
                "layout": {
                    "LARGE": "layout_45",
                    "MEDIUM": "layout_49",
                    "SMALL": "layout_53",
                },
            },
            "view_12": {
                "id": "view_12",
                "layout": {
                    "LARGE": "layout_54",
                    "MEDIUM": "layout_55",
                    "SMALL": "layout_56",
                },
            },
            "view_13": {
                "id": "view_13",
                "layout": {
                    "LARGE": "layout_57",
                    "MEDIUM": "layout_58",
                    "SMALL": "layout_59",
                },
            },
        },
        "sections": {
            "section_2": {
                "id": "section_2",
                "views": ["view_11", "view_12", "view_13"],
                "transition": {"type": "CUBE"},
            }
        },
        "dialogs": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_1": []}],
        "timestamp": 1620587786106,
        "header": {"layout": {"LARGE": "layout_60"}, "height": {"LARGE": 75}},
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "parallax": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {"LARGE": "layout_0", "SMALL": "layout_52"},
                "isDefault": True,
                "isVisible": True,
                "header": True,
                "footer": True,
            }
        },
        "layouts": {
            "layout_0": {
                "type": "FLOW",
                "content": {
                    "0": {"type": "SCREEN_GROUP", "screenGroupId": "screenGroup_1"},
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1280px",
                            "top": "0px",
                            "bottom": "320px",
                            "width": "300px",
                            "height": 320,
                        },
                        "widgetId": "widget_104",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-34px",
                            "right": "326px",
                            "top": "288px",
                            "bottom": "358px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_106",
                        "setting": {"heightMode": "auto"},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-71px",
                            "right": "289px",
                            "top": "387px",
                            "bottom": "457px",
                            "width": "360px",
                            "height": 128,
                        },
                        "widgetId": "widget_108",
                        "setting": {"heightMode": "auto"},
                    },
                },
                "order": ["1", "2", "3", "0"],
            },
            "layout_1": {
                "type": "FIXED",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "15px",
                            "top": "50%",
                            "width": "30.00px",
                            "height": 30,
                        },
                        "widgetId": "widget_1",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"top": False, "bottom": True},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "61px",
                            "top": 0,
                            "width": "295px",
                            "height": "100.0000%",
                            "right": "924px",
                            "bottom": 0,
                        },
                        "widgetId": "widget_2",
                        "setting": {"hCenter": False, "vCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 379,
                            "top": 0,
                            "width": "226px",
                            "height": "100.0000%",
                            "right": 675,
                            "bottom": 0,
                        },
                        "widgetId": "widget_3",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                    },
                },
            },
            "layout_3": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_7",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "100.00px", "height": "50px"},
                        "widgetId": "widget_9",
                        "setting": {
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": "350.00px"},
                        "widgetId": "widget_10",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_110",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": "350.00px"},
                        "widgetId": "widget_111",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["0", "2", "4", "3", "5"],
                "type": "COLUMN",
            },
            "layout_4": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_6",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
                "type": "FIXED",
            },
            "layout_5": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_84",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "100.00px", "height": "50px"},
                        "widgetId": "widget_85",
                        "setting": {
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 262},
                        "widgetId": "widget_86",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 262},
                        "widgetId": "widget_90",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 262},
                        "widgetId": "widget_149",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "type": "COLUMN",
            },
            "layout_6": {"content": {}, "order": [], "type": "FIXED"},
            "layout_7": {"content": {}, "order": [], "type": "COLUMN"},
            "layout_8": {
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_96",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
                "type": "FIXED",
            },
            "layout_9": {
                "type": "ROW",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "330.00px"},
                        "widgetId": "widget_11",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "330.00px"},
                        "widgetId": "widget_72",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "330.00px"},
                        "widgetId": "widget_78",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
            },
            "layout_25": {
                "type": "FIXED",
                "content": {
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_67",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_68",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_69",
                        "setting": {},
                    },
                },
                "order": ["4", "5", "6"],
            },
            "layout_26": {
                "type": "FIXED",
                "content": {
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_70",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_71",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["2", "3"],
            },
            "layout_27": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_73",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_74",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_75",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_28": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_76",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_77",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_29": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_79",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_80",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_81",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_30": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_82",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_83",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_31": {
                "type": "ROW",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "240.00px"},
                        "widgetId": "widget_87",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "240.00px"},
                        "widgetId": "widget_88",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "240.00px"},
                        "widgetId": "widget_89",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
            },
            "layout_32": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "240.00px"},
                        "widgetId": "widget_91",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "240.00px"},
                        "widgetId": "widget_92",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "240.00px"},
                        "widgetId": "widget_93",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_33": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "right": False,
                                "left": True,
                                "bottom": True,
                                "top": False,
                            },
                        },
                        "bbox": {
                            "left": "73.6291%",
                            "right": "2.00%",
                            "top": "50%",
                            "bottom": "40.00px",
                            "width": "300.00px",
                            "height": "400.00px",
                        },
                        "widgetId": "widget_154",
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_34": {
                "type": "FIXED",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "top": "60.8236%",
                            "width": 300,
                            "height": "28.3333%",
                        },
                        "widgetId": "widget_102",
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "top": "20px",
                            "width": "165.00px",
                            "height": "40px",
                            "right": 593,
                        },
                        "widgetId": "widget_103",
                        "setting": {
                            "hCenter": True,
                            "autoProps": {
                                "left": False,
                                "width": "AUTO",
                                "height": "AUTO",
                                "right": True,
                            },
                            "vCenter": False,
                        },
                        "isPending": False,
                    },
                },
            },
            "layout_36": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600.00px"},
                        "widgetId": "widget_105",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
            },
            "layout_37": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_107",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_38": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": "70px"},
                        "widgetId": "widget_109",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
            },
            "layout_39": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "330.00px"},
                        "widgetId": "widget_112",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "330.00px"},
                        "widgetId": "widget_118",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "330.00px"},
                        "widgetId": "widget_124",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_40": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_113",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_114",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_115",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_41": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_116",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_117",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_42": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_119",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_120",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_121",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_43": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_122",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_123",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_44": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_125",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_126",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_127",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_45": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_128",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_129",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_46": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_130",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "100.00px", "height": "50px"},
                        "widgetId": "widget_131",
                        "setting": {
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 262},
                        "widgetId": "widget_132",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 128},
                        "widgetId": "widget_141",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "100.00px", "height": "50px"},
                        "widgetId": "widget_142",
                        "setting": {
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 111},
                        "widgetId": "widget_143",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_144",
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "isPending": False,
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 262},
                        "widgetId": "widget_146",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "120.00px"},
                        "widgetId": "widget_153",
                    },
                },
                "order": ["0", "1", "2", "9", "10", "5", "6", "7", "8"],
            },
            "layout_47": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "240.00px"},
                        "widgetId": "widget_133",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "240.00px"},
                        "widgetId": "widget_145",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_49": {"type": "FIXED"},
            "layout_50": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "240.00px"},
                        "widgetId": "widget_147",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "240.00px"},
                        "widgetId": "widget_148",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_51": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "240.00px"},
                        "widgetId": "widget_150",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "240.00px"},
                        "widgetId": "widget_151",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "240.00px"},
                        "widgetId": "widget_152",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_52": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1280px",
                            "top": "0px",
                            "bottom": "320px",
                            "width": "300px",
                            "height": 320,
                        },
                        "widgetId": "widget_155",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-34px",
                            "right": "326px",
                            "top": "288px",
                            "bottom": "358px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_156",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-71px",
                            "right": "289px",
                            "top": "387px",
                            "bottom": "457px",
                            "width": "360px",
                            "height": 128,
                        },
                        "widgetId": "widget_157",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                    },
                    "3": {
                        "type": "SCREEN_GROUP",
                        "screenGroupId": "screenGroup_1",
                        "setting": {"heightMode": "auto"},
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_53": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_165",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_54": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_166",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_55": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_167",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_56": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_7",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "100.00px", "height": "50px"},
                        "widgetId": "widget_9",
                        "setting": {
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_110",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": "350.00px"},
                        "widgetId": "widget_158",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": "350.00px"},
                        "widgetId": "widget_159",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
            },
            "layout_57": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "330.00px"},
                        "widgetId": "widget_11",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "330.00px"},
                        "widgetId": "widget_72",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "330.00px"},
                        "widgetId": "widget_78",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_58": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_67",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_68",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_69",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_59": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_70",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_71",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_60": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_73",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_74",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_75",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_61": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_76",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_77",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_62": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_79",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_80",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_81",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_63": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_82",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_83",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_64": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "330.00px"},
                        "widgetId": "widget_112",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "330.00px"},
                        "widgetId": "widget_118",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "330.00px"},
                        "widgetId": "widget_124",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_65": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_113",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_114",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_115",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_66": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_116",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_117",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_67": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_119",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_120",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_121",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_68": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_122",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_123",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_69": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_125",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "63.9943%",
                            "bottom": "3.5652%",
                            "width": "90%",
                            "height": "32.4405%",
                        },
                        "widgetId": "widget_126",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_127",
                        "setting": {},
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_70": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "4.1824%",
                            "bottom": 341,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_128",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "15.8998%",
                            "bottom": "4.3682%",
                            "width": "90.00%",
                            "height": "79.7320%",
                        },
                        "widgetId": "widget_129",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                    },
                },
                "order": ["0", "1"],
            },
            "layout_71": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_6",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_72": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_84",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "100.00px", "height": "50px"},
                        "widgetId": "widget_85",
                        "setting": {
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 262},
                        "widgetId": "widget_160",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 262},
                        "widgetId": "widget_161",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 262},
                        "widgetId": "widget_162",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
            },
            "layout_73": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_87",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_88",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_89",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_74": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_91",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_92",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_93",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_75": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_150",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_151",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_152",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_76": {"type": "FIXED"},
            "layout_77": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_130",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "100.00px", "height": "50px"},
                        "widgetId": "widget_131",
                        "setting": {
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 262},
                        "widgetId": "widget_163",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 262},
                        "widgetId": "widget_164",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "120.00px"},
                        "widgetId": "widget_153",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 128},
                        "widgetId": "widget_141",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "100.00px", "height": "50px"},
                        "widgetId": "widget_142",
                        "setting": {
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                        "isPending": False,
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "762px", "height": 111},
                        "widgetId": "widget_143",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_144",
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "isPending": False,
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8"],
            },
            "layout_78": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_133",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_145",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_79": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_147",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "240.00px"},
                        "widgetId": "widget_148",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    },
                },
                "order": ["0", "1"],
            },
            "layout_80": {"type": "FIXED"},
            "layout_81": {"type": "COLUMN"},
            "layout_82": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": 0,
                            "top": 0,
                            "bottom": 0,
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_96",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "top": False,
                                "bottom": False,
                                "left": False,
                                "right": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_83": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "right": False,
                                "left": True,
                                "bottom": True,
                                "top": False,
                            },
                        },
                        "bbox": {
                            "left": "73.6291%",
                            "right": "2.00%",
                            "top": "50%",
                            "bottom": "40.00px",
                            "width": "300.00px",
                            "height": "400.00px",
                        },
                        "widgetId": "widget_154",
                        "isPending": False,
                    }
                },
                "order": ["0"],
            },
            "layout_84": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "240.00px"},
                        "widgetId": "widget_105",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_85": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_107",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
            "layout_86": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_109",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                    }
                },
                "order": ["0"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "parallax",
        "widgets": {
            "widget_1": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_1",
            },
            "widget_2": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px;">Here is the title</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_2",
            },
            "widget_3": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="line-height: 2.5;"><span style="font-size: 1rem;">Here is the subtitle</span></p>',
                    "style": {"verticalAlign": "center", "wrap": True},
                },
                "id": "widget_3",
            },
            "widget_6": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_6",
            },
            "widget_7": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 28px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_7",
            },
            "widget_9": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--primary)",
                        "size": "3px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_9",
            },
            "widget_10": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_10",
                "layouts": {"DEFAULT": {"LARGE": "layout_9"}},
            },
            "widget_11": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": False,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 355.515625,
                    "widgetHeight": 330,
                },
                "id": "widget_11",
                "layouts": {
                    "REGULAR": {"LARGE": "layout_25", "SMALL": "layout_58"},
                    "HOVER": {"LARGE": "layout_26", "SMALL": "layout_59"},
                },
            },
            "widget_67": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_67",
            },
            "widget_68": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_68",
            },
            "widget_69": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_69",
            },
            "widget_70": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_70",
            },
            "widget_71": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_71",
            },
            "widget_72": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": False,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 355.515625,
                    "widgetHeight": 330,
                },
                "id": "widget_72",
                "layouts": {
                    "REGULAR": {"LARGE": "layout_27", "SMALL": "layout_60"},
                    "HOVER": {"LARGE": "layout_28", "SMALL": "layout_61"},
                },
            },
            "widget_73": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_73",
            },
            "widget_74": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_74",
            },
            "widget_75": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_75",
            },
            "widget_76": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_76",
            },
            "widget_77": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_77",
            },
            "widget_78": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": False,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 355.515625,
                    "widgetHeight": 330,
                },
                "id": "widget_78",
                "layouts": {
                    "REGULAR": {"LARGE": "layout_29", "SMALL": "layout_62"},
                    "HOVER": {"LARGE": "layout_30", "SMALL": "layout_63"},
                },
            },
            "widget_79": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_79",
            },
            "widget_80": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_80",
            },
            "widget_81": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_81",
            },
            "widget_82": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_82",
            },
            "widget_83": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_83",
            },
            "widget_84": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 28px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_84",
            },
            "widget_85": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--primary)",
                        "size": "3px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_85",
            },
            "widget_86": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_86",
                "layouts": {"DEFAULT": {"LARGE": "layout_31"}},
            },
            "widget_87": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_87",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_88": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_88",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_89": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_89",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_90": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_90",
                "layouts": {"DEFAULT": {"LARGE": "layout_32"}},
            },
            "widget_91": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_91",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_92": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_92",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_93": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_93",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_96": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": False,
                        "canHome": False,
                        "canSearch": False,
                        "canNavigation": False,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": None,
                    "disableScroll": True,
                },
                "id": "widget_96",
                "layouts": {
                    "MapFixedLayout": {"LARGE": "layout_33", "SMALL": "layout_83"}
                },
            },
            "widget_102": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 13px;">Copyright info</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_102",
            },
            "widget_103": {
                "uri": "widgets/common/share/",
                "version": "1.5.0",
                "config": {
                    "uiMode": "INLINE",
                    "popup": {
                        "icon": {
                            "svg": '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16"><path d="M8 0v2H2v12h12V8h2v6a2 2 0 01-2 2H2a2 2 0 01-2-2V2a2 2 0 012-2h6zm2 0v2h2.5l-8 8L6 11.5l8-8V6h2V0h-6z" fill-rule="evenodd"></path></svg>',
                            "properties": {
                                "filename": None,
                                "originalName": None,
                                "path": None,
                                "color": "#585858",
                                "size": 24,
                                "inlineSvg": True,
                            },
                        },
                        "items": ["embed", "qrcode", "email", "facebook", "twitter"],
                        "tooltip": "",
                    },
                    "inline": {
                        "items": ["facebook", "twitter", "qrcode", "sharelink"],
                        "design": {
                            "direction": "HORIZONTAL",
                            "hideLabel": True,
                            "btnRad": "5px",
                            "btnColor": "rgba(0,0,0,0.4)",
                            "iconColor": "black",
                            "size": "lg",
                        },
                    },
                },
                "id": "widget_103",
            },
            "widget_104": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_104",
                "layouts": {"DEFAULT": {"LARGE": "layout_36"}},
            },
            "widget_105": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_105",
            },
            "widget_106": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_106",
                "layouts": {"DEFAULT": {"LARGE": "layout_37"}},
            },
            "widget_107": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 28px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_107",
            },
            "widget_108": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 40, 10], "unit": "px"}},
                },
                "id": "widget_108",
                "layouts": {"DEFAULT": {"LARGE": "layout_38"}},
            },
            "widget_109": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;">Lorem ipsum the donllor siiten amet, consectetur adipiscing elit. Aenean euismod bibendum laoreet. Proin gravida dolor sit amet lacus accumsan et viverra justo commodo. Cum sociis natoque in the penatibus et magnis dis parturient montes, nascetur ridiculus mus. Nam fermentum, nulla luctus pharetra vulputate, felis tellus mollis orci, sed rhoncus sapien nunc eget.<span style="font-size: 16px;">\ufeff</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_109",
            },
            "widget_110": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_110",
            },
            "widget_111": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_111",
                "layouts": {"DEFAULT": {"LARGE": "layout_39"}},
            },
            "widget_112": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": False,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 355.515625,
                    "widgetHeight": 330,
                },
                "id": "widget_112",
                "layouts": {
                    "REGULAR": {"LARGE": "layout_40", "SMALL": "layout_65"},
                    "HOVER": {"LARGE": "layout_41", "SMALL": "layout_66"},
                },
            },
            "widget_113": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_113",
            },
            "widget_114": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_114",
            },
            "widget_115": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_115",
            },
            "widget_116": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_116",
            },
            "widget_117": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_117",
            },
            "widget_118": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": False,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 355.515625,
                    "widgetHeight": 330,
                },
                "id": "widget_118",
                "layouts": {
                    "REGULAR": {"LARGE": "layout_42", "SMALL": "layout_67"},
                    "HOVER": {"LARGE": "layout_43", "SMALL": "layout_68"},
                },
            },
            "widget_119": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_119",
            },
            "widget_120": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_120",
            },
            "widget_121": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_121",
            },
            "widget_122": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_122",
            },
            "widget_123": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_123",
            },
            "widget_124": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": False,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                    "widgetWidth": 355.515625,
                    "widgetHeight": 330,
                },
                "id": "widget_124",
                "layouts": {
                    "REGULAR": {"LARGE": "layout_44", "SMALL": "layout_69"},
                    "HOVER": {"LARGE": "layout_45", "SMALL": "layout_70"},
                },
            },
            "widget_125": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_125",
            },
            "widget_126": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_126",
            },
            "widget_127": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_127",
            },
            "widget_128": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_128",
            },
            "widget_129": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_129",
            },
            "widget_130": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 28px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_130",
            },
            "widget_131": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--primary)",
                        "size": "3px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_131",
            },
            "widget_132": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_132",
                "layouts": {"DEFAULT": {"LARGE": "layout_47"}},
            },
            "widget_133": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_133",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_141": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 28px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_141",
            },
            "widget_142": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--primary)",
                        "size": "3px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_142",
            },
            "widget_143": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;">Lorem ipsum the donllor siiten amet, consectetur adipiscing elit. Aenean euismod bibendum laoreet. Proin gravida dolor sit amet lacus accumsan et viverra justo commodo. Cum sociis natoque in the penatibus et magnis dis parturient montes, nascetur ridiculus mus. Nam fermentum, nulla luctus pharetra vulputate, felis tellus mollis orci, sed rhoncus sapien nunc eget.<span style="font-size: 16px;">\ufeff</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_143",
            },
            "widget_144": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "themeStyle": {"quickStyleType": "primary"},
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#fff", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                    },
                },
                "id": "widget_144",
            },
            "widget_145": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_145",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_146": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_146",
                "layouts": {"DEFAULT": {"LARGE": "layout_50"}},
            },
            "widget_147": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_147",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_148": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_148",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_149": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_149",
                "layouts": {"DEFAULT": {"LARGE": "layout_51"}},
            },
            "widget_150": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_150",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_151": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_151",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_152": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_152",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_153": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {"type": "Style0", "color": "", "size": "2px"},
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_153",
            },
            "widget_154": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "SLIDE2",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_154-snap0.jpg",
                                "originalName": "widget_154-snap0",
                                "url": "${appResourceUrl}/images/widget_154/widget_154-snap0.jpg",
                                "type": "IMAGE",
                                "size": 222875,
                                "created": 1600060459576,
                                "resourcesPrefix": "images/widget_154",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -15937917.702014264,
                                "ymin": -6438922.485051181,
                                "xmax": 5156256.119783648,
                                "ymax": 6828099.640346764,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -5390830.791115307,
                                    "y": 194588.5776477917,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_154editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "description": "Lorem ipsum the donllor siiten amet, consectetur adipiscing elit. Aenean euismod bibendum laoreet. ",
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_154-snap1.jpg",
                                "originalName": "widget_154-snap1",
                                "url": "${appResourceUrl}/images/widget_154/widget_154-snap1.jpg",
                                "type": "IMAGE",
                                "size": 161725,
                                "created": 1600060463338,
                                "resourcesPrefix": "images/widget_154",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -22503801.042256042,
                                "ymin": -12620607.688665751,
                                "xmax": -1409627.2204581294,
                                "ymax": 646414.4367321935,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -11956714.131357085,
                                    "y": -5987096.625966779,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_154editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "description": "Lorem ipsum the donllor siiten amet, consectetur adipiscing elit. Aenean euismod bibendum laoreet. ",
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_154-snap2.jpg",
                                "originalName": "widget_154-snap2",
                                "url": "${appResourceUrl}/images/widget_154/widget_154-snap2.jpg",
                                "type": "IMAGE",
                                "size": 208453,
                                "created": 1600060467874,
                                "resourcesPrefix": "images/widget_154",
                                "fileFormat": "image/jpeg",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -23412581.08136347,
                                "ymin": 4187803.2148644933,
                                "xmax": -2318407.25956556,
                                "ymax": 17454825.34026244,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 73957190.948944,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -12865494.170464516,
                                    "y": 10821314.277563466,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_154editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "description": "Lorem ipsum the donllor siiten amet, consectetur adipiscing elit. Aenean euismod bibendum laoreet. ",
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": True,
                    "autoInterval": 3,
                    "autoLoopAllow": False,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "runtimeUuid": "exb-/builder",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_154",
                "layouts": {},
                "useMapWidgetIds": ["widget_96"],
            },
            "widget_155": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_155",
                "layouts": {"DEFAULT": {"SMALL": "layout_53"}},
            },
            "widget_156": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [40, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_156",
                "layouts": {"DEFAULT": {"SMALL": "layout_54"}},
            },
            "widget_157": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 40, 10], "unit": "px"}},
                },
                "id": "widget_157",
                "layouts": {"DEFAULT": {"SMALL": "layout_55"}},
            },
            "widget_158": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_158",
                "layouts": {"DEFAULT": {"SMALL": "layout_57"}},
            },
            "widget_159": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_159",
                "layouts": {"DEFAULT": {"SMALL": "layout_64"}},
            },
            "widget_160": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_160",
                "layouts": {"DEFAULT": {"SMALL": "layout_73"}},
            },
            "widget_161": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_161",
                "layouts": {"DEFAULT": {"SMALL": "layout_74"}},
            },
            "widget_162": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_162",
                "layouts": {"DEFAULT": {"SMALL": "layout_75"}},
            },
            "widget_163": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_163",
                "layouts": {"DEFAULT": {"SMALL": "layout_78"}},
            },
            "widget_164": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 30,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_164",
                "layouts": {"DEFAULT": {"SMALL": "layout_79"}},
            },
            "widget_165": {
                "uri": "widgets/layout/column/",
                "id": "widget_165",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_84"}},
            },
            "widget_166": {
                "uri": "widgets/layout/column/",
                "id": "widget_166",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_85"}},
            },
            "widget_167": {
                "uri": "widgets/layout/column/",
                "id": "widget_167",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_86"}},
            },
        },
        "widgetsManifest": {},
        "views": {},
        "sections": {},
        "dialogs": {},
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_0": []}],
        "header": {
            "layout": {"LARGE": "layout_1"},
            "height": {"LARGE": "50"},
            "sticky": True,
        },
        "screenGroups": {
            "screenGroup_1": {
                "transitionType": "COVER",
                "panelTransitionType": "PUSH",
                "triggerType": "TOP",
                "stretchPanel": True,
                "verticalSpace": "30%",
                "screens": ["screen_1", "screen_2", "screen_4", "screen_3"],
                "id": "screenGroup_1",
                "label": "Screen group 1",
            }
        },
        "screens": {
            "screen_1": {
                "panel": {
                    "side": "center",
                    "size": "90%",
                    "offset": "0",
                    "overlay": True,
                    "layout": {"LARGE": "layout_3", "SMALL": "layout_56"},
                },
                "main": {"layout": {"LARGE": "layout_4", "SMALL": "layout_71"}},
                "id": "screen_1",
                "label": "Screen1",
            },
            "screen_2": {
                "panel": {
                    "side": "center",
                    "size": "90%",
                    "offset": "0",
                    "overlay": True,
                    "layout": {"LARGE": "layout_5", "SMALL": "layout_72"},
                },
                "main": {"layout": {"LARGE": "layout_6", "SMALL": "layout_76"}},
                "id": "screen_2",
                "label": "Screen2",
            },
            "screen_3": {
                "panel": {
                    "side": "center",
                    "size": "64px",
                    "offset": "0",
                    "overlay": True,
                    "layout": {"LARGE": "layout_7", "SMALL": "layout_81"},
                },
                "main": {"layout": {"LARGE": "layout_8", "SMALL": "layout_82"}},
                "id": "screen_3",
                "label": "Screen4",
            },
            "screen_4": {
                "panel": {
                    "side": "center",
                    "size": "90%",
                    "offset": "0",
                    "overlay": True,
                    "layout": {"LARGE": "layout_46", "SMALL": "layout_77"},
                },
                "main": {"layout": {"LARGE": "layout_49", "SMALL": "layout_80"}},
                "id": "screen_4",
                "label": "Screen3",
            },
        },
        "footer": {
            "layout": {"LARGE": "layout_34"},
            "height": {"LARGE": "120", "SMALL": "120"},
        },
        "timestamp": 1619660381789,
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "dash": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_0",
                    "MEDIUM": "layout_187",
                    "SMALL": "layout_288",
                },
                "isDefault": True,
                "isVisible": True,
                "header": True,
                "footer": True,
                "icon": {
                    "svg": '<svg viewBox="0 0 10 12" xmlns="http://www.w3.org/2000/svg"><path d="M1 0h8a1 1 0 011 1v10a1 1 0 01-1 1H1a1 1 0 01-1-1V1a1 1 0 011-1zm0 1v10h8V1H1zm2 2h4a.5.5 0 010 1H3a.5.5 0 010-1zm0 2.5h4a.5.5 0 010 1H3a.5.5 0 010-1z" fill="#000" fill-rule="nonzero"></path></svg>',
                    "properties": {
                        "filename": None,
                        "originalName": None,
                        "path": None,
                        "color": "#fff",
                        "size": 16,
                        "inlineSvg": True,
                    },
                },
                "bodyBackgroundColor": "var(--light-200)",
            }
        },
        "layouts": {
            "layout_0": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "723px",
                            "right": "1073px",
                            "top": "-78px",
                            "bottom": "322px",
                            "width": "350px",
                            "height": "60px",
                        },
                        "widgetId": "widget_8",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_0",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "145px",
                            "right": "505px",
                            "top": "79px",
                            "bottom": "149px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_10",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "105px",
                            "right": "465px",
                            "top": "156px",
                            "bottom": "226px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_12",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-71px",
                            "right": "729px",
                            "top": "88px",
                            "bottom": "488px",
                            "width": "800px",
                            "height": 420,
                        },
                        "widgetId": "widget_14",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "179px",
                            "right": "629px",
                            "top": "413px",
                            "bottom": "713px",
                            "width": "450px",
                            "height": "500px",
                        },
                        "widgetId": "widget_40",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1168px",
                            "bottom": "1233px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_49",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1292px",
                            "bottom": "1335px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_51",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "126px",
                            "right": "426px",
                            "top": "980px",
                            "bottom": "1580px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_53",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1908px",
                            "bottom": "1972px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_206",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1958px",
                            "bottom": "2011px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_208",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "316px",
                            "right": "616px",
                            "top": "2260px",
                            "bottom": "2860px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_213",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3004px",
                            "bottom": "3068px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_220",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3044px",
                            "bottom": "3097px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_222",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "15": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "131px",
                            "right": "731px",
                            "top": "2900px",
                            "bottom": "3300px",
                            "width": "600px",
                            "height": 420,
                        },
                        "widgetId": "widget_224",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_0",
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "78px",
                            "right": "378px",
                            "top": "3248px",
                            "bottom": "3848px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_226",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                },
                "order": [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "12",
                    "13",
                    "14",
                    "15",
                    "16",
                ],
                "parent": {"type": "pages", "id": "page_0"},
            },
            "layout_1": {
                "type": "FIXED",
                "order": ["0", "1", "2", "3", "4"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50.00px",
                            "right": "94.8%",
                            "top": "40.00px",
                            "bottom": "54.3%",
                            "width": "40px",
                            "height": "40px",
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_1",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "100px",
                            "right": "64.1%",
                            "top": "40px",
                            "bottom": "49.6%",
                            "width": "360px",
                            "height": "41px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.5%",
                            "right": "50.00px",
                            "top": "30.00px",
                            "bottom": "-52.8%",
                            "width": "50.00px",
                            "height": "50.00px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "65.7%",
                            "right": "140.00px",
                            "top": "40.00px",
                            "bottom": "50.3%",
                            "width": "32.00px",
                            "height": "32.00px",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.6%",
                            "right": "200.00px",
                            "top": "45.00px",
                            "bottom": "24.1%",
                            "width": "24.00px",
                            "height": "24.00px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                },
                "parent": {"type": "header", "id": "header"},
            },
            "layout_2": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "42.6%",
                            "top": "50%",
                            "bottom": "29.6%",
                            "width": "14.9%",
                            "height": "37.9%",
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "parent": "layout_2",
                    }
                },
                "parent": {"type": "footer", "id": "footer"},
            },
            "layout_4": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "400px"},
                        "widgetId": "widget_9",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_4",
                    }
                },
                "parent": {"type": "widget", "id": "widget_8"},
            },
            "layout_5": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_11",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_5",
                    }
                },
                "parent": {"type": "widget", "id": "widget_10"},
            },
            "layout_6": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "70px",
                        },
                        "widgetId": "widget_13",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_6",
                    }
                },
                "parent": {"type": "widget", "id": "widget_12"},
            },
            "layout_7": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 8, "height": "400px"},
                        "widgetId": "widget_15",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_7",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_36",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_7",
                    },
                },
                "parent": {"type": "widget", "id": "widget_14"},
            },
            "layout_8": {
                "type": "ROW",
                "order": ["0", "4", "1", "3", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "600px"},
                        "widgetId": "widget_16",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 1, "height": "50px"},
                        "widgetId": "widget_22",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 3, "height": "600px"},
                        "widgetId": "widget_23",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 1, "height": "50px"},
                        "widgetId": "widget_29",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 3, "height": "600px"},
                        "widgetId": "widget_30",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                    },
                },
                "parent": {"type": "widget", "id": "widget_15"},
            },
            "layout_9": {
                "type": "COLUMN",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_17",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_9",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_18",
                        "parent": "layout_9",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_19",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_9",
                    },
                },
                "setting": {"space": 20},
                "parent": {"type": "widget", "id": "widget_16"},
            },
            "layout_10": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_20",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_10",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "70px"},
                        "widgetId": "widget_21",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_10",
                    },
                },
                "parent": {"type": "widget", "id": "widget_19"},
            },
            "layout_11": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_24",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_11",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_25",
                        "parent": "layout_11",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_26",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_11",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_23"},
            },
            "layout_12": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_27",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_12",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "70px"},
                        "widgetId": "widget_28",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_12",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_26"},
            },
            "layout_13": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_31",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_13",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_32",
                        "parent": "layout_13",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_33",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                        "parent": "layout_13",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_30"},
            },
            "layout_14": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_34",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_14",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "70px"},
                        "widgetId": "widget_35",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_14",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_33"},
            },
            "layout_15": {
                "type": "COLUMN",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_37",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_15",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_38",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                        "parent": "layout_15",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "200.00px", "height": "50.00px"},
                        "widgetId": "widget_39",
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "isPending": False,
                        "parent": "layout_15",
                    },
                },
                "setting": {
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                    "space": 20,
                    "justifyContent": "center",
                },
                "parent": {"type": "widget", "id": "widget_36"},
            },
            "layout_16": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 9, "height": "460.00px"},
                        "widgetId": "widget_41",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_16",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "600px"},
                        "widgetId": "widget_42",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_16",
                    },
                },
                "parent": {"type": "widget", "id": "widget_40"},
            },
            "layout_17": {
                "type": "COLUMN",
                "setting": {
                    "justifyContent": "center",
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                    "space": 20,
                },
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "40.00px", "height": "40.00px"},
                        "widgetId": "widget_43",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_17",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_44",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_17",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_45",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_17",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_46",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_17",
                    },
                },
                "parent": {"type": "widget", "id": "widget_42"},
            },
            "layout_18": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 10, "height": "70px"},
                        "widgetId": "widget_47",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_18",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": "10px", "width": "2px", "height": "300px"},
                        "widgetId": "widget_48",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_18",
                    },
                },
                "parent": {"type": "widget", "id": "widget_46"},
            },
            "layout_19": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_50",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_19",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_49"},
            },
            "layout_20": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "70px",
                        },
                        "widgetId": "widget_52",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_20",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_51"},
            },
            "layout_21": {
                "type": "ROW",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "600px"},
                        "widgetId": "widget_54",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_21",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "600px"},
                        "widgetId": "widget_79",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_21",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "391px"},
                        "widgetId": "widget_90",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_21",
                    },
                },
                "parent": {"type": "widget", "id": "widget_53"},
            },
            "layout_22": {
                "type": "COLUMN",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_55",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_22",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_74",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_22",
                    },
                },
                "setting": {"space": 40},
                "parent": {"type": "widget", "id": "widget_54"},
            },
            "layout_24": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_55"},
            },
            "layout_26": {
                "type": "FIXED",
                "content": {
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_67",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_68",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_69",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_73",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                },
                "order": ["4", "5", "6", "10"],
                "parent": {"type": "widget", "id": "widget_55"},
            },
            "layout_27": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_75",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_76",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_77",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_78",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_74"},
            },
            "layout_28": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_74"},
            },
            "layout_29": {
                "type": "COLUMN",
                "setting": {"space": 40},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_80",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_29",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_85",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_29",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_79"},
            },
            "layout_30": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_81",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_82",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_83",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_84",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_80"},
            },
            "layout_31": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_80"},
            },
            "layout_32": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_86",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_87",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_88",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_89",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_85"},
            },
            "layout_33": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_85"},
            },
            "layout_58": {
                "type": "FIXED",
                "content": {
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "59.6642%",
                            "bottom": 121,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_197",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "69.7132%",
                            "bottom": "22.4142%",
                            "width": "90.00%",
                            "height": "7.8727%",
                        },
                        "widgetId": "widget_198",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "55.36%",
                        },
                        "widgetId": "widget_199",
                        "setting": {},
                        "parent": "layout_58",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "vCenter": False, "hCenter": False},
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "75.9082%",
                            "bottom": "16.8080%",
                            "width": "90.00%",
                            "height": "7.2838%",
                        },
                        "widgetId": "widget_200",
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_201",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "4.9648%",
                            "top": "87.2095%",
                            "bottom": "4.7503%",
                            "width": "40.00%",
                            "height": 32,
                        },
                        "widgetId": "widget_202",
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "height": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_58",
                    },
                },
                "order": ["4", "5", "6", "7", "9", "10"],
                "parent": {"type": "widget", "id": "widget_90"},
            },
            "layout_59": {
                "type": "FIXED",
                "content": {},
                "order": [],
                "parent": {"type": "widget", "id": "widget_90"},
            },
            "layout_60": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_207",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_60",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_206"},
            },
            "layout_61": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "70px",
                        },
                        "widgetId": "widget_209",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_61",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_208"},
            },
            "layout_63": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_215",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_63",
                    }
                },
                "parent": {"type": "view", "id": "view_1"},
            },
            "layout_65": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600px"},
                        "widgetId": "widget_214",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_65",
                    }
                },
                "parent": {"type": "widget", "id": "widget_213"},
            },
            "layout_66": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "SECTION",
                        "setting": {"style": {}, "autoProps": {"height": "CUSTOM"}},
                        "bbox": {"width": "1860px", "height": "800.00px"},
                        "sectionId": "section_1",
                        "isPending": False,
                        "parent": "layout_66",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                        "bbox": {"width": "305px", "height": "43px"},
                        "widgetId": "widget_212",
                        "parent": "layout_66",
                    },
                },
                "order": ["1", "0"],
                "setting": {
                    "space": 30,
                    "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                },
                "parent": {"type": "widget", "id": "widget_214"},
            },
            "layout_67": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "400px"},
                        "widgetId": "widget_216",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_67",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "600px"},
                        "widgetId": "widget_217",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_67",
                    },
                },
                "parent": {"type": "widget", "id": "widget_215"},
            },
            "layout_68": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_216"},
            },
            "layout_69": {
                "type": "COLUMN",
                "setting": {"space": 40, "justifyContent": "flex-start"},
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "735px", "height": "300.00px"},
                        "widgetId": "widget_218",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_69",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_219",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "parent": "layout_69",
                    },
                },
                "parent": {"type": "widget", "id": "widget_217"},
            },
            "layout_70": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_221",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_70",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_220"},
            },
            "layout_71": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "70px",
                        },
                        "widgetId": "widget_223",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_71",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_222"},
            },
            "layout_72": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_225",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_72",
                    }
                },
                "parent": {"type": "widget", "id": "widget_224"},
            },
            "layout_73": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "600px"},
                        "widgetId": "widget_227",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_73",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "520.00px"},
                        "widgetId": "widget_238",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_73",
                    },
                },
                "parent": {"type": "widget", "id": "widget_226"},
            },
            "layout_74": {
                "type": "COLUMN",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_228",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_74",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_229",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_74",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_232",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_74",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_235",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_74",
                    },
                },
                "setting": {
                    "padding": {"number": [30, 20, 30, 20], "unit": "px"},
                    "space": 20,
                },
                "parent": {"type": "widget", "id": "widget_227"},
            },
            "layout_75": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_230",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_75",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "70px"},
                        "widgetId": "widget_231",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_75",
                    },
                },
                "parent": {"type": "widget", "id": "widget_229"},
            },
            "layout_76": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_233",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_76",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "70px"},
                        "widgetId": "widget_234",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_76",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_232"},
            },
            "layout_77": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_236",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_77",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "70px"},
                        "widgetId": "widget_237",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_77",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_235"},
            },
            "layout_78": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_239",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_78",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_2"},
            },
            "layout_79": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "400px"},
                        "widgetId": "widget_240",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_79",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "600px"},
                        "widgetId": "widget_241",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_79",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_239"},
            },
            "layout_80": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_240"},
            },
            "layout_81": {
                "type": "COLUMN",
                "setting": {"space": 40, "justifyContent": "flex-start"},
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_243",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "isPending": False,
                        "parent": "layout_81",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "620px", "height": "275px"},
                        "widgetId": "widget_244",
                        "parent": "layout_81",
                    },
                },
                "order": ["1", "2"],
                "parent": {"type": "widget", "id": "widget_241"},
            },
            "layout_83": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_83",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_83",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_83",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_84": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_89": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_89",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_89",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_89",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_90": {
                "type": "FIXED",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "6.7%",
                            "top": "11.8%",
                            "width": "254px",
                            "height": "30px",
                            "right": "82px",
                            "bottom": "75.7%",
                        },
                        "widgetId": "widget_259",
                        "isPending": False,
                        "parent": "layout_90",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": False,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "6.8%",
                            "top": "30.5%",
                            "width": "86.3%",
                            "height": "32.5%",
                            "right": "6.9%",
                            "bottom": "89px",
                        },
                        "widgetId": "widget_260",
                        "isPending": False,
                        "parent": "layout_90",
                    },
                },
                "order": ["1", "2"],
                "setting": {"lockDescendants": True},
                "parent": {"type": "dialog", "id": "dialog_1"},
            },
            "layout_91": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                                "width": "CUSTOM",
                            },
                            "lockParent": False,
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "bbox": {
                            "left": "6.00%",
                            "top": "11.43%",
                            "width": "36.00px",
                            "height": "36.00px",
                            "right": 305,
                            "bottom": "65.8960%",
                        },
                        "widgetId": "widget_261",
                        "isPending": False,
                        "parent": "layout_91",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "20.00%",
                            "top": "12.00%",
                            "width": "254px",
                            "height": "30px",
                            "right": 30,
                            "bottom": "61.6541%",
                        },
                        "widgetId": "widget_262",
                        "isPending": False,
                        "parent": "layout_91",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": False,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "20.00%",
                            "top": "33.3214%",
                            "width": "70.8333%",
                            "height": "45.7143%",
                            "right": "8.61%",
                            "bottom": "37px",
                        },
                        "widgetId": "widget_263",
                        "isPending": False,
                        "parent": "layout_91",
                    },
                },
                "order": ["0", "1", "2"],
                "setting": {"lockDescendants": True},
                "parent": {"type": "dialog", "id": "dialog_2"},
            },
            "layout_186": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20.00px",
                            "right": "94.8%",
                            "top": "40.00px",
                            "bottom": "54.3%",
                            "width": "40px",
                            "height": "40px",
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_186",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "70.00px",
                            "right": "64.1%",
                            "top": "40px",
                            "bottom": "49.6%",
                            "width": "360px",
                            "height": "41px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.5%",
                            "right": "20.00px",
                            "top": "30.00px",
                            "bottom": "-52.8%",
                            "width": "50.00px",
                            "height": "50.00px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "65.7%",
                            "right": "80.00px",
                            "top": "40.00px",
                            "bottom": "50.3%",
                            "width": "32.00px",
                            "height": "32.00px",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.6%",
                            "right": "130.00px",
                            "top": "45.00px",
                            "bottom": "24.1%",
                            "width": "24.00px",
                            "height": "24.00px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "header", "id": "header"},
            },
            "layout_187": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "723px",
                            "right": "1073px",
                            "top": "-78px",
                            "bottom": "322px",
                            "width": "350px",
                            "height": "60px",
                        },
                        "widgetId": "widget_432",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_187",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "145px",
                            "right": "505px",
                            "top": "79px",
                            "bottom": "149px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_433",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_187",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "105px",
                            "right": "465px",
                            "top": "156px",
                            "bottom": "226px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_434",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_187",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-71px",
                            "right": "729px",
                            "top": "88px",
                            "bottom": "488px",
                            "width": "800px",
                            "height": "360px",
                        },
                        "widgetId": "widget_435",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_187",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "179px",
                            "right": "629px",
                            "top": "413px",
                            "bottom": "713px",
                            "width": "450px",
                            "height": "400px",
                        },
                        "widgetId": "widget_444",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_187",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1168px",
                            "bottom": "1233px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_447",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1292px",
                            "bottom": "1335px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_448",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "126px",
                            "right": "426px",
                            "top": "980px",
                            "bottom": "1580px",
                            "width": "300px",
                            "height": 386,
                        },
                        "widgetId": "widget_449",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_187",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1908px",
                            "bottom": "1972px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_452",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1958px",
                            "bottom": "2011px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_453",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "316px",
                            "right": "616px",
                            "top": "2260px",
                            "bottom": "2860px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_454",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_187",
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3004px",
                            "bottom": "3068px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_460",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3044px",
                            "bottom": "3097px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_461",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "131px",
                            "right": "731px",
                            "top": "2900px",
                            "bottom": "3300px",
                            "width": "600px",
                            "height": 420,
                        },
                        "widgetId": "widget_462",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_187",
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "78px",
                            "right": "378px",
                            "top": "3248px",
                            "bottom": "3848px",
                            "width": "300px",
                            "height": "400px",
                        },
                        "widgetId": "widget_463",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_187",
                    },
                },
                "order": [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                ],
                "parent": {"type": "pages", "id": "page_0"},
            },
            "layout_188": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "400px"},
                        "widgetId": "widget_9",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_188",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_432"},
            },
            "layout_189": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_11",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_189",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_433"},
            },
            "layout_190": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_13",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_190",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_434"},
            },
            "layout_191": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 8, "height": "400px"},
                        "widgetId": "widget_436",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_191",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_443",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_191",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_435"},
            },
            "layout_192": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "600px"},
                        "widgetId": "widget_437",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_192",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 3, "height": "600px"},
                        "widgetId": "widget_439",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_192",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 1, "height": "50px"},
                        "widgetId": "widget_22",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_192",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 1, "height": "50px"},
                        "widgetId": "widget_29",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_192",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 3, "height": "600px"},
                        "widgetId": "widget_441",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_192",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_436"},
            },
            "layout_193": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_17",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_193",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_18",
                        "parent": "layout_193",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_438",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_193",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_437"},
            },
            "layout_194": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_20",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_194",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_21",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_194",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_438"},
            },
            "layout_195": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_31",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_195",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_32",
                        "parent": "layout_195",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_440",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "isPending": False,
                        "parent": "layout_195",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_439"},
            },
            "layout_196": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_34",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_196",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_35",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_196",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_440"},
            },
            "layout_197": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_24",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_197",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_25",
                        "parent": "layout_197",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_442",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_197",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_441"},
            },
            "layout_198": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_27",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_198",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_28",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_198",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_442"},
            },
            "layout_199": {
                "type": "COLUMN",
                "setting": {
                    "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                    "space": 20,
                    "justifyContent": "center",
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_37",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_199",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_38",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                        "parent": "layout_199",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "140.00px", "height": "40.00px"},
                        "widgetId": "widget_39",
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "isPending": False,
                        "parent": "layout_199",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_443"},
            },
            "layout_200": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 9, "height": "300px"},
                        "widgetId": "widget_41",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_200",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "600px"},
                        "widgetId": "widget_445",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_200",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_444"},
            },
            "layout_201": {
                "type": "COLUMN",
                "setting": {
                    "justifyContent": "center",
                    "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                    "space": 20,
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "40.00px", "height": "40.00px"},
                        "widgetId": "widget_43",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_201",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "60.00px"},
                        "widgetId": "widget_44",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_201",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "40.00px"},
                        "widgetId": "widget_45",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_201",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_446",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_201",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_445"},
            },
            "layout_202": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_47",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_202",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": "10px", "width": "2px", "height": "300px"},
                        "widgetId": "widget_48",
                        "setting": {"heightMode": "fit"},
                        "isPending": True,
                        "parent": "layout_202",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_446"},
            },
            "layout_203": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_50",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_203",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_447"},
            },
            "layout_204": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_52",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_204",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_448"},
            },
            "layout_205": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "360.00px"},
                        "widgetId": "widget_450",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_205",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "360.00px"},
                        "widgetId": "widget_451",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_205",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "360.00px"},
                        "widgetId": "widget_90",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_205",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_449"},
            },
            "layout_206": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "59.6642%",
                            "bottom": 121,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_197",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_206",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "69.7132%",
                            "bottom": "22.4142%",
                            "width": "90.00%",
                            "height": "7.8727%",
                        },
                        "widgetId": "widget_198",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_206",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "55.36%",
                        },
                        "widgetId": "widget_199",
                        "setting": {},
                        "parent": "layout_206",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "vCenter": False, "hCenter": False},
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "75.9082%",
                            "bottom": "16.8080%",
                            "width": "90.00%",
                            "height": "7.2838%",
                        },
                        "widgetId": "widget_200",
                        "isPending": False,
                        "parent": "layout_206",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_201",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_206",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "4.9648%",
                            "top": "87.2095%",
                            "bottom": "4.7503%",
                            "width": "40.00%",
                            "height": 32,
                        },
                        "widgetId": "widget_202",
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "height": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_206",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_90"},
            },
            "layout_207": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_90"},
            },
            "layout_208": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "170.00px"},
                        "widgetId": "widget_55",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_208",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "170.00px"},
                        "widgetId": "widget_74",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_208",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_450"},
            },
            "layout_209": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_67",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_209",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "38.6%",
                            "right": "0.8%",
                            "top": "39.3%",
                            "bottom": "6.00%",
                            "width": "60.6%",
                            "height": "90px",
                        },
                        "widgetId": "widget_68",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                                "top": True,
                                "bottom": False,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_209",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_69",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_209",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_73",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": True,
                        "parent": "layout_209",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_55"},
            },
            "layout_210": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_55"},
            },
            "layout_211": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_75",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_211",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "6.00%",
                            "width": "60.6343%",
                            "height": "90.00px",
                        },
                        "widgetId": "widget_76",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_211",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_77",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_211",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_78",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": True,
                        "parent": "layout_211",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_74"},
            },
            "layout_212": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_74"},
            },
            "layout_213": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "170.00px"},
                        "widgetId": "widget_80",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_213",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "170.00px"},
                        "widgetId": "widget_85",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_213",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_451"},
            },
            "layout_214": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_81",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_214",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.2%",
                            "right": "0.2%",
                            "top": "35.8%",
                            "bottom": "6.00%",
                            "width": "60.6%",
                            "height": "90.00px",
                        },
                        "widgetId": "widget_82",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                                "top": True,
                                "bottom": False,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_214",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_83",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_214",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_84",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": True,
                        "parent": "layout_214",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_80"},
            },
            "layout_215": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_80"},
            },
            "layout_216": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_86",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_216",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "6.00%",
                            "width": "60.6343%",
                            "height": "90.00px",
                        },
                        "widgetId": "widget_87",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_216",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_88",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_216",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_89",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": True,
                        "parent": "layout_216",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_85"},
            },
            "layout_217": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_85"},
            },
            "layout_218": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_207",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_218",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_452"},
            },
            "layout_219": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_209",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_219",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_453"},
            },
            "layout_220": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600px"},
                        "widgetId": "widget_455",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_220",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_454"},
            },
            "layout_221": {
                "type": "COLUMN",
                "setting": {
                    "space": 30,
                    "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                        "bbox": {"width": "305px", "height": "43px"},
                        "widgetId": "widget_212",
                        "parent": "layout_221",
                    },
                    "1": {
                        "type": "SECTION",
                        "setting": {"style": {}, "autoProps": {"height": "CUSTOM"}},
                        "bbox": {"width": "1860px", "height": "800.00px"},
                        "sectionId": "section_1",
                        "isPending": False,
                        "parent": "layout_221",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_455"},
            },
            "layout_222": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_456",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_222",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_1"},
            },
            "layout_223": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "400px"},
                        "widgetId": "widget_216",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_223",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "600px"},
                        "widgetId": "widget_457",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_223",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_456"},
            },
            "layout_224": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_216"},
            },
            "layout_225": {
                "type": "COLUMN",
                "setting": {"space": 20, "justifyContent": "flex-start"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "735px", "height": "300.00px"},
                        "widgetId": "widget_218",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_225",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_219",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "parent": "layout_225",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_457"},
            },
            "layout_226": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_458",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_226",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_2"},
            },
            "layout_227": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "400px"},
                        "widgetId": "widget_240",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_227",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "600px"},
                        "widgetId": "widget_459",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_227",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_458"},
            },
            "layout_228": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_240"},
            },
            "layout_229": {
                "type": "COLUMN",
                "setting": {"space": 20, "justifyContent": "flex-start"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_243",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "isPending": False,
                        "parent": "layout_229",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "620px", "height": "275px"},
                        "widgetId": "widget_244",
                        "parent": "layout_229",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_459"},
            },
            "layout_230": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_230",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_230",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_230",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_231": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_231",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_231",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_231",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_232": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_233": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_221",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_233",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_460"},
            },
            "layout_234": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_223",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_234",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_461"},
            },
            "layout_235": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_225",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_235",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_462"},
            },
            "layout_236": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "600px"},
                        "widgetId": "widget_464",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_236",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "300px"},
                        "widgetId": "widget_238",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_236",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_463"},
            },
            "layout_237": {
                "type": "COLUMN",
                "setting": {
                    "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                    "space": 10,
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "60.00px"},
                        "widgetId": "widget_228",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_237",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_465",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_237",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_466",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_237",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_467",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_237",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_464"},
            },
            "layout_238": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_230",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_238",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_231",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_238",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_465"},
            },
            "layout_239": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_233",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_239",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_234",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_239",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_466"},
            },
            "layout_240": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_236",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_240",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_237",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_240",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_467"},
            },
            "layout_241": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10.00px",
                            "right": "94.8%",
                            "top": "30.00px",
                            "bottom": "54.3%",
                            "width": "40px",
                            "height": "40px",
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_241",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50.00px",
                            "right": "64.1%",
                            "top": "30.00px",
                            "bottom": "49.6%",
                            "width": "240.00px",
                            "height": "41px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_241",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.5%",
                            "right": "10.00px",
                            "top": "20.00px",
                            "bottom": "-52.8%",
                            "width": "50.00px",
                            "height": "50.00px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_241",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "85.8%",
                            "right": "19px",
                            "top": "87px",
                            "bottom": "25.9%",
                            "width": "32px",
                            "height": "32px",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_241",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.8%",
                            "right": "70px",
                            "top": "95px",
                            "bottom": "29.09999999999999%",
                            "width": "24px",
                            "height": "24px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_241",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "header", "id": "header"},
            },
            "layout_286": {
                "type": "FIXED",
                "parent": {"type": "footer", "id": "footer"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "42.6%",
                            "top": "50%",
                            "bottom": "29.6%",
                            "width": "90.00%",
                            "height": "37.9%",
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "parent": "layout_286",
                    }
                },
                "order": ["0"],
            },
            "layout_287": {
                "type": "FIXED",
                "parent": {"type": "footer", "id": "footer"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "42.6%",
                            "top": "50%",
                            "bottom": "29.6%",
                            "width": "100.00%",
                            "height": "37.9%",
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "parent": "layout_287",
                    }
                },
                "order": ["0"],
            },
            "layout_288": {
                "type": "FLOW",
                "parent": {"type": "pages", "id": "page_0"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "723px",
                            "right": "1073px",
                            "top": "-78px",
                            "bottom": "322px",
                            "width": "350px",
                            "height": "60px",
                        },
                        "widgetId": "widget_468",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "145px",
                            "right": "505px",
                            "top": "79px",
                            "bottom": "149px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_469",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "105px",
                            "right": "465px",
                            "top": "156px",
                            "bottom": "226px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_470",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-71px",
                            "right": "729px",
                            "top": "88px",
                            "bottom": "488px",
                            "width": "800px",
                            "height": 420,
                        },
                        "widgetId": "widget_471",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "179px",
                            "right": "629px",
                            "top": "413px",
                            "bottom": "713px",
                            "width": "450px",
                            "height": "500px",
                        },
                        "widgetId": "widget_480",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1168px",
                            "bottom": "1233px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_483",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1292px",
                            "bottom": "1335px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_484",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "126px",
                            "right": "426px",
                            "top": "980px",
                            "bottom": "1580px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_485",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1908px",
                            "bottom": "1972px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_488",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1958px",
                            "bottom": "2011px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_489",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "316px",
                            "right": "616px",
                            "top": "2260px",
                            "bottom": "2860px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_490",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3004px",
                            "bottom": "3068px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_496",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3044px",
                            "bottom": "3097px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_497",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "131px",
                            "right": "731px",
                            "top": "2900px",
                            "bottom": "3300px",
                            "width": "600px",
                            "height": 420,
                        },
                        "widgetId": "widget_498",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "78px",
                            "right": "378px",
                            "top": "3248px",
                            "bottom": "3848px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_499",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                },
                "order": [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                ],
            },
            "layout_289": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_8"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_504",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_290": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_10"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_505",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_291": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_12"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_506",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_292": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_14"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_507",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_293": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_15"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600px"},
                        "widgetId": "widget_473",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "600px"},
                        "widgetId": "widget_475",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "50px"},
                        "widgetId": "widget_22",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                        "isPending": True,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 36, "width": 12, "height": "50px"},
                        "widgetId": "widget_29",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                        "isPending": True,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"left": 48, "width": 12, "height": "600px"},
                        "widgetId": "widget_477",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
            },
            "layout_294": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "parent": {"type": "widget", "id": "widget_16"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_17",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_9",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_18",
                        "parent": "layout_9",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_474",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_9",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_295": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_19"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_508",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_296": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "parent": {"type": "widget", "id": "widget_30"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_31",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_13",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_32",
                        "parent": "layout_13",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_476",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "isPending": False,
                        "parent": "layout_13",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_297": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_33"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_509",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_298": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "parent": {"type": "widget", "id": "widget_23"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_24",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_11",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_25",
                        "parent": "layout_11",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_478",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_11",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_299": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_26"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_510",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_300": {
                "type": "COLUMN",
                "setting": {
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                    "space": 20,
                    "justifyContent": "center",
                },
                "parent": {"type": "widget", "id": "widget_36"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_37",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_15",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_38",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                        "parent": "layout_15",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "200.00px", "height": "50.00px"},
                        "widgetId": "widget_39",
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "isPending": False,
                        "parent": "layout_15",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_301": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_40"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_511",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_302": {
                "type": "COLUMN",
                "setting": {
                    "justifyContent": "center",
                    "padding": {"number": [40, 20, 40, 20], "unit": "px"},
                    "space": 40,
                },
                "parent": {"type": "widget", "id": "widget_42"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "40.00px", "height": "40.00px"},
                        "widgetId": "widget_43",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_17",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_44",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_17",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_45",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_17",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_482",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_17",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_303": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_46"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_512",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_304": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_49"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_513",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_305": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_51"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_514",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_306": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_53"},
                "content": {
                    "3": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_515",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["3"],
            },
            "layout_307": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_90"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "59.6642%",
                            "bottom": 121,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_197",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "69.7132%",
                            "bottom": "22.4142%",
                            "width": "90.00%",
                            "height": "7.8727%",
                        },
                        "widgetId": "widget_198",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "55.36%",
                        },
                        "widgetId": "widget_199",
                        "setting": {},
                        "parent": "layout_58",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "vCenter": False, "hCenter": False},
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "75.9082%",
                            "bottom": "16.8080%",
                            "width": "90.00%",
                            "height": "7.2838%",
                        },
                        "widgetId": "widget_200",
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_201",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "4.9648%",
                            "top": "87.2095%",
                            "bottom": "4.7503%",
                            "width": "40.00%",
                            "height": 32,
                        },
                        "widgetId": "widget_202",
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "height": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_58",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
            },
            "layout_308": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_90"},
            },
            "layout_309": {
                "type": "COLUMN",
                "setting": {"space": 40},
                "parent": {"type": "widget", "id": "widget_54"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_55",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_22",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_74",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_22",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_310": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_55"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_67",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_68",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_69",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_73",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_311": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_55"},
            },
            "layout_312": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_74"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_75",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_76",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_77",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_78",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_313": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_74"},
            },
            "layout_314": {
                "type": "COLUMN",
                "setting": {"space": 40},
                "parent": {"type": "widget", "id": "widget_79"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_80",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_29",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_85",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_29",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_315": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_80"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_81",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_82",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_83",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_84",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_316": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_80"},
            },
            "layout_317": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_85"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_86",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_87",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_88",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_89",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_318": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_85"},
            },
            "layout_319": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_206"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_516",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_320": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_208"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_517",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_321": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_213"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_518",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_322": {
                "type": "COLUMN",
                "setting": {
                    "space": 30,
                    "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                },
                "parent": {"type": "widget", "id": "widget_214"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                        "bbox": {"width": "305px", "height": "43px"},
                        "widgetId": "widget_212",
                        "parent": "layout_66",
                    },
                    "1": {
                        "type": "SECTION",
                        "setting": {"style": {}, "autoProps": {"height": "CUSTOM"}},
                        "bbox": {"width": "1860px", "height": "800.00px"},
                        "sectionId": "section_1",
                        "isPending": False,
                        "parent": "layout_66",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_323": {
                "type": "FIXED",
                "parent": {"type": "view", "id": "view_1"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_492",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_63",
                    }
                },
                "order": ["0"],
            },
            "layout_324": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_215"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_216",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_67",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "600px"},
                        "widgetId": "widget_493",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_67",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_325": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_216"},
            },
            "layout_326": {
                "type": "COLUMN",
                "setting": {"space": 40, "justifyContent": "flex-start"},
                "parent": {"type": "widget", "id": "widget_217"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "735px", "height": "300.00px"},
                        "widgetId": "widget_218",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_69",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_219",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "parent": "layout_69",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_327": {
                "type": "FIXED",
                "parent": {"type": "view", "id": "view_2"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_494",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_78",
                    }
                },
                "order": ["0"],
            },
            "layout_328": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_239"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_240",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_79",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "600px"},
                        "widgetId": "widget_495",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_79",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_329": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_240"},
            },
            "layout_330": {
                "type": "COLUMN",
                "setting": {"space": 40, "justifyContent": "flex-start"},
                "parent": {"type": "widget", "id": "widget_241"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_243",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "isPending": False,
                        "parent": "layout_81",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "620px", "height": "275px"},
                        "widgetId": "widget_244",
                        "parent": "layout_81",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_331": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_244"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_89",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_89",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_89",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_332": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_244"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_83",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_83",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_83",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_333": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_334": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_220"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_519",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_335": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_222"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_520",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_336": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_224"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_521",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_337": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_226"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_522",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_338": {
                "type": "COLUMN",
                "setting": {
                    "padding": {"number": [30, 20, 30, 20], "unit": "px"},
                    "space": 20,
                },
                "parent": {"type": "widget", "id": "widget_227"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_228",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_74",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_501",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_74",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_502",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_74",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_503",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_74",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_339": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_229"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_523",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_340": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_232"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_524",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_341": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_235"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_525",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_342": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "400px"},
                        "widgetId": "widget_9",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_4",
                    }
                },
                "order": ["0"],
            },
            "layout_343": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_11",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_5",
                    }
                },
                "order": ["0"],
            },
            "layout_344": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_13",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_6",
                    }
                },
                "order": ["0"],
            },
            "layout_345": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_472",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_7",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_479",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_7",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_346": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_20",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_10",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_21",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_10",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_347": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_34",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_14",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_35",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_14",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_348": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_27",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_12",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_28",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_12",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_349": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_41",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_16",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_481",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_16",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_350": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_47",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_18",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "300px"},
                        "widgetId": "widget_48",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_18",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_351": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_50",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_19",
                    }
                },
                "order": ["0"],
            },
            "layout_352": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_52",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_20",
                    }
                },
                "order": ["0"],
            },
            "layout_353": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_486",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_21",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_487",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_21",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "391px"},
                        "widgetId": "widget_90",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_21",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_354": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_207",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_60",
                    }
                },
                "order": ["0"],
            },
            "layout_355": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_209",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_61",
                    }
                },
                "order": ["0"],
            },
            "layout_356": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_491",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_65",
                    }
                },
                "order": ["0"],
            },
            "layout_357": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_221",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_70",
                    }
                },
                "order": ["0"],
            },
            "layout_358": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_223",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_71",
                    }
                },
                "order": ["0"],
            },
            "layout_359": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "400px"},
                        "widgetId": "widget_225",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_72",
                    }
                },
                "order": ["0"],
            },
            "layout_360": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_500",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_73",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "300px"},
                        "widgetId": "widget_238",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_73",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_361": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_230",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_75",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_231",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_75",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_362": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_233",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_76",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_234",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_76",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_363": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_236",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_77",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_237",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_77",
                    },
                },
                "order": ["0", "1"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "dash",
        "widgets": {
            "widget_1": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_1",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "0"}],
                },
            },
            "widget_2": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px; color: var(--light);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_2",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "1"}],
                },
            },
            "widget_3": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {"linkType": "DIALOG", "value": "dialog_2"},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_3",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "2"}],
                },
            },
            "widget_4": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 22 28"><path d="M22 19.375c0 .562-.25 1.656-.484 2.172-.328.766-1.203 1.266-1.906 1.656-.922.5-1.859.797-2.906.797-1.453 0-2.766-.594-4.094-1.078a14.627 14.627 0 01-2.734-1.297c-2.656-1.641-5.859-4.844-7.5-7.5a14.627 14.627 0 01-1.297-2.734C.595 10.063.001 8.75.001 7.297c0-1.047.297-1.984.797-2.906.391-.703.891-1.578 1.656-1.906.516-.234 1.609-.484 2.172-.484.109 0 .219 0 .328.047.328.109.672.875.828 1.188.5.891.984 1.797 1.5 2.672.25.406.719.906.719 1.391 0 .953-2.828 2.344-2.828 3.187 0 .422.391.969.609 1.344 1.578 2.844 3.547 4.813 6.391 6.391.375.219.922.609 1.344.609.844 0 2.234-2.828 3.187-2.828.484 0 .984.469 1.391.719.875.516 1.781 1 2.672 1.5.313.156 1.078.5 1.188.828.047.109.047.219.047.328z"></path></svg>',
                            "position": "LEFT",
                        },
                        "linkParam": {"linkType": "DIALOG", "value": "dialog_1"},
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "#c5c5c5",
                                    "width": "0px",
                                },
                                "text": {"color": "#181818", "size": "14px"},
                                "borderRadius": {"number": [0], "unit": "px"},
                                "iconProps": {"color": "var(--light)", "size": 28},
                            },
                            "hover": {
                                "background": {"color": "", "fillType": "fill"},
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 28},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_4",
                "useDataSources": [],
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "3"}],
                },
            },
            "widget_5": {
                "uri": "widgets/common/share/",
                "version": "1.5.0",
                "config": {
                    "uiMode": "POPUP",
                    "popup": {
                        "icon": {
                            "svg": '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16"><path d="M8 0v2H2v12h12V8h2v6a2 2 0 01-2 2H2a2 2 0 01-2-2V2a2 2 0 012-2h6zm2 0v2h2.5l-8 8L6 11.5l8-8V6h2V0h-6z" fill-rule="evenodd"></path></svg>',
                            "properties": {
                                "filename": None,
                                "originalName": None,
                                "path": None,
                                "color": "var(--light)",
                                "size": 24,
                                "inlineSvg": True,
                            },
                        },
                        "items": [
                            "embed",
                            "qrcode",
                            "email",
                            "facebook",
                            "twitter",
                            "pinterest",
                            "linkedin",
                        ],
                        "tooltip": "",
                    },
                    "inline": {
                        "items": [
                            "facebook",
                            "twitter",
                            "pinterest",
                            "linkedin",
                            "embed",
                            "qrcode",
                            "email",
                            "sharelink",
                        ],
                        "design": {
                            "direction": "HORIZONTAL",
                            "hideLabel": False,
                            "btnRad": 0,
                            "btnColor": "",
                            "iconColor": "",
                            "size": "default",
                        },
                    },
                },
                "id": "widget_5",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "4"}],
                },
            },
            "widget_7": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="color: var(--dark-800); font-size: 16px;">Powered by Esri</span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_7",
                "parent": {
                    "LARGE": [{"layoutId": "layout_2", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_286", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_287", "layoutItemId": "0"}],
                },
            },
            "widget_8": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 0, 50], "unit": "px"}},
                },
                "id": "widget_8",
                "layouts": {"DEFAULT": {"LARGE": "layout_4"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "0"}]},
            },
            "widget_9": {
                "uri": "widgets/common/filter/",
                "version": "1.5.0",
                "config": {
                    "groupByLayer": False,
                    "logicalOperator": "AND",
                    "custom": False,
                    "resetAll": False,
                    "turnOffAll": False,
                    "arrangeType": "BLOCK",
                    "triggerType": "TOGGLE",
                    "omitInternalStyle": False,
                    "filterItems": [],
                },
                "id": "widget_9",
                "style": {
                    "background": {
                        "color": "",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_4", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_188", "layoutItemId": "0"}],
                },
            },
            "widget_10": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 0, 50], "unit": "px"}},
                },
                "id": "widget_10",
                "layouts": {"DEFAULT": {"LARGE": "layout_5"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "1"}]},
            },
            "widget_11": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark); font-size: 24px;">\ufeff\ufeffTitle</span></p>',
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_11",
                "parent": {
                    "LARGE": [{"layoutId": "layout_5", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_189", "layoutItemId": "0"}],
                },
            },
            "widget_12": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_12",
                "layouts": {"DEFAULT": {"LARGE": "layout_6"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "2"}]},
            },
            "widget_13": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-300);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_13",
                "parent": {
                    "LARGE": [{"layoutId": "layout_6", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_190", "layoutItemId": "0"}],
                },
            },
            "widget_14": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_14",
                "layouts": {"DEFAULT": {"LARGE": "layout_7"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "3"}]},
            },
            "widget_15": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {"padding": {"number": [20, 0, 20, 40], "unit": "px"}},
                },
                "id": "widget_15",
                "layouts": {"DEFAULT": {"LARGE": "layout_8"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_7", "layoutItemId": "0"}]},
            },
            "widget_16": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_16",
                "layouts": {"DEFAULT": {"LARGE": "layout_9"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "0"}]},
            },
            "widget_17": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-100);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_17",
                "parent": {
                    "LARGE": [{"layoutId": "layout_9", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_193", "layoutItemId": "0"}],
                },
            },
            "widget_18": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-800); font-size: 42px;">\ufeffNumber</span></p>',
                    "placeholder": '<p><span style="font-size: 42px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_18",
                "parent": {
                    "LARGE": [{"layoutId": "layout_9", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_193", "layoutItemId": "1"}],
                },
            },
            "widget_19": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_19",
                "layouts": {"DEFAULT": {"LARGE": "layout_10"}},
                "parent": {"LARGE": [{"layoutId": "layout_9", "layoutItemId": "2"}]},
            },
            "widget_20": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_20",
                "parent": {
                    "LARGE": [{"layoutId": "layout_10", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_194", "layoutItemId": "0"}],
                },
            },
            "widget_21": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-400);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_21",
                "parent": {
                    "LARGE": [{"layoutId": "layout_10", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_194", "layoutItemId": "1"}],
                },
            },
            "widget_22": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Vertical",
                    "strokeStyle": {
                        "type": "Style1",
                        "color": "var(--light-400)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_22",
                "parent": {
                    "LARGE": [{"layoutId": "layout_8", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_192", "layoutItemId": "2"}],
                },
            },
            "widget_23": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_23",
                "layouts": {"DEFAULT": {"LARGE": "layout_11"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "2"}]},
            },
            "widget_24": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-100);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_24",
                "parent": {
                    "LARGE": [{"layoutId": "layout_11", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_197", "layoutItemId": "0"}],
                },
            },
            "widget_25": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-800); font-size: 42px;">\ufeffNumber</span></p>',
                    "placeholder": '<p><span style="font-size: 42px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_25",
                "parent": {
                    "LARGE": [{"layoutId": "layout_11", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_197", "layoutItemId": "1"}],
                },
            },
            "widget_26": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_26",
                "layouts": {"DEFAULT": {"LARGE": "layout_12"}},
                "parent": {"LARGE": [{"layoutId": "layout_11", "layoutItemId": "2"}]},
            },
            "widget_27": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_27",
                "parent": {
                    "LARGE": [{"layoutId": "layout_12", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_198", "layoutItemId": "0"}],
                },
            },
            "widget_28": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-400);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_28",
                "parent": {
                    "LARGE": [{"layoutId": "layout_12", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_198", "layoutItemId": "1"}],
                },
            },
            "widget_29": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Vertical",
                    "strokeStyle": {
                        "type": "Style1",
                        "color": "var(--light-400)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_29",
                "parent": {
                    "LARGE": [{"layoutId": "layout_8", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_192", "layoutItemId": "3"}],
                },
            },
            "widget_30": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_30",
                "layouts": {"DEFAULT": {"LARGE": "layout_13"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "4"}]},
            },
            "widget_31": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-100);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_31",
                "parent": {
                    "LARGE": [{"layoutId": "layout_13", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_195", "layoutItemId": "0"}],
                },
            },
            "widget_32": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-800); font-size: 42px;">\ufeffNumber</span></p>',
                    "placeholder": '<p><span style="font-size: 42px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_32",
                "parent": {
                    "LARGE": [{"layoutId": "layout_13", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_195", "layoutItemId": "1"}],
                },
            },
            "widget_33": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_33",
                "layouts": {"DEFAULT": {"LARGE": "layout_14"}},
                "parent": {"LARGE": [{"layoutId": "layout_13", "layoutItemId": "2"}]},
            },
            "widget_34": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_34",
                "parent": {
                    "LARGE": [{"layoutId": "layout_14", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_196", "layoutItemId": "0"}],
                },
            },
            "widget_35": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-400);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_35",
                "parent": {
                    "LARGE": [{"layoutId": "layout_14", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_196", "layoutItemId": "1"}],
                },
            },
            "widget_36": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_36",
                "layouts": {"DEFAULT": {"LARGE": "layout_15"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_7", "layoutItemId": "1"}]},
            },
            "widget_37": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="color: var(--light); font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_37",
                "parent": {
                    "LARGE": [{"layoutId": "layout_15", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_199", "layoutItemId": "0"}],
                },
            },
            "widget_38": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 48px; color: var(--light);">\ufeffNumber</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="color: var(--light); font-size: 48px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                    "tooltip": {"name": "", "parts": [{"type": "STRING", "exp": '""'}]},
                },
                "id": "widget_38",
                "parent": {
                    "LARGE": [{"layoutId": "layout_15", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_199", "layoutItemId": "1"}],
                },
            },
            "widget_39": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(240,240,240,0.2)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "#c5c5c5",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--light)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "rgba(240,240,240,0.3)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                    },
                },
                "id": "widget_39",
                "parent": {
                    "LARGE": [{"layoutId": "layout_15", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_199", "layoutItemId": "2"}],
                },
            },
            "widget_40": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_40",
                "layouts": {"DEFAULT": {"LARGE": "layout_16"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "4"}]},
            },
            "widget_41": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_41",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_16", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_200", "layoutItemId": "0"}],
                },
            },
            "widget_42": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_42",
                "layouts": {"DEFAULT": {"LARGE": "layout_17"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_16", "layoutItemId": "1"}]},
            },
            "widget_43": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_43",
                "parent": {
                    "LARGE": [{"layoutId": "layout_17", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_201", "layoutItemId": "0"}],
                },
            },
            "widget_44": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 32px;">\ufeffNumber</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 32px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_44",
                "parent": {
                    "LARGE": [{"layoutId": "layout_17", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_201", "layoutItemId": "1"}],
                },
            },
            "widget_45": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px; color: var(--dark-300);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_45",
                "parent": {
                    "LARGE": [{"layoutId": "layout_17", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_201", "layoutItemId": "2"}],
                },
            },
            "widget_46": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_46",
                "layouts": {"DEFAULT": {"LARGE": "layout_18"}},
                "style": {
                    "background": {
                        "color": "rgba(7,111,229,0.1)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
                "parent": {"LARGE": [{"layoutId": "layout_17", "layoutItemId": "3"}]},
            },
            "widget_47": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_47",
                "parent": {
                    "LARGE": [{"layoutId": "layout_18", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_202", "layoutItemId": "0"}],
                },
            },
            "widget_48": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_48",
                "parent": {
                    "LARGE": [{"layoutId": "layout_18", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_202", "layoutItemId": "1"}],
                },
            },
            "widget_49": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 0, 50], "unit": "px"}},
                },
                "id": "widget_49",
                "layouts": {"DEFAULT": {"LARGE": "layout_19"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "5"}]},
            },
            "widget_50": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 24px; color: var(--dark-800);">\ufeff</span><span style="font-size: 24px; color: rgb(24, 24, 24);">\ufeffTitle</span></p>',
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_50",
                "parent": {
                    "LARGE": [{"layoutId": "layout_19", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_203", "layoutItemId": "0"}],
                },
            },
            "widget_51": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_51",
                "layouts": {"DEFAULT": {"LARGE": "layout_20"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "6"}]},
            },
            "widget_52": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-300);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_52",
                "parent": {
                    "LARGE": [{"layoutId": "layout_20", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_204", "layoutItemId": "0"}],
                },
            },
            "widget_53": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_53",
                "layouts": {"DEFAULT": {"LARGE": "layout_21"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "7"}]},
            },
            "widget_54": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_54",
                "layouts": {"DEFAULT": {"LARGE": "layout_22"}},
                "parent": {"LARGE": [{"layoutId": "layout_21", "layoutItemId": "0"}]},
            },
            "widget_55": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--primary)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE8",
                    "direction": "Vertical",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_55",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_26",
                        "MEDIUM": "layout_209",
                        "SMALL": "layout_310",
                    },
                    "HOVER": {
                        "LARGE": "layout_24",
                        "MEDIUM": "layout_210",
                        "SMALL": "layout_311",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_22", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_208", "layoutItemId": "0"}],
                },
            },
            "widget_67": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1;"><span style="color: var(--light); font-size: 18px;">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_67",
                "parent": {
                    "LARGE": [{"layoutId": "layout_26", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_209", "layoutItemId": "0"}],
                },
            },
            "widget_68": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="color: var(--light); font-size: 14px;">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_68",
                "parent": {
                    "LARGE": [{"layoutId": "layout_26", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_209", "layoutItemId": "1"}],
                },
            },
            "widget_69": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_69",
                "parent": {
                    "LARGE": [{"layoutId": "layout_26", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_209", "layoutItemId": "2"}],
                },
            },
            "widget_73": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M2 8c0-.291.231-.527.516-.527h9.508L8.527 3.9a.535.535 0 010-.746.508.508 0 01.73 0L14 8l-4.743 4.846a.508.508 0 01-.73 0 .535.535 0 010-.746l3.497-3.573H2.516A.522.522 0 012 8z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(240,240,240,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 24},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(255,255,255,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "#c5c5c5",
                                    "width": "0px",
                                },
                                "text": {"color": "#181818", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 24},
                            },
                        },
                    },
                },
                "id": "widget_73",
                "parent": {
                    "LARGE": [{"layoutId": "layout_26", "layoutItemId": "10"}],
                    "MEDIUM": [{"layoutId": "layout_209", "layoutItemId": "3"}],
                },
            },
            "widget_74": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE8",
                    "direction": "Vertical",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_74",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_27",
                        "MEDIUM": "layout_211",
                        "SMALL": "layout_312",
                    },
                    "HOVER": {
                        "LARGE": "layout_28",
                        "MEDIUM": "layout_212",
                        "SMALL": "layout_313",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_22", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_208", "layoutItemId": "1"}],
                },
            },
            "widget_75": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1;"><span style="font-size: 18px; color: var(--dark);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_75",
                "parent": {
                    "LARGE": [{"layoutId": "layout_27", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_211", "layoutItemId": "0"}],
                },
            },
            "widget_76": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_76",
                "parent": {
                    "LARGE": [{"layoutId": "layout_27", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_211", "layoutItemId": "1"}],
                },
            },
            "widget_77": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_77",
                "parent": {
                    "LARGE": [{"layoutId": "layout_27", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_211", "layoutItemId": "2"}],
                },
            },
            "widget_78": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M2 8c0-.291.231-.527.516-.527h9.508L8.527 3.9a.535.535 0 010-.746.508.508 0 01.73 0L14 8l-4.743 4.846a.508.508 0 01-.73 0 .535.535 0 010-.746l3.497-3.573H2.516A.522.522 0 012 8z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(240,240,240,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark)", "size": 24},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark)", "size": 24},
                            },
                        },
                    },
                },
                "id": "widget_78",
                "parent": {
                    "LARGE": [{"layoutId": "layout_27", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_211", "layoutItemId": "3"}],
                },
            },
            "widget_79": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_79",
                "layouts": {"DEFAULT": {"LARGE": "layout_29"}},
                "parent": {"LARGE": [{"layoutId": "layout_21", "layoutItemId": "1"}]},
            },
            "widget_80": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE8",
                    "direction": "Vertical",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_80",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_30",
                        "MEDIUM": "layout_214",
                        "SMALL": "layout_315",
                    },
                    "HOVER": {
                        "LARGE": "layout_31",
                        "MEDIUM": "layout_215",
                        "SMALL": "layout_316",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_29", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_213", "layoutItemId": "0"}],
                },
            },
            "widget_81": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1;"><span style="font-size: 18px; color: var(--dark);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_81",
                "parent": {
                    "LARGE": [{"layoutId": "layout_30", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_214", "layoutItemId": "0"}],
                },
            },
            "widget_82": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_82",
                "parent": {
                    "LARGE": [{"layoutId": "layout_30", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_214", "layoutItemId": "1"}],
                },
            },
            "widget_83": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_83",
                "parent": {
                    "LARGE": [{"layoutId": "layout_30", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_214", "layoutItemId": "2"}],
                },
            },
            "widget_84": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M2 8c0-.291.231-.527.516-.527h9.508L8.527 3.9a.535.535 0 010-.746.508.508 0 01.73 0L14 8l-4.743 4.846a.508.508 0 01-.73 0 .535.535 0 010-.746l3.497-3.573H2.516A.522.522 0 012 8z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(240,240,240,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark)", "size": 24},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark)", "size": 24},
                            },
                        },
                    },
                },
                "id": "widget_84",
                "parent": {
                    "LARGE": [{"layoutId": "layout_30", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_214", "layoutItemId": "3"}],
                },
            },
            "widget_85": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--primary)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE8",
                    "direction": "Vertical",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_85",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_32",
                        "MEDIUM": "layout_216",
                        "SMALL": "layout_317",
                    },
                    "HOVER": {
                        "LARGE": "layout_33",
                        "MEDIUM": "layout_217",
                        "SMALL": "layout_318",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_29", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_213", "layoutItemId": "1"}],
                },
            },
            "widget_86": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1;"><span style="color: var(--light); font-size: 18px;">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_86",
                "parent": {
                    "LARGE": [{"layoutId": "layout_32", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_216", "layoutItemId": "0"}],
                },
            },
            "widget_87": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="color: var(--light); font-size: 14px;">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_87",
                "parent": {
                    "LARGE": [{"layoutId": "layout_32", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_216", "layoutItemId": "1"}],
                },
            },
            "widget_88": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_88",
                "parent": {
                    "LARGE": [{"layoutId": "layout_32", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_216", "layoutItemId": "2"}],
                },
            },
            "widget_89": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M2 8c0-.291.231-.527.516-.527h9.508L8.527 3.9a.535.535 0 010-.746.508.508 0 01.73 0L14 8l-4.743 4.846a.508.508 0 01-.73 0 .535.535 0 010-.746l3.497-3.573H2.516A.522.522 0 012 8z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(240,240,240,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 24},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 24},
                            },
                        },
                    },
                },
                "id": "widget_89",
                "parent": {
                    "LARGE": [{"layoutId": "layout_32", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_216", "layoutItemId": "3"}],
                },
            },
            "widget_90": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE2",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_90",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_58",
                        "MEDIUM": "layout_206",
                        "SMALL": "layout_307",
                    },
                    "HOVER": {
                        "LARGE": "layout_59",
                        "MEDIUM": "layout_207",
                        "SMALL": "layout_308",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_21", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_205", "layoutItemId": "2"}],
                },
            },
            "widget_197": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_197",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "0"}],
                },
            },
            "widget_198": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark-800);">Created author: admin</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_198",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "1"}],
                },
            },
            "widget_199": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_199",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "2"}],
                },
            },
            "widget_200": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Created date: xxx</span></p></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_200",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "7"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "3"}],
                },
            },
            "widget_201": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "success-100",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_201",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "9"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "4"}],
                },
            },
            "widget_202": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "var(--primary-400)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-400)",
                                    "width": "1px",
                                },
                                "text": {"color": "var(--light)", "size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--primary-500)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"size": 14},
                            },
                        },
                    },
                },
                "id": "widget_202",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "10"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "5"}],
                },
            },
            "widget_206": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 0, 50], "unit": "px"}},
                },
                "id": "widget_206",
                "layouts": {"DEFAULT": {"LARGE": "layout_60"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "8"}]},
            },
            "widget_207": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 24px; color: var(--dark-800);">\ufeff</span><span style="font-size: 24px; color: rgb(24, 24, 24);">\ufeffTitle</span></p>',
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_207",
                "parent": {
                    "LARGE": [{"layoutId": "layout_60", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_218", "layoutItemId": "0"}],
                },
            },
            "widget_208": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_208",
                "layouts": {"DEFAULT": {"LARGE": "layout_61"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "9"}]},
            },
            "widget_209": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-300);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_209",
                "parent": {
                    "LARGE": [{"layoutId": "layout_61", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_219", "layoutItemId": "0"}],
                },
            },
            "widget_212": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_1",
                        "views": ["view_1", "view_2"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "pills",
                        "standard": {
                            "gap": "0px",
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": True,
                        "variant": {
                            "size": {
                                "fontSize": "0.875rem",
                                "lineHeight": 1.5,
                                "paddingX": "1rem",
                                "paddingY": "0.25rem",
                            },
                            "gutter": "0.25rem",
                            "icon": {"spacing": "0.5rem"},
                            "root": {"borderRadius": "50rem"},
                            "item": {
                                "default": {
                                    "color": "var(--dark)",
                                    "border": {"width": "0px"},
                                    "borderRadius": "50rem",
                                    "size": "24px",
                                },
                                "hover": {
                                    "color": "var(--primary)",
                                    "borderRadius": "50rem",
                                    "size": "24px",
                                },
                                "active": {
                                    "color": "var(--light)",
                                    "bg": "var(--primary)",
                                    "borderRadius": "50rem",
                                    "size": "24px",
                                },
                                "disabled": {
                                    "color": "#989898",
                                    "shadow": "none",
                                    "borderRadius": "50rem",
                                },
                            },
                        },
                    },
                },
                "id": "widget_212",
                "parent": {
                    "LARGE": [{"layoutId": "layout_66", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_221", "layoutItemId": "0"}],
                },
            },
            "widget_213": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_213",
                "layouts": {"DEFAULT": {"LARGE": "layout_65"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "12"}]},
            },
            "widget_214": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 30},
                    "style": {
                        "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_214",
                "layouts": {"DEFAULT": {"LARGE": "layout_66"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
                "parent": {"LARGE": [{"layoutId": "layout_65", "layoutItemId": "0"}]},
            },
            "widget_215": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_215",
                "layouts": {"DEFAULT": {"LARGE": "layout_67"}},
                "parent": {"LARGE": [{"layoutId": "layout_63", "layoutItemId": "0"}]},
            },
            "widget_216": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 36978595.474472,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": 963565.178562682,
                                "y": 2749694.3734205654,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -929627.13800406,
                            "ymin": -165919.63348842273,
                            "xmax": 2856757.495129424,
                            "ymax": 5665308.380329553,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_216",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_68",
                        "MEDIUM": "layout_224",
                        "SMALL": "layout_325",
                    }
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_67", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_223", "layoutItemId": "0"}],
                },
            },
            "widget_217": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_217",
                "layouts": {"DEFAULT": {"LARGE": "layout_69"}},
                "parent": {"LARGE": [{"layoutId": "layout_67", "layoutItemId": "1"}]},
            },
            "widget_218": {
                "uri": "widgets/arcgis/query/",
                "version": "1.5.0",
                "config": {
                    "queryItems": [],
                    "arrangeType": "BLOCK",
                    "arrangeWrap": False,
                    "sizeMap": {
                        "arrangementIconPopper": {
                            "minSize": {"width": 300, "height": 300},
                            "defaultSize": {"width": 360, "height": 480},
                        }
                    },
                },
                "id": "widget_218",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_69", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_225", "layoutItemId": "0"}],
                },
            },
            "widget_219": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_219",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_69", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_225", "layoutItemId": "1"}],
                },
            },
            "widget_220": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 0, 50], "unit": "px"}},
                },
                "id": "widget_220",
                "layouts": {"DEFAULT": {"LARGE": "layout_70"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "13"}]},
            },
            "widget_221": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 24px; color: var(--dark-800);">\ufeff</span><span style="font-size: 24px; color: rgb(24, 24, 24);">\ufeffTitle</span></p>',
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_221",
                "parent": {
                    "LARGE": [{"layoutId": "layout_70", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_233", "layoutItemId": "0"}],
                },
            },
            "widget_222": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_222",
                "layouts": {"DEFAULT": {"LARGE": "layout_71"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "14"}]},
            },
            "widget_223": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-300);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_223",
                "parent": {
                    "LARGE": [{"layoutId": "layout_71", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_234", "layoutItemId": "0"}],
                },
            },
            "widget_224": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_224",
                "layouts": {"DEFAULT": {"LARGE": "layout_72"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "15"}]},
            },
            "widget_225": {
                "uri": "widgets/common/table/",
                "version": "1.5.0",
                "config": {"layersConfig": [], "arrangeType": "TABS"},
                "id": "widget_225",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_72", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_235", "layoutItemId": "0"}],
                },
            },
            "widget_226": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_226",
                "layouts": {"DEFAULT": {"LARGE": "layout_73"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "16"}]},
            },
            "widget_227": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [30, 20, 30, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_227",
                "layouts": {"DEFAULT": {"LARGE": "layout_74"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_73", "layoutItemId": "0"}]},
            },
            "widget_228": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_228",
                "parent": {
                    "LARGE": [{"layoutId": "layout_74", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_237", "layoutItemId": "0"}],
                },
            },
            "widget_229": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_229",
                "layouts": {"DEFAULT": {"LARGE": "layout_75"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "1"}]},
            },
            "widget_230": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_230",
                "parent": {
                    "LARGE": [{"layoutId": "layout_75", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_238", "layoutItemId": "0"}],
                },
            },
            "widget_231": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_231",
                "parent": {
                    "LARGE": [{"layoutId": "layout_75", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_238", "layoutItemId": "1"}],
                },
            },
            "widget_232": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_232",
                "layouts": {"DEFAULT": {"LARGE": "layout_76"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "2"}]},
            },
            "widget_233": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_233",
                "parent": {
                    "LARGE": [{"layoutId": "layout_76", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_239", "layoutItemId": "0"}],
                },
            },
            "widget_234": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_234",
                "parent": {
                    "LARGE": [{"layoutId": "layout_76", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_239", "layoutItemId": "1"}],
                },
            },
            "widget_235": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_235",
                "layouts": {"DEFAULT": {"LARGE": "layout_77"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "3"}]},
            },
            "widget_236": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_236",
                "parent": {
                    "LARGE": [{"layoutId": "layout_77", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_240", "layoutItemId": "0"}],
                },
            },
            "widget_237": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_237",
                "parent": {
                    "LARGE": [{"layoutId": "layout_77", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_240", "layoutItemId": "1"}],
                },
            },
            "widget_238": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_238",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_73", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_236", "layoutItemId": "1"}],
                },
            },
            "widget_239": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_239",
                "layouts": {"DEFAULT": {"LARGE": "layout_79"}},
                "parent": {"LARGE": [{"layoutId": "layout_78", "layoutItemId": "0"}]},
            },
            "widget_240": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 73957190.948944,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -3045556.9549937723,
                                "y": -2194864.1122896317,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -6831941.588127256,
                            "ymin": -8026092.126107608,
                            "xmax": 740827.6781397117,
                            "ymax": 3636363.9015283445,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_240",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_80",
                        "MEDIUM": "layout_228",
                        "SMALL": "layout_329",
                    }
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_79", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_227", "layoutItemId": "0"}],
                },
            },
            "widget_241": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_241",
                "layouts": {"DEFAULT": {"LARGE": "layout_81"}},
                "parent": {"LARGE": [{"layoutId": "layout_79", "layoutItemId": "1"}]},
            },
            "widget_243": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_243",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_81", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_229", "layoutItemId": "0"}],
                },
            },
            "widget_244": {
                "uri": "widgets/common/list/",
                "version": "1.5.0",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE5",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "scrollBarOpen": True,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {"LARGE": {"width": 320, "height": 88}},
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {"LARGE": {"width": 320, "height": 88}},
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {"LARGE": {"width": 320, "height": 88}},
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-500)",
                                    "width": "1px",
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                },
                "id": "widget_244",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_89",
                        "MEDIUM": "layout_230",
                        "SMALL": "layout_331",
                    },
                    "SELECTED": {
                        "LARGE": "layout_83",
                        "MEDIUM": "layout_231",
                        "SMALL": "layout_332",
                    },
                    "HOVER": {
                        "LARGE": "layout_84",
                        "MEDIUM": "layout_232",
                        "SMALL": "layout_333",
                    },
                },
                "useDataSourcesEnabled": True,
                "style": {
                    "background": {
                        "color": "rgba(240,240,240,0)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {"type": "solid", "color": "", "width": "0px"},
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_81", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_229", "layoutItemId": "1"}],
                },
            },
            "widget_255": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "cropShape": "circle",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "svgViewBox": "0 0 28 28",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_255",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_89", "layoutItemId": "0"},
                        {"layoutId": "layout_83", "layoutItemId": "0"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_230", "layoutItemId": "0"},
                        {"layoutId": "layout_231", "layoutItemId": "0"},
                    ],
                },
            },
            "widget_256": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 16px;color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_256",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_89", "layoutItemId": "1"},
                        {"layoutId": "layout_83", "layoutItemId": "1"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_230", "layoutItemId": "1"},
                        {"layoutId": "layout_231", "layoutItemId": "1"},
                    ],
                },
            },
            "widget_257": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "Button",
                        "toolTip": "",
                        "linkParam": {},
                    },
                    "styleConfig": {
                        "name": "Default",
                        "themeStyle": {
                            "className": "jimu-widget d-block",
                            "color": "primary",
                        },
                        "customStyle": {"regular": {}, "hover": {}, "clicked": {}},
                    },
                },
                "id": "widget_257",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_89", "layoutItemId": "2"},
                        {"layoutId": "layout_83", "layoutItemId": "2"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_230", "layoutItemId": "2"},
                        {"layoutId": "layout_231", "layoutItemId": "2"},
                    ],
                },
            },
            "widget_259": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><strong style="font-size: 14px; color: var(--dark-800);">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_259",
                "parent": {"LARGE": [{"layoutId": "layout_90", "layoutItemId": "1"}]},
            },
            "widget_260": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_260",
                "parent": {"LARGE": [{"layoutId": "layout_90", "layoutItemId": "2"}]},
            },
            "widget_261": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_261",
                "parent": {"LARGE": [{"layoutId": "layout_91", "layoutItemId": "0"}]},
            },
            "widget_262": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><strong style="font-size: 14px; color: var(--dark-800);">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_262",
                "parent": {"LARGE": [{"layoutId": "layout_91", "layoutItemId": "1"}]},
            },
            "widget_263": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_263",
                "parent": {"LARGE": [{"layoutId": "layout_91", "layoutItemId": "2"}]},
            },
            "widget_432": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 20, 0, 20], "unit": "px"}},
                },
                "id": "widget_432",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_188"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "0"}]},
            },
            "widget_433": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 20, 0, 20], "unit": "px"}},
                },
                "id": "widget_433",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_189"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "1"}]},
            },
            "widget_434": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_434",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_190"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "2"}]},
            },
            "widget_435": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_435",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_191"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "3"}]},
            },
            "widget_436": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {"padding": {"number": [20, 0, 20, 20], "unit": "px"}},
                },
                "id": "widget_436",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_192"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_191", "layoutItemId": "0"}]},
            },
            "widget_437": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_437",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_193"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_192", "layoutItemId": "0"}]},
            },
            "widget_438": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_438",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_194"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_193", "layoutItemId": "2"}]},
            },
            "widget_439": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_439",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_195"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_192", "layoutItemId": "1"}]},
            },
            "widget_440": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_440",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_196"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_195", "layoutItemId": "2"}]},
            },
            "widget_441": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_441",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_197"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_192", "layoutItemId": "4"}]},
            },
            "widget_442": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_442",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_198"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_197", "layoutItemId": "2"}]},
            },
            "widget_443": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_443",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_199"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_191", "layoutItemId": "1"}]},
            },
            "widget_444": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_444",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_200"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "4"}]},
            },
            "widget_445": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_445",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_201"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_200", "layoutItemId": "1"}]},
            },
            "widget_446": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_446",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_202"}},
                "style": {
                    "background": {
                        "color": "rgba(7,111,229,0.1)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_201", "layoutItemId": "3"}]},
            },
            "widget_447": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 20, 0, 20], "unit": "px"}},
                },
                "id": "widget_447",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_203"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "5"}]},
            },
            "widget_448": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_448",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_204"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "6"}]},
            },
            "widget_449": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_449",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_205"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "7"}]},
            },
            "widget_450": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_450",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_208"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_205", "layoutItemId": "0"}]},
            },
            "widget_451": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_451",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_213"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_205", "layoutItemId": "1"}]},
            },
            "widget_452": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 20, 0, 20], "unit": "px"}},
                },
                "id": "widget_452",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_218"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "8"}]},
            },
            "widget_453": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_453",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_219"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "9"}]},
            },
            "widget_454": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_454",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_220"}},
                "parent": {
                    "MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "10"}]
                },
            },
            "widget_455": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 30},
                    "style": {
                        "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_455",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_221"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_220", "layoutItemId": "0"}]},
            },
            "widget_456": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_456",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_223"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_222", "layoutItemId": "0"}]},
            },
            "widget_457": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_457",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_225"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_223", "layoutItemId": "1"}]},
            },
            "widget_458": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_458",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_227"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_226", "layoutItemId": "0"}]},
            },
            "widget_459": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_459",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_229"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_227", "layoutItemId": "1"}]},
            },
            "widget_460": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 20, 0, 20], "unit": "px"}},
                },
                "id": "widget_460",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_233"}},
                "parent": {
                    "MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "11"}]
                },
            },
            "widget_461": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 20, 20, 20], "unit": "px"}},
                },
                "id": "widget_461",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_234"}},
                "parent": {
                    "MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "12"}]
                },
            },
            "widget_462": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_462",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_235"}},
                "parent": {
                    "MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "13"}]
                },
            },
            "widget_463": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 20, 20, 20], "unit": "px"}},
                },
                "id": "widget_463",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_236"}},
                "parent": {
                    "MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "14"}]
                },
            },
            "widget_464": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 10},
                    "style": {
                        "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_464",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_237"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_236", "layoutItemId": "0"}]},
            },
            "widget_465": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_465",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_238"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_237", "layoutItemId": "1"}]},
            },
            "widget_466": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_466",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_239"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_237", "layoutItemId": "2"}]},
            },
            "widget_467": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_467",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_240"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_237", "layoutItemId": "3"}]},
            },
            "widget_468": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 0, 10], "unit": "px"}},
                },
                "id": "widget_468",
                "layouts": {"DEFAULT": {"SMALL": "layout_289"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "0"}]},
            },
            "widget_469": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 0, 10], "unit": "px"}},
                },
                "id": "widget_469",
                "layouts": {"DEFAULT": {"SMALL": "layout_290"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "1"}]},
            },
            "widget_470": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_470",
                "layouts": {"DEFAULT": {"SMALL": "layout_291"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "2"}]},
            },
            "widget_471": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_471",
                "layouts": {"DEFAULT": {"SMALL": "layout_292"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "3"}]},
            },
            "widget_472": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_472",
                "layouts": {"DEFAULT": {"SMALL": "layout_293"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_7", "layoutItemId": "0"}]},
            },
            "widget_473": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_473",
                "layouts": {"DEFAULT": {"SMALL": "layout_294"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "0"}]},
            },
            "widget_474": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_474",
                "layouts": {"DEFAULT": {"SMALL": "layout_295"}},
                "parent": {"LARGE": [{"layoutId": "layout_9", "layoutItemId": "2"}]},
            },
            "widget_475": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_475",
                "layouts": {"DEFAULT": {"SMALL": "layout_296"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "4"}]},
            },
            "widget_476": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_476",
                "layouts": {"DEFAULT": {"SMALL": "layout_297"}},
                "parent": {"LARGE": [{"layoutId": "layout_13", "layoutItemId": "2"}]},
            },
            "widget_477": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_477",
                "layouts": {"DEFAULT": {"SMALL": "layout_298"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "2"}]},
            },
            "widget_478": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_478",
                "layouts": {"DEFAULT": {"SMALL": "layout_299"}},
                "parent": {"LARGE": [{"layoutId": "layout_11", "layoutItemId": "2"}]},
            },
            "widget_479": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_479",
                "layouts": {"DEFAULT": {"SMALL": "layout_300"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_7", "layoutItemId": "1"}]},
            },
            "widget_480": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_480",
                "layouts": {"DEFAULT": {"SMALL": "layout_301"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "4"}]},
            },
            "widget_481": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"number": [40, 20, 40, 20], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_481",
                "layouts": {"DEFAULT": {"SMALL": "layout_302"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_16", "layoutItemId": "1"}]},
            },
            "widget_482": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_482",
                "layouts": {"DEFAULT": {"SMALL": "layout_303"}},
                "style": {
                    "background": {
                        "color": "rgba(7,111,229,0.1)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
                "parent": {"LARGE": [{"layoutId": "layout_17", "layoutItemId": "3"}]},
            },
            "widget_483": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 0, 10], "unit": "px"}},
                },
                "id": "widget_483",
                "layouts": {"DEFAULT": {"SMALL": "layout_304"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "5"}]},
            },
            "widget_484": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_484",
                "layouts": {"DEFAULT": {"SMALL": "layout_305"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "6"}]},
            },
            "widget_485": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_485",
                "layouts": {"DEFAULT": {"SMALL": "layout_306"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "7"}]},
            },
            "widget_486": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_486",
                "layouts": {"DEFAULT": {"SMALL": "layout_309"}},
                "parent": {"LARGE": [{"layoutId": "layout_21", "layoutItemId": "0"}]},
            },
            "widget_487": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_487",
                "layouts": {"DEFAULT": {"SMALL": "layout_314"}},
                "parent": {"LARGE": [{"layoutId": "layout_21", "layoutItemId": "1"}]},
            },
            "widget_488": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 0, 10], "unit": "px"}},
                },
                "id": "widget_488",
                "layouts": {"DEFAULT": {"SMALL": "layout_319"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "8"}]},
            },
            "widget_489": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_489",
                "layouts": {"DEFAULT": {"SMALL": "layout_320"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "9"}]},
            },
            "widget_490": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_490",
                "layouts": {"DEFAULT": {"SMALL": "layout_321"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "12"}]},
            },
            "widget_491": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 30},
                    "style": {
                        "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_491",
                "layouts": {"DEFAULT": {"SMALL": "layout_322"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
                "parent": {"LARGE": [{"layoutId": "layout_65", "layoutItemId": "0"}]},
            },
            "widget_492": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_492",
                "layouts": {"DEFAULT": {"SMALL": "layout_324"}},
                "parent": {"LARGE": [{"layoutId": "layout_63", "layoutItemId": "0"}]},
            },
            "widget_493": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_493",
                "layouts": {"DEFAULT": {"SMALL": "layout_326"}},
                "parent": {"LARGE": [{"layoutId": "layout_67", "layoutItemId": "1"}]},
            },
            "widget_494": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_494",
                "layouts": {"DEFAULT": {"SMALL": "layout_328"}},
                "parent": {"LARGE": [{"layoutId": "layout_78", "layoutItemId": "0"}]},
            },
            "widget_495": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_495",
                "layouts": {"DEFAULT": {"SMALL": "layout_330"}},
                "parent": {"LARGE": [{"layoutId": "layout_79", "layoutItemId": "1"}]},
            },
            "widget_496": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 0, 10], "unit": "px"}},
                },
                "id": "widget_496",
                "layouts": {"DEFAULT": {"SMALL": "layout_334"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "13"}]},
            },
            "widget_497": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_497",
                "layouts": {"DEFAULT": {"SMALL": "layout_335"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "14"}]},
            },
            "widget_498": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_498",
                "layouts": {"DEFAULT": {"SMALL": "layout_336"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "15"}]},
            },
            "widget_499": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_499",
                "layouts": {"DEFAULT": {"SMALL": "layout_337"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "16"}]},
            },
            "widget_500": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [30, 20, 30, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_500",
                "layouts": {"DEFAULT": {"SMALL": "layout_338"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_73", "layoutItemId": "0"}]},
            },
            "widget_501": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_501",
                "layouts": {"DEFAULT": {"SMALL": "layout_339"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "1"}]},
            },
            "widget_502": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_502",
                "layouts": {"DEFAULT": {"SMALL": "layout_340"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "2"}]},
            },
            "widget_503": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_503",
                "layouts": {"DEFAULT": {"SMALL": "layout_341"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "3"}]},
            },
            "widget_504": {
                "uri": "widgets/layout/column/",
                "id": "widget_504",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_342"}},
            },
            "widget_505": {
                "uri": "widgets/layout/column/",
                "id": "widget_505",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_343"}},
            },
            "widget_506": {
                "uri": "widgets/layout/column/",
                "id": "widget_506",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_344"}},
            },
            "widget_507": {
                "uri": "widgets/layout/column/",
                "id": "widget_507",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_345"}},
            },
            "widget_508": {
                "uri": "widgets/layout/column/",
                "id": "widget_508",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_346"}},
            },
            "widget_509": {
                "uri": "widgets/layout/column/",
                "id": "widget_509",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_347"}},
            },
            "widget_510": {
                "uri": "widgets/layout/column/",
                "id": "widget_510",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_348"}},
            },
            "widget_511": {
                "uri": "widgets/layout/column/",
                "id": "widget_511",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_349"}},
            },
            "widget_512": {
                "uri": "widgets/layout/column/",
                "id": "widget_512",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_350"}},
            },
            "widget_513": {
                "uri": "widgets/layout/column/",
                "id": "widget_513",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_351"}},
            },
            "widget_514": {
                "uri": "widgets/layout/column/",
                "id": "widget_514",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_352"}},
            },
            "widget_515": {
                "uri": "widgets/layout/column/",
                "id": "widget_515",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_353"}},
            },
            "widget_516": {
                "uri": "widgets/layout/column/",
                "id": "widget_516",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_354"}},
            },
            "widget_517": {
                "uri": "widgets/layout/column/",
                "id": "widget_517",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_355"}},
            },
            "widget_518": {
                "uri": "widgets/layout/column/",
                "id": "widget_518",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_356"}},
            },
            "widget_519": {
                "uri": "widgets/layout/column/",
                "id": "widget_519",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_357"}},
            },
            "widget_520": {
                "uri": "widgets/layout/column/",
                "id": "widget_520",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_358"}},
            },
            "widget_521": {
                "uri": "widgets/layout/column/",
                "id": "widget_521",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_359"}},
            },
            "widget_522": {
                "uri": "widgets/layout/column/",
                "id": "widget_522",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_360"}},
            },
            "widget_523": {
                "uri": "widgets/layout/column/",
                "id": "widget_523",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_361"}},
            },
            "widget_524": {
                "uri": "widgets/layout/column/",
                "id": "widget_524",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_362"}},
            },
            "widget_525": {
                "uri": "widgets/layout/column/",
                "id": "widget_525",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_363"}},
            },
        },
        "widgetsManifest": {},
        "views": {
            "view_1": {
                "id": "view_1",
                "layout": {
                    "LARGE": "layout_63",
                    "MEDIUM": "layout_222",
                    "SMALL": "layout_323",
                },
                "backgroundPosition": "fit",
                "backgroundColor": None,
                "backgroundIMage": {},
            },
            "view_2": {
                "id": "view_2",
                "layout": {
                    "LARGE": "layout_78",
                    "MEDIUM": "layout_226",
                    "SMALL": "layout_327",
                },
                "backgroundPosition": "fit",
                "backgroundColor": None,
                "backgroundIMage": {},
            },
        },
        "sections": {
            "section_1": {
                "id": "section_1",
                "views": ["view_1", "view_2"],
                "parent": {
                    "LARGE": [{"layoutId": "layout_66", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_221", "layoutItemId": "1"}],
                },
            }
        },
        "dialogs": {
            "dialog_1": {
                "id": "dialog_1",
                "mode": "ANCHORED",
                "sizeMode": {
                    "LARGE": {"position": None, "width": "360px", "height": "240px"}
                },
                "interactionType": None,
                "interactionStyles": None,
                "confirmBeforeClose": True,
                "alwaysShowConfirmation": False,
                "dialogBackground": {
                    "color": "var(--light)",
                    "fillType": "fit",
                    "image": {"url": ""},
                },
                "overlayColor": None,
                "layout": {"LARGE": "layout_90"},
                "index": 0,
                "isSplash": False,
                "closeWhenClickOverlay": True,
                "effect": {
                    "type": "FADE_IN",
                    "option": {"direction": None, "configType": "DEFAULT"},
                },
                "oneByOneEffect": {"type": "NONE", "option": {"direction": None}},
                "boxShadow": {
                    "offsetX": "0px",
                    "offsetY": "0px",
                    "blur": "0px",
                    "spread": "0px",
                    "color": "",
                },
            },
            "dialog_2": {
                "id": "dialog_2",
                "mode": "ANCHORED",
                "sizeMode": {
                    "LARGE": {"position": None, "width": "360px", "height": "240px"}
                },
                "interactionType": None,
                "interactionStyles": None,
                "confirmBeforeClose": True,
                "alwaysShowConfirmation": False,
                "dialogBackground": {
                    "color": "var(--light)",
                    "fillType": "fit",
                    "image": {"url": ""},
                },
                "overlayColor": None,
                "layout": {"LARGE": "layout_91"},
                "index": 1,
                "isSplash": False,
                "closeWhenClickOverlay": True,
                "effect": {
                    "type": "FADE_IN",
                    "option": {"direction": None, "configType": "DEFAULT"},
                },
                "oneByOneEffect": {"type": "NONE", "option": {"direction": None}},
            },
        },
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_0": []}],
        "header": {
            "layout": {
                "LARGE": "layout_1",
                "MEDIUM": "layout_186",
                "SMALL": "layout_241",
            },
            "height": {"LARGE": "120px", "SMALL": "180px"},
            "backgroundColor": "var(--dark-700)",
        },
        "footer": {
            "layout": {
                "LARGE": "layout_2",
                "MEDIUM": "layout_286",
                "SMALL": "layout_287",
            },
            "height": {"LARGE": "80px"},
            "backgroundColor": "var(--light)",
        },
        "timestamp": 1619659964012,
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "indicator": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_0",
                    "MEDIUM": "layout_187",
                    "SMALL": "layout_288",
                },
                "isDefault": True,
                "isVisible": True,
                "header": True,
                "footer": True,
                "icon": {
                    "svg": '<svg viewBox="0 0 10 12" xmlns="http://www.w3.org/2000/svg"><path d="M1 0h8a1 1 0 011 1v10a1 1 0 01-1 1H1a1 1 0 01-1-1V1a1 1 0 011-1zm0 1v10h8V1H1zm2 2h4a.5.5 0 010 1H3a.5.5 0 010-1zm0 2.5h4a.5.5 0 010 1H3a.5.5 0 010-1z" fill="#000" fill-rule="nonzero"></path></svg>',
                    "properties": {
                        "filename": None,
                        "originalName": None,
                        "path": None,
                        "color": "#fff",
                        "size": 16,
                        "inlineSvg": True,
                    },
                },
                "bodyBackgroundColor": "var(--light-200)",
            },
            "page_1": {
                "id": "page_1",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_3",
                    "MEDIUM": "layout_144",
                    "SMALL": "layout_242",
                },
                "isDefault": False,
                "isVisible": True,
                "header": True,
                "footer": True,
                "icon": {
                    "svg": '<svg viewBox="0 0 10 12" xmlns="http://www.w3.org/2000/svg"><path d="M1 0h8a1 1 0 011 1v10a1 1 0 01-1 1H1a1 1 0 01-1-1V1a1 1 0 011-1zm0 1v10h8V1H1zm2 2h4a.5.5 0 010 1H3a.5.5 0 010-1zm0 2.5h4a.5.5 0 010 1H3a.5.5 0 010-1z" fill="#000" fill-rule="nonzero"></path></svg>',
                    "properties": {
                        "filename": None,
                        "originalName": None,
                        "path": None,
                        "color": "#fff",
                        "size": 16,
                        "inlineSvg": True,
                    },
                },
                "bodyBackgroundColor": "var(--light-200)",
            },
        },
        "layouts": {
            "layout_0": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "723px",
                            "right": "1073px",
                            "top": "-78px",
                            "bottom": "322px",
                            "width": "350px",
                            "height": "60px",
                        },
                        "widgetId": "widget_8",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_0",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "145px",
                            "right": "505px",
                            "top": "79px",
                            "bottom": "149px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_10",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "105px",
                            "right": "465px",
                            "top": "156px",
                            "bottom": "226px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_12",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-71px",
                            "right": "729px",
                            "top": "88px",
                            "bottom": "488px",
                            "width": "800px",
                            "height": 420,
                        },
                        "widgetId": "widget_14",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "179px",
                            "right": "629px",
                            "top": "413px",
                            "bottom": "713px",
                            "width": "450px",
                            "height": "500px",
                        },
                        "widgetId": "widget_40",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1168px",
                            "bottom": "1233px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_49",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1292px",
                            "bottom": "1335px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_51",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "126px",
                            "right": "426px",
                            "top": "980px",
                            "bottom": "1580px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_53",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1908px",
                            "bottom": "1972px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_206",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1958px",
                            "bottom": "2011px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_208",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "316px",
                            "right": "616px",
                            "top": "2260px",
                            "bottom": "2860px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_213",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3004px",
                            "bottom": "3068px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_220",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3044px",
                            "bottom": "3097px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_222",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "15": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "131px",
                            "right": "731px",
                            "top": "2900px",
                            "bottom": "3300px",
                            "width": "600px",
                            "height": 420,
                        },
                        "widgetId": "widget_224",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_0",
                    },
                    "16": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "78px",
                            "right": "378px",
                            "top": "3248px",
                            "bottom": "3848px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_226",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_0",
                    },
                },
                "order": [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "12",
                    "13",
                    "14",
                    "15",
                    "16",
                ],
                "parent": {"type": "pages", "id": "page_0"},
            },
            "layout_1": {
                "type": "FIXED",
                "order": ["0", "1", "2", "3", "4", "5"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50.00px",
                            "right": "94.8%",
                            "top": "40.00px",
                            "bottom": "54.3%",
                            "width": "40px",
                            "height": "40px",
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_1",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "100px",
                            "right": "64.1%",
                            "top": "40px",
                            "bottom": "49.6%",
                            "width": "360px",
                            "height": "41px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.5%",
                            "right": "50.00px",
                            "top": "30.00px",
                            "bottom": "-52.8%",
                            "width": "50.00px",
                            "height": "50.00px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "65.7%",
                            "right": "140.00px",
                            "top": "40.00px",
                            "bottom": "50.3%",
                            "width": "32.00px",
                            "height": "32.00px",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.6%",
                            "right": "200.00px",
                            "top": "45.00px",
                            "bottom": "24.1%",
                            "width": "24.00px",
                            "height": "24.00px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "30px",
                            "right": "76.6%",
                            "top": "68.7%",
                            "bottom": "0px",
                            "width": "21.1%",
                            "height": "31.3%",
                        },
                        "widgetId": "widget_6",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": True,
                                "bottom": False,
                                "width": "AUTO",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_1",
                    },
                },
                "parent": {"type": "header", "id": "header"},
            },
            "layout_2": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "42.6%",
                            "top": "50%",
                            "bottom": "29.6%",
                            "width": "14.9%",
                            "height": "37.9%",
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "parent": "layout_2",
                    }
                },
                "parent": {"type": "footer", "id": "footer"},
            },
            "layout_3": {
                "type": "FLOW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "740px",
                            "right": "1120px",
                            "top": "85px",
                            "bottom": "145px",
                            "width": "380px",
                            "height": 80,
                        },
                        "widgetId": "widget_264",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_3",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "419px",
                            "right": "919px",
                            "top": "36px",
                            "bottom": "536px",
                            "width": "500px",
                            "height": "3150px",
                        },
                        "widgetId": "widget_266",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_3",
                    },
                },
                "parent": {"type": "pages", "id": "page_1"},
            },
            "layout_4": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "400px"},
                        "widgetId": "widget_9",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_4",
                    }
                },
                "parent": {"type": "widget", "id": "widget_8"},
            },
            "layout_5": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_11",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_5",
                    }
                },
                "parent": {"type": "widget", "id": "widget_10"},
            },
            "layout_6": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "70px",
                        },
                        "widgetId": "widget_13",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_6",
                    }
                },
                "parent": {"type": "widget", "id": "widget_12"},
            },
            "layout_7": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 8, "height": "400px"},
                        "widgetId": "widget_15",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_7",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_36",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_7",
                    },
                },
                "parent": {"type": "widget", "id": "widget_14"},
            },
            "layout_8": {
                "type": "ROW",
                "order": ["0", "4", "1", "3", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "600px"},
                        "widgetId": "widget_16",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_8",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 1, "height": "50px"},
                        "widgetId": "widget_22",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 3, "height": "600px"},
                        "widgetId": "widget_23",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_8",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 1, "height": "50px"},
                        "widgetId": "widget_29",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 3, "height": "600px"},
                        "widgetId": "widget_30",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_8",
                    },
                },
                "parent": {"type": "widget", "id": "widget_15"},
            },
            "layout_9": {
                "type": "COLUMN",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_17",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_9",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_18",
                        "parent": "layout_9",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_19",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_9",
                    },
                },
                "setting": {"space": 20},
                "parent": {"type": "widget", "id": "widget_16"},
            },
            "layout_10": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_20",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_10",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "70px"},
                        "widgetId": "widget_21",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_10",
                    },
                },
                "parent": {"type": "widget", "id": "widget_19"},
            },
            "layout_11": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_24",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_11",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_25",
                        "parent": "layout_11",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_26",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_11",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_23"},
            },
            "layout_12": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_27",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_12",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "70px"},
                        "widgetId": "widget_28",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_12",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_26"},
            },
            "layout_13": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_31",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_13",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_32",
                        "parent": "layout_13",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_33",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                        "parent": "layout_13",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_30"},
            },
            "layout_14": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_34",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_14",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "70px"},
                        "widgetId": "widget_35",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_14",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_33"},
            },
            "layout_15": {
                "type": "COLUMN",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_37",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_15",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_38",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                        "parent": "layout_15",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "200px", "height": "50.00px"},
                        "widgetId": "widget_39",
                        "setting": {
                            "autoProps": {"width": "CUSTOM", "height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "isPending": False,
                        "parent": "layout_15",
                    },
                },
                "setting": {
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                    "space": 20,
                    "justifyContent": "center",
                },
                "parent": {"type": "widget", "id": "widget_36"},
            },
            "layout_16": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 9, "height": "460.00px"},
                        "widgetId": "widget_41",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_16",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "600px"},
                        "widgetId": "widget_42",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_16",
                    },
                },
                "parent": {"type": "widget", "id": "widget_40"},
            },
            "layout_17": {
                "type": "COLUMN",
                "setting": {
                    "justifyContent": "center",
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                    "space": 20,
                },
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "40.00px", "height": "40.00px"},
                        "widgetId": "widget_43",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_17",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_44",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_17",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_45",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_17",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_46",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_17",
                    },
                },
                "parent": {"type": "widget", "id": "widget_42"},
            },
            "layout_18": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 10, "height": "70px"},
                        "widgetId": "widget_47",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_18",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": "10px", "width": "2px", "height": "300px"},
                        "widgetId": "widget_48",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_18",
                    },
                },
                "parent": {"type": "widget", "id": "widget_46"},
            },
            "layout_19": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_50",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_19",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_49"},
            },
            "layout_20": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "70px",
                        },
                        "widgetId": "widget_52",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_20",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_51"},
            },
            "layout_21": {
                "type": "ROW",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "600px"},
                        "widgetId": "widget_54",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_21",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "600px"},
                        "widgetId": "widget_79",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_21",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "391px"},
                        "widgetId": "widget_90",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_21",
                    },
                },
                "parent": {"type": "widget", "id": "widget_53"},
            },
            "layout_22": {
                "type": "COLUMN",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_55",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_22",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_74",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_22",
                    },
                },
                "setting": {"space": 40},
                "parent": {"type": "widget", "id": "widget_54"},
            },
            "layout_24": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_55"},
            },
            "layout_26": {
                "type": "FIXED",
                "content": {
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_67",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_68",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_69",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_73",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                },
                "order": ["4", "5", "6", "10"],
                "parent": {"type": "widget", "id": "widget_55"},
            },
            "layout_27": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_75",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_76",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_77",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_78",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_74"},
            },
            "layout_28": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_74"},
            },
            "layout_29": {
                "type": "COLUMN",
                "setting": {"space": 40},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_80",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_29",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_85",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_29",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_79"},
            },
            "layout_30": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_81",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_82",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_83",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_84",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_80"},
            },
            "layout_31": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_80"},
            },
            "layout_32": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_86",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_87",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_88",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_89",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_85"},
            },
            "layout_33": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_85"},
            },
            "layout_58": {
                "type": "FIXED",
                "content": {
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "59.6642%",
                            "bottom": 121,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_197",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "69.7132%",
                            "bottom": "22.4142%",
                            "width": "90.00%",
                            "height": "7.8727%",
                        },
                        "widgetId": "widget_198",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "55.36%",
                        },
                        "widgetId": "widget_199",
                        "setting": {},
                        "parent": "layout_58",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "vCenter": False, "hCenter": False},
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "75.9082%",
                            "bottom": "16.8080%",
                            "width": "90.00%",
                            "height": "7.2838%",
                        },
                        "widgetId": "widget_200",
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_201",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "4.9648%",
                            "top": "87.2095%",
                            "bottom": "4.7503%",
                            "width": "40.00%",
                            "height": 32,
                        },
                        "widgetId": "widget_202",
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "height": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_58",
                    },
                },
                "order": ["4", "5", "6", "7", "9", "10"],
                "parent": {"type": "widget", "id": "widget_90"},
            },
            "layout_59": {
                "type": "FIXED",
                "content": {},
                "order": [],
                "parent": {"type": "widget", "id": "widget_90"},
            },
            "layout_60": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_207",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_60",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_206"},
            },
            "layout_61": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "70px",
                        },
                        "widgetId": "widget_209",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_61",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_208"},
            },
            "layout_63": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_215",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_63",
                    }
                },
                "parent": {"type": "view", "id": "view_1"},
            },
            "layout_65": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600px"},
                        "widgetId": "widget_214",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_65",
                    }
                },
                "parent": {"type": "widget", "id": "widget_213"},
            },
            "layout_66": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "SECTION",
                        "setting": {"style": {}, "autoProps": {"height": "CUSTOM"}},
                        "bbox": {"width": "1860px", "height": "800.00px"},
                        "sectionId": "section_1",
                        "isPending": False,
                        "parent": "layout_66",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                        "bbox": {"width": "305px", "height": "43px"},
                        "widgetId": "widget_212",
                        "parent": "layout_66",
                    },
                },
                "order": ["1", "0"],
                "setting": {
                    "space": 30,
                    "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                },
                "parent": {"type": "widget", "id": "widget_214"},
            },
            "layout_67": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "400px"},
                        "widgetId": "widget_216",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_67",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "600px"},
                        "widgetId": "widget_217",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_67",
                    },
                },
                "parent": {"type": "widget", "id": "widget_215"},
            },
            "layout_68": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_216"},
            },
            "layout_69": {
                "type": "COLUMN",
                "setting": {"space": 40, "justifyContent": "flex-start"},
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "735px", "height": "300.00px"},
                        "widgetId": "widget_218",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_69",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_219",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "parent": "layout_69",
                    },
                },
                "parent": {"type": "widget", "id": "widget_217"},
            },
            "layout_70": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_221",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_70",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_220"},
            },
            "layout_71": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "NaNpx",
                            "top": "NaNpx",
                            "bottom": "NaNpx",
                            "width": "3px",
                            "height": "70px",
                        },
                        "widgetId": "widget_223",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_71",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_222"},
            },
            "layout_72": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_225",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_72",
                    }
                },
                "parent": {"type": "widget", "id": "widget_224"},
            },
            "layout_73": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "600px"},
                        "widgetId": "widget_227",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_73",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "520.00px"},
                        "widgetId": "widget_238",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_73",
                    },
                },
                "parent": {"type": "widget", "id": "widget_226"},
            },
            "layout_74": {
                "type": "COLUMN",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_228",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_74",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_229",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_74",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_232",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_74",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_235",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_74",
                    },
                },
                "setting": {
                    "padding": {"number": [30, 20, 30, 20], "unit": "px"},
                    "space": 20,
                },
                "parent": {"type": "widget", "id": "widget_227"},
            },
            "layout_75": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_230",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_75",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "70px"},
                        "widgetId": "widget_231",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_75",
                    },
                },
                "parent": {"type": "widget", "id": "widget_229"},
            },
            "layout_76": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_233",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_76",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "70px"},
                        "widgetId": "widget_234",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_76",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_232"},
            },
            "layout_77": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_236",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_77",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "70px"},
                        "widgetId": "widget_237",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_77",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_235"},
            },
            "layout_78": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_239",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_78",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_2"},
            },
            "layout_79": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "400px"},
                        "widgetId": "widget_240",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_79",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "600px"},
                        "widgetId": "widget_241",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_79",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_239"},
            },
            "layout_80": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_240"},
            },
            "layout_81": {
                "type": "COLUMN",
                "setting": {"space": 40, "justifyContent": "flex-start"},
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_243",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "isPending": False,
                        "parent": "layout_81",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "620px", "height": "275px"},
                        "widgetId": "widget_244",
                        "parent": "layout_81",
                    },
                },
                "order": ["1", "2"],
                "parent": {"type": "widget", "id": "widget_241"},
            },
            "layout_83": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_83",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_83",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_83",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_84": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_89": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_89",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_89",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_89",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_90": {
                "type": "FIXED",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "6.7%",
                            "top": "11.8%",
                            "width": "254px",
                            "height": "30px",
                            "right": "82px",
                            "bottom": "75.7%",
                        },
                        "widgetId": "widget_259",
                        "isPending": False,
                        "parent": "layout_90",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": False,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "6.8%",
                            "top": "30.5%",
                            "width": "86.3%",
                            "height": "32.5%",
                            "right": "6.9%",
                            "bottom": "89px",
                        },
                        "widgetId": "widget_260",
                        "isPending": False,
                        "parent": "layout_90",
                    },
                },
                "order": ["1", "2"],
                "setting": {"lockDescendants": True},
                "parent": {"type": "dialog", "id": "dialog_1"},
            },
            "layout_91": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                                "width": "CUSTOM",
                            },
                            "lockParent": False,
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "bbox": {
                            "left": "6.00%",
                            "top": "11.43%",
                            "width": "36.00px",
                            "height": "36.00px",
                            "right": 305,
                            "bottom": "65.8960%",
                        },
                        "widgetId": "widget_261",
                        "isPending": False,
                        "parent": "layout_91",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "20.00%",
                            "top": "12.00%",
                            "width": "254px",
                            "height": "30px",
                            "right": 30,
                            "bottom": "61.6541%",
                        },
                        "widgetId": "widget_262",
                        "isPending": False,
                        "parent": "layout_91",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": False,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "20.00%",
                            "top": "33.3214%",
                            "width": "70.8333%",
                            "height": "45.7143%",
                            "right": "8.61%",
                            "bottom": "37px",
                        },
                        "widgetId": "widget_263",
                        "isPending": False,
                        "parent": "layout_91",
                    },
                },
                "order": ["0", "1", "2"],
                "setting": {"lockDescendants": True},
                "parent": {"type": "dialog", "id": "dialog_2"},
            },
            "layout_92": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": "200px"},
                        "widgetId": "widget_265",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "aspectRatio": 1,
                        },
                        "parent": "layout_92",
                    }
                },
                "parent": {"type": "widget", "id": "widget_264"},
            },
            "layout_93": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "500px"},
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_93",
                    }
                },
                "parent": {"type": "widget", "id": "widget_266"},
            },
            "layout_94": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_267",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_94",
                    }
                },
                "parent": {"type": "view", "id": "view_3"},
            },
            "layout_95": {
                "type": "COLUMN",
                "order": ["0", "1", "2", "3", "4", "5"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_268",
                        "parent": "layout_95",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_271",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_95",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_299",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_95",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_313",
                        "parent": "layout_95",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_314",
                        "parent": "layout_95",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210.00px"},
                        "widgetId": "widget_315",
                        "parent": "layout_95",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                    },
                },
                "parent": {"type": "widget", "id": "widget_267"},
            },
            "layout_96": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_269",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_96",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "300px"},
                        "widgetId": "widget_270",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_96",
                    },
                },
                "parent": {"type": "widget", "id": "widget_268"},
            },
            "layout_97": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 5, "height": "391px"},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_97",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 7, "height": "275px"},
                        "widgetId": "widget_284",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_97",
                    },
                },
                "parent": {"type": "widget", "id": "widget_271"},
            },
            "layout_99": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_100": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_273",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_274",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_275",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_276",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_277",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_278",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_279",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_280",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_281",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_282",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "12": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_283",
                        "parent": "layout_100",
                    },
                },
                "order": ["0", "3", "4", "5", "8", "9", "10", "12", "6", "7", "11"],
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_102": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_102",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_102",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "31.48%",
                            "width": "214px",
                            "height": "57px",
                            "right": "5.58%",
                            "bottom": "15.74%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": True,
                                "right": False,
                                "bottom": False,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_102",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_103": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_108": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_108",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_108",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.9%",
                            "top": "35.0%",
                            "width": "309px",
                            "height": "87px",
                            "right": "5.6%",
                            "bottom": "15.7%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": True,
                                "right": False,
                                "bottom": False,
                                "left": True,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_108",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_109": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "540.00px"},
                        "widgetId": "widget_300",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_109",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "600px"},
                        "widgetId": "widget_301",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "parent": "layout_109",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_307",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_109",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_299"},
            },
            "layout_110": {
                "type": "COLUMN",
                "order": ["0", "1", "2", "3", "4"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_302",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_110",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_303",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_110",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_304",
                        "parent": "layout_110",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_305",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_110",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220.00px", "height": "50px"},
                        "widgetId": "widget_306",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_110",
                    },
                },
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "parent": {"type": "widget", "id": "widget_301"},
            },
            "layout_111": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_308",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_111",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_309",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_111",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_310",
                        "parent": "layout_111",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_311",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_111",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_312",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_111",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_307"},
            },
            "layout_112": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_314"},
            },
            "layout_113": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_316",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_113",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_4"},
            },
            "layout_114": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_317",
                        "parent": "layout_114",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_320",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_114",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_340",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_114",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_354",
                        "parent": "layout_114",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_355",
                        "parent": "layout_114",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": 235},
                        "widgetId": "widget_356",
                        "parent": "layout_114",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_316"},
            },
            "layout_115": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "300px"},
                        "widgetId": "widget_318",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_115",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "300px"},
                        "widgetId": "widget_319",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_115",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_317"},
            },
            "layout_116": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "391px"},
                        "widgetId": "widget_321",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_116",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "275px"},
                        "widgetId": "widget_333",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_116",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_320"},
            },
            "layout_117": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_322",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_323",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_324",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_325",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_326",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_327",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_328",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_329",
                        "parent": "layout_117",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_330",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_331",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_332",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_118": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_120": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_120",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_120",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_120",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_121": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_122": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "540.00px"},
                        "widgetId": "widget_341",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_122",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "600px"},
                        "widgetId": "widget_342",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_122",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_348",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_122",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_340"},
            },
            "layout_123": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_343",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_123",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_344",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_123",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_345",
                        "parent": "layout_123",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_346",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_123",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_347",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_123",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_342"},
            },
            "layout_124": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_349",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_124",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_350",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_124",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_351",
                        "parent": "layout_124",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_352",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_124",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_353",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_124",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_348"},
            },
            "layout_125": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_355"},
            },
            "layout_126": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_357",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_126",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_5"},
            },
            "layout_127": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_358",
                        "parent": "layout_127",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_361",
                        "isPending": False,
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"}
                        },
                        "parent": "layout_127",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_381",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_127",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_395",
                        "parent": "layout_127",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "440.00px"},
                        "widgetId": "widget_396",
                        "parent": "layout_127",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_397",
                        "parent": "layout_127",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_357"},
            },
            "layout_128": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "300px"},
                        "widgetId": "widget_359",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_128",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "300px"},
                        "widgetId": "widget_360",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_128",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_358"},
            },
            "layout_129": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "391px"},
                        "widgetId": "widget_362",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_129",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "275px"},
                        "widgetId": "widget_374",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_129",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_361"},
            },
            "layout_130": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_363",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_364",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_365",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_366",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_367",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_368",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_369",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_370",
                        "parent": "layout_130",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_371",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_372",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_373",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_131": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_133": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_133",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_133",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_133",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_134": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_135": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "620.00px"},
                        "widgetId": "widget_382",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_135",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": "600px"},
                        "widgetId": "widget_383",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_135",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "600px"},
                        "widgetId": "widget_389",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_135",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_381"},
            },
            "layout_136": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_384",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_136",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_385",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_136",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_386",
                        "parent": "layout_136",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_387",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_136",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_388",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_136",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_383"},
            },
            "layout_137": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_390",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_137",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_391",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_137",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_392",
                        "parent": "layout_137",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_393",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_137",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_394",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_137",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_389"},
            },
            "layout_138": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_396"},
            },
            "layout_140": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_140",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_140",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_140",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_143": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_143",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_143",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_143",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_144": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "740px",
                            "right": "1120px",
                            "top": "85px",
                            "bottom": "145px",
                            "width": "380px",
                            "height": 80,
                        },
                        "widgetId": "widget_412",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_144",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "419px",
                            "right": "919px",
                            "top": "36px",
                            "bottom": "536px",
                            "width": "500px",
                            "height": "3100px",
                        },
                        "widgetId": "widget_413",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_144",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "pages", "id": "page_1"},
            },
            "layout_145": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "200px"},
                        "widgetId": "widget_265",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "aspectRatio": 1,
                        },
                        "parent": "layout_145",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_412"},
            },
            "layout_146": {
                "type": "ROW",
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "500px"},
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_146",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_413"},
            },
            "layout_147": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_414",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_147",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_3"},
            },
            "layout_148": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_415",
                        "parent": "layout_148",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_416",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_148",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_417",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_148",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_313",
                        "parent": "layout_148",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_314",
                        "parent": "layout_148",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_315",
                        "parent": "layout_148",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "setting": {"space": 0},
                "parent": {"type": "widget", "id": "widget_414"},
            },
            "layout_149": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_314"},
            },
            "layout_150": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_269",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_150",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "300px"},
                        "widgetId": "widget_270",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_150",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_415"},
            },
            "layout_151": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 5, "height": "391px"},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_151",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 7, "height": "275px"},
                        "widgetId": "widget_284",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_151",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_416"},
            },
            "layout_152": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_273",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_274",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_275",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_276",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_279",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_280",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_281",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_283",
                        "parent": "layout_152",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_277",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_278",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_282",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_153": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_154": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_154",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30.00%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": "AUTO"},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_154",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35%",
                            "top": "2.00%",
                            "width": "162px",
                            "height": "150px",
                            "right": "5.5%",
                            "bottom": "6.200000000000033%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": "AUTO",
                                "right": False,
                                "bottom": True,
                                "left": True,
                                "top": False,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": True,
                        "parent": "layout_154",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_155": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_155",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_155",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "31.48%",
                            "width": "214px",
                            "height": "57px",
                            "right": "5.58%",
                            "bottom": "15.74%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": True,
                                "right": False,
                                "bottom": False,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_155",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_156": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_157": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_300",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_157",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "600px"},
                        "widgetId": "widget_418",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "parent": "layout_157",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_419",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_157",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_417"},
            },
            "layout_158": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_302",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_158",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_303",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_158",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_304",
                        "parent": "layout_158",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_305",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_158",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_306",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_158",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_418"},
            },
            "layout_159": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_308",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_159",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_309",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_159",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_310",
                        "parent": "layout_159",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_311",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_159",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_312",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_159",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_419"},
            },
            "layout_160": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_420",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_160",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_4"},
            },
            "layout_161": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_421",
                        "parent": "layout_161",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_422",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_161",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_423",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_161",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_354",
                        "parent": "layout_161",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_355",
                        "parent": "layout_161",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": 235},
                        "widgetId": "widget_356",
                        "parent": "layout_161",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "setting": {"space": 0},
                "parent": {"type": "widget", "id": "widget_420"},
            },
            "layout_162": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_355"},
            },
            "layout_163": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "300px"},
                        "widgetId": "widget_318",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_163",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "300px"},
                        "widgetId": "widget_319",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_163",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_421"},
            },
            "layout_164": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "391px"},
                        "widgetId": "widget_321",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_164",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "275px"},
                        "widgetId": "widget_333",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_164",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_422"},
            },
            "layout_165": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_322",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_323",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_324",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_325",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_326",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_327",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_328",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_329",
                        "parent": "layout_165",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_330",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_331",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_332",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_166": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_167": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_167",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "24.2%",
                            "width": "47.4%",
                            "height": "35.0%",
                            "right": "28.4%",
                            "top": "40px",
                            "bottom": "75px",
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_167",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "195px",
                            "top": "71.5%",
                            "width": "27.9%",
                            "height": "17.5%",
                            "right": "10.6%",
                            "bottom": "11.0%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_167",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_168": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_168",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_168",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_168",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_169": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_170": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_341",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_170",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "600px"},
                        "widgetId": "widget_424",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_170",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_425",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_170",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_423"},
            },
            "layout_171": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_343",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_171",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_344",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_171",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_345",
                        "parent": "layout_171",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_346",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_171",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_347",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_171",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_424"},
            },
            "layout_172": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_349",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_172",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_350",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_172",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_351",
                        "parent": "layout_172",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_352",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_172",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_353",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_172",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_425"},
            },
            "layout_173": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_426",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_173",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_5"},
            },
            "layout_174": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_427",
                        "parent": "layout_174",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_428",
                        "isPending": False,
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"}
                        },
                        "parent": "layout_174",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_429",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_174",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_395",
                        "parent": "layout_174",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_396",
                        "parent": "layout_174",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_397",
                        "parent": "layout_174",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "setting": {"space": 0},
                "parent": {"type": "widget", "id": "widget_426"},
            },
            "layout_175": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_396"},
            },
            "layout_176": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "300px"},
                        "widgetId": "widget_359",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_176",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "300px"},
                        "widgetId": "widget_360",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_176",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_427"},
            },
            "layout_177": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "391px"},
                        "widgetId": "widget_362",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_177",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "275px"},
                        "widgetId": "widget_374",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_177",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_428"},
            },
            "layout_178": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_363",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_364",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_365",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_366",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_367",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_368",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_369",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_370",
                        "parent": "layout_178",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_371",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_372",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_373",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_179": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_180": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_180",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "22.6%",
                            "width": "71.7%",
                            "height": "35.0%",
                            "right": "5.7%",
                            "top": "36px",
                            "bottom": "80px",
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_180",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "240px",
                            "top": "71.7%",
                            "width": "27.9%",
                            "height": "18.2%",
                            "right": "5.9%",
                            "bottom": "10.1%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_180",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_181": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_181",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_181",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_181",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_182": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_183": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "300px"},
                        "widgetId": "widget_382",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_183",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": "600px"},
                        "widgetId": "widget_430",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_183",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "600px"},
                        "widgetId": "widget_431",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_183",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_429"},
            },
            "layout_184": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_384",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_184",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_385",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_184",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_386",
                        "parent": "layout_184",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_387",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_184",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_388",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_184",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_430"},
            },
            "layout_185": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_390",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_185",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_391",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_185",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_392",
                        "parent": "layout_185",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_393",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_185",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_394",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_185",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_431"},
            },
            "layout_186": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20.00px",
                            "right": "94.8%",
                            "top": "40.00px",
                            "bottom": "54.3%",
                            "width": "40px",
                            "height": "40px",
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_186",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "70.00px",
                            "right": "64.1%",
                            "top": "40px",
                            "bottom": "49.6%",
                            "width": "360px",
                            "height": "41px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.5%",
                            "right": "20.00px",
                            "top": "30.00px",
                            "bottom": "-52.8%",
                            "width": "50.00px",
                            "height": "50.00px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "65.7%",
                            "right": "80.00px",
                            "top": "40.00px",
                            "bottom": "50.3%",
                            "width": "32.00px",
                            "height": "32.00px",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.6%",
                            "right": "130.00px",
                            "top": "45.00px",
                            "bottom": "24.1%",
                            "width": "24.00px",
                            "height": "24.00px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20.00px",
                            "right": "76.6%",
                            "top": "68.7%",
                            "bottom": "0px",
                            "width": "21.1%",
                            "height": "31.3%",
                        },
                        "widgetId": "widget_6",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": True,
                                "bottom": False,
                                "width": "AUTO",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_186",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "header", "id": "header"},
            },
            "layout_187": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "723px",
                            "right": "1073px",
                            "top": "-78px",
                            "bottom": "322px",
                            "width": "350px",
                            "height": "60px",
                        },
                        "widgetId": "widget_432",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_187",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "145px",
                            "right": "505px",
                            "top": "79px",
                            "bottom": "149px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_433",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_187",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "105px",
                            "right": "465px",
                            "top": "156px",
                            "bottom": "226px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_434",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_187",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-71px",
                            "right": "729px",
                            "top": "88px",
                            "bottom": "488px",
                            "width": "800px",
                            "height": "360px",
                        },
                        "widgetId": "widget_435",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_187",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "179px",
                            "right": "629px",
                            "top": "413px",
                            "bottom": "713px",
                            "width": "450px",
                            "height": "400px",
                        },
                        "widgetId": "widget_444",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_187",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1168px",
                            "bottom": "1233px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_447",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1292px",
                            "bottom": "1335px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_448",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "126px",
                            "right": "426px",
                            "top": "980px",
                            "bottom": "1580px",
                            "width": "300px",
                            "height": 386,
                        },
                        "widgetId": "widget_449",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_187",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1908px",
                            "bottom": "1972px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_452",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1958px",
                            "bottom": "2011px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_453",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "316px",
                            "right": "616px",
                            "top": "2260px",
                            "bottom": "2860px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_454",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_187",
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3004px",
                            "bottom": "3068px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_460",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3044px",
                            "bottom": "3097px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_461",
                        "setting": {"heightMode": "auto"},
                        "isPending": False,
                        "parent": "layout_187",
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "131px",
                            "right": "731px",
                            "top": "2900px",
                            "bottom": "3300px",
                            "width": "600px",
                            "height": 420,
                        },
                        "widgetId": "widget_462",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_187",
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "78px",
                            "right": "378px",
                            "top": "3248px",
                            "bottom": "3848px",
                            "width": "300px",
                            "height": "400px",
                        },
                        "widgetId": "widget_463",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_187",
                    },
                },
                "order": [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                ],
                "parent": {"type": "pages", "id": "page_0"},
            },
            "layout_188": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "400px"},
                        "widgetId": "widget_9",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_188",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_432"},
            },
            "layout_189": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_11",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_189",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_433"},
            },
            "layout_190": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_13",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_190",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_434"},
            },
            "layout_191": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 8, "height": "400px"},
                        "widgetId": "widget_436",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_191",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_443",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_191",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_435"},
            },
            "layout_192": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "600px"},
                        "widgetId": "widget_437",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_192",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 3, "height": "600px"},
                        "widgetId": "widget_439",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_192",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 1, "height": "50px"},
                        "widgetId": "widget_22",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_192",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 1, "height": "50px"},
                        "widgetId": "widget_29",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_192",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 3, "height": "600px"},
                        "widgetId": "widget_441",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_192",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_436"},
            },
            "layout_193": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_17",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_193",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_18",
                        "parent": "layout_193",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_438",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_193",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_437"},
            },
            "layout_194": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_20",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_194",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_21",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_194",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_438"},
            },
            "layout_195": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_31",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_195",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_32",
                        "parent": "layout_195",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_440",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "isPending": False,
                        "parent": "layout_195",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_439"},
            },
            "layout_196": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_34",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_196",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_35",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_196",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_440"},
            },
            "layout_197": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_24",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_197",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_25",
                        "parent": "layout_197",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_442",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_197",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_441"},
            },
            "layout_198": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "300px"},
                        "widgetId": "widget_27",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_198",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_28",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_198",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_442"},
            },
            "layout_199": {
                "type": "COLUMN",
                "setting": {
                    "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                    "space": 20,
                    "justifyContent": "center",
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_37",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_199",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_38",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                        "parent": "layout_199",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "140.00px", "height": "40.00px"},
                        "widgetId": "widget_39",
                        "setting": {
                            "autoProps": {"width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "isPending": False,
                        "parent": "layout_199",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_443"},
            },
            "layout_200": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 9, "height": "300px"},
                        "widgetId": "widget_41",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_200",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "600px"},
                        "widgetId": "widget_445",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_200",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_444"},
            },
            "layout_201": {
                "type": "COLUMN",
                "setting": {
                    "justifyContent": "center",
                    "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                    "space": 20,
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "40.00px", "height": "40.00px"},
                        "widgetId": "widget_43",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_201",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "60.00px"},
                        "widgetId": "widget_44",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_201",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "40.00px"},
                        "widgetId": "widget_45",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_201",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_446",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_201",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_445"},
            },
            "layout_202": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_47",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_202",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": "10px", "width": "2px", "height": "300px"},
                        "widgetId": "widget_48",
                        "setting": {"heightMode": "fit"},
                        "isPending": True,
                        "parent": "layout_202",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_446"},
            },
            "layout_203": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_50",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_203",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_447"},
            },
            "layout_204": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_52",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_204",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_448"},
            },
            "layout_205": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "360.00px"},
                        "widgetId": "widget_450",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_205",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 3, "height": "360.00px"},
                        "widgetId": "widget_451",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_205",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "360.00px"},
                        "widgetId": "widget_90",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_205",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_449"},
            },
            "layout_206": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "59.6642%",
                            "bottom": 121,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_197",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_206",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "69.7132%",
                            "bottom": "22.4142%",
                            "width": "90.00%",
                            "height": "7.8727%",
                        },
                        "widgetId": "widget_198",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_206",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "55.36%",
                        },
                        "widgetId": "widget_199",
                        "setting": {},
                        "parent": "layout_206",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "vCenter": False, "hCenter": False},
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "75.9082%",
                            "bottom": "16.8080%",
                            "width": "90.00%",
                            "height": "7.2838%",
                        },
                        "widgetId": "widget_200",
                        "isPending": False,
                        "parent": "layout_206",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_201",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_206",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "4.9648%",
                            "top": "87.2095%",
                            "bottom": "4.7503%",
                            "width": "40.00%",
                            "height": 32,
                        },
                        "widgetId": "widget_202",
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "height": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_206",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_90"},
            },
            "layout_207": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_90"},
            },
            "layout_208": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "170.00px"},
                        "widgetId": "widget_55",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_208",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "170.00px"},
                        "widgetId": "widget_74",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_208",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_450"},
            },
            "layout_209": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_67",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_209",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "38.6%",
                            "right": "0.8%",
                            "top": "39.3%",
                            "bottom": "6.00%",
                            "width": "60.6%",
                            "height": "90px",
                        },
                        "widgetId": "widget_68",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                                "top": True,
                                "bottom": False,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_209",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_69",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_209",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_73",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": True,
                        "parent": "layout_209",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_55"},
            },
            "layout_210": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_55"},
            },
            "layout_211": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_75",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_211",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "6.00%",
                            "width": "60.6343%",
                            "height": "90.00px",
                        },
                        "widgetId": "widget_76",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_211",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_77",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_211",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_78",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": True,
                        "parent": "layout_211",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_74"},
            },
            "layout_212": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_74"},
            },
            "layout_213": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "170.00px"},
                        "widgetId": "widget_80",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_213",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "170.00px"},
                        "widgetId": "widget_85",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_213",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_451"},
            },
            "layout_214": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_81",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_214",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.2%",
                            "right": "0.2%",
                            "top": "35.8%",
                            "bottom": "6.00%",
                            "width": "60.6%",
                            "height": "90.00px",
                        },
                        "widgetId": "widget_82",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                                "top": True,
                                "bottom": False,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_214",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_83",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_214",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_84",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": True,
                        "parent": "layout_214",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_80"},
            },
            "layout_215": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_80"},
            },
            "layout_216": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_86",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_216",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "6.00%",
                            "width": "60.6343%",
                            "height": "90.00px",
                        },
                        "widgetId": "widget_87",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_216",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_88",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_216",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_89",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": True,
                        "parent": "layout_216",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_85"},
            },
            "layout_217": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_85"},
            },
            "layout_218": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_207",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_218",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_452"},
            },
            "layout_219": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_209",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_219",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_453"},
            },
            "layout_220": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600px"},
                        "widgetId": "widget_455",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_220",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_454"},
            },
            "layout_221": {
                "type": "COLUMN",
                "setting": {
                    "space": 30,
                    "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                        "bbox": {"width": "305px", "height": "43px"},
                        "widgetId": "widget_212",
                        "parent": "layout_221",
                    },
                    "1": {
                        "type": "SECTION",
                        "setting": {"style": {}, "autoProps": {"height": "CUSTOM"}},
                        "bbox": {"width": "1860px", "height": "800.00px"},
                        "sectionId": "section_1",
                        "isPending": False,
                        "parent": "layout_221",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_455"},
            },
            "layout_222": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_456",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_222",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_1"},
            },
            "layout_223": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "400px"},
                        "widgetId": "widget_216",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_223",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "600px"},
                        "widgetId": "widget_457",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_223",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_456"},
            },
            "layout_224": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_216"},
            },
            "layout_225": {
                "type": "COLUMN",
                "setting": {"space": 20, "justifyContent": "flex-start"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "735px", "height": "300.00px"},
                        "widgetId": "widget_218",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_225",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_219",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "parent": "layout_225",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_457"},
            },
            "layout_226": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_458",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_226",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_2"},
            },
            "layout_227": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "400px"},
                        "widgetId": "widget_240",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_227",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "600px"},
                        "widgetId": "widget_459",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_227",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_458"},
            },
            "layout_228": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_240"},
            },
            "layout_229": {
                "type": "COLUMN",
                "setting": {"space": 20, "justifyContent": "flex-start"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_243",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "isPending": False,
                        "parent": "layout_229",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "620px", "height": "275px"},
                        "widgetId": "widget_244",
                        "parent": "layout_229",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_459"},
            },
            "layout_230": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_230",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_230",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_230",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_231": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_231",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_231",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_231",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_232": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_233": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_221",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_233",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_460"},
            },
            "layout_234": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_223",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_234",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_461"},
            },
            "layout_235": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_225",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_235",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_462"},
            },
            "layout_236": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "600px"},
                        "widgetId": "widget_464",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_236",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "300px"},
                        "widgetId": "widget_238",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_236",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_463"},
            },
            "layout_237": {
                "type": "COLUMN",
                "setting": {
                    "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                    "space": 10,
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "60.00px"},
                        "widgetId": "widget_228",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_237",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_465",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_237",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_466",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_237",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_467",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_237",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "widget", "id": "widget_464"},
            },
            "layout_238": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_230",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_238",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_231",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_238",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_465"},
            },
            "layout_239": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_233",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_239",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_234",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_239",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_466"},
            },
            "layout_240": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_236",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "isPending": True,
                        "parent": "layout_240",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "70px"},
                        "widgetId": "widget_237",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_240",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_467"},
            },
            "layout_241": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10.00px",
                            "right": "94.8%",
                            "top": "30.00px",
                            "bottom": "54.3%",
                            "width": "40px",
                            "height": "40px",
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_241",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50.00px",
                            "right": "64.1%",
                            "top": "30.00px",
                            "bottom": "49.6%",
                            "width": "240.00px",
                            "height": "41px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_241",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.5%",
                            "right": "10.00px",
                            "top": "20.00px",
                            "bottom": "-52.8%",
                            "width": "50.00px",
                            "height": "50.00px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_241",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "85.8%",
                            "right": "19px",
                            "top": "87px",
                            "bottom": "25.9%",
                            "width": "32px",
                            "height": "32px",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_241",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.8%",
                            "right": "70px",
                            "top": "95px",
                            "bottom": "29.09999999999999%",
                            "width": "24px",
                            "height": "24px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_241",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "30px",
                            "right": "76.6%",
                            "top": "68.7%",
                            "bottom": "0px",
                            "width": "21.1%",
                            "height": "31.3%",
                        },
                        "widgetId": "widget_6",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": True,
                                "bottom": False,
                                "width": "AUTO",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_241",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "header", "id": "header"},
            },
            "layout_242": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "740px",
                            "right": "1120px",
                            "top": "85px",
                            "bottom": "145px",
                            "width": "380px",
                            "height": 80,
                        },
                        "widgetId": "widget_468",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_242",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "419px",
                            "right": "919px",
                            "top": "36px",
                            "bottom": "536px",
                            "width": "500px",
                            "height": "3100px",
                        },
                        "widgetId": "widget_469",
                        "setting": {
                            "heightMode": "fixed",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_242",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "pages", "id": "page_1"},
            },
            "layout_243": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_488",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_243",
                    }
                },
                "order": ["1"],
                "parent": {"type": "widget", "id": "widget_468"},
            },
            "layout_244": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_489",
                        "setting": {
                            "autoProps": {"height": "STRETCH"},
                            "autoHeight": True,
                        },
                        "parent": "layout_244",
                    }
                },
                "order": ["1"],
                "parent": {"type": "widget", "id": "widget_469"},
            },
            "layout_245": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_470",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_245",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_3"},
            },
            "layout_246": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "335px", "height": "600.00px"},
                        "widgetId": "widget_471",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_246",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "340px", "height": 676},
                        "widgetId": "widget_472",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_246",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "1500.00px"},
                        "widgetId": "widget_473",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_246",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_313",
                        "parent": "layout_246",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_314",
                        "parent": "layout_246",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_315",
                        "parent": "layout_246",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_470"},
            },
            "layout_247": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_314"},
            },
            "layout_248": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_269",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_248",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "300px"},
                        "widgetId": "widget_270",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_248",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_471"},
            },
            "layout_249": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "391px"},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_249",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "275px"},
                        "widgetId": "widget_284",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_249",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_472"},
            },
            "layout_250": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_273",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_274",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_275",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_276",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_279",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_280",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_281",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_283",
                        "parent": "layout_250",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_277",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_278",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_282",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_251": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_252": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_252",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_252",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.9%",
                            "top": "0.0%",
                            "width": "621px",
                            "height": "150px",
                            "right": "5.6%",
                            "bottom": "15.7%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": True,
                                "right": False,
                                "bottom": True,
                                "left": True,
                                "top": False,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": True,
                        "parent": "layout_252",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_253": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_253",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_253",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "31.48%",
                            "width": "214px",
                            "height": "57px",
                            "right": "5.58%",
                            "bottom": "15.74%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": True,
                                "right": False,
                                "bottom": False,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_253",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_254": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_255": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_300",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_255",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "600px"},
                        "widgetId": "widget_474",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "parent": "layout_255",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "600px"},
                        "widgetId": "widget_475",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_255",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_473"},
            },
            "layout_256": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_302",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_256",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_303",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_256",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_304",
                        "parent": "layout_256",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_305",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_256",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_306",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_256",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_474"},
            },
            "layout_257": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_308",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_257",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_309",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_257",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_310",
                        "parent": "layout_257",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_311",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_257",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_312",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_257",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_475"},
            },
            "layout_258": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_476",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_258",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_4"},
            },
            "layout_259": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_477",
                        "parent": "layout_259",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "660.00px"},
                        "widgetId": "widget_478",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_259",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "1700.00px"},
                        "widgetId": "widget_479",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_259",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_354",
                        "parent": "layout_259",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_355",
                        "parent": "layout_259",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": 235},
                        "widgetId": "widget_356",
                        "parent": "layout_259",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_476"},
            },
            "layout_260": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_355"},
            },
            "layout_261": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_318",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_261",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "300px"},
                        "widgetId": "widget_319",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_261",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_477"},
            },
            "layout_262": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "391px"},
                        "widgetId": "widget_321",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_262",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "275px"},
                        "widgetId": "widget_333",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_262",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_478"},
            },
            "layout_263": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_322",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_323",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_324",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_325",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_326",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_327",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_328",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_329",
                        "parent": "layout_263",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_330",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_331",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_332",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_264": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_265": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_265",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_265",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_265",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_266": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_266",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_266",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_266",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_267": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_268": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_341",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_268",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "600px"},
                        "widgetId": "widget_480",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_268",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "600px"},
                        "widgetId": "widget_481",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_268",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_479"},
            },
            "layout_269": {
                "type": "COLUMN",
                "setting": {
                    "space": 10,
                    "padding": {"number": [0, 0, 10, 0], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_343",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_269",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_344",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_269",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_345",
                        "parent": "layout_269",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_346",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_269",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_347",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_269",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_480"},
            },
            "layout_270": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_349",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                        "parent": "layout_270",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_350",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_270",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_351",
                        "parent": "layout_270",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_352",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_270",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_353",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_270",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_481"},
            },
            "layout_271": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_482",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_271",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_5"},
            },
            "layout_272": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_483",
                        "parent": "layout_272",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "700.00px"},
                        "widgetId": "widget_484",
                        "isPending": False,
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"}
                        },
                        "parent": "layout_272",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "1500.00px"},
                        "widgetId": "widget_485",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_272",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_395",
                        "parent": "layout_272",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_396",
                        "parent": "layout_272",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_397",
                        "parent": "layout_272",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_482"},
            },
            "layout_273": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_396"},
            },
            "layout_274": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_359",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_274",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "300px"},
                        "widgetId": "widget_360",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_274",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_483"},
            },
            "layout_275": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "391px"},
                        "widgetId": "widget_362",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_275",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "275px"},
                        "widgetId": "widget_374",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_275",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_484"},
            },
            "layout_276": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_363",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_364",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_365",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_366",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_367",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_368",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_369",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_370",
                        "parent": "layout_276",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_371",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_372",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_373",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_277": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_278": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_278",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "18.7%",
                            "width": "47.4%",
                            "height": "35.0%",
                            "right": "33.9%",
                            "top": "33px",
                            "bottom": "33px",
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_278",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "265px",
                            "top": "33.5%",
                            "width": "27.9%",
                            "height": "34.9%",
                            "right": "6.1%",
                            "bottom": "31.6%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_278",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_279": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_279",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_279",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_279",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_280": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_281": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_382",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_281",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "600px"},
                        "widgetId": "widget_486",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_281",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "600px"},
                        "widgetId": "widget_487",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_281",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_485"},
            },
            "layout_282": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_384",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_282",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_385",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_282",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_386",
                        "parent": "layout_282",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_387",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_282",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_388",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_282",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_486"},
            },
            "layout_283": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_390",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_283",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_391",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_283",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_392",
                        "parent": "layout_283",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_393",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_283",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_394",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_283",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_487"},
            },
            "layout_284": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "200px"},
                        "widgetId": "widget_265",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "aspectRatio": 1,
                        },
                        "parent": "layout_284",
                        "isPending": True,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "334px", "height": "280.00px"},
                        "widgetId": "widget_548",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "aspectRatio": 1,
                        },
                        "parent": "layout_284",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_488"},
            },
            "layout_285": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {"width": "auto", "height": "500px"},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_285",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_489"},
            },
            "layout_286": {
                "type": "FIXED",
                "parent": {"type": "footer", "id": "footer"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "42.6%",
                            "top": "50%",
                            "bottom": "29.6%",
                            "width": "90.00%",
                            "height": "37.9%",
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "parent": "layout_286",
                    }
                },
                "order": ["0"],
            },
            "layout_287": {
                "type": "FIXED",
                "parent": {"type": "footer", "id": "footer"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "42.6%",
                            "top": "50%",
                            "bottom": "29.6%",
                            "width": "100.00%",
                            "height": "37.9%",
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "parent": "layout_287",
                    }
                },
                "order": ["0"],
            },
            "layout_288": {
                "type": "FLOW",
                "parent": {"type": "pages", "id": "page_0"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "723px",
                            "right": "1073px",
                            "top": "-78px",
                            "bottom": "322px",
                            "width": "350px",
                            "height": "60px",
                        },
                        "widgetId": "widget_490",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "145px",
                            "right": "505px",
                            "top": "79px",
                            "bottom": "149px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_491",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "105px",
                            "right": "465px",
                            "top": "156px",
                            "bottom": "226px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_492",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "-71px",
                            "right": "729px",
                            "top": "88px",
                            "bottom": "488px",
                            "width": "800px",
                            "height": 420,
                        },
                        "widgetId": "widget_493",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "179px",
                            "right": "629px",
                            "top": "413px",
                            "bottom": "713px",
                            "width": "450px",
                            "height": "500px",
                        },
                        "widgetId": "widget_502",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1168px",
                            "bottom": "1233px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_505",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1292px",
                            "bottom": "1335px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_506",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "126px",
                            "right": "426px",
                            "top": "980px",
                            "bottom": "1580px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_507",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1908px",
                            "bottom": "1972px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_510",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "1958px",
                            "bottom": "2011px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_511",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "316px",
                            "right": "616px",
                            "top": "2260px",
                            "bottom": "2860px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_512",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3004px",
                            "bottom": "3068px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_518",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "12": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0px",
                            "right": "1920px",
                            "top": "3044px",
                            "bottom": "3097px",
                            "width": "360px",
                            "height": 90,
                        },
                        "widgetId": "widget_519",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_0",
                    },
                    "13": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "131px",
                            "right": "731px",
                            "top": "2900px",
                            "bottom": "3300px",
                            "width": "600px",
                            "height": 420,
                        },
                        "widgetId": "widget_520",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                    "14": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "78px",
                            "right": "378px",
                            "top": "3248px",
                            "bottom": "3848px",
                            "width": "300px",
                            "height": 620,
                        },
                        "widgetId": "widget_521",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_0",
                    },
                },
                "order": [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                ],
            },
            "layout_289": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_8"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_526",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_290": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_10"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_527",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_291": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_12"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_528",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_292": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_14"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_529",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_293": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_15"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "600px"},
                        "widgetId": "widget_495",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_8",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "600px"},
                        "widgetId": "widget_497",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_8",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "50px"},
                        "widgetId": "widget_22",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                        "isPending": True,
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"left": 36, "width": 12, "height": "50px"},
                        "widgetId": "widget_29",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_8",
                        "isPending": True,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"left": 48, "width": 12, "height": "600px"},
                        "widgetId": "widget_499",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_8",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
            },
            "layout_294": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "parent": {"type": "widget", "id": "widget_16"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_17",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_9",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_18",
                        "parent": "layout_9",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_496",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_9",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_295": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_19"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_530",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_296": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "parent": {"type": "widget", "id": "widget_30"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_31",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_13",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_32",
                        "parent": "layout_13",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_498",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "isPending": False,
                        "parent": "layout_13",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_297": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_33"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_531",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_298": {
                "type": "COLUMN",
                "setting": {"space": 20},
                "parent": {"type": "widget", "id": "widget_23"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_24",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_11",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_25",
                        "parent": "layout_11",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_500",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_11",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_299": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_26"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_532",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_300": {
                "type": "COLUMN",
                "setting": {
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                    "space": 20,
                    "justifyContent": "center",
                },
                "parent": {"type": "widget", "id": "widget_36"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_37",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_15",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_38",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                        "parent": "layout_15",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "200px", "height": "50.00px"},
                        "widgetId": "widget_39",
                        "setting": {
                            "autoProps": {"width": "CUSTOM", "height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "isPending": False,
                        "parent": "layout_15",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_301": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_40"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_533",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_302": {
                "type": "COLUMN",
                "setting": {
                    "justifyContent": "center",
                    "padding": {"number": [40, 20, 40, 20], "unit": "px"},
                    "space": 40,
                },
                "parent": {"type": "widget", "id": "widget_42"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "40.00px", "height": "40.00px"},
                        "widgetId": "widget_43",
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_17",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_44",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_17",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_45",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_17",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "100.00px"},
                        "widgetId": "widget_504",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_17",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_303": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_46"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_534",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_304": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_49"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_535",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_305": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_51"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_536",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_306": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_53"},
                "content": {
                    "3": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_537",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["3"],
            },
            "layout_307": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_90"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "59.6642%",
                            "bottom": 121,
                            "width": "90.00%",
                            "height": "10.0503%",
                        },
                        "widgetId": "widget_197",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "69.7132%",
                            "bottom": "22.4142%",
                            "width": "90.00%",
                            "height": "7.8727%",
                        },
                        "widgetId": "widget_198",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": 0,
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "55.36%",
                        },
                        "widgetId": "widget_199",
                        "setting": {},
                        "parent": "layout_58",
                    },
                    "3": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "vCenter": False, "hCenter": False},
                        "bbox": {
                            "left": "5.00%",
                            "right": "5.0360%",
                            "top": "75.9082%",
                            "bottom": "16.8080%",
                            "width": "90.00%",
                            "height": "7.2838%",
                        },
                        "widgetId": "widget_200",
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 0,
                            "right": "75.8993%",
                            "top": 0,
                            "bottom": "93.4673%",
                            "width": 67,
                            "height": 26,
                        },
                        "widgetId": "widget_201",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_58",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "4.9648%",
                            "top": "87.2095%",
                            "bottom": "4.7503%",
                            "width": "40.00%",
                            "height": 32,
                        },
                        "widgetId": "widget_202",
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {
                                "width": "CUSTOM",
                                "left": False,
                                "right": True,
                                "height": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_58",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
            },
            "layout_308": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_90"},
            },
            "layout_309": {
                "type": "COLUMN",
                "setting": {"space": 40},
                "parent": {"type": "widget", "id": "widget_54"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_55",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_22",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_74",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_22",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_310": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_55"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_67",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_68",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_69",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_26",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_73",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_26",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_311": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_55"},
            },
            "layout_312": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_74"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_75",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_76",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_77",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_27",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_78",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_27",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_313": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_74"},
            },
            "layout_314": {
                "type": "COLUMN",
                "setting": {"space": 40},
                "parent": {"type": "widget", "id": "widget_79"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_80",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_29",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "280.00px"},
                        "widgetId": "widget_85",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_29",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_315": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_80"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_81",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_82",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_83",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_30",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_84",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_30",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_316": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_80"},
            },
            "layout_317": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_85"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "212px",
                            "right": 0,
                            "top": "7.0470%",
                            "bottom": "156px",
                            "width": "60.6238%",
                            "height": "28px",
                        },
                        "widgetId": "widget_86",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"right": False, "left": True},
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "39.3676%",
                            "right": "0.00%",
                            "top": "25.4996%",
                            "bottom": "48.9955%",
                            "width": "60.6343%",
                            "height": "50.00%",
                        },
                        "widgetId": "widget_87",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "height": "CUSTOM",
                                "width": "STRETCH",
                            },
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "right": "70.1493%",
                            "top": "0.00%",
                            "bottom": 0,
                            "width": "37.1747%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_88",
                        "setting": {},
                        "isPending": False,
                        "parent": "layout_32",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "81.2%",
                            "right": "20.00px",
                            "top": "79.2%",
                            "bottom": "20.00px",
                            "width": "40.00px",
                            "height": "40.00px",
                        },
                        "widgetId": "widget_89",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "bottom": False,
                                "top": True,
                                "left": True,
                                "right": False,
                            },
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_32",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_318": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_85"},
            },
            "layout_319": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_206"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_538",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_320": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_208"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_539",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_321": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_213"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_540",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_322": {
                "type": "COLUMN",
                "setting": {
                    "space": 30,
                    "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                },
                "parent": {"type": "widget", "id": "widget_214"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {"alignSelf": "center"},
                            "autoProps": {"width": "CUSTOM", "height": "AUTO"},
                        },
                        "bbox": {"width": "305px", "height": "43px"},
                        "widgetId": "widget_212",
                        "parent": "layout_66",
                    },
                    "1": {
                        "type": "SECTION",
                        "setting": {"style": {}, "autoProps": {"height": "CUSTOM"}},
                        "bbox": {"width": "1860px", "height": "800.00px"},
                        "sectionId": "section_1",
                        "isPending": False,
                        "parent": "layout_66",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_323": {
                "type": "FIXED",
                "parent": {"type": "view", "id": "view_1"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_514",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_63",
                    }
                },
                "order": ["0"],
            },
            "layout_324": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_215"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_216",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_67",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "600px"},
                        "widgetId": "widget_515",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_67",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_325": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_216"},
            },
            "layout_326": {
                "type": "COLUMN",
                "setting": {"space": 40, "justifyContent": "flex-start"},
                "parent": {"type": "widget", "id": "widget_217"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "735px", "height": "300.00px"},
                        "widgetId": "widget_218",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_69",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_219",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "parent": "layout_69",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_327": {
                "type": "FIXED",
                "parent": {"type": "view", "id": "view_2"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_516",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "height": "STRETCH",
                                "width": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_78",
                    }
                },
                "order": ["0"],
            },
            "layout_328": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_239"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "400px"},
                        "widgetId": "widget_240",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_79",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "600px"},
                        "widgetId": "widget_517",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_79",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_329": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_240"},
            },
            "layout_330": {
                "type": "COLUMN",
                "setting": {"space": 40, "justifyContent": "flex-start"},
                "parent": {"type": "widget", "id": "widget_241"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "450px", "height": "300px"},
                        "widgetId": "widget_243",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "isPending": False,
                        "parent": "layout_81",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "620px", "height": "275px"},
                        "widgetId": "widget_244",
                        "parent": "layout_81",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_331": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_244"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_89",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_89",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_89",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_332": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_244"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_255",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_83",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_256",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_83",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_257",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_83",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_333": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_244"},
            },
            "layout_334": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_220"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_541",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_335": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_222"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_542",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_336": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_224"},
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_543",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["1"],
            },
            "layout_337": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_226"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_544",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_338": {
                "type": "COLUMN",
                "setting": {
                    "padding": {"number": [30, 20, 30, 20], "unit": "px"},
                    "space": 20,
                },
                "parent": {"type": "widget", "id": "widget_227"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_228",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_74",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_523",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_74",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_524",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_74",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_525",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_74",
                    },
                },
                "order": ["0", "1", "2", "3"],
            },
            "layout_339": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_229"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_545",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_340": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_232"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_546",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_341": {
                "type": "ROW",
                "parent": {"type": "widget", "id": "widget_235"},
                "content": {
                    "2": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_547",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                    }
                },
                "order": ["2"],
            },
            "layout_342": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "400px"},
                        "widgetId": "widget_9",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_4",
                    }
                },
                "order": ["0"],
            },
            "layout_343": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_11",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_5",
                    }
                },
                "order": ["0"],
            },
            "layout_344": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_13",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_6",
                    }
                },
                "order": ["0"],
            },
            "layout_345": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_494",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_7",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_501",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_7",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_346": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_20",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_10",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_21",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_10",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_347": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_34",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_14",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_35",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_14",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_348": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_27",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_12",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_28",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_12",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_349": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "300px"},
                        "widgetId": "widget_41",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_16",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_503",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_16",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_350": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_47",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_18",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "300px"},
                        "widgetId": "widget_48",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_18",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_351": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_50",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_19",
                    }
                },
                "order": ["0"],
            },
            "layout_352": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_52",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_20",
                    }
                },
                "order": ["0"],
            },
            "layout_353": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_508",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_21",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_509",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_21",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "391px"},
                        "widgetId": "widget_90",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_21",
                    },
                },
                "order": ["0", "1", "2"],
            },
            "layout_354": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_207",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_60",
                    }
                },
                "order": ["0"],
            },
            "layout_355": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_209",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_61",
                    }
                },
                "order": ["0"],
            },
            "layout_356": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_513",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_65",
                    }
                },
                "order": ["0"],
            },
            "layout_357": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_221",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_70",
                    }
                },
                "order": ["0"],
            },
            "layout_358": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_223",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_71",
                    }
                },
                "order": ["0"],
            },
            "layout_359": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "400px"},
                        "widgetId": "widget_225",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_72",
                    }
                },
                "order": ["0"],
            },
            "layout_360": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": None},
                        "widgetId": "widget_522",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_73",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "300px"},
                        "widgetId": "widget_238",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_73",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_361": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_230",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_75",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_231",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_75",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_362": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_233",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_76",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_234",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_76",
                    },
                },
                "order": ["0", "1"],
            },
            "layout_363": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": 200},
                        "widgetId": "widget_236",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_77",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "70px"},
                        "widgetId": "widget_237",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_77",
                    },
                },
                "order": ["0", "1"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "indicator",
        "widgets": {
            "widget_1": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_1",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "0"}],
                },
            },
            "widget_2": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px; color: var(--light);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_2",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "1"}],
                },
            },
            "widget_3": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {"linkType": "DIALOG", "value": "dialog_2"},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_3",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "2"}],
                },
            },
            "widget_4": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 22 28"><path d="M22 19.375c0 .562-.25 1.656-.484 2.172-.328.766-1.203 1.266-1.906 1.656-.922.5-1.859.797-2.906.797-1.453 0-2.766-.594-4.094-1.078a14.627 14.627 0 01-2.734-1.297c-2.656-1.641-5.859-4.844-7.5-7.5a14.627 14.627 0 01-1.297-2.734C.595 10.063.001 8.75.001 7.297c0-1.047.297-1.984.797-2.906.391-.703.891-1.578 1.656-1.906.516-.234 1.609-.484 2.172-.484.109 0 .219 0 .328.047.328.109.672.875.828 1.188.5.891.984 1.797 1.5 2.672.25.406.719.906.719 1.391 0 .953-2.828 2.344-2.828 3.187 0 .422.391.969.609 1.344 1.578 2.844 3.547 4.813 6.391 6.391.375.219.922.609 1.344.609.844 0 2.234-2.828 3.187-2.828.484 0 .984.469 1.391.719.875.516 1.781 1 2.672 1.5.313.156 1.078.5 1.188.828.047.109.047.219.047.328z"></path></svg>',
                            "position": "LEFT",
                        },
                        "linkParam": {"linkType": "DIALOG", "value": "dialog_1"},
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "#c5c5c5",
                                    "width": "0px",
                                },
                                "text": {"color": "#181818", "size": "14px"},
                                "borderRadius": {"number": [0], "unit": "px"},
                                "iconProps": {"color": "var(--light)", "size": 28},
                            },
                            "hover": {
                                "background": {"color": "", "fillType": "fill"},
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 28},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_4",
                "useDataSources": [],
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "3"}],
                },
            },
            "widget_5": {
                "uri": "widgets/common/share/",
                "version": "1.5.0",
                "config": {
                    "uiMode": "POPUP",
                    "popup": {
                        "icon": {
                            "svg": '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16"><path d="M8 0v2H2v12h12V8h2v6a2 2 0 01-2 2H2a2 2 0 01-2-2V2a2 2 0 012-2h6zm2 0v2h2.5l-8 8L6 11.5l8-8V6h2V0h-6z" fill-rule="evenodd"></path></svg>',
                            "properties": {
                                "filename": None,
                                "originalName": None,
                                "path": None,
                                "color": "var(--light)",
                                "size": 24,
                                "inlineSvg": True,
                            },
                        },
                        "items": [
                            "embed",
                            "qrcode",
                            "email",
                            "facebook",
                            "twitter",
                            "pinterest",
                            "linkedin",
                        ],
                        "tooltip": "",
                    },
                    "inline": {
                        "items": [
                            "facebook",
                            "twitter",
                            "pinterest",
                            "linkedin",
                            "embed",
                            "qrcode",
                            "email",
                            "sharelink",
                        ],
                        "design": {
                            "direction": "HORIZONTAL",
                            "hideLabel": False,
                            "btnRad": 0,
                            "btnColor": "",
                            "iconColor": "",
                            "size": "default",
                        },
                    },
                },
                "id": "widget_5",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "4"}],
                },
            },
            "widget_6": {
                "uri": "widgets/common/menu/",
                "version": "1.5.0",
                "config": {
                    "type": "nav",
                    "menuStyle": "underline",
                    "vertical": False,
                    "advanced": True,
                    "standard": {
                        "textAlign": "center",
                        "gap": "40px",
                        "showText": True,
                        "showIcon": True,
                        "iconPosition": "start",
                    },
                    "variant": {
                        "size": {
                            "fontSize": "0.875rem",
                            "lineHeight": 1.5,
                            "paddingX": "1rem",
                            "paddingY": "0.25rem",
                        },
                        "gutter": "0.25rem",
                        "icon": {"spacing": "0.5rem"},
                        "root": {"border": {"width": 0}, "borderRadius": "0rem"},
                        "item": {
                            "default": {
                                "color": "var(--light)",
                                "bg": "transparent",
                                "border": {"color": "transparent", "width": "2px"},
                                "borderRadius": "0rem",
                                "icon": {"color": "var(--light)", "size": "16px"},
                                "size": "18px",
                            },
                            "hover": {
                                "color": "var(--light-700)",
                                "border": {"color": "transparent", "width": "2px"},
                                "borderRadius": "0rem",
                                "icon": {"size": "16px", "color": "var(--light-700)"},
                                "size": "18px",
                            },
                            "active": {
                                "color": "var(--light)",
                                "border": {"color": "var(--light)", "width": "2px"},
                                "borderRadius": "0rem",
                                "icon": {"color": "var(--light)", "size": "16px"},
                                "size": "18px",
                            },
                            "disabled": {
                                "color": "#989898",
                                "shadow": "none",
                                "borderRadius": "0rem",
                            },
                        },
                    },
                    "paper": {
                        "bg": "#f0f0f0",
                        "color": "#050505",
                        "paddingX": 0,
                        "paddingY": 0,
                        "border": {"width": 0, "color": "transparent"},
                    },
                },
                "id": "widget_6",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "5"}],
                },
            },
            "widget_7": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="color: var(--dark-800); font-size: 16px;">Powered by Esri</span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_7",
                "parent": {
                    "LARGE": [{"layoutId": "layout_2", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_286", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_287", "layoutItemId": "0"}],
                },
            },
            "widget_8": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 0, 50], "unit": "px"}},
                },
                "id": "widget_8",
                "layouts": {"DEFAULT": {"LARGE": "layout_4"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "0"}]},
            },
            "widget_9": {
                "uri": "widgets/common/filter/",
                "version": "1.5.0",
                "config": {
                    "groupByLayer": False,
                    "logicalOperator": "AND",
                    "custom": False,
                    "resetAll": False,
                    "turnOffAll": False,
                    "arrangeType": "BLOCK",
                    "triggerType": "TOGGLE",
                    "omitInternalStyle": False,
                    "filterItems": [],
                },
                "id": "widget_9",
                "style": {
                    "background": {
                        "color": "",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_4", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_188", "layoutItemId": "0"}],
                },
            },
            "widget_10": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 0, 50], "unit": "px"}},
                },
                "id": "widget_10",
                "layouts": {"DEFAULT": {"LARGE": "layout_5"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "1"}]},
            },
            "widget_11": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark); font-size: 24px;">\ufeff\ufeffTitle</span></p>',
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_11",
                "parent": {
                    "LARGE": [{"layoutId": "layout_5", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_189", "layoutItemId": "0"}],
                },
            },
            "widget_12": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_12",
                "layouts": {"DEFAULT": {"LARGE": "layout_6"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "2"}]},
            },
            "widget_13": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-300);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_13",
                "parent": {
                    "LARGE": [{"layoutId": "layout_6", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_190", "layoutItemId": "0"}],
                },
            },
            "widget_14": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_14",
                "layouts": {"DEFAULT": {"LARGE": "layout_7"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "3"}]},
            },
            "widget_15": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {"padding": {"number": [20, 0, 20, 40], "unit": "px"}},
                },
                "id": "widget_15",
                "layouts": {"DEFAULT": {"LARGE": "layout_8"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_7", "layoutItemId": "0"}]},
            },
            "widget_16": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_16",
                "layouts": {"DEFAULT": {"LARGE": "layout_9"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "0"}]},
            },
            "widget_17": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-100);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_17",
                "parent": {
                    "LARGE": [{"layoutId": "layout_9", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_193", "layoutItemId": "0"}],
                },
            },
            "widget_18": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-800); font-size: 42px;">\ufeffNumber</span></p>',
                    "placeholder": '<p><span style="font-size: 42px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_18",
                "parent": {
                    "LARGE": [{"layoutId": "layout_9", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_193", "layoutItemId": "1"}],
                },
            },
            "widget_19": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_19",
                "layouts": {"DEFAULT": {"LARGE": "layout_10"}},
                "parent": {"LARGE": [{"layoutId": "layout_9", "layoutItemId": "2"}]},
            },
            "widget_20": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_20",
                "parent": {
                    "LARGE": [{"layoutId": "layout_10", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_194", "layoutItemId": "0"}],
                },
            },
            "widget_21": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-400);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_21",
                "parent": {
                    "LARGE": [{"layoutId": "layout_10", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_194", "layoutItemId": "1"}],
                },
            },
            "widget_22": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Vertical",
                    "strokeStyle": {
                        "type": "Style1",
                        "color": "var(--light-400)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_22",
                "parent": {
                    "LARGE": [{"layoutId": "layout_8", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_192", "layoutItemId": "2"}],
                },
            },
            "widget_23": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_23",
                "layouts": {"DEFAULT": {"LARGE": "layout_11"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "2"}]},
            },
            "widget_24": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-100);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_24",
                "parent": {
                    "LARGE": [{"layoutId": "layout_11", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_197", "layoutItemId": "0"}],
                },
            },
            "widget_25": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-800); font-size: 42px;">\ufeffNumber</span></p>',
                    "placeholder": '<p><span style="font-size: 42px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_25",
                "parent": {
                    "LARGE": [{"layoutId": "layout_11", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_197", "layoutItemId": "1"}],
                },
            },
            "widget_26": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_26",
                "layouts": {"DEFAULT": {"LARGE": "layout_12"}},
                "parent": {"LARGE": [{"layoutId": "layout_11", "layoutItemId": "2"}]},
            },
            "widget_27": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_27",
                "parent": {
                    "LARGE": [{"layoutId": "layout_12", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_198", "layoutItemId": "0"}],
                },
            },
            "widget_28": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-400);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_28",
                "parent": {
                    "LARGE": [{"layoutId": "layout_12", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_198", "layoutItemId": "1"}],
                },
            },
            "widget_29": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Vertical",
                    "strokeStyle": {
                        "type": "Style1",
                        "color": "var(--light-400)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_29",
                "parent": {
                    "LARGE": [{"layoutId": "layout_8", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_192", "layoutItemId": "3"}],
                },
            },
            "widget_30": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_30",
                "layouts": {"DEFAULT": {"LARGE": "layout_13"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "4"}]},
            },
            "widget_31": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-100);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_31",
                "parent": {
                    "LARGE": [{"layoutId": "layout_13", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_195", "layoutItemId": "0"}],
                },
            },
            "widget_32": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-800); font-size: 42px;">\ufeffNumber</span></p>',
                    "placeholder": '<p><span style="font-size: 42px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_32",
                "parent": {
                    "LARGE": [{"layoutId": "layout_13", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_195", "layoutItemId": "1"}],
                },
            },
            "widget_33": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_33",
                "layouts": {"DEFAULT": {"LARGE": "layout_14"}},
                "parent": {"LARGE": [{"layoutId": "layout_13", "layoutItemId": "2"}]},
            },
            "widget_34": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_34",
                "parent": {
                    "LARGE": [{"layoutId": "layout_14", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_196", "layoutItemId": "0"}],
                },
            },
            "widget_35": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-400);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_35",
                "parent": {
                    "LARGE": [{"layoutId": "layout_14", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_196", "layoutItemId": "1"}],
                },
            },
            "widget_36": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_36",
                "layouts": {"DEFAULT": {"LARGE": "layout_15"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_7", "layoutItemId": "1"}]},
            },
            "widget_37": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="color: var(--light); font-size: 18px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_37",
                "parent": {
                    "LARGE": [{"layoutId": "layout_15", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_199", "layoutItemId": "0"}],
                },
            },
            "widget_38": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 48px; color: var(--light);">\ufeffNumber</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="color: var(--light); font-size: 48px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_38",
                "parent": {
                    "LARGE": [{"layoutId": "layout_15", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_199", "layoutItemId": "1"}],
                },
            },
            "widget_39": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(240,240,240,0.2)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "#c5c5c5",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--light)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#181818", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "rgba(240,240,240,0.3)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                    },
                },
                "id": "widget_39",
                "parent": {
                    "LARGE": [{"layoutId": "layout_15", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_199", "layoutItemId": "2"}],
                },
            },
            "widget_40": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_40",
                "layouts": {"DEFAULT": {"LARGE": "layout_16"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "4"}]},
            },
            "widget_41": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_41",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_16", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_200", "layoutItemId": "0"}],
                },
            },
            "widget_42": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_42",
                "layouts": {"DEFAULT": {"LARGE": "layout_17"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_16", "layoutItemId": "1"}]},
            },
            "widget_43": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_43",
                "parent": {
                    "LARGE": [{"layoutId": "layout_17", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_201", "layoutItemId": "0"}],
                },
            },
            "widget_44": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 32px;">\ufeffNumber</strong></p>',
                    "placeholder": '<p style="text-align: center;"><strong style="font-size: 32px;">Double click to edit text</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_44",
                "parent": {
                    "LARGE": [{"layoutId": "layout_17", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_201", "layoutItemId": "1"}],
                },
            },
            "widget_45": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px; color: var(--dark-300);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_45",
                "parent": {
                    "LARGE": [{"layoutId": "layout_17", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_201", "layoutItemId": "2"}],
                },
            },
            "widget_46": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_46",
                "layouts": {"DEFAULT": {"LARGE": "layout_18"}},
                "style": {
                    "background": {
                        "color": "rgba(7,111,229,0.1)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
                "parent": {"LARGE": [{"layoutId": "layout_17", "layoutItemId": "3"}]},
            },
            "widget_47": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_47",
                "parent": {
                    "LARGE": [{"layoutId": "layout_18", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_202", "layoutItemId": "0"}],
                },
            },
            "widget_48": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_48",
                "parent": {
                    "LARGE": [{"layoutId": "layout_18", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_202", "layoutItemId": "1"}],
                },
            },
            "widget_49": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 0, 50], "unit": "px"}},
                },
                "id": "widget_49",
                "layouts": {"DEFAULT": {"LARGE": "layout_19"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "5"}]},
            },
            "widget_50": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 24px; color: var(--dark-800);">\ufeffTitle</span></p>',
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_50",
                "parent": {
                    "LARGE": [{"layoutId": "layout_19", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_203", "layoutItemId": "0"}],
                },
            },
            "widget_51": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_51",
                "layouts": {"DEFAULT": {"LARGE": "layout_20"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "6"}]},
            },
            "widget_52": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-300);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_52",
                "parent": {
                    "LARGE": [{"layoutId": "layout_20", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_204", "layoutItemId": "0"}],
                },
            },
            "widget_53": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_53",
                "layouts": {"DEFAULT": {"LARGE": "layout_21"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "7"}]},
            },
            "widget_54": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_54",
                "layouts": {"DEFAULT": {"LARGE": "layout_22"}},
                "parent": {"LARGE": [{"layoutId": "layout_21", "layoutItemId": "0"}]},
            },
            "widget_55": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--primary)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE8",
                    "direction": "Vertical",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_55",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_26",
                        "MEDIUM": "layout_209",
                        "SMALL": "layout_310",
                    },
                    "HOVER": {
                        "LARGE": "layout_24",
                        "MEDIUM": "layout_210",
                        "SMALL": "layout_311",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_22", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_208", "layoutItemId": "0"}],
                },
            },
            "widget_67": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1;"><span style="color: var(--light); font-size: 18px;">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_67",
                "parent": {
                    "LARGE": [{"layoutId": "layout_26", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_209", "layoutItemId": "0"}],
                },
            },
            "widget_68": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="color: var(--light); font-size: 14px;">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_68",
                "parent": {
                    "LARGE": [{"layoutId": "layout_26", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_209", "layoutItemId": "1"}],
                },
            },
            "widget_69": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_69",
                "parent": {
                    "LARGE": [{"layoutId": "layout_26", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_209", "layoutItemId": "2"}],
                },
            },
            "widget_73": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M2 8c0-.291.231-.527.516-.527h9.508L8.527 3.9a.535.535 0 010-.746.508.508 0 01.73 0L14 8l-4.743 4.846a.508.508 0 01-.73 0 .535.535 0 010-.746l3.497-3.573H2.516A.522.522 0 012 8z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(240,240,240,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 24},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(255,255,255,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "#c5c5c5",
                                    "width": "0px",
                                },
                                "text": {"color": "#181818", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 24},
                            },
                        },
                    },
                },
                "id": "widget_73",
                "parent": {
                    "LARGE": [{"layoutId": "layout_26", "layoutItemId": "10"}],
                    "MEDIUM": [{"layoutId": "layout_209", "layoutItemId": "3"}],
                },
            },
            "widget_74": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE8",
                    "direction": "Vertical",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_74",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_27",
                        "MEDIUM": "layout_211",
                        "SMALL": "layout_312",
                    },
                    "HOVER": {
                        "LARGE": "layout_28",
                        "MEDIUM": "layout_212",
                        "SMALL": "layout_313",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_22", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_208", "layoutItemId": "1"}],
                },
            },
            "widget_75": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1;"><span style="font-size: 18px; color: var(--dark);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_75",
                "parent": {
                    "LARGE": [{"layoutId": "layout_27", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_211", "layoutItemId": "0"}],
                },
            },
            "widget_76": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_76",
                "parent": {
                    "LARGE": [{"layoutId": "layout_27", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_211", "layoutItemId": "1"}],
                },
            },
            "widget_77": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_77",
                "parent": {
                    "LARGE": [{"layoutId": "layout_27", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_211", "layoutItemId": "2"}],
                },
            },
            "widget_78": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M2 8c0-.291.231-.527.516-.527h9.508L8.527 3.9a.535.535 0 010-.746.508.508 0 01.73 0L14 8l-4.743 4.846a.508.508 0 01-.73 0 .535.535 0 010-.746l3.497-3.573H2.516A.522.522 0 012 8z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(240,240,240,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark)", "size": 24},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark)", "size": 24},
                            },
                        },
                    },
                },
                "id": "widget_78",
                "parent": {
                    "LARGE": [{"layoutId": "layout_27", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_211", "layoutItemId": "3"}],
                },
            },
            "widget_79": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_79",
                "layouts": {"DEFAULT": {"LARGE": "layout_29"}},
                "parent": {"LARGE": [{"layoutId": "layout_21", "layoutItemId": "1"}]},
            },
            "widget_80": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE8",
                    "direction": "Vertical",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_80",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_30",
                        "MEDIUM": "layout_214",
                        "SMALL": "layout_315",
                    },
                    "HOVER": {
                        "LARGE": "layout_31",
                        "MEDIUM": "layout_215",
                        "SMALL": "layout_316",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_29", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_213", "layoutItemId": "0"}],
                },
            },
            "widget_81": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1;"><span style="font-size: 18px; color: var(--dark);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_81",
                "parent": {
                    "LARGE": [{"layoutId": "layout_30", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_214", "layoutItemId": "0"}],
                },
            },
            "widget_82": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_82",
                "parent": {
                    "LARGE": [{"layoutId": "layout_30", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_214", "layoutItemId": "1"}],
                },
            },
            "widget_83": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_83",
                "parent": {
                    "LARGE": [{"layoutId": "layout_30", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_214", "layoutItemId": "2"}],
                },
            },
            "widget_84": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M2 8c0-.291.231-.527.516-.527h9.508L8.527 3.9a.535.535 0 010-.746.508.508 0 01.73 0L14 8l-4.743 4.846a.508.508 0 01-.73 0 .535.535 0 010-.746l3.497-3.573H2.516A.522.522 0 012 8z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(240,240,240,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark)", "size": 24},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark)", "size": 24},
                            },
                        },
                    },
                },
                "id": "widget_84",
                "parent": {
                    "LARGE": [{"layoutId": "layout_30", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_214", "layoutItemId": "3"}],
                },
            },
            "widget_85": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--primary)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE8",
                    "direction": "Vertical",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_85",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_32",
                        "MEDIUM": "layout_216",
                        "SMALL": "layout_317",
                    },
                    "HOVER": {
                        "LARGE": "layout_33",
                        "MEDIUM": "layout_217",
                        "SMALL": "layout_318",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_29", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_213", "layoutItemId": "1"}],
                },
            },
            "widget_86": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1;"><span style="color: var(--light); font-size: 18px;">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_86",
                "parent": {
                    "LARGE": [{"layoutId": "layout_32", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_216", "layoutItemId": "0"}],
                },
            },
            "widget_87": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="color: var(--light); font-size: 14px;">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_87",
                "parent": {
                    "LARGE": [{"layoutId": "layout_32", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_216", "layoutItemId": "1"}],
                },
            },
            "widget_88": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_88",
                "parent": {
                    "LARGE": [{"layoutId": "layout_32", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_216", "layoutItemId": "2"}],
                },
            },
            "widget_89": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox="0 0 16 16" xmlns="http://www.w3.org/2000/svg"><path d="M2 8c0-.291.231-.527.516-.527h9.508L8.527 3.9a.535.535 0 010-.746.508.508 0 01.73 0L14 8l-4.743 4.846a.508.508 0 01-.73 0 .535.535 0 010-.746l3.497-3.573H2.516A.522.522 0 012 8z" fill="#000" fill-rule="nonzero"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(240,240,240,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 24},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 24},
                            },
                        },
                    },
                },
                "id": "widget_89",
                "parent": {
                    "LARGE": [{"layoutId": "layout_32", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_216", "layoutItemId": "3"}],
                },
            },
            "widget_90": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE2",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_90",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_58",
                        "MEDIUM": "layout_206",
                        "SMALL": "layout_307",
                    },
                    "HOVER": {
                        "LARGE": "layout_59",
                        "MEDIUM": "layout_207",
                        "SMALL": "layout_308",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_21", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_205", "layoutItemId": "2"}],
                },
            },
            "widget_197": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_197",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "0"}],
                },
            },
            "widget_198": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark-800);">Created author: admin</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_198",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "1"}],
                },
            },
            "widget_199": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_199",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "2"}],
                },
            },
            "widget_200": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Created date: xxx</span></p></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_200",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "7"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "3"}],
                },
            },
            "widget_201": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "123",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg " ><g fill= "none " fill-rule= "evenodd "><path d= "M8 3c2.583 0 5.166 1.45 7.75 4.349a.972.972 0 010 1.302C13.166 11.551 10.583 13 8 13S2.834 11.55.25 8.651a.972.972 0 010-1.302C2.834 4.449 5.417 3 8 3zm0 1C5.763 4 3.442 5.303 1.039 8 3.442 10.697 5.763 12 8 12s4.558-1.303 6.961-4C12.558 5.303 10.237 4 8 4zm0 7a3 3 0 110-6 3 3 0 010 6zm0-1a2 2 0 100-4 2 2 0 000 4z " fill= "currentColor " fill-rule= "nonzero "></path><path d= "M0 0h16v16H0z "></path></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "success-100",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"size": "13px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--secondary-100)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_201",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "9"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "4"}],
                },
            },
            "widget_202": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "var(--primary-400)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-400)",
                                    "width": "1px",
                                },
                                "text": {"color": "var(--light)", "size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--primary-500)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"size": 14},
                            },
                        },
                    },
                },
                "id": "widget_202",
                "parent": {
                    "LARGE": [{"layoutId": "layout_58", "layoutItemId": "10"}],
                    "MEDIUM": [{"layoutId": "layout_206", "layoutItemId": "5"}],
                },
            },
            "widget_206": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 0, 50], "unit": "px"}},
                },
                "id": "widget_206",
                "layouts": {"DEFAULT": {"LARGE": "layout_60"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "8"}]},
            },
            "widget_207": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 24px; color: var(--dark-800);">\ufeffTitle</span></p>',
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_207",
                "parent": {
                    "LARGE": [{"layoutId": "layout_60", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_218", "layoutItemId": "0"}],
                },
            },
            "widget_208": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_208",
                "layouts": {"DEFAULT": {"LARGE": "layout_61"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "9"}]},
            },
            "widget_209": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-300);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_209",
                "parent": {
                    "LARGE": [{"layoutId": "layout_61", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_219", "layoutItemId": "0"}],
                },
            },
            "widget_212": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_1",
                        "views": ["view_1", "view_2"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "pills",
                        "standard": {
                            "gap": "0px",
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": True,
                        "variant": {
                            "size": {
                                "fontSize": "0.875rem",
                                "lineHeight": 1.5,
                                "paddingX": "1rem",
                                "paddingY": "0.25rem",
                            },
                            "gutter": "0.25rem",
                            "icon": {"spacing": "0.5rem"},
                            "root": {"borderRadius": "50rem"},
                            "item": {
                                "default": {
                                    "color": "var(--dark)",
                                    "border": {"width": "0px"},
                                    "borderRadius": "50rem",
                                    "size": "24px",
                                },
                                "hover": {
                                    "color": "var(--primary-700)",
                                    "borderRadius": "50rem",
                                    "size": "24px",
                                },
                                "active": {
                                    "color": "var(--light)",
                                    "bg": "var(--primary)",
                                    "borderRadius": "50rem",
                                    "size": "24px",
                                },
                                "disabled": {
                                    "color": "#989898",
                                    "shadow": "none",
                                    "borderRadius": "50rem",
                                },
                            },
                        },
                    },
                },
                "id": "widget_212",
                "parent": {
                    "LARGE": [{"layoutId": "layout_66", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_221", "layoutItemId": "0"}],
                },
            },
            "widget_213": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_213",
                "layouts": {"DEFAULT": {"LARGE": "layout_65"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "12"}]},
            },
            "widget_214": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 30},
                    "style": {
                        "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_214",
                "layouts": {"DEFAULT": {"LARGE": "layout_66"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
                "parent": {"LARGE": [{"layoutId": "layout_65", "layoutItemId": "0"}]},
            },
            "widget_215": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_215",
                "layouts": {"DEFAULT": {"LARGE": "layout_67"}},
                "parent": {"LARGE": [{"layoutId": "layout_63", "layoutItemId": "0"}]},
            },
            "widget_216": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 36978595.474472,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -11165153.70636449,
                                "y": 4749944.28890593,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -13938900.588776229,
                            "ymin": 1834330.2819969417,
                            "xmax": -8391406.823952751,
                            "ymax": 7665558.295814918,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_216",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_68",
                        "MEDIUM": "layout_224",
                        "SMALL": "layout_325",
                    }
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_67", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_223", "layoutItemId": "0"}],
                },
            },
            "widget_217": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_217",
                "layouts": {"DEFAULT": {"LARGE": "layout_69"}},
                "parent": {"LARGE": [{"layoutId": "layout_67", "layoutItemId": "1"}]},
            },
            "widget_218": {
                "uri": "widgets/arcgis/query/",
                "version": "1.5.0",
                "config": {
                    "queryItems": [],
                    "arrangeType": "BLOCK",
                    "arrangeWrap": False,
                    "sizeMap": {
                        "arrangementIconPopper": {
                            "minSize": {"width": 300, "height": 300},
                            "defaultSize": {"width": 360, "height": 480},
                        }
                    },
                },
                "id": "widget_218",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_69", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_225", "layoutItemId": "0"}],
                },
            },
            "widget_219": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_219",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_69", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_225", "layoutItemId": "1"}],
                },
            },
            "widget_220": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 0, 50], "unit": "px"}},
                },
                "id": "widget_220",
                "layouts": {"DEFAULT": {"LARGE": "layout_70"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "13"}]},
            },
            "widget_221": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 24px; color: var(--dark-800);">\ufeffTitle</span></p>',
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_221",
                "parent": {
                    "LARGE": [{"layoutId": "layout_70", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_233", "layoutItemId": "0"}],
                },
            },
            "widget_222": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_222",
                "layouts": {"DEFAULT": {"LARGE": "layout_71"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "14"}]},
            },
            "widget_223": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 16px; color: var(--dark-300);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_223",
                "parent": {
                    "LARGE": [{"layoutId": "layout_71", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_234", "layoutItemId": "0"}],
                },
            },
            "widget_224": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_224",
                "layouts": {"DEFAULT": {"LARGE": "layout_72"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "15"}]},
            },
            "widget_225": {
                "uri": "widgets/common/table/",
                "version": "1.5.0",
                "config": {"layersConfig": [], "arrangeType": "TABS"},
                "id": "widget_225",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_72", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_235", "layoutItemId": "0"}],
                },
            },
            "widget_226": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10, 50, 20, 50], "unit": "px"}},
                },
                "id": "widget_226",
                "layouts": {"DEFAULT": {"LARGE": "layout_73"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "16"}]},
            },
            "widget_227": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [30, 20, 30, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_227",
                "layouts": {"DEFAULT": {"LARGE": "layout_74"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_73", "layoutItemId": "0"}]},
            },
            "widget_228": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 18px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_228",
                "parent": {
                    "LARGE": [{"layoutId": "layout_74", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_237", "layoutItemId": "0"}],
                },
            },
            "widget_229": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_229",
                "layouts": {"DEFAULT": {"LARGE": "layout_75"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "1"}]},
            },
            "widget_230": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_230",
                "parent": {
                    "LARGE": [{"layoutId": "layout_75", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_238", "layoutItemId": "0"}],
                },
            },
            "widget_231": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_231",
                "parent": {
                    "LARGE": [{"layoutId": "layout_75", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_238", "layoutItemId": "1"}],
                },
            },
            "widget_232": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_232",
                "layouts": {"DEFAULT": {"LARGE": "layout_76"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "2"}]},
            },
            "widget_233": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_233",
                "parent": {
                    "LARGE": [{"layoutId": "layout_76", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_239", "layoutItemId": "0"}],
                },
            },
            "widget_234": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_234",
                "parent": {
                    "LARGE": [{"layoutId": "layout_76", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_239", "layoutItemId": "1"}],
                },
            },
            "widget_235": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_235",
                "layouts": {"DEFAULT": {"LARGE": "layout_77"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "3"}]},
            },
            "widget_236": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_236",
                "parent": {
                    "LARGE": [{"layoutId": "layout_77", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_240", "layoutItemId": "0"}],
                },
            },
            "widget_237": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px; color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_237",
                "parent": {
                    "LARGE": [{"layoutId": "layout_77", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_240", "layoutItemId": "1"}],
                },
            },
            "widget_238": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_238",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_73", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_236", "layoutItemId": "1"}],
                },
            },
            "widget_239": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_239",
                "layouts": {"DEFAULT": {"LARGE": "layout_79"}},
                "parent": {"LARGE": [{"layoutId": "layout_78", "layoutItemId": "0"}]},
            },
            "widget_240": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 36978595.474472,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -7087117.052428734,
                                "y": -1029392.3150045087,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -9860863.934840472,
                            "ymin": -3945006.3219134966,
                            "xmax": -4313370.170016995,
                            "ymax": 1886221.6919044794,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_240",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_80",
                        "MEDIUM": "layout_228",
                        "SMALL": "layout_329",
                    }
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_79", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_227", "layoutItemId": "0"}],
                },
            },
            "widget_241": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_241",
                "layouts": {"DEFAULT": {"LARGE": "layout_81"}},
                "parent": {"LARGE": [{"layoutId": "layout_79", "layoutItemId": "1"}]},
            },
            "widget_243": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_243",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_81", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_229", "layoutItemId": "0"}],
                },
            },
            "widget_244": {
                "uri": "widgets/common/list/",
                "version": "1.5.0",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE5",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "scrollBarOpen": True,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {"LARGE": {"width": 320, "height": 88}},
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {"LARGE": {"width": 320, "height": 88}},
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {"LARGE": {"width": 320, "height": 88}},
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-500)",
                                    "width": "1px",
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                },
                "id": "widget_244",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_89",
                        "MEDIUM": "layout_230",
                        "SMALL": "layout_331",
                    },
                    "SELECTED": {
                        "LARGE": "layout_83",
                        "MEDIUM": "layout_231",
                        "SMALL": "layout_332",
                    },
                    "HOVER": {
                        "LARGE": "layout_84",
                        "MEDIUM": "layout_232",
                        "SMALL": "layout_333",
                    },
                },
                "useDataSourcesEnabled": True,
                "style": {
                    "background": {
                        "color": "rgba(240,240,240,0)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_81", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_229", "layoutItemId": "1"}],
                },
            },
            "widget_255": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "cropShape": "circle",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "svgViewBox": "0 0 28 28",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_255",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_89", "layoutItemId": "0"},
                        {"layoutId": "layout_83", "layoutItemId": "0"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_230", "layoutItemId": "0"},
                        {"layoutId": "layout_231", "layoutItemId": "0"},
                    ],
                },
            },
            "widget_256": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 16px;color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_256",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_89", "layoutItemId": "1"},
                        {"layoutId": "layout_83", "layoutItemId": "1"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_230", "layoutItemId": "1"},
                        {"layoutId": "layout_231", "layoutItemId": "1"},
                    ],
                },
            },
            "widget_257": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "Button",
                        "toolTip": "",
                        "linkParam": {},
                    },
                    "styleConfig": {
                        "name": "Default",
                        "themeStyle": {
                            "className": "jimu-widget d-block",
                            "color": "primary",
                        },
                        "customStyle": {"regular": {}, "hover": {}, "clicked": {}},
                    },
                },
                "id": "widget_257",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_89", "layoutItemId": "2"},
                        {"layoutId": "layout_83", "layoutItemId": "2"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_230", "layoutItemId": "2"},
                        {"layoutId": "layout_231", "layoutItemId": "2"},
                    ],
                },
            },
            "widget_259": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><strong style="font-size: 14px; color: var(--dark-800);">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_259",
                "parent": {"LARGE": [{"layoutId": "layout_90", "layoutItemId": "1"}]},
            },
            "widget_260": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_260",
                "parent": {"LARGE": [{"layoutId": "layout_90", "layoutItemId": "2"}]},
            },
            "widget_261": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_261",
                "parent": {"LARGE": [{"layoutId": "layout_91", "layoutItemId": "0"}]},
            },
            "widget_262": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><strong style="font-size: 14px; color: var(--dark-800);">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_262",
                "parent": {"LARGE": [{"layoutId": "layout_91", "layoutItemId": "1"}]},
            },
            "widget_263": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_263",
                "parent": {"LARGE": [{"layoutId": "layout_91", "layoutItemId": "2"}]},
            },
            "widget_264": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [60, 50, 40, 50], "unit": "px"}},
                },
                "id": "widget_264",
                "layouts": {"DEFAULT": {"LARGE": "layout_92"}},
                "parent": {"LARGE": [{"layoutId": "layout_3", "layoutItemId": "0"}]},
            },
            "widget_265": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_2",
                        "views": ["view_3", "view_4", "view_5"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "pills",
                        "standard": {
                            "gap": "40px",
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": True,
                        "variant": {
                            "size": {
                                "fontSize": "0.875rem",
                                "lineHeight": 1.5,
                                "paddingX": "1rem",
                                "paddingY": "0.25rem",
                            },
                            "gutter": "0.25rem",
                            "icon": {"spacing": "0.5rem"},
                            "root": {"borderRadius": "50rem", "bg": ""},
                            "item": {
                                "default": {
                                    "color": "var(--dark-800)",
                                    "border": {"width": "0px"},
                                    "borderRadius": "2px",
                                    "size": "32px",
                                },
                                "hover": {
                                    "color": "var(--dark-300)",
                                    "borderRadius": "2px",
                                    "size": "32px",
                                },
                                "active": {
                                    "color": "var(--light)",
                                    "bg": "var(--primary)",
                                    "borderRadius": "2px",
                                    "size": "32px",
                                    "border": {
                                        "type": "solid",
                                        "color": "",
                                        "width": "0px",
                                    },
                                },
                                "disabled": {
                                    "color": "#989898",
                                    "shadow": "none",
                                    "borderRadius": "50rem",
                                },
                            },
                        },
                    },
                },
                "id": "widget_265",
                "parent": {
                    "LARGE": [{"layoutId": "layout_92", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_145", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_284", "layoutItemId": "0"}],
                },
            },
            "widget_266": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 40, 50], "unit": "px"}},
                },
                "id": "widget_266",
                "layouts": {"DEFAULT": {"LARGE": "layout_93"}},
                "parent": {"LARGE": [{"layoutId": "layout_3", "layoutItemId": "1"}]},
            },
            "widget_267": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_267",
                "layouts": {"DEFAULT": {"LARGE": "layout_95"}},
                "parent": {"LARGE": [{"layoutId": "layout_94", "layoutItemId": "0"}]},
            },
            "widget_268": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_268",
                "layouts": {"DEFAULT": {"LARGE": "layout_96"}},
                "parent": {"LARGE": [{"layoutId": "layout_95", "layoutItemId": "0"}]},
            },
            "widget_269": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_269",
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_96", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_150", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_248", "layoutItemId": "0"}],
                },
            },
            "widget_270": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_270",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_96", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_150", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_248", "layoutItemId": "1"}],
                },
            },
            "widget_271": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_271",
                "layouts": {"DEFAULT": {"LARGE": "layout_97"}},
                "parent": {"LARGE": [{"layoutId": "layout_95", "layoutItemId": "1"}]},
            },
            "widget_272": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0",
                                "offsetY": "0",
                                "blur": "0",
                                "spread": "0",
                                "color": "var(--dark-100)",
                            },
                        },
                    },
                    "itemStyle": "STYLE1",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "isOpenAdvabceSetting": False,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_272",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_100",
                        "MEDIUM": "layout_152",
                        "SMALL": "layout_250",
                    },
                    "HOVER": {
                        "LARGE": "layout_99",
                        "MEDIUM": "layout_153",
                        "SMALL": "layout_251",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_97", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_151", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_249", "layoutItemId": "0"}],
                },
            },
            "widget_273": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_273",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "0"}],
                },
            },
            "widget_274": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                    "tooltip": {"name": "", "parts": [{"type": "STRING", "exp": '""'}]},
                },
                "id": "widget_274",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "1"}],
                },
            },
            "widget_275": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark-800);">Created by:</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_275",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "2"}],
                },
            },
            "widget_276": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "var(--primary)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--light)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#fff", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--primary-600)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                    },
                },
                "id": "widget_276",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "3"}],
                },
            },
            "widget_277": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="26" height="28" viewBox="0 0 26 28"><path d="M17.766 15.687l4.781-4.641-6.594-0.969-2.953-5.969-2.953 5.969-6.594 0.969 4.781 4.641-1.141 6.578 5.906-3.109 5.891 3.109zM26 10.109c0 0.281-0.203 0.547-0.406 0.75l-5.672 5.531 1.344 7.812c0.016 0.109 0.016 0.203 0.016 0.313 0 0.422-0.187 0.781-0.641 0.781-0.219 0-0.438-0.078-0.625-0.187l-7.016-3.687-7.016 3.687c-0.203 0.109-0.406 0.187-0.625 0.187-0.453 0-0.656-0.375-0.656-0.781 0-0.109 0.016-0.203 0.031-0.313l1.344-7.812-5.688-5.531c-0.187-0.203-0.391-0.469-0.391-0.75 0-0.469 0.484-0.656 0.875-0.719l7.844-1.141 3.516-7.109c0.141-0.297 0.406-0.641 0.766-0.641s0.625 0.344 0.766 0.641l3.516 7.109 7.844 1.141c0.375 0.063 0.875 0.25 0.875 0.719z"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_277",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "8"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "8"}],
                },
            },
            "widget_278": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg width="16px" height="16px" viewBox="0 0 16 16"  xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"><g id="Style" stroke="none" stroke-width="1" fill="none" fill-rule="evenodd"><g id="0-UI-icons" transform="translate(-244.000000, -245.000000)" fill="var(--black)000" fill-rule="nonzero"><path d="M252,260 C255.865993,260 259,256.865993 259,253 C259,249.134007 255.865993,246 252,246 C248.134007,246 245,249.134007 245,253 C245,256.865993 248.134007,260 252,260 Z M252,261 C247.581722,261 244,257.418278 244,253 C244,248.581722 247.581722,245 252,245 C256.418278,245 260,248.581722 260,253 C260,257.418278 256.418278,261 252,261 Z M254.544453,252.999962 L250.455473,249.985987 L250.455473,256.013937 L254.544453,252.999962 Z M255.09981,253.802292 L251.01083,256.816267 C250.581429,257.132778 249.984688,257.030145 249.677973,256.58703 C249.562223,256.419806 249.5,256.219439 249.5,256.013937 L249.5,249.985987 C249.5,249.441441 249.92778,249 250.455473,249 C250.654615,249 250.848781,249.064211 251.01083,249.183656 L255.09981,252.197631 C255.529212,252.514142 255.628669,253.12994 255.321953,253.573055 C255.260669,253.661592 255.185607,253.739051 255.09981,253.802292 Z" id="app-launch"></path></g></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {"color": "", "fillType": "fill"},
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_278",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "7"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "9"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "9"}],
                },
            },
            "widget_279": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Created date: xxx</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_279",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "8"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "4"}],
                },
            },
            "widget_280": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Views number: 123</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_280",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "9"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "5"}],
                },
            },
            "widget_281": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--secondary-200)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_281",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "10"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "6"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "6"}],
                },
            },
            "widget_282": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="32" height="11" viewBox="0 0 32 11"><path d="M4 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S5.7 8.5 4 8.5z M16 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S17.7 8.5 16 8.5z M28 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S29.7 8.5 28 8.5z"/></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_282",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "11"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "10"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "10"}],
                },
            },
            "widget_283": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg "><path d= "M8 7a3 3 0 100-6 3 3 0 000 6zm0 1a4 4 0 110-8 4 4 0 010 8zm1 2H7a6.002 6.002 0 00-5.917 5h13.834A6.002 6.002 0 009 10zM7 9h2a7 7 0 017 7H0a7 7 0 017-7zm.646 3.721l1.768-1.767a.5.5 0 11.707.707l-2.475 2.475-1.767-1.768a.5.5 0 01.707-.707l1.06 1.06z " fill= "currentColor " fill-rule= "nonzero "></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_283",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "12"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "7"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "7"}],
                },
            },
            "widget_284": {
                "uri": "widgets/common/list/",
                "version": "1.5.0",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE4",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "scrollBarOpen": True,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {
                                "LARGE": {"width": 1044.98, "height": 179.96},
                                "MEDIUM": {"width": 419.67, "height": 187.02},
                                "SMALL": {"width": 380, "height": 106.55},
                            },
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {
                                "LARGE": {"width": 1044.98, "height": 179.96},
                                "MEDIUM": {"width": 419.67, "height": 187.02},
                                "SMALL": {"width": 380, "height": 106.55},
                            },
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {
                                "LARGE": {"width": 1044.98, "height": 179.96},
                                "MEDIUM": {"width": 419.67, "height": 187.02},
                                "SMALL": {"width": 380, "height": 106.55},
                            },
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-500)",
                                    "width": "1px",
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                },
                "id": "widget_284",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_108",
                        "MEDIUM": "layout_154",
                        "SMALL": "layout_252",
                    },
                    "SELECTED": {
                        "LARGE": "layout_102",
                        "MEDIUM": "layout_155",
                        "SMALL": "layout_253",
                    },
                    "HOVER": {
                        "LARGE": "layout_103",
                        "MEDIUM": "layout_156",
                        "SMALL": "layout_254",
                    },
                },
                "useDataSourcesEnabled": True,
                "style": {"border": {"type": "solid", "color": "", "width": "1px"}},
                "parent": {
                    "LARGE": [{"layoutId": "layout_97", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_151", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_249", "layoutItemId": "1"}],
                },
            },
            "widget_296": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_296",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_108", "layoutItemId": "0"},
                        {"layoutId": "layout_102", "layoutItemId": "0"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_154", "layoutItemId": "0"},
                        {"layoutId": "layout_155", "layoutItemId": "0"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_252", "layoutItemId": "0"},
                        {"layoutId": "layout_253", "layoutItemId": "0"},
                    ],
                },
            },
            "widget_297": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<h5><span style="font-size: 15px;color: var(--dark-800);">Double click to edit text</span></h5>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_297",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_108", "layoutItemId": "1"},
                        {"layoutId": "layout_102", "layoutItemId": "1"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_154", "layoutItemId": "1"},
                        {"layoutId": "layout_155", "layoutItemId": "1"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_252", "layoutItemId": "1"},
                        {"layoutId": "layout_253", "layoutItemId": "1"},
                    ],
                },
            },
            "widget_298": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p ><span style="font-size: 13px;color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit. Aenean euismod</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_298",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_108", "layoutItemId": "2"},
                        {"layoutId": "layout_102", "layoutItemId": "2"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_154", "layoutItemId": "2"},
                        {"layoutId": "layout_155", "layoutItemId": "2"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_252", "layoutItemId": "2"},
                        {"layoutId": "layout_253", "layoutItemId": "2"},
                    ],
                },
            },
            "widget_299": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_299",
                "layouts": {"DEFAULT": {"LARGE": "layout_109"}},
                "parent": {"LARGE": [{"layoutId": "layout_95", "layoutItemId": "2"}]},
            },
            "widget_300": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_300",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_109", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_157", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_255", "layoutItemId": "0"}],
                },
            },
            "widget_301": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_301",
                "layouts": {"DEFAULT": {"LARGE": "layout_110"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_109", "layoutItemId": "1"}]},
            },
            "widget_302": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="font-size: 32px;">\ufeffTitle</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_302",
                "parent": {
                    "LARGE": [{"layoutId": "layout_110", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_158", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_256", "layoutItemId": "0"}],
                },
            },
            "widget_303": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="color: var(--primary); font-size: 24px;">\ufeffNumber</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px; color: var(--primary);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_303",
                "parent": {
                    "LARGE": [{"layoutId": "layout_110", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_158", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_256", "layoutItemId": "1"}],
                },
            },
            "widget_304": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_304",
                "parent": {
                    "LARGE": [{"layoutId": "layout_110", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_158", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_256", "layoutItemId": "2"}],
                },
            },
            "widget_305": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="font-size: 16px; color: var(--dark-900);">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo, ad quo torquatos sententiae reprimique. </span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_305",
                "parent": {
                    "LARGE": [{"layoutId": "layout_110", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_158", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_256", "layoutItemId": "3"}],
                },
            },
            "widget_306": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "default"},
                    },
                },
                "id": "widget_306",
                "parent": {
                    "LARGE": [{"layoutId": "layout_110", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_158", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_256", "layoutItemId": "4"}],
                },
            },
            "widget_307": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_307",
                "layouts": {"DEFAULT": {"LARGE": "layout_111"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_109", "layoutItemId": "2"}]},
            },
            "widget_308": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="font-size: 32px;">\ufeffTitle</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_308",
                "parent": {
                    "LARGE": [{"layoutId": "layout_111", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_159", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_257", "layoutItemId": "0"}],
                },
            },
            "widget_309": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="color: var(--primary); font-size: 24px;">\ufeffNumber</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px; color: var(--primary);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_309",
                "parent": {
                    "LARGE": [{"layoutId": "layout_111", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_159", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_257", "layoutItemId": "1"}],
                },
            },
            "widget_310": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_310",
                "parent": {
                    "LARGE": [{"layoutId": "layout_111", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_159", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_257", "layoutItemId": "2"}],
                },
            },
            "widget_311": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="font-size: 16px; color: var(--dark-900);">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo, ad quo torquatos sententiae reprimique. </span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_311",
                "parent": {
                    "LARGE": [{"layoutId": "layout_111", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_159", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_257", "layoutItemId": "3"}],
                },
            },
            "widget_312": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "default"},
                    },
                },
                "id": "widget_312",
                "parent": {
                    "LARGE": [{"layoutId": "layout_111", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_159", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_257", "layoutItemId": "4"}],
                },
            },
            "widget_313": {
                "uri": "widgets/common/table/",
                "version": "1.5.0",
                "config": {"layersConfig": [], "arrangeType": "TABS"},
                "id": "widget_313",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_95", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_246", "layoutItemId": "3"}],
                },
            },
            "widget_314": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 73957190.948944,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": 880554.5658450173,
                                "y": -1785467.3887943064,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -11642888.148394931,
                            "ymin": -7421016.610202284,
                            "xmax": 13403997.280084964,
                            "ymax": 3850081.8326136703,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_314",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_112",
                        "MEDIUM": "layout_149",
                        "SMALL": "layout_247",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_95", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_246", "layoutItemId": "4"}],
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_315": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap1-1612496096511.jpg",
                                "originalName": "widget_315-snap1-1612496096511",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap1-1612496096511.jpg",
                                "type": "IMAGE",
                                "size": 534041,
                                "created": 1612496096520,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -32026230.541417267,
                                "ymin": -2350382.2791066626,
                                "xmax": 10162117.10217883,
                                "ymax": 11581947.74048537,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 147914381.897889,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -10932056.719619218,
                                    "y": 4615782.730689353,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap2-1612496117082.jpg",
                                "originalName": "widget_315-snap2-1612496117082",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap2-1612496117082.jpg",
                                "type": "IMAGE",
                                "size": 517756,
                                "created": 1612496117089,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 10302264.393539697,
                                "ymin": -976350.2586526736,
                                "xmax": 20849351.30443865,
                                "ymax": 2506732.246245312,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 15575807.848989174,
                                    "y": 765190.9937963192,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap3-1612496125303.jpg",
                                "originalName": "widget_315-snap3-1612496125303",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap3-1612496125303.jpg",
                                "type": "IMAGE",
                                "size": 420471,
                                "created": 1612496125308,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 11006918.248043481,
                                "ymin": -4489921.5832079,
                                "xmax": 21554005.158942435,
                                "ymax": -1006839.0783099146,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 16280461.703492958,
                                    "y": -2748380.3307589074,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 4,
                            "name": "Bookmark-4",
                            "title": "Bookmark-4",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap4-1612496130941.jpg",
                                "originalName": "widget_315-snap4-1612496130941",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap4-1612496130941.jpg",
                                "type": "IMAGE",
                                "size": 115777,
                                "created": 1612496130942,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 12609008.183549203,
                                "ymin": -13415137.953714337,
                                "xmax": 23156095.094448157,
                                "ymax": -9932055.448816352,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 17882551.63899868,
                                    "y": -11673596.701265344,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 5,
                            "name": "Bookmark-5",
                            "title": "Bookmark-5",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap5-1612496138211.jpg",
                                "originalName": "widget_315-snap5-1612496138211",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap5-1612496138211.jpg",
                                "type": "IMAGE",
                                "size": 64288,
                                "created": 1612496138212,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 15307521.920871284,
                                "ymin": -19538508.289638177,
                                "xmax": 25854608.831770238,
                                "ymax": -16055425.784740193,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 20581065.37632076,
                                    "y": -17796967.037189186,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 6,
                            "name": "Bookmark-6",
                            "title": "Bookmark-6",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap6-1612496145998.jpg",
                                "originalName": "widget_315-snap6-1612496145998",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap6-1612496145998.jpg",
                                "type": "IMAGE",
                                "size": 428023,
                                "created": 1612496146003,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 36283391.31498645,
                                "ymin": -3377431.897210424,
                                "xmax": 46830478.225885406,
                                "ymax": 105650.607687562,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 41556934.77043593,
                                    "y": -1635890.6447614308,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 7,
                            "name": "Bookmark-7",
                            "title": "Bookmark-7",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap7-1612496158371.jpg",
                                "originalName": "widget_315-snap7-1612496158371",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap7-1612496158371.jpg",
                                "type": "IMAGE",
                                "size": 626113,
                                "created": 1612496158378,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -8870329.77287056,
                                "ymin": -1763933.611939429,
                                "xmax": -6233558.045145821,
                                "ymax": -893162.9857149327,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7551943.90900819,
                                    "y": -1328548.2988271809,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 8,
                            "name": "Bookmark-8",
                            "title": "Bookmark-8",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap8-1612496164229.jpg",
                                "originalName": "widget_315-snap8-1612496164229",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap8-1612496164229.jpg",
                                "type": "IMAGE",
                                "size": 534610,
                                "created": 1612496164235,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -6674948.543320371,
                                "ymin": -2972099.0698932745,
                                "xmax": -4038176.8155956324,
                                "ymax": -2101328.443668778,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -5356562.679458002,
                                    "y": -2536713.7567810263,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 9,
                            "name": "Bookmark-9",
                            "title": "Bookmark-9",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap9-1612496169112.jpg",
                                "originalName": "widget_315-snap9-1612496169112",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap9-1612496169112.jpg",
                                "type": "IMAGE",
                                "size": 459797,
                                "created": 1612496169118,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -5716782.491468654,
                                "ymin": -1144639.4521959827,
                                "xmax": -3080010.7637439156,
                                "ymax": -273868.8259714863,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -4398396.627606285,
                                    "y": -709254.1390837345,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 10,
                            "name": "Bookmark-10",
                            "title": "Bookmark-10",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap10-1612496173280.jpg",
                                "originalName": "widget_315-snap10-1612496173280",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap10-1612496173280.jpg",
                                "type": "IMAGE",
                                "size": 606004,
                                "created": 1612496173287,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -7710209.16444735,
                                "ymin": -1181014.7935382156,
                                "xmax": -5073437.436722612,
                                "ymax": -310244.16731371905,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -6391823.300584981,
                                    "y": -745629.4804259673,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": True,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_315",
                "useMapWidgetIds": ["widget_314"],
                "parent": {
                    "LARGE": [{"layoutId": "layout_95", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_246", "layoutItemId": "5"}],
                },
            },
            "widget_316": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_316",
                "layouts": {"DEFAULT": {"LARGE": "layout_114"}},
                "parent": {"LARGE": [{"layoutId": "layout_113", "layoutItemId": "0"}]},
            },
            "widget_317": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_317",
                "layouts": {"DEFAULT": {"LARGE": "layout_115"}},
                "parent": {"LARGE": [{"layoutId": "layout_114", "layoutItemId": "0"}]},
            },
            "widget_318": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_318",
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_115", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_163", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_261", "layoutItemId": "0"}],
                },
            },
            "widget_319": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_319",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_115", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_163", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_261", "layoutItemId": "1"}],
                },
            },
            "widget_320": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_320",
                "layouts": {"DEFAULT": {"LARGE": "layout_116"}},
                "parent": {"LARGE": [{"layoutId": "layout_114", "layoutItemId": "1"}]},
            },
            "widget_321": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0",
                                "offsetY": "0",
                                "blur": "0",
                                "spread": "0",
                                "color": "var(--dark-100)",
                            },
                        },
                    },
                    "itemStyle": "STYLE1",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "isOpenAdvabceSetting": False,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_321",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_117",
                        "MEDIUM": "layout_165",
                        "SMALL": "layout_263",
                    },
                    "HOVER": {
                        "LARGE": "layout_118",
                        "MEDIUM": "layout_166",
                        "SMALL": "layout_264",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_116", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_164", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_262", "layoutItemId": "0"}],
                },
            },
            "widget_322": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_322",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "0"}],
                },
            },
            "widget_323": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                    "tooltip": {"name": "", "parts": [{"type": "STRING", "exp": '""'}]},
                },
                "id": "widget_323",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "1"}],
                },
            },
            "widget_324": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark-800);">Created by:</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_324",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "2"}],
                },
            },
            "widget_325": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "var(--primary)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-400)",
                                    "width": "1px",
                                },
                                "text": {"color": "var(--light)", "size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--primary-600)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"size": 14},
                            },
                        },
                    },
                },
                "id": "widget_325",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "3"}],
                },
            },
            "widget_326": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Created date: xxx</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_326",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "4"}],
                },
            },
            "widget_327": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Views number: 123</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_327",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "5"}],
                },
            },
            "widget_328": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--secondary-200)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_328",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "6"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "6"}],
                },
            },
            "widget_329": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg "><path d= "M8 7a3 3 0 100-6 3 3 0 000 6zm0 1a4 4 0 110-8 4 4 0 010 8zm1 2H7a6.002 6.002 0 00-5.917 5h13.834A6.002 6.002 0 009 10zM7 9h2a7 7 0 017 7H0a7 7 0 017-7zm.646 3.721l1.768-1.767a.5.5 0 11.707.707l-2.475 2.475-1.767-1.768a.5.5 0 01.707-.707l1.06 1.06z " fill= "currentColor " fill-rule= "nonzero "></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_329",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "7"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "7"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "7"}],
                },
            },
            "widget_330": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="26" height="28" viewBox="0 0 26 28"><path d="M17.766 15.687l4.781-4.641-6.594-0.969-2.953-5.969-2.953 5.969-6.594 0.969 4.781 4.641-1.141 6.578 5.906-3.109 5.891 3.109zM26 10.109c0 0.281-0.203 0.547-0.406 0.75l-5.672 5.531 1.344 7.812c0.016 0.109 0.016 0.203 0.016 0.313 0 0.422-0.187 0.781-0.641 0.781-0.219 0-0.438-0.078-0.625-0.187l-7.016-3.687-7.016 3.687c-0.203 0.109-0.406 0.187-0.625 0.187-0.453 0-0.656-0.375-0.656-0.781 0-0.109 0.016-0.203 0.031-0.313l1.344-7.812-5.688-5.531c-0.187-0.203-0.391-0.469-0.391-0.75 0-0.469 0.484-0.656 0.875-0.719l7.844-1.141 3.516-7.109c0.141-0.297 0.406-0.641 0.766-0.641s0.625 0.344 0.766 0.641l3.516 7.109 7.844 1.141c0.375 0.063 0.875 0.25 0.875 0.719z"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_330",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "8"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "8"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "8"}],
                },
            },
            "widget_331": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg width="16px" height="16px" viewBox="0 0 16 16"  xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"><g id="Style" stroke="none" stroke-width="1" fill="none" fill-rule="evenodd"><g id="0-UI-icons" transform="translate(-244.000000, -245.000000)" fill="var(--black)000" fill-rule="nonzero"><path d="M252,260 C255.865993,260 259,256.865993 259,253 C259,249.134007 255.865993,246 252,246 C248.134007,246 245,249.134007 245,253 C245,256.865993 248.134007,260 252,260 Z M252,261 C247.581722,261 244,257.418278 244,253 C244,248.581722 247.581722,245 252,245 C256.418278,245 260,248.581722 260,253 C260,257.418278 256.418278,261 252,261 Z M254.544453,252.999962 L250.455473,249.985987 L250.455473,256.013937 L254.544453,252.999962 Z M255.09981,253.802292 L251.01083,256.816267 C250.581429,257.132778 249.984688,257.030145 249.677973,256.58703 C249.562223,256.419806 249.5,256.219439 249.5,256.013937 L249.5,249.985987 C249.5,249.441441 249.92778,249 250.455473,249 C250.654615,249 250.848781,249.064211 251.01083,249.183656 L255.09981,252.197631 C255.529212,252.514142 255.628669,253.12994 255.321953,253.573055 C255.260669,253.661592 255.185607,253.739051 255.09981,253.802292 Z" id="app-launch"></path></g></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {"color": "", "fillType": "fill"},
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_331",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "9"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "9"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "9"}],
                },
            },
            "widget_332": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="32" height="11" viewBox="0 0 32 11"><path d="M4 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S5.7 8.5 4 8.5z M16 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S17.7 8.5 16 8.5z M28 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S29.7 8.5 28 8.5z"/></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_332",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "10"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "10"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "10"}],
                },
            },
            "widget_333": {
                "uri": "widgets/common/list/",
                "version": "1.5.0",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE5",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "scrollBarOpen": True,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {
                                "LARGE": {"width": 1200, "height": 179.21},
                                "MEDIUM": {"width": 482.5, "height": 189.63},
                                "SMALL": {"width": 403.75, "height": 81.5},
                            },
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {
                                "LARGE": {"width": 1200, "height": 179.21},
                                "MEDIUM": {"width": 482.5, "height": 189.63},
                                "SMALL": {"width": 403.75, "height": 81.5},
                            },
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {
                                "LARGE": {"width": 1200, "height": 179.21},
                                "MEDIUM": {"width": 482.5, "height": 189.63},
                                "SMALL": {"width": 403.75, "height": 81.5},
                            },
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-500)",
                                    "width": "1px",
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                },
                "id": "widget_333",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_140",
                        "MEDIUM": "layout_167",
                        "SMALL": "layout_265",
                    },
                    "SELECTED": {
                        "LARGE": "layout_120",
                        "MEDIUM": "layout_168",
                        "SMALL": "layout_266",
                    },
                    "HOVER": {
                        "LARGE": "layout_121",
                        "MEDIUM": "layout_169",
                        "SMALL": "layout_267",
                    },
                },
                "useDataSourcesEnabled": True,
                "style": {"border": {"type": "solid", "color": "", "width": "1px"}},
                "parent": {
                    "LARGE": [{"layoutId": "layout_116", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_164", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_262", "layoutItemId": "1"}],
                },
            },
            "widget_340": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_340",
                "layouts": {"DEFAULT": {"LARGE": "layout_122"}},
                "parent": {"LARGE": [{"layoutId": "layout_114", "layoutItemId": "2"}]},
            },
            "widget_341": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_341",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_122", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_170", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_268", "layoutItemId": "0"}],
                },
            },
            "widget_342": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_342",
                "layouts": {"DEFAULT": {"LARGE": "layout_123"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_122", "layoutItemId": "1"}]},
            },
            "widget_343": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_343",
                "parent": {
                    "LARGE": [{"layoutId": "layout_123", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_171", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_269", "layoutItemId": "0"}],
                },
            },
            "widget_344": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="color: var(--primary-700); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_344",
                "parent": {
                    "LARGE": [{"layoutId": "layout_123", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_171", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_269", "layoutItemId": "1"}],
                },
            },
            "widget_345": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_345",
                "parent": {
                    "LARGE": [{"layoutId": "layout_123", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_171", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_269", "layoutItemId": "2"}],
                },
            },
            "widget_346": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="font-size: 16px; color: var(--dark-900);">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo, ad quo torquatos sententiae reprimique. </span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_346",
                "parent": {
                    "LARGE": [{"layoutId": "layout_123", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_171", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_269", "layoutItemId": "3"}],
                },
            },
            "widget_347": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "themeStyle": {"quickStyleType": "default"},
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                    },
                },
                "id": "widget_347",
                "parent": {
                    "LARGE": [{"layoutId": "layout_123", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_171", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_269", "layoutItemId": "4"}],
                },
            },
            "widget_348": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_348",
                "layouts": {"DEFAULT": {"LARGE": "layout_124"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_122", "layoutItemId": "2"}]},
            },
            "widget_349": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_349",
                "parent": {
                    "LARGE": [{"layoutId": "layout_124", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_172", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_270", "layoutItemId": "0"}],
                },
            },
            "widget_350": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="color: var(--primary-700); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_350",
                "parent": {
                    "LARGE": [{"layoutId": "layout_124", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_172", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_270", "layoutItemId": "1"}],
                },
            },
            "widget_351": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_351",
                "parent": {
                    "LARGE": [{"layoutId": "layout_124", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_172", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_270", "layoutItemId": "2"}],
                },
            },
            "widget_352": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="font-size: 16px; color: var(--dark-900);">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo, ad quo torquatos sententiae reprimique. </span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_352",
                "parent": {
                    "LARGE": [{"layoutId": "layout_124", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_172", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_270", "layoutItemId": "3"}],
                },
            },
            "widget_353": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "default"},
                    },
                },
                "id": "widget_353",
                "parent": {
                    "LARGE": [{"layoutId": "layout_124", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_172", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_270", "layoutItemId": "4"}],
                },
            },
            "widget_354": {
                "uri": "widgets/common/table/",
                "version": "1.5.0",
                "config": {"layersConfig": [], "arrangeType": "TABS"},
                "id": "widget_354",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_114", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_259", "layoutItemId": "3"}],
                },
            },
            "widget_355": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 36978595.474472,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": -215246.6716509963,
                                "y": 2563646.6465745047,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -6476968.02877097,
                            "ymin": -254127.9641294838,
                            "xmax": 6046474.685468977,
                            "ymax": 5381421.257278493,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_355",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_125",
                        "MEDIUM": "layout_162",
                        "SMALL": "layout_260",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_114", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_259", "layoutItemId": "4"}],
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_356": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "LIST",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap1-1612496096511.jpg",
                                "originalName": "widget_315-snap1-1612496096511",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap1-1612496096511.jpg",
                                "type": "IMAGE",
                                "size": 534041,
                                "created": 1612496096520,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "SNAPSHOT",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -32026230.541417267,
                                "ymin": -2350382.2791066626,
                                "xmax": 10162117.10217883,
                                "ymax": 11581947.74048537,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 147914381.897889,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -10932056.719619218,
                                    "y": 4615782.730689353,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "layoutName": "",
                            "layoutId": "",
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap2-1612496117082.jpg",
                                "originalName": "widget_315-snap2-1612496117082",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap2-1612496117082.jpg",
                                "type": "IMAGE",
                                "size": 517756,
                                "created": 1612496117089,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "SNAPSHOT",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 10302264.393539697,
                                "ymin": -976350.2586526736,
                                "xmax": 20849351.30443865,
                                "ymax": 2506732.246245312,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 15575807.848989174,
                                    "y": 765190.9937963192,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "layoutName": "",
                            "layoutId": "",
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap3-1612496125303.jpg",
                                "originalName": "widget_315-snap3-1612496125303",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap3-1612496125303.jpg",
                                "type": "IMAGE",
                                "size": 420471,
                                "created": 1612496125308,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "SNAPSHOT",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 11006918.248043481,
                                "ymin": -4489921.5832079,
                                "xmax": 21554005.158942435,
                                "ymax": -1006839.0783099146,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 16280461.703492958,
                                    "y": -2748380.3307589074,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "layoutName": "",
                            "layoutId": "",
                        },
                        {
                            "id": 4,
                            "name": "Bookmark-4",
                            "title": "Bookmark-4",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap4-1612496130941.jpg",
                                "originalName": "widget_315-snap4-1612496130941",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap4-1612496130941.jpg",
                                "type": "IMAGE",
                                "size": 115777,
                                "created": 1612496130942,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "SNAPSHOT",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 12609008.183549203,
                                "ymin": -13415137.953714337,
                                "xmax": 23156095.094448157,
                                "ymax": -9932055.448816352,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 17882551.63899868,
                                    "y": -11673596.701265344,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "layoutName": "",
                            "layoutId": "",
                        },
                        {
                            "id": 5,
                            "name": "Bookmark-5",
                            "title": "Bookmark-5",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap5-1612496138211.jpg",
                                "originalName": "widget_315-snap5-1612496138211",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap5-1612496138211.jpg",
                                "type": "IMAGE",
                                "size": 64288,
                                "created": 1612496138212,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "SNAPSHOT",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 15307521.920871284,
                                "ymin": -19538508.289638177,
                                "xmax": 25854608.831770238,
                                "ymax": -16055425.784740193,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 20581065.37632076,
                                    "y": -17796967.037189186,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "layoutName": "",
                            "layoutId": "",
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": True,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_356",
                "useMapWidgetIds": ["widget_314"],
                "layouts": {},
                "parent": {
                    "LARGE": [{"layoutId": "layout_114", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_259", "layoutItemId": "5"}],
                },
            },
            "widget_357": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_357",
                "layouts": {"DEFAULT": {"LARGE": "layout_127"}},
                "parent": {"LARGE": [{"layoutId": "layout_126", "layoutItemId": "0"}]},
            },
            "widget_358": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_358",
                "layouts": {"DEFAULT": {"LARGE": "layout_128"}},
                "parent": {"LARGE": [{"layoutId": "layout_127", "layoutItemId": "0"}]},
            },
            "widget_359": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_359",
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_128", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_176", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_274", "layoutItemId": "0"}],
                },
            },
            "widget_360": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_360",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_128", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_176", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_274", "layoutItemId": "1"}],
                },
            },
            "widget_361": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_361",
                "layouts": {"DEFAULT": {"LARGE": "layout_129"}},
                "parent": {"LARGE": [{"layoutId": "layout_127", "layoutItemId": "1"}]},
            },
            "widget_362": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0",
                                "offsetY": "0",
                                "blur": "0",
                                "spread": "0",
                                "color": "var(--dark-100)",
                            },
                        },
                    },
                    "itemStyle": "STYLE1",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "isOpenAdvabceSetting": False,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_362",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_130",
                        "MEDIUM": "layout_178",
                        "SMALL": "layout_276",
                    },
                    "HOVER": {
                        "LARGE": "layout_131",
                        "MEDIUM": "layout_179",
                        "SMALL": "layout_277",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_129", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_177", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_275", "layoutItemId": "0"}],
                },
            },
            "widget_363": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_363",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "0"}],
                },
            },
            "widget_364": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                    "tooltip": {"name": "", "parts": [{"type": "STRING", "exp": '""'}]},
                },
                "id": "widget_364",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "1"}],
                },
            },
            "widget_365": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark-800);">Created by:</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_365",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "2"}],
                },
            },
            "widget_366": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "var(--primary)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-400)",
                                    "width": "1px",
                                },
                                "text": {"color": "var(--light)", "size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--primary-600)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"size": 14},
                            },
                        },
                    },
                },
                "id": "widget_366",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "3"}],
                },
            },
            "widget_367": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Created date: xxx</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_367",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "4"}],
                },
            },
            "widget_368": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Views number: 123</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_368",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "5"}],
                },
            },
            "widget_369": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--secondary-200)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_369",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "6"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "6"}],
                },
            },
            "widget_370": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg "><path d= "M8 7a3 3 0 100-6 3 3 0 000 6zm0 1a4 4 0 110-8 4 4 0 010 8zm1 2H7a6.002 6.002 0 00-5.917 5h13.834A6.002 6.002 0 009 10zM7 9h2a7 7 0 017 7H0a7 7 0 017-7zm.646 3.721l1.768-1.767a.5.5 0 11.707.707l-2.475 2.475-1.767-1.768a.5.5 0 01.707-.707l1.06 1.06z " fill= "currentColor " fill-rule= "nonzero "></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_370",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "7"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "7"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "7"}],
                },
            },
            "widget_371": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="26" height="28" viewBox="0 0 26 28"><path d="M17.766 15.687l4.781-4.641-6.594-0.969-2.953-5.969-2.953 5.969-6.594 0.969 4.781 4.641-1.141 6.578 5.906-3.109 5.891 3.109zM26 10.109c0 0.281-0.203 0.547-0.406 0.75l-5.672 5.531 1.344 7.812c0.016 0.109 0.016 0.203 0.016 0.313 0 0.422-0.187 0.781-0.641 0.781-0.219 0-0.438-0.078-0.625-0.187l-7.016-3.687-7.016 3.687c-0.203 0.109-0.406 0.187-0.625 0.187-0.453 0-0.656-0.375-0.656-0.781 0-0.109 0.016-0.203 0.031-0.313l1.344-7.812-5.688-5.531c-0.187-0.203-0.391-0.469-0.391-0.75 0-0.469 0.484-0.656 0.875-0.719l7.844-1.141 3.516-7.109c0.141-0.297 0.406-0.641 0.766-0.641s0.625 0.344 0.766 0.641l3.516 7.109 7.844 1.141c0.375 0.063 0.875 0.25 0.875 0.719z"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_371",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "8"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "8"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "8"}],
                },
            },
            "widget_372": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg width="16px" height="16px" viewBox="0 0 16 16"  xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"><g id="Style" stroke="none" stroke-width="1" fill="none" fill-rule="evenodd"><g id="0-UI-icons" transform="translate(-244.000000, -245.000000)" fill="var(--black)000" fill-rule="nonzero"><path d="M252,260 C255.865993,260 259,256.865993 259,253 C259,249.134007 255.865993,246 252,246 C248.134007,246 245,249.134007 245,253 C245,256.865993 248.134007,260 252,260 Z M252,261 C247.581722,261 244,257.418278 244,253 C244,248.581722 247.581722,245 252,245 C256.418278,245 260,248.581722 260,253 C260,257.418278 256.418278,261 252,261 Z M254.544453,252.999962 L250.455473,249.985987 L250.455473,256.013937 L254.544453,252.999962 Z M255.09981,253.802292 L251.01083,256.816267 C250.581429,257.132778 249.984688,257.030145 249.677973,256.58703 C249.562223,256.419806 249.5,256.219439 249.5,256.013937 L249.5,249.985987 C249.5,249.441441 249.92778,249 250.455473,249 C250.654615,249 250.848781,249.064211 251.01083,249.183656 L255.09981,252.197631 C255.529212,252.514142 255.628669,253.12994 255.321953,253.573055 C255.260669,253.661592 255.185607,253.739051 255.09981,253.802292 Z" id="app-launch"></path></g></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {"color": "", "fillType": "fill"},
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_372",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "9"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "9"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "9"}],
                },
            },
            "widget_373": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="32" height="11" viewBox="0 0 32 11"><path d="M4 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S5.7 8.5 4 8.5z M16 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S17.7 8.5 16 8.5z M28 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S29.7 8.5 28 8.5z"/></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_373",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "10"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "10"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "10"}],
                },
            },
            "widget_374": {
                "uri": "widgets/common/list/",
                "version": "1.5.0",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE5",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "scrollBarOpen": True,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {
                                "LARGE": {"width": 1355, "height": 180.29},
                                "MEDIUM": {"width": 545.31, "height": 186.92},
                                "SMALL": {"width": 403.75, "height": 103.76},
                            },
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {
                                "LARGE": {"width": 1355, "height": 180.29},
                                "MEDIUM": {"width": 545.31, "height": 186.92},
                                "SMALL": {"width": 403.75, "height": 103.76},
                            },
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {
                                "LARGE": {"width": 1355, "height": 180.29},
                                "MEDIUM": {"width": 545.31, "height": 186.92},
                                "SMALL": {"width": 403.75, "height": 103.76},
                            },
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-500)",
                                    "width": "1px",
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                },
                "id": "widget_374",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_143",
                        "MEDIUM": "layout_180",
                        "SMALL": "layout_278",
                    },
                    "SELECTED": {
                        "LARGE": "layout_133",
                        "MEDIUM": "layout_181",
                        "SMALL": "layout_279",
                    },
                    "HOVER": {
                        "LARGE": "layout_134",
                        "MEDIUM": "layout_182",
                        "SMALL": "layout_280",
                    },
                },
                "useDataSourcesEnabled": True,
                "style": {"border": {"type": "solid", "color": "", "width": "1px"}},
                "parent": {
                    "LARGE": [{"layoutId": "layout_129", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_177", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_275", "layoutItemId": "1"}],
                },
            },
            "widget_381": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_381",
                "layouts": {"DEFAULT": {"LARGE": "layout_135"}},
                "parent": {"LARGE": [{"layoutId": "layout_127", "layoutItemId": "2"}]},
            },
            "widget_382": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_382",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_135", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_183", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_281", "layoutItemId": "0"}],
                },
            },
            "widget_383": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_383",
                "layouts": {"DEFAULT": {"LARGE": "layout_136"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_135", "layoutItemId": "1"}]},
            },
            "widget_384": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_384",
                "parent": {
                    "LARGE": [{"layoutId": "layout_136", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_184", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_282", "layoutItemId": "0"}],
                },
            },
            "widget_385": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px; color: var(--primary);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_385",
                "parent": {
                    "LARGE": [{"layoutId": "layout_136", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_184", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_282", "layoutItemId": "1"}],
                },
            },
            "widget_386": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_386",
                "parent": {
                    "LARGE": [{"layoutId": "layout_136", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_184", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_282", "layoutItemId": "2"}],
                },
            },
            "widget_387": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="color: var(--dark-900); font-size: 16px;">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo.</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_387",
                "parent": {
                    "LARGE": [{"layoutId": "layout_136", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_184", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_282", "layoutItemId": "3"}],
                },
            },
            "widget_388": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "themeStyle": {"quickStyleType": "default"},
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                    },
                },
                "id": "widget_388",
                "parent": {
                    "LARGE": [{"layoutId": "layout_136", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_184", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_282", "layoutItemId": "4"}],
                },
            },
            "widget_389": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_389",
                "layouts": {"DEFAULT": {"LARGE": "layout_137"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_135", "layoutItemId": "2"}]},
            },
            "widget_390": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_390",
                "parent": {
                    "LARGE": [{"layoutId": "layout_137", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_185", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_283", "layoutItemId": "0"}],
                },
            },
            "widget_391": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px; color: var(--primary);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_391",
                "parent": {
                    "LARGE": [{"layoutId": "layout_137", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_185", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_283", "layoutItemId": "1"}],
                },
            },
            "widget_392": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_392",
                "parent": {
                    "LARGE": [{"layoutId": "layout_137", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_185", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_283", "layoutItemId": "2"}],
                },
            },
            "widget_393": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="color: var(--dark-900); font-size: 16px;">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo.</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_393",
                "parent": {
                    "LARGE": [{"layoutId": "layout_137", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_185", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_283", "layoutItemId": "3"}],
                },
            },
            "widget_394": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "default"},
                    },
                },
                "id": "widget_394",
                "parent": {
                    "LARGE": [{"layoutId": "layout_137", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_185", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_283", "layoutItemId": "4"}],
                },
            },
            "widget_395": {
                "uri": "widgets/common/table/",
                "version": "1.5.0",
                "config": {"layersConfig": [], "arrangeType": "TABS"},
                "id": "widget_395",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_127", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_272", "layoutItemId": "3"}],
                },
            },
            "widget_396": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 73957190.948944,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": 3052589.1615960076,
                                "y": -449042.3862566479,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -9470853.55264394,
                            "ymin": -4577864.906107631,
                            "xmax": 15576031.875835955,
                            "ymax": 3679780.133594335,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_396",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_138",
                        "MEDIUM": "layout_175",
                        "SMALL": "layout_273",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_127", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_272", "layoutItemId": "4"}],
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_397": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap1-1612496096511.jpg",
                                "originalName": "widget_315-snap1-1612496096511",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap1-1612496096511.jpg",
                                "type": "IMAGE",
                                "size": 534041,
                                "created": 1612496096520,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -32026230.541417267,
                                "ymin": -2350382.2791066626,
                                "xmax": 10162117.10217883,
                                "ymax": 11581947.74048537,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 147914381.897889,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -10932056.719619218,
                                    "y": 4615782.730689353,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap2-1612496117082.jpg",
                                "originalName": "widget_315-snap2-1612496117082",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap2-1612496117082.jpg",
                                "type": "IMAGE",
                                "size": 517756,
                                "created": 1612496117089,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 10302264.393539697,
                                "ymin": -976350.2586526736,
                                "xmax": 20849351.30443865,
                                "ymax": 2506732.246245312,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 15575807.848989174,
                                    "y": 765190.9937963192,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap3-1612496125303.jpg",
                                "originalName": "widget_315-snap3-1612496125303",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap3-1612496125303.jpg",
                                "type": "IMAGE",
                                "size": 420471,
                                "created": 1612496125308,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 11006918.248043481,
                                "ymin": -4489921.5832079,
                                "xmax": 21554005.158942435,
                                "ymax": -1006839.0783099146,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 16280461.703492958,
                                    "y": -2748380.3307589074,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 4,
                            "name": "Bookmark-4",
                            "title": "Bookmark-4",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap4-1612496130941.jpg",
                                "originalName": "widget_315-snap4-1612496130941",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap4-1612496130941.jpg",
                                "type": "IMAGE",
                                "size": 115777,
                                "created": 1612496130942,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 12609008.183549203,
                                "ymin": -13415137.953714337,
                                "xmax": 23156095.094448157,
                                "ymax": -9932055.448816352,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 17882551.63899868,
                                    "y": -11673596.701265344,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 5,
                            "name": "Bookmark-5",
                            "title": "Bookmark-5",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap5-1612496138211.jpg",
                                "originalName": "widget_315-snap5-1612496138211",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap5-1612496138211.jpg",
                                "type": "IMAGE",
                                "size": 64288,
                                "created": 1612496138212,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 15307521.920871284,
                                "ymin": -19538508.289638177,
                                "xmax": 25854608.831770238,
                                "ymax": -16055425.784740193,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 20581065.37632076,
                                    "y": -17796967.037189186,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 6,
                            "name": "Bookmark-6",
                            "title": "Bookmark-6",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap6-1612496145998.jpg",
                                "originalName": "widget_315-snap6-1612496145998",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap6-1612496145998.jpg",
                                "type": "IMAGE",
                                "size": 428023,
                                "created": 1612496146003,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 36283391.31498645,
                                "ymin": -3377431.897210424,
                                "xmax": 46830478.225885406,
                                "ymax": 105650.607687562,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 41556934.77043593,
                                    "y": -1635890.6447614308,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 7,
                            "name": "Bookmark-7",
                            "title": "Bookmark-7",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap7-1612496158371.jpg",
                                "originalName": "widget_315-snap7-1612496158371",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap7-1612496158371.jpg",
                                "type": "IMAGE",
                                "size": 626113,
                                "created": 1612496158378,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -8870329.77287056,
                                "ymin": -1763933.611939429,
                                "xmax": -6233558.045145821,
                                "ymax": -893162.9857149327,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7551943.90900819,
                                    "y": -1328548.2988271809,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 8,
                            "name": "Bookmark-8",
                            "title": "Bookmark-8",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap8-1612496164229.jpg",
                                "originalName": "widget_315-snap8-1612496164229",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap8-1612496164229.jpg",
                                "type": "IMAGE",
                                "size": 534610,
                                "created": 1612496164235,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -6674948.543320371,
                                "ymin": -2972099.0698932745,
                                "xmax": -4038176.8155956324,
                                "ymax": -2101328.443668778,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -5356562.679458002,
                                    "y": -2536713.7567810263,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 9,
                            "name": "Bookmark-9",
                            "title": "Bookmark-9",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap9-1612496169112.jpg",
                                "originalName": "widget_315-snap9-1612496169112",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap9-1612496169112.jpg",
                                "type": "IMAGE",
                                "size": 459797,
                                "created": 1612496169118,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -5716782.491468654,
                                "ymin": -1144639.4521959827,
                                "xmax": -3080010.7637439156,
                                "ymax": -273868.8259714863,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -4398396.627606285,
                                    "y": -709254.1390837345,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 10,
                            "name": "Bookmark-10",
                            "title": "Bookmark-10",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap10-1612496173280.jpg",
                                "originalName": "widget_315-snap10-1612496173280",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap10-1612496173280.jpg",
                                "type": "IMAGE",
                                "size": 606004,
                                "created": 1612496173287,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -7710209.16444735,
                                "ymin": -1181014.7935382156,
                                "xmax": -5073437.436722612,
                                "ymax": -310244.16731371905,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -6391823.300584981,
                                    "y": -745629.4804259673,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": True,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_397",
                "useMapWidgetIds": ["widget_314"],
                "parent": {
                    "LARGE": [{"layoutId": "layout_127", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_272", "layoutItemId": "5"}],
                },
            },
            "widget_400": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "cropShape": "circle",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "svgViewBox": "0 0 28 28",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_400",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_140", "layoutItemId": "0"},
                        {"layoutId": "layout_120", "layoutItemId": "0"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_167", "layoutItemId": "0"},
                        {"layoutId": "layout_168", "layoutItemId": "0"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_265", "layoutItemId": "0"},
                        {"layoutId": "layout_266", "layoutItemId": "0"},
                    ],
                },
            },
            "widget_401": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 16px;color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_401",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_140", "layoutItemId": "1"},
                        {"layoutId": "layout_120", "layoutItemId": "1"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_167", "layoutItemId": "1"},
                        {"layoutId": "layout_168", "layoutItemId": "1"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_265", "layoutItemId": "1"},
                        {"layoutId": "layout_266", "layoutItemId": "1"},
                    ],
                },
            },
            "widget_402": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "Button",
                        "toolTip": "",
                        "linkParam": {},
                    },
                    "styleConfig": {
                        "name": "Default",
                        "themeStyle": {
                            "className": "jimu-widget d-block",
                            "color": "primary",
                        },
                        "customStyle": {"regular": {}, "hover": {}, "clicked": {}},
                    },
                },
                "id": "widget_402",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_140", "layoutItemId": "2"},
                        {"layoutId": "layout_120", "layoutItemId": "2"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_167", "layoutItemId": "2"},
                        {"layoutId": "layout_168", "layoutItemId": "2"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_265", "layoutItemId": "2"},
                        {"layoutId": "layout_266", "layoutItemId": "2"},
                    ],
                },
            },
            "widget_409": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "cropShape": "circle",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "svgViewBox": "0 0 28 28",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_409",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_143", "layoutItemId": "0"},
                        {"layoutId": "layout_133", "layoutItemId": "0"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_180", "layoutItemId": "0"},
                        {"layoutId": "layout_181", "layoutItemId": "0"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_278", "layoutItemId": "0"},
                        {"layoutId": "layout_279", "layoutItemId": "0"},
                    ],
                },
            },
            "widget_410": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 16px;color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_410",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_143", "layoutItemId": "1"},
                        {"layoutId": "layout_133", "layoutItemId": "1"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_180", "layoutItemId": "1"},
                        {"layoutId": "layout_181", "layoutItemId": "1"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_278", "layoutItemId": "1"},
                        {"layoutId": "layout_279", "layoutItemId": "1"},
                    ],
                },
            },
            "widget_411": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "Button",
                        "toolTip": "",
                        "linkParam": {},
                    },
                    "styleConfig": {
                        "name": "Default",
                        "themeStyle": {
                            "className": "jimu-widget d-block",
                            "color": "primary",
                        },
                        "customStyle": {"regular": {}, "hover": {}, "clicked": {}},
                    },
                },
                "id": "widget_411",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_143", "layoutItemId": "2"},
                        {"layoutId": "layout_133", "layoutItemId": "2"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_180", "layoutItemId": "2"},
                        {"layoutId": "layout_181", "layoutItemId": "2"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_278", "layoutItemId": "2"},
                        {"layoutId": "layout_279", "layoutItemId": "2"},
                    ],
                },
            },
            "widget_412": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [60, 20, 40, 20], "unit": "px"}},
                },
                "id": "widget_412",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_145"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_144", "layoutItemId": "0"}]},
            },
            "widget_413": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_413",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_146"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_144", "layoutItemId": "1"}]},
            },
            "widget_414": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_414",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_148"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_147", "layoutItemId": "0"}]},
            },
            "widget_415": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_415",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_150"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "0"}]},
            },
            "widget_416": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_416",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_151"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "1"}]},
            },
            "widget_417": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_417",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_157"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "2"}]},
            },
            "widget_418": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_418",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_158"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_157", "layoutItemId": "1"}]},
            },
            "widget_419": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_419",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_159"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_157", "layoutItemId": "2"}]},
            },
            "widget_420": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_420",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_161"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_160", "layoutItemId": "0"}]},
            },
            "widget_421": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_421",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_163"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "0"}]},
            },
            "widget_422": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_422",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_164"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "1"}]},
            },
            "widget_423": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_423",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_170"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "2"}]},
            },
            "widget_424": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_424",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_171"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_170", "layoutItemId": "1"}]},
            },
            "widget_425": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_425",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_172"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_170", "layoutItemId": "2"}]},
            },
            "widget_426": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_426",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_174"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_173", "layoutItemId": "0"}]},
            },
            "widget_427": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_427",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_176"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "0"}]},
            },
            "widget_428": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_428",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_177"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "1"}]},
            },
            "widget_429": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_429",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_183"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "2"}]},
            },
            "widget_430": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_430",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_184"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_183", "layoutItemId": "1"}]},
            },
            "widget_431": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_431",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_185"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_183", "layoutItemId": "2"}]},
            },
            "widget_432": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 20, 0, 20], "unit": "px"}},
                },
                "id": "widget_432",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_188"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "0"}]},
            },
            "widget_433": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 20, 0, 20], "unit": "px"}},
                },
                "id": "widget_433",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_189"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "1"}]},
            },
            "widget_434": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_434",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_190"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "2"}]},
            },
            "widget_435": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_435",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_191"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "3"}]},
            },
            "widget_436": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {"padding": {"number": [20, 0, 20, 20], "unit": "px"}},
                },
                "id": "widget_436",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_192"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_191", "layoutItemId": "0"}]},
            },
            "widget_437": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_437",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_193"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_192", "layoutItemId": "0"}]},
            },
            "widget_438": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_438",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_194"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_193", "layoutItemId": "2"}]},
            },
            "widget_439": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_439",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_195"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_192", "layoutItemId": "1"}]},
            },
            "widget_440": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_440",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_196"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_195", "layoutItemId": "2"}]},
            },
            "widget_441": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_441",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_197"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_192", "layoutItemId": "4"}]},
            },
            "widget_442": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_442",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_198"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_197", "layoutItemId": "2"}]},
            },
            "widget_443": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_443",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_199"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_191", "layoutItemId": "1"}]},
            },
            "widget_444": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_444",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_200"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "4"}]},
            },
            "widget_445": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_445",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_201"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_200", "layoutItemId": "1"}]},
            },
            "widget_446": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_446",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_202"}},
                "style": {
                    "background": {
                        "color": "rgba(7,111,229,0.1)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_201", "layoutItemId": "3"}]},
            },
            "widget_447": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 20, 0, 20], "unit": "px"}},
                },
                "id": "widget_447",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_203"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "5"}]},
            },
            "widget_448": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_448",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_204"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "6"}]},
            },
            "widget_449": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_449",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_205"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "7"}]},
            },
            "widget_450": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_450",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_208"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_205", "layoutItemId": "0"}]},
            },
            "widget_451": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_451",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_213"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_205", "layoutItemId": "1"}]},
            },
            "widget_452": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 20, 0, 20], "unit": "px"}},
                },
                "id": "widget_452",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_218"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "8"}]},
            },
            "widget_453": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_453",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_219"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "9"}]},
            },
            "widget_454": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_454",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_220"}},
                "parent": {
                    "MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "10"}]
                },
            },
            "widget_455": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 30},
                    "style": {
                        "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_455",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_221"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_220", "layoutItemId": "0"}]},
            },
            "widget_456": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_456",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_223"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_222", "layoutItemId": "0"}]},
            },
            "widget_457": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_457",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_225"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_223", "layoutItemId": "1"}]},
            },
            "widget_458": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_458",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_227"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_226", "layoutItemId": "0"}]},
            },
            "widget_459": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_459",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_229"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_227", "layoutItemId": "1"}]},
            },
            "widget_460": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 20, 0, 20], "unit": "px"}},
                },
                "id": "widget_460",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_233"}},
                "parent": {
                    "MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "11"}]
                },
            },
            "widget_461": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 20, 20, 20], "unit": "px"}},
                },
                "id": "widget_461",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_234"}},
                "parent": {
                    "MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "12"}]
                },
            },
            "widget_462": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_462",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_235"}},
                "parent": {
                    "MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "13"}]
                },
            },
            "widget_463": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 20, 20, 20], "unit": "px"}},
                },
                "id": "widget_463",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_236"}},
                "parent": {
                    "MEDIUM": [{"layoutId": "layout_187", "layoutItemId": "14"}]
                },
            },
            "widget_464": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 10},
                    "style": {
                        "padding": {"number": [20, 10, 20, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_464",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_237"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_236", "layoutItemId": "0"}]},
            },
            "widget_465": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_465",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_238"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_237", "layoutItemId": "1"}]},
            },
            "widget_466": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_466",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_239"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_237", "layoutItemId": "2"}]},
            },
            "widget_467": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_467",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_240"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_237", "layoutItemId": "3"}]},
            },
            "widget_468": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_468",
                "layouts": {"DEFAULT": {"SMALL": "layout_243"}},
                "parent": {"SMALL": [{"layoutId": "layout_242", "layoutItemId": "0"}]},
            },
            "widget_469": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_469",
                "layouts": {"DEFAULT": {"SMALL": "layout_244"}},
                "parent": {"SMALL": [{"layoutId": "layout_242", "layoutItemId": "1"}]},
            },
            "widget_470": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_470",
                "layouts": {"DEFAULT": {"SMALL": "layout_246"}},
                "parent": {"SMALL": [{"layoutId": "layout_245", "layoutItemId": "0"}]},
            },
            "widget_471": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_471",
                "layouts": {"DEFAULT": {"SMALL": "layout_248"}},
                "parent": {"SMALL": [{"layoutId": "layout_246", "layoutItemId": "0"}]},
            },
            "widget_472": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_472",
                "layouts": {"DEFAULT": {"SMALL": "layout_249"}},
                "parent": {"SMALL": [{"layoutId": "layout_246", "layoutItemId": "1"}]},
            },
            "widget_473": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_473",
                "layouts": {"DEFAULT": {"SMALL": "layout_255"}},
                "parent": {"SMALL": [{"layoutId": "layout_246", "layoutItemId": "2"}]},
            },
            "widget_474": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_474",
                "layouts": {"DEFAULT": {"SMALL": "layout_256"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_255", "layoutItemId": "1"}]},
            },
            "widget_475": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_475",
                "layouts": {"DEFAULT": {"SMALL": "layout_257"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_255", "layoutItemId": "2"}]},
            },
            "widget_476": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_476",
                "layouts": {"DEFAULT": {"SMALL": "layout_259"}},
                "parent": {"SMALL": [{"layoutId": "layout_258", "layoutItemId": "0"}]},
            },
            "widget_477": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_477",
                "layouts": {"DEFAULT": {"SMALL": "layout_261"}},
                "parent": {"SMALL": [{"layoutId": "layout_259", "layoutItemId": "0"}]},
            },
            "widget_478": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_478",
                "layouts": {"DEFAULT": {"SMALL": "layout_262"}},
                "parent": {"SMALL": [{"layoutId": "layout_259", "layoutItemId": "1"}]},
            },
            "widget_479": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_479",
                "layouts": {"DEFAULT": {"SMALL": "layout_268"}},
                "parent": {"SMALL": [{"layoutId": "layout_259", "layoutItemId": "2"}]},
            },
            "widget_480": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 10},
                    "style": {
                        "padding": {"number": [0, 0, 10, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_480",
                "layouts": {"DEFAULT": {"SMALL": "layout_269"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_268", "layoutItemId": "1"}]},
            },
            "widget_481": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_481",
                "layouts": {"DEFAULT": {"SMALL": "layout_270"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_268", "layoutItemId": "2"}]},
            },
            "widget_482": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_482",
                "layouts": {"DEFAULT": {"SMALL": "layout_272"}},
                "parent": {"SMALL": [{"layoutId": "layout_271", "layoutItemId": "0"}]},
            },
            "widget_483": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_483",
                "layouts": {"DEFAULT": {"SMALL": "layout_274"}},
                "parent": {"SMALL": [{"layoutId": "layout_272", "layoutItemId": "0"}]},
            },
            "widget_484": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_484",
                "layouts": {"DEFAULT": {"SMALL": "layout_275"}},
                "parent": {"SMALL": [{"layoutId": "layout_272", "layoutItemId": "1"}]},
            },
            "widget_485": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_485",
                "layouts": {"DEFAULT": {"SMALL": "layout_281"}},
                "parent": {"SMALL": [{"layoutId": "layout_272", "layoutItemId": "2"}]},
            },
            "widget_486": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_486",
                "layouts": {"DEFAULT": {"SMALL": "layout_282"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_281", "layoutItemId": "1"}]},
            },
            "widget_487": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_487",
                "layouts": {"DEFAULT": {"SMALL": "layout_283"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_281", "layoutItemId": "2"}]},
            },
            "widget_488": {
                "uri": "widgets/layout/column/",
                "id": "widget_488",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_284"}},
                "parent": {"SMALL": [{"layoutId": "layout_243", "layoutItemId": "1"}]},
            },
            "widget_489": {
                "uri": "widgets/layout/column/",
                "id": "widget_489",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_285"}},
                "parent": {"SMALL": [{"layoutId": "layout_244", "layoutItemId": "1"}]},
            },
            "widget_490": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 0, 10], "unit": "px"}},
                },
                "id": "widget_490",
                "layouts": {"DEFAULT": {"SMALL": "layout_289"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "0"}]},
            },
            "widget_491": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 0, 10], "unit": "px"}},
                },
                "id": "widget_491",
                "layouts": {"DEFAULT": {"SMALL": "layout_290"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "1"}]},
            },
            "widget_492": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_492",
                "layouts": {"DEFAULT": {"SMALL": "layout_291"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "2"}]},
            },
            "widget_493": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_493",
                "layouts": {"DEFAULT": {"SMALL": "layout_292"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "3"}]},
            },
            "widget_494": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 0,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_494",
                "layouts": {"DEFAULT": {"SMALL": "layout_293"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_7", "layoutItemId": "0"}]},
            },
            "widget_495": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_495",
                "layouts": {"DEFAULT": {"SMALL": "layout_294"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "0"}]},
            },
            "widget_496": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_496",
                "layouts": {"DEFAULT": {"SMALL": "layout_295"}},
                "parent": {"LARGE": [{"layoutId": "layout_9", "layoutItemId": "2"}]},
            },
            "widget_497": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_497",
                "layouts": {"DEFAULT": {"SMALL": "layout_296"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "4"}]},
            },
            "widget_498": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_498",
                "layouts": {"DEFAULT": {"SMALL": "layout_297"}},
                "parent": {"LARGE": [{"layoutId": "layout_13", "layoutItemId": "2"}]},
            },
            "widget_499": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_499",
                "layouts": {"DEFAULT": {"SMALL": "layout_298"}},
                "parent": {"LARGE": [{"layoutId": "layout_8", "layoutItemId": "2"}]},
            },
            "widget_500": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_500",
                "layouts": {"DEFAULT": {"SMALL": "layout_299"}},
                "parent": {"LARGE": [{"layoutId": "layout_11", "layoutItemId": "2"}]},
            },
            "widget_501": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_501",
                "layouts": {"DEFAULT": {"SMALL": "layout_300"}},
                "style": {
                    "background": {
                        "color": "var(--primary)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_7", "layoutItemId": "1"}]},
            },
            "widget_502": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_502",
                "layouts": {"DEFAULT": {"SMALL": "layout_301"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "4"}]},
            },
            "widget_503": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"number": [40, 20, 40, 20], "unit": "px"},
                        "justifyContent": "center",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_503",
                "layouts": {"DEFAULT": {"SMALL": "layout_302"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_16", "layoutItemId": "1"}]},
            },
            "widget_504": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10], "unit": "px"}},
                },
                "id": "widget_504",
                "layouts": {"DEFAULT": {"SMALL": "layout_303"}},
                "style": {
                    "background": {
                        "color": "rgba(7,111,229,0.1)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    }
                },
                "parent": {"LARGE": [{"layoutId": "layout_17", "layoutItemId": "3"}]},
            },
            "widget_505": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 0, 10], "unit": "px"}},
                },
                "id": "widget_505",
                "layouts": {"DEFAULT": {"SMALL": "layout_304"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "5"}]},
            },
            "widget_506": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_506",
                "layouts": {"DEFAULT": {"SMALL": "layout_305"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "6"}]},
            },
            "widget_507": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_507",
                "layouts": {"DEFAULT": {"SMALL": "layout_306"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "7"}]},
            },
            "widget_508": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_508",
                "layouts": {"DEFAULT": {"SMALL": "layout_309"}},
                "parent": {"LARGE": [{"layoutId": "layout_21", "layoutItemId": "0"}]},
            },
            "widget_509": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_509",
                "layouts": {"DEFAULT": {"SMALL": "layout_314"}},
                "parent": {"LARGE": [{"layoutId": "layout_21", "layoutItemId": "1"}]},
            },
            "widget_510": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 0, 10], "unit": "px"}},
                },
                "id": "widget_510",
                "layouts": {"DEFAULT": {"SMALL": "layout_319"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "8"}]},
            },
            "widget_511": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_511",
                "layouts": {"DEFAULT": {"SMALL": "layout_320"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "9"}]},
            },
            "widget_512": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_512",
                "layouts": {"DEFAULT": {"SMALL": "layout_321"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "12"}]},
            },
            "widget_513": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 30},
                    "style": {
                        "padding": {"number": [40, 0, 0, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_513",
                "layouts": {"DEFAULT": {"SMALL": "layout_322"}},
                "style": {
                    "background": {"color": "", "fillType": "fit", "image": {"url": ""}}
                },
                "parent": {"LARGE": [{"layoutId": "layout_65", "layoutItemId": "0"}]},
            },
            "widget_514": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_514",
                "layouts": {"DEFAULT": {"SMALL": "layout_324"}},
                "parent": {"LARGE": [{"layoutId": "layout_63", "layoutItemId": "0"}]},
            },
            "widget_515": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_515",
                "layouts": {"DEFAULT": {"SMALL": "layout_326"}},
                "parent": {"LARGE": [{"layoutId": "layout_67", "layoutItemId": "1"}]},
            },
            "widget_516": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_516",
                "layouts": {"DEFAULT": {"SMALL": "layout_328"}},
                "parent": {"LARGE": [{"layoutId": "layout_78", "layoutItemId": "0"}]},
            },
            "widget_517": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 40},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_517",
                "layouts": {"DEFAULT": {"SMALL": "layout_330"}},
                "parent": {"LARGE": [{"layoutId": "layout_79", "layoutItemId": "1"}]},
            },
            "widget_518": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 0, 10], "unit": "px"}},
                },
                "id": "widget_518",
                "layouts": {"DEFAULT": {"SMALL": "layout_334"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "13"}]},
            },
            "widget_519": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_519",
                "layouts": {"DEFAULT": {"SMALL": "layout_335"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "14"}]},
            },
            "widget_520": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_520",
                "layouts": {"DEFAULT": {"SMALL": "layout_336"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "15"}]},
            },
            "widget_521": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [10, 10, 20, 10], "unit": "px"}},
                },
                "id": "widget_521",
                "layouts": {"DEFAULT": {"SMALL": "layout_337"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "16"}]},
            },
            "widget_522": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [30, 20, 30, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_522",
                "layouts": {"DEFAULT": {"SMALL": "layout_338"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_73", "layoutItemId": "0"}]},
            },
            "widget_523": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_523",
                "layouts": {"DEFAULT": {"SMALL": "layout_339"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "1"}]},
            },
            "widget_524": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_524",
                "layouts": {"DEFAULT": {"SMALL": "layout_340"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "2"}]},
            },
            "widget_525": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_525",
                "layouts": {"DEFAULT": {"SMALL": "layout_341"}},
                "parent": {"LARGE": [{"layoutId": "layout_74", "layoutItemId": "3"}]},
            },
            "widget_526": {
                "uri": "widgets/layout/column/",
                "id": "widget_526",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_342"}},
            },
            "widget_527": {
                "uri": "widgets/layout/column/",
                "id": "widget_527",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_343"}},
            },
            "widget_528": {
                "uri": "widgets/layout/column/",
                "id": "widget_528",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_344"}},
            },
            "widget_529": {
                "uri": "widgets/layout/column/",
                "id": "widget_529",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_345"}},
            },
            "widget_530": {
                "uri": "widgets/layout/column/",
                "id": "widget_530",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_346"}},
            },
            "widget_531": {
                "uri": "widgets/layout/column/",
                "id": "widget_531",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_347"}},
            },
            "widget_532": {
                "uri": "widgets/layout/column/",
                "id": "widget_532",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_348"}},
            },
            "widget_533": {
                "uri": "widgets/layout/column/",
                "id": "widget_533",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_349"}},
            },
            "widget_534": {
                "uri": "widgets/layout/column/",
                "id": "widget_534",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_350"}},
            },
            "widget_535": {
                "uri": "widgets/layout/column/",
                "id": "widget_535",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_351"}},
            },
            "widget_536": {
                "uri": "widgets/layout/column/",
                "id": "widget_536",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_352"}},
            },
            "widget_537": {
                "uri": "widgets/layout/column/",
                "id": "widget_537",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_353"}},
            },
            "widget_538": {
                "uri": "widgets/layout/column/",
                "id": "widget_538",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_354"}},
            },
            "widget_539": {
                "uri": "widgets/layout/column/",
                "id": "widget_539",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_355"}},
            },
            "widget_540": {
                "uri": "widgets/layout/column/",
                "id": "widget_540",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_356"}},
            },
            "widget_541": {
                "uri": "widgets/layout/column/",
                "id": "widget_541",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_357"}},
            },
            "widget_542": {
                "uri": "widgets/layout/column/",
                "id": "widget_542",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_358"}},
            },
            "widget_543": {
                "uri": "widgets/layout/column/",
                "id": "widget_543",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_359"}},
            },
            "widget_544": {
                "uri": "widgets/layout/column/",
                "id": "widget_544",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_360"}},
            },
            "widget_545": {
                "uri": "widgets/layout/column/",
                "id": "widget_545",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_361"}},
            },
            "widget_546": {
                "uri": "widgets/layout/column/",
                "id": "widget_546",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_362"}},
            },
            "widget_547": {
                "uri": "widgets/layout/column/",
                "id": "widget_547",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_363"}},
            },
            "widget_548": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_2",
                        "views": ["view_3", "view_4", "view_5"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "pills",
                        "standard": {
                            "gap": "40px",
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": True,
                        "advanced": True,
                        "variant": {
                            "size": {
                                "fontSize": "0.875rem",
                                "lineHeight": 1.5,
                                "paddingX": "1rem",
                                "paddingY": "0.25rem",
                            },
                            "gutter": "0.25rem",
                            "icon": {"spacing": "0.5rem"},
                            "root": {"borderRadius": "50rem", "bg": ""},
                            "item": {
                                "default": {
                                    "color": "var(--dark-800)",
                                    "border": {"width": "0px"},
                                    "borderRadius": "2px",
                                    "size": "32px",
                                },
                                "hover": {
                                    "color": "var(--dark-300)",
                                    "borderRadius": "2px",
                                    "size": "32px",
                                },
                                "active": {
                                    "color": "var(--light)",
                                    "bg": "var(--primary)",
                                    "borderRadius": "2px",
                                    "size": "32px",
                                    "border": {
                                        "type": "solid",
                                        "color": "",
                                        "width": "0px",
                                    },
                                },
                                "disabled": {
                                    "color": "#989898",
                                    "shadow": "none",
                                    "borderRadius": "50rem",
                                },
                            },
                        },
                    },
                },
                "id": "widget_548",
                "parent": {
                    "LARGE": [{"layoutId": "layout_92", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_145", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_284", "layoutItemId": "0"}],
                },
            },
        },
        "widgetsManifest": {},
        "views": {
            "view_1": {
                "id": "view_1",
                "layout": {
                    "LARGE": "layout_63",
                    "MEDIUM": "layout_222",
                    "SMALL": "layout_323",
                },
                "parent": "section_1",
            },
            "view_2": {
                "id": "view_2",
                "layout": {
                    "LARGE": "layout_78",
                    "MEDIUM": "layout_226",
                    "SMALL": "layout_327",
                },
                "parent": "section_1",
            },
            "view_3": {
                "id": "view_3",
                "layout": {
                    "LARGE": "layout_94",
                    "MEDIUM": "layout_147",
                    "SMALL": "layout_245",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "rgba(240,240,240,0)",
                "backgroundIMage": {},
                "parent": "section_2",
            },
            "view_4": {
                "id": "view_4",
                "layout": {
                    "LARGE": "layout_113",
                    "MEDIUM": "layout_160",
                    "SMALL": "layout_258",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "rgba(240,240,240,0)",
                "backgroundIMage": {},
                "parent": "section_2",
            },
            "view_5": {
                "id": "view_5",
                "layout": {
                    "LARGE": "layout_126",
                    "MEDIUM": "layout_173",
                    "SMALL": "layout_271",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "rgba(240,240,240,0)",
                "backgroundIMage": {},
                "parent": "section_2",
            },
        },
        "sections": {
            "section_1": {
                "id": "section_1",
                "views": ["view_1", "view_2"],
                "parent": {
                    "LARGE": [{"layoutId": "layout_66", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_221", "layoutItemId": "1"}],
                },
            },
            "section_2": {
                "id": "section_2",
                "views": ["view_3", "view_4", "view_5"],
                "parent": {
                    "LARGE": [{"layoutId": "layout_93", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_146", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_285", "layoutItemId": "0"}],
                },
            },
        },
        "dialogs": {
            "dialog_1": {
                "id": "dialog_1",
                "mode": "ANCHORED",
                "sizeMode": {
                    "LARGE": {"position": None, "width": "360px", "height": "240px"}
                },
                "interactionType": None,
                "interactionStyles": None,
                "confirmBeforeClose": True,
                "alwaysShowConfirmation": False,
                "dialogBackground": {
                    "color": "var(--light)",
                    "fillType": "fit",
                    "image": {"url": ""},
                },
                "overlayColor": None,
                "layout": {"LARGE": "layout_90"},
                "index": 0,
                "isSplash": False,
                "closeWhenClickOverlay": True,
                "effect": {
                    "type": "FADE_IN",
                    "option": {"direction": None, "configType": "DEFAULT"},
                },
                "oneByOneEffect": {"type": "NONE", "option": {"direction": None}},
                "boxShadow": {
                    "offsetX": "0px",
                    "offsetY": "0px",
                    "blur": "0px",
                    "spread": "0px",
                    "color": "",
                },
            },
            "dialog_2": {
                "id": "dialog_2",
                "mode": "ANCHORED",
                "sizeMode": {
                    "LARGE": {"position": None, "width": "360px", "height": "240px"}
                },
                "interactionType": None,
                "interactionStyles": None,
                "confirmBeforeClose": True,
                "alwaysShowConfirmation": False,
                "dialogBackground": {
                    "color": "var(--light)",
                    "fillType": "fit",
                    "image": {"url": ""},
                },
                "overlayColor": None,
                "layout": {"LARGE": "layout_91"},
                "index": 1,
                "isSplash": False,
                "closeWhenClickOverlay": True,
                "effect": {
                    "type": "FADE_IN",
                    "option": {"direction": None, "configType": "DEFAULT"},
                },
                "oneByOneEffect": {"type": "NONE", "option": {"direction": None}},
            },
        },
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_0": []}, {"page_1": []}],
        "header": {
            "layout": {
                "LARGE": "layout_1",
                "MEDIUM": "layout_186",
                "SMALL": "layout_241",
            },
            "height": {"LARGE": "160px", "SMALL": "180px"},
            "backgroundColor": "var(--dark-700)",
        },
        "footer": {
            "layout": {
                "LARGE": "layout_2",
                "MEDIUM": "layout_286",
                "SMALL": "layout_287",
            },
            "height": {"LARGE": "80px"},
            "backgroundColor": "var(--light)",
        },
        "timestamp": 1619661337558,
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "monitor": {
        "pages": {
            "page_0": {
                "id": "page_0",
                "mode": "FIT_WINDOW",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_0",
                    "MEDIUM": "layout_51",
                    "SMALL": "layout_73",
                },
                "isDefault": True,
                "isVisible": True,
                "header": True,
                "bodyBackgroundColor": "var(--light)",
            }
        },
        "layouts": {
            "layout_0": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.7726%",
                            "right": "78.3963%",
                            "top": "1.7078%",
                            "bottom": "1.8112%",
                            "width": "20.8312%",
                            "height": "96.4811%",
                        },
                        "widgetId": "widget_126",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_0",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "22.4881%",
                            "right": "0.8757%",
                            "top": "1.7078%",
                            "bottom": "34.1417%",
                            "width": "76.6363%",
                            "height": "64.1505%",
                        },
                        "widgetId": "widget_146",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_0",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "22.4881%",
                            "right": "0.8757%",
                            "top": "67.0716%",
                            "bottom": "1.8176%",
                            "width": "76.6363%",
                            "height": "31.1108%",
                        },
                        "widgetId": "widget_165",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_0",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "pages", "id": "page_0"},
            },
            "layout_24": {
                "type": "FIXED",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2.00%",
                            "right": "94.5116%",
                            "top": "50%",
                            "bottom": "16.6667%",
                            "width": "40.00px",
                            "height": "40px",
                        },
                        "widgetId": "widget_70",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                            "autoProps": {"top": False, "bottom": True},
                        },
                        "parent": "layout_24",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.9983%",
                            "right": "18.9598%",
                            "top": "50%",
                            "bottom": "9.0365%",
                            "width": "75.0419%",
                            "height": "81.9271%",
                        },
                        "widgetId": "widget_71",
                        "setting": {
                            "vCenter": True,
                            "autoProps": {"top": False, "bottom": True},
                        },
                        "parent": "layout_24",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "90.6327%",
                            "right": "2.00%",
                            "top": "16.6667%",
                            "bottom": "16.6667%",
                            "width": "117px",
                            "height": "40px",
                        },
                        "widgetId": "widget_125",
                        "setting": {
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "top": False,
                                "bottom": True,
                                "right": False,
                                "left": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_24",
                    },
                },
                "parent": {"type": "header", "id": "header"},
            },
            "layout_25": {
                "type": "FIXED",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "4.00%",
                            "top": "5.00%",
                            "width": "92.00%",
                            "height": 30,
                            "right": 340,
                            "bottom": "86.4774%",
                        },
                        "widgetId": "widget_75",
                        "isPending": False,
                        "parent": "layout_25",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                                "height": "STRETCH",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": False,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "3.9968%",
                            "top": "14.0000%",
                            "width": "91.9988%",
                            "height": "8.0000%",
                            "right": "17px",
                            "bottom": "468px",
                        },
                        "widgetId": "widget_76",
                        "isPending": False,
                        "parent": "layout_25",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "4.0000%",
                            "right": "21.0000%",
                            "top": "24.00%",
                            "bottom": "8.6667%",
                            "width": "92.00%",
                            "height": "71.00%",
                        },
                        "widgetId": "widget_77",
                        "parent": "layout_25",
                    },
                },
                "order": ["1", "2", "3"],
                "setting": {"lockDescendants": True},
                "parent": {"type": "dialog", "id": "dialog_1"},
            },
            "layout_27": {
                "type": "COLUMN",
                "order": ["0", "3", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": 307},
                        "widgetId": "widget_127",
                        "setting": {
                            "autoProps": {"width": "STRETCH", "height": "STRETCH"}
                        },
                        "parent": "layout_27",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "220.00px"},
                        "widgetId": "widget_129",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_27",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_199",
                    },
                },
                "parent": {"type": "widget", "id": "widget_126"},
            },
            "layout_31": {
                "type": "FIXED",
                "content": {
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_141",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_31",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "4.9974%",
                            "right": "5.0026%",
                            "top": "68.1786%",
                            "bottom": "3.5792%",
                            "width": "90.0000%",
                            "height": "28.2421%",
                        },
                        "widgetId": "widget_142",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_31",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_143",
                        "setting": {},
                        "parent": "layout_31",
                    },
                },
                "order": ["4", "5", "6"],
                "parent": {"type": "widget", "id": "widget_129"},
            },
            "layout_32": {
                "type": "FIXED",
                "content": {},
                "order": [],
                "parent": {"type": "widget", "id": "widget_129"},
            },
            "layout_33": {
                "type": "ROW",
                "order": ["2", "1"],
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_148",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_33",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 8, "height": "400px"},
                        "widgetId": "widget_198",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_33",
                    },
                },
                "parent": {"type": "widget", "id": "widget_146"},
            },
            "layout_35": {
                "type": "COLUMN",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "361px", "height": 314},
                        "widgetId": "widget_155",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "isPending": False,
                        "parent": "layout_35",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "355px", "height": 198},
                        "widgetId": "widget_156",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_35",
                    },
                },
                "setting": {"space": 10},
                "parent": {"type": "widget", "id": "widget_148"},
            },
            "layout_40": {
                "type": "COLUMN",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {"width": "355px", "height": 299},
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "parent": "layout_40",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "33.4831%", "height": 26},
                        "widgetId": "widget_157",
                        "setting": {"autoProps": {"height": "AUTO", "width": "CUSTOM"}},
                        "parent": "layout_40",
                    },
                },
                "setting": {"space": 0},
                "parent": {"type": "widget", "id": "widget_155"},
            },
            "layout_41": {
                "type": "COLUMN",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "353px", "height": 33},
                        "widgetId": "widget_158",
                        "parent": "layout_41",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "353px", "height": 76},
                        "widgetId": "widget_159",
                        "parent": "layout_41",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "353px", "height": 43},
                        "widgetId": "widget_160",
                        "parent": "layout_41",
                    },
                },
                "setting": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                "parent": {"type": "widget", "id": "widget_156"},
            },
            "layout_42": {
                "type": "FIXED",
                "order": ["0", "1", "2", "3"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "41.6295%",
                            "top": "6.00%",
                            "bottom": "68.6152%",
                            "width": "40.00px",
                            "height": "40px",
                        },
                        "widgetId": "widget_161",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "parent": "layout_42",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0.0522%",
                            "top": "28.00%",
                            "bottom": "40.5898%",
                            "width": "99.9478%",
                            "height": "26.7764%",
                        },
                        "widgetId": "widget_162",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_42",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "6.0251%",
                            "top": "54.00%",
                            "bottom": "18.844699999999996%",
                            "width": "88.00%",
                            "height": "24.2215%",
                        },
                        "widgetId": "widget_163",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_42",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "19.9983%",
                            "right": "20.0026%",
                            "top": "80.00%",
                            "bottom": "4.8497%",
                            "width": "60.00%",
                            "height": "14.00%",
                        },
                        "widgetId": "widget_164",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"left": False, "right": True},
                            "vCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_42",
                    },
                },
                "parent": {"type": "view", "id": "view_2"},
            },
            "layout_43": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 0, "width": 5, "height": "307px"},
                        "widgetId": "widget_166",
                        "parent": "layout_43",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 7, "height": "400px"},
                        "widgetId": "widget_167",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_43",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_165"},
            },
            "layout_48": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "41.6295%",
                            "top": "6.00%",
                            "bottom": "68.6152%",
                            "width": "40.00px",
                            "height": "40px",
                        },
                        "widgetId": "widget_173",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "parent": "layout_48",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0.0522%",
                            "top": "28.00%",
                            "bottom": "40.5898%",
                            "width": "99.9478%",
                            "height": "26.7764%",
                        },
                        "widgetId": "widget_174",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_48",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "6.0251%",
                            "top": "54.00%",
                            "bottom": "18.844699999999996%",
                            "width": "88.00%",
                            "height": "24.2215%",
                        },
                        "widgetId": "widget_175",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_48",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "19.9983%",
                            "right": "20.0026%",
                            "top": "80.00%",
                            "bottom": "4.8497%",
                            "width": "60.00%",
                            "height": "14.00%",
                        },
                        "widgetId": "widget_176",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"left": False, "right": True},
                            "vCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_48",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "view", "id": "view_4"},
            },
            "layout_49": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "41.6295%",
                            "top": "6.00%",
                            "bottom": "68.6152%",
                            "width": "40.00px",
                            "height": "40px",
                        },
                        "widgetId": "widget_177",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "parent": "layout_49",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0.0522%",
                            "top": "28.00%",
                            "bottom": "40.5898%",
                            "width": "99.9478%",
                            "height": "26.7764%",
                        },
                        "widgetId": "widget_178",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_49",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "6.0251%",
                            "top": "54.00%",
                            "bottom": "18.844699999999996%",
                            "width": "88.00%",
                            "height": "24.2215%",
                        },
                        "widgetId": "widget_179",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_49",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "19.9983%",
                            "right": "20.0026%",
                            "top": "80.00%",
                            "bottom": "4.8497%",
                            "width": "60.00%",
                            "height": "14.00%",
                        },
                        "widgetId": "widget_180",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"left": False, "right": True},
                            "vCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_49",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "view", "id": "view_5"},
            },
            "layout_50": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2.00%",
                            "right": "94.5116%",
                            "top": "50%",
                            "bottom": "16.6667%",
                            "width": "40.00px",
                            "height": "40px",
                        },
                        "widgetId": "widget_70",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                            "autoProps": {"top": False, "bottom": True},
                        },
                        "parent": "layout_50",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "11.0000%",
                            "right": "22.8008%",
                            "top": "50%",
                            "bottom": "9.0365%",
                            "width": "66.1992%",
                            "height": "74.2969%",
                        },
                        "widgetId": "widget_71",
                        "setting": {
                            "vCenter": True,
                            "autoProps": {"top": False, "bottom": True},
                        },
                        "parent": "layout_50",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "90.6327%",
                            "right": "2.00%",
                            "top": "16.6667%",
                            "bottom": "16.6667%",
                            "width": "117px",
                            "height": "40px",
                        },
                        "widgetId": "widget_125",
                        "setting": {
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "top": False,
                                "bottom": True,
                                "right": False,
                                "left": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_50",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "header", "id": "header"},
            },
            "layout_51": {
                "type": "FIXED",
                "content": {
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0%",
                            "right": "50.1%",
                            "top": "0.0%",
                            "bottom": "33.4%",
                            "width": "100.00%",
                            "height": "100.00%",
                        },
                        "widgetId": "widget_200",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            }
                        },
                    }
                },
                "order": ["3"],
                "parent": {"type": "pages", "id": "page_0"},
            },
            "layout_52": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": 307},
                        "widgetId": "widget_127",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "parent": "layout_52",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "125px", "height": "240.00px"},
                        "widgetId": "widget_129",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_52",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "125px", "height": "160.00px"},
                        "widgetId": "widget_199",
                    },
                },
                "order": ["0", "3", "2"],
                "parent": {"type": "widget", "id": "widget_184"},
            },
            "layout_53": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_141",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_53",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "4.9974%",
                            "right": "5.0026%",
                            "top": "68.1786%",
                            "bottom": "3.5792%",
                            "width": "90.0000%",
                            "height": "28.2421%",
                        },
                        "widgetId": "widget_142",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_53",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_143",
                        "setting": {},
                        "parent": "layout_53",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_129"},
            },
            "layout_54": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_129"},
            },
            "layout_55": {
                "type": "ROW",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "600px"},
                        "widgetId": "widget_187",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_55",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "400px"},
                        "widgetId": "widget_198",
                        "setting": {"heightMode": "fit"},
                    },
                },
                "order": ["2", "1"],
                "parent": {"type": "widget", "id": "widget_185"},
            },
            "layout_65": {
                "type": "COLUMN",
                "setting": {"space": 10},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "361px", "height": 314},
                        "widgetId": "widget_188",
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "isPending": False,
                        "parent": "layout_65",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "355px", "height": "180.00px"},
                        "widgetId": "widget_189",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_65",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_187"},
            },
            "layout_66": {
                "type": "COLUMN",
                "setting": {"space": 0},
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {"width": "355px", "height": 299},
                        "setting": {"autoProps": {"height": "STRETCH"}},
                        "parent": "layout_66",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "50.00%", "height": "60px"},
                        "widgetId": "widget_157",
                        "setting": {"autoProps": {"height": "AUTO", "width": "CUSTOM"}},
                        "parent": "layout_66",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_188"},
            },
            "layout_67": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "41.6295%",
                            "top": "6.00%",
                            "bottom": "68.6152%",
                            "width": "60.00px",
                            "height": "60.00px",
                        },
                        "widgetId": "widget_161",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_67",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0.0522%",
                            "top": "28.00%",
                            "bottom": "40.5898%",
                            "width": "99.9478%",
                            "height": "26.7764%",
                        },
                        "widgetId": "widget_162",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_67",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "6.0251%",
                            "top": "54.00%",
                            "bottom": "18.844699999999996%",
                            "width": "88.00%",
                            "height": "24.2215%",
                        },
                        "widgetId": "widget_163",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_67",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "19.9983%",
                            "right": "20.0026%",
                            "top": "80.00%",
                            "bottom": "4.8497%",
                            "width": "60.00%",
                            "height": "14.00%",
                        },
                        "widgetId": "widget_164",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"left": False, "right": True},
                            "vCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_67",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "view", "id": "view_2"},
            },
            "layout_68": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "41.6295%",
                            "top": "6.00%",
                            "bottom": "68.6152%",
                            "width": "60.00px",
                            "height": "60.00px",
                        },
                        "widgetId": "widget_173",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_68",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0.0522%",
                            "top": "28.00%",
                            "bottom": "40.5898%",
                            "width": "99.9478%",
                            "height": "26.7764%",
                        },
                        "widgetId": "widget_174",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_68",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "6.0251%",
                            "top": "54.00%",
                            "bottom": "18.844699999999996%",
                            "width": "88.00%",
                            "height": "24.2215%",
                        },
                        "widgetId": "widget_175",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_68",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "19.9983%",
                            "right": "20.0026%",
                            "top": "80.00%",
                            "bottom": "4.8497%",
                            "width": "60.00%",
                            "height": "14.00%",
                        },
                        "widgetId": "widget_176",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"left": False, "right": True},
                            "vCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_68",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "view", "id": "view_4"},
            },
            "layout_69": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "41.6295%",
                            "top": "6.00%",
                            "bottom": "68.6152%",
                            "width": "60.00px",
                            "height": "60.00px",
                        },
                        "widgetId": "widget_177",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_69",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0.0522%",
                            "top": "28.00%",
                            "bottom": "40.5898%",
                            "width": "99.9478%",
                            "height": "26.7764%",
                        },
                        "widgetId": "widget_178",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_69",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "6.0251%",
                            "top": "54.00%",
                            "bottom": "18.844699999999996%",
                            "width": "88.00%",
                            "height": "24.2215%",
                        },
                        "widgetId": "widget_179",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_69",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "19.9983%",
                            "right": "20.0026%",
                            "top": "80.00%",
                            "bottom": "4.8497%",
                            "width": "60.00%",
                            "height": "14.00%",
                        },
                        "widgetId": "widget_180",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"left": False, "right": True},
                            "vCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_69",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "view", "id": "view_5"},
            },
            "layout_70": {
                "type": "COLUMN",
                "setting": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "353px", "height": 33},
                        "widgetId": "widget_158",
                        "parent": "layout_70",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "353px", "height": 76},
                        "widgetId": "widget_159",
                        "parent": "layout_70",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "353px", "height": "30.00px"},
                        "widgetId": "widget_160",
                        "parent": "layout_70",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_189"},
            },
            "layout_71": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {"style": {}, "heightMode": "fit"},
                        "bbox": {"left": 0, "width": 5, "height": "307px"},
                        "widgetId": "widget_166",
                        "parent": "layout_71",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 7, "height": "400px"},
                        "widgetId": "widget_167",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_71",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_190"},
            },
            "layout_72": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "2.00%",
                            "right": "94.5116%",
                            "top": "50%",
                            "bottom": "16.6667%",
                            "width": "40.00px",
                            "height": "40px",
                        },
                        "widgetId": "widget_70",
                        "isPending": False,
                        "setting": {
                            "vCenter": True,
                            "hCenter": False,
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                            "autoProps": {"top": False, "bottom": True},
                        },
                        "parent": "layout_72",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "16.2999%",
                            "right": "37.6888%",
                            "top": "16.6667%",
                            "bottom": "9.0625%",
                            "width": "46.0113%",
                            "height": "74.2708%",
                        },
                        "widgetId": "widget_71",
                        "setting": {
                            "vCenter": True,
                            "autoProps": {"top": False, "bottom": True},
                        },
                        "parent": "layout_72",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "90.6327%",
                            "right": "2.00%",
                            "top": "16.6667%",
                            "bottom": "16.6667%",
                            "width": "117px",
                            "height": "40px",
                        },
                        "widgetId": "widget_125",
                        "setting": {
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "top": False,
                                "bottom": True,
                                "right": False,
                                "left": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_72",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "header", "id": "header"},
            },
            "layout_73": {
                "type": "FIXED",
                "content": {
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0%",
                            "right": "16.7%",
                            "top": "0.0%",
                            "bottom": "0.0%",
                            "width": "100.00%",
                            "height": "100.0%",
                        },
                        "widgetId": "widget_201",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": True,
                                "bottom": False,
                            }
                        },
                    }
                },
                "order": ["3"],
                "parent": {"type": "pages", "id": "page_0"},
            },
            "layout_75": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "right": "5%",
                            "top": "51.9593%",
                            "bottom": 127,
                            "width": "90%",
                            "height": "10.1190%",
                        },
                        "widgetId": "widget_141",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"height": "AUTO"},
                        },
                        "isPending": False,
                        "parent": "layout_75",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "4.9918%",
                            "right": "5.0152%",
                            "top": "76.7624%",
                            "bottom": "3.5853%",
                            "width": "89.9930%",
                            "height": "19.6524%",
                        },
                        "widgetId": "widget_142",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_75",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.0000%",
                            "right": 0,
                            "top": "0.0000%",
                            "bottom": "36.2647%",
                            "width": "100.0000%",
                            "height": "47.53%",
                        },
                        "widgetId": "widget_143",
                        "setting": {},
                        "parent": "layout_75",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_129"},
            },
            "layout_76": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_129"},
            },
            "layout_89": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "41.6295%",
                            "top": "6.00%",
                            "bottom": "68.6152%",
                            "width": "60.00px",
                            "height": "60.00px",
                        },
                        "widgetId": "widget_161",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_89",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0.0522%",
                            "top": "28.00%",
                            "bottom": "40.5898%",
                            "width": "99.9478%",
                            "height": "26.7764%",
                        },
                        "widgetId": "widget_162",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_89",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "6.0251%",
                            "top": "52.00%",
                            "bottom": "18.844699999999996%",
                            "width": "88.00%",
                            "height": "24.2215%",
                        },
                        "widgetId": "widget_163",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_89",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "19.9983%",
                            "right": "20.0026%",
                            "top": "80.00%",
                            "bottom": "4.8497%",
                            "width": "60.00%",
                            "height": "14.00%",
                        },
                        "widgetId": "widget_164",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"left": False, "right": True},
                            "vCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_89",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "view", "id": "view_2"},
            },
            "layout_90": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "41.6295%",
                            "top": "6.00%",
                            "bottom": "68.6152%",
                            "width": "60.00px",
                            "height": "60.00px",
                        },
                        "widgetId": "widget_173",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_90",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0.0522%",
                            "top": "28.00%",
                            "bottom": "40.5898%",
                            "width": "99.9478%",
                            "height": "26.7764%",
                        },
                        "widgetId": "widget_174",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_90",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "6.0251%",
                            "top": "52.00%",
                            "bottom": "18.844699999999996%",
                            "width": "88.00%",
                            "height": "24.2215%",
                        },
                        "widgetId": "widget_175",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_90",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "19.9983%",
                            "right": "20.0026%",
                            "top": "80.00%",
                            "bottom": "4.8497%",
                            "width": "60.00%",
                            "height": "14.00%",
                        },
                        "widgetId": "widget_176",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"left": False, "right": True},
                            "vCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_90",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "view", "id": "view_4"},
            },
            "layout_91": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "41.6295%",
                            "top": "6.00%",
                            "bottom": "68.6152%",
                            "width": "60.00px",
                            "height": "60.00px",
                        },
                        "widgetId": "widget_177",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_91",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0.0522%",
                            "top": "28.00%",
                            "bottom": "40.5898%",
                            "width": "99.9478%",
                            "height": "26.7764%",
                        },
                        "widgetId": "widget_178",
                        "isPending": False,
                        "setting": {"vCenter": False, "hCenter": False},
                        "parent": "layout_91",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "6.0251%",
                            "top": "52.00%",
                            "bottom": "18.844699999999996%",
                            "width": "88.00%",
                            "height": "24.2215%",
                        },
                        "widgetId": "widget_179",
                        "isPending": False,
                        "setting": {
                            "vCenter": False,
                            "hCenter": True,
                            "autoProps": {"left": False, "right": True},
                        },
                        "parent": "layout_91",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "19.9983%",
                            "right": "20.0026%",
                            "top": "80.00%",
                            "bottom": "4.8497%",
                            "width": "60.00%",
                            "height": "14.00%",
                        },
                        "widgetId": "widget_180",
                        "setting": {
                            "hCenter": False,
                            "autoProps": {"left": False, "right": True},
                            "vCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_91",
                    },
                },
                "order": ["0", "1", "2", "3"],
                "parent": {"type": "view", "id": "view_5"},
            },
            "layout_94": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_198"},
            },
            "layout_95": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_198"},
            },
            "layout_96": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_198"},
            },
            "layout_97": {
                "type": "COLUMN",
                "setting": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                "order": ["1", "0", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "601px", "height": "400px"},
                        "widgetId": "widget_190",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "601px", "height": "420.00px"},
                        "widgetId": "widget_185",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "800.00px"},
                        "widgetId": "widget_184",
                    },
                },
            },
            "layout_98": {
                "type": "COLUMN",
                "order": ["0", "9", "1", "2", "5", "6", "7", "4", "8"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "400px"},
                        "widgetId": "widget_198",
                    },
                    "1": {
                        "type": "SECTION",
                        "bbox": {"width": "200px", "height": "300.00px"},
                        "sectionId": "section_2",
                        "isPending": False,
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "355px", "height": 38},
                        "widgetId": "widget_157",
                        "isPending": False,
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "360.00px"},
                        "widgetId": "widget_129",
                        "isPending": False,
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "300px"},
                        "widgetId": "widget_166",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "200.00px"},
                        "widgetId": "widget_202",
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {"width": "350px", "height": "250.00px"},
                        "widgetId": "widget_199",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "277.00px"},
                        "widgetId": "widget_167",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "300px"},
                        "widgetId": "widget_127",
                    },
                },
            },
            "layout_99": {
                "type": "COLUMN",
                "order": ["0", "1", "2"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "355px", "height": "35.00px"},
                        "widgetId": "widget_158",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "355px", "height": "70px"},
                        "widgetId": "widget_159",
                        "setting": {"autoProps": {"height": "AUTO"}},
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "355px", "height": "45.00px"},
                        "widgetId": "widget_160",
                    },
                },
                "setting": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "monitor",
        "widgets": {
            "widget_70": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                        "imageFillMode": "FIT",
                    },
                    "styleConfig": {},
                },
                "id": "widget_70",
                "parent": {
                    "LARGE": [{"layoutId": "layout_24", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_50", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_72", "layoutItemId": "0"}],
                },
            },
            "widget_71": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 28px; color: var(--light);">Title Lorem ipsum dolor sit amet, vix ne solum deleniti euripidis </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_71",
                "parent": {
                    "LARGE": [{"layoutId": "layout_24", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_50", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_72", "layoutItemId": "1"}],
                },
            },
            "widget_75": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><strong style="color: var(--dark-800); font-size: 24px;">\ufeffTABLE</strong></p>',
                    "placeholder": '<p><strong style="color: var(--dark-800); font-size: 24px;">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_75",
                "parent": {"LARGE": [{"layoutId": "layout_25", "layoutItemId": "1"}]},
            },
            "widget_76": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800); font-size: 16px;">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_76",
                "parent": {"LARGE": [{"layoutId": "layout_25", "layoutItemId": "2"}]},
            },
            "widget_77": {
                "uri": "widgets/common/table/",
                "version": "1.5.0",
                "config": {"layersConfig": [], "arrangeType": "TABS"},
                "id": "widget_77",
                "parent": {"LARGE": [{"layoutId": "layout_25", "layoutItemId": "3"}]},
            },
            "widget_125": {
                "uri": "widgets/common/share/",
                "version": "1.5.0",
                "config": {
                    "uiMode": "INLINE",
                    "popup": {
                        "icon": "",
                        "items": ["email", "facebook", "twitter", "linkedin"],
                        "tooltip": "",
                    },
                    "inline": {
                        "items": ["facebook", "twitter", "email"],
                        "design": {
                            "direction": "HORIZONTAL",
                            "hideLabel": True,
                            "btnRad": 0,
                            "btnColor": "",
                            "iconColor": "",
                            "size": "default",
                        },
                    },
                },
                "id": "widget_125",
                "parent": {
                    "LARGE": [{"layoutId": "layout_24", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_50", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_72", "layoutItemId": "2"}],
                },
            },
            "widget_126": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_126",
                "layouts": {"DEFAULT": {"LARGE": "layout_27"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "0"}]},
            },
            "widget_127": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_127",
                "style": {
                    "border": {"type": "solid", "color": "var(--dark)", "width": "1px"},
                    "background": {
                        "color": "",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_27", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_52", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_74", "layoutItemId": "0"}],
                },
            },
            "widget_129": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0px",
                                "offsetY": "0px",
                                "blur": "0px",
                                "spread": "0px",
                                "color": "",
                            },
                            "background": {
                                "color": "",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "itemStyle": "STYLE0",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isOpenAdvabceSetting": False,
                    "isInitialed": True,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_129",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_31",
                        "MEDIUM": "layout_53",
                        "SMALL": "layout_75",
                    },
                    "HOVER": {
                        "LARGE": "layout_32",
                        "MEDIUM": "layout_54",
                        "SMALL": "layout_76",
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_27", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_52", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_74", "layoutItemId": "2"}],
                },
                "style": {
                    "border": {"type": "solid", "color": "var(--dark)", "width": "1px"}
                },
            },
            "widget_141": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="color: var(--dark-800); font-size: 20px;">\ufeffCard</span></p>',
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_141",
                "parent": {
                    "LARGE": [{"layoutId": "layout_31", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_53", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_75", "layoutItemId": "0"}],
                },
            },
            "widget_142": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p style="line-height: 1.5;"><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_142",
                "parent": {
                    "LARGE": [{"layoutId": "layout_31", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_53", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_75", "layoutItemId": "1"}],
                },
            },
            "widget_143": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_143",
                "parent": {
                    "LARGE": [{"layoutId": "layout_31", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_53", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_75", "layoutItemId": "2"}],
                },
            },
            "widget_146": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_146",
                "layouts": {"DEFAULT": {"LARGE": "layout_33"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "1"}]},
            },
            "widget_148": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 10},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_148",
                "layouts": {"DEFAULT": {"LARGE": "layout_35"}},
                "parent": {"LARGE": [{"layoutId": "layout_33", "layoutItemId": "1"}]},
            },
            "widget_155": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_155",
                "layouts": {"DEFAULT": {"LARGE": "layout_40"}},
                "parent": {"LARGE": [{"layoutId": "layout_35", "layoutItemId": "0"}]},
            },
            "widget_156": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [20, 0, 20, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_156",
                "layouts": {"DEFAULT": {"LARGE": "layout_41"}},
                "style": {
                    "background": {
                        "color": "",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {"type": "solid", "color": "var(--dark)", "width": "1px"},
                },
                "parent": {"LARGE": [{"layoutId": "layout_35", "layoutItemId": "1"}]},
            },
            "widget_157": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_2",
                        "views": ["view_2", "view_4", "view_5"],
                    },
                    "display": {
                        "type": "navButtonGroup",
                        "navStyle": "default",
                        "standard": {
                            "showPageNumber": True,
                            "previousText": "",
                            "previousIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="M2.479 5.47L8.019.162a.592.592 0 01.813 0 .534.534 0 010 .779L3.552 6l5.28 5.06a.534.534 0 010 .779.592.592 0 01-.813 0l-5.54-5.31a.726.726 0 010-1.058z" fill="#000"></path><path d="M0 0h12v12H0z"></path></g></svg>',
                                "properties": {"filename": "left-arrow-12", "size": 12},
                            },
                            "nextText": "",
                            "nextIcon": {
                                "svg": '<svg viewBox="0 0 12 12" xmlns="http://www.w3.org/2000/svg"><g fill="none" fill-rule="evenodd"><path d="M9.521 5.47L3.981.162a.592.592 0 00-.813 0 .534.534 0 000 .779L8.448 6l-5.28 5.06a.534.534 0 000 .779.592.592 0 00.813 0l5.54-5.31a.726.726 0 000-1.058z" fill="#000"></path><path d="M0 0h12v12H0z"></path></g></svg>',
                                "properties": {
                                    "filename": "right-arrow-12",
                                    "size": 12,
                                },
                            },
                        },
                        "vertical": False,
                        "advanced": False,
                        "variant": None,
                    },
                },
                "id": "widget_157",
                "parent": {
                    "LARGE": [{"layoutId": "layout_40", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_66", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_88", "layoutItemId": "1"}],
                },
            },
            "widget_158": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="font-size: 18px; color: var(--dark-100);">Eos id amet inani </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_158",
                "parent": {
                    "LARGE": [{"layoutId": "layout_41", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_70", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_92", "layoutItemId": "0"}],
                },
                "useDataSourcesEnabled": False,
            },
            "widget_159": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 42px; color: var(--primary);">Statistics</strong></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_159",
                "parent": {
                    "LARGE": [{"layoutId": "layout_41", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_70", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_92", "layoutItemId": "1"}],
                },
            },
            "widget_160": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="font-size: 18px;">Lorem ipsum dolor sit amet </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_160",
                "parent": {
                    "LARGE": [{"layoutId": "layout_41", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_70", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_92", "layoutItemId": "2"}],
                },
            },
            "widget_161": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                        "imageFillMode": "FIT",
                    },
                    "styleConfig": {},
                },
                "id": "widget_161",
                "parent": {
                    "LARGE": [{"layoutId": "layout_42", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_67", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_89", "layoutItemId": "0"}],
                },
            },
            "widget_162": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="font-size: 42px; color: var(--primary);">Number 1</strong></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_162",
                "parent": {
                    "LARGE": [{"layoutId": "layout_42", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_67", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_89", "layoutItemId": "1"}],
                },
            },
            "widget_163": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 14px;">Lorem ipsum dolor sit amet, vix ne solum dniti euripidis. Ut has facer indoctum adversarium, vis ei epicurei. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_163",
                "parent": {
                    "LARGE": [{"layoutId": "layout_42", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_67", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_89", "layoutItemId": "2"}],
                },
            },
            "widget_164": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "Button",
                        "textExpression": None,
                        "linkParam": {"linkType": "DIALOG", "value": "dialog_1"},
                    },
                    "styleConfig": {
                        "themeStyle": {"quickStyleType": "secondary"},
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#076fe5", "size": 14}},
                            "hover": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                    },
                },
                "id": "widget_164",
                "useDataSources": [],
                "parent": {
                    "LARGE": [{"layoutId": "layout_42", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_67", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_89", "layoutItemId": "3"}],
                },
            },
            "widget_165": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_165",
                "layouts": {"DEFAULT": {"LARGE": "layout_43"}},
                "parent": {"LARGE": [{"layoutId": "layout_0", "layoutItemId": "2"}]},
            },
            "widget_166": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_166",
                "style": {
                    "border": {"type": "solid", "color": "var(--dark)", "width": "1px"},
                    "background": {
                        "color": "",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_43", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_71", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_93", "layoutItemId": "0"}],
                },
            },
            "widget_167": {
                "uri": "widgets/common/table/",
                "version": "1.5.0",
                "config": {"layersConfig": [], "arrangeType": "TABS"},
                "id": "widget_167",
                "style": {
                    "border": {"type": "solid", "color": "var(--dark)", "width": "1px"}
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_43", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_71", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_93", "layoutItemId": "1"}],
                },
            },
            "widget_173": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                        "imageFillMode": "FIT",
                    },
                    "styleConfig": {},
                },
                "id": "widget_173",
                "parent": {
                    "LARGE": [{"layoutId": "layout_48", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_68", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_90", "layoutItemId": "0"}],
                },
            },
            "widget_174": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="color: var(--primary); font-size: 42px;">Number 2</strong></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_174",
                "parent": {
                    "LARGE": [{"layoutId": "layout_48", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_68", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_90", "layoutItemId": "1"}],
                },
            },
            "widget_175": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 14px;">Eos id amet inani. Iudicabit conceptam at nec, quo magna dicam dictas eu.</span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_175",
                "parent": {
                    "LARGE": [{"layoutId": "layout_48", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_68", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_90", "layoutItemId": "2"}],
                },
            },
            "widget_176": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {"text": "Button", "textExpression": None},
                    "styleConfig": {
                        "themeStyle": {"quickStyleType": "secondary"},
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#076fe5", "size": 14}},
                            "hover": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                    },
                },
                "id": "widget_176",
                "useDataSources": [],
                "parent": {
                    "LARGE": [{"layoutId": "layout_48", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_68", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_90", "layoutItemId": "3"}],
                },
            },
            "widget_177": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                        "imageFillMode": "FIT",
                    },
                    "styleConfig": {},
                },
                "id": "widget_177",
                "parent": {
                    "LARGE": [{"layoutId": "layout_49", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_69", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_91", "layoutItemId": "0"}],
                },
            },
            "widget_178": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><strong style="color: var(--primary); font-size: 42px;">Number 3</strong></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_178",
                "parent": {
                    "LARGE": [{"layoutId": "layout_49", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_69", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_91", "layoutItemId": "1"}],
                },
            },
            "widget_179": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p><span style="font-size: 14px;">Ut has facer indoctum adversarium, vis ei epicurei. Lorem ipsum dolor sit amet, vix ne solum </span>ad quo torquatos&nbsp; <span style="font-size: 14px;">dniti euripidis. </span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_179",
                "parent": {
                    "LARGE": [{"layoutId": "layout_49", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_69", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_91", "layoutItemId": "2"}],
                },
            },
            "widget_180": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {"text": "Button", "textExpression": None},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#076fe5", "size": 14}},
                            "hover": {"iconProps": {"color": "#fff", "size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "secondary"},
                    },
                },
                "id": "widget_180",
                "useDataSources": [],
                "parent": {
                    "LARGE": [{"layoutId": "layout_49", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_69", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_91", "layoutItemId": "3"}],
                },
            },
            "widget_184": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_184",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_52"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_51", "layoutItemId": "0"}]},
            },
            "widget_185": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_185",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_55"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_51", "layoutItemId": "1"}]},
            },
            "widget_187": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 10},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_187",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_65"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_55", "layoutItemId": "1"}]},
            },
            "widget_188": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_188",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_66"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_65", "layoutItemId": "0"}]},
            },
            "widget_189": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 0, 10, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_189",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_70"}},
                "style": {
                    "background": {
                        "color": "#ffffff",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--light-400)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_65", "layoutItemId": "1"}]},
            },
            "widget_190": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 0, 0], "unit": "px"}},
                },
                "id": "widget_190",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_71"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_51", "layoutItemId": "2"}]},
            },
            "widget_198": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": False,
                },
                "id": "widget_198",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_94",
                        "MEDIUM": "layout_95",
                        "SMALL": "layout_96",
                    }
                },
                "parent": {"LARGE": [{"layoutId": "layout_33", "layoutItemId": "2"}]},
                "style": {
                    "border": {"type": "solid", "color": "var(--dark)", "width": "1px"}
                },
            },
            "widget_199": {
                "uri": "widgets/arcgis/query/",
                "version": "1.5.0",
                "config": {
                    "queryItems": [],
                    "arrangeType": "BLOCK",
                    "arrangeWrap": False,
                    "sizeMap": {
                        "arrangementIconPopper": {
                            "minSize": {"width": 300, "height": 300},
                            "defaultSize": {"width": 360, "height": 480},
                        }
                    },
                },
                "id": "widget_199",
                "style": {
                    "border": {"type": "solid", "color": "var(--dark)", "width": "1px"}
                },
            },
            "widget_200": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [10, 10, 10, 10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_200",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_97"}},
            },
            "widget_201": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_201",
                "layouts": {"DEFAULT": {"SMALL": "layout_98"}},
            },
            "widget_202": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"number": [20, 0, 20, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_202",
                "layouts": {"DEFAULT": {"SMALL": "layout_99"}},
            },
        },
        "widgetsManifest": {},
        "views": {
            "view_2": {
                "id": "view_2",
                "layout": {
                    "LARGE": "layout_42",
                    "MEDIUM": "layout_67",
                    "SMALL": "layout_89",
                },
                "backgroundPosition": "fit",
                "backgroundColor": None,
                "backgroundIMage": {},
            },
            "view_4": {
                "id": "view_4",
                "layout": {
                    "LARGE": "layout_48",
                    "MEDIUM": "layout_68",
                    "SMALL": "layout_90",
                },
                "backgroundPosition": "fit",
                "backgroundColor": None,
                "backgroundIMage": {},
            },
            "view_5": {
                "id": "view_5",
                "layout": {
                    "LARGE": "layout_49",
                    "MEDIUM": "layout_69",
                    "SMALL": "layout_91",
                },
                "backgroundPosition": "fit",
                "backgroundColor": None,
                "backgroundIMage": {},
            },
        },
        "sections": {
            "section_2": {
                "id": "section_2",
                "views": ["view_2", "view_4", "view_5"],
                "style": {
                    "border": {"type": "solid", "color": "var(--dark)", "width": "1px"}
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_40", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_66", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_88", "layoutItemId": "0"}],
                },
            }
        },
        "dialogs": {
            "dialog_1": {
                "id": "dialog_1",
                "mode": "FIXED",
                "sizeMode": {
                    "LARGE": {"position": "CENTER", "width": "800px", "height": "600px"}
                },
                "interactionType": None,
                "interactionStyles": None,
                "confirmBeforeClose": True,
                "alwaysShowConfirmation": False,
                "dialogBackground": {
                    "color": "var(--light)",
                    "fillType": "fit",
                    "image": {"url": ""},
                },
                "overlayColor": "var(--dark-900)",
                "layout": {"LARGE": "layout_25"},
                "index": 0,
                "isSplash": False,
                "closeWhenClickOverlay": True,
            }
        },
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_0": []}],
        "timestamp": 1619661155928,
        "header": {
            "layout": {
                "LARGE": "layout_24",
                "MEDIUM": "layout_50",
                "SMALL": "layout_72",
            },
            "height": {"LARGE": "60px"},
            "backgroundColor": "var(--primary-900)",
        },
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
    "reveal": {
        "pages": {
            "page_1": {
                "id": "page_1",
                "mode": "AUTO_SCROLL",
                "type": "NORMAL",
                "layout": {
                    "LARGE": "layout_3",
                    "MEDIUM": "layout_144",
                    "SMALL": "layout_242",
                },
                "isDefault": True,
                "isVisible": True,
                "header": True,
                "footer": True,
                "icon": {
                    "svg": '<svg viewBox="0 0 10 12" xmlns="http://www.w3.org/2000/svg"><path d="M1 0h8a1 1 0 011 1v10a1 1 0 01-1 1H1a1 1 0 01-1-1V1a1 1 0 011-1zm0 1v10h8V1H1zm2 2h4a.5.5 0 010 1H3a.5.5 0 010-1zm0 2.5h4a.5.5 0 010 1H3a.5.5 0 010-1z" fill="#000" fill-rule="nonzero"></path></svg>',
                    "properties": {
                        "filename": None,
                        "originalName": None,
                        "path": None,
                        "color": "#fff",
                        "size": 16,
                        "inlineSvg": True,
                    },
                },
                "bodyBackgroundColor": "var(--light-200)",
            }
        },
        "layouts": {
            "layout_1": {
                "type": "FIXED",
                "order": ["0", "1", "2", "3", "4"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50.00px",
                            "right": "94.8%",
                            "top": "40.00px",
                            "bottom": "54.3%",
                            "width": "40px",
                            "height": "40px",
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_1",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "100px",
                            "right": "64.1%",
                            "top": "40px",
                            "bottom": "49.6%",
                            "width": "360px",
                            "height": "41px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.5%",
                            "right": "50.00px",
                            "top": "30.00px",
                            "bottom": "-52.8%",
                            "width": "50.00px",
                            "height": "50.00px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "65.7%",
                            "right": "140.00px",
                            "top": "40.00px",
                            "bottom": "50.3%",
                            "width": "32.00px",
                            "height": "32.00px",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.6%",
                            "right": "200.00px",
                            "top": "45.00px",
                            "bottom": "24.1%",
                            "width": "24.00px",
                            "height": "24.00px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_1",
                    },
                },
                "parent": {"type": "header", "id": "header"},
            },
            "layout_2": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "42.6%",
                            "top": "50%",
                            "bottom": "29.6%",
                            "width": "14.9%",
                            "height": "37.9%",
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "parent": "layout_2",
                    }
                },
                "parent": {"type": "footer", "id": "footer"},
            },
            "layout_3": {
                "type": "FLOW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "740px",
                            "right": "1120px",
                            "top": "85px",
                            "bottom": "145px",
                            "width": "380px",
                            "height": 80,
                        },
                        "widgetId": "widget_264",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_3",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "419px",
                            "right": "919px",
                            "top": "36px",
                            "bottom": "536px",
                            "width": "500px",
                            "height": "3150px",
                        },
                        "widgetId": "widget_266",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_3",
                    },
                },
                "parent": {"type": "pages", "id": "page_1"},
            },
            "layout_90": {
                "type": "FIXED",
                "content": {
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "6.7%",
                            "top": "11.8%",
                            "width": "254px",
                            "height": "30px",
                            "right": "82px",
                            "bottom": "75.7%",
                        },
                        "widgetId": "widget_259",
                        "isPending": False,
                        "parent": "layout_90",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": False,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "6.8%",
                            "top": "30.5%",
                            "width": "86.3%",
                            "height": "32.5%",
                            "right": "6.9%",
                            "bottom": "89px",
                        },
                        "widgetId": "widget_260",
                        "isPending": False,
                        "parent": "layout_90",
                    },
                },
                "order": ["1", "2"],
                "setting": {"lockDescendants": True},
                "parent": {"type": "dialog", "id": "dialog_1"},
            },
            "layout_91": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                                "width": "CUSTOM",
                            },
                            "lockParent": False,
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "bbox": {
                            "left": "6.00%",
                            "top": "11.43%",
                            "width": "36.00px",
                            "height": "36.00px",
                            "right": 305,
                            "bottom": "65.8960%",
                        },
                        "widgetId": "widget_261",
                        "isPending": False,
                        "parent": "layout_91",
                    },
                    "1": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": True,
                                "top": False,
                                "width": "CUSTOM",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "lockParent": False,
                        },
                        "bbox": {
                            "left": "20.00%",
                            "top": "12.00%",
                            "width": "254px",
                            "height": "30px",
                            "right": 30,
                            "bottom": "61.6541%",
                        },
                        "widgetId": "widget_262",
                        "isPending": False,
                        "parent": "layout_91",
                    },
                    "2": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "bottom": True,
                                "top": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockParent": False,
                            "lockLayout": False,
                        },
                        "bbox": {
                            "left": "20.00%",
                            "top": "33.3214%",
                            "width": "70.8333%",
                            "height": "45.7143%",
                            "right": "8.61%",
                            "bottom": "37px",
                        },
                        "widgetId": "widget_263",
                        "isPending": False,
                        "parent": "layout_91",
                    },
                },
                "order": ["0", "1", "2"],
                "setting": {"lockDescendants": True},
                "parent": {"type": "dialog", "id": "dialog_2"},
            },
            "layout_92": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 2, "width": 8, "height": "200px"},
                        "widgetId": "widget_265",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "aspectRatio": 1,
                        },
                        "parent": "layout_92",
                    }
                },
                "parent": {"type": "widget", "id": "widget_264"},
            },
            "layout_93": {
                "type": "ROW",
                "order": ["0"],
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "500px"},
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_93",
                    }
                },
                "parent": {"type": "widget", "id": "widget_266"},
            },
            "layout_94": {
                "type": "FIXED",
                "order": ["0"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_267",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_94",
                    }
                },
                "parent": {"type": "view", "id": "view_3"},
            },
            "layout_95": {
                "type": "COLUMN",
                "order": ["0", "1", "2", "3", "4", "5"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_268",
                        "parent": "layout_95",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_271",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_95",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_299",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_95",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_313",
                        "parent": "layout_95",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_314",
                        "parent": "layout_95",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_315",
                        "parent": "layout_95",
                    },
                },
                "parent": {"type": "widget", "id": "widget_267"},
            },
            "layout_96": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_269",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_96",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "300px"},
                        "widgetId": "widget_270",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_96",
                    },
                },
                "parent": {"type": "widget", "id": "widget_268"},
            },
            "layout_97": {
                "type": "ROW",
                "order": ["0", "1"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 5, "height": "391px"},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_97",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 7, "height": "275px"},
                        "widgetId": "widget_284",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_97",
                    },
                },
                "parent": {"type": "widget", "id": "widget_271"},
            },
            "layout_99": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_100": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_273",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_274",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_275",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_276",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_277",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "7": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_278",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_279",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_280",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_281",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "11": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_282",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_100",
                    },
                    "12": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_283",
                        "parent": "layout_100",
                    },
                },
                "order": ["0", "3", "4", "5", "8", "9", "10", "12", "6", "7", "11"],
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_102": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_102",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_102",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "31.48%",
                            "width": "214px",
                            "height": "57px",
                            "right": "5.58%",
                            "bottom": "15.74%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": True,
                                "right": False,
                                "bottom": False,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_102",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_103": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_108": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_108",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_108",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.9%",
                            "top": "32.9%",
                            "width": "309px",
                            "height": "91px",
                            "right": "5.6%",
                            "bottom": "15.7%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": True,
                                "right": False,
                                "bottom": False,
                                "left": True,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_108",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_109": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "540.00px"},
                        "widgetId": "widget_300",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_109",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "600px"},
                        "widgetId": "widget_301",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "parent": "layout_109",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_307",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_109",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_299"},
            },
            "layout_110": {
                "type": "COLUMN",
                "order": ["0", "1", "2", "3", "4"],
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_302",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_110",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_303",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_110",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_304",
                        "parent": "layout_110",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_305",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_110",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_306",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_110",
                    },
                },
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "parent": {"type": "widget", "id": "widget_301"},
            },
            "layout_111": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_308",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_111",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_309",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_111",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_310",
                        "parent": "layout_111",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_311",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_111",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_312",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_111",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_307"},
            },
            "layout_112": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_314"},
            },
            "layout_113": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_316",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_113",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_4"},
            },
            "layout_114": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_317",
                        "parent": "layout_114",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_320",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_114",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_340",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_114",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_354",
                        "parent": "layout_114",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_355",
                        "parent": "layout_114",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": 235},
                        "widgetId": "widget_356",
                        "parent": "layout_114",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_316"},
            },
            "layout_115": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "300px"},
                        "widgetId": "widget_318",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_115",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "300px"},
                        "widgetId": "widget_319",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_115",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_317"},
            },
            "layout_116": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "391px"},
                        "widgetId": "widget_321",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_116",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "275px"},
                        "widgetId": "widget_333",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_116",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_320"},
            },
            "layout_117": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_322",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_323",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_324",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_325",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_326",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_327",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_328",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_329",
                        "parent": "layout_117",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_330",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_331",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_332",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_117",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_118": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_120": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_120",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_120",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_120",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_121": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_122": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "540.00px"},
                        "widgetId": "widget_341",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_122",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "600px"},
                        "widgetId": "widget_342",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_122",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_348",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_122",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_340"},
            },
            "layout_123": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_343",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_123",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_344",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_123",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_345",
                        "parent": "layout_123",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_346",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_123",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_347",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_123",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_342"},
            },
            "layout_124": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_349",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_124",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_350",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_124",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_351",
                        "parent": "layout_124",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_352",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_124",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_353",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_124",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_348"},
            },
            "layout_125": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_355"},
            },
            "layout_126": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_357",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "AUTO",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_126",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_5"},
            },
            "layout_127": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_358",
                        "parent": "layout_127",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_361",
                        "isPending": False,
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"}
                        },
                        "parent": "layout_127",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_381",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_127",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_395",
                        "parent": "layout_127",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "440.00px"},
                        "widgetId": "widget_396",
                        "parent": "layout_127",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_397",
                        "parent": "layout_127",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_357"},
            },
            "layout_128": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "300px"},
                        "widgetId": "widget_359",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_128",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "300px"},
                        "widgetId": "widget_360",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_128",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_358"},
            },
            "layout_129": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "391px"},
                        "widgetId": "widget_362",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_129",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "275px"},
                        "widgetId": "widget_374",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_129",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_361"},
            },
            "layout_130": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_363",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_364",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_365",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_366",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_367",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_368",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_369",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_370",
                        "parent": "layout_130",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_371",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_372",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_373",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_130",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_131": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_133": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_133",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_133",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_133",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_134": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_135": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "620.00px"},
                        "widgetId": "widget_382",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_135",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": "600px"},
                        "widgetId": "widget_383",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_135",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "600px"},
                        "widgetId": "widget_389",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_135",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_381"},
            },
            "layout_136": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_384",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_136",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_385",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_136",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_386",
                        "parent": "layout_136",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_387",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_136",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_388",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_136",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_383"},
            },
            "layout_137": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_390",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_137",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_391",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_137",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_392",
                        "parent": "layout_137",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_393",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_137",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_394",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_137",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_389"},
            },
            "layout_138": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_396"},
            },
            "layout_140": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_140",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_140",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_140",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_143": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_143",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_143",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_143",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_144": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "740px",
                            "right": "1120px",
                            "top": "85px",
                            "bottom": "145px",
                            "width": "380px",
                            "height": 80,
                        },
                        "widgetId": "widget_412",
                        "setting": {"heightMode": "auto"},
                        "parent": "layout_144",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "419px",
                            "right": "919px",
                            "top": "36px",
                            "bottom": "536px",
                            "width": "500px",
                            "height": "3100px",
                        },
                        "widgetId": "widget_413",
                        "setting": {"heightMode": "fixed"},
                        "parent": "layout_144",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "pages", "id": "page_1"},
            },
            "layout_145": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "200px"},
                        "widgetId": "widget_265",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "aspectRatio": 1,
                        },
                        "parent": "layout_145",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_412"},
            },
            "layout_146": {
                "type": "ROW",
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {"left": 0, "width": 12, "height": "500px"},
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_146",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_413"},
            },
            "layout_147": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_414",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_147",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_3"},
            },
            "layout_148": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_415",
                        "parent": "layout_148",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_416",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_148",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_417",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_148",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_313",
                        "parent": "layout_148",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_314",
                        "parent": "layout_148",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_315",
                        "parent": "layout_148",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "setting": {"space": 0},
                "parent": {"type": "widget", "id": "widget_414"},
            },
            "layout_149": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_314"},
            },
            "layout_150": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_269",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_150",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "300px"},
                        "widgetId": "widget_270",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_150",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_415"},
            },
            "layout_151": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 5, "height": "391px"},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_151",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 5, "width": 7, "height": "275px"},
                        "widgetId": "widget_284",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_151",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_416"},
            },
            "layout_152": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_273",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_274",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_275",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_276",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_279",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_280",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_281",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_283",
                        "parent": "layout_152",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_277",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_278",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_282",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_152",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_153": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_154": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_154",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30.00%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": "AUTO"},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_154",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "35%",
                            "top": "2.00%",
                            "width": "162px",
                            "height": "150px",
                            "right": "5.5%",
                            "bottom": "6.200000000000033%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": "AUTO",
                                "right": False,
                                "bottom": True,
                                "left": True,
                                "top": False,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": True,
                        "parent": "layout_154",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_155": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_155",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_155",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "31.48%",
                            "width": "214px",
                            "height": "57px",
                            "right": "5.58%",
                            "bottom": "15.74%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": True,
                                "right": False,
                                "bottom": False,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_155",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_156": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_157": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_300",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_157",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "600px"},
                        "widgetId": "widget_418",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "parent": "layout_157",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_419",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_157",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_417"},
            },
            "layout_158": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_302",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_158",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_303",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_158",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_304",
                        "parent": "layout_158",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_305",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_158",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_306",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_158",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_418"},
            },
            "layout_159": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_308",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_159",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_309",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_159",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_310",
                        "parent": "layout_159",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_311",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_159",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_312",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_159",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_419"},
            },
            "layout_160": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_420",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_160",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_4"},
            },
            "layout_161": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_421",
                        "parent": "layout_161",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_422",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_161",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_423",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_161",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_354",
                        "parent": "layout_161",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_355",
                        "parent": "layout_161",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": 235},
                        "widgetId": "widget_356",
                        "parent": "layout_161",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "setting": {"space": 0},
                "parent": {"type": "widget", "id": "widget_420"},
            },
            "layout_162": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_355"},
            },
            "layout_163": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 7, "height": "300px"},
                        "widgetId": "widget_318",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_163",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 7, "width": 5, "height": "300px"},
                        "widgetId": "widget_319",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_163",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_421"},
            },
            "layout_164": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "391px"},
                        "widgetId": "widget_321",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_164",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 8, "height": "275px"},
                        "widgetId": "widget_333",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_164",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_422"},
            },
            "layout_165": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_322",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_323",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_324",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_325",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_326",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_327",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_328",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_329",
                        "parent": "layout_165",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_330",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_331",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_332",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_165",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_166": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_167": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_167",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "24.2%",
                            "width": "47.4%",
                            "height": "35.0%",
                            "right": "28.4%",
                            "top": "40px",
                            "bottom": "75px",
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_167",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "195px",
                            "top": "71.5%",
                            "width": "27.9%",
                            "height": "17.5%",
                            "right": "10.6%",
                            "bottom": "11.0%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_167",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_168": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_168",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_168",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_168",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_169": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_170": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 4, "height": "300px"},
                        "widgetId": "widget_341",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_170",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 4, "width": 4, "height": "600px"},
                        "widgetId": "widget_424",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_170",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 8, "width": 4, "height": "600px"},
                        "widgetId": "widget_425",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_170",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_423"},
            },
            "layout_171": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_343",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_171",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_344",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_171",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_345",
                        "parent": "layout_171",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_346",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_171",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220.00px", "height": "50px"},
                        "widgetId": "widget_347",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_171",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_424"},
            },
            "layout_172": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_349",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_172",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_350",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_172",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_351",
                        "parent": "layout_172",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_352",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_172",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_353",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_172",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_425"},
            },
            "layout_173": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_426",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_173",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_5"},
            },
            "layout_174": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "400px"},
                        "widgetId": "widget_427",
                        "parent": "layout_174",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_428",
                        "isPending": False,
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"}
                        },
                        "parent": "layout_174",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "500.00px"},
                        "widgetId": "widget_429",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_174",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_395",
                        "parent": "layout_174",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_396",
                        "parent": "layout_174",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_397",
                        "parent": "layout_174",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "setting": {"space": 0},
                "parent": {"type": "widget", "id": "widget_426"},
            },
            "layout_175": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_396"},
            },
            "layout_176": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "300px"},
                        "widgetId": "widget_359",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_176",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 6, "height": "300px"},
                        "widgetId": "widget_360",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_176",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_427"},
            },
            "layout_177": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 3, "height": "391px"},
                        "widgetId": "widget_362",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_177",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 3, "width": 9, "height": "275px"},
                        "widgetId": "widget_374",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_177",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_428"},
            },
            "layout_178": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_363",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_364",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_365",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_366",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_367",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_368",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_369",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_370",
                        "parent": "layout_178",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_371",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_372",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_373",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_178",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_179": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_180": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_180",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "22.6%",
                            "width": "71.7%",
                            "height": "35.0%",
                            "right": "5.7%",
                            "top": "36px",
                            "bottom": "80px",
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_180",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "240px",
                            "top": "71.7%",
                            "width": "27.9%",
                            "height": "18.2%",
                            "right": "5.9%",
                            "bottom": "10.1%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_180",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_181": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_181",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_181",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_181",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_182": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_183": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 6, "height": "658px"},
                        "widgetId": "widget_382",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "CUSTOM"},
                        },
                        "parent": "layout_183",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 6, "width": 3, "height": "600px"},
                        "widgetId": "widget_430",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_183",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 9, "width": 3, "height": "600px"},
                        "widgetId": "widget_431",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_183",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_429"},
            },
            "layout_184": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_384",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_184",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_385",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_184",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_386",
                        "parent": "layout_184",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_387",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_184",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_388",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_184",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_430"},
            },
            "layout_185": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_390",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_185",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_391",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_185",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_392",
                        "parent": "layout_185",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_393",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_185",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_394",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_185",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_431"},
            },
            "layout_186": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "20.00px",
                            "right": "94.8%",
                            "top": "40.00px",
                            "bottom": "54.3%",
                            "width": "40px",
                            "height": "40px",
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_186",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "70.00px",
                            "right": "64.1%",
                            "top": "40px",
                            "bottom": "49.6%",
                            "width": "360px",
                            "height": "41px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.5%",
                            "right": "20.00px",
                            "top": "30.00px",
                            "bottom": "-52.8%",
                            "width": "50.00px",
                            "height": "50.00px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "65.7%",
                            "right": "80.00px",
                            "top": "40.00px",
                            "bottom": "50.3%",
                            "width": "32.00px",
                            "height": "32.00px",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.6%",
                            "right": "130.00px",
                            "top": "45.00px",
                            "bottom": "24.1%",
                            "width": "24.00px",
                            "height": "24.00px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_186",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "header", "id": "header"},
            },
            "layout_241": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "10.00px",
                            "right": "94.8%",
                            "top": "30.00px",
                            "bottom": "54.3%",
                            "width": "40px",
                            "height": "40px",
                        },
                        "widgetId": "widget_1",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_241",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50.00px",
                            "right": "64.1%",
                            "top": "30.00px",
                            "bottom": "49.6%",
                            "width": "240.00px",
                            "height": "41px",
                        },
                        "widgetId": "widget_2",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_241",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.5%",
                            "right": "10.00px",
                            "top": "20.00px",
                            "bottom": "-52.8%",
                            "width": "50.00px",
                            "height": "50.00px",
                        },
                        "widgetId": "widget_3",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": False,
                                "bottom": True,
                            }
                        },
                        "parent": "layout_241",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "85.8%",
                            "right": "19px",
                            "top": "87px",
                            "bottom": "25.9%",
                            "width": "32px",
                            "height": "32px",
                        },
                        "widgetId": "widget_4",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_241",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "73.8%",
                            "right": "70px",
                            "top": "95px",
                            "bottom": "29.09999999999999%",
                            "width": "24px",
                            "height": "24px",
                        },
                        "widgetId": "widget_5",
                        "setting": {
                            "autoProps": {
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_241",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "header", "id": "header"},
            },
            "layout_242": {
                "type": "FLOW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "740px",
                            "right": "1120px",
                            "top": "85px",
                            "bottom": "145px",
                            "width": "380px",
                            "height": 80,
                        },
                        "widgetId": "widget_468",
                        "setting": {
                            "heightMode": "auto",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_242",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "419px",
                            "right": "919px",
                            "top": "36px",
                            "bottom": "536px",
                            "width": "500px",
                            "height": "3100px",
                        },
                        "widgetId": "widget_469",
                        "setting": {
                            "heightMode": "fixed",
                            "autoHeight": True,
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_242",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "pages", "id": "page_1"},
            },
            "layout_243": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_488",
                        "setting": {
                            "autoProps": {"height": "AUTO"},
                            "autoHeight": True,
                        },
                        "parent": "layout_243",
                    }
                },
                "order": ["1"],
                "parent": {"type": "widget", "id": "widget_468"},
            },
            "layout_244": {
                "type": "ROW",
                "content": {
                    "1": {
                        "bbox": {"left": 0, "width": 12},
                        "type": "WIDGET",
                        "widgetId": "widget_489",
                        "setting": {
                            "autoProps": {"height": "STRETCH"},
                            "autoHeight": True,
                        },
                        "parent": "layout_244",
                    }
                },
                "order": ["1"],
                "parent": {"type": "widget", "id": "widget_469"},
            },
            "layout_245": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_470",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_245",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_3"},
            },
            "layout_246": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "335px", "height": "600.00px"},
                        "widgetId": "widget_471",
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_246",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "340px", "height": 676},
                        "widgetId": "widget_472",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_246",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "1500.00px"},
                        "widgetId": "widget_473",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_246",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_313",
                        "parent": "layout_246",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_314",
                        "parent": "layout_246",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_315",
                        "parent": "layout_246",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_470"},
            },
            "layout_247": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_314"},
            },
            "layout_248": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_269",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_248",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "300px"},
                        "widgetId": "widget_270",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_248",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_471"},
            },
            "layout_249": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "391px"},
                        "widgetId": "widget_272",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_249",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "275px"},
                        "widgetId": "widget_284",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_249",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_472"},
            },
            "layout_250": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_273",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_274",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_275",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_276",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_279",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_280",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_281",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_283",
                        "parent": "layout_250",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_277",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_278",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_282",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_250",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_251": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_272"},
            },
            "layout_252": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_252",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_252",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.9%",
                            "top": "0.0%",
                            "width": "621px",
                            "height": "150px",
                            "right": "5.6%",
                            "bottom": "15.7%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": True,
                                "right": False,
                                "bottom": True,
                                "left": True,
                                "top": False,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": True,
                        "parent": "layout_252",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_253": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0.00%",
                            "top": "0.00%",
                            "width": "30.68%",
                            "height": "133px",
                            "right": "2.79%",
                            "bottom": "0.00%",
                        },
                        "widgetId": "widget_296",
                        "setting": {
                            "lockParent": True,
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": True},
                        },
                        "isPending": False,
                        "parent": "layout_253",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "9.26%",
                            "width": "465px",
                            "height": "30%",
                            "right": "5.58%",
                        },
                        "widgetId": "widget_297",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_253",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "34.87%",
                            "top": "31.48%",
                            "width": "214px",
                            "height": "57px",
                            "right": "5.58%",
                            "bottom": "15.74%",
                        },
                        "widgetId": "widget_298",
                        "setting": {
                            "lockParent": True,
                            "hCenter": False,
                            "vCenter": False,
                            "autoProps": {
                                "width": True,
                                "height": True,
                                "right": False,
                                "bottom": False,
                            },
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_253",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_254": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_284"},
            },
            "layout_255": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_300",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_255",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "600px"},
                        "widgetId": "widget_474",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                            "style": {"alignSelf": "flex-start"},
                        },
                        "parent": "layout_255",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "600px"},
                        "widgetId": "widget_475",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_255",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_473"},
            },
            "layout_256": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_302",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_256",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_303",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_256",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_304",
                        "parent": "layout_256",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_305",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_256",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_306",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_256",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_474"},
            },
            "layout_257": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_308",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_257",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_309",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_257",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_310",
                        "parent": "layout_257",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_311",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_257",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_312",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_257",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_475"},
            },
            "layout_258": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_476",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_258",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_4"},
            },
            "layout_259": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_477",
                        "parent": "layout_259",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "660.00px"},
                        "widgetId": "widget_478",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_259",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "1700.00px"},
                        "widgetId": "widget_479",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_259",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_354",
                        "parent": "layout_259",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_355",
                        "parent": "layout_259",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": 235},
                        "widgetId": "widget_356",
                        "parent": "layout_259",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_476"},
            },
            "layout_260": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_355"},
            },
            "layout_261": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_318",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_261",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "300px"},
                        "widgetId": "widget_319",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_261",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_477"},
            },
            "layout_262": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "391px"},
                        "widgetId": "widget_321",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_262",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "275px"},
                        "widgetId": "widget_333",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_262",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_478"},
            },
            "layout_263": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_322",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_323",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_324",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_325",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_326",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_327",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_328",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_329",
                        "parent": "layout_263",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_330",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_331",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_332",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_263",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_264": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_321"},
            },
            "layout_265": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_265",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_265",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_265",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_266": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_400",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_266",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_401",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_266",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_402",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_266",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_267": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_333"},
            },
            "layout_268": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_341",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_268",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "600px"},
                        "widgetId": "widget_480",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_268",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "600px"},
                        "widgetId": "widget_481",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_268",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_479"},
            },
            "layout_269": {
                "type": "COLUMN",
                "setting": {
                    "space": 10,
                    "padding": {"number": [0, 0, 10, 0], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_343",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_269",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_344",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_269",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_345",
                        "parent": "layout_269",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_346",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_269",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_347",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_269",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_480"},
            },
            "layout_270": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_349",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "isPending": False,
                        "parent": "layout_270",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_350",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_270",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_351",
                        "parent": "layout_270",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_352",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_270",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_353",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_270",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_481"},
            },
            "layout_271": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "0%",
                            "right": "0%",
                            "top": "0%",
                            "bottom": "0%",
                            "width": "100%",
                            "height": "100%",
                        },
                        "widgetId": "widget_482",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": False,
                                "top": False,
                                "bottom": False,
                                "width": "STRETCH",
                                "height": "STRETCH",
                            },
                            "hCenter": False,
                            "vCenter": False,
                        },
                        "parent": "layout_271",
                    }
                },
                "order": ["0"],
                "parent": {"type": "view", "id": "view_5"},
            },
            "layout_272": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "600.00px"},
                        "widgetId": "widget_483",
                        "parent": "layout_272",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "700.00px"},
                        "widgetId": "widget_484",
                        "isPending": False,
                        "setting": {
                            "autoProps": {"height": "CUSTOM", "width": "STRETCH"}
                        },
                        "parent": "layout_272",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "800px", "height": "1500.00px"},
                        "widgetId": "widget_485",
                        "isPending": False,
                        "setting": {"autoProps": {"height": "CUSTOM"}},
                        "parent": "layout_272",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "1820px", "height": "600.00px"},
                        "widgetId": "widget_395",
                        "parent": "layout_272",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "400px", "height": "600.00px"},
                        "widgetId": "widget_396",
                        "parent": "layout_272",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {"width": "516px", "height": "210px"},
                        "widgetId": "widget_397",
                        "parent": "layout_272",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5"],
                "parent": {"type": "widget", "id": "widget_482"},
            },
            "layout_273": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_396"},
            },
            "layout_274": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_359",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_274",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "300px"},
                        "widgetId": "widget_360",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_274",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_483"},
            },
            "layout_275": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "391px"},
                        "widgetId": "widget_362",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_275",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "275px"},
                        "widgetId": "widget_374",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "AUTO"},
                        },
                        "parent": "layout_275",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_484"},
            },
            "layout_276": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "top": 0,
                            "width": "100.0000%",
                            "height": "50.00%",
                            "left": 0,
                            "right": "0.0000%",
                            "bottom": "30.3226%",
                        },
                        "widgetId": "widget_363",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": True,
                                "left": False,
                                "top": False,
                                "bottom": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "53.3030%",
                            "width": "90%",
                            "height": "34px",
                            "right": "5.0360%",
                            "bottom": "38.2512%",
                        },
                        "widgetId": "widget_364",
                        "setting": {
                            "autoProps": {
                                "width": "CUSTOM",
                                "height": True,
                                "right": True,
                                "bottom": False,
                                "left": False,
                                "top": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                            "order": 0,
                            "lockLayout": False,
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5%",
                            "top": "62.6597%",
                            "width": "90%",
                            "height": "6.7839%",
                            "right": "5%",
                            "bottom": "30.5590%",
                        },
                        "widgetId": "widget_365",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "top": 354,
                            "width": "25.5034%",
                            "height": "32px",
                            "right": 188,
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_366",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.0360%",
                            "right": "5.0360%",
                            "top": "69.4423%",
                            "bottom": "23.5226%",
                            "width": "89.9281%",
                            "height": 28,
                        },
                        "widgetId": "widget_367",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "5": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "76.3845%",
                            "bottom": "15.8291%",
                            "width": "90.00%",
                            "height": "7.7889%",
                        },
                        "widgetId": "widget_368",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "6": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "5.00%",
                            "right": 14,
                            "top": "84.1709%",
                            "bottom": "11.0553%",
                            "width": "90.00%",
                            "height": "4.7739%",
                        },
                        "widgetId": "widget_369",
                        "setting": {"vCenter": False, "hCenter": False},
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "7": {
                        "type": "WIDGET",
                        "setting": {
                            "style": {},
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "bbox": {
                            "left": "48.9209%",
                            "right": "39.5683%",
                            "top": "88.9447%",
                            "bottom": "3.00%",
                            "width": "32.00px",
                            "height": "32px",
                        },
                        "widgetId": "widget_370",
                        "parent": "layout_276",
                    },
                    "8": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "60.4317%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "28.0576%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_371",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "9": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "71.9387%",
                            "top": "89.0494%",
                            "width": 32,
                            "height": 32,
                            "right": "16.5505%",
                            "bottom": "3.00%",
                        },
                        "widgetId": "widget_372",
                        "setting": {
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                            "order": 0,
                            "lockLayout": False,
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                    "10": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "84.2229%",
                            "right": "5.0388%",
                            "top": "87.4690%",
                            "bottom": "3.00%",
                            "width": 32,
                            "height": 32,
                        },
                        "widgetId": "widget_373",
                        "setting": {
                            "vCenter": False,
                            "hCenter": False,
                            "autoProps": {
                                "right": False,
                                "left": True,
                                "bottom": False,
                                "top": True,
                            },
                        },
                        "isPending": False,
                        "parent": "layout_276",
                    },
                },
                "order": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"],
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_277": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_362"},
            },
            "layout_278": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_278",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "18.7%",
                            "width": "47.4%",
                            "height": "35.0%",
                            "right": "33.9%",
                            "top": "33px",
                            "bottom": "33px",
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": True,
                                "height": False,
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_278",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "265px",
                            "top": "33.5%",
                            "width": "27.9%",
                            "height": "34.9%",
                            "right": "6.1%",
                            "bottom": "31.6%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                                "top": True,
                                "bottom": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_278",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_279": {
                "type": "FIXED",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "3.56%",
                            "top": "23.26%",
                            "width": "53px",
                            "height": "53px",
                            "right": "2.79%",
                            "bottom": "23.26%",
                        },
                        "widgetId": "widget_409",
                        "setting": {
                            "lockParent": True,
                            "vCenter": True,
                            "hCenter": False,
                            "autoProps": {"width": False, "height": False},
                            "heightMode": "ratio",
                            "aspectRatio": 1,
                        },
                        "isPending": False,
                        "parent": "layout_279",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "13.8029%",
                            "width": "36.5%",
                            "height": "35%",
                            "right": "38.8333%",
                            "top": 31,
                            "bottom": 24,
                        },
                        "widgetId": "widget_410",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {"width": True, "height": False},
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "isPending": False,
                        "parent": "layout_279",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": 400,
                            "top": "36.6340%",
                            "width": "27.89%",
                            "height": "34.88%",
                            "right": "7.3574%",
                            "bottom": "28.4944%",
                        },
                        "widgetId": "widget_411",
                        "setting": {
                            "lockParent": True,
                            "autoProps": {
                                "width": False,
                                "height": False,
                                "left": True,
                                "right": False,
                            },
                            "vCenter": False,
                            "hCenter": False,
                        },
                        "parent": "layout_279",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_280": {
                "type": "FIXED",
                "parent": {"type": "widget", "id": "widget_374"},
            },
            "layout_281": {
                "type": "ROW",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"left": 0, "width": 12, "height": "300px"},
                        "widgetId": "widget_382",
                        "setting": {"heightMode": "fit"},
                        "parent": "layout_281",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"left": 12, "width": 12, "height": "600px"},
                        "widgetId": "widget_486",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_281",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"left": 24, "width": 12, "height": "600px"},
                        "widgetId": "widget_487",
                        "setting": {
                            "heightMode": "fit",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_281",
                    },
                },
                "order": ["0", "1", "2"],
                "parent": {"type": "widget", "id": "widget_485"},
            },
            "layout_282": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_384",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_282",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_385",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_282",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_386",
                        "parent": "layout_282",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_387",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_282",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_388",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_282",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_486"},
            },
            "layout_283": {
                "type": "COLUMN",
                "setting": {
                    "space": 20,
                    "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                },
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_390",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_283",
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_391",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_283",
                    },
                    "2": {
                        "type": "WIDGET",
                        "bbox": {"width": "300px", "height": "160.00px"},
                        "widgetId": "widget_392",
                        "parent": "layout_283",
                    },
                    "3": {
                        "type": "WIDGET",
                        "bbox": {"width": "360px", "height": "70px"},
                        "widgetId": "widget_393",
                        "setting": {"autoProps": {"height": "AUTO"}},
                        "parent": "layout_283",
                    },
                    "4": {
                        "type": "WIDGET",
                        "bbox": {"width": "220px", "height": "50px"},
                        "widgetId": "widget_394",
                        "setting": {
                            "autoProps": {"width": "STRETCH"},
                            "style": {"alignSelf": "center"},
                        },
                        "parent": "layout_283",
                    },
                },
                "order": ["0", "1", "2", "3", "4"],
                "parent": {"type": "widget", "id": "widget_487"},
            },
            "layout_284": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "200px"},
                        "widgetId": "widget_265",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "aspectRatio": 1,
                        },
                        "parent": "layout_284",
                        "isPending": True,
                    },
                    "1": {
                        "type": "WIDGET",
                        "bbox": {"width": "auto", "height": "270.00px"},
                        "widgetId": "widget_490",
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "CUSTOM"},
                            "style": {"alignSelf": "center"},
                            "aspectRatio": 1,
                        },
                        "parent": "layout_284",
                    },
                },
                "order": ["0", "1"],
                "parent": {"type": "widget", "id": "widget_488"},
            },
            "layout_285": {
                "type": "COLUMN",
                "content": {
                    "0": {
                        "sectionId": "section_2",
                        "type": "SECTION",
                        "bbox": {"width": "auto", "height": "500px"},
                        "setting": {
                            "heightMode": "fixed",
                            "autoProps": {"height": "STRETCH"},
                        },
                        "parent": "layout_285",
                    }
                },
                "order": ["0"],
                "parent": {"type": "widget", "id": "widget_489"},
            },
            "layout_286": {
                "type": "FIXED",
                "parent": {"type": "footer", "id": "footer"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "42.6%",
                            "top": "50%",
                            "bottom": "29.6%",
                            "width": "90.00%",
                            "height": "37.9%",
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "parent": "layout_286",
                    }
                },
                "order": ["0"],
            },
            "layout_287": {
                "type": "FIXED",
                "parent": {"type": "footer", "id": "footer"},
                "content": {
                    "0": {
                        "type": "WIDGET",
                        "bbox": {
                            "left": "50%",
                            "right": "42.6%",
                            "top": "50%",
                            "bottom": "29.6%",
                            "width": "100.00%",
                            "height": "37.9%",
                        },
                        "widgetId": "widget_7",
                        "setting": {
                            "autoProps": {
                                "left": False,
                                "right": True,
                                "top": False,
                                "bottom": True,
                            },
                            "vCenter": True,
                            "hCenter": True,
                        },
                        "parent": "layout_287",
                    }
                },
                "order": ["0"],
            },
        },
        "exbVersion": "1.5.0",
        "mainSizeMode": "LARGE",
        "theme": "themes/default/",
        "forBuilderAttributes": {"lockLayout": True},
        "template": "reveal",
        "widgets": {
            "widget_1": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_1",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "0"}],
                },
            },
            "widget_2": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p><span style="font-size: 24px; color: var(--light);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_2",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "1"}],
                },
            },
            "widget_3": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {"linkType": "DIALOG", "value": "dialog_2"},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_3",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "2"}],
                },
            },
            "widget_4": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 22 28"><path d="M22 19.375c0 .562-.25 1.656-.484 2.172-.328.766-1.203 1.266-1.906 1.656-.922.5-1.859.797-2.906.797-1.453 0-2.766-.594-4.094-1.078a14.627 14.627 0 01-2.734-1.297c-2.656-1.641-5.859-4.844-7.5-7.5a14.627 14.627 0 01-1.297-2.734C.595 10.063.001 8.75.001 7.297c0-1.047.297-1.984.797-2.906.391-.703.891-1.578 1.656-1.906.516-.234 1.609-.484 2.172-.484.109 0 .219 0 .328.047.328.109.672.875.828 1.188.5.891.984 1.797 1.5 2.672.25.406.719.906.719 1.391 0 .953-2.828 2.344-2.828 3.187 0 .422.391.969.609 1.344 1.578 2.844 3.547 4.813 6.391 6.391.375.219.922.609 1.344.609.844 0 2.234-2.828 3.187-2.828.484 0 .984.469 1.391.719.875.516 1.781 1 2.672 1.5.313.156 1.078.5 1.188.828.047.109.047.219.047.328z"></path></svg>',
                            "position": "LEFT",
                        },
                        "linkParam": {"linkType": "DIALOG", "value": "dialog_1"},
                    },
                    "styleConfig": {
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "#c5c5c5",
                                    "width": "0px",
                                },
                                "text": {"color": "#181818", "size": "14px"},
                                "borderRadius": {"number": [0], "unit": "px"},
                                "iconProps": {"color": "var(--light)", "size": 28},
                            },
                            "hover": {
                                "background": {"color": "", "fillType": "fill"},
                                "border": {"type": "solid", "width": "0px"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--light)", "size": 28},
                            },
                        },
                        "useCustom": True,
                    },
                },
                "id": "widget_4",
                "useDataSources": [],
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "3"}],
                },
            },
            "widget_5": {
                "uri": "widgets/common/share/",
                "version": "1.5.0",
                "config": {
                    "uiMode": "POPUP",
                    "popup": {
                        "icon": {
                            "svg": '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16"><path d="M8 0v2H2v12h12V8h2v6a2 2 0 01-2 2H2a2 2 0 01-2-2V2a2 2 0 012-2h6zm2 0v2h2.5l-8 8L6 11.5l8-8V6h2V0h-6z" fill-rule="evenodd"></path></svg>',
                            "properties": {
                                "filename": None,
                                "originalName": None,
                                "path": None,
                                "color": "var(--light)",
                                "size": 24,
                                "inlineSvg": True,
                            },
                        },
                        "items": [
                            "embed",
                            "qrcode",
                            "email",
                            "facebook",
                            "twitter",
                            "pinterest",
                            "linkedin",
                        ],
                        "tooltip": "",
                    },
                    "inline": {
                        "items": [
                            "facebook",
                            "twitter",
                            "pinterest",
                            "linkedin",
                            "embed",
                            "qrcode",
                            "email",
                            "sharelink",
                        ],
                        "design": {
                            "direction": "HORIZONTAL",
                            "hideLabel": False,
                            "btnRad": 0,
                            "btnColor": "",
                            "iconColor": "",
                            "size": "default",
                        },
                    },
                },
                "id": "widget_5",
                "parent": {
                    "LARGE": [{"layoutId": "layout_1", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_186", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_241", "layoutItemId": "4"}],
                },
            },
            "widget_7": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="color: var(--dark-800); font-size: 16px;">Powered by Esri</span></p>',
                    "placeholder": "Double click to edit text",
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_7",
                "parent": {
                    "LARGE": [{"layoutId": "layout_2", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_286", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_287", "layoutItemId": "0"}],
                },
            },
            "widget_259": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><strong style="font-size: 14px; color: var(--dark-800);">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_259",
                "parent": {"LARGE": [{"layoutId": "layout_90", "layoutItemId": "1"}]},
            },
            "widget_260": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_260",
                "parent": {"LARGE": [{"layoutId": "layout_90", "layoutItemId": "2"}]},
            },
            "widget_261": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "svgViewBox": "0 0 28 28",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "cropShape": "circle",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_261",
                "parent": {"LARGE": [{"layoutId": "layout_91", "layoutItemId": "0"}]},
            },
            "widget_262": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><strong style="font-size: 14px; color: var(--dark-800);">Title</strong></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_262",
                "parent": {"LARGE": [{"layoutId": "layout_91", "layoutItemId": "1"}]},
            },
            "widget_263": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_263",
                "parent": {"LARGE": [{"layoutId": "layout_91", "layoutItemId": "2"}]},
            },
            "widget_264": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [60, 50, 40, 50], "unit": "px"}},
                },
                "id": "widget_264",
                "layouts": {"DEFAULT": {"LARGE": "layout_92"}},
                "parent": {"LARGE": [{"layoutId": "layout_3", "layoutItemId": "0"}]},
            },
            "widget_265": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_2",
                        "views": ["view_3", "view_4", "view_5"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "pills",
                        "standard": {
                            "gap": "40px",
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": False,
                        "advanced": True,
                        "variant": {
                            "size": {
                                "fontSize": "0.875rem",
                                "lineHeight": 1.5,
                                "paddingX": "1rem",
                                "paddingY": "0.25rem",
                            },
                            "gutter": "0.25rem",
                            "icon": {"spacing": "0.5rem"},
                            "root": {"borderRadius": "50rem", "bg": ""},
                            "item": {
                                "default": {
                                    "color": "var(--dark-800)",
                                    "border": {"width": "0px"},
                                    "borderRadius": "2px",
                                    "size": "32px",
                                },
                                "hover": {
                                    "color": "var(--dark-300)",
                                    "borderRadius": "2px",
                                    "size": "32px",
                                },
                                "active": {
                                    "color": "var(--light)",
                                    "bg": "var(--primary)",
                                    "borderRadius": "2px",
                                    "size": "32px",
                                    "border": {
                                        "type": "solid",
                                        "color": "",
                                        "width": "0px",
                                    },
                                },
                                "disabled": {
                                    "color": "#989898",
                                    "shadow": "none",
                                    "borderRadius": "50rem",
                                },
                            },
                        },
                    },
                },
                "id": "widget_265",
                "parent": {
                    "LARGE": [{"layoutId": "layout_92", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_145", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_284", "layoutItemId": "0"}],
                },
            },
            "widget_266": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 50, 40, 50], "unit": "px"}},
                },
                "id": "widget_266",
                "layouts": {"DEFAULT": {"LARGE": "layout_93"}},
                "parent": {"LARGE": [{"layoutId": "layout_3", "layoutItemId": "1"}]},
            },
            "widget_267": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_267",
                "layouts": {"DEFAULT": {"LARGE": "layout_95"}},
                "parent": {"LARGE": [{"layoutId": "layout_94", "layoutItemId": "0"}]},
            },
            "widget_268": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_268",
                "layouts": {"DEFAULT": {"LARGE": "layout_96"}},
                "parent": {"LARGE": [{"layoutId": "layout_95", "layoutItemId": "0"}]},
            },
            "widget_269": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_269",
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_96", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_150", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_248", "layoutItemId": "0"}],
                },
            },
            "widget_270": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_270",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_96", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_150", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_248", "layoutItemId": "1"}],
                },
            },
            "widget_271": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_271",
                "layouts": {"DEFAULT": {"LARGE": "layout_97"}},
                "parent": {"LARGE": [{"layoutId": "layout_95", "layoutItemId": "1"}]},
            },
            "widget_272": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0",
                                "offsetY": "0",
                                "blur": "0",
                                "spread": "0",
                                "color": "var(--dark-100)",
                            },
                        },
                    },
                    "itemStyle": "STYLE1",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "isOpenAdvabceSetting": False,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_272",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_100",
                        "MEDIUM": "layout_152",
                        "SMALL": "layout_250",
                    },
                    "HOVER": {
                        "LARGE": "layout_99",
                        "MEDIUM": "layout_153",
                        "SMALL": "layout_251",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_97", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_151", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_249", "layoutItemId": "0"}],
                },
            },
            "widget_273": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_273",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "0"}],
                },
            },
            "widget_274": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                    "tooltip": {"name": "", "parts": [{"type": "STRING", "exp": '""'}]},
                },
                "id": "widget_274",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "1"}],
                },
            },
            "widget_275": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark-800);">Created by:</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_275",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "2"}],
                },
            },
            "widget_276": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "var(--primary)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--light)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "#fff", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--primary-600)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"size": 14},
                            },
                        },
                    },
                },
                "id": "widget_276",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "3"}],
                },
            },
            "widget_277": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="26" height="28" viewBox="0 0 26 28"><path d="M17.766 15.687l4.781-4.641-6.594-0.969-2.953-5.969-2.953 5.969-6.594 0.969 4.781 4.641-1.141 6.578 5.906-3.109 5.891 3.109zM26 10.109c0 0.281-0.203 0.547-0.406 0.75l-5.672 5.531 1.344 7.812c0.016 0.109 0.016 0.203 0.016 0.313 0 0.422-0.187 0.781-0.641 0.781-0.219 0-0.438-0.078-0.625-0.187l-7.016-3.687-7.016 3.687c-0.203 0.109-0.406 0.187-0.625 0.187-0.453 0-0.656-0.375-0.656-0.781 0-0.109 0.016-0.203 0.031-0.313l1.344-7.812-5.688-5.531c-0.187-0.203-0.391-0.469-0.391-0.75 0-0.469 0.484-0.656 0.875-0.719l7.844-1.141 3.516-7.109c0.141-0.297 0.406-0.641 0.766-0.641s0.625 0.344 0.766 0.641l3.516 7.109 7.844 1.141c0.375 0.063 0.875 0.25 0.875 0.719z"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_277",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "8"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "8"}],
                },
            },
            "widget_278": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg width="16px" height="16px" viewBox="0 0 16 16"  xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"><g id="Style" stroke="none" stroke-width="1" fill="none" fill-rule="evenodd"><g id="0-UI-icons" transform="translate(-244.000000, -245.000000)" fill="var(--black)000" fill-rule="nonzero"><path d="M252,260 C255.865993,260 259,256.865993 259,253 C259,249.134007 255.865993,246 252,246 C248.134007,246 245,249.134007 245,253 C245,256.865993 248.134007,260 252,260 Z M252,261 C247.581722,261 244,257.418278 244,253 C244,248.581722 247.581722,245 252,245 C256.418278,245 260,248.581722 260,253 C260,257.418278 256.418278,261 252,261 Z M254.544453,252.999962 L250.455473,249.985987 L250.455473,256.013937 L254.544453,252.999962 Z M255.09981,253.802292 L251.01083,256.816267 C250.581429,257.132778 249.984688,257.030145 249.677973,256.58703 C249.562223,256.419806 249.5,256.219439 249.5,256.013937 L249.5,249.985987 C249.5,249.441441 249.92778,249 250.455473,249 C250.654615,249 250.848781,249.064211 251.01083,249.183656 L255.09981,252.197631 C255.529212,252.514142 255.628669,253.12994 255.321953,253.573055 C255.260669,253.661592 255.185607,253.739051 255.09981,253.802292 Z" id="app-launch"></path></g></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {"color": "", "fillType": "fill"},
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_278",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "7"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "9"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "9"}],
                },
            },
            "widget_279": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Created date: xxx</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_279",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "8"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "4"}],
                },
            },
            "widget_280": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Views number: 123</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_280",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "9"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "5"}],
                },
            },
            "widget_281": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--secondary-200)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_281",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "10"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "6"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "6"}],
                },
            },
            "widget_282": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="32" height="11" viewBox="0 0 32 11"><path d="M4 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S5.7 8.5 4 8.5z M16 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S17.7 8.5 16 8.5z M28 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S29.7 8.5 28 8.5z"/></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_282",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "11"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "10"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "10"}],
                },
            },
            "widget_283": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg "><path d= "M8 7a3 3 0 100-6 3 3 0 000 6zm0 1a4 4 0 110-8 4 4 0 010 8zm1 2H7a6.002 6.002 0 00-5.917 5h13.834A6.002 6.002 0 009 10zM7 9h2a7 7 0 017 7H0a7 7 0 017-7zm.646 3.721l1.768-1.767a.5.5 0 11.707.707l-2.475 2.475-1.767-1.768a.5.5 0 01.707-.707l1.06 1.06z " fill= "currentColor " fill-rule= "nonzero "></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_283",
                "parent": {
                    "LARGE": [{"layoutId": "layout_100", "layoutItemId": "12"}],
                    "MEDIUM": [{"layoutId": "layout_152", "layoutItemId": "7"}],
                    "SMALL": [{"layoutId": "layout_250", "layoutItemId": "7"}],
                },
            },
            "widget_284": {
                "uri": "widgets/common/list/",
                "version": "1.5.0",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE4",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "scrollBarOpen": True,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {
                                "LARGE": {"width": 1044.98, "height": 179.96},
                                "MEDIUM": {"width": 419.67, "height": 187.02},
                                "SMALL": {"width": 380, "height": 106.55},
                            },
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {
                                "LARGE": {"width": 1044.98, "height": 179.96},
                                "MEDIUM": {"width": 419.67, "height": 187.02},
                                "SMALL": {"width": 380, "height": 106.55},
                            },
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {
                                "LARGE": {"width": 1044.98, "height": 179.96},
                                "MEDIUM": {"width": 419.67, "height": 187.02},
                                "SMALL": {"width": 380, "height": 106.55},
                            },
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-500)",
                                    "width": "1px",
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                },
                "id": "widget_284",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_108",
                        "MEDIUM": "layout_154",
                        "SMALL": "layout_252",
                    },
                    "SELECTED": {
                        "LARGE": "layout_102",
                        "MEDIUM": "layout_155",
                        "SMALL": "layout_253",
                    },
                    "HOVER": {
                        "LARGE": "layout_103",
                        "MEDIUM": "layout_156",
                        "SMALL": "layout_254",
                    },
                },
                "useDataSourcesEnabled": True,
                "style": {
                    "border": {"type": "solid", "color": "", "width": "1px"},
                    "background": {
                        "color": "",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_97", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_151", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_249", "layoutItemId": "1"}],
                },
            },
            "widget_296": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_296",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_108", "layoutItemId": "0"},
                        {"layoutId": "layout_102", "layoutItemId": "0"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_154", "layoutItemId": "0"},
                        {"layoutId": "layout_155", "layoutItemId": "0"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_252", "layoutItemId": "0"},
                        {"layoutId": "layout_253", "layoutItemId": "0"},
                    ],
                },
            },
            "widget_297": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<h5><span style="font-size: 15px;color: var(--dark-800);">Double click to edit text</span></h5>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_297",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_108", "layoutItemId": "1"},
                        {"layoutId": "layout_102", "layoutItemId": "1"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_154", "layoutItemId": "1"},
                        {"layoutId": "layout_155", "layoutItemId": "1"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_252", "layoutItemId": "1"},
                        {"layoutId": "layout_253", "layoutItemId": "1"},
                    ],
                },
            },
            "widget_298": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p ><span style="font-size: 13px;color: var(--dark-800);">Lorem ipsum dolor sit amet, consectetur adipiscing elit. Aenean euismod</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_298",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_108", "layoutItemId": "2"},
                        {"layoutId": "layout_102", "layoutItemId": "2"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_154", "layoutItemId": "2"},
                        {"layoutId": "layout_155", "layoutItemId": "2"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_252", "layoutItemId": "2"},
                        {"layoutId": "layout_253", "layoutItemId": "2"},
                    ],
                },
            },
            "widget_299": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_299",
                "layouts": {"DEFAULT": {"LARGE": "layout_109"}},
                "parent": {"LARGE": [{"layoutId": "layout_95", "layoutItemId": "2"}]},
            },
            "widget_300": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_300",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_109", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_157", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_255", "layoutItemId": "0"}],
                },
            },
            "widget_301": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_301",
                "layouts": {"DEFAULT": {"LARGE": "layout_110"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_109", "layoutItemId": "1"}]},
            },
            "widget_302": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="font-size: 32px;">\ufeffTitle</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_302",
                "parent": {
                    "LARGE": [{"layoutId": "layout_110", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_158", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_256", "layoutItemId": "0"}],
                },
            },
            "widget_303": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="color: var(--primary); font-size: 24px;">\ufeffNumber</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px; color: var(--primary);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_303",
                "parent": {
                    "LARGE": [{"layoutId": "layout_110", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_158", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_256", "layoutItemId": "1"}],
                },
            },
            "widget_304": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_304",
                "parent": {
                    "LARGE": [{"layoutId": "layout_110", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_158", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_256", "layoutItemId": "2"}],
                },
            },
            "widget_305": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="font-size: 16px; color: var(--dark-900);">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo, ad quo torquatos sententiae reprimique. </span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_305",
                "parent": {
                    "LARGE": [{"layoutId": "layout_110", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_158", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_256", "layoutItemId": "3"}],
                },
            },
            "widget_306": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "default"},
                    },
                },
                "id": "widget_306",
                "parent": {
                    "LARGE": [{"layoutId": "layout_110", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_158", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_256", "layoutItemId": "4"}],
                },
            },
            "widget_307": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_307",
                "layouts": {"DEFAULT": {"LARGE": "layout_111"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_109", "layoutItemId": "2"}]},
            },
            "widget_308": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="font-size: 32px;">\ufeffTitle</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_308",
                "parent": {
                    "LARGE": [{"layoutId": "layout_111", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_159", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_257", "layoutItemId": "0"}],
                },
            },
            "widget_309": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style="text-align: center;"><span style="color: var(--primary); font-size: 24px;">\ufeffNumber</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px; color: var(--primary);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_309",
                "parent": {
                    "LARGE": [{"layoutId": "layout_111", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_159", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_257", "layoutItemId": "1"}],
                },
            },
            "widget_310": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_310",
                "parent": {
                    "LARGE": [{"layoutId": "layout_111", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_159", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_257", "layoutItemId": "2"}],
                },
            },
            "widget_311": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="font-size: 16px; color: var(--dark-900);">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo, ad quo torquatos sententiae reprimique. </span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_311",
                "parent": {
                    "LARGE": [{"layoutId": "layout_111", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_159", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_257", "layoutItemId": "3"}],
                },
            },
            "widget_312": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "default"},
                    },
                },
                "id": "widget_312",
                "parent": {
                    "LARGE": [{"layoutId": "layout_111", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_159", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_257", "layoutItemId": "4"}],
                },
            },
            "widget_313": {
                "uri": "widgets/common/table/",
                "version": "1.5.0",
                "config": {"layersConfig": [], "arrangeType": "TABS"},
                "id": "widget_313",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_95", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_246", "layoutItemId": "3"}],
                },
            },
            "widget_314": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 73957190.948944,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": 58703.63772301114,
                                "y": 2253159.4376802025,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -12464739.076516936,
                            "ymin": -3353037.964866274,
                            "xmax": 12582146.35196296,
                            "ymax": 7859356.840226679,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_314",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_112",
                        "MEDIUM": "layout_149",
                        "SMALL": "layout_247",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_95", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_246", "layoutItemId": "4"}],
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_315": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap1-1612496096511.jpg",
                                "originalName": "widget_315-snap1-1612496096511",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap1-1612496096511.jpg",
                                "type": "IMAGE",
                                "size": 534041,
                                "created": 1612496096520,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -32026230.541417267,
                                "ymin": -2350382.2791066626,
                                "xmax": 10162117.10217883,
                                "ymax": 11581947.74048537,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 147914381.897889,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -10932056.719619218,
                                    "y": 4615782.730689353,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap2-1612496117082.jpg",
                                "originalName": "widget_315-snap2-1612496117082",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap2-1612496117082.jpg",
                                "type": "IMAGE",
                                "size": 517756,
                                "created": 1612496117089,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 10302264.393539697,
                                "ymin": -976350.2586526736,
                                "xmax": 20849351.30443865,
                                "ymax": 2506732.246245312,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 15575807.848989174,
                                    "y": 765190.9937963192,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap3-1612496125303.jpg",
                                "originalName": "widget_315-snap3-1612496125303",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap3-1612496125303.jpg",
                                "type": "IMAGE",
                                "size": 420471,
                                "created": 1612496125308,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 11006918.248043481,
                                "ymin": -4489921.5832079,
                                "xmax": 21554005.158942435,
                                "ymax": -1006839.0783099146,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 16280461.703492958,
                                    "y": -2748380.3307589074,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 4,
                            "name": "Bookmark-4",
                            "title": "Bookmark-4",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap4-1612496130941.jpg",
                                "originalName": "widget_315-snap4-1612496130941",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap4-1612496130941.jpg",
                                "type": "IMAGE",
                                "size": 115777,
                                "created": 1612496130942,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 12609008.183549203,
                                "ymin": -13415137.953714337,
                                "xmax": 23156095.094448157,
                                "ymax": -9932055.448816352,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 17882551.63899868,
                                    "y": -11673596.701265344,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 5,
                            "name": "Bookmark-5",
                            "title": "Bookmark-5",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap5-1612496138211.jpg",
                                "originalName": "widget_315-snap5-1612496138211",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap5-1612496138211.jpg",
                                "type": "IMAGE",
                                "size": 64288,
                                "created": 1612496138212,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 15307521.920871284,
                                "ymin": -19538508.289638177,
                                "xmax": 25854608.831770238,
                                "ymax": -16055425.784740193,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 20581065.37632076,
                                    "y": -17796967.037189186,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 6,
                            "name": "Bookmark-6",
                            "title": "Bookmark-6",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap6-1612496145998.jpg",
                                "originalName": "widget_315-snap6-1612496145998",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap6-1612496145998.jpg",
                                "type": "IMAGE",
                                "size": 428023,
                                "created": 1612496146003,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 36283391.31498645,
                                "ymin": -3377431.897210424,
                                "xmax": 46830478.225885406,
                                "ymax": 105650.607687562,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 41556934.77043593,
                                    "y": -1635890.6447614308,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 7,
                            "name": "Bookmark-7",
                            "title": "Bookmark-7",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap7-1612496158371.jpg",
                                "originalName": "widget_315-snap7-1612496158371",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap7-1612496158371.jpg",
                                "type": "IMAGE",
                                "size": 626113,
                                "created": 1612496158378,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -8870329.77287056,
                                "ymin": -1763933.611939429,
                                "xmax": -6233558.045145821,
                                "ymax": -893162.9857149327,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7551943.90900819,
                                    "y": -1328548.2988271809,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 8,
                            "name": "Bookmark-8",
                            "title": "Bookmark-8",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap8-1612496164229.jpg",
                                "originalName": "widget_315-snap8-1612496164229",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap8-1612496164229.jpg",
                                "type": "IMAGE",
                                "size": 534610,
                                "created": 1612496164235,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -6674948.543320371,
                                "ymin": -2972099.0698932745,
                                "xmax": -4038176.8155956324,
                                "ymax": -2101328.443668778,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -5356562.679458002,
                                    "y": -2536713.7567810263,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 9,
                            "name": "Bookmark-9",
                            "title": "Bookmark-9",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap9-1612496169112.jpg",
                                "originalName": "widget_315-snap9-1612496169112",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap9-1612496169112.jpg",
                                "type": "IMAGE",
                                "size": 459797,
                                "created": 1612496169118,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -5716782.491468654,
                                "ymin": -1144639.4521959827,
                                "xmax": -3080010.7637439156,
                                "ymax": -273868.8259714863,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -4398396.627606285,
                                    "y": -709254.1390837345,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 10,
                            "name": "Bookmark-10",
                            "title": "Bookmark-10",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap10-1612496173280.jpg",
                                "originalName": "widget_315-snap10-1612496173280",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap10-1612496173280.jpg",
                                "type": "IMAGE",
                                "size": 606004,
                                "created": 1612496173287,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -7710209.16444735,
                                "ymin": -1181014.7935382156,
                                "xmax": -5073437.436722612,
                                "ymax": -310244.16731371905,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -6391823.300584981,
                                    "y": -745629.4804259673,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": True,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_315",
                "useMapWidgetIds": ["widget_314"],
                "parent": {
                    "LARGE": [{"layoutId": "layout_95", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_246", "layoutItemId": "5"}],
                },
                "style": {"border": {"type": "solid", "color": "", "width": "0px"}},
            },
            "widget_316": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_316",
                "layouts": {"DEFAULT": {"LARGE": "layout_114"}},
                "parent": {"LARGE": [{"layoutId": "layout_113", "layoutItemId": "0"}]},
            },
            "widget_317": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_317",
                "layouts": {"DEFAULT": {"LARGE": "layout_115"}},
                "parent": {"LARGE": [{"layoutId": "layout_114", "layoutItemId": "0"}]},
            },
            "widget_318": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_318",
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_115", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_163", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_261", "layoutItemId": "0"}],
                },
            },
            "widget_319": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_319",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_115", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_163", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_261", "layoutItemId": "1"}],
                },
            },
            "widget_320": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_320",
                "layouts": {"DEFAULT": {"LARGE": "layout_116"}},
                "parent": {"LARGE": [{"layoutId": "layout_114", "layoutItemId": "1"}]},
            },
            "widget_321": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0",
                                "offsetY": "0",
                                "blur": "0",
                                "spread": "0",
                                "color": "var(--dark-100)",
                            },
                        },
                    },
                    "itemStyle": "STYLE1",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "isOpenAdvabceSetting": False,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_321",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_117",
                        "MEDIUM": "layout_165",
                        "SMALL": "layout_263",
                    },
                    "HOVER": {
                        "LARGE": "layout_118",
                        "MEDIUM": "layout_166",
                        "SMALL": "layout_264",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_116", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_164", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_262", "layoutItemId": "0"}],
                },
            },
            "widget_322": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_322",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "0"}],
                },
            },
            "widget_323": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                    "tooltip": {"name": "", "parts": [{"type": "STRING", "exp": '""'}]},
                },
                "id": "widget_323",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "1"}],
                },
            },
            "widget_324": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark-800);">Created by:</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_324",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "2"}],
                },
            },
            "widget_325": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "var(--primary)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-400)",
                                    "width": "1px",
                                },
                                "text": {"color": "var(--light)", "size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--primary-600)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"size": 14},
                            },
                        },
                    },
                },
                "id": "widget_325",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "3"}],
                },
            },
            "widget_326": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Created date: xxx</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_326",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "4"}],
                },
            },
            "widget_327": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Views number: 123</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_327",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "5"}],
                },
            },
            "widget_328": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--secondary-200)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_328",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "6"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "6"}],
                },
            },
            "widget_329": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg "><path d= "M8 7a3 3 0 100-6 3 3 0 000 6zm0 1a4 4 0 110-8 4 4 0 010 8zm1 2H7a6.002 6.002 0 00-5.917 5h13.834A6.002 6.002 0 009 10zM7 9h2a7 7 0 017 7H0a7 7 0 017-7zm.646 3.721l1.768-1.767a.5.5 0 11.707.707l-2.475 2.475-1.767-1.768a.5.5 0 01.707-.707l1.06 1.06z " fill= "currentColor " fill-rule= "nonzero "></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_329",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "7"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "7"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "7"}],
                },
            },
            "widget_330": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="26" height="28" viewBox="0 0 26 28"><path d="M17.766 15.687l4.781-4.641-6.594-0.969-2.953-5.969-2.953 5.969-6.594 0.969 4.781 4.641-1.141 6.578 5.906-3.109 5.891 3.109zM26 10.109c0 0.281-0.203 0.547-0.406 0.75l-5.672 5.531 1.344 7.812c0.016 0.109 0.016 0.203 0.016 0.313 0 0.422-0.187 0.781-0.641 0.781-0.219 0-0.438-0.078-0.625-0.187l-7.016-3.687-7.016 3.687c-0.203 0.109-0.406 0.187-0.625 0.187-0.453 0-0.656-0.375-0.656-0.781 0-0.109 0.016-0.203 0.031-0.313l1.344-7.812-5.688-5.531c-0.187-0.203-0.391-0.469-0.391-0.75 0-0.469 0.484-0.656 0.875-0.719l7.844-1.141 3.516-7.109c0.141-0.297 0.406-0.641 0.766-0.641s0.625 0.344 0.766 0.641l3.516 7.109 7.844 1.141c0.375 0.063 0.875 0.25 0.875 0.719z"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_330",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "8"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "8"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "8"}],
                },
            },
            "widget_331": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg width="16px" height="16px" viewBox="0 0 16 16"  xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"><g id="Style" stroke="none" stroke-width="1" fill="none" fill-rule="evenodd"><g id="0-UI-icons" transform="translate(-244.000000, -245.000000)" fill="var(--black)000" fill-rule="nonzero"><path d="M252,260 C255.865993,260 259,256.865993 259,253 C259,249.134007 255.865993,246 252,246 C248.134007,246 245,249.134007 245,253 C245,256.865993 248.134007,260 252,260 Z M252,261 C247.581722,261 244,257.418278 244,253 C244,248.581722 247.581722,245 252,245 C256.418278,245 260,248.581722 260,253 C260,257.418278 256.418278,261 252,261 Z M254.544453,252.999962 L250.455473,249.985987 L250.455473,256.013937 L254.544453,252.999962 Z M255.09981,253.802292 L251.01083,256.816267 C250.581429,257.132778 249.984688,257.030145 249.677973,256.58703 C249.562223,256.419806 249.5,256.219439 249.5,256.013937 L249.5,249.985987 C249.5,249.441441 249.92778,249 250.455473,249 C250.654615,249 250.848781,249.064211 251.01083,249.183656 L255.09981,252.197631 C255.529212,252.514142 255.628669,253.12994 255.321953,253.573055 C255.260669,253.661592 255.185607,253.739051 255.09981,253.802292 Z" id="app-launch"></path></g></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {"color": "", "fillType": "fill"},
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_331",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "9"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "9"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "9"}],
                },
            },
            "widget_332": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="32" height="11" viewBox="0 0 32 11"><path d="M4 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S5.7 8.5 4 8.5z M16 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S17.7 8.5 16 8.5z M28 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S29.7 8.5 28 8.5z"/></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_332",
                "parent": {
                    "LARGE": [{"layoutId": "layout_117", "layoutItemId": "10"}],
                    "MEDIUM": [{"layoutId": "layout_165", "layoutItemId": "10"}],
                    "SMALL": [{"layoutId": "layout_263", "layoutItemId": "10"}],
                },
            },
            "widget_333": {
                "uri": "widgets/common/list/",
                "version": "1.5.0",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE5",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "scrollBarOpen": True,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {
                                "LARGE": {"width": 1200, "height": 179.21},
                                "MEDIUM": {"width": 482.5, "height": 189.63},
                                "SMALL": {"width": 403.75, "height": 81.5},
                            },
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {
                                "LARGE": {"width": 1200, "height": 179.21},
                                "MEDIUM": {"width": 482.5, "height": 189.63},
                                "SMALL": {"width": 403.75, "height": 81.5},
                            },
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {
                                "LARGE": {"width": 1200, "height": 179.21},
                                "MEDIUM": {"width": 482.5, "height": 189.63},
                                "SMALL": {"width": 403.75, "height": 81.5},
                            },
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-500)",
                                    "width": "1px",
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                },
                "id": "widget_333",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_140",
                        "MEDIUM": "layout_167",
                        "SMALL": "layout_265",
                    },
                    "SELECTED": {
                        "LARGE": "layout_120",
                        "MEDIUM": "layout_168",
                        "SMALL": "layout_266",
                    },
                    "HOVER": {
                        "LARGE": "layout_121",
                        "MEDIUM": "layout_169",
                        "SMALL": "layout_267",
                    },
                },
                "useDataSourcesEnabled": True,
                "style": {"border": {"type": "solid", "color": "", "width": "1px"}},
                "parent": {
                    "LARGE": [{"layoutId": "layout_116", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_164", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_262", "layoutItemId": "1"}],
                },
            },
            "widget_340": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_340",
                "layouts": {"DEFAULT": {"LARGE": "layout_122"}},
                "parent": {"LARGE": [{"layoutId": "layout_114", "layoutItemId": "2"}]},
            },
            "widget_341": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_341",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_122", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_170", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_268", "layoutItemId": "0"}],
                },
            },
            "widget_342": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_342",
                "layouts": {"DEFAULT": {"LARGE": "layout_123"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_122", "layoutItemId": "1"}]},
            },
            "widget_343": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_343",
                "parent": {
                    "LARGE": [{"layoutId": "layout_123", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_171", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_269", "layoutItemId": "0"}],
                },
            },
            "widget_344": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="color: var(--primary-700); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_344",
                "parent": {
                    "LARGE": [{"layoutId": "layout_123", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_171", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_269", "layoutItemId": "1"}],
                },
            },
            "widget_345": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_345",
                "parent": {
                    "LARGE": [{"layoutId": "layout_123", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_171", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_269", "layoutItemId": "2"}],
                },
            },
            "widget_346": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="font-size: 16px; color: var(--dark-900);">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo, ad quo torquatos sententiae reprimique. </span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_346",
                "parent": {
                    "LARGE": [{"layoutId": "layout_123", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_171", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_269", "layoutItemId": "3"}],
                },
            },
            "widget_347": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "themeStyle": {"quickStyleType": "default"},
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                    },
                },
                "id": "widget_347",
                "parent": {
                    "LARGE": [{"layoutId": "layout_123", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_171", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_269", "layoutItemId": "4"}],
                },
            },
            "widget_348": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_348",
                "layouts": {"DEFAULT": {"LARGE": "layout_124"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_122", "layoutItemId": "2"}]},
            },
            "widget_349": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_349",
                "parent": {
                    "LARGE": [{"layoutId": "layout_124", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_172", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_270", "layoutItemId": "0"}],
                },
            },
            "widget_350": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="color: var(--primary-700); font-size: 24px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_350",
                "parent": {
                    "LARGE": [{"layoutId": "layout_124", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_172", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_270", "layoutItemId": "1"}],
                },
            },
            "widget_351": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_351",
                "parent": {
                    "LARGE": [{"layoutId": "layout_124", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_172", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_270", "layoutItemId": "2"}],
                },
            },
            "widget_352": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="font-size: 16px; color: var(--dark-900);">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo, ad quo torquatos sententiae reprimique. </span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_352",
                "parent": {
                    "LARGE": [{"layoutId": "layout_124", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_172", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_270", "layoutItemId": "3"}],
                },
            },
            "widget_353": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "default"},
                    },
                },
                "id": "widget_353",
                "parent": {
                    "LARGE": [{"layoutId": "layout_124", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_172", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_270", "layoutItemId": "4"}],
                },
            },
            "widget_354": {
                "uri": "widgets/common/table/",
                "version": "1.5.0",
                "config": {"layersConfig": [], "arrangeType": "TABS"},
                "id": "widget_354",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_114", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_259", "layoutItemId": "3"}],
                },
            },
            "widget_355": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 73957190.948944,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": 2113330.9580280073,
                                "y": 2010701.18395969,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -10410111.75621194,
                            "ymin": -3624848.037448287,
                            "xmax": 14636773.672267955,
                            "ymax": 7646250.405367667,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_355",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_125",
                        "MEDIUM": "layout_162",
                        "SMALL": "layout_260",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_114", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_259", "layoutItemId": "4"}],
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_356": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "LIST",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_356-snap1-1616384651569.jpg",
                                "originalName": "widget_356-snap1-1616384651569",
                                "url": "${appResourceUrl}/images/widget_356/widget_356-snap1-1616384651569.jpg",
                                "type": "IMAGE",
                                "size": 103015,
                                "created": 1616384651574,
                                "resourcesPrefix": "images/widget_356",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "SNAPSHOT",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -429270.3508494331,
                                "ymin": -8890747.79006928,
                                "xmax": 3327762.4634225518,
                                "ymax": -5123931.036176797,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 18489297.737236,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 1449246.0562865592,
                                    "y": -7007339.413123039,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_356editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap2-1612496117082.jpg",
                                "originalName": "widget_315-snap2-1612496117082",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap2-1612496117082.jpg",
                                "type": "IMAGE",
                                "size": 517756,
                                "created": 1612496117089,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "SNAPSHOT",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 10302264.393539697,
                                "ymin": -976350.2586526736,
                                "xmax": 20849351.30443865,
                                "ymax": 2506732.246245312,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 15575807.848989174,
                                    "y": 765190.9937963192,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "layoutName": "",
                            "layoutId": "",
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap3-1612496125303.jpg",
                                "originalName": "widget_315-snap3-1612496125303",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap3-1612496125303.jpg",
                                "type": "IMAGE",
                                "size": 420471,
                                "created": 1612496125308,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "SNAPSHOT",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 11006918.248043481,
                                "ymin": -4489921.5832079,
                                "xmax": 21554005.158942435,
                                "ymax": -1006839.0783099146,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 16280461.703492958,
                                    "y": -2748380.3307589074,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "layoutName": "",
                            "layoutId": "",
                        },
                        {
                            "id": 4,
                            "name": "Bookmark-4",
                            "title": "Bookmark-4",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap4-1612496130941.jpg",
                                "originalName": "widget_315-snap4-1612496130941",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap4-1612496130941.jpg",
                                "type": "IMAGE",
                                "size": 115777,
                                "created": 1612496130942,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "SNAPSHOT",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 12609008.183549203,
                                "ymin": -13415137.953714337,
                                "xmax": 23156095.094448157,
                                "ymax": -9932055.448816352,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 17882551.63899868,
                                    "y": -11673596.701265344,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "layoutName": "",
                            "layoutId": "",
                        },
                        {
                            "id": 5,
                            "name": "Bookmark-5",
                            "title": "Bookmark-5",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap5-1612496138211.jpg",
                                "originalName": "widget_315-snap5-1612496138211",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap5-1612496138211.jpg",
                                "type": "IMAGE",
                                "size": 64288,
                                "created": 1612496138212,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "SNAPSHOT",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 15307521.920871284,
                                "ymin": -19538508.289638177,
                                "xmax": 25854608.831770238,
                                "ymax": -16055425.784740193,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 20581065.37632076,
                                    "y": -17796967.037189186,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                            "layoutName": "",
                            "layoutId": "",
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": True,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_356",
                "useMapWidgetIds": ["widget_314"],
                "layouts": {},
                "parent": {
                    "LARGE": [{"layoutId": "layout_114", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_259", "layoutItemId": "5"}],
                },
            },
            "widget_357": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_357",
                "layouts": {"DEFAULT": {"LARGE": "layout_127"}},
                "parent": {"LARGE": [{"layoutId": "layout_126", "layoutItemId": "0"}]},
            },
            "widget_358": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [0, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_358",
                "layouts": {"DEFAULT": {"LARGE": "layout_128"}},
                "parent": {"LARGE": [{"layoutId": "layout_127", "layoutItemId": "0"}]},
            },
            "widget_359": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_359",
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_128", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_176", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_274", "layoutItemId": "0"}],
                },
            },
            "widget_360": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_360",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_128", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_176", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_274", "layoutItemId": "1"}],
                },
            },
            "widget_361": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_361",
                "layouts": {"DEFAULT": {"LARGE": "layout_129"}},
                "parent": {"LARGE": [{"layoutId": "layout_127", "layoutItemId": "1"}]},
            },
            "widget_362": {
                "uri": "widgets/common/card/",
                "version": "1.5.0",
                "config": {
                    "REGULAR": {
                        "enable": True,
                        "backgroundStyle": {
                            "border": {"type": "solid", "color": "", "width": "0px"},
                            "background": {
                                "color": "var(--light)",
                                "fillType": "fit",
                                "image": {"url": ""},
                            },
                        },
                    },
                    "HOVER": {
                        "enable": False,
                        "backgroundStyle": {
                            "border": {
                                "type": "solid",
                                "color": "var(--light-400)",
                                "width": "1px",
                            },
                            "boxShadow": {
                                "offsetX": "0",
                                "offsetY": "0",
                                "blur": "0",
                                "spread": "0",
                                "color": "var(--dark-100)",
                            },
                        },
                    },
                    "itemStyle": "STYLE1",
                    "direction": "Horizontal",
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "isOpenAdvabceSetting": False,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "oneByOneEffect": None,
                    },
                },
                "id": "widget_362",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_130",
                        "MEDIUM": "layout_178",
                        "SMALL": "layout_276",
                    },
                    "HOVER": {
                        "LARGE": "layout_131",
                        "MEDIUM": "layout_179",
                        "SMALL": "layout_277",
                    },
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_129", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_177", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_275", "layoutItemId": "0"}],
                },
            },
            "widget_363": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_363",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "0"}],
                },
            },
            "widget_364": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 20px; color: var(--dark-800);">Card title</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                    "tooltip": {"name": "", "parts": [{"type": "STRING", "exp": '""'}]},
                },
                "id": "widget_364",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "1"}],
                },
            },
            "widget_365": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 14px; color: var(--dark-800);">Created by:</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_365",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "2"}],
                },
            },
            "widget_366": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "regular": {
                                "background": {
                                    "color": "var(--primary)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-400)",
                                    "width": "1px",
                                },
                                "text": {"color": "var(--light)", "size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"color": "var(--black)", "size": 14},
                            },
                            "hover": {
                                "background": {
                                    "color": "var(--primary-600)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {
                                    "unit": "px",
                                    "number": [100, 100, 100, 100],
                                },
                                "iconProps": {"size": 14},
                            },
                        },
                    },
                },
                "id": "widget_366",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "3"}],
                },
            },
            "widget_367": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Created date: xxx</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_367",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "4"}],
                },
            },
            "widget_368": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="color: var(--dark-800);">Views number: 123</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_368",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "5"}],
                },
            },
            "widget_369": {
                "uri": "widgets/common/divider/",
                "version": "1.5.0",
                "config": {
                    "direction": "Horizontal",
                    "strokeStyle": {
                        "type": "Style0",
                        "color": "var(--secondary-200)",
                        "size": "1px",
                    },
                    "pointStart": {"pointStyle": "None", "pointSize": 2},
                    "pointEnd": {"pointStyle": "None", "pointSize": 2},
                    "themeStyle": {"quickStyleType": "None"},
                },
                "id": "widget_369",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "6"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "6"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "6"}],
                },
            },
            "widget_370": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg viewBox= "0 0 16 16 " xmlns= "http://www.w3.org/2000/svg "><path d= "M8 7a3 3 0 100-6 3 3 0 000 6zm0 1a4 4 0 110-8 4 4 0 010 8zm1 2H7a6.002 6.002 0 00-5.917 5h13.834A6.002 6.002 0 009 10zM7 9h2a7 7 0 017 7H0a7 7 0 017-7zm.646 3.721l1.768-1.767a.5.5 0 11.707.707l-2.475 2.475-1.767-1.768a.5.5 0 01.707-.707l1.06 1.06z " fill= "currentColor " fill-rule= "nonzero "></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_370",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "7"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "7"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "7"}],
                },
            },
            "widget_371": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="26" height="28" viewBox="0 0 26 28"><path d="M17.766 15.687l4.781-4.641-6.594-0.969-2.953-5.969-2.953 5.969-6.594 0.969 4.781 4.641-1.141 6.578 5.906-3.109 5.891 3.109zM26 10.109c0 0.281-0.203 0.547-0.406 0.75l-5.672 5.531 1.344 7.812c0.016 0.109 0.016 0.203 0.016 0.313 0 0.422-0.187 0.781-0.641 0.781-0.219 0-0.438-0.078-0.625-0.187l-7.016-3.687-7.016 3.687c-0.203 0.109-0.406 0.187-0.625 0.187-0.453 0-0.656-0.375-0.656-0.781 0-0.109 0.016-0.203 0.031-0.313l1.344-7.812-5.688-5.531c-0.187-0.203-0.391-0.469-0.391-0.75 0-0.469 0.484-0.656 0.875-0.719l7.844-1.141 3.516-7.109c0.141-0.297 0.406-0.641 0.766-0.641s0.625 0.344 0.766 0.641l3.516 7.109 7.844 1.141c0.375 0.063 0.875 0.25 0.875 0.719z"></path></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-300)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_371",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "8"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "8"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "8"}],
                },
            },
            "widget_372": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg width="16px" height="16px" viewBox="0 0 16 16"  xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"><g id="Style" stroke="none" stroke-width="1" fill="none" fill-rule="evenodd"><g id="0-UI-icons" transform="translate(-244.000000, -245.000000)" fill="var(--black)000" fill-rule="nonzero"><path d="M252,260 C255.865993,260 259,256.865993 259,253 C259,249.134007 255.865993,246 252,246 C248.134007,246 245,249.134007 245,253 C245,256.865993 248.134007,260 252,260 Z M252,261 C247.581722,261 244,257.418278 244,253 C244,248.581722 247.581722,245 252,245 C256.418278,245 260,248.581722 260,253 C260,257.418278 256.418278,261 252,261 Z M254.544453,252.999962 L250.455473,249.985987 L250.455473,256.013937 L254.544453,252.999962 Z M255.09981,253.802292 L251.01083,256.816267 C250.581429,257.132778 249.984688,257.030145 249.677973,256.58703 C249.562223,256.419806 249.5,256.219439 249.5,256.013937 L249.5,249.985987 C249.5,249.441441 249.92778,249 250.455473,249 C250.654615,249 250.848781,249.064211 251.01083,249.183656 L255.09981,252.197631 C255.529212,252.514142 255.628669,253.12994 255.321953,253.573055 C255.260669,253.661592 255.185607,253.739051 255.09981,253.802292 Z" id="app-launch"></path></g></g></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {"color": "", "fillType": "fill"},
                                "border": {"type": "solid", "width": {"unit": "px"}},
                                "text": {"size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 16},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": {"distance": 0, "unit": "px"},
                                },
                                "text": {"color": "var(--black)", "size": "16px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_372",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "9"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "9"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "9"}],
                },
            },
            "widget_373": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "",
                        "textExpression": None,
                        "icon": {
                            "data": '<svg xmlns="http://www.w3.org/2000/svg" width="32" height="11" viewBox="0 0 32 11"><path d="M4 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S5.7 8.5 4 8.5z M16 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S17.7 8.5 16 8.5z M28 8.5c-1.7 0-3-1.3-3-3s1.3-3 3-3s3 1.3 3 3S29.7 8.5 28 8.5z"/></svg>',
                            "position": "LEFT",
                        },
                    },
                    "styleConfig": {
                        "useCustom": True,
                        "customStyle": {
                            "hover": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {"type": "solid", "width": "undefinedpx"},
                                "text": {"size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-900)", "size": 14},
                            },
                            "regular": {
                                "background": {
                                    "color": "rgba(0,0,0,0)",
                                    "fillType": "fill",
                                },
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "0px",
                                },
                                "text": {"color": "var(--dark-800)", "size": "14px"},
                                "borderRadius": {"unit": "px", "number": [2, 2, 2, 2]},
                                "iconProps": {"color": "var(--dark-400)", "size": 16},
                            },
                        },
                    },
                },
                "id": "widget_373",
                "parent": {
                    "LARGE": [{"layoutId": "layout_130", "layoutItemId": "10"}],
                    "MEDIUM": [{"layoutId": "layout_178", "layoutItemId": "10"}],
                    "SMALL": [{"layoutId": "layout_276", "layoutItemId": "10"}],
                },
            },
            "widget_374": {
                "uri": "widgets/common/list/",
                "version": "1.5.0",
                "config": {
                    "maxItems": 100,
                    "space": 10,
                    "itemsPerPage": 10,
                    "isItemStyleConfirm": True,
                    "isInitialed": True,
                    "pageStyle": "SCROLL",
                    "itemStyle": "STYLE5",
                    "hoverType": "HOVER0",
                    "selectedStyle": "STYLE0",
                    "differentOddEven": False,
                    "scrollBarOpen": True,
                    "direction": "VERTICAL",
                    "alignType": "CENTER",
                    "scrollStep": 2,
                    "cardConfigs": {
                        "REGULAR": {
                            "cardSize": {
                                "LARGE": {"width": 1355, "height": 180.29},
                                "MEDIUM": {"width": 545.31, "height": 186.92},
                                "SMALL": {"width": 403.75, "height": 103.76},
                            },
                            "enable": True,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "HOVER": {
                            "cardSize": {
                                "LARGE": {"width": 1355, "height": 180.29},
                                "MEDIUM": {"width": 545.31, "height": 186.92},
                                "SMALL": {"width": 403.75, "height": 103.76},
                            },
                            "enable": False,
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--light-400)",
                                    "width": "1px",
                                }
                            },
                        },
                        "SELECTED": {
                            "cardSize": {
                                "LARGE": {"width": 1355, "height": 180.29},
                                "MEDIUM": {"width": 545.31, "height": 186.92},
                                "SMALL": {"width": 403.75, "height": 103.76},
                            },
                            "backgroundStyle": {
                                "border": {
                                    "type": "solid",
                                    "color": "var(--primary-500)",
                                    "width": "1px",
                                }
                            },
                            "selectionMode": "SINGLE",
                        },
                    },
                    "style": {"id": "default"},
                },
                "id": "widget_374",
                "layouts": {
                    "REGULAR": {
                        "LARGE": "layout_143",
                        "MEDIUM": "layout_180",
                        "SMALL": "layout_278",
                    },
                    "SELECTED": {
                        "LARGE": "layout_133",
                        "MEDIUM": "layout_181",
                        "SMALL": "layout_279",
                    },
                    "HOVER": {
                        "LARGE": "layout_134",
                        "MEDIUM": "layout_182",
                        "SMALL": "layout_280",
                    },
                },
                "useDataSourcesEnabled": True,
                "style": {"border": {"type": "solid", "color": "", "width": "1px"}},
                "parent": {
                    "LARGE": [{"layoutId": "layout_129", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_177", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_275", "layoutItemId": "1"}],
                },
            },
            "widget_381": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_381",
                "layouts": {"DEFAULT": {"LARGE": "layout_135"}},
                "parent": {"LARGE": [{"layoutId": "layout_127", "layoutItemId": "2"}]},
            },
            "widget_382": {
                "uri": "widgets/common/chart/",
                "version": "1.5.0",
                "config": {},
                "id": "widget_382",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_135", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_183", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_281", "layoutItemId": "0"}],
                },
            },
            "widget_383": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_383",
                "layouts": {"DEFAULT": {"LARGE": "layout_136"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_135", "layoutItemId": "1"}]},
            },
            "widget_384": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_384",
                "parent": {
                    "LARGE": [{"layoutId": "layout_136", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_184", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_282", "layoutItemId": "0"}],
                },
            },
            "widget_385": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px; color: var(--primary);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_385",
                "parent": {
                    "LARGE": [{"layoutId": "layout_136", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_184", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_282", "layoutItemId": "1"}],
                },
            },
            "widget_386": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_386",
                "parent": {
                    "LARGE": [{"layoutId": "layout_136", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_184", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_282", "layoutItemId": "2"}],
                },
            },
            "widget_387": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="color: var(--dark-900); font-size: 16px;">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo.</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_387",
                "parent": {
                    "LARGE": [{"layoutId": "layout_136", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_184", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_282", "layoutItemId": "3"}],
                },
            },
            "widget_388": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "themeStyle": {"quickStyleType": "default"},
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                    },
                },
                "id": "widget_388",
                "parent": {
                    "LARGE": [{"layoutId": "layout_136", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_184", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_282", "layoutItemId": "4"}],
                },
            },
            "widget_389": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_389",
                "layouts": {"DEFAULT": {"LARGE": "layout_137"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"LARGE": [{"layoutId": "layout_135", "layoutItemId": "2"}]},
            },
            "widget_390": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 32px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_390",
                "parent": {
                    "LARGE": [{"layoutId": "layout_137", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_185", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_283", "layoutItemId": "0"}],
                },
            },
            "widget_391": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "<p></p>",
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 24px; color: var(--primary);">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_391",
                "parent": {
                    "LARGE": [{"layoutId": "layout_137", "layoutItemId": "1"}],
                    "MEDIUM": [{"layoutId": "layout_185", "layoutItemId": "1"}],
                    "SMALL": [{"layoutId": "layout_283", "layoutItemId": "1"}],
                },
            },
            "widget_392": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {},
                    },
                    "styleConfig": {},
                },
                "id": "widget_392",
                "parent": {
                    "LARGE": [{"layoutId": "layout_137", "layoutItemId": "2"}],
                    "MEDIUM": [{"layoutId": "layout_185", "layoutItemId": "2"}],
                    "SMALL": [{"layoutId": "layout_283", "layoutItemId": "2"}],
                },
            },
            "widget_393": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": '<p style=""><span style="color: var(--dark-900); font-size: 16px;">\ufeffUt has facer indoctum adversm, vis ei epicurei tetur. Eos id amet inani.&nbsp; Etiam tamquam detraco his et, eu noluisse tractatos duo.</span></p>',
                    "placeholder": '<p style="text-align: center;"><span style="font-size: 16px;">Double click to edit text</span></p>',
                    "style": {"verticalAlign": "start", "wrap": True},
                },
                "id": "widget_393",
                "parent": {
                    "LARGE": [{"layoutId": "layout_137", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_185", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_283", "layoutItemId": "3"}],
                },
            },
            "widget_394": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {},
                    "styleConfig": {
                        "useCustom": False,
                        "customStyle": {
                            "regular": {"iconProps": {"color": "#181818", "size": 14}},
                            "hover": {"iconProps": {"size": 14}},
                        },
                        "themeStyle": {"quickStyleType": "default"},
                    },
                },
                "id": "widget_394",
                "parent": {
                    "LARGE": [{"layoutId": "layout_137", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_185", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_283", "layoutItemId": "4"}],
                },
            },
            "widget_395": {
                "uri": "widgets/common/table/",
                "version": "1.5.0",
                "config": {"layersConfig": [], "arrangeType": "TABS"},
                "id": "widget_395",
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_127", "layoutItemId": "3"}],
                    "MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "3"}],
                    "SMALL": [{"layoutId": "layout_272", "layoutItemId": "3"}],
                },
            },
            "widget_396": {
                "uri": "widgets/arcgis/arcgis-map/",
                "version": "1.5.0",
                "config": {
                    "toolConifg": {
                        "canZoom": True,
                        "canHome": True,
                        "canSearch": True,
                        "canNavigation": True,
                    },
                    "isUseCustomMapState": True,
                    "initialMapState": {
                        "viewPoint": {
                            "rotation": 0,
                            "scale": 36978595.474472,
                            "targetGeometry": {
                                "spatialReference": {"wkid": 102100},
                                "x": 273950.3093740021,
                                "y": -583265.8079253773,
                            },
                        },
                        "extent": {
                            "spatialReference": {"wkid": 102100},
                            "xmin": -5987771.047745972,
                            "ymin": -2647677.0678508687,
                            "xmax": 6535671.666493976,
                            "ymax": 1481145.4520001141,
                        },
                        "rotation": 0,
                        "viewType": "2d",
                    },
                },
                "id": "widget_396",
                "layouts": {
                    "MapFixedLayout": {
                        "LARGE": "layout_138",
                        "MEDIUM": "layout_175",
                        "SMALL": "layout_273",
                    }
                },
                "parent": {
                    "LARGE": [{"layoutId": "layout_127", "layoutItemId": "4"}],
                    "MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "4"}],
                    "SMALL": [{"layoutId": "layout_272", "layoutItemId": "4"}],
                },
                "style": {
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    }
                },
            },
            "widget_397": {
                "uri": "widgets/arcgis/bookmark/",
                "version": "1.5.0",
                "config": {
                    "templateType": "GALLERY",
                    "isTemplateConfirm": True,
                    "isInitialed": True,
                    "bookmarks": [
                        {
                            "id": 1,
                            "name": "Bookmark-1",
                            "title": "Bookmark-1",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap1-1612496096511.jpg",
                                "originalName": "widget_315-snap1-1612496096511",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap1-1612496096511.jpg",
                                "type": "IMAGE",
                                "size": 534041,
                                "created": 1612496096520,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -32026230.541417267,
                                "ymin": -2350382.2791066626,
                                "xmax": 10162117.10217883,
                                "ymax": 11581947.74048537,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 147914381.897889,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -10932056.719619218,
                                    "y": 4615782.730689353,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 2,
                            "name": "Bookmark-2",
                            "title": "Bookmark-2",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap2-1612496117082.jpg",
                                "originalName": "widget_315-snap2-1612496117082",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap2-1612496117082.jpg",
                                "type": "IMAGE",
                                "size": 517756,
                                "created": 1612496117089,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 10302264.393539697,
                                "ymin": -976350.2586526736,
                                "xmax": 20849351.30443865,
                                "ymax": 2506732.246245312,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 15575807.848989174,
                                    "y": 765190.9937963192,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 3,
                            "name": "Bookmark-3",
                            "title": "Bookmark-3",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap3-1612496125303.jpg",
                                "originalName": "widget_315-snap3-1612496125303",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap3-1612496125303.jpg",
                                "type": "IMAGE",
                                "size": 420471,
                                "created": 1612496125308,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 11006918.248043481,
                                "ymin": -4489921.5832079,
                                "xmax": 21554005.158942435,
                                "ymax": -1006839.0783099146,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 16280461.703492958,
                                    "y": -2748380.3307589074,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 4,
                            "name": "Bookmark-4",
                            "title": "Bookmark-4",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap4-1612496130941.jpg",
                                "originalName": "widget_315-snap4-1612496130941",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap4-1612496130941.jpg",
                                "type": "IMAGE",
                                "size": 115777,
                                "created": 1612496130942,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 12609008.183549203,
                                "ymin": -13415137.953714337,
                                "xmax": 23156095.094448157,
                                "ymax": -9932055.448816352,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 17882551.63899868,
                                    "y": -11673596.701265344,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 5,
                            "name": "Bookmark-5",
                            "title": "Bookmark-5",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap5-1612496138211.jpg",
                                "originalName": "widget_315-snap5-1612496138211",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap5-1612496138211.jpg",
                                "type": "IMAGE",
                                "size": 64288,
                                "created": 1612496138212,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 15307521.920871284,
                                "ymin": -19538508.289638177,
                                "xmax": 25854608.831770238,
                                "ymax": -16055425.784740193,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 20581065.37632076,
                                    "y": -17796967.037189186,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 6,
                            "name": "Bookmark-6",
                            "title": "Bookmark-6",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap6-1612496145998.jpg",
                                "originalName": "widget_315-snap6-1612496145998",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap6-1612496145998.jpg",
                                "type": "IMAGE",
                                "size": 428023,
                                "created": 1612496146003,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": 36283391.31498645,
                                "ymin": -3377431.897210424,
                                "xmax": 46830478.225885406,
                                "ymax": 105650.607687562,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 36978595.474472,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": 41556934.77043593,
                                    "y": -1635890.6447614308,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 7,
                            "name": "Bookmark-7",
                            "title": "Bookmark-7",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap7-1612496158371.jpg",
                                "originalName": "widget_315-snap7-1612496158371",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap7-1612496158371.jpg",
                                "type": "IMAGE",
                                "size": 626113,
                                "created": 1612496158378,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -8870329.77287056,
                                "ymin": -1763933.611939429,
                                "xmax": -6233558.045145821,
                                "ymax": -893162.9857149327,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -7551943.90900819,
                                    "y": -1328548.2988271809,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 8,
                            "name": "Bookmark-8",
                            "title": "Bookmark-8",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap8-1612496164229.jpg",
                                "originalName": "widget_315-snap8-1612496164229",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap8-1612496164229.jpg",
                                "type": "IMAGE",
                                "size": 534610,
                                "created": 1612496164235,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -6674948.543320371,
                                "ymin": -2972099.0698932745,
                                "xmax": -4038176.8155956324,
                                "ymax": -2101328.443668778,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -5356562.679458002,
                                    "y": -2536713.7567810263,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 9,
                            "name": "Bookmark-9",
                            "title": "Bookmark-9",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap9-1612496169112.jpg",
                                "originalName": "widget_315-snap9-1612496169112",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap9-1612496169112.jpg",
                                "type": "IMAGE",
                                "size": 459797,
                                "created": 1612496169118,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -5716782.491468654,
                                "ymin": -1144639.4521959827,
                                "xmax": -3080010.7637439156,
                                "ymax": -273868.8259714863,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -4398396.627606285,
                                    "y": -709254.1390837345,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                        {
                            "id": 10,
                            "name": "Bookmark-10",
                            "title": "Bookmark-10",
                            "type": "2d",
                            "imgParam": {},
                            "snapParam": {
                                "fileName": "widget_315-snap10-1612496173280.jpg",
                                "originalName": "widget_315-snap10-1612496173280",
                                "url": "${appResourceUrl}/images/widget_315/widget_315-snap10-1612496173280.jpg",
                                "type": "IMAGE",
                                "size": 606004,
                                "created": 1612496173287,
                                "resourcesPrefix": "images/widget_315",
                                "fileFormat": "image/png",
                            },
                            "imagePosition": "FILL",
                            "imgSourceType": "CUSTOM",
                            "extent": {
                                "spatialReference": {"wkid": 102100},
                                "xmin": -7710209.16444735,
                                "ymin": -1181014.7935382156,
                                "xmax": -5073437.436722612,
                                "ymax": -310244.16731371905,
                            },
                            "viewpoint": {
                                "rotation": 0,
                                "scale": 9244648.868618,
                                "targetGeometry": {
                                    "spatialReference": {"wkid": 102100},
                                    "x": -6391823.300584981,
                                    "y": -745629.4804259673,
                                },
                            },
                            "graphics": [],
                            "showFlag": True,
                            "mapViewId": "widget_315editor-null",
                            "mapDataSourceId": None,
                            "layersConfig": {},
                        },
                    ],
                    "initBookmark": False,
                    "runtimeAddAllow": False,
                    "autoPlayAllow": False,
                    "autoInterval": 3,
                    "autoLoopAllow": True,
                    "style": {"id": "default"},
                    "direction": "HORIZON",
                    "pageStyle": "PAGING",
                    "space": 10,
                    "scrollBarOpen": True,
                    "navigatorOpen": False,
                    "transition": "None",
                    "transitionDirection": "H",
                    "displayType": "SELECTED",
                    "itemHeight": 240,
                    "itemWidth": 240,
                    "transitionInfo": {
                        "transition": {"type": "None", "direction": "H"},
                        "effect": {
                            "type": "None",
                            "option": {"direction": "TOP", "configType": "DEFAULT"},
                        },
                        "oneByOneEffect": None,
                        "previewId": None,
                    },
                },
                "id": "widget_397",
                "useMapWidgetIds": ["widget_314"],
                "parent": {
                    "LARGE": [{"layoutId": "layout_127", "layoutItemId": "5"}],
                    "MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "5"}],
                    "SMALL": [{"layoutId": "layout_272", "layoutItemId": "5"}],
                },
            },
            "widget_400": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "cropShape": "circle",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "svgViewBox": "0 0 28 28",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_400",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_140", "layoutItemId": "0"},
                        {"layoutId": "layout_120", "layoutItemId": "0"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_167", "layoutItemId": "0"},
                        {"layoutId": "layout_168", "layoutItemId": "0"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_265", "layoutItemId": "0"},
                        {"layoutId": "layout_266", "layoutItemId": "0"},
                    ],
                },
            },
            "widget_401": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 16px;color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_401",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_140", "layoutItemId": "1"},
                        {"layoutId": "layout_120", "layoutItemId": "1"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_167", "layoutItemId": "1"},
                        {"layoutId": "layout_168", "layoutItemId": "1"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_265", "layoutItemId": "1"},
                        {"layoutId": "layout_266", "layoutItemId": "1"},
                    ],
                },
            },
            "widget_402": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "Button",
                        "toolTip": "",
                        "linkParam": {},
                    },
                    "styleConfig": {
                        "name": "Default",
                        "themeStyle": {
                            "className": "jimu-widget d-block",
                            "color": "primary",
                        },
                        "customStyle": {"regular": {}, "hover": {}, "clicked": {}},
                    },
                },
                "id": "widget_402",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_140", "layoutItemId": "2"},
                        {"layoutId": "layout_120", "layoutItemId": "2"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_167", "layoutItemId": "2"},
                        {"layoutId": "layout_168", "layoutItemId": "2"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_265", "layoutItemId": "2"},
                        {"layoutId": "layout_266", "layoutItemId": "2"},
                    ],
                },
            },
            "widget_409": {
                "uri": "widgets/common/image/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "altText": "",
                        "toolTip": "",
                        "linkParam": {},
                        "scale": "Fit",
                        "imageParam": {
                            "cropParam": {
                                "cropShape": "circle",
                                "svgPath": "M28 14c0 7.732-6.268 14-14 14S0 21.732 0 14 6.268 0 14 0s14 6.268 14 14z",
                                "svgViewBox": "0 0 28 28",
                            }
                        },
                    },
                    "styleConfig": {},
                },
                "id": "widget_409",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_143", "layoutItemId": "0"},
                        {"layoutId": "layout_133", "layoutItemId": "0"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_180", "layoutItemId": "0"},
                        {"layoutId": "layout_181", "layoutItemId": "0"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_278", "layoutItemId": "0"},
                        {"layoutId": "layout_279", "layoutItemId": "0"},
                    ],
                },
            },
            "widget_410": {
                "uri": "widgets/common/text/",
                "version": "1.5.0",
                "config": {
                    "text": "",
                    "placeholder": '<p><span style="font-size: 16px;color: var(--dark-800);">Double click to edit text</span></p>',
                    "style": {
                        "verticalAlign": "start",
                        "wrap": True,
                        "overflow": "scroll",
                        "padding": "p-1",
                    },
                },
                "id": "widget_410",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_143", "layoutItemId": "1"},
                        {"layoutId": "layout_133", "layoutItemId": "1"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_180", "layoutItemId": "1"},
                        {"layoutId": "layout_181", "layoutItemId": "1"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_278", "layoutItemId": "1"},
                        {"layoutId": "layout_279", "layoutItemId": "1"},
                    ],
                },
            },
            "widget_411": {
                "uri": "widgets/common/button/",
                "version": "1.5.0",
                "config": {
                    "functionConfig": {
                        "text": "Button",
                        "toolTip": "",
                        "linkParam": {},
                    },
                    "styleConfig": {
                        "name": "Default",
                        "themeStyle": {
                            "className": "jimu-widget d-block",
                            "color": "primary",
                        },
                        "customStyle": {"regular": {}, "hover": {}, "clicked": {}},
                    },
                },
                "id": "widget_411",
                "parent": {
                    "LARGE": [
                        {"layoutId": "layout_143", "layoutItemId": "2"},
                        {"layoutId": "layout_133", "layoutItemId": "2"},
                    ],
                    "MEDIUM": [
                        {"layoutId": "layout_180", "layoutItemId": "2"},
                        {"layoutId": "layout_181", "layoutItemId": "2"},
                    ],
                    "SMALL": [
                        {"layoutId": "layout_278", "layoutItemId": "2"},
                        {"layoutId": "layout_279", "layoutItemId": "2"},
                    ],
                },
            },
            "widget_412": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [60, 20, 40, 20], "unit": "px"}},
                },
                "id": "widget_412",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_145"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_144", "layoutItemId": "0"}]},
            },
            "widget_413": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 20, 10, 20], "unit": "px"}},
                },
                "id": "widget_413",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_146"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_144", "layoutItemId": "1"}]},
            },
            "widget_414": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_414",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_148"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_147", "layoutItemId": "0"}]},
            },
            "widget_415": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_415",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_150"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "0"}]},
            },
            "widget_416": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_416",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_151"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "1"}]},
            },
            "widget_417": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_417",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_157"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_148", "layoutItemId": "2"}]},
            },
            "widget_418": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_418",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_158"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_157", "layoutItemId": "1"}]},
            },
            "widget_419": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_419",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_159"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_157", "layoutItemId": "2"}]},
            },
            "widget_420": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_420",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_161"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_160", "layoutItemId": "0"}]},
            },
            "widget_421": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_421",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_163"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "0"}]},
            },
            "widget_422": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_422",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_164"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "1"}]},
            },
            "widget_423": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_423",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_170"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_161", "layoutItemId": "2"}]},
            },
            "widget_424": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_424",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_171"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_170", "layoutItemId": "1"}]},
            },
            "widget_425": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_425",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_172"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_170", "layoutItemId": "2"}]},
            },
            "widget_426": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 0},
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_426",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_174"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_173", "layoutItemId": "0"}]},
            },
            "widget_427": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_427",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_176"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "0"}]},
            },
            "widget_428": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_428",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_177"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "1"}]},
            },
            "widget_429": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 20,
                    "style": {"padding": {"number": [10, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_429",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_183"}},
                "parent": {"MEDIUM": [{"layoutId": "layout_174", "layoutItemId": "2"}]},
            },
            "widget_430": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_430",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_184"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_183", "layoutItemId": "1"}]},
            },
            "widget_431": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_431",
                "layouts": {"DEFAULT": {"MEDIUM": "layout_185"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"MEDIUM": [{"layoutId": "layout_183", "layoutItemId": "2"}]},
            },
            "widget_468": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [20, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_468",
                "layouts": {"DEFAULT": {"SMALL": "layout_243"}},
                "parent": {"SMALL": [{"layoutId": "layout_242", "layoutItemId": "0"}]},
            },
            "widget_469": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_469",
                "layouts": {"DEFAULT": {"SMALL": "layout_244"}},
                "parent": {"SMALL": [{"layoutId": "layout_242", "layoutItemId": "1"}]},
            },
            "widget_470": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_470",
                "layouts": {"DEFAULT": {"SMALL": "layout_246"}},
                "parent": {"SMALL": [{"layoutId": "layout_245", "layoutItemId": "0"}]},
            },
            "widget_471": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_471",
                "layouts": {"DEFAULT": {"SMALL": "layout_248"}},
                "parent": {"SMALL": [{"layoutId": "layout_246", "layoutItemId": "0"}]},
            },
            "widget_472": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 40,
                    "style": {"padding": {"number": [20, 0, 20, 0], "unit": "px"}},
                },
                "id": "widget_472",
                "layouts": {"DEFAULT": {"SMALL": "layout_249"}},
                "parent": {"SMALL": [{"layoutId": "layout_246", "layoutItemId": "1"}]},
            },
            "widget_473": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [10, 10, 10, 10], "unit": "px"}},
                },
                "id": "widget_473",
                "layouts": {"DEFAULT": {"SMALL": "layout_255"}},
                "parent": {"SMALL": [{"layoutId": "layout_246", "layoutItemId": "2"}]},
            },
            "widget_474": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_474",
                "layouts": {"DEFAULT": {"SMALL": "layout_256"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_255", "layoutItemId": "1"}]},
            },
            "widget_475": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_475",
                "layouts": {"DEFAULT": {"SMALL": "layout_257"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_255", "layoutItemId": "2"}]},
            },
            "widget_476": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_476",
                "layouts": {"DEFAULT": {"SMALL": "layout_259"}},
                "parent": {"SMALL": [{"layoutId": "layout_258", "layoutItemId": "0"}]},
            },
            "widget_477": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_477",
                "layouts": {"DEFAULT": {"SMALL": "layout_261"}},
                "parent": {"SMALL": [{"layoutId": "layout_259", "layoutItemId": "0"}]},
            },
            "widget_478": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_478",
                "layouts": {"DEFAULT": {"SMALL": "layout_262"}},
                "parent": {"SMALL": [{"layoutId": "layout_259", "layoutItemId": "1"}]},
            },
            "widget_479": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_479",
                "layouts": {"DEFAULT": {"SMALL": "layout_268"}},
                "parent": {"SMALL": [{"layoutId": "layout_259", "layoutItemId": "2"}]},
            },
            "widget_480": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 10},
                    "style": {
                        "padding": {"number": [0, 0, 10, 0], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_480",
                "layouts": {"DEFAULT": {"SMALL": "layout_269"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_268", "layoutItemId": "1"}]},
            },
            "widget_481": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_481",
                "layouts": {"DEFAULT": {"SMALL": "layout_270"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_268", "layoutItemId": "2"}]},
            },
            "widget_482": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_482",
                "layouts": {"DEFAULT": {"SMALL": "layout_272"}},
                "parent": {"SMALL": [{"layoutId": "layout_271", "layoutItemId": "0"}]},
            },
            "widget_483": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_483",
                "layouts": {"DEFAULT": {"SMALL": "layout_274"}},
                "parent": {"SMALL": [{"layoutId": "layout_272", "layoutItemId": "0"}]},
            },
            "widget_484": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_484",
                "layouts": {"DEFAULT": {"SMALL": "layout_275"}},
                "parent": {"SMALL": [{"layoutId": "layout_272", "layoutItemId": "1"}]},
            },
            "widget_485": {
                "uri": "widgets/layout/row/",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {"padding": {"number": [0, 0, 10, 0], "unit": "px"}},
                },
                "id": "widget_485",
                "layouts": {"DEFAULT": {"SMALL": "layout_281"}},
                "parent": {"SMALL": [{"layoutId": "layout_272", "layoutItemId": "2"}]},
            },
            "widget_486": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_486",
                "layouts": {"DEFAULT": {"SMALL": "layout_282"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_281", "layoutItemId": "1"}]},
            },
            "widget_487": {
                "uri": "widgets/layout/column/",
                "version": "1.5.0",
                "config": {
                    "space": {"distance": 20},
                    "style": {
                        "padding": {"number": [20, 20, 20, 20], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "id": "widget_487",
                "layouts": {"DEFAULT": {"SMALL": "layout_283"}},
                "style": {
                    "background": {
                        "color": "var(--light)",
                        "fillType": "fit",
                        "image": {"url": ""},
                    },
                    "border": {
                        "type": "solid",
                        "color": "var(--dark-100)",
                        "width": "1px",
                    },
                },
                "parent": {"SMALL": [{"layoutId": "layout_281", "layoutItemId": "2"}]},
            },
            "widget_488": {
                "uri": "widgets/layout/column/",
                "id": "widget_488",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_284"}},
                "parent": {"SMALL": [{"layoutId": "layout_243", "layoutItemId": "1"}]},
            },
            "widget_489": {
                "uri": "widgets/layout/column/",
                "id": "widget_489",
                "version": "1.5.0",
                "config": {
                    "space": 10,
                    "style": {
                        "padding": {"space": [10], "unit": "px"},
                        "justifyContent": "flex-start",
                        "alignItems": "stretch",
                    },
                },
                "layouts": {"DEFAULT": {"SMALL": "layout_285"}},
                "parent": {"SMALL": [{"layoutId": "layout_244", "layoutItemId": "1"}]},
            },
            "widget_490": {
                "uri": "widgets/common/navigator/",
                "version": "1.5.0",
                "config": {
                    "data": {
                        "type": "AUTO",
                        "section": "section_2",
                        "views": ["view_3", "view_4", "view_5"],
                    },
                    "display": {
                        "type": "nav",
                        "navStyle": "pills",
                        "standard": {
                            "gap": "40px",
                            "scrollable": True,
                            "showIcon": False,
                            "showText": True,
                            "iconPosition": "start",
                            "textAlign": "center",
                        },
                        "vertical": True,
                        "advanced": True,
                        "variant": {
                            "size": {
                                "fontSize": "0.875rem",
                                "lineHeight": 1.5,
                                "paddingX": "1rem",
                                "paddingY": "0.25rem",
                            },
                            "gutter": "0.25rem",
                            "icon": {"spacing": "0.5rem"},
                            "root": {"borderRadius": "50rem", "bg": ""},
                            "item": {
                                "default": {
                                    "color": "var(--dark-800)",
                                    "border": {"width": "0px"},
                                    "borderRadius": "2px",
                                    "size": "32px",
                                },
                                "hover": {
                                    "color": "var(--dark-300)",
                                    "borderRadius": "2px",
                                    "size": "32px",
                                },
                                "active": {
                                    "color": "var(--light)",
                                    "bg": "var(--primary)",
                                    "borderRadius": "2px",
                                    "size": "32px",
                                    "border": {
                                        "type": "solid",
                                        "color": "",
                                        "width": "0px",
                                    },
                                },
                                "disabled": {
                                    "color": "#989898",
                                    "shadow": "none",
                                    "borderRadius": "50rem",
                                },
                            },
                        },
                    },
                },
                "id": "widget_490",
                "parent": {
                    "LARGE": [{"layoutId": "layout_92", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_145", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_284", "layoutItemId": "0"}],
                },
            },
        },
        "widgetsManifest": {},
        "views": {
            "view_3": {
                "id": "view_3",
                "layout": {
                    "LARGE": "layout_94",
                    "MEDIUM": "layout_147",
                    "SMALL": "layout_245",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "rgba(240,240,240,0)",
                "backgroundIMage": {},
                "parent": "section_2",
            },
            "view_4": {
                "id": "view_4",
                "layout": {
                    "LARGE": "layout_113",
                    "MEDIUM": "layout_160",
                    "SMALL": "layout_258",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "rgba(240,240,240,0)",
                "backgroundIMage": {},
                "parent": "section_2",
            },
            "view_5": {
                "id": "view_5",
                "layout": {
                    "LARGE": "layout_126",
                    "MEDIUM": "layout_173",
                    "SMALL": "layout_271",
                },
                "backgroundPosition": "fit",
                "backgroundColor": "rgba(240,240,240,0)",
                "backgroundIMage": {},
                "parent": "section_2",
            },
        },
        "sections": {
            "section_2": {
                "id": "section_2",
                "views": ["view_3", "view_4", "view_5"],
                "parent": {
                    "LARGE": [{"layoutId": "layout_93", "layoutItemId": "0"}],
                    "MEDIUM": [{"layoutId": "layout_146", "layoutItemId": "0"}],
                    "SMALL": [{"layoutId": "layout_285", "layoutItemId": "0"}],
                },
            }
        },
        "dialogs": {
            "dialog_1": {
                "id": "dialog_1",
                "mode": "ANCHORED",
                "sizeMode": {
                    "LARGE": {"position": None, "width": "360px", "height": "240px"}
                },
                "interactionType": None,
                "interactionStyles": None,
                "confirmBeforeClose": True,
                "alwaysShowConfirmation": False,
                "dialogBackground": {
                    "color": "var(--light)",
                    "fillType": "fit",
                    "image": {"url": ""},
                },
                "overlayColor": None,
                "layout": {"LARGE": "layout_90"},
                "index": 0,
                "isSplash": False,
                "closeWhenClickOverlay": True,
                "effect": {
                    "type": "FADE_IN",
                    "option": {"direction": None, "configType": "DEFAULT"},
                },
                "oneByOneEffect": {"type": "NONE", "option": {"direction": None}},
                "boxShadow": {
                    "offsetX": "0px",
                    "offsetY": "0px",
                    "blur": "0px",
                    "spread": "0px",
                    "color": "",
                },
            },
            "dialog_2": {
                "id": "dialog_2",
                "mode": "ANCHORED",
                "sizeMode": {
                    "LARGE": {"position": None, "width": "360px", "height": "240px"}
                },
                "interactionType": None,
                "interactionStyles": None,
                "confirmBeforeClose": True,
                "alwaysShowConfirmation": False,
                "dialogBackground": {
                    "color": "var(--light)",
                    "fillType": "fit",
                    "image": {"url": ""},
                },
                "overlayColor": None,
                "layout": {"LARGE": "layout_91"},
                "index": 1,
                "isSplash": False,
                "closeWhenClickOverlay": True,
                "effect": {
                    "type": "FADE_IN",
                    "option": {"direction": None, "configType": "DEFAULT"},
                },
                "oneByOneEffect": {"type": "NONE", "option": {"direction": None}},
            },
        },
        "dataSources": {},
        "messageConfigs": {},
        "pageStructure": [{"page_1": []}],
        "header": {
            "layout": {
                "LARGE": "layout_1",
                "MEDIUM": "layout_186",
                "SMALL": "layout_241",
            },
            "height": {"LARGE": "120px", "SMALL": "180px"},
            "backgroundColor": "var(--dark-700)",
        },
        "footer": {
            "layout": {
                "LARGE": "layout_2",
                "MEDIUM": "layout_286",
                "SMALL": "layout_287",
            },
            "height": {"LARGE": "80px"},
            "backgroundColor": "var(--light)",
        },
        "timestamp": 1619661236140,
        "attributes": {"portalUrl": "https://pythonapi.playground.esri.com/portal"},
        "customTheme": {
            "sharedThemeVariables": {
                "header": {},
                "body": {},
                "button": {},
                "logo": {"small": None, "link": None},
            }
        },
    },
}
